/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.adm;

import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NameClassPair;
import org.objectweb.common.Env;
import org.objectweb.common.RemoteObject;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.container.EJBServiceImpl;
import org.objectweb.jonas.ear.EarServiceException;
import org.objectweb.jonas.ear.EarServiceImplMBean;
import org.objectweb.jonas.jtm.TransactionServiceImpl;
import org.objectweb.jonas.mail.MailServiceImplMBean;
import org.objectweb.jonas.mail.MailServiceException;
import org.objectweb.jonas.mail.factory.JMailSession;
import org.objectweb.jonas.mail.factory.JMailMimePartDS;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.registry.RegistryManager;
import org.objectweb.jonas.service.Service;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.AbsJWebContainerServiceImplMBean;
import org.objectweb.jonas.web.JWebContainerServiceException;
import org.objectweb.jonas_ejb.container.Container;


/**
 * This class implements a remote interface used for administering
 *      the server.
 * @author Philippe Coq, Philippe Durieux
 * Contributor(s): Marc Dutoo, 
 * JOnAS 2.4 Murad Meghani (Murad.Meghani@compuware.com) killServer and stopServer
 * JOnAS 2.5 2002.06 Florent Benoit & Ludovic Bert : Methods for wars and ear files
 * 
 */
public class Adm extends RemoteObject implements AdmInterface {

    public static final String ADMNAME_SUFFIX = "_Adm";

    private ServiceManager sm = null;
    private EJBServiceImpl ejbserv  = null;
    private TransactionServiceImpl tm = null;
    private JProp jonasProperties = null;
    private InitialContext ictx = null;
    private boolean isEJBContainer;

    // Server State
    public static final int NOT_READY = 0;
    public static final int READY = 1;
    public static final int STOPPED = 2;
    private int serverState = NOT_READY;

    /**
     * Reference to the web container service
     */
    private AbsJWebContainerServiceImplMBean webContainerService = null;

    /**
     * Reference to the ear service
     */
    private EarServiceImplMBean earService = null;

    /**
     * Reference to the mail service
     */
    private MailServiceImplMBean mailService = null;


    /**
     * Adm constructor
     */
    public Adm(JProp jp) throws RemoteException, NamingException, ServiceException {

	jonasProperties = jp;

	// Get Service references
	sm = ServiceManager.getInstance();

	// Get an Initial Context and
	// registers the JOnAS Server Name to JNDI
	String name = null;
        name = jonasProperties.getValue("jonas.name", "jonas") + ADMNAME_SUFFIX;
	try {
	    if (Env.isDavid()) 
		ictx = NamingManager.getInstance().getCorbaInitialContext();
	    else
		ictx = NamingManager.getInstance().getInitialContext();

	    ictx.rebind(name, this);
	} catch (NamingException e) {
	    throw e;
	}
    }

    private TransactionServiceImpl getTM() {
	if (tm == null) {
	    try {
		tm = (TransactionServiceImpl) sm.getTransactionService();
	    } catch(ServiceException e) {
	    }
	}
	return tm;
    }

    /** 
     * Get the web container service
     * We can use methods provided by the interface for MBean
     */
    private AbsJWebContainerServiceImplMBean getWebContainerService() {
        if (webContainerService == null) {
            try {
                webContainerService = (AbsJWebContainerServiceImplMBean) sm.getWebContainerService();
            } catch (ServiceException e) {
                //not started
            }
        }
        return webContainerService;
    }

    /**
     * Get the mail service
     * We can use methods provided by the interface for MBean
     */
    private MailServiceImplMBean getMailService() {
        if (mailService == null) {
            try {
                mailService = (MailServiceImplMBean) sm.getMailService();
            } catch (Exception e) {
                //not started
            }
        }
        return mailService;
    }

    /**
     * Get the ear service
     * We can use methods provided by the interface for MBean
     */
    private EarServiceImplMBean getEarService() {
        if (earService == null) {
            try {
                earService = (EarServiceImplMBean) sm.getEarService();
            } catch (Exception e) {
                //not started
            }
        }
        return earService;
    }

    // ------------------------------------------------------------------
    //  AdmInterface implementation
    // ------------------------------------------------------------------

    /**
     * get Topics. 
     * Assumes that all Loggers are TopicalLoggers.
     */
    public String [] getTopics() throws RemoteException {
	return Log.getInstance().getTopics();
    }

    /**
     * get Topic Level
     */
    public String getTopicLevel(String topic) throws RemoteException {
	return Log.getInstance().getTopicLevel(topic);
    }

    /**
     * set Topic Level
     */
    public void setTopicLevel(String topic, String l) throws RemoteException {
	Log.getInstance().setTopicLevel(topic, l);
    }
	
    /**
     * Create a container and load beans in it
     * @parameter fileName name of the ejb-jar or xml file
     */
    public void addBeans(String fileName) throws RemoteException {
	if (!isEJBContainer) return;
	ejbserv.createContainer(fileName);
    }


   /**
     * Deploy a given ear file with the help of the ear service.
     * @param fileName the name of the ear file.
     * @throw RemoteException if rmi call failed.
     * @throw EarServiceException if the deployment failed.
    */
    public void addEar(String fileName) throws RemoteException, EarServiceException {
	if (getEarService() != null) {
	    getEarService().deployEarMBean(fileName);
	} else {
	    throw new EarServiceException("Can't add the Ear file " + fileName + ". The ear service is not started");
	}
    }

    /**
     * Deploy a given war file with the help of the web container service.
     * @param fileName the name of the war file.
     * @throw RemoteException if rmi call failed.
     * @throw JWebContainerServiceException if the deployment failed.
    */
    public void addWar(String fileName) throws RemoteException, JWebContainerServiceException {
	if (getWebContainerService() != null) {
	    getWebContainerService().registerWarMBean(fileName);
	} else {
	    throw new JWebContainerServiceException("Can't add the war file " + fileName + ". The web container service is not started");
	}
    }

    /**
     * UnDeploy a given ear file with the help of the ear service.
     * @param fileName the name of the ear file.
     * @throw RemoteException if rmi call failed.
     * @throw EarServiceException if the undeployment failed.
    */
    public void removeEar(String fileName) throws RemoteException, EarServiceException {
	if (getEarService() != null) {
	    getEarService().unDeployEarMBean(fileName);
	} else {
	    throw new EarServiceException("Can't remove of the Ear file " + fileName + ". The ear service is not started");
	}
    }

    /**
     * UnDeploy a given war file with the help of the web container service.
     * @param fileName the name of the war file.
     * @throw RemoteException if rmi call failed.
     * @throw JWebContainerServiceException if the undeployment failed.
    */
    public void removeWar(String fileName) throws RemoteException, JWebContainerServiceException {
	if (getWebContainerService() != null) {
	    getWebContainerService().unRegisterWarMBean(fileName);
	} else {
	    throw new JWebContainerServiceException("Can't remove of the war file " + fileName + ". The web container service is not started");
	}
    }


    /**
     * Remove the container identified by fileName and remove all beans in it
     * @parameter fileName name of the ejb-jar or xml file 
     */
    public void removeBeans(String fileName) throws RemoteException {
	if (!isEJBContainer) return;
	ejbserv.removeContainer(fileName);
    }

    /**
     * returns true if beans are already loaded in server.
     * @parameter fileName name of the ejb-jar or xml file
     */
    public boolean isLoaded(String fileName) throws RemoteException {
	return (ejbserv.getContainer(fileName) != null);
    }
	
    /**
     * List beans of all JOnAS containers
     */
    public String [] listBeans() throws RemoteException {

	if (!isEJBContainer) return new String[0];

	Container[] lcont = ejbserv.listContainers();
	Vector lbeans = new Vector();
	for( int j = 0; j<lcont.length; j++) {
	    String[] lbn = lcont[j].listBeanNames();
	    String contname = lcont[j].getName();
	    for( int k = 0; k<lbn.length; k++){
		String s = new String(contname + ": " + lbn[k]);
		lbeans.addElement(s);
	    } 
	}
	String[] ret = new String[lbeans.size()];
	lbeans.copyInto(ret);

	return ret;
    }

    /**
     * List JNDI context
     */
    public Vector listContext() throws RemoteException {
	String name = null;
	Vector ret  = new Vector();
	try {
	    NamingEnumeration ne;

	    if (Env.isDavid()) {
		// list CosNaming and rmiregistry
		ne = NamingManager.getInstance().getCorbaInitialContext().list("");
		for (Enumeration e = ne; e.hasMoreElements(); ) {
		    name = ((NameClassPair) e.nextElement()).getName();
		ret.addElement(name);
		}
	    }

	    ne = NamingManager.getInstance().getInitialContext().list("");
	    for (Enumeration e = ne; e.hasMoreElements(); ) {
		name = ((NameClassPair) e.nextElement()).getName();
		ret.addElement(name);
	    }
	} catch (Exception ex) {
	    throw new RemoteException("Cannot list JNDI context", ex);
	}
	return ret;
    }

    /**
     * List Environment
     */
    public Properties listEnv() {
	return jonasProperties.getFilesEnv();
    }

    /**
     * Stop the Server without stopping the JVM
     */
    public void stopServer() throws RemoteException {
	try {
            serverState = STOPPED;
            // keep registry alive a little while (for jonas admin) in case of error at launch
            Thread.sleep(5000);
	    sm.stopServices();
	} catch (ServiceException e) {
	    throw new RemoteException("Cannot stop services: ", e);
	} catch (InterruptedException e) {
	}


    }
 

    /**
     * Stop  the Server and  stop the JVM
     */
    public void killServer() throws RemoteException {
	stopServer();
	System.exit(0);
    }

    /**
     * To test if the server is ready
     * @returns int 0=not ready, 1=ready, 2=stopped
     */
    public int getServerState() throws RemoteException {
	return serverState;
    }

    /**
     * To test if the server is an EJB container
     */
    public boolean isEJBContainer() throws RemoteException {
	return isEJBContainer;
    }

    /**
     * set the default value for transaction timeout
     */
    public void setTransactionTimeout(int timeout) throws RemoteException {
            getTM().setCurrentTimeOut(new Integer(timeout));
    }

    /**
     * run the garbage collector
     */
    public void runGC() throws RemoteException {
	Runtime.getRuntime().gc();
    }
    
    /**
     * sync all entity instances outside transactions
     * @param passivate passivate instances after synchronization.
     */
    public void syncAllEntities(boolean passivate) throws RemoteException {
        if (!isEJBContainer) return;
	ejbserv.syncAllEntities(passivate);
    }


    // ------------------------------------------------------------------
    //  other public methods
    // ------------------------------------------------------------------

    /**
     * server is ready
     */
    public void serverReady(boolean isEJB) {
	serverState = READY;
	isEJBContainer = isEJB;
	if (isEJBContainer) {
	    ejbserv = (EJBServiceImpl) sm.getEjbService();
	}
    }

}

 

