/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.ear.lib;

import java.io.File;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Enumeration;
import java.util.Vector;


/**
 * JOnAS Ear unused direcotry cleaner class.
 * This class provides a  way for removing directories wich are unconsistent 
 * directories for ear files.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public class EarCleaner extends Thread {

    /**
     * the instance of this VM.
     * There is only one instance of the cleaner running in the same VM.
     */
    private static EarCleaner unique = null;

    /**
     * reference to the deployerLog
     */
    private static EarDeployerLog earDeployerLog = null;

    /**
     * Url of the jonas_ROOT/apps directory
     */
    private static URL earAppsUrl = null;


    /**
     * Default constructor : Construct a new cleaner.
     * @param earApps the jonasroot apps directory
     * @param earDeployerLog the deployer logger
     */
    private EarCleaner(URL earApps, EarDeployerLog earDeployerLog) {
        super();
        EarCleaner.earAppsUrl = earApps;
        EarCleaner.earDeployerLog = earDeployerLog;
    }


    /**
     * Get an instance of the EarCleaner.
     * @param earApps the jonasroot apps directory
     * @param earDeployerLog the deployer logger
     * @return the instance of the EarCleaner.
     */
    public static EarCleaner getInstance(URL earApps, EarDeployerLog earDeployerLog) {
        if (unique == null) {
            unique = new EarCleaner(earApps, earDeployerLog);
        }

        return unique;
    }
    

    /**
     * Start the thread of this class
     * It will remove all directories
     */
    public void run() {

        for (; ;) {
            try {
                removeUnusedDirectories();
            } catch (EarCleanerException e) {
                e.printStackTrace();
            }
            try {
                this.sleep(300000);
            } catch (InterruptedException e) {

            }
        }

    }


    /**
     * Remove the unused directories and return the new Vector of ears.
     * @return the new earEntry vector.
     * @throws EarCleanerException if it failed.
     */ 
    public Vector removeUnusedDirectories() throws EarCleanerException {
 
        //Get the directory of the jonas root /ear
        File earRootDir = new File(earAppsUrl.getFile());

        //get the entries from the logger
        Vector earsLogEntries = earDeployerLog.getEntries();
 

        //Check if the files in this vector exists
        File dirEar = null;
        String fTimeStamp = null;
        EarLogEntry earLogEntry = null;
        File earLogEntryFile = null;
        String earLogEntryUnpackedDir = null;

        for (Enumeration e = earsLogEntries.elements(); e.hasMoreElements();) {
            earLogEntry = (EarLogEntry) e.nextElement();

            earLogEntryFile = earLogEntry.getFile();
            earLogEntryUnpackedDir = earLogEntry.getUnpackedDir();

            dirEar = new File(earAppsUrl.getFile() + File.separator + earLogEntryUnpackedDir);

            //The source ear file doesn't exists anymore
            if (!earLogEntry.getFile().exists()) {
                removeRecursiveDirectory(dirEar);
                try {
                    earDeployerLog.removeEntry(earLogEntry);
                } catch (EarDeployerLogException edle) {
                    throw new EarCleanerException("Can't remove an entry" + edle.getMessage());
                }
            } else {
                //the ear file is still present
                // is it the right timestamp
  
                //get the timestamp
                try {
                    fTimeStamp = EarFileManager.fileToTimeStampDir(earLogEntryFile.toURL());
                } catch (EarFileManagerException efme) {
                    throw new EarCleanerException("Can't get the timestamp of the file " + earLogEntryFile + " : " + efme.getMessage());
                } catch (MalformedURLException mue) {
                    throw new EarCleanerException("Can't get the timestamp of the file " + earLogEntryFile + " : " + mue.getMessage());
                }
                //compare
                if (!fTimeStamp.equalsIgnoreCase(earLogEntryUnpackedDir)) {
                    //time stamp has changed
                    //mark the directory as removed and delete it
                    removeRecursiveDirectory(dirEar);
                    try {
                        earDeployerLog.removeEntry(earLogEntry);
                    } catch (EarDeployerLogException edle) {
                        throw new EarCleanerException("Can't remove an entry" + edle.getMessage());
                    }
                }
                //else time stamp always the same
            }
        }
        return earsLogEntries;

    }

    /**
     * Remove a directory with all its child (recursive)
     * @param file the file or directory which must be deleted
     */
    private void removeRecursiveDirectory(File file) {

        //File or directory doesn't exists, exit.
        if (!file.exists()) {
            return;
        }
 
        //Remove the child before the current file(directory)
        if (file.isDirectory()) {
            //remove all the child
            File[] childFiles = file.listFiles();
            for (int i = 0; i < childFiles.length; i++) {
                removeRecursiveDirectory(childFiles[i]);
            }
        }
        //Since all childs are removed , remove us
        file.delete();
    }
}
