/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.ear.lib;

//import java
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * Class which permits to store or load the  association between the name of 
 * an ear and the timestamped dir associated.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public class EarDeployerLog {

    /**
     * File for logging
     */
    private File logFile;

    /**
     * the current entries of the logfile
     */
    private Vector earLogEntries = null;

    /**
     * separator char of a entry in the log file
     */
    private static final String  SEPARATOR_ENTRY = ";";


    /**
     * Constructor for the deployerLog.
     * @param logFile the file which is used for read/write entries
     * @throws EarDeployerLogException if the loadentries failed.
     */
    public EarDeployerLog(File logFile) throws EarDeployerLogException {
        this.logFile = logFile;
        earLogEntries = new Vector();
        loadEntries();
    }

    /**
     * load the entries of the log file.
     * @throws EarDeployerLogException if the load failed.
     */
    private synchronized void loadEntries() throws EarDeployerLogException {

        BufferedReader br = null;
        try {
            br = new BufferedReader(new FileReader (logFile));
        } catch (FileNotFoundException e) {
            throw new EarDeployerLogException("Can not read the " + logFile + " file");
        }
        String line = null;
 
        String fNameEar = null;
        File fEar = null;
        String sEar = null;
        StringTokenizer st = null;
 
        try {
            //Read the text file
            while ((line = br.readLine()) != null) {
  
                //parse the String
                st = new StringTokenizer(line, SEPARATOR_ENTRY);
                fNameEar = st.nextToken();
                if (fNameEar == null) {
                    throw new EarDeployerLogException("Inconsistent line in the file " + logFile);
                }
                fEar = new File(fNameEar);
  
                sEar = st.nextToken();
                if (sEar == null) {
                    throw new EarDeployerLogException("Inconsistent line in the file " + logFile);
                }
                earLogEntries.add(new EarLogEntry(fEar, sEar));
            }
            // Close the input stream
            br.close();
        } catch (IOException ioe) {
            throw new EarDeployerLogException("Error while reading the log file " +  logFile + " :" + ioe.getMessage());
        }
    }

    /**
     * Dump(save) the entries to the log file.
     * @throws EarDeployerLogException if the save failed.
     */
    private synchronized void saveEntries() throws EarDeployerLogException {

        PrintWriter pw = null;
        try {
            pw = new PrintWriter(new BufferedWriter(new FileWriter(logFile)));
        } catch (IOException e) {
            throw new EarDeployerLogException("Problem while trying to get an output stream for the " + logFile + " file");
        }

        EarLogEntry earLogEntry = null;
        String earLogEntryFile = null;
        String earLogEntryUnpackedDir = null;
        String line = null;
        for (Enumeration e = earLogEntries.elements(); e.hasMoreElements();) {
            earLogEntry = (EarLogEntry) e.nextElement();
     
            //get the infos
            earLogEntryFile = earLogEntry.getFile().getPath();
            earLogEntryUnpackedDir = earLogEntry.getUnpackedDir();

            //create the line
            line = earLogEntryFile + SEPARATOR_ENTRY + earLogEntryUnpackedDir;
 
            //dump the line
            pw.println(line);
        }
        // Close the stream
        pw.close();
    }


    /**
     * Return the entries of the file.
     * @return a vector of EarLogEntry item.
     */
    public synchronized Vector getEntries() {
        return earLogEntries;
    }

    /**
     * Remove the given entry and return the entries of the file.
     * @param entry the EarLogEntry which must be remove.
     * @return the new vector of EarLogEntry item.
     * @throws EarDeployerLogException if the remove can't be done
     */
    public synchronized Vector removeEntry(EarLogEntry entry) throws EarDeployerLogException {
        if (earLogEntries == null) {
            throw new EarDeployerLogException("Can not remove a entry, the vector is null");
        }

        if (!earLogEntries.contains(entry)) {
            throw new EarDeployerLogException("Can not remove entry " +  entry + ". There is no such entry");
        }

        //remove can be done
        earLogEntries.remove(entry);

        //write to the file.
        saveEntries();

        //return the new vector
        return earLogEntries;
    }
    

    /**
     * Add the entry and return the new entries
     * @param fileName the name of the ear file
     * @param earUnpackedDirectory the timestamped directory
     * @return the new vector of EarLogEntry item.
     * @throws EarDeployerLogException if the add can't be done
     */
    public synchronized Vector addEntry(String fileName, URL earUnpackedDirectory) throws EarDeployerLogException {
        if (earLogEntries == null) {
            throw new EarDeployerLogException("Can not add an entry, the vector is null");
        }

        File f = new File(fileName);
        String d = new File(earUnpackedDirectory.getFile()).getName();
 
        //add only if it's not already present
        EarLogEntry earLogEntry = null;
        File earLogEntryFile = null;
        String earLogEntryDir = null;
        boolean found = false;
        Enumeration e = earLogEntries.elements();

        //add only if the entry is not found
        while (e.hasMoreElements() && !found) {
            earLogEntry = (EarLogEntry) e.nextElement();

            earLogEntryFile = earLogEntry.getFile();
            earLogEntryDir = earLogEntry.getUnpackedDir();

            if (earLogEntryFile.getPath().equals(f.getPath()) && earLogEntryDir.equals(d)) {
                found = true;
            }
        }
        if (found) {
            return earLogEntries;
        }
 
        //add entry
        earLogEntry = new EarLogEntry(f, d);

        //add can be done
        earLogEntries.add(earLogEntry);

        //write to the file.
        saveEntries();

        //return the new vector
        return earLogEntries;
    }
     


}
