/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.ear.lib;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * JOnAS Ear File manager.
 * This class provides a  way for managing the EAR files.
 * @author Florent Benoit
 * @author Ludovic Bert
 */

public class EarFileManager {


    /**
     * Format of a timestamp dir.
     */
    private static final String TIMESTAMP_FORMAT = "_yyyy.MM.dd-HH.mm.ss";

    /**
     * Size of the buffer.
     */
    private static final int BUFFER_SIZE = 2048;


    /**
     * true If an unpacked directory has the same timestamp than the EAR file,
     * false otherwise.
     * @param urlFileName the url of the name of the EAR file (ends with the
     * .ear extension).
     * @param urlDirName the url of the directory where the file must be unpacked.
     * @return true If an unpacked directory has the same timestamp than the 
     * EAR file,
     * false otherwise.
     * @throws EarFileManagerException if the file doesn't exist
     */
    protected static boolean isUnpackedEar(URL urlFileName, URL urlDirName) throws EarFileManagerException {

        File urlFile = new File(urlFileName.getFile());
  
        if (!urlFile.exists()) {
            throw new EarFileManagerException("File " + urlFileName + "doesn't exist");
        }
        String timeStampDir = fileToTimeStampDir(urlFileName);

        File f = new File(urlDirName.getPath() + File.separator + timeStampDir);

        if (f.exists()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Give the unpack destination directory of the specified file.
     * @param urlFileName the url of the name of the EAR file (ends with the
     * .ear extension).
     * @return the timestamp destination directory of the specified file.
     * @throws EarFileManagerException if we can't get the timestamp.
     */
    public static String fileToTimeStampDir(URL urlFileName) throws EarFileManagerException {
 
        //check protocol
        if (!urlFileName.getProtocol().equalsIgnoreCase("file")) {
            throw new EarFileManagerException("Only the file:/ URL can be used");
        }

        File urlFile = null;
        urlFile = new File(urlFileName.getFile());

        if (!urlFile.exists()) {
            throw new EarFileManagerException("The file " + urlFileName.getFile() + " was not found.");
        }

        //Create a format
        SimpleDateFormat sdf = new SimpleDateFormat(TIMESTAMP_FORMAT);

        //get the date of the Ear file
        long lastModified = urlFile.lastModified();

        if (lastModified == 0L) {
            throw new EarFileManagerException("Can't read the last modified time for the file" + urlFile + ".");
        }

        //Date
        Date d = new Date(lastModified);

        //Return String
        String stReturn = urlFile.getName();
        //Remove extension .ear
        int lastIndex = stReturn.lastIndexOf(".");
        if (lastIndex == -1) {
            throw new EarFileManagerException("The specified file " + urlFileName.getFile() + " is not a file with the format XXX.ear.");
        }
        
        stReturn = stReturn.substring(0, lastIndex);
            
        stReturn += sdf.format(d);
 
        return stReturn;
    }

    /**
     * Unpack the given EAR file to the specified directory.
     * @param urlFileName the url of the name of the EAR file to unpack.
     * @param urlDirName the url of the destination directory where is 
     * unpacked the EAR file.
     * @return the url of the unpacked directory
     * @throws EarFileManagerException if we can't unpack the file.
     */
    public static URL unpackEar(URL urlFileName, URL urlDirName) throws EarFileManagerException {

        //check protocol
        if ((!urlFileName.getProtocol().equalsIgnoreCase("file")) || (!urlDirName.getProtocol().equalsIgnoreCase("file"))) {
            throw new EarFileManagerException("Only the file:/ URL can be used");
        }

        //Get the dirName
        String timeStampDir = fileToTimeStampDir(urlFileName);

        boolean unPacked = false;
        unPacked = isUnpackedEar(urlFileName, urlDirName);
 
        //EAR file and directory
        JarFile earFile = null;
        File parentDirectoryFile = null;
        URL parentDirectoryUrl = null;
        try {
            earFile = new JarFile(urlFileName.getFile());
            parentDirectoryFile = new File(urlDirName.getPath() + File.separator + timeStampDir);
            //Protocol file: no .toURL() method because we have a / at the end of the url otherwise
            parentDirectoryUrl = new URL("file:" + parentDirectoryFile.getPath());
        } catch (IOException e) {
            throw new EarFileManagerException("Error while creating file for reading the ear file :" + urlFileName + ": " + e.getMessage());
        }
 
        //Nothing to do if it's unpack
        if (unPacked) {
            return parentDirectoryUrl;
        }
 
        JarEntry earEntry = null;
        try {
            try {
                //get entries of the EAR file
                for (Enumeration earEntries = earFile.entries(); earEntries.hasMoreElements();) {
                    earEntry = (JarEntry) earEntries.nextElement();
     
                    //File entry
                    File earEntryFile = new File(parentDirectoryFile, earEntry.getName());

                    //Create directory
                    if (earEntry.isDirectory()) {
                        //create parent directories (with mkdirs)
                        earEntryFile.mkdirs();
                        continue;
                    }
  
                    //If it's a file, we must extract the file
                    InputStream is = null;
                    try {
                        //get the input stream
                        is = earFile.getInputStream(earEntry);

                        //Dump to the file
                        dump(is, earEntryFile);
                    } finally {
                        is.close();
                    }
                } 
            } finally {
                earFile.close();
            }
        }  catch (IOException e) {
            throw new EarFileManagerException("Error while uncompressing entry " + earEntry + ": " + e.getMessage());
        }
        return parentDirectoryUrl;
    }
    
    /**
     * Write an input stream to a given file.
     * @param in the inputStream.
     * @param earEntryFile the file where the inputstream must be dumped.
     * @throws EarFileManagerException if the dump failed.
     */
    private static void dump(InputStream in, File earEntryFile) throws EarFileManagerException {

        try {
            //File output
            FileOutputStream out = new FileOutputStream(earEntryFile);
            int n = 0;
            try {
                //buffer
                byte buffer[] = new byte[BUFFER_SIZE];
  
                while ((n = in.read(buffer)) > 0) {
                    out.write(buffer, 0, n);
                }
            } finally {
                out.close();
            }
        } catch (IOException e) {
            throw new EarFileManagerException("Error while uncompressing the file " +  earEntryFile + ": " + e.getMessage());
        }
    }


}
