/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jdbc_xa;

import java.sql.*;
import java.util.Map;

import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a wrapper class on a standard java.sql.Connection class.
 * Its goal is to prevent explicit close of a connection and transaction
 * management (should be done via XAResource)
 * @author Philippe Durieux
 * Contributor(s): 
 * oe Gittings/Dmitry Melekhov for Sybase pb.
 */
public class ConnectionImpl implements Connection {
	
    static private Logger logger = null;
	
    protected Connection actConn = null;
    protected XAConnectionImpl xac = null;
    protected boolean autocommit_set = false;
    protected boolean autocommit_unset = false;
    static private Class implclass = null;
	
    // -----------------------------------------------------------------
    // Constructors
    // -----------------------------------------------------------------

    /**
     */
    public ConnectionImpl(XAConnectionImpl xac, Connection actual) {
	this.xac = xac;
	actConn = actual;
        logger = Log.getLogger(Log.JONAS_JDBCXA_PREFIX);
    }

    

    // -----------------------------------------------------------------
    // Accessors
    // -----------------------------------------------------------------

    /**
     * Get the actual connection on database
     */
    public Connection getConnection() {
	logger.log(BasicLevel.DEBUG, "entering");
	
	return actConn;
    }
	
    // -----------------------------------------------------------------
    // Connection implementation
    // Most of the methods just forward the call to the actual connection.
    // -----------------------------------------------------------------
	
    public Statement createStatement() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");

	try {
	    return actConn.createStatement();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }



    public PreparedStatement prepareStatement(String sql) throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.prepareStatement(sql);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    public CallableStatement prepareCall(String sql) throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.prepareCall(sql);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
	
    public String nativeSQL(String sql) throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.nativeSQL(sql);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
	
    public boolean isClosed() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.isClosed();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    public DatabaseMetaData getMetaData() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.getMetaData();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    public void setReadOnly(boolean readOnly) throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
		
	try {
	    actConn.setReadOnly(readOnly);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    public boolean isReadOnly() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.isReadOnly();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
    
    public void setCatalog(String catalog) throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    actConn.setCatalog(catalog);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    public String getCatalog() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.getCatalog();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
	
    /**
     * Trigger an event to the listener.
     */
    public void close() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	xac.notifyClose();
    }
	
    public void setTransactionIsolation(int level) throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    actConn.setTransactionIsolation(level);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
	
    public int getTransactionIsolation() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.getTransactionIsolation();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
	
    public SQLWarning getWarnings() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    return actConn.getWarnings();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
	
    public void clearWarnings() throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	
	try {
	    actConn.clearWarnings();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
	
    /**
     * In a JDBC-XA driver, Connection.commit is only called if we
     * are outside a global transaction.
     */
    public void commit() throws SQLException {
	logger.log(BasicLevel.DEBUG, "local transaction");
	try {
	    actConn.commit();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
    
    /**
     * In a JDBC-XA driver, Connection.rollback is only called if we
     * are outside a global transaction.
     */
    public void rollback() throws SQLException {
	logger.log(BasicLevel.DEBUG, "local transaction");
	try {
	    actConn.rollback();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    /**
     * In a JDBC-XA driver, autocommit is false if we are in a global Tx
     */    
    public void setAutoCommit(boolean autoCommit) throws SQLException {
	logger.log(BasicLevel.DEBUG, " entering");
	try {
	    if (autoCommit == false) {
		if (autocommit_unset == false) { // cache for optimization
		    actConn.setAutoCommit(false);
		    autocommit_set = false;
		    autocommit_unset = true;
		}
	    } else {
		if (autocommit_set == false) { // cache for optimization
		    actConn.setAutoCommit(true);
		    autocommit_set = true;
		    autocommit_unset = false;
		}
	    }
	} catch (SQLException e) {
	    String s = e.getMessage().toLowerCase();
	    if (s.indexOf("set chained command not allowed") != -1) {
         	logger.log(BasicLevel.DEBUG, "failed...");
		logger.log(BasicLevel.DEBUG, "Committing then retrying");
		try {
		    // These lines for Sybase only, hoping they don't broke
		    // anything for others DBs.
		    actConn.commit();
		    actConn.setAutoCommit(autoCommit); // Shouldn't fail now.
		    logger.log(BasicLevel.DEBUG, "succeeded after retry");
		} catch (SQLException se) {
		    logger.log(BasicLevel.ERROR,"" + autoCommit + ") failed again after retry");
		    logger.log(BasicLevel.ERROR,"Exception: "+se);
		    xac.notifyError(e);
		    throw(e);
		}
	    } else {
		logger.log(BasicLevel.ERROR,"setAutoCommit(" + autoCommit + ") failed: "+e);
		xac.notifyError(e);
		throw(e);
	    }
	}
    }

    /**
     * In a JDBC-XA driver, autocommit is false if we are in a global Tx
     */    
    public boolean getAutoCommit() throws SQLException {
	try {
	    return actConn.getAutoCommit();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    public Statement createStatement(int resultSetType, int resultSetConcurrency) 
	throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
	try {
	    return actConn.createStatement(resultSetType, resultSetConcurrency);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
    
    public Map getTypeMap() throws SQLException {
        logger.log(BasicLevel.DEBUG, "entering");
        try {
	    return actConn.getTypeMap();
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    public void setTypeMap(Map map) throws SQLException {
        logger.log(BasicLevel.DEBUG, "entering");
        try {
	    setTypeMap(map);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }
    
    public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency) 
	throws SQLException {
        logger.log(BasicLevel.DEBUG, "entering");
        
	try {
	    return actConn.prepareStatement(sql, resultSetType, resultSetConcurrency);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

    public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency) 
	throws SQLException {
	logger.log(BasicLevel.DEBUG, "entering");
        
	try {
	    return actConn.prepareCall(sql, resultSetType, resultSetConcurrency);
	} catch (SQLException e) {
	    xac.notifyError(e);
	    throw(e);
	}
    }

   /*
    * Add those following methods to compile on JDK 1.4.
    * Instead those methods are defined in the java.sql.Connection interface
    * only since JDK 1.4.
    */
    public Statement createStatement(int resultSetType,
                                     int resultSetConcurrency,
                                     int resultSetHoldability) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public int getHoldability() throws SQLException {
        throw new UnsupportedOperationException();
    }
    public CallableStatement prepareCall(String sql,
                                         int resultSetType,
                                         int resultSetConcurrency,
                                         int resultSetHoldability) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public PreparedStatement prepareStatement(String sql,
                                              int autoGeneratedKeys) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public PreparedStatement prepareStatement(String sql,
                                              int resultSetType,
                                              int resultSetConcurrency,
                                              int resultSetHoldability) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public PreparedStatement prepareStatement(String sql,
                                              int[] columnIndexes) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public PreparedStatement prepareStatement(String sql,
                                              String[] columnNames) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public void releaseSavepoint(java.sql.Savepoint savepoint) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public void rollback(java.sql.Savepoint savepoint) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public void setHoldability(int holdability) throws SQLException {
        throw new UnsupportedOperationException();
    }
    public java.sql.Savepoint setSavepoint() throws SQLException {
        throw new UnsupportedOperationException();
    }
    public java.sql.Savepoint setSavepoint(String name) throws SQLException {
        throw new UnsupportedOperationException();
    }

}

