/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.jmx.vjadmin.servlet;

import java.util.Iterator;
import java.util.Set;
import java.util.Vector;
import javax.management.ObjectName;
import javax.management.MalformedObjectNameException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.context.Context;
import org.apache.velocity.servlet.VelocityServlet;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.vjadmin.JonasManagementRepr;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Abstract class inherited by all servlets managing EJBs
 * @author Philippe Durieux, Adriana Danes (JOnAS team)
 */
public abstract class EjbServlet extends JadminServlet {

    /**
     * Get info about a bean
     */
    void getBeanInfo(Context ctx, ObjectName mbean) {
	ctx.put("EJBClass", (String) JonasManagementRepr.getAttribute(mbean, "EJBClass"));
	ctx.put("DisplayName", (String) JonasManagementRepr.getAttribute(mbean, "DisplayName"));
	ctx.put("JndiName", (String) JonasManagementRepr.getAttribute(mbean, "JndiName"));
	String home = (String) JonasManagementRepr.getAttribute(mbean, "HomeClass");
	ctx.put("HomeClass", home != null ? home : "none");
	String remote = (String) JonasManagementRepr.getAttribute(mbean, "RemoteClass");
	ctx.put("RemoteClass", remote != null ? remote : "none");
	String localhome = (String) JonasManagementRepr.getAttribute(mbean, "LocalHomeClass");
	ctx.put("LocalHomeClass", localhome != null ? localhome : "none");
	String local = (String) JonasManagementRepr.getAttribute(mbean, "LocalClass");
	ctx.put("LocalClass", local != null ? local : "none");
    }

    /**
     * Get information about DataSources referenced by the bean
     */
    void getDatasourceInfo(Context ctx, ObjectName mbean) {
	String error = "";
	try {
	   // Get data sources
	    Vector dataSources = new Vector();
	    Set allDataSources = (Set)JonasManagementRepr.getAttribute(mbean, "AllDataSourceName"); 
	    ctx.put("nbDatasources", new Integer(allDataSources.size()));
	    String dataSourceName = null;
	    for (Iterator i = allDataSources.iterator(); i.hasNext() ;) {
		dataSourceName = (String)i.next();
		dataSources.add(dataSourceName);
	    }
	    ctx.put("dataSources", dataSources);
	} catch (Exception e) {
	    error = "Error while getting attribute:" + e.getMessage();
	}
	ctx.put("datasource_error", error);
    }

    /**
     * Get information about JMS Connection Factories referenced by the bean
     */
    void getConnFactoryInfo(Context ctx, ObjectName mbean) {
	String error = "";
	try {
	    // Get JMS Connection Factories
	    Vector jmsConnFactories = new Vector();
	    Set allJmsConnFactories = (Set)JonasManagementRepr.getAttribute(mbean, "AllJMSConnectionFactoryName");
	    ctx.put("nbConnFactories", new Integer(allJmsConnFactories.size()));
	    String jmsConnFactoryName = null;
	    for (Iterator i = allJmsConnFactories.iterator(); i.hasNext() ;) {
		jmsConnFactoryName = (String)i.next();
		jmsConnFactories.add(jmsConnFactoryName);
	    }
	    ctx.put("jmsConnFactories", jmsConnFactories);
	} catch (Exception e) {
	    error = "Error while getting attribute:" + e.getMessage();
	}
	ctx.put("jmsConnFact_error", error);
    }

    /**
     * Get information about JMS Destinations referenced by the bean
     */
    void getDestinationInfo(Context ctx, ObjectName mbean) {
	String error = "";
	try {
	    // Get JMS Destinations
	    Vector destinations = new Vector();
	    Set allDestinations = (Set) JonasManagementRepr.getAttribute(mbean, "AllJMSDestinationName");
	    ctx.put("nbDestinations", new Integer(allDestinations.size()));
	    String destinationName = null;
	    for (Iterator i = allDestinations.iterator(); i.hasNext() ;) {
		destinationName = (String)i.next();
		destinations.add(destinationName);
	    }
	    ctx.put("destinations", destinations);
	} catch (Exception e) {
	    error = "Error while getting attribute:" + e.getMessage();
	}
	ctx.put("jmsDest_error", error);
    }
    /**
     * Get information about Mail Factories referenced by the bean
     */
    void getMailFactoryInfo(Context ctx, ObjectName mbean) {
	String error = "";
	try {
	    // Get Session Mail Factories
	    Set sessMailFactories = (Set)JonasManagementRepr.getAttribute(mbean, "AllMailFactorySName");
	    Set mimeMailFactories = (Set)JonasManagementRepr.getAttribute(mbean, "AllMailFactoryMName");
	    int nbSessMailFactories = sessMailFactories.size();
	    int nbMimeMailFactories = mimeMailFactories.size();
	    ctx.put("nbSessMailFactories", new Integer(nbSessMailFactories));
	    ctx.put("nbMimeMailFactories", new Integer(nbMimeMailFactories));
	    if (nbSessMailFactories > 0) {
		String[] sessMailFactArray = new String[nbSessMailFactories];
		int j = 0; 
		for (Iterator i = sessMailFactories.iterator(); i.hasNext() ;) {
		    sessMailFactArray[j++] = (String)i.next();
		}
		ctx.put("sessMailFactories", sessMailFactArray);
	    }
	    if (nbMimeMailFactories > 0) {
		String[] mimeMailFactArray = new String[nbMimeMailFactories];
		int j = 0; 
		for (Iterator i = mimeMailFactories.iterator(); i.hasNext() ;) {
		    mimeMailFactArray[j++] = (String)i.next();
		}
		ctx.put("mimeMailFactories", mimeMailFactArray);
	    }
	} catch (Exception e) {
	    error = "Error while getting attribute:" + e.getMessage();
	}
	ctx.put("mailFact_error", error);
    }
}
