/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.mail;


//import java
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Iterator;
import java.io.FileNotFoundException;
//import javax
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.MalformedObjectNameException;
import javax.management.MBeanServer;
import javax.management.MBeanRegistrationException;
import javax.management.NotCompliantMBeanException;
import javax.management.InstanceNotFoundException;
import javax.management.RuntimeOperationsException;
import javax.management.ObjectName;

//import objectweb
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;

//import objectweb.util
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

//import jonas
import org.objectweb.jonas.mail.factory.JMailSession;
import org.objectweb.jonas.mail.factory.JMailMimePartDS;


/**
 * This class provides an implementation of the javaMail service.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public class MailServiceImpl extends AbsServiceImpl
    implements MailService, MailServiceImplMBean {

    /**
     * The list of the registered mail session factories.
     */
    private Vector mailSessionList = null;

    /**
     * The list of the registered mime part data source factories.
     */
    private Vector mailMimePartDSList = null;

    /**
     * Reference to a MBean server.
     */
    private MBeanServer mbeanServer = null;

   /**
    * Initial Context for Naming
    */
    private Context ictx = null;

    /**
     * Logger for this service.
     */
    private static Logger logger = null;

    /**
     * List of the factories names to load when starting the MailService
     */
    private Vector factoryNames = new Vector();

    /**
     *  List of the javax.mail.Session factories currently loaded
     */
    private Hashtable jMailSessionFactories = new Hashtable();

    /**
     *  List of the javax.mail.internet.MimePartDataSource factories currently loaded
     */
    private Hashtable jMailMimePartDSFactories = new Hashtable();

    /**
     *  List of the binded factories (jndi name -> factory name)
     *  This is needed by EJB components in order to find out the factory name 
     *  based upon their jndi name
     */
    private Hashtable bindedFactories = new Hashtable();

    //Two kind of factories
    static final int JAVAX_MAIL_SESSION_FACTORY = 1;
    static final int JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE = 2;

    /**
     * Init the Mail service.
     * @param ctx the configuration of the Ear service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(Context ctx) throws ServiceException {

        // get logger for this service
        logger = Log.getLogger(Log.JONAS_MAIL_PREFIX);

        //Get the inital Context
        try {
            ictx = new InitialContext();
        } catch (NamingException e) {
            logger.log(BasicLevel.ERROR, "Cannot create initial context during the mail service initializing");
            throw new ServiceException("Cannot create initial context during the mail service initializing", e);
        }

        // Get the JMX Server via JMX Service
        try {
            mbeanServer = 
                ((JmxService) ServiceManager.getJmxService()).getJmxServer();
        } catch (ServiceException e) {
            // the JMX service may not be started
            mbeanServer = null;
        }
        
        // Get the list of the factory names
        String factories = null;
        try {
            factories = (String) ctx.lookup("factories");
        } catch (NamingException e) {
            // No problem if there is no value for 'factories'
        }
        if (factories != null) {
            StringTokenizer st = new StringTokenizer(factories, ",");
            while (st.hasMoreTokens()) {
                factoryNames.add(st.nextToken().trim());
            }
        }
        logger.log(BasicLevel.DEBUG, "Mail service initialized");
     
    }

    /**
     * Start the Mail Service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doStart() throws ServiceException {

        // creates each factory
        String factoryName = null;
        for (int i = 0; i < factoryNames.size(); i++) {
            factoryName = (String) factoryNames.elementAt(i);
            try {
                JProp prop = new JProp(factoryName);
                logger.log(BasicLevel.DEBUG, "Creating mail factory " + factoryName);
                createMailFactory(factoryName, prop.getFilesEnv());
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "JOnAS: Cannot create mail factory : " + factoryName + " " + e);
                logger.log(BasicLevel.ERROR, "Please check if " + factoryName + ".properties is available");
            }
        }

        try {
            // Register MailService MBean : MailServiceImplMBean
            if (mbeanServer != null) {
                mbeanServer.registerMBean(this, JonasObjectName.mailService());
            }
        } catch (InstanceAlreadyExistsException iae) {
            throw new ServiceException("Cannot start the MailService (Already Exists)" , iae);
        } catch (MBeanRegistrationException mbre) {
            throw new ServiceException("Cannot start the MailService (MBean registration error)" , mbre);
        } catch (NotCompliantMBeanException ncmbe) {
            throw new ServiceException("Cannot start the MailService (MBean Not compliant error)" , ncmbe);
        }


    }
    
    /**
     * Stop the Mail service.
     * @throws ServiceException if the stop failed.
     */
    protected void doStop() throws ServiceException {

        try {
            unbindMailFactories();
        } catch (MailServiceException e) {
            logger.log(BasicLevel.ERROR, "Cannot unbind mail factories " + e);
            throw new ServiceException("Cannot unbind mail factories ", e);
        }
    
        if (mbeanServer != null) {
            try {
                mbeanServer.unregisterMBean(JonasObjectName.mailService());
                // unregister all mail factories
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "Cannot stop the Mail Service (JMX): " + e.getMessage());
                throw new ServiceException("Cannot stop the mail service (JMX).", e);
            }
        }
        logger.log(BasicLevel.DEBUG, "mail service stopped");
    }

    /**
     * Create a mail factory with the specified properties and register it 
     * into the registry.
     * @param props the properties used to configure the mail factory.
     * @throws MailServiceException if the creation or the registration of 
     * the factory failed.
     */
    public void createMailFactory(String factoryName, Properties props) 
            throws MailServiceException {

        Object factory = null;

        //Factory type/jndi name must be non null
        String factoryType = props.getProperty("mail.factory.type");
        String jndiName = props.getProperty("mail.factory.name");

        if (jndiName == null) {
	    logger.log(BasicLevel.ERROR, "The property 'mail.factory.name' is a required property");
            throw new MailServiceException("The property 'mail.factory.name' is a required property for this factory.");
        }

        if (factoryType == null) {
	    logger.log(BasicLevel.ERROR, "The property 'mail.factory.type' is a required property.");
            throw new MailServiceException("The property 'mail.factory.type' is a required property for this factory.");
        }

	// Verify that jndi name not already used
	if (bindedFactories.containsKey(jndiName)) {
	    logger.log(BasicLevel.ERROR, "There is already a factory bound with the name " + jndiName);
	    throw new MailServiceException("There is already a factory bound with the name '" + jndiName +  "', please correct the provided configuration properties");
	}

        //Define the type of our factory
        int typeOfFactory;
        if (factoryType.equalsIgnoreCase("javax.mail.Session")) {
            typeOfFactory = JAVAX_MAIL_SESSION_FACTORY;
        } else if (factoryType.equalsIgnoreCase("javax.mail.internet.MimePartDataSource")) {
            typeOfFactory = JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE;
        } else {
            typeOfFactory = 0;
        }
      
	// Create the factory object and register it in the internal data structure
        switch (typeOfFactory) {
            
        case JAVAX_MAIL_SESSION_FACTORY :
            JMailSession jMailSession = new JMailSession(factoryName, jndiName, props);
            jMailSessionFactories.put(factoryName, jMailSession);
            factory = jMailSession;
            break;
            
        case JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE :
            JMailMimePartDS jMailMimePartDs = new JMailMimePartDS(factoryName, jndiName, props);
            jMailMimePartDSFactories.put(factoryName, jMailMimePartDs);
            factory = jMailMimePartDs;
            break;

        default :
            throw new MailServiceException("Can not create a factory of the type '" + factoryType + "'. This type is incorrect");
        }

	// Bind the factory object in the naming context
        try {
            ictx.rebind(jndiName, factory);
	    bindedFactories.put(jndiName, factoryName);
        } catch (NamingException e) {
            logger.log(BasicLevel.ERROR, "Cannot bind mail factory '" + jndiName + "'  :" + e.getMessage());
            throw new MailServiceException("Cannot bind mail factory " + jndiName + ".", e);
        }

        logger.log(BasicLevel.WARN, "Mapping Mail Factory " + factoryType + " on " + jndiName);

	// Register the factory object as an MBean with the jmx server
        try {
            if (mbeanServer != null) {
		switch (typeOfFactory) {
		case JAVAX_MAIL_SESSION_FACTORY :
		    mbeanServer.registerMBean(factory, JonasObjectName.sessionMailFactory(factoryName));
		    logger.log(BasicLevel.DEBUG, "Register session mail factory with name " + factoryName);
		    break;
		case JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE :  
		    mbeanServer.registerMBean(factory, JonasObjectName.mimeMailFactory(factoryName));
		    logger.log(BasicLevel.DEBUG, "Register mime mail factory with name " + factoryName);
		}
            }
        } catch (MalformedObjectNameException mone) {
            throw new MailServiceException("Cannot register mail factory '" + factoryName + "' in JMX server", mone);
        } catch (InstanceAlreadyExistsException iaee) {
            throw new MailServiceException("Cannot register mail factory '" + factoryName + "' in JMX server", iaee);
        } catch (MBeanRegistrationException mre) {
            throw new MailServiceException("Cannot register mail factory '" + factoryName + "' in JMX server", mre);
        } catch (NotCompliantMBeanException ncme) {
            throw new MailServiceException("Cannot register mail factory '" + factoryName + "' in JMX server", ncme);
        }

    }

    /**
     * Create a mail factory with the specified properties and register it 
     * into the registry.
     * @param name the mail factory name
     * @param props the properties used to configure the mail factory.
     * @throws MailServiceException if the creation or the registration of 
     * the factory failed.     
     * @throws RemoteException if rmi call failed.
     */
    public void createMailFactoryMBean(String name, Properties props) 
            throws MailServiceException {

        //call the internal proc
        createMailFactory(name, props);
    }

    /**
     * Unregister all the binding factories on the server and in JMX Server.
     * @throws MailServiceException if the unregistration of the factories 
     * failed.
     */
    public void unbindMailFactories() throws MailServiceException {
	//Unbind all factories
	for (Enumeration e = jMailSessionFactories.keys(); e.hasMoreElements();) {
	    unbindMailFactoryMBean((String)e.nextElement());
	}
	for (Enumeration e = jMailMimePartDSFactories.keys(); e.hasMoreElements();) {
	    unbindMailFactoryMBean((String)e.nextElement());
	}
    }

    /**
     * Unregister the factory with the given name.
     * @param factoryName the name of the factory to unbind.
     * @throws RemoteException if rmi call failed.
     * @throws MailServiceException if the unregistration of the factory 
     * failed.
     */
    public void unbindMailFactoryMBean(String factoryName) 
            throws MailServiceException {

	// determine the type of the factory and the jndi name
	String name = null;
	int typeOfFactory;
        JMailSession jmailSession = (JMailSession) jMailSessionFactories.get(factoryName);
        JMailMimePartDS jMailMimePartDS = (JMailMimePartDS) jMailMimePartDSFactories.get(factoryName);
        if (jmailSession != null) {
	    name = jmailSession.getName();
	    typeOfFactory = JAVAX_MAIL_SESSION_FACTORY;
        } else if (jMailMimePartDS != null) {
	    name = jMailMimePartDS.getName();
	    typeOfFactory = JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE;
        } else {
            throw new MailServiceException("Can not unload the mail factory '" + factoryName + "' (this is not a known factory name");
        }

	// unbind the factory
        try {
            ictx.unbind(name);
	    bindedFactories.remove(name);
        } catch (NamingException e) {
            throw new MailServiceException("Can not unbind the factory '" + name + "'.", e);
        }

	// De-register the factory object from the jmx server and remove the factory from the internal data structure
        try {
            if (mbeanServer != null) {
		switch (typeOfFactory) {
		case JAVAX_MAIL_SESSION_FACTORY :
		    mbeanServer.unregisterMBean(JonasObjectName.sessionMailFactory(factoryName));
		    logger.log(BasicLevel.DEBUG, "Unregister session mail factory with name " + factoryName);
		    jMailSessionFactories.remove(factoryName);
		    break;
		case JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE :  
		    mbeanServer.unregisterMBean(JonasObjectName.mimeMailFactory(factoryName));
		    logger.log(BasicLevel.DEBUG, "Unregister mime mail factory with name " + factoryName);
		    jMailMimePartDSFactories.remove(factoryName);
		}
            }
	} catch (MalformedObjectNameException mone) {
            throw new MailServiceException("Cannot unregister mail factory '" + factoryName + "' frpm the JMX server", mone);
        } catch (InstanceNotFoundException infe) {
            throw new MailServiceException("Cannot unegister mail factory '" + factoryName + "' from the JMX server", infe);
        } catch (MBeanRegistrationException mre) {
            throw new MailServiceException("Cannot unregister mail factory '" + factoryName + "' from the JMX server", mre);
        } catch (RuntimeOperationsException e) {
            throw new MailServiceException("Cannot unregister mail factory '" + factoryName + "' from the JMX server", e);
        }


    }

    /**
     * @return the factory name given the jndi name. Null is returned if the given name is not binded.
     */
    public String getFactoryName(String jndiName) {
	return (String)bindedFactories.get(jndiName);
    }

    /**
     * @return Integer Total number of mail factories available in JOnAS
     */ 
    public Integer getCurrentNumberOfMailFactories() {
	return new Integer(jMailSessionFactories.size() + jMailMimePartDSFactories.size());
    }

    /**
     * @return Integer Number of Session mail factories available in JOnAS
     */ 
    public Integer getCurrentNumberOfSessionMailFactories() {
	return new Integer(jMailSessionFactories.size());
    }

    /**
     * @return Integer Number of internet.MimePartDataSource mail factories available in JOnAS
     */ 
    public Integer getCurrentNumberOfMimeMailFactories() {
	return new Integer(jMailMimePartDSFactories.size());
    }

    /**
     * @return mail factory configuration properties from a local file
     */
    public Properties getMailFactoryPropertiesFile(String configFile) throws Exception {
	try {
	    return new JProp(configFile).getFilesEnv();
	} catch (Exception e) {
	    if (e instanceof FileNotFoundException) {
		logger.log(BasicLevel.ERROR, "Please check if "+ configFile + ".properties is available in JONAS_ROOT/config, HOME, or .");
	    } else {
		logger.log(BasicLevel.ERROR, "Error occured when reading file " + configFile);
	    }
	    throw e;
	}	
    }

}
