/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.mail.factory;

//import java
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Properties;
import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;

//import javax
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;

//import objectweb.util
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

//import jonas
import org.objectweb.jonas.common.Log;

/**
 * This class implements JOnAS mail factory objects. It gets the properties from mail factory properties file
 * and build a properties object for the Session.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public abstract class JMail implements Serializable, Referenceable, JMailMBean {

    /**
     * The logger used in JOnAS
     */
    protected static Logger logger = null;

    /**
     * The name of the factory.
     */
    private String factoryName = null;

    /**
     * The jndi name of this object.
     */
    private String name = null;

    /**
     * Properties of the mail factory properties file
     */
    protected Properties mailProperties = null;

    /**
     * Properties for the javax.mail.Session object.
     * Keep only the var used for configuring the javax.mail.Session object.
     */
    protected Properties mailSessionProperties = null;
    
    /**
     * Properties for the authentication.
     */
    protected Properties authenticationProperties = null;

    /**
     * List of all matching properties for configure the javax.mail.Session object
     */
    protected static final String[] PATTERN_MATCHING_SESSION = 
        new String[] { "^mail.debug$",
                       "^mail.from$",
                       "^mail.mime.address.strict$",
                       "^mail.host$",
                       "^mail.store.protocol$",
                       "^mail.transport.protocol$",
                       "^mail.user$",
                       "^mail.(.*).class$",
                       "^mail.(.*).host$",
                       "^mail.(.*).port$",
                       "^mail.(.*).user$"
        };
    
    /**
     * List of all matching properties for configure the
     * authentication of the javax.mail.Session objects.
     */
    protected static final String[] PATTERN_MATCHING_AUTHENTICATION = 
        new String[] { "^mail.authentication.username$",
                       "^mail.authentication.password$"
        };


    /**
     * Constructor of a JMail Object with the given name and properties.
     * @param name the name of this object.
     * @param mailProperties properties for configuring this object.
     */
    public JMail(String factoryName, String name, Properties mailProperties) {
	//Get the logger
        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_MAIL_PREFIX);
        }
	this.factoryName = factoryName;
        this.name = name;
        this.mailProperties = mailProperties;
        this.mailSessionProperties = findMailSessionProperties(mailProperties);
        this.authenticationProperties = findAuthenticationProperties(mailProperties);
    }

    /**
     * Return the jndi name of this object
     * @return jndi name of this Jmail.
     */
    public String getName() {
        return name;
    }

    /**
     * Return the name of this mail factory
     * @return name of this mail factory.
     */
    public String getFactoryName() {
        return factoryName;
    }

    /** 
     * Get the mail server host
     * @return the mail.host property
     */
    public String getHost() {
	String key = null;
	String hostName = null;
	for (Enumeration e = mailSessionProperties.propertyNames(); e.hasMoreElements();) {
	    key = (String) e.nextElement();
            //Check if the property is matching mail.host
	    if (key.equals("mail.host")) {
		hostName = mailSessionProperties.getProperty(key);
		logger.log(BasicLevel.DEBUG, "JMail: mail.host property = " + hostName);
		break;
	    }
	}
	return hostName;
    }

    /** 
     * Set the mail server host
     * @param host the new mail.host property
     */
    public void setHost(String host) {
	logger.log(BasicLevel.DEBUG, "JMail.setHost");
	Properties sessionProps = new Properties();
	String key = null;
	boolean done = false;
	for (Enumeration e = mailSessionProperties.propertyNames(); e.hasMoreElements();) {
	    key = (String) e.nextElement();
	    logger.log(BasicLevel.DEBUG, "JMail.setHost - reading mail session props, prop " + key);
	    //Check if the property is matching mail.host
	    if (key.equals("mail.host")) {
		// replace with the new value
		logger.log(BasicLevel.DEBUG, "JMail: replace mail.host " + mailSessionProperties.getProperty(key) + " by: " + host);
		sessionProps.put(key, host);
		done = true;
	    } else {
		// keep the old value
		sessionProps.put(key, mailSessionProperties.getProperty(key));
	    }
	}
	if (!done) {
	    logger.log(BasicLevel.DEBUG, "JMail: had to add a new property - mail.host");
	    sessionProps.put("mail.host", host);
	}
	mailSessionProperties = sessionProps;  
    }


    /**
     * Return all the properties required for the configuration of
     * javax.mail.Session objects.
     * Used a pattern matching for find the good properties.
     * @param props all the properties in which, javax.mail.Session properties
     * must be found.
     * @return all the properties required for the configuration of Session objects.
     */
    protected Properties findMailSessionProperties(Properties props) {
        Properties sessionProps = new Properties();
        String key = null;

        for (Enumeration e = props.propertyNames(); e.hasMoreElements();) {
            //Check if the property is matching
            key = (String) e.nextElement();
            if (isMatchingSessionProperty(key)) {
                sessionProps.put(key, props.getProperty(key));
            }
        }
        return sessionProps;
    }
    
    /**
     * Return all the properties required for the configuration of
     * the authentication for the javax.mail.Session objects.
     * Used a pattern matching for find the good properties.
     * @param props all the properties defined in the properties file.
     * @return all the properties required for the configuration of Session objects.
     */
    protected Properties findAuthenticationProperties(Properties props) {
        Properties authProps = new Properties();
        String key = null;

        for (Enumeration e = props.propertyNames(); e.hasMoreElements();) {
            //Check if the property is matching
            key = (String) e.nextElement();
            if (isMatchingAuthenticationProperty(key)) {
                authProps.put(key, props.getProperty(key));
            }
        }
        return authProps;
    }

     /**
     * Return if the property is a property of the specified array.
     * @param property to check.
     * @param regexps in order to do the check.
     * @return true if the property is matching, else false.
     */
    protected boolean isMatchingProperty(String property, String[] regexps) {
        boolean matching = false;
        int i = 0;
        RE r = null;

        //In JDK 1.4, we could use string.matches() method

        while (i < regexps.length && !matching) {
            //compare for each regexp
            try {
                r = new RE(regexps[i]);
            } catch (RESyntaxException res) {
                return false;
            }
            if (r.match(property)) {
                matching = true;
            }
            i++;
        }
        return matching;
    }

    /**
     * Return if the property is a javax.mail.Session property.
     * @param property to do the check with this property.
     * @return true if the property is matching, else false.
     */
    protected boolean isMatchingSessionProperty(String property) {
        return isMatchingProperty(property, PATTERN_MATCHING_SESSION);
    }

    /**
     * Return if the property is a authentication property.
     * @param property to do the check with this property.
     * @return true if the property is matching, else false.
     */
    protected boolean isMatchingAuthenticationProperty(String property) {
        return isMatchingProperty(property, PATTERN_MATCHING_AUTHENTICATION);
    }

    /**
     * Retrieves the Reference of the object. 
     * The Reference contains the factory used to create this object 
     * and the optional parameters used to configure the factory.
     * @return the non-null Reference of the object. 
     * @throws NamingException if a naming exception was encountered while 
     * retrieving the reference.
     */
    public abstract Reference getReference() throws NamingException;

    /**
     * Return the type of the factory
     * @return the type of the mail factory
     */
    public abstract String getType();

    /**
     * Retrieves all the properties of this object. 
     * @return the properties of this object.
     */
    public Properties getProperties() {
        return mailProperties;
    }

    /**
     * Retrieves the session properties of this object. 
     * @return the session properties of this object.
     */
    public Properties getMailSessionProperties() {
        return mailSessionProperties;
    }

    /**
     * Retrieves the authentication properties of this object. 
     * @return the authentication properties of this object.
     */
    public Properties getAuthenticationProperties() {
        return authenticationProperties;
    }

    /**
     * Set the authentication properties.
     * @param props the authentication properties.
     */
    public void setAuthenticationProperties(Properties props) {
        this.authenticationProperties = props;
    }

    /**
     * Set the session properties.
     * @param props the session properties.
     */
    public void setSessionProperties(Properties props) {
        this.mailSessionProperties = props;
    }


    /**
     * Return an array of byte from a given object
     * @param obj the object from which we must extract the bytes.
     * @return the byte[] from an object
     */
    protected byte[] getBytesFromObject(Object obj) {

        if (obj == null) {
            return null;
        }

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        
        ObjectOutputStream oos = null;
        byte[] bytes = null;
      
        try {
            oos = new ObjectOutputStream(baos);
            oos.writeObject(obj);
            bytes = baos.toByteArray();
        } catch (Exception e) {
            //Can't tranform
            return null;
        } finally {
            try {
                oos.close();
                baos.close();
            } catch (Exception e) {
                //Not a problem
            }
        }
        return bytes;
    }



    // MBean methods

    /**
     * Get the jndi name of the factory.
     * @return jndi name of the mail factory.
     * @throws RemoteException if rmi call failed.
     */
    public String getNameMBean() throws RemoteException {
        return this.getName();
    }

    /**
     * Get the jndi name of the factory.
     * @return jndi name of the mail factory.
     * @throws RemoteException if rmi call failed.
     */
    public String getFactoryNameMBean() throws RemoteException {
        return this.getFactoryName();
    }

    /**
     * Get the type of the factory.
     * @return type of the mail factory.
     * @throws RemoteException if rmi call failed.
     */
    public String getTypeMBean() throws RemoteException {
        return this.getType();
    }

    /**
     * Get the mail server's host
     * @return the name of the host 
     * @throws RemoteException if rmi call failed.
     */
    public String getHostMBean() throws RemoteException {
	return this.getHost();
    }

    /**
     * Set the mail server's host
     * @parm host the name of the host 
     * @throws RemoteException if rmi call failed.
     */
    public void setHostMBean(String host) throws RemoteException {
	logger.log(BasicLevel.DEBUG, "JMail.setHostMBean");
	this.setHost(host);
    }

    /**
     * Get the authentication properties.
     * @return properties of the authentication.
     * @throws RemoteException if rmi call failed.
     */
    public Properties getAuthenticationPropertiesMBean() throws RemoteException {
        return this.getAuthenticationProperties();
    }

    /**
     * Get the session properties.
     * @return the properties of Session.
     * @throws RemoteException if rmi call failed.
     */
    public Properties getSessionPropertiesMBean() throws RemoteException {
        return this.getMailSessionProperties();
    }

    /**
     * Set the authentication properties.
     * @param props the authentication properties.
     * @throws RemoteException if rmi call failed.
     */
    public void setAuthenticationPropertiesMBean(Properties props) throws RemoteException {
        this.setAuthenticationProperties(props);
    }

    /**
     * Set the session properties.
     * @param props the session properties.
     * @throws RemoteException if rmi call failed.
     */
    public void setSessionPropertiesMBean(Properties props) throws RemoteException {
        this.setSessionPropertiesMBean(props);
    }




}

