/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.mail.factory;


//import java
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.OptionalDataException;
import java.util.Hashtable;
import java.util.Properties;
import java.util.StringTokenizer;

//import javax
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimePart;
import javax.mail.internet.MimePartDataSource;
import javax.mail.internet.InternetAddress;
import javax.mail.Message;
import javax.mail.Session;
import javax.naming.Context;
import javax.naming.Name;
import javax.naming.Reference;
import javax.naming.RefAddr;
import javax.naming.spi.ObjectFactory;

//import objectweb.util
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

//import jonas
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.mail.lib.JAuthenticator;

/**
 * This class provides an implementation of a mime part data source factory 
 * for sending mail.
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public class JMailMimePartDSFactory implements ObjectFactory {

    /**
     * The Java type for which this factory knows how to create objects.
     */
    protected static final String factoryType = "javax.mail.internet.MimePartDataSource";

    /**
     * The logger used in JOnAS
     */
    private static Logger logger = null;



    /**
     * Creates a javax.mail.MimePartDataSource object using the location or 
     * reference information specified. 
     * @param obj the possibly null object containing location or reference 
     * information that can be used in creating an object.
     * @param name the name of this object relative to nameCtx, or null if no 
     * name is specified.
     * @param nameCtx the context relative to which the name parameter is 
     * specified, or null if name is relative to the default initial context.
     * @param environment the possibly null environment that is used in 
     * creating the object. 
     * @return a newly created javax.mail.MimePartDataSource object with the 
     * specific configuration; null if an object cannot be created. 
     * @throws Exception if this object factory encountered an exception 
     * while attempting to create an object, and no other object factories are
     * to be tried.
     */
    public Object getObjectInstance(Object obj, Name name, Context nameCtx,
                                    Hashtable environment) throws Exception {

        //Get the logger
        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_MAIL_PREFIX);
        }

        //Get the reference
        Reference ref = (Reference) obj;

        //Get the class name
        String clname = ref.getClassName();

        //Check the class name
        if (!ref.getClassName().equals(factoryType)) {
            logger.log(BasicLevel.ERROR, "Cannot create object : required type is '" + factoryType + "', but found type is '" + clname + "'.");
            return (null);
        }

        Properties sessionProps = new Properties();
        Properties mimeMessageProps = new Properties();    
        Properties authenticationProps = new Properties();
        RefAddr refAddr = null;

        refAddr =  ref.get("javaxmailSession.properties");
        if (refAddr != null) {
            sessionProps = (Properties) getArrayFromBytes((byte[]) refAddr.getContent());
        }

        refAddr =  ref.get("javaxInternetMimeMessage.properties");
        if (refAddr != null) {
            mimeMessageProps = (Properties) getArrayFromBytes((byte[]) refAddr.getContent());
        }

        refAddr =  ref.get("authentication.properties");
        if (refAddr != null) {
            authenticationProps = (Properties) getArrayFromBytes((byte[]) refAddr.getContent());
        }
        
        //username (Authentication)
        String mailAuthenticationUsername =  authenticationProps.getProperty("mail.authentication.username");

        //Password (Authentication)
        String mailAuthenticationPassword = authenticationProps.getProperty("mail.authentication.password");

        //Field 'to'
        String mailTo = mimeMessageProps.getProperty("mail.to");
        InternetAddress[] toRecipients = null;
        if (mailTo != null) {
            toRecipients = getInternetAddressFromString(mailTo);
        }

        //Field 'cc'
        String mailCc = mimeMessageProps.getProperty("mail.cc");
        InternetAddress[] ccRecipients = null;
        if (ccRecipients != null) {
            ccRecipients = getInternetAddressFromString(mailCc);
        }

        //Field 'bcc'
        String mailBcc = mimeMessageProps.getProperty("mail.bcc");
        InternetAddress[] bccRecipients = null;
        if (bccRecipients != null) {
            getInternetAddressFromString(mailBcc);
        }

        //Field 'subject'
        String mailSubject = mimeMessageProps.getProperty("mail.subject");


        JAuthenticator jAuthenticator = null; 
        if ((mailAuthenticationUsername != null) && (mailAuthenticationPassword != null)) {
            jAuthenticator = new JAuthenticator(mailAuthenticationUsername, mailAuthenticationPassword);
        }

        //Build the message from the Session.
        MimeMessage mimeMessage = new MimeMessage(Session.getInstance(sessionProps, jAuthenticator));

        //And set the properties if there are not null

        //Field 'to'
        if (toRecipients != null) {
            mimeMessage.setRecipients(Message.RecipientType.TO, toRecipients);
        }

        //Field 'cc'
        if (ccRecipients != null) {
            mimeMessage.setRecipients(Message.RecipientType.CC, ccRecipients);
        }

        //Field 'bcc'
        if (bccRecipients != null) {
            mimeMessage.setRecipients(Message.RecipientType.BCC, bccRecipients);
        }
      
        //Field 'subject'
        if (mailSubject != null) {
            mimeMessage.setSubject(mailSubject);
        }

        MimePartDataSource mimePartDS = new MimePartDataSource((MimePart) mimeMessage);
        return mimePartDS;
    }
    

    /** Return an object from an array of bytes.
     * Useful for BinaryRefAddr
     * @param bytes an array of bytes
     * @return an object or null if there is an error of if it's empty
     */
    private Object getArrayFromBytes(byte[] bytes) {
        //Declaration
        ByteArrayInputStream bis = null;
        ObjectInputStream ois = null;
        Object obj = null;

        if (bytes == null) {
            return null;
        }

        bis = new ByteArrayInputStream(bytes);
        try {
            ois = new ObjectInputStream(bis);
            obj =  ois.readObject();

        } catch (ClassNotFoundException cfe) {
            //obj is null
        } catch (OptionalDataException ode) {
            //obj is null
        } catch (IOException ioe) {
            //obj is null
        } finally {
            try {
                bis.close();
                ois.close();
            } catch (Exception e) {
                //Ignore it
            }
        }
        return obj;
    }

    /**
     * Convert a comma separated list into an array of InternetAddress
     * @param txt acomma separated string
     * @return an array of InternetAddress
     */
    
    private InternetAddress[] getInternetAddressFromString(String txt) {
        if (txt == null) {
            return null;
        }
        StringTokenizer st = new StringTokenizer(txt, ",");
        InternetAddress[] addresses = new InternetAddress[st.countTokens()];
        int i = 0;
        try {
            while (st.hasMoreTokens()) {
                addresses[i] = new InternetAddress((String) st.nextToken());
                i++;
            }
        } catch (Exception e) {
            return null;
        }
        return addresses;
    }



}
