/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.server;

import java.io.File;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.StringTokenizer;
import org.objectweb.common.Env;
import org.objectweb.jonas.tools.CheckEnv;


/**
 * This class load all the jars needed to start JOnAS and after this, it 
 * launch the JOnAS server (Server class).
 * @author Ludovic Bert   (initial developper)
 * @author Florent Benoit (initial developper)
 * @author Philippe Durieux (new jonas jar files architecture)
 */
public class Bootstrap {

    private static boolean jdk14;
    private static String jonasLib;
    private static String commonLib;
    private static String javaLib;

    /**
     * Return an array containing all the resources declared in the classpath
     * environment variable.
     * @return a list containing all the resources declared in the classpath
     * environment variable.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getClassPath() throws Exception {
        String cpValue = System.getProperty("java.class.path");
	//System.out.println("classpath="+cpValue); // DEBUG
        String sep = File.pathSeparator;
        JURLs jurls = new JURLs();
        StringTokenizer tokenizer = new StringTokenizer(cpValue, sep);
        while (tokenizer.hasMoreTokens()) {
            File file = new File(tokenizer.nextToken());
            if (file.isDirectory()) {
                jurls.addDir(file);
            } else {
                jurls.add(file);
            }
        }
        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Tomcat.
     * @param tomcatHomeDir the home directory of Tomcat.
     * @return a list containing all the resources used for the integration
     * of Tomcat.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getTomcatResources(File tomcatHomeDir) throws Exception {
        
        JURLs jurls = new JURLs();
        String tomcatLib = tomcatHomeDir + File.separator + "lib";
        String tomcatLibContainer = tomcatLib + File.separator + "container";
        String tomcatLibCommon = tomcatLib + File.separator + "common";
	jurls.add(new File(tomcatLibContainer, "tomcat_util.jar"));
	jurls.add(new File(tomcatLibContainer, "tomcat_modules.jar"));
	jurls.add(new File(tomcatLibContainer, "facade22.jar"));
	jurls.add(new File(tomcatLibContainer, "jasper.jar"));
	jurls.add(new File(tomcatLibCommon, "etomcat.jar"));
	jurls.add(new File(tomcatLibCommon, "tomcat_core.jar"));
	jurls.add(new File(commonLib, "ow_jonas_tomcat3.jar"));
        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Catalina.
     * @param catalinaHomeDir the home directory of Catalina.
     * @return a list containing all the resources used for the integration
     * of Catalina.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getCatalinaResources(File catalinaHomeDir) throws Exception {
                    
        JURLs jurls = new JURLs();
	jurls.add(new File(catalinaHomeDir, "bin" + File.separator + "bootstrap.jar"));
	jurls.add(new File(catalinaHomeDir, "server" + File.separator + "lib"));
	jurls.add(new File(catalinaHomeDir, "common" + File.separator + "lib"));
	jurls.add(new File(catalinaHomeDir, "lib"));
	jurls.add(new File(commonLib, "ow_jonas_tomcat4.jar"));
        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Jetty.
     * @param jettyHomeDir the home directory of Jetty.
     * @return a list containing all the resources used for the integration
     * of Jetty.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getJettyResources(File jettyHomeDir) throws Exception {

        JURLs jurls = new JURLs();
        String jettyLib = jettyHomeDir + File.separator + "lib";
        if (jdk14){
	    jurls.add(new File(jettyLib, "org.mortbay.jetty.jar"));
        } else {
	    jurls.add(new File(jettyLib, "org.mortbay.jetty-jdk1.2.jar"));
        }
	jurls.add(new File(jettyLib, "org.apache.jasper.jar"));
	jurls.add(new File(jettyLib, "org.mortbay.jmx.jar"));
	jurls.add(new File(commonLib, "ow_jonas_jetty.jar"));
        return jurls;
    }    

    /**
     * Return an array containing all the resources used for JOnAS
     * @return a list containing all the resources.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getJonasResources() throws Exception {

        JURLs jurls = new JURLs();

	// add resources for RMI
	if (Env.isSunRMI()) {
	    jurls.add(new File(jonasLib, "rmi"));
	}

	// add resources for JEREMIE
	if (Env.isJeremie()) {
	    jurls.add(new File(jonasLib, "jeremie"));
	    jurls.add(new File(jonasLib, "jonathan"));
	}

	// add resources for DAVID
	if (Env.isDavid()) {
	    jurls.add(new File(jonasLib, "david"));
	    jurls.add(new File(jonasLib, "jonathan"));
	}

        // add Orb Independant JOnAS resources
        jurls.add(new File(commonLib, "ow_jonas.jar"));

	// add some methods missing in JDK 1.3
	if (!jdk14) {
	    jurls.add(new File(commonLib, "ow_jonas_jdk13.jar"));
	}

	// add external jars needed for JOnAS
	jurls.add(new File(commonLib, "j2ee"));
	jurls.add(new File(commonLib, "acme"));
	jurls.add(new File(commonLib, "connector"));
	jurls.add(new File(commonLib, "jmx"));
	jurls.add(new File(commonLib, "joram"));
        jurls.add(new File(commonLib, "jorm"));
        jurls.add(new File(commonLib, "medor"));
	jurls.add(new File(commonLib, "log"));
	jurls.add(new File(commonLib, "velocity"));
	jurls.add(new File(commonLib, "xml"));

        return jurls;
    }

    /**
     * Get the list of the resources to load when launching JOnAS.
     * @return the list of the resources to load when launching JOnAS.
     * @throws Exception if a resource cannot be found.
     */
    public static URL[] getResources() throws Exception {
        JURLs jurls = new JURLs();
	boolean webserver = false;

        // Start with resources from the classpath
        jurls.merge(getClassPath());

	// Load now mail.jar because it must be loaded before activation.jar
	jurls.add(new File(commonLib + File.separator + "j2ee", "mail.jar"));

        // add JOnAS resources
	jurls.merge(getJonasResources());

	// add resources for catalina (tomcat 4.0) if required
	String catalinaHome = System.getProperty("catalina.home");
	if (catalinaHome != null) {
	    jurls.merge(getCatalinaResources(new File(catalinaHome)));
	    webserver = true;
        } 

        // add resources for jetty if required
	String jettyHome = System.getProperty("jetty.home");
	if (!webserver && jettyHome != null) {
	    jurls.merge(getJettyResources(new File(jettyHome)));
	    webserver = true;
        }

        // add resources for tomcat 3.3 if required
	String tomcatHome = System.getProperty("tomcat.home");
	if (!webserver && tomcatHome != null) {
	    jurls.merge(getTomcatResources(new File(tomcatHome)));
	    webserver = true;
	}

	// lib/tools.jar needed to dynamically compile JSP's
	// In Windows: Pb because "Program Files" directory contains a space!?
	if (webserver && !Env.isOsWindows()) {
	    // lib/tools.jar is not in $JAVA_HOME/jre but in $JAVA_HOME
	    String javahome = System.getProperty("java.home");
	    javaLib = javahome + File.separator + ".." + File.separator + "lib";
	    //System.out.println("java Lib="+javaLib); // DEBUG
	    jurls.add(new File(javaLib, "tools.jar"));
	}

        return jurls.toURLs();
    }


    /**
     * Server main routine (Load all the jars needed to start jonas)
     * @param args the list of the args to give to the bootstrap class.
     */
    public static void main(String args[]) {
        try {
	    // Check JDK version
	    jdk14 = Env.isJAVA4();

	    // Get the $JONAS_ROOT/lib directories
	    String jonasRoot = System.getProperty("install.root");
	    if (jonasRoot == null) {
		System.err.println("JONAS_ROOT is not set");
		System.exit(2);
	    }
	    jonasLib = jonasRoot + File.separator + "lib";
	    commonLib = jonasLib + File.separator + "common";

	    // Create a ClassLoader and set it for the current thread.
            JClassLoader jonasLoader = new JClassLoader(getResources(), null);
            Thread.currentThread().setContextClassLoader(jonasLoader);

	    //jonasLoader.printURLs(); // DEBUG
	    //System.out.println("CLASSPATH="+jonasLoader.getClassPath());// DEBUG

            // Launch the "class_to_run" by using our classloader.
            String classToRun = args[0];
            Class clazz = jonasLoader.loadClass(classToRun);
            Class[] argList = new Class[] { args.getClass() };
            Method meth = clazz.getMethod("main", argList);
            String[] newArgs = new String [args.length - 1];
            System.arraycopy(args, 1, newArgs, 0, newArgs.length);
            Object appli = meth.invoke(null, new Object[]{newArgs});
        } catch (Exception e) {
            // We don't use here the trace system.
            System.err.println("JOnAS error: " + e.getMessage());
            System.err.println("JOnAS halting");
            e.printStackTrace();
            System.exit(2);
        }
    }
}
