/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web;


//import java
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.rmi.RemoteException;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Set;
import java.util.HashSet;

//import javax
import javax.management.InstanceAlreadyExistsException;
import javax.management.MBeanServer;
import javax.management.MBeanRegistrationException;
import javax.management.NotCompliantMBeanException;
import javax.naming.Context;
import javax.naming.LinkRef;
import javax.naming.NamingException;
import javax.naming.StringRefAddr;

//import jonas_lib
import org.objectweb.jonas_lib.naming.ContainerNaming;

//import objectweb
import org.objectweb.common.Env;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

//import jonas_web
import org.objectweb.jonas_web.deployment.api.EjbLocalRefDesc;
import org.objectweb.jonas_web.deployment.api.EjbRefDesc;
import org.objectweb.jonas_web.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_web.deployment.api.ResourceRefDesc;
import org.objectweb.jonas_web.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDesc;
import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDescException;
import org.objectweb.jonas_web.deployment.lib.WebDeploymentDescManager;

//import jonas
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.naming.CompNamingContext;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;


/**
 * This abstract class provides an implementation for a dynamic 
 * JWebContainerService service.
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public abstract class AbsJWebContainerServiceImpl 
    extends AbsServiceImpl
    implements JWebContainerService, AbsJWebContainerServiceImplMBean {


    /**
     * Class representing a War (a web application) structure which is 
     * composed of :
     *    - the URL of the war file.
     *    - the URL of the ear containing this war.
     *    - the name of the host on which this war is deployed.
     *    - the context root of this war.
     */
    private class War {
 
        /**
         * The URL of the war file.
         */
        private URL warURL = null;

        /**
         * The URL of the ear containing this war.
         */
        private URL earURL = null;

        /**
         * The name of the host on which this war is deployed.
         */
        private String hostName = null;

        /**
         * The context root where this war is deployed.
         */
        private String contextRoot = null;

        /**
         * Construct a War with the specified name, of the specified 
         * ear. (Used in the case of an ear application).
         * @param warURL the URL of the war.
         * @param earURL the URL of the ear containing this war.
         * @param hostName the name of the host on which this war is deployed.
         * @param contextRoot the context root of this war.
         */
        public War(URL warURL, URL earURL, 
                   String hostName, String contextRoot) {    
            this.warURL = warURL;
            this.earURL = earURL;
            this.hostName = hostName;
            this.contextRoot = contextRoot;
        }

        /**
         * Return true if only if this war is in an ear file.
         * @return true if only if this war is in an ear file.
         */
        public boolean isInEarCase() {
            return earURL != null;
        }

        /**
         * Get the URL of the war file.
         * @return the URL of the war file.
         */
        public URL getWarURL() {
            return warURL;
        }

        /**
         * Get the URL of the ear containing this war.
         * @return the URL of the war file.
         */
        public URL getEarURL() {
            return earURL;
        }

        /**
         * Get the name of the host on which this war is deployed.
         * @return the name of the host on which this war is deployed.
         */
        public String getHostName() {
            return hostName;
        }

        /**
         * Get the context root of this war.
         * @return the context root of this war.
         */
        public String getContextRoot() {
            return contextRoot;
        }

        /**
         * Return true if only if the specified war is equal to this war.
         * @param war the war to compare for the equality.
         * @return true if only if the specified war is equal to this war.
         */
        public boolean equals(War war) {
            return war.getWarURL().equals(warURL) 
                && war.getEarURL().equals(earURL)
                && war.getHostName().equals(hostName)
                && war.getContextRoot().equals(contextRoot);
        }

        /**
         * Return a string representation of the war. Used for debug.
         * @return a string representation of the war.
         */
        public String toString() {
            StringBuffer ret = new StringBuffer();
            ret.append("WAR=\n");
            ret.append("\twarURL=" + getWarURL() + "\n");
            ret.append("\tearURL=" + getEarURL() + "\n");
            ret.append("\thostName=" + getHostName() + "\n");
            ret.append("\tcontextRoot=" + getContextRoot());
            return ret.toString();
        }
    }




    /**
     * Reference on the NamingManager.
     */
    private ContainerNaming naming;

    /**
     * Associates an URL of a WAR file to its classloader.
     */
    private Hashtable warBindings = new Hashtable();

    /**
     * Logger for this service.
     */
    protected static Logger logger = null;
    
    /**
     * Reference to a MBean server.
     */
    private MBeanServer mbeanServer = null;

    /**
     * Reference to the web deployment descriptor manager.
     */
    private WebDeploymentDescManager webDDManager = null;

    /**
     * List of the war names to load when starting the Web container Service.
     */
    private Vector warNames = new Vector();
    
    /**
     * List of the war deployed by the Web container Service.
     */
    private Vector warDeployed = new Vector();

    /**
     * The name of the webapps directory.
     */
    protected static String webAppsDir = "webapps";

    /**
     * Name of the jadmin file
     */
    private final static String JADMIN_FILE = "jadmin.war";

    /**
     * Initialize the service.
     * @param ctx the configuration context of the service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(Context ctx) throws ServiceException {

        //we init the logger
        logger = Log.getLogger(Log.JONAS_WEB_PREFIX);

        // Get the JMX Server via JMX Service
        try {
            mbeanServer = ((JmxService) ServiceManager.getJmxService()).getJmxServer();
        } catch (ServiceException e) {
            // the JMX service may not be started
            mbeanServer = null;
        }

        // Set the XML parsing mode to no validation
        String parsingMode = "false";
        try {
            parsingMode = (String) ctx.lookup("parsingwithvalidation");
        } catch (NamingException e) {
            // No problem if there is no value for 'parsingwithvalidation' 
            // (false by default)
        }

        WebContainerDeploymentDesc.setParsingWithValidation("true".equalsIgnoreCase(parsingMode));
        if ("false".equalsIgnoreCase(parsingMode)) {
            logger.log(BasicLevel.DEBUG, "Web XML parsing without validation");
        } else {
     
            logger.log(BasicLevel.DEBUG, "Web XML parsing with validation");
        }

        // Init the war names to be loaded when starting
        String descsValue = null;
        try {
            descsValue = (String) ctx.lookup("descriptors");
        } catch (NamingException e) {
            // No problem if there is no value for 'descriptors' 
            // (no war to load)
        }
        if (descsValue != null) {
            StringTokenizer st = new StringTokenizer(descsValue, ",");
            while (st.hasMoreTokens()) {
                String fileName = st.nextToken().trim();
                warNames.add(fileName);
            }
        }
 
        //Check if jadmin.war is present in the $JONAS_ROOT/webapps/ or the current dir
        try {
            checkWarFile(JADMIN_FILE);
            //file was found
            //check if it's not in the web.descriptors and append the jadmin to the list
            if (!warNames.contains(JADMIN_FILE)) {
                logger.log(BasicLevel.DEBUG, JADMIN_FILE + " was found, so we add to the autoload list of war files.");
                warNames.add(JADMIN_FILE);
            }
        } catch (JWebContainerServiceException webExc) {
            //jadmin.war was not found, so we don't load the file.
        }

        // Init the web deployment desc manager.
        webDDManager = WebDeploymentDescManager.getInstance();

        //Init the naming manager
        try {
            naming = NamingManager.getInstance();
        } catch (NamingException e) {
            throw new ServiceException("Error when getting the reference to the Naming manager");
        }
    }

    /**
     * Start the service.
     * @throws ServiceException if the startup failed.
     */
    protected void doStart() throws ServiceException {
        
        // Deploy all wars which are in descriptors section
        for (int i = 0; i < warNames.size(); i++) {
            String fileName = (String) warNames.elementAt(i);

            URL warURL = checkWarFile(fileName);
            
            Context contctx = null;
            try {
                contctx = new CompNamingContext(fileName);
                contctx.rebind("warURL", warURL);
                // No context root to rebind because not in ear case.
                // => no rebind("contextRoot", ...)
            } catch (NamingException e) {
                String err = "Cannot start the WebContainerService";
                throw new ServiceException(err, e);
            }
            registerWar(contctx);
        }

        try {
            // Register JWebContainerService MBean : JWebContainerServiceImplMBean
            if (mbeanServer != null) {
                mbeanServer.registerMBean(this, JonasObjectName.webContainerService());
            } 
        } catch (InstanceAlreadyExistsException iae) {
            String err = "Cannot start the WebContainerService Already Exists";
            throw new ServiceException(err, iae);
        } catch (MBeanRegistrationException mbre) {
            throw new ServiceException("Cannot start the WebContainerService (MBean registration error)" , mbre);
        } catch (NotCompliantMBeanException ncmbe) {
            throw new ServiceException("Cannot start the WebContainerService (MBean Not compliant error)" , ncmbe);
        }
    }

    /**
     * Stop the service.
     * @throws ServiceException if the stop failed.
     */
    protected void doStop() throws ServiceException {

        // Get all the deployed war which are not in an ear application file
        // and undeploy them.
        Enumeration wars = warDeployed.elements();
        while (wars.hasMoreElements()) {
            War war = (War) wars.nextElement();
            URL warURL = war.getWarURL();
            //String fileName = (new File(warURL.getFile())).getPath();
            String fileName = warURL.getFile();
            try {
                // Test if the war is in an ear application.
                if (!war.isInEarCase()) {  
                    Context ctx = new CompNamingContext(fileName);
                    ctx.rebind("warURL", warURL);
                    ctx.rebind("isEarCase", new Boolean(false));
                    unRegisterWar(ctx);
                }
            } catch (Exception e) {
                //We don't stop the process of undeploying the wars.
                //We only display an error in the logger.
                String err = "Error when undeploying the war :" + fileName;
                logger.log(BasicLevel.ERROR, err + e.getMessage());
            }
        }

        if (mbeanServer != null) {
            try {
                // unregister AbsJWebContainerServiceImpl MBean
                mbeanServer.unregisterMBean(JonasObjectName.webContainerService());
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "Cannot stop the WebContainerService: " + e);
            }
        }
        logger.log(BasicLevel.DEBUG, "WebContainerService stopped");
    }

    /**
     * Create the environment and delegate the operation to the 
     * implementation of the web container.
     * @param ctx the context which contains the configuration in order 
     * to deploy a WAR.
     * @throws JWebContainerServiceException if the registration of the 
     * WAR failed.
     */
    protected abstract void doRegisterWar(Context ctx)
        throws JWebContainerServiceException;

    /**
     * Delegate the unregistration to the implementation of the web container.
     * @param ctx the context which contains the configuration in order to 
     * undeploy a WAR.
     * @throws JWebContainerServiceException if the unregistration failed.
     */
    protected abstract void doUnRegisterWar(Context ctx) 
        throws JWebContainerServiceException;

    /**
     * Create the environment and delegate the operation to the 
     * implementation of the web container.
     * @param ctx the context which contains the configuration in order to 
     * deploy a WAR.
     * @throws JWebContainerServiceException if the registration of the WAR 
     * failed.
     */
    private void registerWar(Context ctx)
        throws JWebContainerServiceException {
        // There are 6 possible parameters :
        //   - warURL is the URL of the war to deploy (required param).
        //   - parentClassLoader is the parent classloader of 
        //     the web classloader (optional param).
        //   - earClassLoader is the ear classloader (optional param).
        //   - earURL is the URL of the ear (optional parameter :
        //     if earURL is set it means that we are in the ear case, else it 
        //     means that we are in the non ear case).
        //   - altDD is the optional deployment descriptor (optional param).
        //   - contextRoot is the context root for the Web application
        //     (optional param).
           
        // Get the URL of the ear application file in the case of an
        // ear application, null otherwise.
        URL earURL = null;
        String contextRoot = null;
        try {
            earURL = (URL) ctx.lookup("earURL");
            contextRoot = (String) ctx.lookup("contextRoot");
        } catch (NamingException e) {
            if (earURL != null || contextRoot != null) {
                String err = "Error while getting parameter from context param :" + e.getMessage();
                logger.log(BasicLevel.ERROR, err);
                throw new JWebContainerServiceException(err, e);
            } // else nothing to do : non-ear case.
        }

        // Get the URL of the war to deploy ...
        URL warURL = null;
        try {
            warURL = (URL) ctx.lookup("warURL");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param :" + e.getMessage();
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err, e);
        }

        // ... and check if the war to deploy exists.
        File warFile = new File(warURL.getFile());
        if (!warFile.exists()) {
            String err = "registerWar: '" + warFile.getPath() + "' not found";
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err);
        }

        // Check if the war to deploy is not already deployed.
        War war = getWar(warURL);
        if (war != null) {
            // The war is already deployed.
            String err = "Cannot deploy war '" + warURL.getFile() + "' is already deployed. You must undeploy the war before a new deployment.";
            throw new JWebContainerServiceException(err);
        }

        String fileName = warURL.getFile();

        // Classloader
        ClassLoader parent = Thread.currentThread().getContextClassLoader();
        URL url[] = {warURL};
        URLClassLoader loaderForCls = null;
        URLClassLoader parentLoader = null;
        URLClassLoader earClassLoader = null;
        try {
            parentLoader = (URLClassLoader) ctx.lookup("parentClassLoader");
            earClassLoader = (URLClassLoader) ctx.lookup("earClassLoader");
            loaderForCls = new URLClassLoader(url, parentLoader);
        } catch (NamingException ne) {
            //Case of non-ear application : only a single war
            loaderForCls = new URLClassLoader(url, parent);
        }

        // Get the deployment descriptor from file
        WebContainerDeploymentDesc webDD = null;
        try {
            webDD = webDDManager.getDeploymentDesc(warURL, loaderForCls, earClassLoader);
        } catch (WebContainerDeploymentDescException e) {
            String err = "Cannot read the deployment descriptors '" + fileName + "'";
            logger.log(BasicLevel.ERROR, err + ": " + e);
            throw new JWebContainerServiceException(err, e);
        }

        // Populate the java:comp/env (ENC) environment.
        URLClassLoader webClassLoader = null;  
        try {
            logger.log(BasicLevel.DEBUG, "Populating environment of the file " + warURL.getFile());
            Context ctxParam = new CompNamingContext(warURL.getFile());
            ctxParam.rebind("DeploymentDesc", webDD);
            ctxParam.rebind("warName", warURL.getFile());
            if (parentLoader == null) {
                webClassLoader = new URLClassLoader(new URL[0], parent);
            } else {
                webClassLoader = new URLClassLoader(new URL[0], parentLoader);
            }
            ctxParam.rebind("parentCL", webClassLoader);
            // Do the populating of the java:comp/env (ENC) environment.
            setWebEnvironment(ctxParam);
        } catch (Exception e) {
            //populating environment failed.
            String err = "Error when populating ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }

        // Set the right context root for the web application, the priority is
        // the following : 
        //    1 - context-root of application.xml
        //    2 - context-root of jonas-web.xml
        //    3 - context-root is the name of the file without .war.
        if (earClassLoader == null && contextRoot == null) {
            String cRoot = webDD.getContextRoot();
            if (cRoot == null) {
                String file = new File(warURL.getFile()).getName();
                if (file.toLowerCase().endsWith(".war")) {
                    contextRoot = file.substring(0, file.length() - 4);
                } else {
                    String err = "Cannot deploy war " + file;
                    err = err + " it does not end with .war";
                    throw new JWebContainerServiceException(err);
                }
            } else {
                contextRoot = cRoot;
            }
        }

        // Set the name of the host where to deploy the war if it is 
        // specified in the jonas-web.xml.
        String hostName = webDD.getHost();
        
        // Configure the context to give to doRegisterWar ...
        Context ctxParam = null;
        try {
            ctxParam = new CompNamingContext(warURL.getFile());
            ctxParam.rebind("warURL", warURL);
            ctxParam.rebind("parentCL", webClassLoader);
            if (hostName != null) {
                ctxParam.rebind("hostName", hostName);
            }
            ctxParam.rebind("contextRoot", contextRoot);
        } catch (NamingException e) {
            String err = "Error when deploying the war  '" + fileName + "'";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }
        // ... and delegate the registration of the war to the wrapper.
        doRegisterWar(ctxParam);
        // Finaly indicates that the war is deployed.
        warDeployed.addElement(new War(url[0], earURL, hostName, contextRoot));
        warBindings.put(url[0], webClassLoader);
        System.out.println("War " + fileName + " available at " + contextRoot);
    } 

    /**
     * Register a WAR by delegating the operation to the registerWar() 
     * method. This is used for JMX management.
     * @param fileName the name of the war to deploy.
     * @throws RemoteException if rmi call failed.
     * @throws JWebContainerServiceException if the registration failed.
     */
    public void registerWarMBean(String fileName) 
        throws RemoteException, JWebContainerServiceException {

        // convert the file name to the appropriate  url and check if the file
        // exists.
        URL warURL = checkWarFile(fileName);

        // create the context to call the registerWar method.
        Context ctx = null;
        try {
             ctx = new CompNamingContext(fileName);
             ctx.rebind("warURL", warURL);
        } catch (NamingException e) {
            String err = "Error when deploying the war  '" + fileName + "'";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }

        // call the registerWar method.
        registerWar(ctx);
    }

    /**
     * Set the environment of the web container inside the given context.
     * @param ctxParam the java:comp/env/ environment where is stored 
     * the values of the web container environment.
     * @throws JWebContainerServiceException if the populating of the 
     * environment failed.
     */
    private  void setWebEnvironment(Context ctxParam)  
        throws JWebContainerServiceException {

        WebContainerDeploymentDesc dd = null;
        String warName = null;
        ClassLoader parentClassLoader = null;

        /* get the parameters
           - Deployment desc of xml
           - Name of the war
           - Parent Class loader
        */
        try {
            dd = (WebContainerDeploymentDesc) ctxParam.lookup("DeploymentDesc");
            warName = (String) ctxParam.lookup("warName");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }

        //Get the parentCL
        try {
            parentClassLoader = (ClassLoader) ctxParam.lookup("parentCL");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }

        //Create the java:comp/env entry and bind the entries
        try {
            // Create a JNDI context for this war
            javax.naming.Context ctx = naming.createEnvironmentContext(warName);
            naming.setComponentContext(ctx, parentClassLoader);

            // War Environment entries
            EnvEntryDesc [] envt = dd.getEnvEntryTags();
            for (int i = 0; i < envt.length; i++) {
                // get information in descriptor
                String name = envt[i].getName();
                Object obj = envt[i].getValue();
     
                // register object in JNDI
                logger.log(BasicLevel.DEBUG, warName + ": Binding object " + name + " -> " + obj);
                ctx.rebind(name, obj);
            }

            // Resource References
            ResourceRefDesc [] resref = dd.getResourceRefTags();
            for (int i = 0; i < resref.length; i++) {
                // get information in descriptor
                String name = resref[i].getName();
                String resname = resref[i].getJndiName();
                // build the LinkRef that will be registered:
                // FactoryClassName = null, size = 1, refAddr = resname.
                LinkRef lref = new LinkRef(resname);

                // if we are using David add a parameter to the link to tell
                // that the object must be found in rmiregistry (and not in 
                // CosNaming) this info is used by CompNamingContext.lookup()
                if (Env.isDavid()) {
                    lref.add(new StringRefAddr("TargetNaming", "RmiRegistry"));
                }

                // register object in JNDI
                logger.log(BasicLevel.DEBUG, warName + ": Linking resource " + name + " -> " + resname);
                ctx.rebind(name, lref);
            }

 
            // Resource Environment References
            ResourceEnvRefDesc [] resEnvref = dd.getResourceEnvRefTags();
            for (int i = 0; i < resEnvref.length; i++) {
                // get information in descriptor
                String name = resEnvref[i].getName();
                String resname = resEnvref[i].getJndiName();
                LinkRef lref = new LinkRef(resname);

                // if we are using David add a parameter to the link to 
                // tell that the object must be found in rmiregistry (and not 
                // in CosNaming) this info is used by 
                // CompNamingContext.lookup()
                if (Env.isDavid()) {
                    lref.add(new StringRefAddr("TargetNaming", "RmiRegistry"));
                }
    
                logger.log(BasicLevel.DEBUG, warName + ": Linking resource environment " + name  + " -> " + resname);
                ctx.rebind(name, lref);
            }

            // EJB References
            EjbRefDesc [] ejbref = dd.getEjbRefTags();
            for (int i = 0; i < ejbref.length; i++) {
                // get information in descriptor
                String name = ejbref[i].getName();
                String ejbname = null;
                //                String ejblink = ejbref[i].getEjbLink();
                ejbname = ejbref[i].getJndiName();
//                 if (ejbname == null) {
//                     throw new JWebContainerServiceException("jonas-ejb-ref missing for ejb-ref '" + name);
//                 }

                LinkRef lref = new LinkRef(ejbname);
                // if we are using David add a parameter to the link to 
                // tell that the object must be found in the CosNaming 
                // (and not in rmiregistry) this info is used by 
                // CompNamingContext.lookup()
                if (Env.isDavid()) {
                    lref.add(new StringRefAddr("TargetNaming", "CosNaming"));
                }

                logger.log(BasicLevel.DEBUG, warName + ": Linking ejb " + name + " -> " + ejbname);
                ctx.rebind(name, lref);
            }

            // EJB Local Refs
            // We use here ejb-link tag. This should be used also for 
            // ejb-ref when we are able to manage references to 
            // another jar file.
            EjbLocalRefDesc [] ejblocalref = dd.getEjbLocalRefTags();
            for (int i = 0; i < ejblocalref.length; i++) {
                String name = ejblocalref[i].getName();
                //                String ejblink = ejblocalref[i].getEjbLink();
                String ejbname = ejblocalref[i].getJndiLocalName();
                LinkRef lref = new LinkRef(ejbname);
                logger.log(BasicLevel.DEBUG, warName + ": Linking ejb " + name + " -> " + ejbname);
                ctx.rebind(name, lref);
            }

        } catch (NamingException e) {
            String err = "Error while populating environment of the war file "  + warName;
            logger.log(BasicLevel.ERROR, err + "  :" + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }
    }

    /**
     * Delegate the unregistration to the implementation of the web 
     * container and delete the environment associated to the WAR file.
     * @param ctx the context which contains the configuration in order 
     * to undeploy a WAR.
     * @throws JWebContainerServiceException if the unregistration failed.
     */
    private void unRegisterWar(Context ctx) 
        throws JWebContainerServiceException {

        // Gets the 2 parameters : 
        //    warURL, isEarCase, hostName and contextRoot.
        URL warURL = null;
        boolean isEarCase = true;
        try {
            warURL = (URL) ctx.lookup("warURL");
            isEarCase = ((Boolean) ctx.lookup("isEarCase")).booleanValue();
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param.";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }
 
        // get the file name for displaying message.
        String fileName = warURL.getFile();

        // Check if the war is deployed.
        War war = null;
        war = getWar(warURL);
        if (war == null) {
            String err = "Cannot undeploy war: '" + fileName + "' is not deployed.";
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err);
        }

        // Check if the war can be undeployed (2 case authorized) :
        //    - case EAR and war is in an EAR.
        //    - case non EAR and war isn't in an EAR.
        if (isEarCase != war.isInEarCase()) {
            String err = "Cannot undeploy war: '" + fileName + "' it is in an ear application. You must undeploy the ear associated.";
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err);
        }

        // Call the specific method of the implementation of the web container.
        try {
            //ctx.rebind("hostName", war.getHostName());
            ctx.rebind("contextRoot", war.getContextRoot());
        } catch (NamingException e) {
            String err = "Error when deploying the war  '" + fileName + "'";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }
        doUnRegisterWar(ctx);

        // Remove the context.
        URLClassLoader loader = (URLClassLoader) warBindings.remove(warURL);
        naming.unSetComponentContext(loader);

        // Indicates that the war is not in the deployed war.
        // (here we are sure that the war is not null and isn't in an ear).
        warDeployed.removeElement(war);
        
        System.out.println("War " + fileName + " no longer available");
    }

    /**
     * Unregister a WAR by delegating the operation to the unRegisterWar() 
     * method. This is used for JMX management.
     * @param fileName the name of the war to undeploy.
     * @throws RemoteException if rmi call failed.
     * @throws JWebContainerServiceException if the unregistration failed.
     */
    public void unRegisterWarMBean(String fileName) 
        throws RemoteException, JWebContainerServiceException {
 
        // Convert the given file name to an url and check if the war is
        // deployed ...
        URL warURL = checkWarDeployed(fileName);

        // ... and do the undeployment.
        Context ctx = null;
        try {
            ctx = new CompNamingContext(fileName);
            ctx.rebind("warURL", warURL);
            ctx.rebind("isEarCase", new Boolean(false));
        } catch (NamingException e) {
            String err = "Error when undeploying the war file '" + fileName + "'";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }
        unRegisterWar(ctx);
    }

    /**
     * Deploy the given wars of an ear file with the specified parent 
     * classloader (ejb classloader or ear classloader). (This method
     * is only used for the ear applications, not for the 
     * web applications).
     * @param ctx the context containing the configuration 
     * to deploy the wars.<BR>
     * This context contains the following parameters :<BR>
     *    - urls the list of the urls of the wars to deploy.<BR>
     *    - earURL the URL of the ear application file.<BR>
     *    - parentClassLoader the parent classLoader of the wars.<BR>
     *    - earClassLoader the ear classLoader of the j2ee app.<BR>
     *    - altDDs the optional URI of deployment descriptor.<BR>
     *    - contextRoots the optional context root of the wars.<BR>
     * @throws JWebContainerServiceException if an error occurs during 
     * the deployment.
     */
    public void deployWars(Context ctx) throws JWebContainerServiceException {
        // Gets the parameters from the context :
        //    - urls the list of the urls of the wars to deploy.
        //    - earURL the URL of the ear application file.
        //    - parentClassLoader the parent classLoader of the wars.
        //    - earClassLoader the ear classLoader of the j2ee app.
        //    - altDDs the optional URI of deployment descriptor.
        //    - contextRoots the optional context root of the wars.
        URL[] urls = null;
        URL earURL = null;
        ClassLoader parentClassLoader = null;
        ClassLoader earClassLoader = null;
        URL[] altDDs = null;
        String[] contextRoots = null;
        try {
            urls = (URL[]) ctx.lookup("urls");
            earURL = (URL) ctx.lookup("earURL");
            parentClassLoader = (ClassLoader) ctx.lookup("parentClassLoader");
            earClassLoader = (ClassLoader) ctx.lookup("earClassLoader");
            altDDs = (URL[]) ctx.lookup("altDDs");
            contextRoots = (String[]) ctx.lookup("contextRoots");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }
        webDDManager.setAltDD(earClassLoader, urls, altDDs);

        // Deploy all the wars of the ear application.
        for (int i = 0; i < urls.length; i++) {
            // Get the name of a war to deploy.
            String fileName = urls[i].getFile();
            logger.log(BasicLevel.DEBUG, "Deploy war '" + fileName + "' for the ear service");

            // The context to give for the creation of the container
            // associated to the ejb-jar.
            Context contctx = null;
            try {
                contctx = new CompNamingContext(fileName);
                contctx.rebind("warURL", urls[i]);     
                contctx.rebind("parentClassLoader", parentClassLoader);
                contctx.rebind("earClassLoader", earClassLoader);
                contctx.rebind("earURL", earURL);
                if (altDDs[i] != null) {
                    contctx.rebind("altDD", altDDs[i]);
                }
                if (contextRoots[i] != null) {
                    contctx.rebind("contextRoot", contextRoots[i]);
                }
                registerWar(contctx);
            } catch (Exception e) {
                // A war is corrupted so undeploy all the deployed war
                // of the ear application.
                logger.log(BasicLevel.ERROR, "Error when deploying '" + fileName + "'");
                logger.log(BasicLevel.ERROR, e.getMessage());
                logger.log(BasicLevel.ERROR, "Undeploy war of the ear application");

                for (int j = 0; j < i; j++) {
                    String warFileName = urls[j].getFile();      
                    try {
                        // Try to undeploy a war of the ear application.
                        CompNamingContext context = new CompNamingContext(warFileName);
                        context.rebind("warURL", urls[j]);
                        context.rebind("isEarCase", new Boolean(true));
                        unRegisterWar(context);
                    } catch (Exception ex) {
                        // Cannot undeploy a war of the ear application
                        // So there is an error message.
                        logger.log(BasicLevel.ERROR, "Error when undeploying '" + warFileName + "'");
                        logger.log(BasicLevel.ERROR, ex.getMessage());
                        logger.log(BasicLevel.ERROR, "Cannot undeploy war of the ear application");
                    }
      
                }
                throw new JWebContainerServiceException("Error during the deployment", e);
            }
        } 
    }

    /**
     * Undeploy the given wars of an ear file with the specified parent 
     * classloader (ejb classloader or ear classloader). (This method
     * is only used for the ear applications, not for the 
     * war applications).
     * @param urls the list of the urls of the wars to undeploy.
     */
    public void unDeployWars(URL[] urls) {
        for (int i = 0; i < urls.length; i++) {
            String warFileName = urls[i].getFile();      
            try {
                // Try to undeploy a war of the ear application.
                CompNamingContext context = new CompNamingContext(warFileName);
                context.rebind("warURL", urls[i]);
                context.rebind("isEarCase", new Boolean(true));
                unRegisterWar(context);
            } catch (Exception ex) {
                // Cannot undeploy a war of the ear application
                // So there is an error message.
                logger.log(BasicLevel.ERROR, "Error when undeploying '" + warFileName + "'");
                logger.log(BasicLevel.ERROR, ex.getMessage());
                logger.log(BasicLevel.ERROR, "Cannot undeploy war of the ear application");
            }     
        }
    }

    /**
     * Get the war identified by its URL (.war).
     * @param url the URL of the war to get.
     * @return the war indentified by its URL, or null if the war
     * is not found.
     */
    private War getWar(URL url) {
        Enumeration wars = warDeployed.elements();
        while (wars.hasMoreElements()) {
            War war  = (War) wars.nextElement();
            if (war.getWarURL().equals(url)) {
                return war;
            }
        }
        return null;
    }

    /**
     * Make a cleanup of the cache of deployment descriptor. This method must 
     * be invoked after the ear deployment by the EAR service.
     * @param earClassLoader the ClassLoader of the ear application to
     * remove from the cache.
     */
    public void removeCache(ClassLoader earClassLoader) {
        webDDManager.removeCache(earClassLoader);
    }


    /**
     * Check if the specified file name correspond to a file which is located 
     * relatively to where the JOnAS server is launched or in the 
     * $JONAS_ROOT/web-apps.
     * @param fileName the file to check if it exists.
     * @return the URL of the file found (either relatively to the JOnAS 
     * server or to the $JONAS_ROOT/web-apps).
     * @throws JWebContainerServiceException if the specified file 
     * does't exists.
     */
    private URL checkWarFile(String fileName) 
        throws JWebContainerServiceException {
        
        File f = null;
        try {
            f = new File(fileName).getCanonicalFile();
            if (!f.exists()) {
                boolean found = false;
                String warFileName = null;
                if (fileName.endsWith(".war")) {
                    // In case of the name is a war file name, check also in
                    // the JONAS_ROOT/webapps directory
                    String jonasRoot = JProp.getInstallRoot();
                    warFileName = jonasRoot + File.separator + webAppsDir + File.separator + fileName;
                    f = new File(warFileName).getCanonicalFile();
                    found = f.exists();
                }
                if (found) {
                    fileName = f.getPath();//warFileName;
                } else {
                    String err = "File '" + f.getPath() + "' not found";
                    logger.log(BasicLevel.ERROR, err);
                    throw new JWebContainerServiceException(err);
                }
            }
        } catch (IOException e) {
            String err = "Invalid war file name '" + fileName;
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err, e);
        }

        URL warURL = null;
        try {
            warURL = f.toURL();
        } catch (MalformedURLException e) {
            String err = "Invalid war file name '" + fileName + "'.";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        }
        return warURL;
    }

    /**
     * Check if the specified file is already deployed in the JOnAS server
     * and return the URL of this deployed war file.
     * @param fileName the name of the WAR file to check.
     * @return the URL of the deployed war file.
     * @throws JWebContainerServiceException if the file name doesn't 
     * correspond to a deployed war.
     */
    private URL checkWarDeployed(String fileName) 
        throws JWebContainerServiceException {

        URL url = null;
        try {
            Enumeration wars = warDeployed.elements();
            while (wars.hasMoreElements()) {
                War war = (War) wars.nextElement();
                url = (new File(fileName).getCanonicalFile()).toURL();
                if (war.getWarURL().equals(url)) {
                    return url;
                }
            }
            if (fileName.endsWith(".war")) {
                // In case of the name is a ejb-jar file name, check also in
                // the JONAS_ROOT/ejbjars directory
                String jonasRoot = JProp.getInstallRoot();
                String warFileName = jonasRoot + File.separator + webAppsDir + File.separator + fileName;
                wars = warDeployed.elements();
                while (wars.hasMoreElements()) {
                    War war  = (War) wars.nextElement();
                    url = (new File(warFileName).getCanonicalFile()).toURL();
                    if (war.getWarURL().equals(url)) {
                        return url;
                    }
                }
            }

            String err = "Cannot undeploy war: '" + fileName + "' is not deployed.";
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err);

        } catch (MalformedURLException e) {
            String err = "Invalid war file name '" + fileName + "'.";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);
        } catch (IOException e) {
            String err = "Invalid war file name '" + fileName;
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err, e);
        }
    }   

    /**
     * @return current number of wars deployed in the JOnAS server
     */
    public Integer getCurrentNumberOfWars() {
	return new Integer(warDeployed.size());
    }

    /**
     * This method is added temporarily. It will disapear when Wars will have their associated MBeans
     * (when Wars will become manageable)
     * @return the names of the wars currently deployed in the JOnAS server
     */
    public Set getWarNames() {
	HashSet names = new HashSet();
	for (Enumeration wars = warDeployed.elements() ; wars.hasMoreElements() ;) {
	    War war = (War) wars.nextElement();
            URL warURL = war.getWarURL();
            names.add(warURL.getFile());
	}
	return names;
    }
}
