/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Ludovic BERT & Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web.catalina;

//import java
import java.util.Stack;
import java.io.File;
import java.io.FileNotFoundException;
import java.net.URL;
import java.net.URLClassLoader;
import java.lang.reflect.Constructor;

//import javax
import javax.servlet.ServletContext;
import javax.naming.NamingException;

//import jonas
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.AbsJWebContainerServiceImpl;
import org.objectweb.jonas.web.JWebContainerServiceException;
import org.objectweb.jonas.web.lib.JarTools;
import org.objectweb.util.monolog.api.BasicLevel;

//import catalina
import org.apache.catalina.Globals;
import org.apache.catalina.Container;
import org.apache.catalina.Deployer;
import org.apache.catalina.Host;
import org.apache.catalina.Loader;
import org.apache.catalina.Lifecycle;
import org.apache.catalina.LifecycleListener;
import org.apache.catalina.loader.WebappClassLoader;
import org.apache.catalina.core.StandardContext;
import org.apache.catalina.util.xml.SaxContext;
import org.apache.catalina.util.xml.XmlAction;
import org.apache.catalina.util.xml.XmlMapper;

// import sax
import org.xml.sax.AttributeList;


/**
 * This class provides an implementation of the Catalina service (as 
 * web container service).
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public class CatalinaJWebContainerServiceImpl 
    extends AbsJWebContainerServiceImpl {


    /**
     * Relative path of the configuration file
     */
    protected static final String CONFIG_FILE = "conf/server.xml";

    /**
     * The reference on the encapsulated embedded catalina.
     */
    protected EmbeddedCatalina embedded = null;

    /**
     * The shared extensions class loader for this server.
     */
    protected ClassLoader parentClassLoader =
        Thread.currentThread().getContextClassLoader();


    /**
     * Initialize the Catalina service.
     * @param ctx the configuration context of the service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(javax.naming.Context ctx) throws ServiceException {
        super.doInit(ctx);
        boolean useSecurity = false;
        try {
            ServiceManager.getSecurityService();
            useSecurity = true;
        } catch (Exception e) {
            // nothing to do 
        }
        
        // if the user do not specify catalina.base,
        // catalina.base = catalina.home.
        String catalinaHome = System.getProperty("catalina.home");
        if (catalinaHome != null) {
            if (System.getProperty("catalina.base") == null) {
                System.setProperty("catalina.base", catalinaHome);
            }
        }
        embedded = new EmbeddedCatalina(new JCatalinaLogger(), useSecurity);
    }

    /**
     * Start the Catalina service in a new thread
     * @throws ServiceException if the startup failed.
     */
    public void doStart() throws ServiceException {

        // Create the mapper for the parsing of the server.xml.
        JXmlMapper mapper = mapperConfig();
        
        // Execute the mapper for the parsing of the server.xml.
        // And configure the embedded catalina.
        File configFile = null;
        try {
            configFile = getConfigFile();
        } catch (FileNotFoundException e) {
            logger.log(BasicLevel.ERROR, "Cannot find config file " + configFile);
            throw new ServiceException(e.getMessage(), e);
        }
        try {
            mapper.readXml(configFile, embedded);
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Error when reading config file " + configFile);
            throw new ServiceException(e.getMessage(), e);
        }

        // Finaly start embedded catalina ...
        try {
            embedded.start();
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Cannot start catalina");
            throw new ServiceException(e.getMessage(), e);
        }
        
        // ... and deploy wars of the jonas.properties
        super.doStart();
    }

    /**
     * Stop the Catalina service.
     * @throws ServiceException if the stop failed.
     */
    protected void doStop() throws ServiceException {
        // Undeploy the wars ...
        super.doStop();
        
        // ... and shut down embedded catalina 
        try {
            embedded.stop();
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Cannot stop catalina");
            throw new ServiceException(e.getMessage(), e);
        }
    }

    /**
     * Deploy a specific WAR file specified in the context.
     * @param ctx the context which contains the configuration in order 
     * to deploy the WAR.
     * @throws JWebContainerServiceException if the registration of the WAR 
     * failed.
     */
    protected void doRegisterWar(javax.naming.Context ctx) 
        throws JWebContainerServiceException {

        // Get the 4 parameters :
        //   - warURL is the URL of the war to register (required param).
        //   - contextRoot is the context root to which this application
        //     should be installed (must be unique) (required param).
        //   - hostName is the name of the host on which deploy the war
        //     (optional param).
        //   - parentCL the war classloader of this war.
        URL warURL = null;
        String contextRoot = null;
        try {
            warURL = (URL) ctx.lookup("warURL");
            contextRoot = (String) ctx.lookup("contextRoot");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
        
        ClassLoader webClassLoader = null;
        try {
            webClassLoader = (ClassLoader) ctx.lookup("parentCL");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
        
        String hostName = null;
        try {
            hostName = (String) ctx.lookup("hostName");
        } catch (NamingException e) {
            hostName = "";
        }
        
        // Install a new web application, whose web application archive is 
        // at the specified URL, into this container with the specified 
        // context root.
        // A context root of "" (the empty string) should be used for the root
        // application for this container.  Otherwise, the context root must
        // start with a slash.
        Deployer deployer = (Deployer) embedded.findHost(hostName);
        if (contextRoot.equals("/")) {
            contextRoot = "";
        }

        org.apache.catalina.Context c = deployer.findDeployedApp(contextRoot);

        if (c != null) {
            String err = "Cannot deploy war: ";
            err = err + "Context " + contextRoot + " already exists.";
            throw new JWebContainerServiceException(err);
        }

        // Finaly install the war.
        String jonasRoot = JProp.getInstallRoot();
        String fileName = new File(warURL.getFile()).getName();
        String destDir = jonasRoot + File.separator + webAppsDir + File.separator + fileName.substring(0, fileName.length() - 4);
        JarTools.unpack(warURL.getFile(), destDir);

        StandardContext ctxWar = (StandardContext) embedded.createContext("/" + contextRoot, destDir);
        ctxWar.setParentClassLoader(webClassLoader);
        deployer.addChild(ctxWar);
        WebappClassLoader cl = (WebappClassLoader) ctxWar.getLoader().getClassLoader();
        cl.setDelegate(true);
        setClassPath(ctxWar, cl);
    }

    /** 
     * Undeploy a specific WAR file specified in the context.
     * @param ctx the context which contains the configuration in order
     * to undeploy a WAR.
     * @throws JWebContainerServiceException if the unregistration failed.
     */
    protected void doUnRegisterWar(javax.naming.Context ctx) 
        throws JWebContainerServiceException {

        // Get the 2 parameters :
        //   - contextRoot is the context root to be removed (required param).
        //   - hostName is the name of the host to remove the war (optional).
        String contextRoot = null;
        try {
            contextRoot = (String) ctx.lookup("contextRoot");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
    
        String hostName = null;
        try {
            hostName = (String) ctx.lookup("hostName");
        } catch (NamingException e) {
            hostName = "";
        }

        // Remove the specified Context from the set of defined Contexts 
        // for its associated Host.
        Host host = embedded.findHost(hostName);
        org.apache.catalina.Context context = host.map("/" + contextRoot);
        if (context != null) {
            embedded.removeContext(context);
        } else {
            String err = "Cannot undeploy war: ";
            err = err + "Context does not exist for named path " + contextRoot;
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err);
        }
    }

    /**
     * Set the classpath for JSP compilation. For more details see (catalina
     * class WebappLoader.java method setClassPath).
     * @param container the context of the JSP.
     * @param webClassLoader the web classloader.
     */
    protected void setClassPath(Container container, 
                                ClassLoader webClassLoader) {
        // Validate our current state information
        if (!(container instanceof org.apache.catalina.Context)) {
            return;
        }
        ServletContext servletContext =
            ((org.apache.catalina.Context) container).getServletContext();
        if (servletContext == null) {
            return;
        }
        StringBuffer classpath = new StringBuffer();

        // Assemble the class path information from our class loader chain
        int n = 0;
        while (webClassLoader != null) {
            if (!(webClassLoader instanceof URLClassLoader)) {
                break;
            }
            URL repositories[] =
                ((URLClassLoader) webClassLoader).getURLs();
            for (int i = 0; i < repositories.length; i++) {
                String repository = repositories[i].toString();
                if (repository.startsWith("file://")) {
                    repository = repository.substring(7);
                } else if (repository.startsWith("file:")) {
                    repository = repository.substring(5);
                } else if (repository.startsWith("jndi:")) {
                    repository =
                        servletContext.getRealPath(repository.substring(5));
                } else {
                    continue;
                }
                if (repository == null) {
                    continue;
                }
                if (n > 0) {
                    classpath.append(File.pathSeparator);
                }
                classpath.append(repository);
                n++;
            }
            webClassLoader = webClassLoader.getParent();
        }

        // Store the assembled class path as a servlet context attribute
        servletContext.setAttribute(Globals.CLASS_PATH_ATTR,
                                    classpath.toString());
    }

    /**
     * Return a File object representing the server.xml configuration file.
     * @return a File object representing the server.xml configuration file.
     * @throws FileNotFoundException if the configuration file is not found.
     */
    protected File getConfigFile() throws FileNotFoundException {
        String fileName = System.getProperty("catalina.base");
        fileName = fileName + File.separator + CONFIG_FILE;
        File file = new File(fileName);
        if (!file.exists()) {
            String err = "Can't find configuration file : " + fileName;
            throw new FileNotFoundException(err);
        }
        return (file);
    }

    /**
     * Create and configure the JXmlMapper that will be used for the xml
     * parsing of the configuration file.
     * @return JXmlMapper the mapper containing the rules for the xml parsing
     * of the server.xml.
     */
    protected JXmlMapper mapperConfig() {

        JXmlMapper mapper = new JXmlMapper();
        mapper.setDebug(999);
        mapper.setValidating(false);

        // Configure the actions we will be using

        mapper.addRule("Server/Listener", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Listener",
                       mapper.setProperties());
        mapper.addRule("Server/Listener", mapper.addChild
                       ("addLifecycleListener",
                        "org.apache.catalina.LifecycleListener"));

        mapper.addRule("Server/Service/Listener", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Listener",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Listener", mapper.addChild
                       ("addLifecycleListener",
                        "org.apache.catalina.LifecycleListener"));

        mapper.addRule("Server/Service/Connector", mapper.objectCreate
                       ("org.apache.catalina.connector.http.HttpConnector",
                        "className"));
        mapper.addRule("Server/Service/Connector", mapper.setProperties());
        mapper.addRule("Server/Service/Connector", mapper.addChild
                       ("addConnector", "org.apache.catalina.Connector"));

        mapper.addRule("Server/Service/Connector/Factory", mapper.objectCreate
                       ("org.apache.catalina.net.DefaultServerSocketFactory",
                        "className"));
        mapper.addRule("Server/Service/Connector/Factory",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Connector/Factory", mapper.addChild
                       ("setFactory",
                        "org.apache.catalina.net.ServerSocketFactory"));

        mapper.addRule("Server/Service/Connector/Listener", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Connector/Listener",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Connector/Listener", mapper.addChild
                       ("addLifecycleListener",
                        "org.apache.catalina.LifecycleListener"));

        mapper.addRule("Server/Service/Engine", mapper.objectCreate
                       ("org.apache.catalina.core.StandardEngine",
                        "className"));
        mapper.addRule("Server/Service/Engine", mapper.setProperties());
        mapper.addRule("Server/Service/Engine",
                       new LifecycleListenerAction
                           ("org.apache.catalina.startup.EngineConfig",
                            "configClass"));
        mapper.addRule("Server/Service/Engine",
                       new SetParentClassLoaderAction(parentClassLoader));
        mapper.addRule("Server/Service/Engine", mapper.addChild
                       ("addEngine", "org.apache.catalina.Engine"));

        mapper.addRule("Server/Service/Engine/Listener", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Listener",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Listener", mapper.addChild
                       ("addLifecycleListener",
                        "org.apache.catalina.LifecycleListener"));

        createStartMapperContext("Server/Service/Engine/Context", mapper);
        createStartMapperDefaultContext(
                                        "Server/Service/Engine/DefaultContext",
                                        mapper);

        mapper.addRule("Server/Service/Engine/Host", mapper.objectCreate
                       ("org.apache.catalina.core.StandardHost",
                        "className"));
        mapper.addRule("Server/Service/Engine/Host", mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Host",
                       new CopyParentClassLoaderAction());
        mapper.addRule("Server/Service/Engine/Host",
                       new LifecycleListenerAction
                           ("org.apache.catalina.startup.HostConfig",
                            "configClass"));
        mapper.addRule("Server/Service/Engine/Host", mapper.addChild
                       ("addChild", "org.apache.catalina.Container"));

        mapper.addRule("Server/Service/Engine/Host/Alias",
                       mapper.methodSetter("addAlias", 0));

//         mapper.addRule("Server/Service/Engine/Host/Cluster",
//                        mapper.objectCreate(null, "className"));
//         mapper.addRule("Server/Service/Engine/Host/Cluster",
//                        mapper.setProperties());
//         mapper.addRule("Server/Service/Engine/Host/Cluster",
//                        mapper.addChild("setCluster",
//                                        "org.apache.catalina.Cluster"));

        createStartMapperContext("Server/Service/Engine/Host/Context", mapper);
        createStartMapperDefaultContext
            ("Server/Service/Engine/Host/DefaultContext", mapper);

        mapper.addRule("Server/Service/Engine/Host/Context/Manager/Store",
                       mapper.objectCreate(null, "className"));
        mapper.addRule("Server/Service/Engine/Host/Context/Manager/Store",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Host/Context/Manager/Store",
                       mapper.addChild("setStore", "org.apache.catalina.Store"));

        mapper.addRule("Server/Service/Engine/Host/Listener",
                       mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Host/Listener",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Host/Listener", mapper.addChild
                       ("addLifecycleListener",
                        "org.apache.catalina.LifecycleListener"));

        mapper.addRule("Server/Service/Engine/Host/Logger", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Host/Logger",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Host/Logger", mapper.addChild
                       ("setLogger", "org.apache.catalina.Logger"));

        mapper.addRule("Server/Service/Engine/Host/Realm", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Host/Realm",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Host/Realm", mapper.addChild
                       ("setRealm", "org.apache.catalina.Realm"));

//         mapper.addRule("Server/Service/Engine/Host/Resources",
//                        mapper.objectCreate
//                        ("org.apache.naming.resources.FileDirContext",
//                         "className"));
//         mapper.addRule("Server/Service/Engine/Host/Resources",
//                        mapper.setProperties());
//         mapper.addRule("Server/Service/Engine/Host/Resources", mapper.addChild
//                        ("setResources", "javax.naming.directory.DirContext"));

        mapper.addRule("Server/Service/Engine/Host/Valve", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Host/Valve",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Host/Valve", mapper.addChild
                       ("addValve", "org.apache.catalina.Valve"));

        mapper.addRule("Server/Service/Engine/Listener", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Listener",
                       mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Listener", mapper.addChild
                       ("addLifecycleListener",
                        "org.apache.catalina.LifecycleListener"));

        mapper.addRule("Server/Service/Engine/Logger", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Logger", mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Logger", mapper.addChild
                       ("setLogger", "org.apache.catalina.Logger"));

        mapper.addRule("Server/Service/Engine/Realm", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Realm", mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Realm", mapper.addChild
                       ("setRealm", "org.apache.catalina.Realm"));

//         mapper.addRule("Server/Service/Engine/Resources", mapper.objectCreate
//                        ("org.apache.naming.resources.FileDirContext",
//                         "className"));
//         mapper.addRule("Server/Service/Engine/Resources",
//                        mapper.setProperties());
//         mapper.addRule("Server/Service/Engine/Resources", mapper.addChild
//                        ("setResources", "javax.naming.directory.DirContext"));

        mapper.addRule("Server/Service/Engine/Valve", mapper.objectCreate
                       (null, "className"));
        mapper.addRule("Server/Service/Engine/Valve", mapper.setProperties());
        mapper.addRule("Server/Service/Engine/Valve", mapper.addChild
                       ("addValve", "org.apache.catalina.Valve"));

        return mapper;
    }

    /**
     * Create the mapper rules for a Context, based on the specified prefix.
     * @param prefix Prefix to rule selectors to be created
     * @param mapper The mapper we are updating
     */
    protected void createStartMapperContext(String prefix, XmlMapper mapper) {

        mapper.addRule(prefix + "", mapper.objectCreate
                       ("org.apache.catalina.core.StandardContext",
                        "className"));
        mapper.addRule(prefix + "", mapper.setProperties());
        mapper.addRule(prefix + "",
                       new CopyParentClassLoaderAction());
        mapper.addRule(prefix + "",
                       new LifecycleListenerAction
                           ("org.apache.catalina.startup.ContextConfig",
                            "configClass"));
        mapper.addRule(prefix + "", mapper.addChild
                       ("addChild", "org.apache.catalina.Container"));

        createContextCommon(prefix, mapper);

    }


    /**
     * Create the mapper rules for a DefaultContext, based on the
     * specified prefix.
     * @param prefix Prefix to rule selectors to be created
     * @param mapper The mapper we are updating
     */
    protected void createStartMapperDefaultContext(String prefix,
        XmlMapper mapper) {

        mapper.addRule(prefix + "", mapper.objectCreate
                       ("org.apache.catalina.core.DefaultContext",
                        "className"));
        mapper.addRule(prefix + "", mapper.setProperties());
        mapper.addRule(prefix + "", mapper.addChild
                       ("addDefaultContext", "org.apache.catalina.core.DefaultContext"));

        createContextCommon(prefix, mapper);

    }


    /**
     * Create the mapper rules for a Context which are common to both
     * a Context and a DefaultContext, based on the specified prefix.
     * @param prefix Prefix to rule selectors to be created
     * @param mapper The mapper we are updating
     */
    protected void createContextCommon(String prefix, XmlMapper mapper) {

//         mapper.addRule(prefix + "/Ejb", mapper.objectCreate
//                        ("org.apache.catalina.deploy.ContextEjb"));
//         mapper.addRule(prefix + "/Ejb",
//                        mapper.setProperties());
//         mapper.addRule(prefix + "/Ejb", mapper.addChild
//                        ("addEjb", "org.apache.catalina.deploy.ContextEjb"));

//         mapper.addRule(prefix + "/Environment",
//                        mapper.objectCreate
//                        ("org.apache.catalina.deploy.ContextEnvironment"));
//         mapper.addRule(prefix + "/Environment",
//                        mapper.setProperties());
//         mapper.addRule(prefix + "/Environment",
//                        mapper.addChild
//                        ("addEnvironment",
//                         "org.apache.catalina.deploy.ContextEnvironment"));

        mapper.addRule(prefix + "/InstanceListener",
                       mapper.methodSetter("addInstanceListener", 0));

        mapper.addRule(prefix + "/Listener",
                       mapper.objectCreate
                       (null, "className"));
        mapper.addRule(prefix + "/Listener",
                       mapper.setProperties());
        mapper.addRule(prefix + "/Listener", mapper.addChild
                       ("addLifecycleListener",
                        "org.apache.catalina.LifecycleListener"));

        mapper.addRule(prefix + "/Loader", new CreateLoaderAction
            ("org.apache.catalina.loader.WebappLoader", "className"));
        mapper.addRule(prefix + "/Loader",
                       mapper.setProperties());
        mapper.addRule(prefix + "/Loader", mapper.addChild
                       ("setLoader", "org.apache.catalina.Loader"));

        mapper.addRule(prefix + "/Logger", mapper.objectCreate
                       (null, "className"));
        mapper.addRule(prefix + "/Logger",
                       mapper.setProperties());
        mapper.addRule(prefix + "/Logger", mapper.addChild
                       ("setLogger", "org.apache.catalina.Logger"));

        mapper.addRule(prefix + "/Manager",
                       mapper.objectCreate
                       ("org.apache.catalina.session.StandardManager",
                        "className"));
        mapper.addRule(prefix + "/Manager",
                       mapper.setProperties());
        mapper.addRule(prefix + "/Manager", mapper.addChild
                       ("setManager", "org.apache.catalina.Manager"));

//         mapper.addRule(prefix + "/Parameter", mapper.objectCreate
//                        ("org.apache.catalina.deploy.ApplicationParameter"));
//         mapper.addRule(prefix + "/Parameter",
//                        mapper.setProperties());
//         mapper.addRule(prefix + "/Parameter", mapper.addChild
//                        ("addApplicationParameter",
//                         "org.apache.catalina.deploy.ApplicationParameter"));

        mapper.addRule(prefix + "/Realm", mapper.objectCreate
                       (null, "className"));
        mapper.addRule(prefix + "/Realm",
                       mapper.setProperties());
        mapper.addRule(prefix + "/Realm", mapper.addChild
                       ("setRealm", "org.apache.catalina.Realm"));

//         mapper.addRule(prefix + "/Resource",
//                        mapper.objectCreate
//                        ("org.apache.catalina.deploy.ContextResource"));
//         mapper.addRule(prefix + "/Resource",
//                        mapper.setProperties());
//         mapper.addRule(prefix + "/Resource",
//                        mapper.addChild
//                        ("addResource",
//                         "org.apache.catalina.deploy.ContextResource"));

//         mapper.addRule(prefix + "/ResourceParams", mapper.objectCreate
//                        ("org.apache.catalina.deploy.ResourceParams"));
//         mapper.addRule(prefix + "/ResourceParams",
//                        mapper.setProperties());
//         mapper.addRule(prefix + "/ResourceParams", mapper.addChild
//                        ("addResourceParams",
//                         "org.apache.catalina.deploy.ResourceParams"));

//         mapper.addRule(prefix + "/ResourceParams/parameter",
//                        mapper.methodSetter("addParameter", 2));
//         mapper.addRule(prefix + "/ResourceParams/parameter/name",
//                        mapper.methodParam(0));
//         mapper.addRule(prefix + "/ResourceParams/parameter/value",
//                        mapper.methodParam(1));

//         mapper.addRule(prefix + "/Resources",
//                        mapper.objectCreate
//                        ("org.apache.naming.resources.FileDirContext",
//                         "className"));
//         mapper.addRule(prefix + "/Resources",
//                        mapper.setProperties());
//         mapper.addRule(prefix + "/Resources", mapper.addChild
//                        ("setResources", "javax.naming.directory.DirContext"));

        mapper.addRule(prefix + "/Valve", mapper.objectCreate
                       (null, "className"));
        mapper.addRule(prefix + "/Valve",
                       mapper.setProperties());
        mapper.addRule(prefix + "/Valve", mapper.addChild
                       ("addValve", "org.apache.catalina.Valve"));

        mapper.addRule(prefix + "/WrapperLifecycle",
                       mapper.methodSetter("addWrapperLifecycle", 0));

        mapper.addRule(prefix + "/WrapperListener",
                       mapper.methodSetter("addWrapperListener", 0));
    }
}



/**
 * Class that copies the parentClassLoader property from the next-to-top
 * item on the stack (which must be a Container) to the top item on the
 * stack (which must also be a Container).
 * @author take from Catalina.
 */
final class CopyParentClassLoaderAction extends XmlAction {

    /**
     * Construct a new action.
     */
    public CopyParentClassLoaderAction() {
        super();
    }

    /**
     * Copy the required class loader.
     * @param context the parsing context.
     * @throws Exception if an error occurs.
     */
    public void start(SaxContext context) throws Exception {
        if (context.getDebug() >= 1) {
            context.log("Copying parent class loader");
        }
        Stack stack = context.getObjectStack();
        Container child = (Container) stack.pop();
        Container parent = (Container) stack.peek();
        stack.push(child);
        child.setParentClassLoader(parent.getParentClassLoader());
    }
}


/**
 * Class that creates a new loader instance, with the parent class
 * loader associated with the top object on the stack (which must be a
 * Container), and pushes it on to the stack.
 * @author take from Catalina.
 */
final class CreateLoaderAction extends XmlAction {

    /**
     * Construct a new action.
     * @param loaderClass see catalina javadoc for more details.
     */
    public CreateLoaderAction(String loaderClass) {
        this(loaderClass, null);
    }


    /**
     * Construct a new action.
     * @param loaderClass see catalina javadoc for more details.
     * @param attributeName see catalina javadoc for more details.
     */
    public CreateLoaderAction(String loaderClass, String attributeName) {
        super();
        this.loaderClass = loaderClass;
        this.attributeName = attributeName;
    }

    /**
     * Classname of the loader.
     */
    protected String loaderClass;

    /**
     * The attribute name of the optional override class (if any).
     */
    protected String attributeName;

    /**
     * Add the requested Loader implemenation.
     * @param context the parsing context.
     * @throws Exception if an error occurs.
     */
    public void start(SaxContext context) throws Exception {
        // Look up the required parent class loader
        Stack stack = context.getObjectStack();
        Container container = (Container) stack.peek();
        ClassLoader parentClassLoader = container.getParentClassLoader();

        // Instantiate a new Loader implementation object
        String className = loaderClass;
        if (attributeName != null) {
            int top = context.getTagCount() - 1;
            AttributeList attributes = context.getAttributeList(top);
            if (attributes.getValue(attributeName) != null) {
                className = attributes.getValue(attributeName);
            }
        }
        Class clazz = Class.forName(className);
        Class[] paramTypes = { ClassLoader.class };
        Object[] arguments = { parentClassLoader };
        Constructor constructor = clazz.getDeclaredConstructor(paramTypes);
        Loader loader = (Loader) constructor.newInstance(arguments);

        // Push the new loader onto the stack
        stack.push(loader);
        if (context.getDebug() >= 1) {
            context.log("new " + loader.getClass().getName());
        }
    }

    /**
     * Pop the loader off the top of the stack.
     * @param context the parsing context.
     */
    public void cleanup(SaxContext context) {
        Stack stack = context.getObjectStack();
        Object o = stack.pop();
        if (context.getDebug() >= 1) {
            context.log("pop " + o.getClass().getName());
        }
    }
}


/**
 * Class that adds a LifecycleListener for the top class on the stack.
 * @author take from Catalina.
 */
final class LifecycleListenerAction extends XmlAction {

    /**
     * Construct a new action.
     * @param listenerClass Name of the listener class to create
     */
    public LifecycleListenerAction(String listenerClass) {
        this(listenerClass, null);
    }

    /**
     * Construct a new action.
     * @param listenerClass Name of the listener class to create
     * @param attributeName Name of an attribute optionally overriding
     * the listener class name
     */
    public LifecycleListenerAction(String listenerClass,
                                   String attributeName) {
        super();
        this.listenerClass = listenerClass;
        this.attributeName = attributeName;

    }

    /**
     * The attribute name of the optional override class (if any).
     */
    private String attributeName = null;

    /**
     * The class name of the listener class to create.
     */
    private String listenerClass = null;

    /**
     * Add the requested lifecycle listener.
     * @param context the parsing context.
     * @throws Exception if an error occurs.
     */
    public void start(SaxContext context) throws Exception {
        // Create a new listener object
        String className = listenerClass;
        if (attributeName != null) {
            int top = context.getTagCount() - 1;
            AttributeList attributes = context.getAttributeList(top);
            if (attributes.getValue(attributeName) != null) {
                className = attributes.getValue(attributeName);
            }
        }
        if (context.getDebug() >= 1) {
            context.log("Add " + className + " listener");
        }
        Class clazz = Class.forName(className);
        LifecycleListener listener =
            (LifecycleListener) clazz.newInstance();

        // Add it to the top object on the stack
        Stack stack = context.getObjectStack();
        Lifecycle top = (Lifecycle) stack.peek();
        top.addLifecycleListener(listener);
    }
}


/**
 * Class that sets the parent class loader for the top class on the stack.
 * @author take from Catalina.
 */
final class SetParentClassLoaderAction extends XmlAction {

    /**
     * Construct a new action.
     * @param parentClassLoader The new parent class loader
     */
    public SetParentClassLoaderAction(ClassLoader parentClassLoader) {
        this.parentClassLoader = parentClassLoader;
    }

    /**
     * The parent class loader to be set.
     */
    private ClassLoader parentClassLoader = null;

    /**
     * Add the requested parent class loader.
     * @param context the parsing context.
     * @throws Exception if an error occurs.
     */
    public void start(SaxContext context) throws Exception {
        if (context.getDebug() >= 1) {
            context.log("Setting parent class loader");
        }

        // Set the parent class loader for this Container
        Stack stack = context.getObjectStack();
        Container top = (Container) stack.peek();
        top.setParentClassLoader(parentClassLoader);
    }
}
