/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Ludovic BERT & Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web.catalina;


//import java
import java.util.Vector;

//import catalina
import org.apache.catalina.Connector;
import org.apache.catalina.Container;
import org.apache.catalina.Engine;
import org.apache.catalina.Host;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.Logger;
import org.apache.catalina.startup.Embedded;

//import jonas
import org.objectweb.jonas.web.JWebContainerServiceException;
import org.objectweb.security.tomcat.interceptor.SecurityCtxInterceptor40;


/**
 * This class provides an implementation of the Embedded Catalina server. 
 * This class is used for the Calatina integration as web container service.
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public class EmbeddedCatalina extends Embedded {

    /**
     * The temporary list of connectors to add in an engine.
     */
    protected Vector connectorList = null;

    /**
     * The JOnAS logger for catalina.
     */
    protected Logger jLogger = null;

    /**
     * Indicates if the embedded catalina use the security of JOnAS.
     */
    protected boolean useSecurity = false;


    /**
     * Construct a new EmbeddedCatalina server (used for the Calatina 
     * integration as web container service).
     */
    public EmbeddedCatalina() {
        super();
        JCatalinaLogger logger = new JCatalinaLogger();
        setLogger(logger);
        this.jLogger = logger;
        this.useSecurity = false;
        connectorList = new Vector();
    }

    /**
     * Construct a new EmbeddedCatalina server (used for the Calatina 
     * integration as web container service).
     * @param logger Logger implementation to be inherited by all components.
     * @param useSecurity true if only if the embedded use the security if no
     * realm has configured for the engine we will use the JOnAS realm.
     */
    public EmbeddedCatalina(Logger logger, boolean useSecurity) {
        super(logger, null);
        setUseNaming(false);
        this.jLogger = logger;
        this.useSecurity = useSecurity;
        connectorList = new Vector();
    }


    /**
     * Add a new Connector to the set of defined Connectors. The newly added 
     * Connector will be associated with the most recently added Engine.
     * @param connector the connector to be added.
     */
    public synchronized void addConnector(Connector connector) {
        connectorList.addElement(connector);
    }

    /**
     * Add a new Engine to the set of defined Engines.
     * @param engine the engine to be added.
     */
    public synchronized void addEngine(Engine engine) {
        super.addEngine(engine);
        if (engine.getLogger() == null) {
            engine.setLogger(jLogger);
        }
        if ((engine.getRealm() == null) && useSecurity) {
            engine.setRealm(new SecurityCtxInterceptor40());
        }        
        for (int i = 0; i < connectorList.size(); i++) {
            Connector connector = (Connector) connectorList.elementAt(i);
            super.addConnector(connector);
        }
        connectorList = new Vector();
    }

    /**
     * Start the embedded catalina.
     * @exception LifecycleException if this component detects a fatal error.
     */
    public void start() throws LifecycleException {
        // Check if there is an engine in the embedded catalina.
        if (engines.length < 1) {
            throw new IllegalStateException 
                (sm.getString("embedded.noEngines"));
        }
        // All is right then start embedded catalina.
        super.start();
    }

    /**
     * Find the specified host.
     * @param hostName the name of the host to find.
     * @return the host found.
     * @throws JWebContainerServiceException if the specified host cannot be
     * found.
     */
    public Host findHost(String hostName) 
        throws JWebContainerServiceException {
        
        if (hostName == null || hostName.equals("")) {
            // no host are specified so get the default host of 
            // the first engine.
            Engine engine = engines[0];
            String defaultHost = engine.getDefaultHost();
            if (defaultHost == null) {
                String err = "Default host must be specified in server.xml or host must be specified in jonas-web.xml";
                throw new JWebContainerServiceException(err);
            }
            Container child = engine.findChild(defaultHost);
            if (child instanceof Host) {
                return (Host) child;
            } else {
                String err = "Default host " + defaultHost + " not found";
                throw new JWebContainerServiceException(err);
            }         
        } else {
            // get the host.
            Vector hosts = new Vector();
            for (int i = 0; i < engines.length; i++) {
                Engine engine = engines[i];
                Container child = engine.findChild(hostName);
                if (child instanceof Host) {
                    hosts.addElement(child);
                }
            }
            if (hosts.size() == 0) {
                // no host found.
                String err = "Host " + hostName + " not found";
                throw new JWebContainerServiceException(err);
            } else {
                return (Host) hosts.elementAt(0);
            }
        } 
    }

    /**
     * Get the temporary connector list. Used for the tests.
     * @return an array of connectors.
     */
    public Connector[] getConnectors() {
        Connector[] ret = new Connector[connectorList.size()];
        return (Connector[]) connectorList.toArray(ret);
    }

    /**
     * Get the engines. Used for the tests.
     * @return an array of engines added in the EmbeddedCatalina.
     */
    public Engine[] getEngines() {
        return engines;
    }

}
