/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web.catalina;


//import java
import java.io.CharArrayWriter;
import java.io.PrintWriter;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;
import javax.servlet.ServletException;

//import jonas
import org.objectweb.jonas.common.Log;

//import monolog
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

//import catalina
import org.apache.catalina.Container;
import org.apache.catalina.LifecycleException;


/**
 * This class provides a JOnAS logger for Catalina.
 * JWebContainerService service.
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public class JCatalinaLogger implements org.apache.catalina.Logger {

    /**
     * Reference on the logger of JOnAS.
     */
    protected Logger logger = null;

    /**
     * The Container with which this Logger has been associated.
     */
    protected Container container = null;

    /**
     * The descriptive information about this implementation.
     */
    protected static final String INFO =
        "org.objectweb.jonas.web.JCatalinaLogger/1.0";

    /**
     * The property change support for this component.
     */
    protected PropertyChangeSupport support = new PropertyChangeSupport(this);

    /**
     * The verbosity level for above which log messages may be filtered.
     */
    protected int verbosity = ERROR;


    /**
     * Contruct a new JCatalinaLogger : JOnAS logger for Catalina.
     */
    public JCatalinaLogger() {
        logger = Log.getLogger(Log.JONAS_WEB_PREFIX);
    }


    /**
     * Return the Container with which this Logger has been associated.
     * @return the Container with which this Logger has been associated.
     */
    public Container getContainer() {
        return container;
    }

    /**
     * Set the Container with which this Logger has been associated.
     * @param container the associated Container.
     */
    public void setContainer(Container container) {
        Container oldContainer = this.container;
        this.container = container;
        support.firePropertyChange("container", oldContainer, this.container);
    }

    /**
     * Return descriptive information about this Logger implementation and
     * the corresponding version number, in the format
     * <code>&lt;description&gt;/&lt;version&gt;</code>.
     * @return a descriptive information about this Logger implementation.
     */
    public String getInfo() {
        return (INFO);
    }

    /**
     * Return the verbosity level of this logger. Messages logged with a
     * higher verbosity than this level will be silently ignored.
     * @return the verbosity level of this logger.
     */
    public int getVerbosity() {
        return (this.verbosity);
    }

    /**
     * Set the verbosity level of this logger. Messages logged with a
     * higher verbosity than this level will be silently ignored.
     * @param verbosity The new verbosity level.
     */
    public void setVerbosity(int verbosity) {
        this.verbosity = verbosity;
    }

    /**
     * Set the verbosity level of this logger. Messages logged with a
     * higher verbosity than this level will be silently ignored.
     * @param verbosity The new verbosity level, as a string.
     */
    public void setVerbosityLevel(String verbosity) {
        if ("FATAL".equalsIgnoreCase(verbosity)) {
            this.verbosity = FATAL;
        } else if ("ERROR".equalsIgnoreCase(verbosity)) {
            this.verbosity = ERROR;
        } else if ("WARNING".equalsIgnoreCase(verbosity)) {
            this.verbosity = WARNING;
        } else if ("INFORMATION".equalsIgnoreCase(verbosity)) {
            this.verbosity = INFORMATION;
        } else if ("DEBUG".equalsIgnoreCase(verbosity)) {
            this.verbosity = DEBUG;
        }
    }

    /**
     * Add a property change listener to this component.
     * @param listener the listener to add.
     */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        support.addPropertyChangeListener(listener);
    }

    /**
     * Writes the specified message to the log.
     * @param msg the message to be written to the log.
     */
    public void log(String msg) {
        logger.log(BasicLevel.DEBUG, msg);
    }

    /**
     * Writes the specified exception, and message, to a servlet log file.
     * The implementation of this method should call
     * <code>log(msg, exception)</code> instead.  This method is deprecated
     * in the ServletContext interface, but not deprecated here to avoid
     * many useless compiler warnings.  This message will be logged
     * unconditionally.
     * @param exception An <code>Exception</code> to be reported.
     * @param msg The associated message string.
     */
    public void log(Exception exception, String msg) {
        log(msg, exception);
    }

    /**
     * Writes an explanatory message and a stack trace for a given
     * <code>Throwable</code> exception to the servlet log file.  The name
     * and type of the servlet log file is specific to the servlet container,
     * usually an event log.  This message will be logged unconditionally.
     * @param msg A <code>String</code> that describes the error or
     * exception.
     * @param throwable The <code>Throwable</code> error or exception.
     */
    public void log(String msg, Throwable throwable) {
        CharArrayWriter buf = new CharArrayWriter();
        PrintWriter writer = new PrintWriter(buf);
        writer.println(msg);
        throwable.printStackTrace(writer);
        Throwable rootCause = null;
        if (throwable instanceof LifecycleException) {
            rootCause = ((LifecycleException) throwable).getThrowable();
        } else if (throwable instanceof ServletException) {
            rootCause = ((ServletException) throwable).getRootCause();
        }
        if (rootCause != null) {
            writer.println("----- Root Cause -----");
            rootCause.printStackTrace(writer);
        }
        log(buf.toString());
    }

    /**
     * Writes the specified message to the servlet log file, usually an event
     * log, if the logger is set to a verbosity level equal to or higher than
     * the specified value for this message.
     * @param message A <code>String</code> specifying the message to be
     * written to the log file.
     * @param verbosity Verbosity level of this message.
     */
    public void log(String message, int verbosity) {
        if (this.verbosity >= verbosity) {
            log(message);
        }
    }


    /**
     * Writes the specified message and exception to the servlet log file,
     * usually an event log, if the logger is set to a verbosity level equal
     * to or higher than the specified value for this message.
     * @param message A <code>String</code> that describes the error or
     * exception.
     * @param throwable The <code>Throwable</code> error or exception.
     * @param verbosity Verbosity level of this message.
     */
    public void log(String message, Throwable throwable, int verbosity) {
        if (this.verbosity >= verbosity) {
            log(message, throwable);
        }
    }

    /**
     * Remove a property change listener from this component.
     * @param listener The listener to remove.
     */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        support.removePropertyChangeListener(listener);
    }

}
