/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Ludovic BERT & Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web.lib;


//import java
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileOutputStream;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.Enumeration;

//import jonas
import org.objectweb.jonas.web.JWebContainerServiceException;


/**
 * This class provides an implementation of tools methods than permit to 
 * unpack a jar.
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public class JarTools {

    /**
     * Size of the buffer.
     */
    private static final int BUFFER_SIZE = 2048;

    /**
     * Unpack the specified jar to the specified directory.
     * @param fileName the name of the archive file to unpack.
     * @param destDir the destination directory of the unpacking.
     * @throws JWebContainerServiceException if an error occors during 
     * the unpacking.
     */
    public static void unpack(String fileName, String destDir) 
        throws JWebContainerServiceException {

        boolean doUnpack = false;
        
        // Check if the file to unpack exists.
        File jarFile = new File(fileName);
        if (!jarFile.exists()) {
            String err = "Cannot unpack " + fileName + " file does not exist";
            throw new JWebContainerServiceException(err);
        }

        // Check if the destination dir exists.
        File dest = new File(destDir);
        if (dest.exists()) {
            // Check if the destination directory is up to date and unpack it
            // if necessary.
            if (dest.lastModified() < jarFile.lastModified()) {
                if (removeDirectory(dest)) {
                    doUnpack = true;
                } else {
                    String err = "Cannot unpack " + fileName;
                    err = err + " cannot delete existing unpacked directory";
                    new JWebContainerServiceException(err);
                }
            } // else nothing to do the dest dir is up to date.
        } else {
            doUnpack = true;
        }

        // Unpack if necessary
        if (doUnpack) {
            JarFile packedJar = null;
            JarEntry entry = null;
            try {
                try {
                    packedJar = new JarFile(fileName);

                    //get entries of the jar file
                    Enumeration entries = packedJar.entries();
                    while (entries.hasMoreElements()) {
                        entry = (JarEntry) entries.nextElement();
                        
                        //File entry
                        File entryFile = new File(destDir, entry.getName());

                        //Create directory
                        if (entry.isDirectory()) {
                            //create parent directories (with mkdirs)
                            entryFile.mkdirs();
                            continue;
                        }
                        
                        //If it's a file, we must extract the file
                        InputStream is = null;
                        try {
                            //get the input stream
                            is = packedJar.getInputStream(entry);
                            
                            //Dump to the file
                            dump(is, entryFile);
                        } finally {
                            is.close();
                        }
                    } 
                } finally {
                    packedJar.close();
                }
            }  catch (IOException e) {
                String err = "Error while unpacking entry " + entry + " : ";
                err = err + e.getMessage();
                throw new JWebContainerServiceException(err);
            }
        }
    }

    /**
     * Write the given input stream in the given file.
     * @param in the inputStream to copy.
     * @param entryFile the file where the inputStream must be dumped.
     * @throws JWebContainerServiceException if the dump failed.
     */
    private static void dump(InputStream in, File entryFile) 
            throws JWebContainerServiceException {
        try {
            //File output
            FileOutputStream out = new FileOutputStream(entryFile);
            int n = 0;
            try {
                //buffer
                byte buffer[] = new byte[BUFFER_SIZE];
                while ((n = in.read(buffer)) > 0) {
                    out.write(buffer, 0, n);
                }
            } finally {
                out.close();
            }
        } catch (IOException e) {
            String err = "Error while unpacking entry " + entryFile + " : ";
            err = err + e.getMessage();
            throw new JWebContainerServiceException(err);
        }
    }

    /**
     * Remove a directory with all its child (recursive remove).
     * @param fileName the name of the file or directory to remove.
     * @return true if only if the directory is removed.
     */
    private static boolean removeDirectory(File file) {
        boolean removeOk = true;
        String fileName = file.getName();

        //File or directory doesn't exists, exit.
        if (!file.exists()) {
            return false;
        }
 
        //Remove the child before the current file(directory)
        if (file.isDirectory()) {
            //remove all the children
            File[] childFiles = file.listFiles();
            for (int i = 0; i < childFiles.length; i++) {
                removeOk = removeOk && removeDirectory(childFiles[i]);
            }
        }
        //Since all childs are removed , remove us
        removeOk = removeOk && file.delete();
        return removeOk;
    }
}
