/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ear.deployment.api;

//import jonas_lib.deployment
import org.objectweb.jonas_lib.deployment.AbsDeploymentDesc;
import org.objectweb.jonas_lib.deployment.SAXParserHelper;

//xml of application.dtd
import org.objectweb.jonas_ear.deployment.xml.Application;
import org.objectweb.jonas_ear.deployment.xml.ApplicationUnmarshaller;
import org.objectweb.jonas_ear.deployment.xml.AltDd;
import org.objectweb.jonas_ear.deployment.xml.Ejb;
import org.objectweb.jonas_ear.deployment.xml.Module;
import org.objectweb.jonas_ear.deployment.xml.SecurityRole;
import org.objectweb.jonas_ear.deployment.xml.Web;

// SAX imports
import org.xml.sax.ErrorHandler;
import org.xml.sax.EntityResolver;

import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

/**
 * This class extends the AbsDeploymentDescriptor class of JOnAS
 * It provides a description of the specific EAR desployment descriptor
 * @author Florent Benoit
 * @author Ludovic Bert
 */

public class EarDeploymentDesc extends AbsDeploymentDesc {

    /** 
     *  The error handler to use
     */
    private static ErrorHandler errorHandler = null;

    /** 
     * The entity resolver to use
     */
    private static EntityResolver entityResolver = null;

    /**
     * Flag for parser validation
     */
    private static boolean parsingWithValidation = true;

    /**
     * Vector of ejbTags (Ejb objects).
     */
    private Vector ejbTags = null;

    /**
     * Vector of altDD for EJBs
     */
    private Vector altDDEjbs = null;

    /**
     * Vector of webTags (Web objects).
     */
    private Vector webTags = null;

    /**
     * Vector of altDD for wars
     */
    private Vector altDDWebs = null;

    /**
     * Vector of security roles
     */
    private Vector securityRoles = null;


    /**
     * Path of the application.xml deploymnet descriptor file
     */
    private static final String APPLICATION_FILE_NAME = "META-INF/application.xml";



    /**
     * Construct an instance of a EarDeploymentDesc.
     * Called by the getInstance() method.
     * @param classLoaderForCls ClassLoader of the classes .
     * @param application application.xml parsed file.
     * @throws EarDeploymentDescException if we can't build an instance.
     */
    private EarDeploymentDesc(ClassLoader classLoaderForCls, Application application) throws EarDeploymentDescException {
 
        ejbTags = new Vector();
        webTags = new Vector();
        altDDEjbs = new Vector();
        altDDWebs = new Vector();
        securityRoles = new Vector();

        // test classloader
        if (classLoaderForCls == null) {
            throw new EarDeploymentDescException("DeploymentDesc: Classloader is null");
        }

        // display name
        displayName = (application.getDisplayName() != null) ? application.getDisplayName().getValue() : null;

        //Tags
        for (Iterator i = application.getModuleList().iterator(); i.hasNext();) {
            Module module = (Module) i.next();
     
            Ejb ejb = module.getEjb();
            Web web = module.getWeb();
            AltDd altDD = module.getAltDd();
            if (ejb != null) {
                ejbTags.add(ejb);
                altDDEjbs.add(altDD);
            } else if (web != null) {
                webTags.add(web);
                altDDWebs.add(altDD);
            }
        }
        for (Iterator i = application.getSecurityRoleList().iterator(); i.hasNext();) {
            SecurityRole securityRole = (SecurityRole) i.next();
            if (securityRole != null) {
                securityRoles.add(securityRole);
            }
        }
    }
    
    /**
     * Get the ejb tags of the application.xml file.
     * @return the ejb tags of the application.xml file.
     */
    public Ejb[] getEjbTags() {
        Ejb[] tmp = new Ejb[ejbTags.size()];
        ejbTags.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the alt-dd of the ejbs of the application.xml file.
     * @return the alt-dd of the ejbs of the application.xml file.
     */
    public AltDd[] getAltDDEjbs() {
        AltDd[] tmp = new AltDd[altDDEjbs.size()];
        altDDEjbs.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the alt-dd of the wars of the application.xml file.
     * @return the alt-dd of the wars of the application.xml file.
     */
    public AltDd[] getAltDDWebs() {
        AltDd[] tmp = new AltDd[altDDWebs.size()];
        altDDWebs.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the security-role tags
     * @return the security roles.
     */
    public SecurityRole[] getSecurityRoles() {
        SecurityRole[] tmp = new SecurityRole[securityRoles.size()];
        securityRoles.copyInto(tmp);
        return tmp;
    }

    /**
     * Get an instance of an EAR deployment descriptor by parsing the 
     * application.xml deployment descriptor.
     * @param earFileName the fileName of the war file for the deployment descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @return an EAR deployment descriptor by parsing the application.xml 
     * deployment descriptor.
     * @throws EarDeploymentDescException if the deployment descriptor is 
     * corrupted.
     */
    public static EarDeploymentDesc getInstance(String earFileName,
                                                ClassLoader classLoaderForCls)
        throws EarDeploymentDescException {

        //ear file
        JarFile earFile = null;

        //Input Stream
        InputStream appInputStream = null;

        //ZipEntry
        ZipEntry appZipEntry = null;

        Application application;
 
        //Check if the file exists.
        if (!new File(earFileName).exists()) {
            throw new EarDeploymentDescException("The file '" + earFileName + "' was not found.");
        }

        // load deployment descriptor data (META-INF/application.xml)
        try {
            earFile = new JarFile(earFileName);;

            //Check the application entry
            appZipEntry = earFile.getEntry(APPLICATION_FILE_NAME);
            if (appZipEntry == null) {
                throw new EarDeploymentDescException("The entry '" + APPLICATION_FILE_NAME + "' was not found in the file '" + earFileName + "'.");
            }
            //Get the stream
            appInputStream = earFile.getInputStream(appZipEntry);

        } catch (Exception e) {
            if (earFile != null) {
                try {
                    earFile.close();
                } catch (IOException ioe) {
                    //We can't close the file
                }
            }
            throw new EarDeploymentDescException("Cannot read the XML deployment descriptors of the ear file '" + earFileName + "'.", e);
        }

        application = loadApplication(appInputStream, APPLICATION_FILE_NAME);
        try {
            appInputStream.close();
        } catch (IOException e) {
        }

        // instantiate deployment descriptor
        return new EarDeploymentDesc(classLoaderForCls, application);
    }

    /**
     * Get the web tags of the application.xml file.
     * @return the web tags of the application.xml file.
     */
    public Web[] getWebTags() {
        Web[] tmp = new Web[webTags.size()];
        webTags.copyInto(tmp);
        return tmp;
    }

    /**
     * Load the application.xml file.
     * @param is the InputStream of the XML file.
     * @param fileName the name of the file (application.xml).
     * @throws EarDeploymentDescException if the deployment descriptor is 
     * corrupted.
     * @return an application object.
     */
    private static Application loadApplication(InputStream is, String fileName)
        throws EarDeploymentDescException {

        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper = new SAXParserHelper(errorHandler, entityResolver);

        try {
            ApplicationUnmarshaller.setErrorHandler(helper.getWrappedErrorHandler());
            ApplicationUnmarshaller.setEntityResolver(helper.getWrappedEntityResolver());

            // Unmarshal the XML, and validate as it is unmarshalled
            return ApplicationUnmarshaller.unmarshal(reader, parsingWithValidation);
        } catch (IOException e) {
            if (helper.getSAXParseException() != null) {
                String detailMsg = getSAXMsg(fileName, helper.getSAXParseException(), "Invalid XML");
                throw new EarDeploymentDescException(detailMsg, e);
            } else {
                throw new EarDeploymentDescException("Error reading XML document " + fileName, e);
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignored) { }
            }
        }
    }

    /** 
     * Return a String representation of the EarDeploymentDesc.
     * @return a String representation of the EarDeploymentDesc.
     */
    public String toString() {
        String ret = "Ejbs :";
        for (Enumeration e = ejbTags.elements(); e.hasMoreElements();) {
            ret += "\t" + ((Ejb) e.nextElement()).getValue();
        }

        ret += "\nWebs :";
        for (Enumeration e = webTags.elements(); e.hasMoreElements();) {
            ret += "\t" + ((Web) e.nextElement()).getWebUri().getValue();
        }

        return ret;
    }

    /**
     * Get the current ErrorHandler.
     * @return the current ErrorHandler.
     */
    public static ErrorHandler getErrorHandler() {
        return errorHandler;
    }

    /**
     * Get the current EntityResolver.
     * @return the current EntityResolver.
     */
    public static EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @return if true, all external entities will be read.
     */
    public static boolean getParsingWithValidation() {
        return EarDeploymentDesc.parsingWithValidation;
    }

    /**
     * Set the specified ErrorHandler. If the application does not register
     * an ErrorHandler, all error events reported by the reader are ignored.
     * @param errorHandler the Error Handler to set.
     */
    public static void setErrorHandler(ErrorHandler errorHandler) {
        EarDeploymentDesc.errorHandler = errorHandler;
    }

    /**
     * Set the specified EntityResolver.
     * @param entityResolver the Entity Resolver to set.
     */
    public static void setEntityResolver(EntityResolver entityResolver) {
        EarDeploymentDesc.entityResolver = entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @param validation if true, all external entities will be read.
     */
    public static void setParsingWithValidation(boolean validation) {
        EarDeploymentDesc.parsingWithValidation = validation;
    }



}
