/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *

 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.security.Guard;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Enumeration;
import java.util.Set;
import java.util.HashSet;
import java.util.Vector;
import java.util.Properties;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;
import javax.ejb.AccessLocalException;
import javax.ejb.EJBException;
import javax.naming.Context;
import javax.naming.LinkRef;
import javax.naming.StringRefAddr;
import javax.naming.NamingException;
import javax.transaction.NotSupportedException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.TransactionRequiredException;

import org.objectweb.common.Env;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbLocalRefDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbRefDesc;
import org.objectweb.jonas_ejb.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_ejb.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_ejb.deployment.api.ResourceRefDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatefulDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityBmpDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmpDesc;
import org.objectweb.jonas_ejb.deployment.api.MessageDrivenDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp1Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp2Desc;
import org.objectweb.jonas_jms.api.JmsManager;
import org.objectweb.jonas_lib.naming.ContainerNaming;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class represents an EJB container.
 *
 * A container is where an enterprise Bean object lives.
 * All beans from a same ejb-jar file are installed in a single container.
 * For each EJB installed, the container provides a factory and
 * makes it available in the JNDI name space.
 * The JContainer basically manages a set of BeanFactory objects.
 * @author: Philippe Coq
 * Contributor(s):
 *   Jeff Mesnil     : Security
 *   Christophe Ney  : for making easier Enhydra integration
 *   Philippe Durieux: New architecture for local interfaces
 *   Florent Benoit & Ludovic Bert : ear service, ejb-link treatment
 */

public class JContainer implements Container {

    private Swapper swapper;
    private ClassLoader loader = null;
    private ContainerNaming naming = null;
    private TransactionManager tm = null;
    private JmsManager jms = null;
    private DeploymentDesc deploymentDesc = null;

    public String myname; // Container name
    public String fileName; // Container file name x.jar or x.xml
    private HashMap beanList = new HashMap(); // key=EJBName / Value=BeanFactory
    private Guard methodGuard = null;
    private Guard roleGuard = null;
    private PrincipalFactory principalFactory = null;

    // Name of the ear application containing this container.
    private String earFileName = null;

    public JStatefulPool statefulSessionPool = null;

    // Pool of threads for Message Driven Beans
    // We use only 1 pool for all containers.
    private static final int DEFAULT_THREAD_POOL_SIZE = 10;
    int threadpoolsize = DEFAULT_THREAD_POOL_SIZE;
    static ThreadPool threadpool = null;

    /**
     * constructor
     */
    public JContainer(String name, String file, ClassLoader ld) {
	myname = name;
	fileName = file;
	loader = ld;
	swapper = new Swapper(this);
	swapper.start();
    }

    // ---------------------------------------------------------------
    // Container Implementation
    // ---------------------------------------------------------------

    /**
     * Return an array of bean names
     */
    public String [] listBeanNames() {

	String ret[] = new String[beanList.size()];
	Vector namelist = new Vector();
	BeanFactory bf = null;
        return (String[]) beanList.keySet().toArray(new String[0]);
    }

    /**
     * Return the name of the container
     */
    public String getName() {
	return myname;
    }

    /**
     * Return the file name of the container (.xml or .jar)
     */
    public String getFileName() {
	return fileName;
    }

    /**
     * Set a MethodGuard
     */
    public void setMethodGuard(Guard methodGuard) {
	if (TraceEjb.isDebugSecurity())
	    TraceEjb.debugSecurity(myname+": setMethodGuard("+methodGuard+")");
	this.methodGuard = methodGuard;
    }

    /**
     * Get the MethodGuard
     */
    public Guard getMethodGuard() {
	return methodGuard;
    }

    /**
     * Set the RoleGuard
     */
    public void setRoleGuard (Guard roleGuard) {
	if (TraceEjb.isDebugSecurity())
	    TraceEjb.debugSecurity(myname+": setRoleGuard("+roleGuard+")");
	this.roleGuard = roleGuard;
    }

    /**
     * Get the RoleGuard
     */
    public Guard getRoleGuard() {
	return roleGuard;
    }

    /**
     * Returns the classloader used for this Container
     */
    public ClassLoader getClassLoader() {
	if (loader == null) {
	    return Thread.currentThread().getContextClassLoader();
	} else {
	    return loader;
	}
    }

    /**
     * Set the PrincipalFactory
     */
    public void setPrincipalFactory(PrincipalFactory pf) {
	principalFactory = pf;
    }

    /**
     * Get the PrincipalFactory
     */
    public PrincipalFactory getPrincipalFactory() {
	return principalFactory;
    }

    /**
     * set the ContainerNaming object
     * Called by the EJB Server when starting the service.
     */
   public void setContainerNaming(ContainerNaming naming) {
	this.naming = naming;
   }

    /**
     * return the ContainerNaming object
     * used internally by all the EJB Container classes.
     */
    public ContainerNaming getContainerNaming() {
	return naming;
    }

    /**
     * set the Thread pool size.
     */
    public void setThreadPoolSize(int sz) {
	threadpoolsize = sz;
    }

    /**
     * Try to passivate all entity bean instances
     */
    public void syncAll(boolean passivate) {
        TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": syncAll");
        BeanFactory bf = null;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
            if (passivate) {
                bf.reduceCache();
            }
            else {
                bf.sync();
            }
        }
    }

    /**
     * set the Transaction Manager.
     * Called by the EJB Server when starting the service.
     */
    public void setTransactionManager(TransactionManager tm) {
	this.tm = tm;
    }

    /**
     * return the Transaction Manager
     * used internally by all the EJB Container classes.
     */
    public TransactionManager getTransactionManager() {
	return tm;
    }

    /**
     * set the JmsManager object
     * Called by the EJB Server when starting ths service.
     */
   public void setJmsManager(JmsManager jms) {
	this.jms = jms;
   }

    /**
     * return the JmsManager object
     * used internally by all the EJB Container classes.
     */
    public JmsManager getJmsManager() {
	return jms;
    }

    /**
     * set the name of the ear application containing this container.
     * @param fileName the name of the ear application containing this
     * container.
     */
    public void setEarFileName(String fileName) {
	earFileName = fileName;
    }

    /**
     * get the name of the ear application containing this container.
     * @param fileName the name of the ear application containing this
     * container.
     */
    public String getEarFileName() {
	return earFileName;
    }

    /**
     * Remove the JOnAS container.
     */
    public void remove() {
	TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": remove");

	// Stop the swapper.
	swapper.stopIt();

	// synchronize all entity beans before removing them.
	// must be done here, not in the swapper.
	syncAll(false);

	// remove all factories
	Iterator it = beanList.values().iterator();
	while (it.hasNext()) {
	    BeanFactory bf = (BeanFactory) it.next();
	    try {
		bf.stop();
		System.out.println(myname+": "+ bf.getEJBName() + " no longer available");
	    } catch (Exception e) {
		TraceEjb.error(myname+": remove:", e);
	    }
	}
	beanList.clear();

	loader = null;
    }


    // -------------------------------------------------------------------
    // Methods used by JOnAS Server (but not part of JContainer interface)
    // -------------------------------------------------------------------

    /**
     * Set DeploymentDesc. Needed for ejb-link in ejb-ref and ejb-local-ref.
     * LATER: May be better to replace addBean by this.
     */
    public void setDeploymentDesc(DeploymentDesc dd) {
        this.deploymentDesc = dd;
    }

    public DeploymentDesc getDeploymentDesc() {
        return deploymentDesc;
    }
    /**
     * Adds beans in container.
     * This method is not part of JContainer interface, although it
     * is used by JOnAS Server, because we don't want to get all BeanDesc
     * classes (jonas_ejb.deployment.api)
     * @param desc  The Bean Deployment Descriptor
     * @throw EJBException
     */
    public BeanFactory addBean(BeanDesc dd) {
	BeanFactory bf = null;
	String beanName = dd.getEjbName();
	if (dd instanceof SessionStatefulDesc) {
	    TraceEjb.interp.log(BasicLevel.DEBUG, "add SessionStatefulBean " + beanName);
	    if (statefulSessionPool == null) {
		statefulSessionPool = new JStatefulPool();
	    }
	    bf = new JStatefulFactory((SessionStatefulDesc)dd, this, statefulSessionPool);
	} else if (dd instanceof SessionStatelessDesc) {
	    TraceEjb.interp.log(BasicLevel.DEBUG, "add SessionStatelessBean " + beanName);
	    bf = new JStatelessFactory((SessionStatelessDesc)dd, this);
	} else if (dd instanceof MessageDrivenDesc) {
	    TraceEjb.interp.log(BasicLevel.DEBUG, "add MessageDrivenBean " + beanName);
	    if (threadpool == null) {
		threadpool = new ThreadPool(threadpoolsize);
	    }
	    bf = new JMdbFactory((MessageDrivenDesc)dd, this, threadpool);
	} else if (dd instanceof EntityDesc) {
	    TraceEjb.interp.log(BasicLevel.DEBUG, "add EntityBean " + beanName);
        //Instanciate the factory specified in the EntityDesc
		try {
			bf = (JEntityFactory) Class.forName(
                ((EntityDesc)dd).getFactoryClassName() ).newInstance();
		}
		catch (ClassNotFoundException e1) {
            try {
                bf = (JEntityFactory) loader.loadClass(
                    ((EntityDesc)dd).getFactoryClassName()).newInstance();
            }
            catch (Exception e) {
                throw new EJBException(
                    "Impossible to instanciate the specified entity factory: "
                    + ((EntityDesc)dd).getFactoryClassName(), e);
            }
        }
		catch (Exception e) {
            throw new EJBException(
                "Impossible to instanciate the specified entity factory: "
				+ ((EntityDesc)dd).getFactoryClassName(), e);
		}
	    ((JEntityFactory)bf).init((EntityDesc)dd, this);
	    // Set the passivation timeout
	    setSwapTime(((EntityDesc)dd).getPassivationTimeout());
	} else {
	    throw new EJBException("Bad Descriptor Type for "+beanName);
	}
	beanList.put(beanName, bf);
	System.out.println(beanName + " available");
	return bf;
    }

    public BeanFactory getBeanFactory(String ejbName) {
        return (BeanFactory) beanList.get(ejbName);
    }

    /**
     * @return total Number of Beans
     */
    public int getBeanNb() {
	return beanList.size();
    }

    /**
     * @return int Number of BMP type currently in this container
     */
    public int getEntityBMPNb() {
	BeanFactory bf;
	int total = 0;
	Iterator it = beanList.values().iterator();
	while (it.hasNext()) {
	    bf = (BeanFactory) it.next();
	    if (bf.getDeploymentDescriptor() instanceof EntityBmpDesc) total++;
	}
	return total;
    }

    /**
     * @return int Number of CMP type currently in this container
     */
    public int getEntityCMPNb() {
	BeanFactory bf;
	int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
	    if (bf.getDeploymentDescriptor() instanceof EntityCmpDesc) total++;
	}
	return total;
    }

    /**
     * @return int Number of SBF type currently in this container
     */
    public int getStatefulSessionNb() {
	BeanFactory bf;
	int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
	    if (bf.getDeploymentDescriptor() instanceof SessionStatefulDesc) total++;
	}
	return total;
    }

    /**
     * @return int Number of SBL type currently in this container
     */
    public int getStatelessSessionNb() {
	BeanFactory bf;
	int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
	    if (bf.getDeploymentDescriptor() instanceof SessionStatelessDesc) total++;
	}
	return total;
    }

    /**
     * @return int Number of MDB type currently in this container
     */
    public int getMessageDrivenNb() {
	BeanFactory bf;
	int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
	    if (bf.getDeploymentDescriptor() instanceof MessageDrivenDesc) total++;
	}
	return total;
    }

    /**
     * @return passivation time out for this container
     */
    public long getSwapTime() {
	return swapper.getTimeout();
    }

    /**
     * Management method used by the EJBServiceImpl MBean. Determine
     * which are the ejbs using a given data source.
     * @param dsName JNDI name of the data source
     * @ret a set of Properties describing the beans that use the data source
     */
    public Set getDataSourceDependence(String dsName) {
        // the true param corresponds to a resource-ref
        return beansDependence(
            Collections.enumeration(beanList.values()), dsName, true);
    }

    /**
     * Management method used by the EJBServiceImpl MBean. Determine
     * which are the beans using a JMS destination.
     * @param destName JNDI name of a JMS destination
     * @ret a set of Properties describing the ejbs that use the
     * JMS destination.
     */
    public Set getJmsDestinationDependence(String destName) {
        // the true param corresponds to a resource-ref
        return beansDependence(
            Collections.enumeration(beanList.values()), destName, false);
    }

    /**
     * Management method used by the EJBServiceImpl MBean. Determine
     * which are the ejbs using a JMS Connection Factory.
     * @param cfName JNDI name of a JMS Connection Factory.
     * @ret a set of Properties describing the beans that use the
     * JMS Connection Factory.
     */
    public Set getJmsConnectionFactoryDependence(String cfName) {
        // the true param corresponds to a resource-ref
        return beansDependence(
            Collections.enumeration(beanList.values()), cfName, true);
    }

    /**
     * Management method used by the EJBServiceImpl MBean. Determine
     * which are the ejbs using a Mail Factory.
     * @param mfName JNDI name of a Mail Factory.
     * @ret a set of Properties describing the beans that use the
     * given Mail Factory.
     */
    public Set getMailFactoryDependence(String mfName) {
        // the true param corresponds to a resource-ref
        return beansDependence(
            Collections.enumeration(beanList.values()), mfName, true);
    }

    /**
     * Used by the above getXXXDependence() methods.
     * @param enum the bean factories managed by this container.
     * @param rName jndi name of a resource (datasource, JMS destination or
     * connection factory, mail factory).
     * @param isResRef true for resource-ref, false for ressource-env-ref.
     * @return set of Properties, where each Properties provides the bean
     * name, the container's file-name and the type of the bean which uses
     * the resource.
     */
    private Set beansDependence(Enumeration enum, String rName, boolean isResRef) {
	HashSet result = new HashSet();
	BeanFactory bf = null;
	while (enum.hasMoreElements()) {
	    // for each bean,
	    bf = (BeanFactory) enum.nextElement();
	    BeanDesc ejbDesc = bf.getDeploymentDescriptor();

	    boolean isDependent = false; // suppose its not dependent

	    // get its type from the DeploymentDescriptor
	    String ejbType = null;
	    if (bf instanceof JEntityFactory) {
		if (ejbDesc instanceof EntityBmpDesc) ejbType = "ejbbmp";
		if (ejbDesc instanceof EntityCmpDesc ) ejbType = "ejbcmp";
	    } else if (bf instanceof JStatefulFactory) {
		ejbType = "ejbsbf";
	    } else if (bf instanceof JStatelessFactory) {
		ejbType = "ejbsbl";
	    } else if (bf instanceof JMdbFactory) {
		ejbType = "ejbmdb";
	    }

	    if (isResRef) {
		// if this is a cmp bean, directly compare rName with the data source JNDI name
		// from its deployment descriptor
		if (ejbType.equals("ejbcmp")) {
            String jndiName = null;
            if (ejbDesc instanceof EntityJdbcCmp1Desc)
                jndiName = ((EntityJdbcCmp1Desc)ejbDesc).getDatasourceJndiName();
            else if (ejbDesc instanceof EntityJdbcCmp2Desc)
                jndiName = ((EntityJdbcCmp2Desc)ejbDesc).getDatasourceJndiName();
			isDependent = rName.equals(jndiName);
		}
		// get the description of the bean's resource-refs
		ResourceRefDesc[] rrDesc = ejbDesc.getResourceRefDesc();
		for (int i=0 ; i<rrDesc.length ; i++) {
		    // for each resource-ref compare its name with rName
		    if (rrDesc[i].getJndiName().equals(rName)) {
			isDependent = true;
			break;
		    }
		}
	    } else {
		if (ejbType.equals("ejbmdb")) {
		    if (rName.equals(((MessageDrivenDesc)ejbDesc).getDestinationJndiName()))
			isDependent = true;
		}
		// get the description of the bean's  resource-env-refs
		ResourceEnvRefDesc[] rerDesc = ejbDesc.getResourceEnvRefDesc();
		for (int i=0 ; i<rerDesc.length ; i++) {
		    // for each resource-env-ref compare its name with rName
		    if (rerDesc[i].getJndiName().equals(rName)) {
			isDependent = true;
			break;
		    }
		}
	    }

	    if (isDependent) {
		Properties toAdd = new Properties();
		toAdd.setProperty("type", ejbType);
		toAdd.setProperty("fname", getFileName());
		toAdd.setProperty("name", ejbDesc.getEjbName());
		result.add(toAdd);
	    }

	}
	return result;
    }

    // -------------------------------------------------------------------
    // Other public methods, used internally in this package.
    // -------------------------------------------------------------------

    /**
     * register an Entity that must be passivated
     */
    public void registerEJB(JEntitySwitch ejb) {
	TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": registerEJB");

	swapper.addEntity(ejb);
    }

    /**
     * register a BeanFactory
     */
    public void registerBF(BeanFactory bf) {
	TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": registerBF");

	swapper.addBeanFactory(bf);
    }

    /**
     * Take into account the swapping time for the bean.
     * @args time in seconds  (time = 0 no time out)
     */
    public void setSwapTime(int t) {
	TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": setSwapTime "+t);
	if (t > 0) {
	    swapper.setTimeout(1000*t);
	}
    }

    /**
     * Set the bean environment
     * args ctx Context for this bean
     * args dd Bean Deployment Descriptor
     * throws NamingException if could not rebind objects
     */
    public void setBeanEnvironment(Context ctx, BeanDesc dd) throws NamingException {

	// Set bean context
	Context ctxold = naming.setComponentContext(ctx);

	// Bean Environment
	EnvEntryDesc [] envt = dd.getEnvEntryDesc();
	for (int i = 0; i < envt.length; i++) {
	    // get information in descriptor
	    String name = envt[i].getName();
	    Object obj = envt[i].getValue();
	    // register object in JNDI
	    if (TraceEjb.isDebugIc())
		TraceEjb.interp.log(BasicLevel.DEBUG, myname+": Binding object "+name+" -> "+obj);
	    ctx.rebind(name, obj);
	}

	// Resource References
	ResourceRefDesc [] resref = dd.getResourceRefDesc();
	for (int i = 0; i < resref.length; i++) {
	    // get information in descriptor
	    String name = resref[i].getName();
	    String resname = resref[i].getJndiName();
	    // build the LinkRef that will be registered:
	    // FactoryClassName = null, size = 1, refAddr = resname.
	    LinkRef lref = new LinkRef(resname);

	    // if we are using David add a parameter to the link to tell that the
	    // object must be found in rmiregistry (and not in CosNaming)
	    // this info is used by CompNamingContext.lookup()
	    if (Env.isDavid()) {
		lref.add(new StringRefAddr("TargetNaming", "RmiRegistry"));
	    }

	    // register object in JNDI
	    if (TraceEjb.isDebugIc())
		TraceEjb.interp.log(BasicLevel.DEBUG, myname+": Linking resource "+name+" -> "+resname);
	    ctx.rebind(name, lref);
	}

	// Resource Environment References
	ResourceEnvRefDesc [] resEnvref = dd.getResourceEnvRefDesc();
	for (int i = 0; i < resEnvref.length; i++) {
	    // get information in descriptor
	    String name = resEnvref[i].getName();
	    String resname = resEnvref[i].getJndiName();
	    LinkRef lref = new LinkRef(resname);

	    // if we are using David add a parameter to the link to tell that the
	    // object must be found in rmiregistry (and not in CosNaming)
	    // this info is used by CompNamingContext.lookup()
	    if (Env.isDavid()) {
		lref.add(new StringRefAddr("TargetNaming", "RmiRegistry"));
	    }

	    if (TraceEjb.isDebugIc())
		TraceEjb.interp.log(BasicLevel.DEBUG, myname+": Linking resource environment "+name
				     +" -> "+resname);
	    ctx.rebind(name, lref);
	}

	// EJB References
	EjbRefDesc [] ejbref = dd.getEjbRefDesc();
	for (int i = 0; i < ejbref.length; i++) {
	    // get information in descriptor
	    String name = ejbref[i].getName();
	    String ejbname = null;
	    String ejblink = ejbref[i].getEjbLink();
            ejbname = ejbref[i].getJndiName();
	    LinkRef lref = new LinkRef(ejbname);

	    // if we are using David add a parameter to the link to tell that the
	    // object must be found in the CosNaming (and not in rmiregistry)
	    // this info is used by CompNamingContext.lookup()
	    if (Env.isDavid()) {
		lref.add(new StringRefAddr("TargetNaming", "CosNaming"));
	    }

	    if (TraceEjb.isDebugIc())
		TraceEjb.interp.log(BasicLevel.DEBUG, myname+": Linking ejb "+name+" -> "+ejbname);
	    ctx.rebind(name, lref);
	}

	// EJB Local Refs
	// We use here ejb-link tag. This should be used also for ejb-ref when
	// we are able to manage references to another jar file.
	EjbLocalRefDesc [] ejblocalref = dd.getEjbLocalRefDesc();
	for (int i = 0; i < ejblocalref.length; i++) {
	    String name = ejblocalref[i].getName();
	    String ejblink = ejblocalref[i].getEjbLink();
	    String ejbname = ejblocalref[i].getJndiLocalName();
	    LinkRef lref = new LinkRef(ejbname);
	    if (TraceEjb.isDebugIc())
		TraceEjb.interp.log(BasicLevel.DEBUG, myname+": Linking ejb "+name+" -> "+ejbname);
	    ctx.rebind(name, lref);
	}

	// Reset bean context
	naming.setComponentContext(ctxold);
    }

    /**
     * Check Security. No control for Message Driven Beans
     */
    protected void checkSecurity(String methodSignature) {
	// A "null" string means : no security control.
	if (methodSignature.equals("null")) {
	    return;
	}
	// methodGuard == null : no security service available
	if (methodGuard == null) {
	    TraceEjb.debugSecurity("JContainer.checkSecurity:  no security service available");
	    return;
	}
	TraceEjb.debugSecurity("JContainer.checkSecurity()");
	try {
	    methodGuard.checkGuard(methodSignature);
	} catch (SecurityException e) {
	    throw new AccessLocalException("Access Denied");
	}
    }

    // -------------------------------------------------------------------
    // Private methods
    // -------------------------------------------------------------------

}
