/**
 * copyright area
 */
package org.objectweb.jonas_ejb.container.jorm;

import org.objectweb.jonas_ejb.container.JContainer;
import org.objectweb.jonas_ejb.container.TraceEjb;
import org.objectweb.jonas_ejb.deployment.api.DeploymentCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.lib.EjbqlQueryTreeHolder;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.eval.api.ConnectionResources;
import org.objectweb.medor.eval.api.QueryEvaluator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.optim.api.ExecPlanGenerator;
import org.objectweb.medor.optim.api.QueryTransformer;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.IndexesGenerator;
import org.objectweb.medor.optim.jorm.JormQueryRewriter;
import org.objectweb.medor.optim.jorm.RdbJormLeafRewriter;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.ejb.EJBException;
import java.io.IOException;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import java.util.Collection;

/**
 * This class does the initialisation of Medor and permits to access to the
 * query. The optimisation of the query tree is done during the first time that
 * it is used.
 * @author S.Chassande-Barrioz
 */
public abstract class MedorFactory extends JormFactory {
    /**
     * This field references the query transformer which must be used to
     * optimize the medor requests.
     */
    protected QueryTransformer queryTransformer = null;

    protected Manager manager = null;

    protected ExecPlanGenerator indexesGenerator;

    protected boolean optimizeAtInit = false;


    protected static final Collection leafRewriters =
        Collections.singleton(new RdbJormLeafRewriter());

    public MedorFactory() {
        super();
    }

    /**
     * It retrieves a medor request which is evaluable and optimized. The method
     * index is translate into a request index, then the found request is
     * optimized if it is not already.
     *
     * @param methodIndex is the index of the finder or select method.
     * @return the QueryTree optimized which is associated to the method index
     */
    public QueryEvaluator getOptimizedRequest(int methodIndex) throws MedorException {
        MethodCmp2Desc mcd = (MethodCmp2Desc) dd.getMethodDesc(methodIndex);
		TraceEjb.logger.log(BasicLevel.DEBUG, "getOptimizedRequest("
			+ methodIndex + ") method name: " + mcd.getMethod().getName());
        try {
            EjbqlQueryTreeHolder qth = mcd.getQueryTreeHolder();
            // set the optimizer if not set yet
            if (qth.getQueryOptimizer() == null) {
                if (queryTransformer == null) {
                    // This map contains the association between the jorm class
                    // names and the mapper.
                    Map name2mapper = new HashMap();
                    BeanDesc[] bds = dd.getDeploymentDesc().getBeanDesc();
                    for (int i = 0; i<bds.length; i++) {
                        if (bds[i] instanceof EntityCmp2Desc) {
                            String cn = ((EntityCmp2Desc) bds[i]).getAbstractSchemaName();
//TODO: All jorm classes are not mapped on the same mapper that the current class
// The EntityCmp2Desc must provide the mapper name:
// MapperManager.getMapper( ((EntityCmp2Desc) bds[i]).getMapperName() );
                            PMapper mapper = getMapper();
                            name2mapper.put(cn,mapper);
                            if (TraceEjb.logger.isLoggable(BasicLevel.DEBUG))
                                TraceEjb.logger.log(BasicLevel.DEBUG, "Class "
                                    + cn + " is mapped on the mapper "
                                    + mapper.getMapperName());
                        }
                    }
                    queryTransformer = new JormQueryRewriter(
                        name2mapper,leafRewriters, TraceEjb.logger);
                }
                qth.setQueryOptimizer(new QueryTransformer() {
                    public QueryTree transform(QueryTree qt) throws MedorException {
                        return indexesGenerator.transform(queryTransformer.transform(qt));
                    }
                });
            }
            return qth.getOptimizedQueryTree();
        }
        catch (Exception e) {
            throw new MedorException("Impossible to optimize the query " + mcd.getQuery(), e);
        }
    }

    /**
     * It optimized all medor requests which are already not.
     */
    public void optimizeAll() throws MedorException {
        for (Iterator it = dd.getMethodDescIterator(); it.hasNext();) {
            MethodCmp2Desc mcd = (MethodCmp2Desc) it.next();
            if (!mcd.isFindByPrimaryKey() && (mcd.isFinder() || mcd.isSelect()))
                getOptimizedRequest(mcd.getIndex());
        }
    }

    /**
     * It evaluate an optimized medor request according to the specified parameters
     * To evalute the medor request the query evaluator is used.
     * @param methodIndex is method index which permits to find a medor request.
     * @param parameters is the parameters
     * (key=parameter name  /  value=parameter value)
     * @return TupleCollection is the result of the request
     */
    public TupleCollection evaluate(int methodIndex, ParameterOperand[] parameters)
        throws MedorException {
        QueryEvaluator evaluator = getOptimizedRequest(methodIndex);

        // Calculates and gets the required connection ressources for this query
        ConnectionResources cr = evaluator.getRequiredConnectionResources();

        // Gets the QueryLeafs that requierd connections
        QueryLeaf[] leaves = cr.getRequiredQueryLeafConnection();

        // Setting QueryLeaf's appropriated connection Object
        Object conn = null;
        try {
            conn = getConnection(null);
        }
        catch (PException e) {
            throw new MedorException(
                "impossible to obtain connection to the data support", e);
        }
        for (int i = 0; (i < leaves.length); i++) {
            cr.setConnection(leaves[i], conn);
        }
		try {
			return evaluator.evaluate(parameters, cr);
		} catch (EvaluationException e) {
			throw e;
		} finally {
			try {
				if (conn != null)
					releaseConnection(conn);
			} catch (PException e) {
				TraceEjb.error("Impossible to close a connection", e);
			}
		}
	}

    // OVERRIDE JEntityFactory CLASS //
    //-------------------------------//

    public void init(EntityDesc ed, JContainer c) {
        TraceEjb.logger.log(BasicLevel.DEBUG, "MedorFactory.init");
        super.init(ed, c);

        EntityCmp2Desc ecd = (EntityCmp2Desc) ed;

        //------------------- JORM META INFORMATION BUILDING ----------------//
        TraceEjb.logger.log(BasicLevel.DEBUG, "Jorm Meta information building");
        DeploymentCmp2Desc dd = ((DeploymentCmp2Desc) c.getDeploymentDesc());
        try {
            manager = dd.getManager();
        }
        catch (DeploymentDescException e) {
            TraceEjb.logger.log(BasicLevel.ERROR,
                "Impossible to load the jorm meta information into the manager", e);
            throw new EJBException(
                "Impossible to load the jorm meta information into the manager", e);
        }

        //---------------------- MEDOR INITIALIZATION -----------------------//
        TraceEjb.logger.log(BasicLevel.DEBUG, "Medor initialisation");
        indexesGenerator = new IndexesGenerator();


    }


}
