/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Helene Joanin
 * Contributor(s): Lutris Technologies Inc.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.deployment.api;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Vector;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.rmi.RemoteException;
import javax.jms.MessageListener;

import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.ContainerTransaction;
import org.objectweb.jonas_ejb.deployment.xml.DisplayName;
import org.objectweb.jonas_ejb.deployment.xml.EjbClass;
import org.objectweb.jonas_ejb.deployment.xml.EjbName;
import org.objectweb.jonas_ejb.deployment.xml.EjbRef;
import org.objectweb.jonas_ejb.deployment.xml.EnvEntry;
import org.objectweb.jonas_ejb.deployment.xml.Home;
import org.objectweb.jonas_ejb.deployment.xml.JndiName;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbRef;
import org.objectweb.jonas_ejb.deployment.xml.JonasResource;
import org.objectweb.jonas_ejb.deployment.xml.JonasResourceEnv;
import org.objectweb.jonas_ejb.deployment.xml.Local;
import org.objectweb.jonas_ejb.deployment.xml.LocalHome;
import org.objectweb.jonas_ejb.deployment.xml.Method;
import org.objectweb.jonas_ejb.deployment.xml.MethodPermission;
import org.objectweb.jonas_ejb.deployment.xml.Remote;
import org.objectweb.jonas_ejb.deployment.xml.ResourceRef;
import org.objectweb.jonas_ejb.deployment.xml.ResourceEnvRef;
import org.objectweb.jonas_ejb.deployment.xml.RoleName;
import org.objectweb.jonas_ejb.deployment.xml.SecurityRole;
import org.objectweb.jonas_ejb.deployment.xml.SecurityRoleRef;
import org.objectweb.jonas_ejb.deployment.xml.EjbLocalRef;
import org.objectweb.jonas_ejb.lib.BeanNaming;
/*
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbRefDesc;
import org.objectweb.jonas_ejb.deployment.api.ResourceRefDesc;
import org.objectweb.jonas_ejb.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbLocalRefDesc;
*/

/**
 * Class to hold meta-information related to any enterprise-beans
 * Contributor(s):
 * Florent Benoit & Ludovic Bert : ejb_link
 * Jeff Mesnil
 * @author Christophe Ney [cney@batisseurs.com]
 * @author Helene Joanin
 */
public abstract class BeanDesc {

    protected final static int BEAN_TRANSACTION_TYPE = 1;
    protected final static int CONTAINER_TRANSACTION_TYPE = 2;
    protected final static String[] TRANS = {
        null,
        "BEAN_TRANSACTION_TYPE",
        "CONTAINER_TRANSACTION_TYPE"
    };

    private static final String ejb10PropertiesPrefix = "ejb10-properties/";

    protected String displayName = null;
    protected String ejbName = null;
    protected String jndiName = null;
    protected Class home = null;
    protected Class remote = null;
    protected Class localhome = null;
    protected Class local = null;
    protected Class ejbClass = null;
    protected Vector envEntryDesc = new Vector();
    protected Properties ejb10EnvProps = new Properties();
    protected Vector resourceEnvRefDesc = new Vector();
    protected Vector resourceRefDesc = new Vector();
    protected Vector ejbRefDesc = new Vector();
    protected Vector ejbLocalRefDesc = new Vector();

    // array of methods meta-info
    private ArrayList methodDesc = new ArrayList();
    // used by deprecated method
    private HashMap methodDescMap = new HashMap();

    protected HashMap securityRoleRef = new HashMap();

    protected DeploymentDesc deploymentDesc = null;

    private String fullWrpRemoteName = null;
    private String wrpRemoteName = null;

    private String fullWrpHomeName = null;
    private String wrpHomeName = null;

    private String fullWrpLocalName = null;
    private String wrpLocalName = null;

    private String fullWrpLocalHomeName = null;
    private String wrpLocalHomeName = null;

    private String fullWrpHandleName = null;
    private String wrpHandleName = null;

    protected String fullDerivedBeanName = null;
    protected String derivedBeanName = null;

    /**
     * constructor to be used by parent node
     */
    public BeanDesc(ClassLoader classLoader, DisplayName displayname, EjbName ejbname,
             JndiName jndiname, Home home, Remote remote, LocalHome localhome, Local local, EjbClass ejbClass,
             List envEntryList, List resourceRefList, List jonasResourceList,
             List resourceEnvRefList, List jonasResourceEnvList,
             List ejbRefList, List jonasEjbRefList, List ejbLocalRefList, List securityRoleRefList,
             AssemblyDescriptor asd)
            throws DeploymentDescException {

        // names
        displayName = (displayname != null) ? displayname.getValue() : null;
        ejbName = (ejbname != null) ? ejbname.getValue() : null;

        // If no value given specifically, we use ejbName as JNDI name.
        jndiName = (jndiname != null) ? jndiname.getValue() : ejbName;

        // classes
        if (home != null) {
            try {
                this.home = classLoader.loadClass(home.getValue());
            } catch (ClassNotFoundException e) {
                throw new DeploymentDescException("Home class not found for bean " + ejbName, e);
            }
        }

        if (remote != null) {
            try {
                this.remote = classLoader.loadClass(remote.getValue());
            } catch (ClassNotFoundException e) {
                throw new DeploymentDescException("Remote class not found for bean " + ejbName, e);
            }
        }

        if (localhome != null) {
            try {
                this.localhome = classLoader.loadClass(localhome.getValue());
            } catch (ClassNotFoundException e) {
                throw new DeploymentDescException("LocalHome class not found for bean " + ejbName, e);
            }
        }

        if (local != null) {
            try {
                this.local = classLoader.loadClass(local.getValue());
            } catch (ClassNotFoundException e) {
                throw new DeploymentDescException("Local class not found for bean " + ejbName, e);
            }
        }

        try {
            this.ejbClass = classLoader.loadClass(ejbClass.getValue());
        } catch (ClassNotFoundException e) {
            throw new DeploymentDescException("Ejb class not found for bean " + ejbName, e);
        }

        // env-entries and ejb10 environment properties
        for (Iterator i = envEntryList.iterator(); i.hasNext();) {
            envEntryDesc.addElement(new EnvEntryDesc((EnvEntry) i.next()));
        }
        for (Enumeration e = envEntryDesc.elements(); e.hasMoreElements();) {
            EnvEntryDesc ed = ((EnvEntryDesc) e.nextElement());
            if (ed.getName().startsWith(ejb10PropertiesPrefix)) {
                String newName = ed.getName().substring(ejb10PropertiesPrefix.length());
                String value = new String();
                if (ed.hasValue()) {
                    value = ed.getValue().toString();
                }
                ejb10EnvProps.put(newName, value);
            }
        }

        // resources
        HashMap resource = new HashMap();
        for (Iterator i = jonasResourceList.iterator(); i.hasNext();) {
            JonasResource jonasResource = ((JonasResource) i.next());
            String rName = jonasResource.getResRefName().getValue();
            boolean findResRef = false;
            for (Iterator j = resourceRefList.iterator(); j.hasNext() && !findResRef;) {
                if (rName.equals(((ResourceRef) j.next()).getResRefName().getValue())) {
                    findResRef = true;
                }
            }
            if (!findResRef) {
                throw new DeploymentDescException("resource-ref missing for jonas-resource " + rName);
            }
            resource.put(rName, jonasResource);
        }
        for (Iterator i = resourceRefList.iterator(); i.hasNext();) {
            ResourceRef resourceRef = ((ResourceRef) i.next());
            String rName = resourceRef.getResRefName().getValue();
            if (!resource.containsKey(rName))
                throw new DeploymentDescException("jonas-resource missing for res-ref-name " + rName);
            resourceRefDesc.addElement(new ResourceRefDesc(classLoader, resourceRef, (JonasResource) resource.get(rName)));
        }

        // resources-env
        HashMap resourceEnv = new HashMap();
        for (Iterator i = jonasResourceEnvList.iterator(); i.hasNext();) {
            JonasResourceEnv jonasResourceEnv = (JonasResourceEnv) i.next();
            String rName = jonasResourceEnv.getResourceEnvRefName().getValue();
            boolean findResRef = false;
            for (Iterator j = resourceEnvRefList.iterator(); j.hasNext() && !findResRef;) {
                ResourceEnvRef resourceEnvRef = (ResourceEnvRef) j.next();
                if (rName.equals(resourceEnvRef.getResourceEnvRefName().getValue())) {
                    findResRef = true;
                }
            }
            if (!findResRef) {
                throw new DeploymentDescException("resource-env-ref missing for jonas-resource-env " + rName);
            }
            resourceEnv.put(rName, jonasResourceEnv);
        }
        for (Iterator i = resourceEnvRefList.iterator(); i.hasNext();) {
            ResourceEnvRef resourceEnvRef = (ResourceEnvRef) i.next();
            String rName = resourceEnvRef.getResourceEnvRefName().getValue();
            if (!resourceEnv.containsKey(rName))
                throw new DeploymentDescException("jonas-resource-env missing for resource-env-ref-name " + rName);
            resourceEnvRefDesc.addElement(new ResourceEnvRefDesc(classLoader, resourceEnvRef, (JonasResourceEnv) resourceEnv.get(rName)));
        }

        // ejb-ref
        HashMap ejbRef = new HashMap();
        for (Iterator i = jonasEjbRefList.iterator(); i.hasNext();) {
            JonasEjbRef jonasEjbRef = (JonasEjbRef) i.next();
            String brName = jonasEjbRef.getEjbRefName().getValue();
            boolean findBRef = false;
            for (Iterator j = ejbRefList.iterator(); j.hasNext() && !findBRef;) {
                EjbRef eRef = (EjbRef) j.next();
                if (brName.equals(eRef.getEjbRefName().getValue())) {
                    findBRef = true;
                }
            }
            if (!findBRef) {
                throw new DeploymentDescException("ejb-ref missing for jonas-ejb-ref " + brName);
            }
            ejbRef.put(brName, jonasEjbRef);
        }
        for (Iterator i = ejbRefList.iterator(); i.hasNext();) {
            EjbRef eRef = (EjbRef) i.next();
            String brName = eRef.getEjbRefName().getValue();
            JonasEjbRef jonasEjbRef = (JonasEjbRef) ejbRef.get(brName);
            if (jonasEjbRef == null) {
                // The ejb-link must be specified when there is no jonas-ejb-ref
                if (eRef.getEjbLink() == null) {
                    throw new DeploymentDescException("jonas-ejb-ref missing for ejb-ref-name " + brName);
                }
            }
            ejbRefDesc.addElement(new EjbRefDesc(eRef, jonasEjbRef));
        }

        // ejb-local-ref
        HashMap ejbLocalRef = new HashMap();
        for (Iterator i = ejbLocalRefList.iterator(); i.hasNext();) {
            EjbLocalRef elr = (EjbLocalRef) i.next();
            EjbLocalRefDesc elrd = new EjbLocalRefDesc(elr);
            ejbLocalRefDesc.addElement(elrd);
        }

        // populate array with all methods
        try {
            java.lang.reflect.Method[] m;
			int beginIdx = methodDesc.size();
			int len = 0;
            if ((this.home != null) && (this.remote != null)) {
                // session bean or entity bean with remote interface
                m = this.home.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (!m[i].getDeclaringClass().equals(javax.ejb.EJBHome.class) ||
                            m[i].getName().equals("remove")) {
                        addMethodDesc(m[i]);
						len++;
                        // check RemoteException is thrown
                        checkRemoteException(m[i], true);
                    }
                }

                m = this.remote.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (!m[i].getDeclaringClass().equals(javax.ejb.EJBObject.class) ||
                            m[i].getName().equals("remove")) {
                        addMethodDesc(m[i]);
						len++;
                        // check RemoteException is thrown
                        checkRemoteException(m[i], true);
                    }
                }
            }
            if ((this.localhome != null) && (this.local != null)) {
                // session bean or entity bean with local interface
                m = this.localhome.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (!m[i].getDeclaringClass().equals(javax.ejb.EJBLocalHome.class) ||
                            m[i].getName().equals("remove")) {
                        addMethodDesc(m[i]);
						len++;
                        // check RemoteException is NOT thrown
                        checkRemoteException(m[i], false);
                    }
                }
                m = this.local.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (!m[i].getDeclaringClass().equals(javax.ejb.EJBLocalObject.class) ||
                            m[i].getName().equals("remove")) {
                        addMethodDesc(m[i]);
						len++;
                        // check RemoteException is NOT thrown
                        checkRemoteException(m[i], false);
                    }
                }
            }
            if (this.home == null && this.localhome == null) {
                // message driven bean
                addMethodDesc(MessageListener.class.getMethods()[0]);
				len++;
            }
			orderMethodDesc(methodDesc, beginIdx, len);
        } catch (SecurityException e) {
            throw new DeploymentDescException("Cannot use java reflection on " + ejbName);
        }

        // populate HashMap with all security role references
        // for each security-role-ref of the bean
        if (securityRoleRefList != null) {
            for (Iterator i = securityRoleRefList.iterator(); i.hasNext();) {
                SecurityRoleRef sRoleRef = (SecurityRoleRef) i.next();
                // test if the assembly-descriptor element exists
                if (asd == null) {
                    throw new DeploymentDescException("No assembly-descriptor element (and security-role elements) whereas security-role-ref element(s) are defined");
                } else {
                    // test if the assembly-descriptor element contains security-role elements
                    if (asd.getSecurityRoleList().isEmpty()) {
                        throw new DeploymentDescException("No security-role elements defined in the assembly-descriptor element  whereas security-role-ref element(s) are defined");
                    }
                }
                // test if the role-link element of the security-role-ref element exists
                if (sRoleRef.getRoleLink() == null) {
                    throw new DeploymentDescException("No role-link element in a security-role-ref element with role-name element : " + sRoleRef.getRoleName());
                }
                String roleLink = sRoleRef.getRoleLink().getValue();
                // correspond to role-name element in security-role-ref element
                String roleRefName = sRoleRef.getRoleName().getValue();
                // to know if there are corresponding roleLink and roleRefName
                boolean correspondingRoles = false;
                for (Iterator j = asd.getSecurityRoleList().iterator(); j.hasNext();) {
                    SecurityRole securityRole = (SecurityRole) j.next();
                    // role name of a security-role element defined in the assembly descriptor
                    String roleName = securityRole.getRoleName().getValue();
                    if (roleName.equals(roleLink)) {
                        securityRoleRef.put(roleRefName, roleLink);
                        correspondingRoles = true;
                    }
                }
                //if there are no matching roleLink and roleName
                if (!correspondingRoles) {
                    throw new DeploymentDescException("No corresponding role-name in a security-role element for role-link : " + roleLink);
                }
            }
        }

        // check if there are corresponding role-name element in security-role
        // and method-permission element of the Assembly Descriptor
        if (asd != null) {
            verifyRoleName(asd);
            // set  method descriptor values for method permissions
            for (Iterator i = asd.getMethodPermissionList().iterator(); i.hasNext();) {
                MethodPermission methodPermission = (MethodPermission) i.next();
                for (Iterator j = methodPermission.getMethodList().iterator(); j.hasNext();) {
                    Method m = ((Method) j.next());
                    // check method applies to current bean
                    if (!m.getEjbName().getValue().equals(ejbName))
                        continue;
                    // selection of interface home/remote
                    Class pClass = null;
                    if (m.getMethodIntf() != null) {
                        if (m.getMethodIntf().getValue().equals("Home"))
                            pClass = javax.ejb.EJBHome.class;
                        else if (m.getMethodIntf().getValue().equals("Remote"))
                            pClass = javax.ejb.EJBObject.class;
                        else if (m.getMethodIntf().getValue().equals("LocalHome"))
                            pClass = javax.ejb.EJBLocalHome.class;
                        else if (m.getMethodIntf().getValue().equals("Local"))
                            pClass = javax.ejb.EJBLocalObject.class;
                        else
                            throw new DeploymentDescException(m.getMethodIntf().getValue() + " is invalid value for method-intf on bean " + ejbName);
                    }
                    for (int k = 0; k < methodDesc.size(); k++) {
                        MethodDesc md = (MethodDesc) methodDesc.get(k);
                        int matchStatus = md.matchPattern(pClass, m.getMethodName().getValue(), m.getMethodParams());
                        for (Iterator l = methodPermission.getRoleNameList().iterator(); l.hasNext();) {
                            String roleName = ((RoleName) l.next()).getValue();
                            md.addRoleName(roleName, matchStatus);
                        }
                    }
                }
            }
        }

        // set method descriptor values for container transactions
        if (asd != null) {
            // Loop on ContainerTransaction list
            Iterator i = asd.getContainerTransactionList().iterator();
            while (i.hasNext()) {
                ContainerTransaction cotr = (ContainerTransaction) i.next();
                // Loop on Method list
                Iterator j = cotr.getMethodList().iterator();
                while (j.hasNext()) {
                    Method m = (Method) j.next();

                    // check method applies to current bean
                    if (!m.getEjbName().getValue().equals(ejbName))
                        continue;

                    // selection of interface home/remote
                    Class pClass = null;
                    if (m.getMethodIntf() != null) {
                        String mintf = m.getMethodIntf().getValue();
                        if (mintf.equals("Home"))
                            pClass = javax.ejb.EJBHome.class;
                        else if (mintf.equals("Remote"))
                            pClass = javax.ejb.EJBObject.class;
                        else if (mintf.equals("LocalHome"))
                            pClass = javax.ejb.EJBLocalHome.class;
                        else if (mintf.equals("Local"))
                            pClass = javax.ejb.EJBLocalObject.class;
                        else
                            throw new DeploymentDescException(mintf + " is invalid value for method-intf on bean " + ejbName);
                    }

                    // loop on methods in the array
                    boolean found = false;
                    for (int k = 0; k < methodDesc.size(); k++) {
                        MethodDesc md = (MethodDesc) methodDesc.get(k);
                        int matchStatus = md.matchPattern(pClass, m.getMethodName().getValue(), m.getMethodParams());
                        if (matchStatus != MethodDesc.APPLY_TO_NOTHING) {
                            found = true;
                        }
                        // set if proper level of assignation
                        String transAttribute = null;
                        if (cotr.getTransAttribute() != null)
                            transAttribute = cotr.getTransAttribute().getValue();
                        md.overwriteTxAttribute(transAttribute, matchStatus);
                    }
                    if (!found) {
                        // The method defined via the <method> element is not a declared method
                        throw new DeploymentDescException("Invalid container-transaction: method '"
                                + MethodDesc.methodElementToString(m)
                                + "' not declared in bean " + ejbName);
                    }
                }
            }
        }

        String packageName ="";
        String ejbIdentifier = getIdentifierFromEjbName(ejbName);

        if (getRemoteClass() != null) {
            packageName = BeanNaming.getPackageName(getRemoteClass().getName());
            wrpRemoteName = new String("JOnAS" + ejbIdentifier + "Remote");
            wrpHandleName = new String("JOnAS" + ejbIdentifier + "Handle");
            fullWrpRemoteName = BeanNaming.getClassName(packageName, wrpRemoteName);
            fullWrpHandleName = BeanNaming.getClassName(packageName, wrpHandleName);
        }

        if (getLocalClass() != null) {
            packageName = BeanNaming.getPackageName(getLocalClass().getName());
            wrpLocalName = new String("JOnAS" + ejbIdentifier + "Local");
            fullWrpLocalName = BeanNaming.getClassName(packageName, wrpLocalName);
        }

        if (getHomeClass() != null) {
            packageName = BeanNaming.getPackageName(getHomeClass().getName());
            wrpHomeName = new String("JOnAS" + ejbIdentifier + "Home");
            fullWrpHomeName = BeanNaming.getClassName(packageName, wrpHomeName);
        }

        if (getLocalHomeClass() != null) {
            packageName = BeanNaming.getPackageName(getLocalHomeClass().getName());
            wrpLocalHomeName = new String("JOnAS" + ejbIdentifier + "LocalHome");
            fullWrpLocalHomeName = BeanNaming.getClassName(packageName, wrpLocalHomeName);
        }

        fullDerivedBeanName = getEjbClass().getName();
        derivedBeanName = BeanNaming.getBaseName(getEjbClass().getName());

    }

    /**
     * Build a valid java identifier from the ejb name
     */
    protected String getIdentifierFromEjbName(String ejbName) {
        char[] tabChar = new char[ejbName.length()];
        for (int i = 0; i < ejbName.length(); i++) {
            if (!Character.isJavaIdentifierPart(ejbName.charAt(i))) {
                tabChar[i] = '_';
            } else {
                tabChar[i] = ejbName.charAt(i);
            }
        }
        return new String(tabChar);
    }

    /**
     * Check if method throws RemoteException
     * @exception DeploymentDescException thrown for non-valid method
     */
    public void checkRemoteException(java.lang.reflect.Method m, boolean shouldThrow) throws DeploymentDescException {
        Class[] excepts = m.getExceptionTypes();
        for (int j = 0; j < excepts.length; j++) {
            if (RemoteException.class.isAssignableFrom(excepts[j])) {
                if (shouldThrow)
                    return;
                else {
                    throw new DeploymentDescException("Method " + m.getName() + " of "
                            + m.getDeclaringClass() + " should NOT throw RemoteException");
                }
            }
        }
        if (shouldThrow) {
            throw new DeploymentDescException("Method " + m.getName() + " of "
                    + m.getDeclaringClass() + " must throw RemoteException");
        } else {
            return;
        }

    }


    /**
     * Check that the bean descriptor is valid
     * @exception DeploymentDescException thrown for non-valid bean
     */
    public void check() throws DeploymentDescException {
        // check business methods transaction attribute
        for (int i = 0; i < methodDesc.size(); i++) {
            checkTxAttribute(((MethodDesc) methodDesc.get(i)));
        }
    }

    /**
     * check that trans-attribute is valid for bean
     */
    protected abstract void checkTxAttribute(MethodDesc md) throws DeploymentDescException;

    /**
     * check that there are always corresponding role-name element
     * in security-role and method-permission elements
     */
    protected void verifyRoleName(AssemblyDescriptor asd) throws DeploymentDescException {
        boolean verify;
        List methodPermission = asd.getMethodPermissionList();
        if (methodPermission != null) {
            for (Iterator i = methodPermission.iterator(); i.hasNext();) {
                MethodPermission mPermission = (MethodPermission) i.next();
                for (Iterator j = mPermission.getRoleNameList().iterator(); j.hasNext();) {
                    verify = false;
                    String mpRoleName = ((RoleName) j.next()).getValue();
                    for (Iterator k = asd.getSecurityRoleList().iterator(); k.hasNext();) {
                        String srRoleName = ((SecurityRole) k.next()).getRoleName().getValue();
                        if (mpRoleName.equals(srRoleName)) {
                            verify = true;
                        }
                    }
                    if (!verify) {
                        throw new DeploymentDescException("a method-permission element defined a role-name element \"" + mpRoleName + "\" which doesn't correspond to any security-role element of the Deployment Descriptor");
                    }
                }
            }
        }
    }

    /**
     * Get the displayName of the bean.
     * Used by Server
     * @return null if not set
     */
    public String getDisplayName() {
        return displayName;
    }

    /**
     * Get the enterprise bean's name.
     * @return Name of the bean
     */
    public String getEjbName() {
        return ejbName;
    }

    /**
     * Get the name to associate with the enterprise Bean in the JNDI name space.
     * Invoking this method is disallowed for message driven bean.
     * (This information is JOnAS specific). Used by Server
     * @return Jndi name of the bean
     */
    public String getJndiName() {
        if (jndiName == null) {
            throw new Error("Disallowed invocation of BeanDesc.getJndiName()");
        }
        return jndiName;
    }

    /**
     * Get the name to associate with the enterprise Bean in the JNDI name space
     * for local access (inside the same JOnAS Server)
     * @return Jndi name of the bean
     */
    public String getJndiLocalName() {
        if (jndiName == null) {
            throw new Error("Disallowed invocation of BeanDesc.getJndiLocalName()");
        }
        return jndiName + "_L";
    }

    /**
     * Get bean's home interface.
     * Invoking this method is disallowed for message driven bean.
     * Used by GenIC and Server
     * @return Class for the home interface or null
     */
    public Class getHomeClass() {
        return home;
    }

    /**
     * Get bean's remote interface.
     * Invoking this method is disallowed for message driven bean.
     * Used by GenIC and Server
     * @return Class for the remote interface or null
     */
    public Class getRemoteClass() {
        return remote;
    }

    /**
     * Get bean's local home interface.
     * Invoking this method is disallowed for message driven bean.
     * Used by GenIC and Server
     * @return Class for the home interface or null
     */
    public Class getLocalHomeClass() {
        return localhome;
    }

    /**
     * Get bean's local interface.
     * Invoking this method is disallowed for message driven bean.
     * Used by GenIC and Server
     * @return Class for the Local interface or null
     */
    public Class getLocalClass() {
        return local;
    }

    /**
     * Get bean's class.
     * Used by GenIC and Server
     * @return Class for the bean
     */
    public Class getEjbClass() {
        return ejbClass;
    }

    /**
     * Get bean's environment entries.
     * Used by Server
     * @return array of resource reference descriptors (may be empty)
     */
    public EnvEntryDesc[] getEnvEntryDesc() {
        EnvEntryDesc ret[] = new EnvEntryDesc[envEntryDesc.size()];
        envEntryDesc.copyInto(ret);
        return ret;
    }

    /**
     * Get bean's EJB 1.0 environment properties
     */
    public Properties getEjb10Environment() {
        Properties ret = new Properties(ejb10EnvProps);
        return ret;
    }

    /**
     * Get bean's resource environment references.
     * Used by Server
     * @return array of resource environment reference descriptors (may be empty)
     */
    public ResourceEnvRefDesc[] getResourceEnvRefDesc() {
        ResourceEnvRefDesc ret[] = new ResourceEnvRefDesc[resourceEnvRefDesc.size()];
        resourceEnvRefDesc.copyInto(ret);
        return ret;
    }

    /**
     * Get bean's resource manager connection factory references.
     * Used by Server
     * @return array of resource reference descriptors (may be empty)
     */
    public ResourceRefDesc[] getResourceRefDesc() {
        ResourceRefDesc ret[] = new ResourceRefDesc[resourceRefDesc.size()];
        resourceRefDesc.copyInto(ret);
        return ret;
    }

    /**
     * Get bean's ejb references.
     * Used by Server
     * @return array of ejb reference descriptors (may be empty)
     */
    public EjbRefDesc[] getEjbRefDesc() {
        EjbRefDesc ret[] = new EjbRefDesc[ejbRefDesc.size()];
        ejbRefDesc.copyInto(ret);
        return ret;
    }

    /**
     * Get bean's ejb local references.
     * Used by Server
     * @return array of ejb local reference descriptors (may be empty)
     */
    public EjbLocalRefDesc[] getEjbLocalRefDesc() {
        EjbLocalRefDesc ret[] = new EjbLocalRefDesc[ejbLocalRefDesc.size()];
        ejbLocalRefDesc.copyInto(ret);
        return ret;
    }

    /**
     * Get descriptor attached to a given method.
     * Used by GenIC
     * @param method Method from the Home/Remote/MessageListener interface
     * @return The method descriptor (never returns null)
     * @deprecated use getMethodDescIterator or getMethodDesc(int index) instead
     */
    public MethodDesc getMethodDesc(java.lang.reflect.Method method) {

        if (!methodDescMap.containsKey(method))
            throw new Error(method.getName() + " is not a method of home/remote interfaces for bean " + getEjbName());
        return (MethodDesc) methodDescMap.get(method);
    }

    /**
     * get an iterator for all MethodDesc of the bean
     */
    public Iterator getMethodDescIterator() {
        return methodDesc.iterator();
    }

    /**
     * get a MethodDesc from its index
     */
    public MethodDesc getMethodDesc(int index) {
        return (MethodDesc) methodDesc.get(index);
    }

    /**
     * Get the role-link value corresponding to the role-name value of a security-role-ref element
     *
     * @param roleName String corresponding to the role-name value of a security-role-ref element
     * @return String corresponding to the role-link value
     */
    public String getRoleLink(String roleName) {
        String roleLink = (String) securityRoleRef.get(roleName);
        return roleLink;
    }

    /**
     * Get all the role names defined in the security-role-ref elements of the bean
     *
     * @param roleName String corresponding to the role-name value of a security-role-ref element
     * @return String corresponding to the role-link value
     */
    public String[] getSecurityRoleNames() {
        String[] ret = new String[securityRoleRef.size()];
        securityRoleRef.keySet().toArray(ret);
        return ret;
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append("\ngetDisplayName()=" + getDisplayName());
        ret.append("\ngetEjbName()=" + getEjbName());
        if (jndiName != null) {
            ret.append("\ngetJndiName()=" + getJndiName());
        }
        if (home != null) {
            ret.append("\ngetHomeClass()=" + getHomeClass().toString());
        }
        if (remote != null) {
            ret.append("\ngetRemoteClass()=" + getRemoteClass().toString());
        }
        if (localhome != null) {
            ret.append("\ngetLocalHomeClass()=" + getLocalHomeClass().toString());
        }
        if (local != null) {
            ret.append("\ngetLocalClass()=" + getLocalClass().toString());
        }
        ret.append("\ngetEjbClass()=" + getEjbClass().toString());
        EnvEntryDesc[] envE = getEnvEntryDesc();
        for (int i = 0; i < envE.length; i++) {
            ret.append("\ngetEnvEntryDesc(" + i + ")=" + envE[i].toString());
        }
        ResourceEnvRefDesc[] resEnvR = getResourceEnvRefDesc();
        for (int i = 0; i < resEnvR.length; i++) {
            ret.append("\ngetResourceEnvRefDesc(" + i + ")=" + resEnvR[i].toString());
        }
        ResourceRefDesc[] resR = getResourceRefDesc();
        for (int i = 0; i < resR.length; i++) {
            ret.append("\ngetResourceRefDesc(" + i + ")=" + resR[i].toString());
        }
        EjbRefDesc[] ejbR = getEjbRefDesc();
        for (int i = 0; i < ejbR.length; i++) {
            ret.append("\ngetEjbRefDesc(" + i + ")=" + ejbR[i].toString());
        }
        EjbLocalRefDesc[] ejbLR = getEjbLocalRefDesc();
        for (int i = 0; i < ejbLR.length; i++) {
            ret.append("\ngetEjbLocalRefDesc(" + i + ")=" + ejbLR[i].toString());
        }
        for (int i = 0; i < methodDesc.size(); i++) {
            MethodDesc md = (MethodDesc) methodDesc.get(i);
            ret.append("\ngetMethodDesc(" + i + ")=" + md.getClass().getName());
            ret.append(md.toString());
        }
        for (Iterator i = securityRoleRef.keySet().iterator(); i.hasNext();) {
            String roleName = (String) i.next();
            ret.append("\ngetRoleLink(" + roleName + ") = " + getRoleLink(roleName));
        }
        ret.append("getFullWrpRemoteName() = " + getFullWrpRemoteName());
        ret.append("getFullWrpHomeName() = " + getFullWrpHomeName());
        ret.append("getFullWrpLocalName() = " + getFullWrpLocalName());
        ret.append("getFullWrpLocalHomeName() = " + getFullWrpLocalHomeName());
        ret.append("getFullWrpHandleName() = " + getFullWrpHandleName());
        ret.append("getFullDerivedBeanName() = " + getFullDerivedBeanName());
        return ret.toString();
    }

    /**
     * MethodDesc constructor factory method to ovewrite
     */
    protected MethodDesc newMethodDescInstance(java.lang.reflect.Method meth, int index) {
        return new MethodDesc(this, meth, index);
    }

    /**
     * add a new method descriptor to the bean
     */
    private MethodDesc addMethodDesc(java.lang.reflect.Method meth) {
        MethodDesc md = newMethodDescInstance(meth, methodDesc.size());
        methodDesc.add(md);
        methodDescMap.put(meth, md);
        return md;
    }

    /**
     * returns the full name of the wrapper class for the bean
     */
    public String getFullWrpRemoteName() {
        return fullWrpRemoteName;
    }

    /**
     * returns the full name of the wrapper class for the bean
     */
    public String getFullWrpLocalName() {
        return fullWrpLocalName;
    }

    /**
     * returns the name of the wrapper class for the bean
     */
    public String getWrpRemoteName() {
        return wrpRemoteName;
    }

    /**
     * returns the name of the wrapper class for the bean
     */
    public String getWrpLocalName() {
        return wrpLocalName;
    }

    /**
     * returns the full name of the Handle class for the bean
     */
    public String getFullWrpHandleName() {
        return fullWrpHandleName;
    }

    /**
     * returns the name of the Handle class for the bean
     */
    public String getWrpHandleName() {
        return wrpHandleName;
    }

    /**
     * returns the full name of the wrapper home class for the bean
     */
    public String getFullWrpHomeName() {
        return fullWrpHomeName;
    }

    /**
     * returns the full name of the wrapper local home class for the bean
     */
    public String getFullWrpLocalHomeName() {
        return fullWrpLocalHomeName;
    }

    /**
     * returns the name of the wrapper home class for the bean
     */
    public String getWrpHomeName() {
        return wrpHomeName;
    }

    /**
     * returns the name of the wrapper local home class for the bean
     */
    public String getWrpLocalHomeName() {
        return wrpLocalHomeName;
    }

    /**
     * returns the full name of the derived class of the bean
     */
    public String getFullDerivedBeanName() {
        return fullDerivedBeanName;
    }

    /**
     * returns the name of the derived class of the bean
     */
    public String getDerivedBeanName() {
        return derivedBeanName;
    }

    /**
     * get the parent deployment descriptor
     */
    public DeploymentDesc getDeploymentDesc() {
        return deploymentDesc;
    }

    /**
     * set the parent deployment descriptor
     */
    public void setDeploymentDesc(DeploymentDesc deploymentDesc) {
        this.deploymentDesc = deploymentDesc;
    }

	private void orderMethodDesc(ArrayList al, int begin, int length) {
        MethodDesc[] mds = new MethodDesc[length];
		for (int i=0; i<length; i++) {
			mds[i] = (MethodDesc) al.get(begin + i);
		}
		Arrays.sort(mds, 0, length, new MethodDescComparator());
		for (int i=0; i<length; i++) {
			al.set(begin + i, mds[i]);
			mds[i].setIndex(begin + i);
		}
	}
	private class MethodDescComparator implements Comparator {
		public int compare(Object o1, Object o2) {
            MethodDesc md1 = (MethodDesc) o1;
            MethodDesc md2 = (MethodDesc) o2;
			int res = md1.getMethod().getDeclaringClass().getName().compareTo(
				md2.getMethod().getDeclaringClass().getName());
			if (res == 0) {
				res = md1.getMethod().getName().compareTo(
					md2.getMethod().getName());
				if (res == 0) {
					// Same method name => compare the parameter types
					Class[] p1 = md1.getMethod().getParameterTypes();
					Class[] p2 = md2.getMethod().getParameterTypes();
					res = p1.length - p2.length;
					if (res == 0) {
						//Same parameter size
						for (int i = 0;
							 i < p1.length && (res = p1[i].getName().compareTo(p2[i].getName())) == 0;
							 i++)
							;
					}
				}
			}
			return res;
		}
	}
}

