/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Sebastien Chassandre Barrioz
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.deployment.api;

import org.objectweb.jonas_ejb.deployment.xml.EjbJar;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbJar;
import org.objectweb.jonas_ejb.deployment.xml.Entity;
import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.JonasEntity;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbRelation;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelation;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelationName;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbRelationshipRole;
import org.objectweb.jonas_ejb.deployment.xml.JdbcMapper;
import org.objectweb.jonas_ejb.lib.MappingBuilder;
import org.objectweb.jonas_ejb.lib.JormType;
import org.objectweb.jonas_ejb.container.TraceEjb;
import org.objectweb.jonas_ejb.genic.GenICException;
import org.objectweb.jonas_ejb.genic.BeanSources;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.lib.BasicManager;
import org.objectweb.jorm.type.api.PType;

import java.util.Iterator;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.ArrayList;
import java.util.List;

/**
 * Class to hold meta-information related to the deployment of an ejb-jar
 * with JORM based CMP2 persistence
 * @author S.Chassande-Barrioz
 * @author Christophe Ney [cney@batisseurs.com]
 */
public class DeploymentCmp2Desc extends DeploymentDesc {

    /**
     * This constant defines the default mapping name as 'pkNaming'
     */
    public final static String MAPPING_NAME = "pkNaming";

    /**
     * hold the JORM meta-information manager
     */
    private BasicManager manager = null;

	private String[] mapperNames = null;

    /**
     * Build the Meta-Information from the XML data binding trees
     * containing the EJB and JOnAS deployment descriptors.
     */
    public DeploymentCmp2Desc(ClassLoader classLoader,
                              EjbJar ejbJar,
                              JonasEjbJar jonasEjbJar,Logger l)
        throws DeploymentDescException {
        super(classLoader, ejbJar, jonasEjbJar, l);
        if (ejbJar.getRelationships()!=null) {
            // HashMap of jonas-ejb-relation
            HashMap jonasEjbRelation = new HashMap();
            for (Iterator i = jonasEjbJar.getJonasEjbRelationList().iterator(); i.hasNext();) {
                JonasEjbRelation jEr = (JonasEjbRelation) i.next();
                jonasEjbRelation.put(jEr.getEjbRelationName().getValue(), jEr);
            }

            // ejb-relation
            for (Iterator i = ejbJar.getRelationships().getEjbRelationList().iterator(); i.hasNext();) {
                EjbRelationDesc erd = null;
                EjbRelation er = (EjbRelation) i.next();
                EjbRelationName ern = er.getEjbRelationName();
                if (ern==null) {
                    throw new DeploymentDescException("ejb-relation-name missing in ejb-relation "+er.getDescription());
                }
                String name = ern.getValue();
                if (name.length()==0) {
                    throw new DeploymentDescException(
                        "ejb-relation-name value missing in jonas-ejb-jar.xml");
                }
                JonasEjbRelation jer = (JonasEjbRelation) jonasEjbRelation.get(name);
                if (jer==null) {
                    throw new DeploymentDescException(
                        "jonas-ejb-relation missing for ejb-relation \""+name+"\"");
                }
                // check if jdbc mapping
                boolean jdbc = false;
                for (Iterator j = jer.getJonasEjbRelationshipRoleList().iterator();j.hasNext();) {
                    JonasEjbRelationshipRole jerrl = (JonasEjbRelationshipRole)j.next();
                    if (!jerrl.getForeignKeyJdbcMappingList().isEmpty()) {
                        jdbc = true;
                    }
                }
                if (jdbc==true) {
                    erd = new EjbRelationJdbcDesc(er, jer);
                }
                else {
                    erd = new EjbRelationDesc(er, jer);
                }
                // ejb-relationship-role (source,target)
                EjbRelationshipRoleDesc rsd1 = erd.getRelationshipRoleDesc1();
                EjbRelationshipRoleDesc rsd2 = erd.getRelationshipRoleDesc2();
                String ejbName1 = rsd1.getEjbName();
                String ejbName2 = rsd2.getEjbName();
                EntityCmp2Desc ed1 = (EntityCmp2Desc) beanDesc.get(ejbName1);
                EntityCmp2Desc ed2 = (EntityCmp2Desc) beanDesc.get(ejbName2);
                ed1.addEjbRelationshipRoleDesc(rsd1);
                ed2.addEjbRelationshipRoleDesc(rsd2);
                rsd1.setLinkedEntityCmp2Desc(ed2);
                rsd2.setLinkedEntityCmp2Desc(ed1);
            }
        }
		JdbcMapper m = jonasEjbJar.getJdbcMapper();
		if (m == null || m.getValue() == null || m.getValue().length() == 0)
			mapperNames = new String[]{"rdb"};
		else
			mapperNames = new String[] {jonasEjbJar.getJdbcMapper().getValue()};
    }

    public BeanDesc newBeanDesc(String cmpversion,
                                ClassLoader classLoader,
                                Entity ent,
                                AssemblyDescriptor asd,
                                JonasEntity jEnt)
            throws DeploymentDescException {

        if (cmpversion.trim().equals("2.x"))
            return new EntityJdbcCmp2Desc(classLoader, ent, asd, jEnt);
        else
            return super.newBeanDesc(cmpversion, classLoader, ent, asd, jEnt);
    }

    /**
     * get the JORM Manager
     */
    public Manager getManager() throws DeploymentDescException {
        if (manager==null) {
            manager = new BasicManager();
            manager.setLogger(logger);
            manager.init();
            manager.addMappingFactory("rdb","org.objectweb.jorm.metainfo.lib.rdb.BasicRDBMappingFactory");
            // FIXME mapper factory should be set per bean
            // manager.addMappingFactory(getMapperName(ed), getMIMappingFactory(ed));


            BeanDesc[] beansDD = getBeanDesc();
            for (int i = 0; i < beansDD.length; i++) {
                // Load the jorm meta information of the class
                if (beansDD[i] instanceof EntityCmp2Desc) {
                    createFromBeanDesc((EntityCmp2Desc)beansDD[i],
                            ((EntityCmp2Desc)beansDD[i]).getMappingBuilder(),
                            manager);
                }
            }

        }
        return manager;
    }

    /**
     * Create JORM MetaInfo from a JOnAS CMP2 Entity Bean Descriptor
     */
    public void createFromBeanDesc(EntityCmp2Desc bd, MappingBuilder mb, Manager m)
        throws DeploymentDescException {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                "Load Jorm MI for the bean: " + bd.getAbstractSchemaName());
        }
        Schema schema = m.createSchema(bd.getJormSchemaName());

        // create persistent class of name abstract schema name.
        // Be careful: the skeleton of class can be already defined when this
        // class is referenced by another class (scheduling problem)
        Class clazz = schema.getHiddenClass(bd.getAbstractSchemaName());
        if (clazz == null) {
            clazz = schema.createClass(bd.getAbstractSchemaName(), false);
        } else {
            schema.moveHiddenToClass(bd.getAbstractSchemaName());
        }
        mb.createClassMapping(clazz, bd, MAPPING_NAME);

        //----------------------------- CMP fields ---------------------------//
        //--------------------------------------------------------------------//
        for (Iterator i = bd.getCmpFieldDescIterator(); i.hasNext();) {
            FieldDesc fd = (FieldDesc) i.next();
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG,
                    "Create PrimitiveElement for the CMP field: " + fd.getFieldName());
            }
            PType pt = JormType.getPType(fd.getFieldType());
            PrimitiveElement pe =
                clazz.createPrimitiveElement(fd.getFieldName(), pt);
            mb.createValueMapping(pe, fd, MAPPING_NAME);
        }

        NameDef[] nds = {clazz.createNameDef()};

        //--------------------------- Class NameDef --------------------------//
        //--------------------------------------------------------------------//
        for (int i=0; i<nds.length; i++) {
            nds[i].setMappNames(MAPPING_NAME);
        }
        if (bd.hasPrimaryKeyField()) {
            // one primary key field
            String fieldName = bd.getPrimaryKeyFieldName();
            for (int i=0; i<nds.length; i++) {
                nds[i].setFieldName(fieldName);
            }
        }
        else {
            //Define a Composite Name and a NameRef
            // The name of the composite is built with the abstract schema
            // name and the suffix "Pk"
            CompositeName cn =
                schema.createCompositeName(bd.getAbstractSchemaName() + "Pk");

            // Used the new CompositeName into the NameDef
            NameRef[] nrs = new NameRef[nds.length];
            for (int i=0; i<nds.length; i++) {
                nrs[i] = nds[i].createNameRef(cn.getName(), cn);
            }
            // fill the composite name with pk fields and link these fields to
            // to the class fields
            for (Iterator it = bd.getCmpFieldDescIterator(); it.hasNext();) {
                FieldDesc fd = (FieldDesc) it.next();
                if (fd.isPrimaryKey()) {
                    String fieldName = fd.getFieldName();
                    cn.createCompositeNameField(
                        fieldName, JormType.getPType(fd.getFieldType()));
                    for (int i=0; i<nds.length; i++) {
                        nrs[i].addProjection(fieldName, fieldName);
                    }
                }
            }
        }

        //--------------------------- CMR fields -----------------------------//
        //--------------------------------------------------------------------//
        for (Iterator i = bd.getEjbRelationshipRoleDescIterator(); i.hasNext();) {
            EjbRelationshipRoleDesc rsrd = (EjbRelationshipRoleDesc) i.next();
            if (rsrd.hasCmrField()) {
                EntityCmp2Desc bd2 = rsrd.getLinkedEntityCmp2Desc();
                if (rsrd.isRelationMultiple()) {
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG,
                            "Create GenClassReference for the CMR field: " + rsrd.getCmrFieldName());
                    }
                    createGenClassReference(bd, clazz, rsrd, bd2, mb);
                }
                else {
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG,
                            "Create ClassReference for the CMR field: " + rsrd.getCmrFieldName());
                    }
                    createClassReference(clazz, rsrd, bd2, mb,logger);
                }
            }
        }
    }

    /**
     * This method create the jorm meta information which matches with a
     * relation between one bean and several others.
     * @param bd is the descriptor of the first bean
     * @param clazz is the class (jorm meta object) of the first bean.
     * @param fd is the descriptor of the relation
     * @param bd2 is the descriptor of the second bean
     */
    private void createGenClassReference(EntityCmp2Desc bd,
                                         Class clazz,
                                         EjbRelationshipRoleDesc rsrd,
                                         EntityCmp2Desc bd2,
                                         MappingBuilder mb)
        throws DeploymentDescException {
        Schema schema = (Schema) clazz.getParent();

        //create the gen class
        GenClassRef gcr = clazz.createGenClassRef(
            rsrd.getCmrFieldName(), rsrd.getCmrFieldType().getName());

        // cm describes the mapping of the Gen class
        mb.createClassMapping(gcr, rsrd, MAPPING_NAME);

        // gcnd is the namedef which describes how the gen class is
        // identified (the gen class name)
        NameDef gcnd = gcr.createRefNameDef();
        //fieldNd is the nameDef which desribes how the current bean
        // references the gen class
        NameDef fieldNd = gcr.createNameDef();

        NameDef[] nds = {gcnd, fieldNd};
//        fillNameDef(nds,bd, schema, clazz);

        Class clazz2 = schema.getClass(bd2.getAbstractSchemaName());
        if (clazz2==null) {
            // The class is not yet defined => create a hidden class
            clazz = schema.createHiddenClass(bd2.getAbstractSchemaName(), false);
        }

        //This namedef describes how the referenced class
        // is referenced from the gen class.
        // In a implicit mapping this NameDef is equals to the NameDef
        // of the referenced class
        nds = new NameDef[1];
        nds[0] = gcr.createClassRef(clazz2).createNameDef() ;
//        fillNameDef(nds, bd2, schema, null);
    }

    /**
     * This method create the jorm meta information which matches with a
     * simple relation between two bean.
     * @param clazz is the class (jorm meta object) of the first bean.
     * @param rsrd is the descriptor of the relation role
     * @param bd is the descriptor of the second bean
     */
    private void createClassReference(Class clazz,
                                      EjbRelationshipRoleDesc rsrd,
                                      EntityCmp2Desc bd,
                                      MappingBuilder mb, Logger logger)
        throws DeploymentDescException {

        String cmrfieldName = rsrd.getCmrFieldName();
        Schema schema = (Schema) clazz.getParent();
        String className = bd.getAbstractSchemaName();
        Class c = schema.getClass(className);
        if (c == null) {
            throw new DeploymentDescException("Class " + className
                + " is not availlable in the jorm meta information");
        }
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                "Create ClassRef for the CMR field: " + rsrd.getCmrFieldName());
        }
        ClassRef cr = clazz.createClassRef(cmrfieldName, c);

        // Define the NameDef of the field like the NameDef of the
        // referenced class
        NameDef[] nds = {cr.createNameDef()};

        // Fill the NameDef of the class either the primary key
         for (int i=0; i<nds.length; i++) {
             nds[i].setMappNames(MAPPING_NAME);
         }
         if (bd.hasPrimaryKeyField()) {
             // one primary key field
             String clafn = cmrfieldName + "Pk";
             for (int i=0; i<nds.length; i++) {
                 nds[i].setFieldName(clafn);
                 declareHiddenField(clafn,
                     bd.getCmpFieldDesc(bd.getPrimaryKeyFieldName()),
                     rsrd,
                     clazz,
                     mb);
             }
         }
         else {
             //Define a Composite Name and a NameRef
             // The name of the composite is built with the abstract schema
             // name and the suffix "Pk"
             CompositeName cn =
                 schema.createCompositeName(bd.getAbstractSchemaName() + "Pk");

             // Used the new CompositeName into the NameDef
             NameRef[] nrs = new NameRef[nds.length];
             for (int i=0; i<nds.length; i++) {
                 nrs[i] = nds[i].createNameRef(cn.getName(), cn);
             }
             // fill the composite name with pk fields and link these field to
             // to the class field
             for (Iterator it = bd.getCmpFieldDescIterator(); it.hasNext();) {
                 FieldDesc fd = (FieldDesc) it.next();
                 if (fd.isPrimaryKey()) {
                     String cofn = fd.getFieldName(); //composite field name
                     String clafn = cmrfieldName + "_" + cofn; // class field name
                     cn.createCompositeNameField(
                         cofn, JormType.getPType(fd.getFieldType()));
                     for (int i=0; i<nds.length; i++) {
                         nrs[i].addProjection(cofn, clafn);
                     }
                     declareHiddenField(clafn, fd, rsrd, clazz, mb);
                 }
             }
         }
    }

    /**
     * This method declares a new hidden field in a class.
     * @param fieldName is the name of the new field
     * @param desc describes the new field. But the name specified inside must
     * not be used. Indeed in order to prevent conflict the use of this method
     * specifies a name (fieldName param).
     * @param clazz is the class (jorm meta object) inside which the field must
     * be declared.
     * @throw DeploymentDescException when an existing field (hidden or not)
     * has the same name.
     */
    private void declareHiddenField(String fieldName,
                                    FieldDesc desc,
                                    EjbRelationshipRoleDesc rsrd,
                                    Class clazz,
                                    MappingBuilder mb) throws DeploymentDescException {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Declare the hidden field "
                + fieldName + " in the class " + clazz.getName());
        }
        if (clazz.getHiddenField(fieldName) != null
            || clazz.getTypedElement(fieldName) != null) {
            throw new DeploymentDescException(
                "Bad implicit mapping choice: Two fields with same name: "
                + fieldName);
        }
        PrimitiveElement pe = (PrimitiveElement)
            clazz.createHiddenField(fieldName, JormType.getPType(desc.getFieldType()));
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Create ValueMapping for a Class reference");
        }
        mb.createValueMapping(pe, desc, rsrd, MAPPING_NAME);
    }


    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(super.toString());
        return ret.toString();
    }

	public String[] getMapperNames() {
		return mapperNames;
	}
}
