/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): Lutris Technologies Inc.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.deployment.api;

import java.util.Vector;
import java.util.HashMap;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.zip.ZipFile;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.net.URL;

import org.objectweb.util.monolog.api.Logger;
// SAX imports
import org.xml.sax.ErrorHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.EjbJar;
import org.objectweb.jonas_ejb.deployment.xml.EjbJarUnmarshaller;
import org.objectweb.jonas_ejb.deployment.xml.Entity;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbJar;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbJarUnmarshaller;
import org.objectweb.jonas_ejb.deployment.xml.JonasEntity;
import org.objectweb.jonas_ejb.deployment.xml.JonasMessageDriven;
import org.objectweb.jonas_ejb.deployment.xml.JonasSession;
import org.objectweb.jonas_ejb.deployment.xml.JonasSessionImpl;
import org.objectweb.jonas_ejb.deployment.xml.MessageDriven;
import org.objectweb.jonas_ejb.deployment.xml.Session;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbRelation;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelation;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelationName;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbRelationshipRole;


/**
 * Class to hold meta-information related to the deployment of an ejb-jar
 * Contributor(s):
 * Markus Karg: Novell port.
 * Phlippe Durieux: EJB Local References
 * Florent Benoit & Ludovic Bert : Ear
 * @author Christophe Ney [cney@batisseurs.com]
 */
public class DeploymentDesc {

    protected Logger logger;

    /**
     * the role names in the EAR case
     */
    protected String[] earRoleNames = null;

    protected HashMap beanDesc = new HashMap();

    protected String displayName = null;


    /**
     * Build the Meta-Information from the XML data binding trees
     * containing the EJB and JOnAS deployment descriptors.
     */
    public DeploymentDesc(ClassLoader classLoader, EjbJar ejbJar, JonasEjbJar jonasEjbJar, Logger l)
	throws DeploymentDescException {

        logger = l;

        // test classloader
        if (classLoader == null) {
            throw new DeploymentDescException("DeploymentDesc: Classloader is null");
        }

        // test the validity of the ejbJar (EnterpriseBeans must be present)
        if (ejbJar.getEnterpriseBeans()==null) {
                throw new DeploymentDescException("DeploymentDesc: Invalid standard deployment descriptor (<enterprise-beans> element missing)");
        }

        // display name
        displayName = (ejbJar.getDisplayName() != null) ? ejbJar.getDisplayName().getValue() : null;

        // assembly descriptor
        AssemblyDescriptor asd = ejbJar.getAssemblyDescriptor();

        // HashMap of jonas-session
        HashMap jonasSession = new HashMap();
        for (Iterator i = jonasEjbJar.getJonasSessionList().iterator(); i.hasNext(); ) {
            JonasSession jSes = (JonasSession)i.next();
            jonasSession.put(jSes.getEjbName().getValue(), jSes);
        }

        // session beans
        for (Iterator i = ejbJar.getEnterpriseBeans().getSessionList().iterator(); i.hasNext(); ) {
            BeanDesc bd = null;
            Session ses = (Session)i.next();
            // find corresponding jonas session
            JonasSession jSes = (JonasSession)jonasSession.get(ses.getEjbName().getValue());
            if (jSes == null) {
		// Build a default jonas-session if not exist
		jSes = new JonasSessionImpl();
		jSes.setEjbName(ses.getEjbName());
	    }
            if (ses.getSessionType().getValue().equals("Stateful")) {
                // stateful
                bd = new SessionStatefulDesc(classLoader, ses, asd, jSes);
            } else if (ses.getSessionType().getValue().equals("Stateless")) {
                // stateless
                bd = new SessionStatelessDesc(classLoader, ses, asd, jSes);
            } else {
                throw new DeploymentDescException("Invalid session-type content for bean "+ses.getEjbName().getValue());
            }
            bd.check();
            bd.setDeploymentDesc(this);
            beanDesc.put(bd.getEjbName(),bd);
        }

        // HashMap of jonas-entity
        HashMap jonasEntity = new HashMap();
        for (Iterator i = jonasEjbJar.getJonasEntityList().iterator(); i.hasNext(); ) {
            JonasEntity jEnt = (JonasEntity)i.next();
            jonasEntity.put(jEnt.getEjbName().getValue(), jEnt);
        }
        // entity beans
        for (Iterator i = ejbJar.getEnterpriseBeans().getEntityList().iterator(); i.hasNext(); ) {
            BeanDesc bd = null;
            Entity ent = (Entity) i.next();
            // find corresponding jonas session
            JonasEntity jEnt = (JonasEntity)jonasEntity.get(ent.getEjbName().getValue());
            if (jEnt == null)
                throw new DeploymentDescException("jonas-entity missing for bean "+ent.getEjbName().getValue());
            // bean managed
            if (ent.getPersistenceType().getValue().equals("Bean")) {
                bd = new EntityBmpDesc(classLoader, ent, asd, jEnt);
                // container managed (always jdbc)
            } else if (ent.getPersistenceType().getValue().equals("Container")) {
                String cmpVer = (ent.getCmpVersion()==null)?"2.x":ent.getCmpVersion().getValue();
                bd = this.newBeanDesc(cmpVer, classLoader, ent, asd, jEnt);
            } else {
                throw new DeploymentDescException("Invalid persistence-type content for bean "+ent.getEjbName());
            }
            bd.check();
            bd.setDeploymentDesc(this);
            beanDesc.put(bd.getEjbName(),bd);
        }

        // HashMap of jonas-message-driven
        HashMap jonasMessageDriven = new HashMap();
        for (Iterator i = jonasEjbJar.getJonasMessageDrivenList().iterator(); i.hasNext(); ) {
            JonasMessageDriven jMd = (JonasMessageDriven)i.next();
                jonasMessageDriven.put(jMd.getEjbName().getValue(),jMd);
        }

        // message-driven beans
        for (Iterator i = ejbJar.getEnterpriseBeans().getMessageDrivenList().iterator(); i.hasNext(); ) {
            BeanDesc bd = null;
            MessageDriven md = (MessageDriven)i.next();
            /*
             * The Standard DTD allows a message-driven-bean without ejb-name
             * Problem to associate infos to this bean (trans-attribute, jonas-specific infos)
             * So the ejb-name must be specify and we check this
             */
            if (md.getEjbName()==null)
                throw new DeploymentDescException("ejb-name missing for a message driven bean");
            // find corresponding jonas message-driven
            JonasMessageDriven jMd = (JonasMessageDriven)jonasMessageDriven.get(md.getEjbName().getValue());
            if (jMd == null)
                throw new DeploymentDescException("jonas-message-driven-bean missing for bean "+md.getEjbName().getValue());
            bd = new MessageDrivenDesc(classLoader, md, asd, jMd);
            bd.check();
            bd.setDeploymentDesc(this);
            beanDesc.put(bd.getEjbName(), bd);
        }
    }


    /**
     * Get descriptors for all beans contained in jar file
     * @return Array of bean's descriptors
     */
    public BeanDesc[] getBeanDesc() {
        BeanDesc ret[] = new BeanDesc[beanDesc.size()];
        int j=0;
        for (Iterator i = beanDesc.values().iterator();i.hasNext();j++) {
            ret[j] = (BeanDesc)i.next();
        }
        return ret;
    }

    /**
     * Get bean descriptor given its name
     */
    public BeanDesc getBeanDesc(String ejbName) {
        return (BeanDesc)beanDesc.get(ejbName);
    }

    public BeanDesc newBeanDesc(String cmpversion,
                                ClassLoader classLoader,
                                Entity ent,
                                AssemblyDescriptor asd,
                                JonasEntity jEnt)
            throws DeploymentDescException {

        if (cmpversion.trim().equals("1.x")) {
            return new EntityJdbcCmp1Desc(classLoader, ent, asd, jEnt);
        } else {
            throw new DeploymentDescException("Invalid cmp-version content for bean "+ent.getEjbName());
        }
    }


    /**
     * Get the display name
     */
    public String getDisplayName() {
        return displayName;
    }

    /**
     * Get the role names of the deployment desc in the EAR case
     * @return an array of role name
     */
    public String[] getEarRoleNames() {
        return earRoleNames;
    }

    /**
     * Set the role names of the deployment desc in the EAR case
     * @param the role names
     */
     public void setEarRoleNames(String[] earRoleNames) {
         this.earRoleNames = earRoleNames;
    }

    /**
     * get the current logger
     */
    public Logger getLogger() {
        return logger;
    }

    /**
     * set the current logger
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        BeanDesc[] b = getBeanDesc();
        for (int i = 0; i < b.length; i++) {
            ret.append("\ngetBeanDesc("+i+")="+b[i].getClass().getName());
            ret.append(b[i].toString());
        }
        return ret.toString();
    }
}

