/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): Lutris Technologies Inc.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.deployment.api;

import org.objectweb.jonas_ejb.deployment.xml.EjbRelationshipRole;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbRelationshipRole;
import org.objectweb.jonas_ejb.deployment.xml.ForeignKeyJdbcMapping;
import org.objectweb.jonas_ejb.deployment.xml.CmrFieldType;

import java.util.Collection;
import java.util.Set;

/**
 * Class to hold meta-information related to an ejb-relation-role
 * Created on Jul 7, 2002
 * @author Christophe Ney [cney@batisseurs.com]
 */
public class EjbRelationshipRoleDesc {

    boolean cascadeDelete;
    boolean isRelationMultiple;
    String ejbName;
    EjbRelationDesc ejbRelationDesc;
    EntityCmp2Desc linkedEntityCmp2Desc;
    boolean target;  // is the relationship reverse
    protected String cmrFieldName = null;
    protected Class cmrFieldType = null;


    /**
     * constructor to be used by parent node
     */
    public EjbRelationshipRoleDesc(EjbRelationshipRole role, JonasEjbRelationshipRole jRole,
                                   boolean target,EjbRelationDesc ejbRelationDesc) throws DeploymentDescException {
        this.ejbRelationDesc = ejbRelationDesc;
        this.target=target;
        cascadeDelete = (role.getCascadeDelete()!=null)?true:false;
        ejbName = role.getRelationshipRoleSource().getEjbName().getValue();
        // mutiplicity is One or Many
        if (role.getMultiplicity().getValue().equals("Many")) {
            isRelationMultiple = true;
        } else if (role.getMultiplicity().getValue().equals("One")) {
            isRelationMultiple = false;
        } else {
            throw new DeploymentDescException("Invalid multiplicity value for relationshipRole \""
                    +role.getEjbRelationshipRoleName().getValue()+"\". (must be One or Many)");
        }
        // store cmr field if any
        if (role.getCmrField()!=null) {
            cmrFieldName = role.getCmrField().getCmrFieldName().getValue();
            if (isRelationMultiple) {
                CmrFieldType type = role.getCmrField().getCmrFieldType();
                if (type==null) {
                    throw new DeploymentDescException(
                        "You must specify a cmr-field-type in case where the relation is 'Many' in the cmr-field '"
                        + cmrFieldName + "' of bean " + ejbName);
                }
                try {
                    cmrFieldType = Class.forName(type.getValue());
                } catch (ClassNotFoundException e) {
                    throw new DeploymentDescException("class name not found for cmr-field "+cmrFieldName+" of bean "
                            +ejbName,e);
                }
                if (!(cmrFieldType.getName().equals("java.util.Collection")
                        || cmrFieldType.getName().equals("java.util.Set")))
                    throw new DeploymentDescException("value of cmr-field-type "+cmrFieldName+" of bean "
                            +ejbName+" should be java.util.Set or java.util.Collection if set");
            }
        }
    }

    /**
     * get state of field cascade delete
     */
    public boolean isCascadeDelete() {
        return cascadeDelete;
    }

    /**
     * get state of field is relation multiple
     */
    public boolean isRelationMultiple() {
        return isRelationMultiple;
    }

    /**
     * get name of the ejb involved in this relation-ship-role
     */
    public String getEjbName() {
        return ejbName;
    }


    /**
     * get the parent ejb relation
     */
    public EjbRelationDesc getEjbRelationDesc() {
        return ejbRelationDesc;
    }

    /**
     * It retrieves the EntityCmp2Desc which is linked to the EntityCmp2Desc
     * associated to this EjbRelationshipRoleDesc.
     */
    public EntityCmp2Desc getLinkedEntityCmp2Desc() {
        return linkedEntityCmp2Desc;
    }

    /**
     * set the linkedEntityCmp2Desc field.
     */
    public void setLinkedEntityCmp2Desc(EntityCmp2Desc linkedEntityCmp2Desc) {
        this.linkedEntityCmp2Desc = linkedEntityCmp2Desc;
        if (cmrFieldType==null) {
            cmrFieldType = linkedEntityCmp2Desc.getLocalClass();
        }
    }

    /**
     * It retrieves true if the EntityCmp2Desc associated to this
     * EjbRelationshipRoleDesc has a cmr field to the linked EntityCmp2Desc
     */
    public boolean hasCmrField() {
        return cmrFieldName!=null;
    }

    /**
     * It retrieves true if the EntityCmp2Desc associated to this
     * EjbRelationshipRoleDesc is the target of the cmr field associated the
     * linked EntityCmp2Desc.
     */
    public boolean isTarget() {
        return target;
    }

    /**
     * get the name of the cmr-field
     */
    public String getCmrFieldName() {
        return cmrFieldName;
    }

    /**
     * get the type of the cmr-field when set in the deployment descriptor.
     * @return Collection or Set for multiple rel. and null for non multiple rel.
     */
    public Class getCmrFieldType() {
        return cmrFieldType;
    }

   /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
   public String toString() {
       StringBuffer ret = new StringBuffer();
       ret.append("\nisCascadeDelete() = " + isCascadeDelete());
       ret.append("\nisRelationMultiple() = " + isRelationMultiple());
       ret.append("\ngetEjbName() = " + getEjbName());
       ret.append("\ngetEjbRelationDesc() = " + getEjbRelationDesc());
       ret.append("\ngetLinkedEntityCmp2Desc().getEjbName() = " + getLinkedEntityCmp2Desc().getEjbName());
       ret.append("\nisTarget() = " + isTarget());
       ret.append("\ngetCmrFieldName() = " + getCmrFieldName());
       ret.append("\ngetCmrFieldType() = " + getCmrFieldType());
       return ret.toString();
    }

}
