/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.deployment.api;

import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.Entity;
import org.objectweb.jonas_ejb.deployment.xml.JonasEntity;
import org.objectweb.jonas_ejb.deployment.xml.Query;
import org.objectweb.jonas_ejb.deployment.xml.MapperName;
import org.objectweb.jonas_ejb.deployment.ejbql.ParseException;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas_ejb.lib.MappingBuilder;
import org.objectweb.jonas_ejb.lib.RdbMappingBuilder;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.ArrayList;
import java.util.Map;
import java.util.List;

/**
 * Class to hold meta-information related to an entity of type CMP version 2.
 * Created on Jun 24, 2002
 * @author Christophe Ney [cney@batisseurs.com]
 */
public abstract class EntityCmp2Desc extends EntityCmpDesc {

    private ArrayList ejbRelationshipRoleDesc = new ArrayList();
    protected String abstractSchemaName;

    String[] mapperNames = null;
    private static final String JORM_SCHEMA_NAME = "jorm";

    private String jormName = null;
    private String jormAccessorClassName = null;
    private String jormBindingClassName = null;
    private String jormMappingClassName = null;
    private String jormCompositeName = null;
    private String jormPNameClassName = null;
    private String jormPNameGetterClassName = null;
    private String jormBinderClassName = null;

    /**
     * constructor to be used by parent node
     */
    public EntityCmp2Desc(ClassLoader classLoader, Entity ent,
                          AssemblyDescriptor asd, JonasEntity jEnt)
        throws DeploymentDescException {
        super(classLoader, ent, asd, jEnt);

		List l = jEnt.getMapperNameList();
		if (l != null || l.size()>=0) {
			mapperNames = new String[l.size()];
			int idx = 0;
			for(Iterator it = l.iterator(); it.hasNext();) {
				mapperNames[idx++] = ((MapperName) it.next()).getValue();
			}
		}

		// An abstract schema name is required
        if ((ent.getAbstractSchemaName() == null)
            || (ent.getAbstractSchemaName().getValue().length() == 0)) {
            throw new DeploymentDescException(
                "abstract-schema-name must be provided for bean " + this.ejbName);
        }
        abstractSchemaName = ent.getAbstractSchemaName().getValue();


        // check if persistent fields map to getter and setter
        for (Iterator i = fieldDesc.keySet().iterator(); i.hasNext();) {
            String fn = (String) i.next();

            // field should not be defined
            try {
                Field f = ejbClass.getField(fn);
                throw new DeploymentDescException("In cmp-version 2.x, field-name " + fn + " should not be defined in bean " + this.ejbName);
            }
            catch (NoSuchFieldException e) {
                // this is what we expect: nothing to do
            }
            catch (SecurityException e) {
                throw new DeploymentDescException("Cannot use java reflexion on " + this.ejbClass.getName());
            }

            try {
                Method getter = null;
                try {
                    getter = ejbClass.getMethod(getGetterName(fn), null);
                    ((FieldDesc) (fieldDesc.get(fn))).setFieldType(getter.getReturnType());
                }
                catch (NoSuchMethodException e) {
                    throw new DeploymentDescException("Getter method not found for field-name " + fn + " in bean " + this.ejbName, e);
                }
                try {
                    ejbClass.getMethod(getSetterName(fn), new Class[]{getter.getReturnType()});
                }
                catch (NoSuchMethodException e) {
                    throw new DeploymentDescException("Setter method not found for field-name " + fn + " in bean " + this.ejbName, e);
                }
            }
            catch (SecurityException e) {
                throw new DeploymentDescException("Cannot use java reflexion on " + this.ejbClass.getName());
            }
        }

        // isModifiedMethod deprecated for CMP 2.x
        if (jEnt.getIsModifiedMethodName() != null) {
            throw new DeploymentDescException("use of is-modified-method-name deprecated for CMP 2.x");
        }

        // EJB-QL query
        if (ent.getQueryList()!=null) {
            for (Iterator i = ent.getQueryList().iterator(); i.hasNext();) {
                Query q = (Query)i.next();
                boolean foundMatch = false;
                for (Iterator j=getMethodDescIterator(); j.hasNext();){
                    MethodDesc methd = (MethodDesc)j.next();
                    String methName = q.getQueryMethod().getMethodName().getValue();
                    if (methd.matchPattern(null,methName,q.getQueryMethod().getMethodParams())
                            != MethodDesc.APPLY_TO_NOTHING) {
                        foundMatch = true;
                        String query = q.getEjbQl().getValue();
                        if (!(methd instanceof MethodCmp2Desc))
                            throw new DeploymentDescException("ejbql query "+query+" can't apply to method "
                                    +methName+" in bean "+ejbName);

                        try {
                            ((MethodCmp2Desc)methd).setQuery(query);
                        } catch (ParseException e) {
                            throw new DeploymentDescException("Invalid ejbql syntax for bean "+ejbName+":\n"
                                    +e.getMessage(query));
                        }
                        if (q.getResultTypeMapping()!=null) {
                            ((MethodCmp2Desc)methd).setResultTypeMapping(q.getResultTypeMapping().getValue());
                        }
                    }
                }
                if (!foundMatch) {
                    throw new DeploymentDescException("invalid query-method definition for bean"+ejbName+"\nno such method as "
                            +MethodCmp2Desc.queryMethodElementToString(q.getQueryMethod())+"\ncheck method name and method parameters");
                }
            }
        }
        // check that all finder/selecter methods but findByPrimaryKey have a non null query
        for (Iterator j=getMethodDescIterator(); j.hasNext();) {
            MethodDesc md = (MethodDesc)j.next();
            if ((md.isFinder() || md.isSelect()) && !md.isFindByPrimaryKey()) {
                if (((MethodCmp2Desc)md).getQuery()==null) {
                    throw new DeploymentDescException("query not defined for method "+MethodDesc.toString(md.getMethod())
                            +" of bean"+ejbName);

                }
            }
        }

        // All JORM classes of an ejb-jar must be in same package
        // could be module name but module name not available in GenIC
        String packageName = JORM_SCHEMA_NAME;
        jormName = BeanNaming.getClassName(packageName, getAbstractSchemaName());
        jormAccessorClassName = BeanNaming.getClassName(packageName, getAbstractSchemaName() + "Accessor");
        jormBindingClassName = BeanNaming.getClassName(packageName, mapperNames[0] + "." + getAbstractSchemaName() + "Binding");
        jormMappingClassName = BeanNaming.getClassName(packageName, mapperNames[0] + "." + getAbstractSchemaName() + "Mapping");
        if (hasPrimaryKeyField()) {
            jormPNameClassName =
                    "org.objectweb.jorm.facility.naming.basidir.BasidName";
            jormBinderClassName =
                    "org.objectweb.jorm.facility.naming.basidir.BasidBinder";
        } else {
            jormCompositeName = BeanNaming.getClassName(packageName, getAbstractSchemaName() + "Pk");
            jormPNameClassName = BeanNaming.getClassName(packageName, getAbstractSchemaName() + "PkPName");
            jormPNameGetterClassName = BeanNaming.getClassName(packageName, getAbstractSchemaName() + "PkPNG");
            jormBinderClassName = BeanNaming.getClassName(packageName, getAbstractSchemaName() + "PkBinder");
        }
        factoryClassName = jormMappingClassName;
    }

    /**
     * getter for field abstractSchemaName
     */
    public String getAbstractSchemaName() {
        return abstractSchemaName;
    }

    /**
     * setter for field abstractSchemaName
     */
    public void setAbstractSchemaName(String abstractSchemaName) {
        this.abstractSchemaName = abstractSchemaName;
    }

    /**
     * Get descriptor for a given accessor
     * Used by GenIC
     * @param accessor the accessor method of the bean class
     * @return Descriptor for the given field
     */
    public FieldDesc getCmpFieldDesc(Method accessor) {
        FieldDesc ret = (FieldDesc) fieldDesc.get(getFieldName(accessor));
        if (ret == null)
            throw new Error("Field " + getFieldName(accessor) + " is not a cmp field of class " + this.ejbClass.getName());
        return ret;
    }

    /**
     * Get descriptor for a given field name
     * Used by GenIC
     * @param accessor the accessor method of the bean class
     * @return Descriptor for the given field or null
     */
    public FieldDesc getCmpFieldDesc(String fieldName) {
        return (FieldDesc) fieldDesc.get(fieldName);
    }

    /**
     * get iterator on meta information for all cmp fields
     */
    public Iterator getCmpFieldDescIterator() {
        return fieldDesc.values().iterator();
    }

    /**
     * get the java field name for a given accessor method
     * @param accessor the accessor set/get method
     * @return the name of the corresponding field
     */
    public String getFieldName(Method accessor) {
        String ret = accessor.getName();
        if (!ret.startsWith("get") && !ret.startsWith("set")) {
            throw new Error("Method " + ret + " is not an accessor method");
        }
        return Character.toLowerCase(ret.charAt(3)) + ret.substring(4);
    }

    /**
     * get the setter name for a given java field
     */
    public String getSetterName(String fieldName) {
        return "set" + Character.toUpperCase(fieldName.charAt(0)) + fieldName.substring(1);
    }

    /**
     * get the getter name for a given java field
     */
    public String getGetterName(String fieldName) {
        return "get" + Character.toUpperCase(fieldName.charAt(0)) + fieldName.substring(1);
    }

    /**
     * Assessor for a CMP field
     * @param accessor method for which a descriptor is to be returned
     * @return Descriptor for the given field
     */
    public boolean hasCmpFieldDesc(Method accessor) {
        return fieldDesc.containsKey(getFieldName(accessor));
    }

    /**
     * Add meta-info of a relationship in which the bean is involved
     */
    public void addEjbRelationshipRoleDesc(EjbRelationshipRoleDesc ersrd) {
        ejbRelationshipRoleDesc.add(ersrd);
    }

    /**
     * Get Iterator on meta-info for all relationships for which the bean
     * is involved in.
     */
    public Iterator getEjbRelationshipRoleDescIterator() {
        return ejbRelationshipRoleDesc.iterator();
    }


    /**
     * Factory method for MethodDesc.
     * Only method with EJBQL queries are of type <code>MethodCmp2Desc</code>
     */
    protected MethodDesc newMethodDescInstance(Method meth, int index) {
        if (!MethodDesc.isFindByPrimaryKey(meth)
              && (MethodDesc.isFinder(meth) || MethodDesc.isSelect(meth))) {
            return new MethodCmp2Desc(this,meth,index);
        } else {
            return new MethodDesc(this,meth,index);
        }
    }

    /**
     * It retrieves the class name of the generated Jorm accessor interface.
     */
    public String getJormAccessorClassName() {
        return jormAccessorClassName;
    }

    /**
     * It retrieves the class name of the generated Jorm binding.
     */
    public String getJormBindingClassName() {
        return jormBindingClassName;
    }

    /**
     * It retrieves the class name of the generated Jorm mapping.
     */
    public String getJormMappingClassName() {
        return jormMappingClassName;
    }

    /**
     * It retrieves the class name of the generated Jorm PName.
     */
    public String getJormPNameClassName() {
        return jormPNameClassName;
    }

    /**
     * It retrieves the class name of the generated Jorm PNameGetter interface.
     */
    public String getJormPNameGetterClassName() {
        return jormPNameGetterClassName;
    }

    /**
     * It retrieves the class name of the generated Jorm Binder.
     */
    public String getJormBinderClassName() {
        return jormBinderClassName;
    }

    /**
     * fully qualified name of the class in the Jorm
     */
    public String getJormName() {
        return jormName;
    }

    /**
     * Name of the composite name for this class in Jorm
     */
    public String getJormCompositeName() {
        return jormCompositeName;
    }


    /**
     * Name of the schema (package) of this class in Jorm
     */
    public String getJormSchemaName() {
        return JORM_SCHEMA_NAME;
    }

    /**
     * return the mapping builder to use for the entity bean
     */
    public abstract MappingBuilder getMappingBuilder();

	public String[] getMapperNames() {
		return mapperNames;
	}
    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(super.toString());
        for (Iterator i=ejbRelationshipRoleDesc.iterator();i.hasNext();) {
            ret.append("\nejbRelationshipRoleDesc[]="+i.next());
        }
        ret.append("\ngetAbstractSchemaName()="+getAbstractSchemaName());
        ret.append("\ngetJormName() = " + getJormName());
        ret.append("\ngetJormAccessorClassName() = " + getJormAccessorClassName());
        ret.append("\ngetJormBindingClassName() = " + getJormBindingClassName());
        ret.append("\ngetJormMappingClassName() = " + getJormMappingClassName());
        ret.append("\ngetJormPNameGetterClassName() = " + getJormPNameGetterClassName());
        ret.append("\ngetJormBinderClassName() = " + getJormBinderClassName());
        return ret.toString();
    }

}

