/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.deployment.api;

import org.objectweb.jonas_ejb.deployment.xml.*;
import org.objectweb.jonas_ejb.lib.MappingBuilder;
import org.objectweb.jonas_ejb.lib.RdbMappingBuilder;

import java.util.Iterator;
import java.util.Enumeration;
import java.lang.reflect.Method;
import java.lang.reflect.Field;

/**
 * Class to hold meta-information related to an CMP v2 entity bean with jdbc data store.
 * Created on Jun 24, 2002
 * @author Christophe Ney [cney@batisseurs.com]
 */
public class EntityJdbcCmp2Desc extends EntityCmp2Desc {

    String jdbcTableName;
    String datasourceJndiName;

    /**
     * constructor to be used by parent node
     */
    public EntityJdbcCmp2Desc(ClassLoader classLoader, Entity ent,
                              AssemblyDescriptor asd, JonasEntity jEnt)
            throws DeploymentDescException {
        super(classLoader,ent,asd,jEnt);

        // check for jdbcMapping
        JdbcMapping jm = jEnt.getJdbcMapping();
        if (jm==null)
            throw new DeploymentDescException("jdbc-mapping missing for bean "+ent.getEjbName());

        // populate field descriptor map with jdbc information
        for (Iterator i= jm.getCmpFieldJdbcMappingList().iterator(); i.hasNext(); ) {
            CmpFieldJdbcMapping cmpFieldJdbcMapping = (CmpFieldJdbcMapping)i.next();
            String fn = cmpFieldJdbcMapping.getFieldName().getValue();
            FieldDesc f = (FieldDesc)fieldDesc.get(fn);
            if (f==null)
                throw new DeploymentDescException("field-name "+fn
                    +" listed in cmp-field-jdbc-mapping is not of cmp-field of bean "+ent.getEjbName());
            ((FieldJdbcDesc) f).setJdbcFieldName(cmpFieldJdbcMapping.getJdbcFieldName().getValue());
        }

        // check that all cmp fields are jdbc
        for (Iterator j = fieldDesc.values().iterator();j.hasNext();) {
            FieldJdbcDesc fd = (FieldJdbcDesc)j.next();
            if (fd.getJdbcFieldName()==null) {
                throw new DeploymentDescException("field-name "+fd.getFieldName()
                        +" is missing in cmp-field-jdbc-mapping for bean "+ent.getEjbName());
            }
        }

        // jndi name of the datasource
        datasourceJndiName = jm.getJndiName().getValue();

        //table name
        jdbcTableName=jm.getJdbcTableName().getValue();
    }


    /**
     * Get the datasource jndi name
     * @return String representation of the jndi name
     */
    public String getDatasourceJndiName(){
        return datasourceJndiName;
    }

    /**
     * Assessor for existence of a jdbc descriptor for a given field
     * @param accessor public accessor method of the bean class
     * @return true if jdbc descriptor is available for given field
     */
    public boolean hasFieldJdbcDesc(Method accessor) {
        return super.hasCmpFieldDesc(accessor);
    }


    /**
     * Get jdbc specific descriptor for a given field.
     * Used by GenIC
     * @param accessor public accessor method of the bean class
     * @return Descriptor for the given field
     */
    public FieldJdbcDesc getFieldJdbcDesc(Method accessor) {
        return (FieldJdbcDesc)super.getCmpFieldDesc(accessor);
    }

    /**
     * Get the associated DataBase table name in case of container persistence type.
     * Used by GenIC (This information is JOnAS specific).
     * @return Name of the database table where entity bean is stored
     */
    public String getJdbcTableName(){
        return jdbcTableName;
    }

    /**
     * field descriptor factory method
     */
    protected FieldDesc newFieldDescInstance() {
        return new FieldJdbcDesc();
    }


    /**
     * return the mapping builder to use for the entity bean
     */
    public MappingBuilder getMappingBuilder() {
        return new RdbMappingBuilder();
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(super.toString());
        ret.append("\ngetDatasourceJndiName()="+getDatasourceJndiName());
        ret.append("\ngetJdbcTableName()="+getJdbcTableName());
        return ret.toString();
    }


}
