/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): Lutris Tecnologies Inc.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.deployment.api;

import java.lang.reflect.Method;
import org.objectweb.jonas_ejb.deployment.xml.Session;
import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.JonasSession;


/**
 * Base class to hold meta-information related to a session bean.
 * @author Christophe Ney [cney@batisseurs.com]
 */
public abstract class SessionDesc extends BeanDesc {

    protected int transactionType;
    int sessionTimeout = 0;
   
    /**
     * constructor to be used by parent node
     */
    public SessionDesc(ClassLoader classLoader, Session ses, AssemblyDescriptor asd, JonasSession jSes)
            throws DeploymentDescException {

        super(classLoader, ses.getDisplayName(), ses.getEjbName(), jSes.getJndiName(),
              ses.getHome(), ses.getRemote(), ses.getLocalHome(), ses.getLocal(), ses.getEjbClass(),
              ses.getEnvEntryList(), ses.getResourceRefList(), jSes.getJonasResourceList(),
              ses.getResourceEnvRefList(), jSes.getJonasResourceEnvList(),
              ses.getEjbRefList(), jSes.getJonasEjbRefList(), ses.getEjbLocalRefList(), ses.getSecurityRoleRefList(), asd);

        // session timeout
        if (jSes.getSessionTimeout() != null) {
            String tstr = jSes.getSessionTimeout().getValue();
            Integer tval = new Integer(tstr);
            sessionTimeout = tval.intValue();
        }

        // bean or container managed transaction
        if (ses.getTransactionType().getValue().equals("Bean")) {
            transactionType = BEAN_TRANSACTION_TYPE;
        } else if (ses.getTransactionType().getValue().equals("Container")) {
            transactionType = CONTAINER_TRANSACTION_TYPE;
        } else {
            throw new DeploymentDescException("Invalid transaction-type content for ejb-name "+ejbName);
        }
    }

    /**
     * check that trans-attribute is valid for bean
     */
    protected void checkTxAttribute(MethodDesc md) throws DeploymentDescException {
        java.lang.reflect.Method m = md.getMethod();
        if (getTransactionType() == CONTAINER_TRANSACTION_TYPE) {
            // tx-attribute must be set for methods of remote interface
            // excluding the methods of the javax.ejb.EJBObject interface
            if ((md.getTxAttribute() == MethodDesc.TX_NOT_SET) &&
                javax.ejb.EJBObject.class.isAssignableFrom(m.getDeclaringClass()) &&
                !javax.ejb.EJBObject.class.equals(m.getDeclaringClass())) {
                throw new DeploymentDescException("trans-attribute missing for method "
                                                  +m.toString()+" in session bean "+getEjbName());
	    }
            // tx-attribute must not be set for methods of home interface
            if ((md.getTxAttribute() != MethodDesc.TX_NOT_SET) &&
		javax.ejb.EJBHome.class.isAssignableFrom(m.getDeclaringClass()) &&
                ((md.getTxAttributeStatus() == MethodDesc.APPLY_TO_CLASS) ||
		 (md.getTxAttributeStatus() == MethodDesc.APPLY_TO_CLASS_METHOD_NAME) ||
		 (md.getTxAttributeStatus() == MethodDesc.APPLY_TO_CLASS_METHOD))) {
                throw new DeploymentDescException("trans-attribute must not be specified for home interface's method "
                                                  +m.toString()+" in session bean "+getEjbName());
            }
        } else {
            if (md.getTxAttribute() != MethodDesc.TX_NOT_SET)
                throw new DeploymentDescException(md.getTxAttributeName()+" is not a valid trans-attribute for method "+m.toString()+" in session bean "+getEjbName());
        }
    }
        
    /**
     * Get session transaction management type.<BR>
     * @return transaction type value within BEAN_TRANSACTION_TYPE,CONTAINER_TRANSACTION_TYPE
     */
    public int getTransactionType() {
        return transactionType;
    }

    /**
     * Returns true if bean managed transaction. (used by JOnAS Server)
     */
    public boolean isBeanManagedTransaction() {
	return (transactionType == BEAN_TRANSACTION_TYPE);
    }
    
    /**
     * Get the session timeout value
     */
    public int getSessionTimeout() {
        return sessionTimeout;
    }

    /**
     * Check that the bean descriptor is valid
     * @exception DeploymentDescException thrown for non-valid bean
     */
    public void check() throws DeploymentDescException {
        super.check();
        // for BEAN_TRANSACTION_TYPE transactions ejbClass should not implement javax.ejb.SessionSynchronization
        if ((getTransactionType()==BEAN_TRANSACTION_TYPE) && (javax.ejb.SessionSynchronization.class.isAssignableFrom(ejbClass)))
            throw new DeploymentDescException(ejbClass.getName()+" should NOT manage transactions and implement javax.ejb.SessionSynchronization");
    }
    
    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(super.toString());
        ret.append("\ngetTransactionType()"+TRANS[getTransactionType()]);
        ret.append("\nsessionTimeout = " + sessionTimeout);
        return ret.toString();
    }
    
}

