/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.lib;

import org.objectweb.jonas_ejb.deployment.api.DeploymentCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.MethodCmp2Desc;
import org.objectweb.jonas_ejb.deployment.ejbql.ASTEJBQL;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.Schema;
import org.objectweb.medor.eval.api.QueryEvaluator;
import org.objectweb.medor.eval.lib.MedorEvaluator;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.optim.api.QueryTransformer;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.lib.SelectProject;

import java.util.Iterator;
import java.util.Map;

/**
 * Class to hold the query tree of a given EJBQL request. This allows walking the
 * JormExtents of the tree to set the mappers.
 * @author  Christophe Ney - cney@batisseurs.com
 */
public class EjbqlQueryTreeHolder {

    QueryTree queryTree = null;
    QueryTransformer queryOptimizer = null;
    QueryEvaluator optimizedQuery = null;

    private String beanName;
    private String methodName;


    /**
     * constructor of a holder for a specific query.
     * @param mDesc JOnAS meta-information for the corresponding finder/select method
     * @param ejbql root of the lexical tree of the query
     */
    public EjbqlQueryTreeHolder(MethodCmp2Desc mDesc, ASTEJBQL ejbql) throws Exception {

        // get schema (FIXME we use current bean schema)
        EntityCmp2Desc eDesc = (EntityCmp2Desc) mDesc.getBeanDesc();
        String schemaName = eDesc.getJormSchemaName();

        Manager jormManager = ((DeploymentCmp2Desc) eDesc.getDeploymentDesc()).getManager();
        Schema sc = jormManager.getSchema(schemaName);

        // create identifiers Map (FROM clause)
        queryTree = new EjbqlVariableVisitor(ejbql, sc).getQueryTree();

        // create SelectProject (SELECT clause)
        SelectProject sp = new EjbqlSelectVisitor(ejbql, queryTree).getSelectProject();

        Expression qf = new EjbqlQueryFilterVisitor(queryTree, mDesc.getMethod().getParameterTypes(), ejbql).getQueryFilter();
        sp.setQueryFilter(qf);
        queryTree = sp;

        beanName = eDesc.getAbstractSchemaName();
        methodName = mDesc.getMethod().getName();
/*
        System.out.println("QueryTree of the method " + methodName
            + " (bean "+ beanName+")");
        TestQueryTreeHelper.debug = true;
        TestQueryTreeHelper.printQueryTree(queryTree);
*/
    }

    /**
     * get the query tree that was built from visiting the lexical tree
     */
    public QueryTree getQueryTree() {
        return queryTree;
    }

    /**
     * Set the optimizer to be used when optimizing the query tree
     */
    public void setQueryOptimizer(QueryTransformer qtf) {
        optimizedQuery = null;
        queryOptimizer = qtf;
    }

    /**
     * get the current query tree optimizer
     */
    public QueryTransformer getQueryOptimizer() {
        return queryOptimizer;
    }

    /**
     * Get the query evaluator of the optimized query. The query is optimized on first call.
     * All mappers of the query should have been already set with the help of the iterator
     * returned by the <code>getJormExtentIterator</code> method.
     */
    public QueryEvaluator getOptimizedQueryTree() throws Exception {
        if (optimizedQuery == null) {
            QueryTree qt = queryOptimizer.transform(queryTree);
            System.out.println("OPTIMIZED QueryTree of the method " + methodName + " (bean " + beanName + ")");
//			TestQueryTreeHelper.debug = true;
//            TestQueryTreeHelper.printQueryTree(qt);
            optimizedQuery = new MedorEvaluator(qt, 0);
        }
        return optimizedQuery;
    }

}
