/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_web.deployment.api;


//import java
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;




//import jonas_lib
import org.objectweb.jonas_lib.deployment.AbsDeploymentDesc;
import org.objectweb.jonas_lib.deployment.SAXParserHelper;

//import jonas_web
import org.objectweb.jonas_web.deployment.xml.WebApp;
import org.objectweb.jonas_web.deployment.xml.JonasWebApp;
import org.objectweb.jonas_web.deployment.xml.JonasWebAppImpl;
import org.objectweb.jonas_web.deployment.xml.JonasResourceEnv;
import org.objectweb.jonas_web.deployment.xml.ResourceEnvRef;
import org.objectweb.jonas_web.deployment.xml.JonasResource;
import org.objectweb.jonas_web.deployment.xml.ResourceRef;
import org.objectweb.jonas_web.deployment.xml.JonasEjbRef;
import org.objectweb.jonas_web.deployment.xml.EjbRef;
import org.objectweb.jonas_web.deployment.xml.EjbLocalRef;
import org.objectweb.jonas_web.deployment.xml.EnvEntry;
import org.objectweb.jonas_web.deployment.xml.WebAppUnmarshaller;
import org.objectweb.jonas_web.deployment.xml.JonasWebAppUnmarshaller;

//import SAX
import org.xml.sax.ErrorHandler;
import org.xml.sax.EntityResolver;

/**
 * This class do the parsing of the web.xml file and jonas-web.xml files and
 * contruct a data structure associated to these two files.
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public class WebContainerDeploymentDesc extends AbsDeploymentDesc {

    /** 
     * The path to the web.xml file. 
     */
    public static final String WEB_FILE_NAME = "WEB-INF/web.xml";

    /** 
     * The path to the jonas-web.xml file. 
     */
    public static final String JONAS_WEB_FILE_NAME = "WEB-INF/jonas-web.xml";

    /** 
     * Associates a JNDI name to a name of a resource env ref.
     */
    private Vector resourceEnvRefs = new Vector();

    /**
     * Associates a JNDI name to a name of a resource ref.
     */
    private Vector resourceRefs = new Vector();

    /**
     * List of the env-entry
     */
    private Vector envEntries = new Vector();

    /**
     * Associates a JNDI name to a name of an EJB ref.
     */
    private Vector ejbRefs = new Vector();

    /**
     * Associates a JNDI name to a name of a local EJB ref.
     */
    private Vector ejbLocalRefs = new Vector();

    /**
     * The host on which the web application must be depoyed (can be null in
     * ear case).
     */
    private String host = null;

    /**
     * The context root of the web application (can be null in ear case).
     */
    private String contextRoot = null;

    /** 
     *  The error handler to use
     */
    private static ErrorHandler errorHandler = null;

    /** 
     * The entity resolver to use
     */
    private static EntityResolver entityResolver = null;

    /**
     * Flag for parser validation
     */
    private static boolean parsingWithValidation = true;


    /**
     * Construct an instance of a WebContainerDeploymentDesc.<BR>
     * Constructor is private, call one of the static getInstance 
     * method instead.
     * @param classLoader the classloader for the classes.
     * @param webApp the data structure of the web-app (web.xml)
     * @param jonasWebApp the data structure of the jonas-web-app 
     * (jonas-web.xml)
     * @throws WebContainerDeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    private WebContainerDeploymentDesc(ClassLoader classLoader,
                                       WebApp webApp, JonasWebApp jonasWebApp) 
        throws WebContainerDeploymentDescException {
 
        // test classloader
        if (classLoader == null) {
            throw new WebContainerDeploymentDescException("DeploymentDesc: Classloader is null");
        }

        // display name
        displayName = (webApp.getDisplayName() != null) ? webApp.getDisplayName().getValue() : null;
        // host
        host = (jonasWebApp.getHost() != null) ? jonasWebApp.getHost().getValue() : null;
        // context-root
        contextRoot = (jonasWebApp.getContextRoot() != null) ? jonasWebApp.getContextRoot().getValue() : null;
 
        // resource-env-ref
        List resourceEnvRefList = webApp.getResourceEnvRefList();
        List jonasResourceEnvList = jonasWebApp.getJonasResourceEnvList();
        Hashtable resourceEnv = new Hashtable();
        for (Iterator i = jonasResourceEnvList.iterator(); i.hasNext();) {
            JonasResourceEnv jonasResourceEnv = (JonasResourceEnv) i.next();
            String rName = jonasResourceEnv.getResourceEnvRefName().getValue();
            boolean findResRef = false;
            for (Iterator j = resourceEnvRefList.iterator(); j.hasNext() && !findResRef;) {
                ResourceEnvRef resourceEnvRef = (ResourceEnvRef) j.next();
                if (rName.equals(resourceEnvRef.getResourceEnvRefName().getValue())) {
                    findResRef = true;
                }
            }
            if (!findResRef) {
                throw new WebContainerDeploymentDescException("resource-env-ref missing for jonas-resource-env " + rName);
            }
            resourceEnv.put(rName, jonasResourceEnv);
        }
        for (Iterator i = resourceEnvRefList.iterator(); i.hasNext();) {
            ResourceEnvRef resourceEnvRef = (ResourceEnvRef) i.next();
            String rName = resourceEnvRef.getResourceEnvRefName().getValue();
            if (!resourceEnv.containsKey(rName)) {
                throw new WebContainerDeploymentDescException("jonas-resource-env missing for resource-env-ref-name " + rName);
            }
            resourceEnvRefs.addElement(new ResourceEnvRefDesc(classLoader, resourceEnvRef, (JonasResourceEnv) resourceEnv.get(rName)));
        }


        // resource-ref
        List resourceRefList = webApp.getResourceRefList();
        List jonasResourceList = jonasWebApp.getJonasResourceList();
        Hashtable resource = new Hashtable();
        for (Iterator i = jonasResourceList.iterator(); i.hasNext();) {
            JonasResource jonasResource = ((JonasResource) i.next());
            String rName = jonasResource.getResRefName().getValue();
            boolean findResRef = false;
            for (Iterator j = resourceRefList.iterator(); j.hasNext() && !findResRef;) {
                if (rName.equals(((ResourceRef) j.next()).getResRefName().getValue())) {
                    findResRef = true;
                }
            }
            if (!findResRef) {
                throw new WebContainerDeploymentDescException("resource-ref missing for jonas-resource " + rName);
            }
            resource.put(rName, jonasResource);
        }

        for (Iterator i = resourceRefList.iterator(); i.hasNext();) {
            ResourceRef resourceRef = ((ResourceRef) i.next());
            String rName = resourceRef.getResRefName().getValue();
            if (!resource.containsKey(rName)) {
                throw new WebContainerDeploymentDescException("jonas-resource missing for res-ref-name " + rName);
            }
            resourceRefs.addElement(new ResourceRefDesc(classLoader, resourceRef, (JonasResource) resource.get(rName)));
        }

        // env-entry
        for (Iterator i = webApp.getEnvEntryList().iterator(); i.hasNext();) {
            EnvEntryDesc envEntryDesc = new EnvEntryDesc((EnvEntry) i.next());
            envEntries.addElement(envEntryDesc);
        }

        // ejb-ref
        List ejbRefList = webApp.getEjbRefList();
        List jonasEjbRefList = jonasWebApp.getJonasEjbRefList();

        Hashtable ejbRef = new Hashtable();
        for (Iterator i = jonasEjbRefList.iterator(); i.hasNext();) {
            JonasEjbRef jonasEjbRef = (JonasEjbRef) i.next();
            String brName = jonasEjbRef.getEjbRefName().getValue();
            boolean findBRef = false;
            for (Iterator j = ejbRefList.iterator(); j.hasNext() && !findBRef;) {
                EjbRef eRef = (EjbRef) j.next();
                if (brName.equals(eRef.getEjbRefName().getValue())) {
                    findBRef = true;
                }
            }
            if (!findBRef) {
                throw new WebContainerDeploymentDescException("ejb-ref missing for jonas-ejb-ref " + brName);
            }
            ejbRef.put(brName, jonasEjbRef);
        }
        for (Iterator i = ejbRefList.iterator(); i.hasNext();) {
            EjbRef eRef = (EjbRef) i.next();
            String brName = eRef.getEjbRefName().getValue();
            JonasEjbRef jonasEjbRef = (JonasEjbRef) ejbRef.get(brName);
            if (jonasEjbRef == null) {
                // The ejb-link must be specified when there is no jonas-ejb-ref
                if ((eRef.getEjbLink() == null)) {
                    // Ignore the ejb-link syntax "product.jar#ProductEJB"
                    throw new WebContainerDeploymentDescException("ejb-link missing for ejb-ref-name " + brName);
                }
            }
            ejbRefs.addElement(new EjbRefDesc(eRef, jonasEjbRef));
        }

        // ejb-local-ref
        List ejbLocalRefList = webApp.getEjbLocalRefList();
        for (Iterator i = ejbLocalRefList.iterator(); i.hasNext();) {
            EjbLocalRef ejbLocalRef = (EjbLocalRef) i.next();
            if ((ejbLocalRef.getEjbLink() == null)) {
                throw new WebContainerDeploymentDescException("ejb-link missing for ejb-ref-name " + ejbLocalRef.getEjbRefName().getValue());
            }
            ejbLocalRefs.addElement(new EjbLocalRefDesc(ejbLocalRef));
        }
    }

    /** ontainerDeploymentDescException("ejb-lin
     * Get an instance of a WEB deployment descriptor by parsing the web.xml 
     * and jonas-web.xml deployment descriptors.
     * @param warFileName the fileName of the war file for the deployment 
     * descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @param altWebXmlFilename the fileName to the web.xml for the alt-dd
     * tag in the Ear Case. This is used for specify an alternate DDesc file.
     * @return a WEB deployment descriptor by parsing the web.xml and 
     * jonas-web.xml deployment descriptors.
     * @throws WebContainerDeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    public static WebContainerDeploymentDesc getInstance(String warFileName, ClassLoader classLoaderForCls,
                                                         String altWebXmlFilename)
        throws WebContainerDeploymentDescException {
 
        //war file
        JarFile warFile = null;

        //Input streams
        InputStream webInputStream = null;
        InputStream jonasWebInputStream = null;
        
        //ZipEntry
        ZipEntry webZipEntry = null;
        ZipEntry jonasWebZipEntry = null;

        //Webapps
        WebApp webApp;
        JonasWebApp jonasWebApp;

        //Check if the file exists.
        if (!new File(warFileName).exists()) {
            String err = "The file '" + warFileName + "' was not found.";
            throw new WebContainerDeploymentDescException(err);
        }

        //Check if the Alt deploymentDesc file exists.
        //But only if it's a non null value because it's optionnal.
        if ((altWebXmlFilename != null) && (!new File(altWebXmlFilename).exists())) {
            String err = "The file for the altdd tag for the EAR case '" + altWebXmlFilename + "' was not found.";
            throw new WebContainerDeploymentDescException(err);
        }


        // load the web-app deployment descriptor data (WEB-INF/web.xml 
        // and WEB-INF/jonas-web.xml)
        try {
            warFile = new JarFile(warFileName);

            
            //No alt-dd case
            if (altWebXmlFilename == null) {
                
                //Check the web entry
                webZipEntry = warFile.getEntry(WEB_FILE_NAME);
                if (webZipEntry == null) {
                    throw new WebContainerDeploymentDescException("The entry '" + WEB_FILE_NAME + "' was not found in the file '" + warFileName + "'.");
                }
                //Get the stream
                webInputStream = warFile.getInputStream(webZipEntry);
            } else {
                webInputStream = new FileInputStream(altWebXmlFilename);
            }

            //Check the jonas web entry
            jonasWebZipEntry = warFile.getEntry(JONAS_WEB_FILE_NAME);

            //Get the stream
            if (jonasWebZipEntry != null) {
		jonasWebInputStream = warFile.getInputStream(jonasWebZipEntry);
	    }
        } catch (Exception e) {
            if (warFile != null) {
                try {
                    warFile.close();
                } catch (IOException ioe) {
                    //We can't close the file
                }
            }
            throw new WebContainerDeploymentDescException("Cannot read the XML deployment descriptors of the war file '" + warFileName + "'.", e);
        }

        webApp = loadWebApp(webInputStream, WEB_FILE_NAME);
        try {
            webInputStream.close();
        } catch (IOException e) {
            // Nothing to do
        }




        // load jonas-web-app deployment descriptor data 
        // (WEB-INF/jonas-web.xml)
        if (jonasWebInputStream != null) {
	    jonasWebApp = loadJonasWebApp(jonasWebInputStream, JONAS_WEB_FILE_NAME);
	    try {
		jonasWebInputStream.close();
	    } catch (IOException e) {
		// Nothing to do
	    }
	} else {
	    jonasWebApp = new JonasWebAppImpl();
	}


        // instantiate web deployment descriptor
        return new WebContainerDeploymentDesc(classLoaderForCls, webApp, jonasWebApp);
    }

    /** ontainerDeploymentDescException("ejb-lin
     * Get an instance of a WEB deployment descriptor by parsing the web.xml 
     * and jonas-web.xml deployment descriptors.
     * @param warFileName the fileName of the war file for the deployment 
     * descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @return a WEB deployment descriptor by parsing the web.xml and 
     * jonas-web.xml deployment descriptors.
     * @throws WebContainerDeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    public static WebContainerDeploymentDesc getInstance(String warFileName, ClassLoader classLoaderForCls)
        throws WebContainerDeploymentDescException {

        return getInstance(warFileName, classLoaderForCls, null);
    }

    /** 
     * Load the web.xml file.
     * @param is the InputStream of the XML file.
     * @param fileName the name of the file (web.xml).
     * @return a structure containing the result of the web.xml parsing.
     * @throws WebContainerDeploymentDescException if the deployment descriptor 
     * is corrupted.
     */
    private static WebApp loadWebApp(InputStream is, String fileName) 
        throws WebContainerDeploymentDescException {
        
        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper = new SAXParserHelper(errorHandler, entityResolver);
        try {
            WebAppUnmarshaller.setErrorHandler(helper.getWrappedErrorHandler());
            WebAppUnmarshaller.setEntityResolver(helper.getWrappedEntityResolver());
            // Unmarshal the XML, and validate as it is unmarshalled
            return WebAppUnmarshaller.unmarshal(reader, parsingWithValidation);
        } catch (IOException e) {
            if (helper.getSAXParseException() != null) {
                String detailMsg = getSAXMsg(fileName, helper.getSAXParseException(), "Invalid XML");
                throw new WebContainerDeploymentDescException(detailMsg, e);
            } else {
                throw new WebContainerDeploymentDescException("Error reading XML document " + fileName, e);
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignored) { }
            }
        } 
    }

    /**
     * Load the jonas-web.xml file.
     * @param is the InputStream of the XML file.
     * @param fileName the name of the file (jonas-web.xml).
     * @return a structure containing the result of the jonas-web.xml parsing.
     * @throws WebContainerDeploymentDescException if the deployment 
     * descriptor is corrupted.
     */
    private static JonasWebApp loadJonasWebApp(InputStream is, 
                                               String fileName) 
        throws WebContainerDeploymentDescException {

        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper = new SAXParserHelper(errorHandler, entityResolver);
        try {
            JonasWebAppUnmarshaller.setErrorHandler(helper.getWrappedErrorHandler());
            JonasWebAppUnmarshaller.setEntityResolver(helper.getWrappedEntityResolver());
            // Unmarshal the XML, and validate as it is unmarshalled
            return JonasWebAppUnmarshaller.unmarshal(reader, parsingWithValidation);
        } catch (IOException e) {
            if (helper.getSAXParseException() != null) {
                String detailMsg = getSAXMsg(fileName, helper.getSAXParseException(), "Invalid XML");
                throw new WebContainerDeploymentDescException(detailMsg, e);
            } else {
                throw new WebContainerDeploymentDescException("Error reading XML document " + fileName, e);
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignored) { }
            }
        }
    }

    /**
     * Get an array of mapping between a res-env-ref name and its JNDI name.
     * @return an array of mapping between a res-env-ref name and its 
     * JNDI name.
     */
    public ResourceEnvRefDesc[] getResourceEnvRefTags() {
        ResourceEnvRefDesc ret[] = new ResourceEnvRefDesc[resourceEnvRefs.size()];
        resourceEnvRefs.copyInto(ret);
        return ret;
    }

    /**
     * Get an array of mapping between an res-ref name and its JNDI name.
     * @return an array of mapping between an res-ref name and its JNDI name.
     */
    public ResourceRefDesc[] getResourceRefTags() {
        ResourceRefDesc ret[] = new ResourceRefDesc[resourceRefs.size()];
        resourceRefs.copyInto(ret);
        return ret;
    }

    /**
     * Get an array of mapping between an entry name and its value.
     * @return an array of mapping between an entry name and its value.
     */
    public EnvEntryDesc[] getEnvEntryTags() {
        EnvEntryDesc ret[] = new EnvEntryDesc[envEntries.size()];
        envEntries.copyInto(ret);
        return ret;
    }

    /**
     * Get an array of mapping between an ejb name and its JNDI name.
     * @return an array of mapping between an ejb name and its JNDI name.
     */
    public EjbRefDesc[] getEjbRefTags() {
        EjbRefDesc ret[] = new EjbRefDesc[ejbRefs.size()];
        ejbRefs.copyInto(ret);
        return ret;
    }

    /**
     * Get an array of mapping between an ejb local name and its JNDI name.
     * @return an array of mapping between an ejb name local and its JNDI name.
     */
    public EjbLocalRefDesc[] getEjbLocalRefTags() {
        EjbLocalRefDesc ret[] = new EjbLocalRefDesc[ejbLocalRefs.size()];
        ejbLocalRefs.copyInto(ret);
        return ret;
    }

    /**
     * Get the context root of this web application.
     * @return the context root of this web application.
     */
    public String getContextRoot() {
        return contextRoot;
    }

    /**
     * Get the host on which the web application must be deployed.
     * @return the host on which the web application must be deployed.
     */
    public String getHost() {
        return host;
    }

    /**
     * Return a String representation of the WebContainerDeploymentDesc.
     * @return a String representation of the WebContainerDeploymentDesc.
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        
        // Return the displayName
        ret.append("\ngetDisplayName()=" + getDisplayName());
        
        // Return the resource-env-ref
        ResourceEnvRefDesc[] rer = getResourceEnvRefTags();
        for (int i = 0; i < rer.length; i++) {
            ret.append("\ngetResourceEnvRefTags(" + i + ")=" + rer[i].getClass().getName());
            ret.append(rer[i].toString());
        }

        // Return the resource-ref
        ResourceRefDesc[] resourceRefDesc = getResourceRefTags();
        for (int i = 0; i < resourceRefDesc.length; i++) {
            ret.append("\ngetResourceRefTags(" + i + ")=" + resourceRefDesc[i].getClass().getName());
            ret.append(resourceRefDesc[i].toString());
        }
 
        // Return the env-entry
        EnvEntryDesc[] envEntries = getEnvEntryTags();
        for (int i = 0; i < envEntries.length; i++) {
            ret.append("\ngetEnvEntryTags(" + i + ")=" + envEntries[i].getClass().getName());
            ret.append(envEntries[i].toString());
        }

        // Return the ejb-ref
        EjbRefDesc[] ejbRefDesc = getEjbRefTags();
        for (int i = 0; i < ejbRefDesc.length; i++) {
            ret.append("\ngetEjbRefTags(" + i + ")=" + ejbRefDesc[i].getClass().getName());
            ret.append(ejbRefDesc[i].toString());
        }
        
        // Return the ejb-local-ref
        EjbLocalRefDesc[] ejbLocalRefDesc = getEjbLocalRefTags();
        for (int i = 0; i < ejbLocalRefDesc.length; i++) {
            ret.append("\ngetEjbLocalRefTags(" + i + ")=" + ejbLocalRefDesc[i].getClass().getName());
            ret.append(ejbLocalRefDesc[i].toString());
        }

        // Return the host
        ret.append("\ngetHost()=" + getHost());

        // Return the context-root
        ret.append("\ngetContextRoot()=" + getContextRoot());
      
        return ret.toString();
    }

    /**
     * Get the current ErrorHandler.
     * @return the current ErrorHandler.
     */
    public static ErrorHandler getErrorHandler() {
        return errorHandler;
    }
    
    /**
     * Get the current EntityResolver.
     * @return the current EntityResolver.
     */
    public static EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @return if true, all external entities will be read.
     */
    public static boolean getParsingWithValidation() {
        return WebContainerDeploymentDesc.parsingWithValidation;
    }

    /**
     * Set the specified ErrorHandler. If the application does not register
     * an ErrorHandler, all error events reported by the reader are ignored.
     * @param errorHandler the Error Handler to set.
     */
    public static void setErrorHandler(ErrorHandler errorHandler) {
        WebContainerDeploymentDesc.errorHandler = errorHandler;
    }

    /**
     * Set the specified EntityResolver.
     * @param entityResolver the Entity Resolver to set.
     */
    public static void setEntityResolver(EntityResolver entityResolver) {
        WebContainerDeploymentDesc.entityResolver = entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @param validation if true, all external entities will be read.
     */
    public static void setParsingWithValidation(boolean validation) {
        WebContainerDeploymentDesc.parsingWithValidation = validation;
    }

}
