/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *
 * Initial developer(s): Miroslav halas
 * Contributor(s):
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.security.jetty.realm;

import org.mortbay.http.*;

import org.objectweb.security.context.SecurityCurrent;
import org.objectweb.security.context.SecurityContext;

/**
 * Jetty realm which propagates the security context to Jonas container. 
 * It also delegates all functionality to any desired Jetty realm to perform 
 * the true authentication. This will ensure that we can integrate any existing
 * Jetty realm with Jonas just through XML configuration.
 *
 * @author Miroslav Halas
 */
public class JonasRealm implements UserRealm,
                                   java.io.Serializable  // must be serializable since it is exposed through JMX
{   
   /**
    * This is the real Jetty realm this realm will delegate to.
    */
   protected UserRealm m_urRealm;
   
   /** 
    * Creates a new instance of JonasRealm 
    *
    * @param urRealm - the real Jetty realm this realm will delegate to
    */
   public JonasRealm(
      UserRealm urRealm
   )
   {
      m_urRealm = urRealm;
   }
   
   /**
    * Authenticate a users credentials. Implementations of this method may 
    * adorn the calling context to assoicate it with the authenticated principal 
    * (eg ThreadLocals). If such context associations are made, they should be 
    * considered valid until a UserRealm.deAuthenticate(UserPrincipal) call is 
    * made for this UserPrincipal.
    *
    * @param strUsername - The username
    * @param objCredentials - The user credentials, normally a String password. 
    * @param strRequest - The request to be authenticated. Additional parameters 
    *                     may be extracted or set on this request as needed for 
    *                     the authentication mechanism (none required for BASIC 
    *                     and FORM authentication). 
    * @return UserPrincipal - The authenticated UserPrincipal.
    */
   public UserPrincipal authenticate(
      String      strUsername, 
      Object      objCredentials,
      HttpRequest httpRequest
   )
   {
      UserPrincipal upPrincipal;
      
      upPrincipal = m_urRealm.authenticate(strUsername, objCredentials, httpRequest);
         
      if ((upPrincipal != null) && (upPrincipal.isAuthenticated()))
      {
         // Now propagate the principal to Jonas
         SecurityContext ctx = new SecurityContext(strUsername);
         SecurityCurrent current = SecurityCurrent.getCurrent();
         current.setSecurityContext(ctx);
         // End of Jonas propagation
      }
      
      return upPrincipal;
   }
   
   /** 
    * Dissassociate the calling context with a Principal. This method is called 
    * when the calling context is not longer associated with the Principal. It 
    * should be used by an implementation to remove context associations such as 
    * ThreadLocals. The UserPrincipal object remains authenticated, as it may be 
    * associated with other contexts.
    * 
    * @param userPrincipal - A UserPrincipal allocated from this realm.
    */
   public void disassociate(
      UserPrincipal userPrincipal
   )
   {
      m_urRealm.disassociate(userPrincipal);

      // Now propagate the empty principal to Jonas
      SecurityContext ctx = new SecurityContext("");
      SecurityCurrent current = SecurityCurrent.getCurrent();
      current.setSecurityContext(ctx);
      // End of Jonas propagation
   }
   
   /**
    * ???
    *
    * @return String
    */
   public String getName(
   )
   {
      return m_urRealm.getName();
   }
   
   /**
    * Pop role from a Principal.
    *
    * @param  userPrincipal - A UserPrincipal previously returned from pushRole 
    * @return UserPrincipal - The principal without the role. Most often this 
    * will be the original UserPrincipal passed.
    */
   public UserPrincipal popRole(
      UserPrincipal userPrincipal
   )
   {
      return m_urRealm.popRole(userPrincipal);
   }
   
   /**
    * Push role onto a Principal. This method is used to add a role to an 
    * existing principal.
    * 
    * @param userPrincipal - An existing UserPrincipal or null for an anonymous user.
    * @param strRole - The role to add. 
    * @return A new UserPrincipal object that wraps the passed user, but with 
    * the added role.
    */
   public UserPrincipal pushRole(
      UserPrincipal userPrincipal, 
      String        strRole
   )
   {
      return m_urRealm.pushRole(userPrincipal, strRole);
   }
}
