/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.jtests.util;

import java.rmi.RemoteException;
import java.util.Properties;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import javax.transaction.UserTransaction;
import junit.framework.TestCase;
import org.objectweb.jonas.adm.AdmInterface;
import org.objectweb.jonas.jtests.tables.DBEnv;
import org.objectweb.jonas.jtests.tables.DBEnvHome;

/**
 * JTestCase extends TestCase to provide a set of standard routines
 * used in jonas tests.
 */
public abstract class JTestCase extends TestCase {
    protected static String jonasName = "EJB"; // change this XXX
    protected static String testtorun = null;
    protected static Context ictx = null;
    public static UserTransaction utx = null;
    private static AdmInterface admI = null;
    private static DBEnv dbEnv = null;
    private static boolean tableSessionLoaded = false;

    public JTestCase(String name) {
	super(name);
    }

    /**
     * random returns an integer between 0 and max - 1
     */
    public int random(int max) throws RemoteException {
	double d = Math.random();
	int ret = (int) (max * d);
	return ret;
    }

    private Context getInitialContext() throws NamingException {

	String registryPort = System.getProperty("jonas.registryport");
	Context ctx = null;
		
	if (registryPort != null) {
	    // it means we are using David
	    Properties prop = new Properties();
	    prop.put("java.naming.factory.initial", "com.sun.jndi.rmi.registry.RegistryContextFactory");
	    prop.put("java.naming.provider.url", "rmi://localhost:" + registryPort);
	    ctx = new InitialContext(prop);
	} else {
	    ctx = new InitialContext(); 
	}
	return ctx;
    }

    /**
     * common setUp routine, used for every test.
     */
    protected void setUp() {
	try {
	    // get InitialContext
	    if (ictx == null) {
		ictx = getInitialContext();
	    }

	    // get UserTransaction
	    if (utx == null) {
		utx = (UserTransaction) ictx.lookup("javax.transaction.UserTransaction");
	    }

	    // Load Session bean tables.jar
	    if (!tableSessionLoaded) {
		if (admI == null) {
		    admI = (AdmInterface) ictx.lookup(jonasName + "_Adm");
		}
		String filename = "tables";
		if (! admI.isLoaded(filename+".jar")) {
		    admI.addBeans(filename+".jar");
		}
		tableSessionLoaded = true;
	    }
	} catch (NamingException e) {
	    System.err.println("Cannot setup test: "+e);
	    e.printStackTrace();
	} catch (RemoteException e) {
	    System.err.println("Cannot load session bean: "+e);
	}
    }

	protected void tearDown() throws Exception {
		if (utx != null) {
			utx.begin();
			utx.commit();
		}
	}

	/**
     * load a bean jar file in the jonas server
     * 
     */
    public void useBeans(String filename, boolean create) {

	try {
	    // Load bean in EJBServer if not already loaded.
	    if (ictx == null) {
		ictx = getInitialContext();
	    }
	    if (admI == null) {
		admI = (AdmInterface) ictx.lookup(jonasName + "_Adm");
	    }
	    if (! admI.isLoaded(filename+".jar")) {
		admI.addBeans(filename+".jar");

		// Create table if the bean was not loaded
		if (create) {
		    getDBEnv().initTable(filename);
		}
	    }
	} catch (Exception e) {
	    System.err.println("Cannot load bean: "+e);
	}
    }

    /**
     * Get the session bean to manage database tables.
     */
    public DBEnv getDBEnv() {
	if (dbEnv == null) {

	    useBeans("tables", false);

	    // Connect to the DBEnvHome object
	    DBEnvHome dbhome = null;
	    try {
		dbhome = (DBEnvHome) PortableRemoteObject.narrow(ictx.lookup("tablesDBEnvHome"), DBEnvHome.class);
	    } catch (NamingException e) {
                System.err.println(">>> "+e);
		fail("Cannot bind to DBEnvHome");
	    }

	    // Create the table using a session bean 
	    try {
		dbEnv = dbhome.create();
	    } catch (Exception e) {
		fail(e.toString());
	    }
	}
	return dbEnv;
    }

    /**
     * sleep n millisec.
     */
    public void sleep(int msec) {
	try {
	    Thread.currentThread().sleep(msec);
	} catch (InterruptedException e) {
	    fail(e.toString());
	}
    }
}
