/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.deployment.api;

import junit.framework.Assert;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas_ejb.deployment.lib.EjbDeploymentDescManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Level;
import org.objectweb.util.monolog.api.Logger;

import java.io.File;
import java.io.PrintStream;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Iterator;

/**
 * Helper used to load deployment descriptors in test environment
 * @author  Christophe Ney - cney@batisseurs.com
 */
public class AssertDeploymentHelper extends Assert {


    static String jarName = "";
    static DeploymentDesc deploymentDesc = null;
    static Logger logger = new TestLogger();


    /**
     * set archive jar name and load deployment descriptor
     */
    public static void setJarName(String jarName) throws Exception {
        if (AssertDeploymentHelper.jarName.equals(jarName)) return;
        AssertDeploymentHelper.jarName = jarName;
        String dir = System.getProperty("ejbjars.dir");
        if (dir == null) {
            throw new Exception("ejbjars.dir system property must be set");
        }
        String file = new File(dir + File.separator + jarName).getAbsolutePath();
        URL url = new URL("file:" + file);
        URLClassLoader ucl = new URLClassLoader(new URL[]{url});
        EjbDeploymentDescManager.setParsingWithValidation(false);
        deploymentDesc = EjbDeploymentDescManager.getDeploymentDesc(file, ucl);
        deploymentDesc.setLogger(logger);
    }

    /**
     * get the current deployment descriptor
     */
    public static DeploymentDesc getDeploymentDesc() {
        return deploymentDesc;
    }

    /**
     * get the method desc corresponding to a method given its name and params
     */
    public static MethodDesc getMethodDesc(BeanDesc bd, String methodName, String[] methodParams)
            throws Exception {
        for (Iterator i = bd.getMethodDescIterator(); i.hasNext();) {
            MethodDesc md = (MethodDesc) i.next();
            Method m = md.getMethod();
            if (m.getName().equals(methodName)) {
                if (methodParams != null) {
                    Class[] paramTypes = m.getParameterTypes();
                    if (paramTypes.length == methodParams.length) {
                        boolean match = true;
                        for (int j = 0; (j < paramTypes.length); j++) {
                            Class type = paramTypes[j];
                            if (!methodParams.equals(paramTypes[j].getName()))
                                match = false;
                        }
                        if (match) return md;
                    }
                }
                return md;
            }
        }
        String msg = methodName;
        if (methodParams != null) {
            msg += "(";
            for (int i = 0; i < methodParams.length; i++) {
                msg += methodParams[i] + ",";
            }
            msg = msg.substring(0, msg.length() - 1) + ")";
        }
        throw new Exception("Method " + msg + " not found");
    }


    /**
     * simple implementation of a logger
     */
    static class TestLogger implements Logger {

        public static boolean debug = false;
        public static PrintStream out = System.out;

        static {
            debug = System.getProperty("jonas_ejb.test.debug", "false").equalsIgnoreCase("true");
        }

        public static void debug(String m) {
            if (debug) System.out.println(m);
        }

        public boolean isOn;

        public TestLogger() {
            isOn = debug;
        }

        public void setIntLevel(int i) {
        }

        public void setLevel(Level level) {
        }

        public int getCurrentIntLevel() {
            return BasicLevel.DEBUG;
        }

        public Level getCurrentLevel() {
            return BasicLevel.LEVEL_DEBUG;
        }

        public boolean isLoggable(int i) {
            return true;
        }

        public boolean isLoggable(Level level) {
            return true;
        }

        public boolean isOn() {
            return true;
        }

        public void log(int i, Object o) {
            if (isOn) out.println(o);
        }

        public void log(Level level, Object o) {
            if (isOn) out.println(o);
        }

        public void log(int i, Object o, Throwable throwable) {
            if (isOn) {
                throwable.printStackTrace(out);
                out.println(o);
            }
        }

        public void log(Level level, Object o, Throwable throwable) {
            if (isOn) {
                throwable.printStackTrace(out);
                out.println(o);
            }
        }

        public void log(int i, Object o, Object o1, Object o2) {
            if (isOn) out.println(o1 + " ." + o2 + ": " + o);
        }

        public void log(Level level, Object o, Object o1, Object o2) {
            if (isOn) out.println(o1 + " ." + o2 + ": " + o);
        }

        public void log(int i, Object o, Throwable throwable, Object o1, Object o2) {
            if (isOn) {
                throwable.printStackTrace(out);
                out.println(o1 + " ." + o2 + ": " + o);
            }
        }

        public void log(Level level, Object o, Throwable throwable, Object o1, Object o2) {
            if (isOn) {
                throwable.printStackTrace(out);
                out.println(o1 + " ." + o2 + ": " + o);
            }
        }

        public void turnOn() {
            isOn = true;
        }

        public void turnOff() {
            isOn = false;
        }

        public Object getAttribute(String s) {
            return null;
        }

        public String[] getAttributeNames() {
            return new String[0];
        }

        public String getName() {
            return null;
        }

        public byte getType() {
            return 0;
        }

        public Object setAttribute(String s, Object o) {
            return null;
        }

        public void setName(String s) {
        }
    }
}
