/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.deployment.ejbql;

import junit.framework.Assert;

import java.io.CharArrayReader;

/**
 * Helper used to check the validity of query parsing
 * @author  Christophe Ney - cney@batisseurs.com
 */
public class AssertQueryHelper extends Assert {

    /**
     * parse the EJBQL query with the lexical parser
     * @param query the string containing the EJBQL query
     * @return the root node of the lexical tree representing the EJBQL query
     */
    private static SimpleNode parse(String query) throws Exception {
        EJBQL parser = new EJBQL(new CharArrayReader(query.toCharArray()));
        return parser.EJBQL();
    }

    /**
     * Check the validity of the parsing for a given query.So far, the validation
     * consists in performing a first parsing, counting the nodes of the obtained
     * lexical tree, creating a new string representation of the query from the
     * lexical tree, parsing this string again and comparing the number of nodes
     * of the last tree with the first one.
     * @param query the string containing the EJBQL query
     * @return the root node of the lexical tree representing the EJBQL query
     * @throw Exception when the number of nodes does not match
     * @throw ParsingException when one of the query (initial,formated) cannot
     * be parsed.
     */
    public static SimpleNode checkParserFor(String query) throws Exception {
        SimpleNode nodes = parse(query);
        // count nodes
        int nbNodes = CountNodeVisitor.count(nodes);
        // create query string for nodes
        String formated = new EJBQLFormater().format(nodes);
        // parse formated
        SimpleNode fnodes = parse(formated);
        // count nodes
        int nbFNodes = CountNodeVisitor.count(fnodes);
        if (nbNodes != nbFNodes) {
            throw new Exception("Query obtained from parsing \"" + formated + "\" is not identical to \"" + query + "\"");
        }
        return nodes;
    }

    /**
     * count the number of nodes
     */
    static class CountNodeVisitor implements EJBQLVisitor {
        int nb = 0;

        static int count(SimpleNode query) {
            CountNodeVisitor v = new CountNodeVisitor();
            v.visit(query, null);
            return v.nb;
        }

        public Object visit(SimpleNode node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTEJBQL node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTFromClause node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTCollectionMemberDeclaration node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTRangeVariableDeclaration node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTSingleValuedPathExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTSingleValuedNavigation node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTCollectionValuedPathExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTSelectClause node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTWhereClause node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTConditionalExpression node, Object data) {
            node.childrenAccept(this, data);
            // skip this if not relevant
            if (node.jjtGetNumChildren() > 1) nb++;
            return null;
        }

        public Object visit(ASTConditionalTerm node, Object data) {
            node.childrenAccept(this, data);
            // skip this if not relevant
            if (node.jjtGetNumChildren() > 1) nb++;
            return null;
        }

        public Object visit(ASTConditionalFactor node, Object data) {
            node.childrenAccept(this, data);
            // skip this if not relevant
            if (node.not) nb++;
            return null;
        }

        public Object visit(ASTBetweenExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTInExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTLikeExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTNullComparisonExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTEmptyCollectionComparisonExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTCollectionMemberExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTComparisonExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTArithmeticExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTArithmeticTerm node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTArithmeticFactor node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTStringExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTDatetimeExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTBooleanExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTEntityBeanExpression node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTFunctionsReturningStrings node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTFunctionsReturningNumerics node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTAbstractSchemaName node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTIdentificationVariable node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTIdentifier node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTPath node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTStringLiteral node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTArithmeticLiteral node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTBooleanLiteral node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

        public Object visit(ASTInputParameter node, Object data) {
            node.childrenAccept(this, data);
            nb++;
            return null;
        }

    }


}
