/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.deployment.ejbql;

import java.io.CharArrayReader;
import java.util.Stack;

/**
 * Formater that creates an EJBQL string from a given lexical tree.
 * @author  Christophe Ney - cney@batisseurs.com
 */
public class EJBQLFormater implements EJBQLVisitor {

    StringBuffer buf = null;

    /**
     * format am EJBQL string equivalent to a given lexical tree.
     */
    public String format(SimpleNode ejbql) throws Exception {
        buf = new StringBuffer();
        visit(ejbql, new Stack());
        return buf.toString();
    }

    /**
     * Generic visit method that traverses all child nodes
     */
    public Object visit(SimpleNode node, Object data) {
        node.childrenAccept(this, data);
        return null;
    }

    //collection_member_declaration ::= IN (collection_valued_path_expression) [ AS ] identifier
    public Object visit(ASTCollectionMemberDeclaration node, Object data) {
        visit((SimpleNode) node, data);
        String id = (String) ((Stack) data).pop();
        String name = (String) ((Stack) data).pop();
        ((Stack) data).push("IN(" + name + ") AS " + id);
        return null;
    }

    public Object visit(ASTEJBQL node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    //from_clause ::= FROM identification_variable_declaration  [ , identification_variable_declaration ] *
    //identification_variable_declaration ::= collection_member_declaration | range_variable_declaration
    public Object visit(ASTFromClause node, Object data) {
        buf.append("FROM ");
        visit((SimpleNode) node, data);
        String s = "";
        for (int i = 0; i < node.jjtGetNumChildren(); i++) {
            String varDec = (String) ((Stack) data).pop();
            s = varDec + ", " + s;
        }
        buf.append(s);
        buf.deleteCharAt(buf.length() - 1);
        buf.deleteCharAt(buf.length() - 1);
        buf.append(' ');
        return null;
    }

    //range_variable_declaration ::= abstract_schema_name  [ AS ] identifier
    public Object visit(ASTRangeVariableDeclaration node, Object data) {
        visit((SimpleNode) node, data);
        String id = (String) ((Stack) data).pop();
        String name = (String) ((Stack) data).pop();
        ((Stack) data).push(name + " AS " + id);
        return null;
    }

    public Object visit(ASTSingleValuedPathExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTSingleValuedNavigation node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTCollectionValuedPathExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    // select_clause ::= SELECT  [ DISTINCT ] {single_valued_path_expression | OBJECT (identification_variable)}
    public Object visit(ASTSelectClause node, Object data) {
        boolean distinct = node.distinct;
        visit((SimpleNode) node, data);
        String val = (String) ((Stack) data).pop();
        buf.append("SELECT " + (distinct?"DISTINCT ":""));
        if (val.indexOf('.') != -1) {
            buf.append(val + " ");
        } else {
            buf.append("OBJECT(" + val + ") ");
        }
        return null;
    }

    public Object visit(ASTWhereClause node, Object data) {
        buf.append("WHERE ");
        visit((SimpleNode) node, data);
        buf.append((String) ((Stack) data).pop());
        return null;
    }

    //where_clause ::= WHERE conditional_expression
    //conditional_expression ::= conditional_term [ OR conditional_term ]*
    public Object visit(ASTConditionalExpression node, Object data) {
        visit((SimpleNode) node, data);
        String s = "";
        if (node.jjtGetNumChildren() > 1) {
            for (int i = 0; i < node.jjtGetNumChildren(); i++) {
                String term = (String) ((Stack) data).pop();
                s = term + " OR " + s;
            }
            s = "(" + s.substring(0, s.length() - 4) + ")";
        } else {
            s = (String) ((Stack) data).pop();
        }
        ((Stack) data).push(s);
        return null;
    }

    //conditional_term ::= conditional_factor [ AND conditional_factor ]*
    public Object visit(ASTConditionalTerm node, Object data) {
        visit((SimpleNode) node, data);
        String s = "";
        if (node.jjtGetNumChildren() > 1) {
            for (int i = 0; i < node.jjtGetNumChildren(); i++) {
                String term = (String) ((Stack) data).pop();
                s = term + " AND " + s;
            }
            s = "(" + s.substring(0, s.length() - 5) + ")";
        } else {
            s = (String) ((Stack) data).pop();
        }
        ((Stack) data).push(s);
        return null;
    }

    //conditional_factor ::= [ NOT ] conditional_test
    public Object visit(ASTConditionalFactor node, Object data) {
        visit((SimpleNode) node, data);
        if (node.not) {
            String s = (String) ((Stack) data).pop();
            ((Stack) data).push("NOT " + s);
        }
        return null;
    }

    //between_expression ::= arithmetic_expression  [ NOT ] BETWEEN arithmetic_expression AND arithmetic_expression
    public Object visit(ASTBetweenExpression node, Object data) {
        visit((SimpleNode) node, data);
        String c = (String) ((Stack) data).pop();
        String b = (String) ((Stack) data).pop();
        String a = (String) ((Stack) data).pop();
        if (node.not) {
            ((Stack) data).push(a + " NOT BETWEEN " + b + " AND " + c);
        } else {
            ((Stack) data).push(a + " BETWEEN " + b + " AND " + c);
        }
        return null;
    }

    // in_expression ::= single_valued_path_expression  [ NOT ] IN (string_literal  [ , string_literal ] * )
    public Object visit(ASTInExpression node, Object data) {
        visit((SimpleNode) node, data);
        String s = "";
        for (int i = 0; i < node.jjtGetNumChildren() - 1; i++) {
            String literal = (String) ((Stack) data).pop();
            s = literal + "," + s;
        }
        s = "IN (" + s.substring(0, s.length() - 1) + ")";
        String svp = (String) ((Stack) data).pop();
        if (node.not) {
            ((Stack) data).push(svp + " NOT " + s);
        } else {
            ((Stack) data).push(svp + " " + s);
        }
        return null;
    }

    //like_expression ::= single_valued_path_expression  [ NOT ] LIKE pattern_value  [ ESCAPE escape_character ]
    public Object visit(ASTLikeExpression node, Object data) {
        visit((SimpleNode) node, data);
        String esc = "";
        if (node.jjtGetNumChildren() == 3) {
            esc = "ESCAPE " + (String) ((Stack) data).pop() + " ";
        }
        String pat = (String) ((Stack) data).pop();
        String svp = (String) ((Stack) data).pop();
        if (node.not) {
            ((Stack) data).push(svp + " NOT LIKE " + pat + " " + esc);
        } else {
            ((Stack) data).push(svp + " LIKE " + pat + " " + esc);
        }
        return null;
    }

    // null_comparison_expression ::= single_valued_path_expression IS  [ NOT ] NULL
    public Object visit(ASTNullComparisonExpression node, Object data) {
        visit((SimpleNode) node, data);
        String svp = (String) ((Stack) data).pop();
        if (node.not) {
            ((Stack) data).push(svp + " IS NOT NULL ");
        } else {
            ((Stack) data).push(svp + " IS NULL ");
        }
        return null;
    }

    // empty_collection_comparison_expression ::= collection_valued_path_expression IS  [ NOT ] EMPTY
    public Object visit(ASTEmptyCollectionComparisonExpression node, Object data) {
        visit((SimpleNode) node, data);
        String cvp = (String) ((Stack) data).pop();
        String not = node.not?"NOT ":"";
        ((Stack) data).push(cvp + " IS " + not + "EMPTY");
        return null;
    }

    //collection_member_expression ::= {single_valued_navigation | identification_variable | input_parameter}
    //   [ NOT ] MEMBER  [ OF ] collection_valued_path_expression
    public Object visit(ASTCollectionMemberExpression node, Object data) {
        visit((SimpleNode) node, data);
        String cvp = (String) ((Stack) data).pop();
        String a = (String) ((Stack) data).pop();
        String not = node.not?"NOT ":"";
        ((Stack) data).push(a + " " + not + "MEMBER OF " + cvp);
        return null;
    }

    //comparison_expression ::= string_value { =|<>} string_expression
    //  | boolean_value { =|<>} boolean_expression
    //  | datetime_value { = | <> | > | < } datetime_expression
    //  | entity_bean_value { = | <> } entity_bean_expression
    //  | arithmetic_value { = | > | >= | < | <= | <> } single_value_designator
    public Object visit(ASTComparisonExpression node, Object data) {
        visit((SimpleNode) node, data);
        String b = (String) ((Stack) data).pop();
        String a = (String) ((Stack) data).pop();
        String op = null;
        switch (((Integer) node.ops.get(0)).intValue()) {
            case EJBQLConstants.EQ:
                op = "=";
                break;
            case EJBQLConstants.NE:
                op = "<>";
                break;
            case EJBQLConstants.GT:
                op = ">";
                break;
            case EJBQLConstants.GE:
                op = ">=";
                break;
            case EJBQLConstants.LT:
                op = "<";
                break;
            case EJBQLConstants.LE:
                op = "<=";
                break;
        }
        ((Stack) data).push(a + op + b);
        return null;
    }

    //arithmetic_expression ::= arithmetic_term [ { + | - } arithmetic_term ] *
    public Object visit(ASTArithmeticExpression node, Object data) {
        visit((SimpleNode) node, data);
        String s = (String) ((Stack) data).pop();
        for (int i = 0; i < node.jjtGetNumChildren() - 1; i++) {
            String literal = (String) ((Stack) data).pop();
            String op = null;
            switch (((Integer) node.ops.get(node.jjtGetNumChildren() - 2 - i)).intValue()) {
                case EJBQLConstants.PLUS:
                    op = "+";
                    break;
                case EJBQLConstants.MINUS:
                    op = "-";
                    break;
            }
            s = literal + op + s;
        }
        ((Stack) data).push(s);
        return null;
    }

    // arithmetic_term ::= arithmetic_factor [ { * | / } arithmetic_factor ]*
    public Object visit(ASTArithmeticTerm node, Object data) {
        visit((SimpleNode) node, data);
        String s = (String) ((Stack) data).pop();
        for (int i = 0; i < node.jjtGetNumChildren() - 1; i++) {
            String literal = (String) ((Stack) data).pop();
            String op = null;
            switch (((Integer) node.ops.get(node.jjtGetNumChildren() - 2 - i)).intValue()) {
                case EJBQLConstants.MULT:
                    op = "*";
                    break;
                case EJBQLConstants.DIV:
                    op = "/";
                    break;
            }
            s = literal + op + s;
        }
        ((Stack) data).push(s);
        return null;
    }

    //arithmetic_factor ::= [ + |- ] arithmetic_primary
    public Object visit(ASTArithmeticFactor node, Object data) {
        visit((SimpleNode) node, data);
        if (node.ops.size() > 0) {
            if (((Integer) node.ops.get(0)).intValue() == EJBQLConstants.MINUS) {
                String primary = (String) ((Stack) data).pop();
                ((Stack) data).push("-" + primary);
            }
        }
        return null;
    }

    public Object visit(ASTStringExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTDatetimeExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTBooleanExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTEntityBeanExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTFunctionsReturningStrings node, Object data) {
        visit((SimpleNode) node, data);
        switch (((Integer) node.ops.get(0)).intValue()) {
            case EJBQLConstants.SUBSTRING:
                {
                    String length = (String) ((Stack) data).pop();
                    String start = (String) ((Stack) data).pop();
                    String literal = (String) ((Stack) data).pop();
                    ((Stack) data).push("SUBSTRING(" + literal + "," + start + "," + length + ")");
                }
                break;
            case EJBQLConstants.CONCAT:
                {
                    String l2 = (String) ((Stack) data).pop();
                    String l1 = (String) ((Stack) data).pop();
                    ((Stack) data).push("CONCAT(" + l1 + "," + l2 + ")");
                }
                break;
        }
        return null;
    }

    public Object visit(ASTFunctionsReturningNumerics node, Object data) {
        visit((SimpleNode) node, data);
        switch (((Integer) node.ops.get(0)).intValue()) {
            case EJBQLConstants.LOCATE:
                {
                    if (node.jjtGetNumChildren() == 3) {
                        String start = (String) ((Stack) data).pop();
                        String l2 = (String) ((Stack) data).pop();
                        String l1 = (String) ((Stack) data).pop();
                        ((Stack) data).push("LOCATE(" + l1 + "," + l2 + "," + start + ")");
                    } else {
                        String l2 = (String) ((Stack) data).pop();
                        String l1 = (String) ((Stack) data).pop();
                        ((Stack) data).push("LOCATE(" + l1 + "," + l2 + ")");
                    }
                }
                break;
            case EJBQLConstants.LENGTH:
                {
                    String l = (String) ((Stack) data).pop();
                    ((Stack) data).push("LENGTH(" + l + ")");
                }
                break;
            case EJBQLConstants.SQRT:
                {
                    String l = (String) ((Stack) data).pop();
                    ((Stack) data).push("SQRT(" + l + ")");
                }
                break;
            case EJBQLConstants.ABS:
                {
                    String l = (String) ((Stack) data).pop();
                    ((Stack) data).push("ABS(" + l + ")");
                }
                break;
        }
        return null;
    }

    public Object visit(ASTAbstractSchemaName node, Object data) {
        String name = (String) node.value;
        ((Stack) data).push(name);
        return null;
    }

    public Object visit(ASTIdentificationVariable node, Object data) {
        visit((SimpleNode) node, data);
        ((Stack) data).push(node.value);
        return null;
    }

    public Object visit(ASTIdentifier node, Object data) {
        visit((SimpleNode) node, data);
        String id = (String) node.value;
        ((Stack) data).push(id);
        return null;
    }

    public Object visit(ASTPath node, Object data) {
        String path = (String) node.value;
        ((Stack) data).push(path);
        return null;
    }

    public Object visit(ASTStringLiteral node, Object data) {
        String val = node.value.toString();
        ((Stack) data).push("'" + val + "'");
        return null;
    }

    public Object visit(ASTArithmeticLiteral node, Object data) {
        String val = node.value.toString();
        ((Stack) data).push(val);
        return null;
    }

    public Object visit(ASTBooleanLiteral node, Object data) {
        String val = node.value.toString();
        ((Stack) data).push(val);
        return null;
    }

    public Object visit(ASTInputParameter node, Object data) {
        String index = node.value.toString();
        ((Stack) data).push("?" + index);
        return null;
    }

    /**
     * display query after it has been formated
     */
    public static void main(String[] args) throws Exception {
        String query = args[0];
        System.out.println(query);
        EJBQL parser = new EJBQL(new CharArrayReader(query.toCharArray()));
        System.out.println(new EJBQLFormater().format(parser.EJBQL()));
    }
}


