/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.deployment.ejbql;

import java.io.CharArrayReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Utility class that can analyse an EJBQL query and print the
 * characteristics of the persistent schema on which the query
 * applies.
 * @author  Christophe Ney - cney@batisseurs.com
 */
public class QueryAnalyser implements EJBQLVisitor {

    HashMap paths = new HashMap();
    HashMap ids = new HashMap();

    /**
     * main method to call the class from the command line
     * @params args the double-quoted EJBQL query to analyze
     */
    public static void main(String args[]) {
        if (args.length != 1) {
            System.out.println("usage: QueryAnalyser [EJB-QL query]");
            System.exit(-1);
        }
        try {
            System.out.println("Analyzing " + args[0]);
            EJBQL parser = new EJBQL(new CharArrayReader(args[0].toCharArray()));
            SimpleNode n = parser.EJBQL();
            n.dump("");
            // visit all nodes
            QueryAnalyser qa = new QueryAnalyser();
            qa.visit(n);
            qa.report();
            System.out.println("Thank you.");
        } catch (Throwable e) {
            System.out.println("Oops.");
            System.out.println(e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * dump a full report
     */
    public void report() {
        System.out.println("Identifiers ");
        for (Iterator i = ids.entrySet().iterator(); i.hasNext();) {
            Map.Entry e = (Map.Entry) i.next();
            System.out.println("  " + e.getKey() + "=" + e.getValue());
        }
        System.out.println("Schema Structure");
        reportSchema();
    }

    /**
     * dump a report on the expected schema struture
     */
    public void reportSchema() {
        for (Iterator i = paths.entrySet().iterator(); i.hasNext();) {
            Map.Entry e = (Map.Entry) i.next();
            String key = (String) e.getKey();
            String path = expendPath(key);
            int dotPos = key.indexOf('.');
            String id = (dotPos != -1)? key.substring(0, dotPos):key;
            String tail = (dotPos != -1)? key.substring(dotPos):"";
            if (!ids.containsKey(id)) {
                System.out.println("Invalid path : " + key);
            } else {
                System.out.println("  " + key + " -> " + path + " : " + e.getValue());
            }
        }
    }

    /**
     * recursively expend the path with declared ids
     */
    String expendPath(String path) {
        int dotPos = path.indexOf('.');
        String id = (dotPos != -1)? path.substring(0, dotPos):path;
        String tail = (dotPos != -1)? path.substring(dotPos):"";
        if (!ids.containsKey(id)) {
            return path; // stop expanding
        } else {
            return expendPath(ids.get(id) + tail);
        }
    }

    /**
     * visitor pattern implementation
     */
    public void visit(SimpleNode node) {
        visit(node, new ArrayList());
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(SimpleNode node, Object data) {
        node.childrenAccept(this, data);
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTEJBQL node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTFromClause node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTCollectionMemberDeclaration node, Object data) {
        ((ArrayList) data).clear();
        visit((SimpleNode) node, data);
        for (Iterator i = ((ArrayList) data).iterator(); i.hasNext();) {
            String id = (String) i.next();
            String key = (String) i.next();
            ids.put(key, id);
        }
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTRangeVariableDeclaration node, Object data) {
        ((ArrayList) data).clear();
        visit((SimpleNode) node, data);
        for (Iterator i = ((ArrayList) data).iterator(); i.hasNext();) {
            String id = (String) i.next();
            String key = (String) i.next();
            ids.put(key, id);
            paths.put(key, "AbstractSchemaName");
        }
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTSingleValuedPathExpression node, Object data) {
        visit((SimpleNode) node, data);
        ArrayList list = ((ArrayList) data);
        String path = (String) list.get(list.size() - 1);
        paths.put(path, "SingleValuedPathExpression");
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTSingleValuedNavigation node, Object data) {
        visit((SimpleNode) node, data);
        ArrayList list = ((ArrayList) data);
        String path = (String) list.get(list.size() - 1);
        paths.put(path, "SingleValuedNavigation");
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTCollectionValuedPathExpression node, Object data) {
        visit((SimpleNode) node, data);
        ArrayList list = ((ArrayList) data);
        String path = (String) list.get(list.size() - 1);
        paths.put(path, "CollectionValuedPathExpression");
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTSelectClause node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTWhereClause node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTConditionalExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTConditionalTerm node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTConditionalFactor node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTBetweenExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTInExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTLikeExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTNullComparisonExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTEmptyCollectionComparisonExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTCollectionMemberExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTComparisonExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTArithmeticExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTArithmeticTerm node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTArithmeticFactor node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTStringExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTDatetimeExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTBooleanExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTEntityBeanExpression node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTFunctionsReturningStrings node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTFunctionsReturningNumerics node, Object data) {
        return visit((SimpleNode) node, data);
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTAbstractSchemaName node, Object data) {
        ((ArrayList) data).add(node.value);
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTIdentificationVariable node, Object data) {
        ((ArrayList) data).add(node.value);
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTIdentifier node, Object data) {
        ((ArrayList) data).add(node.value);
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTStringLiteral node, Object data) {
        ((ArrayList) data).add(node.value);
        return data;
    }

    /**
     * visitor pattern implementation
     */
    public Object visit(ASTArithmeticLiteral node, Object data) {
        try {
            ((ArrayList) data).add(node.value);
        } catch (NumberFormatException e) {
            ((ArrayList) data).add(null);
        }
        return data;
    }

    /**
     * return the literal value as Boolean
     */
    public Object visit(ASTBooleanLiteral node, Object data) {
        ((ArrayList) data).add(node.value);
        return data;
    }

    /**
     * return the parameter number as Integer
     */
    public Object visit(ASTInputParameter node, Object data) {
        try {
            ((ArrayList) data).add(node.value);
        } catch (NumberFormatException e) {
            ((ArrayList) data).add(null);
        }
        return data;
    }

    /**
     * return the String containing the path
     */
    public Object visit(ASTPath node, Object data) {
        ((ArrayList) data).add(node.value);
        return data;
    }

}

