/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.api;

import org.objectweb.jorm.type.api.PExceptionTyping;

import java.io.Serializable;
import java.sql.Date;

/**
 * The <b>PAccessorGen</b> interface defines a generic accessor to be used by
 * persistent bindings to access persistent variable within memory instances.
 * This interface is implemented for each JORM class in order to define
 * specific accessors associated to such a memory instance.
 * @author	R. Basset, P. Dchamboux
 */
public interface PAccessorGen extends PAccessor {
	/**
	 * It retrieves the value of a boolean field.
	 * @param fn	A String containing the name of the field.
	 * @return	The boolean value of the field as stored within the memory
	 * instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	boolean paGetBooleanField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a byte field.
	 * @param fn	A String containing the name of the field.
	 * @return	The byte value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	byte paGetByteField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a char field.
	 * @param fn	A String containing the name of the field.
	 * @return	The char value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	char paGetCharField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a short field.
	 * @param fn	A String containing the name of the field.
	 * @return	The short value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	short paGetShortField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of an int field.
	 * @param fn	A String containing the name of the field.
	 * @return	The int value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	int paGetIntField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of an int field.
	 * @param fn	A String containing the name of the field.
	 * @return	The int value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Integer paGetIntObjField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a long field.
	 * @param fn	A String containing the name of the field.
	 * @return	The long value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	long paGetLongField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a float field.
	 * @param fn	A String containing the name of the field.
	 * @return	The float value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	float paGetFloatField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a double field.
	 * @param fn	A String containing the name of the field.
	 * @return	The double value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	double paGetDoubleField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a string field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	String paGetStringField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a date field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.sql.Date value of the field as stored within the
	 * 			memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Date paGetDateField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a serialized field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.io.Serializable value of the field as stored within
	 * 			the memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Serializable paGetSerializedField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a reference field.
	 * @param fn	A String containing the name of the field.
	 * @param connection The connection to access to support. This connection
	 * can be used to resolve a pname for example.
	 * @return	The reference value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Object paGetRefField(String fn, Object connection) throws PExceptionTyping;

	/**
	 * It assigns the value to a boolean field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The boolean value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetBooleanField(String fn, boolean value) throws PExceptionTyping;

	/**
	 * It assigns the value to a byte field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The byte value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetByteField(String fn, byte value) throws PExceptionTyping;

	/**
	 * It assigns the value to a char field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The char value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetCharField(String fn, char value) throws PExceptionTyping;

	/**
	 * It assigns the value to a short field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The short value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetShortField(String fn, short value) throws PExceptionTyping;

	/**
	 * It assigns the value to an int field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The int value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetIntField(String fn, int value) throws PExceptionTyping;

	/**
	 * It assigns the value to an int field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The int value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetIntObjField(String fn, Integer value) throws PExceptionTyping;

	/**
	 * It assigns the value to a long field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The long value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetLongField(String fn, long value) throws PExceptionTyping;

	/**
	 * It assigns the value to a float field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The float value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetFloatField(String fn, float value) throws PExceptionTyping;

	/**
	 * It assigns the value to a double field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The double value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetDoubleField(String fn, double value) throws PExceptionTyping;

	/**
	 * It assigns the value to a string field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.lang.String value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetStringField(String fn, String value) throws PExceptionTyping;

	/**
	 * It assigns the value to a date field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.sql.Date value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetDateField(String fn, Date value) throws PExceptionTyping;

	/**
	 * It assigns the value to a serialized field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.io.Serializable value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetSerializedField(String fn, Serializable value)
		throws PExceptionTyping;

	/**
	 * It assigns the value to a reference field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The reference value to assign.
	 * @param connection The connection to access to support. This connection
	 * can be used to resolve a pname for example.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetRefField(String fn, Object value, Object connection)
		throws PExceptionTyping;
}
