/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.api;

import java.util.Iterator;

/**
 * The <b>PAccessorGenClass</b> interface defines the Accessor for all
 * generic classes. These methods are the same for all generic classes, they
 * are defined by this unique interface that extends the <code>PAccessor</code>
 * interface. Notice that all accessor methods are also prefixed with "pa".
 * It is also a factory of PIndexedElem that defines the object used by the
 * bindings in order to store the generic class element, and that are
 * implemented by generic class MI.
 * @author	R. Basset, P. Dchamboux
 */
public interface PAccessorGenClass extends PAccessor, PIndexedElemFactory {
	/**
	 * It passes one of the elements extracted from the data store to the
	 * memory instance.
	 * @param	elem	the PIndexElenm to be added to the memory instance.

	 * @param   conn    the connection to the DS
	 */
	void paAdd(PIndexedElem elem, Object conn) throws PExceptionIO;

	/**
	 * It tells if the generic class memory instance supports delta
	 * modifications, which means that paadded, padeleted and pamodified are
	 * supported.
	 * @return	It return true is delta is supported by the memory instance.
	 */
	boolean paDeltaSupported();

	/**
	 * It retrieves the number of elements contained within the generic class
	 * memory instance before storing it. This is the number of element that
	 * should remain into the DSI after writing.
	 * @return	The number of elements within the memory instance of this
	 * 			generic class.
	 */
	int paGetNbElem();

	/**
	 * It returns an Iterator over the elements of a generic class memory
	 * instance. It scans all elements of the MI, their modification flag be
	 * ELEM_CREATED, ELEM_DELETED, ELEM_MODIFIED or ELEM_UNMODIFIED. The
	 * PBinding can then store them into the related DSI.
	 * @return	The Iterator over the memory instance elements.
	 */
	Iterator paIterator();

	/**
	 * It passes the number of elements contained within the generic class DSI
	 * to the memory instance. This is done just before lements are passed to
	 * the memory instance using the paAdd method.
	 * @param nbelem	The number of elements within this generic class DSI.
	 */
	void paSetNbElem(int nbelem);
}
