/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.api;

import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.type.api.PExceptionTyping;

/**
 * The <b>PBinding</b> interface defines the behaviour of binding objects used
 * in order to control the I/Os between a data store instance and its
 * associated memory instance.
 * @author	R. Basset, P. Dchamboux
 */
public interface PBinding {
	static final byte ACTION_SETPA = 0x01;
	static final byte ACTION_UNSETPA = 0x02;
	static final byte ACTION_BIND = 0x03;
	static final byte ACTION_UNBIND = 0x04;
	static final byte ACTION_EXPORT = 0x05;
	static final byte ACTION_UNEXPORT = 0x06;
	static final byte ACTION_EXIST = 0x07;
	static final byte ACTION_READ = 0x08;
	static final byte ACTION_WRITE = 0x09;
	/**
	 * The PBinding can perform read and write to the data store:
	 * It has a PName to reach an existing data store instance and
	 * a PAccessor to acces a memory instance
	 */
	static final byte LIFECYCLE_ACTIVEFORIO = 0x01;
	/**
	 * On the next <code>write</code> call, the PBinding will delete
	 * the data store instance associated to its PName, and release
	 * its PName:
	 * It has a PName to reach an existing data store instance and
	 * a PAccessor to access to a memory instance, even if this last link is
	 * not required for deletion.
	 */
	static final byte LIFECYCLE_DELTOWRITE = 0x02;
	/**
	 * On the next <code>write</code> call the PBinding will delete
	 * the data store instance associated to its PName, and release
	 * its PName:
	 * It has a PName to reach an existing data store instance,
	 * but no PAccessor to acces to a memory instance that is not necessary.
	 */
	static final byte LIFECYCLE_DELTOWRITENOACC = 0x03;
	/**
	 * The PBinding has a PName to reach the data store, but the
	 * corresponding data store instance does not exist yet. It has a PAccessor
	 * to access to a memory instance. It is thus ready to create the new data
	 * store instance.
	 *
	 */
	static final byte LIFECYCLE_NEWTOWRITE = 0x04;
	/**
	 * The PBinding has a PName to reach an existing data store
	 * instance. It does not have any PAccessor to access to a memory instance.
	 */
	static final byte LIFECYCLE_NOACC = 0x05;
	/**
	 * The PBinding has a PAccessor to access to a memory instance. It does not
	 * have any PName to reach a data store instance.
	 */
	static final byte LIFECYCLE_NOTBOUND = 0x06;
	/**
	 * The PBinding has neither a PName to designate a data store
	 * instance, nor a PAccessor to acces a memeory instance.
	 */
	static final byte LIFECYCLE_NOTBOUNDNOACC = 0x07;
	/**
	 * This is not a real state for a PBinding. It is used when a transition
	 * leads to a protocol error.
	 */
	static final byte LIFECYCLE_ERROR = 0x08;

	/**
	 * It assigns a DSI represented by the PName passed as a parameter to this
	 * PBinding. This thus changes the link to the data store instance if one
	 * has been already assigned. The state either changed to
	 * LIFECYCLE_NOACCESSOR and LIFECYCLE_ACTIVEFORIO.
	 * @param pn    The new PName to associate to this binding. This name
	 *              must belong to the binder associated to this binding.
	 * @exception PExceptionNaming	It is raised when there is a problem for the
	 * 								associated binder to manage this PName.
	 */
	void bind(PName pn) throws PExceptionNaming;

	/**
	 * It verifies if the DSI attached to this binding exists within the DS.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @return	It returns true if the DSI exists.
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionNaming	It is raised when there is a problem for the
	 * 								associated binder to manage this PName.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_NOACC
	 */
	boolean exist(Object conn)
		throws PExceptionIO, PExceptionNaming, PExceptionProtocol;

	/**
	 * It creates a new PName for this binding, the binding state
	 * is normally changed to LIFECYCLE_NEWTOWRITE, and the link to a data
	 * store instance if it existed is removed.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @return	The new PName assigned to this binding.
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionNaming	It is raised when the associated binder
	 * 								does not manage to create a new name.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_DELTOWRITE
	 * 									LIFECYCLE_NEWTOWRITE
	 * 									LIFECYCLE_NOACC
	 */
	PName export(Object conn)
		throws PExceptionIO, PExceptionNaming, PExceptionProtocol;

	/**
	 * It gives access to the PAccessor associated to this binding. It is a
	 * static one with a couple of accessor methods for each field.
	 * @return	The PAccessor the binding uses to access the memory instance
	 * 			when it read/write the data store instance. It may be null.
	 */
	PAccessor getPAccessor();

	/**
	 * It gives access to the PAccessorGen associated to this binding. It is a
	 * generic one with accessor methods parameterized with the field name.
	 * @return	The PAccessorGen the binding uses to access the memory instance
	 * 			when it read/write the data store instance. It may be null.
	 */
	PAccessorGen getPAccessorGen();

	/**
	 * It gives access to the PClassMapping which manages this binding.
	 * @return	The PClassMapping which creates this binding.
	 */
	PClassMapping getPClassMapping();

	/**
	 * It gives access to the PName, which designate a particular DSI,
	 * associated to this binding. It returns <code>null</code> if the binding
	 * has not been given a PName using a bind or export method.
	 * @return	The PName associated to this binding.
	 */
	PName getPName();

	/**
	 * It yields the current status associated with this PBinding.
	 * @return	The corresponding status.
	 */
	byte getStatus();

	/**
	 * It initialises a PBinding.
	 * @param pcm	The PBinding to initialise.
	 * @return	The new PBinding with lifecycle state set to
	 * 			LIFECYCLE_NOTBOUNDNOACC or LIFECYCLE_NOTBOUND.
	 * @exception PExceptionProtocol	This PBinding cannot be managed by this
	 * 									PClassMapping.
	 */
	void init(PClassMapping pcm) throws PExceptionProtocol;

	/**
	 * It reads the data store instance designated by the PName of this
	 * binding and transfers each attribute value to the accessor object.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionNaming	It is raised when there is a problem for the
	 * 								associated binder to manage this PName.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 */
	void read(Object conn)
		throws PExceptionIO, PExceptionNaming, PExceptionProtocol;

	/**
	 * It associates a PAccessor to this binding in order to access to the
	 * memory instance.
	 * @param pa   The accessor object to be used by the binding for
	 *             accessing the variables of the memory instance when
	 *             loading and storing persistent data.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									null accessor and with a state
	 * 									different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_DELTOWRITE
	 * 									LIFECYCLE_DELTOWRITENOACC
	 * 									LIFECYCLE_NOACC
	 * 									LIFECYCLE_NOTBOUND
	 * 									LIFECYCLE_NOTBOUNDNOACC
	 */
	void setPAccessor(PAccessor pa) throws PExceptionProtocol;

	/**
	 * It associates a PAccessorGen to this binding in order to access to the
	 * memory instance.
	 * @param pa   The generic accessor object to be used by the binding for
	 *             accessing the variables of the memory instance when
	 *             loading and storing persistent data.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									null accessor and with a state
	 * 									different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_DELTOWRITE
	 * 									LIFECYCLE_DELTOWRITENOACC
	 * 									LIFECYCLE_NOACC
	 * 									LIFECYCLE_NOTBOUND
	 * 									LIFECYCLE_NOTBOUNDNOACC
	 */
	void setPAccessorGen(PAccessorGen pa) throws PExceptionProtocol;

	/**
	 * It remove the association between a PName and a PBinding. No more PName
	 * is associated with this binding (it has no more link with a DSI).
	 * @exception PExceptionNaming	It is raised when there is a problem for the
	 * 								associated binder to manage this PName.
	 */
	void unbind() throws PExceptionNaming;

	/**
	 * It prepares this PBinding to remove the data store instance from the
	 * data store. In order to actually remove this DSI, a call to write must
	 * follow.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @exception PExceptionNaming	It is raised when there is a problem for the
	 * 								associated binder to manage this PName.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_DELTOWRITE
	 * 									LIFECYCLE_DELTOWRITENOACC
	 * 									LIFECYCLE_NEWTOWRITE
	 * 									LIFECYCLE_NOACC
	 */
	void unexport(Object conn)
		throws PExceptionNaming, PExceptionProtocol, PExceptionIO;

	/**
	 * It transfers each attribute value taken from the PAccessor
	 * object to the data store instance designated by the PName
	 * associated with this binding.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @exception   e1 if the state of the binding does not allow writing
	 *                 to be completed (no valid name for example)
	 *              e2 if the pc parameter does not allow acces to the
	 *                 data store
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionNaming	It is raised when there is a problem for the
	 * 								associated binder to manage this PName.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_NEWTOWRITE
	 * 									LIFECYCLE_DELTOWRITE
	 * 									LIFECYCLE_DELTOWRITENOACC
	 * @exception PExceptionTyping	It is raised when a reference to be stored
	 * 								has a type incompatible with the respective
	 * 								field definition.
	 */
	void write(Object conn)
		throws PExceptionIO, PExceptionNaming, PExceptionProtocol,
		PExceptionTyping;
}
