/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.api;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.GenClass;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PExceptionTyping;

/**
 * The <b>PClassMapping</b> interface defines the behaviour of objects that
 * contain all information required to map a JORM class. It is a PBinder but
 * mainly delegates this behaviour to a nested PBinder that it must reference.
 * It allows the creation or deletion of the persistent structures required to
 * map this class within a particular hosting data store.
 * It also holds references to naming contexts associated to reference fields,
 * as well as to PClassMapping associated to generic class instances of
 * relevant fields.
 * @author	P. Dchamboux
 */
public interface PClassMapping extends PBinder {
	/**
	 * It tells that a cleanup is required for a JORM class with
	 * the removal of existing data as well as hosting structures.
	 */
	final static byte CLEANUP_REMOVEALL = 1;
	/**
	 * It tells that a cleanup is required for a JORM class with
	 * the removal of only existing data.
	 */
	final static byte CLEANUP_REMOVEDATA = 2;
	/**
	 * It tells that a cleanup is required for a JORM class with
	 * nothing to do on existing data or hosting structures.
	 */
	final static byte CLEANUP_DONOTHING = 3;

	/**
	 * It creates a new PBinding.
	 * @return	The new PBinding with lifecycle state set to
	 * 			LIFECYCLE_NOTBOUNDNOACC.
	 */
	PBinding createPBinding();

	/**
	 * It intialises the class mapping. It especially assigns a mapper as well
	 * as a meta-object describing either a class or a generic class. It also
	 * defines all type information related to this class mapping.
	 * @param mapper	The mapper to be assigned to this class mapping.
	 * @param metaclass	The meta-object describing the associated class. It may
	 * 					be null.
	 */
	void init(PMapper mapper, MetaObject metaclass) throws PExceptionTyping;

	/**
	 * It initializes the persistent structures required to map a JORM class
	 * within a particular Data Store.
	 * This method has no effect when the structures already exist, or when the
	 * structure must not be manipulated via JORM (e.g., case of a legacy DS).
	 * The deletion of these structures may be requested before initializing
	 * them.
	 * @param conn		The connection that can be used to access the DS.
	 * @param cleanup	Tells if structures and/or data should be removed at
	 * 					initialization time.
	 * @exception PExceptionProtocol	A problem occured during initialisation
	 * 									that is not related to I/O.
	 * @exception PExceptionIO	A problem occured during the initialization of
	 * 							the DS.
	 */
	void initMappingStructures(Object conn, byte cleanup)
		throws PExceptionProtocol, PExceptionIO;

	/**
	 * It removes the persistent structures required to map a JORM class
	 * within a particular Data Store.
	 * This method has no effect when the structures do not exist, or when the
	 * structure must not be manipulated via JORM (e.g., case of a legacy DS).
	 * @param conn	The connection that can be used to access the DS.
	 * @param contentonly	It specifies if only the content of the persistent
	 * 						structures should be removed (it only removes data).
	 * @exception PExceptionProtocol	A problem occured during the deletion
	 * 									that is not related to I/O.
	 * @exception PExceptionIO	A problem occured during the deletion of
	 * 							the DS.
	 */
	void removeMappingStructures(Object conn, boolean contentonly)
		throws PExceptionProtocol, PExceptionIO;

	/**
	 * It retrieves the name of the class mapped by this PClassMapping. It
	 * should be the same as the one get when calling
	 * "getPType().getJormName()" if the PType exists, that is a fully qualified
	 * name.
	 * @return	The name of this class.
	 */
	String getClassName();

	/**
	 * It gives access to the PClassMapping associated to the element of a
	 * generic class whose type is a generic class.
	 * @return	The PClassMapping of the generic class instance.
	 * @exception UnsupportedOperationException	Raised when the method is not
	 * 											supported by the PClassMapping.
	 */
	PClassMapping getGenClassMapping() throws UnsupportedOperationException;

	/**
	 * It gives access to the PClassMapping associated to a field whose type is
	 * a generic class.
	 * @param fn	The name of the generic class field.
	 * @return	The PClassMapping of the generic class instance.
	 * @exception UnsupportedOperationException	Raised when the method is not
	 * 											supported by the PClassMapping.
	 */
	PClassMapping getGenClassMapping(String fn)
		throws UnsupportedOperationException;

	/**
	 * It gives access to the mapper that has mapped the associated JORM class
	 * as defined by this PClassMapping.
	 * @return	The associated PMapper.
	 */
	PMapper getPMapper();

	/**
	 * It returns the memory instance manager. This manager could manage the
	 * life cycle of memorty instances. A manager is not always required,
	 * therefore this method can return null.
	 */
	Object getMIManager();

	/**
	 * It yields the PBinder associated to this PClassMapping. As a PBinder
	 * itself, this PClassMapping delegates this behaviour to this PBinder.
	 * @return	The PBinder associated to this PClassMapping.
	 */
	PBinder getPBinder();

	/**
	 * It returns the Class/GenClass object representing the meta-information
	 * associated to the JORM class/generic class whose mapping is defined by
	 * this PClassMapping. It may be null if not requested at mapping time.
	 * @return	The Class/GenClass meta-object or null.
	 * @see	org.objectweb.jorm.api.PMapper#map
	 */
	MetaObject getMetaInfo();

	/**
	 * It returns the naming context that manages the references contained by
	 * the generic class instance whose mapping is defined by this
	 * PClassMapping.
	 * @return	The naming context that manages references contained by a
	 * 			generic class instance.
	 * @exception	UnsupportedOperationException	It is raised when this
	 * 												PClassMapping does not map
	 * 												a generic class instance, or
	 * 												when the elements of this
	 * 												generic class instance are
	 * 												not references.
	 */
	PNamingContext getPNamingContext()
		throws UnsupportedOperationException;

	/**
	 * It returns the naming context that manages the reference associated to
	 * a field whose type is a JORM class.
	 * @return	The naming context that manages the reference defined by a
	 * 			field.
	 * @exception	UnsupportedOperationException	It is raised when no naming
	 * 												context is associated to
	 * 												the corresponding field.
	 */
	PNamingContext getPNamingContext(String fn)
		throws UnsupportedOperationException;

	/**
	 * It specifies if the given kind of PMapper is supported by this
	 * PClassMapping.
	 * @param mappername	The name defining the PMapper kind.
	 * @return	true if this kind of mapper is supported.
	 */
	boolean isConform(String mappername);

	/**
	 * It assigns the PBinder associated to this PClassMapping. As a PBinder
	 * itself, this PClassMapping delegates this behaviour to this PBinder.
	 * @param pb	The PBinder associated to this PClassMapping.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the name that identifies the DSI for
	 * 									the PClassMapping is not supported by
	 * 									this PBinder that should be connected
	 * 									to this PClassMapping.
	 */
	void setPBinder(PBinder pb) throws PExceptionTyping;

	/**
	 * It assigns the GenClassMapping used to manage the reference hold by a
	 * field whose type is a JORM class.
	 */
	void setGenClassMapping(PClassMapping pcm)
		throws UnsupportedOperationException;

	/**
	 * It assigns the GenClassMapping used to manage the reference hold by a
	 * field whose type is a JORM class.
	 */
	void setGenClassMapping(String fn, PClassMapping pcm)
		throws UnsupportedOperationException;

	/**
	 * It assigns the memory instance manager. This manager could manage the
	 * life cycle of memorty instances. A manager is not always required.
	 */
	void setMIManager(Object manager);

	/**
	 * It assigns the PNamingContext used to manage the references hold by a
	 * generic class instance whose mapping is described by this PClassMapping.
	 * It should verify that the name structure for this reference is supported
	 * by the corresponding naming context.
	 * @param pnc	The PNamingContext to be associated to this generic class
	 * 				PClassMapping for managing its reference elements.
	 * @param codingonly	It specifies that when using this naming context for
	 * 						managing a reference, it only executes coding or
	 * 						both coding and export/resolve phase.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the names that reference the DSI stored
	 * 									within the generic class instance is not
	 * 									supported by this naming context.
	 * @exception	UnsupportedOperationException	It is raised when this
	 * 												PClassMapping does not map
	 * 												a generic class instance, or
	 * 												when the elements of this
	 * 												generic class instance are
	 * 												not references.
	 */
	void setPNamingContext(PNamingContext pnc, boolean codingonly)
		throws PExceptionTyping, UnsupportedOperationException;

	/**
	 * It assigns the PNamingContext used to manage the reference hold by a
	 * field whose type is a JORM class.
	 * It should verify that the name structure for this reference is supported
	 * by the corresponding naming context.
	 * @param fn	The name of the involved field.
	 * @param pnc	The PNamingContext to be associated to this field for
	 * 				managing the reference it may hold.
	 * @param codingonly	It specifies that when using this naming context for
	 * 						managing a reference, it only executes coding or
	 * 						both coding and export/resolve phase.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the names that reference the DSI stored
	 * 									within the generic class instance is not
	 * 									supported by this naming context.
	 * @exception	UnsupportedOperationException	It is raised when no naming
	 * 												context is associated to
	 * 												the corresponding field.
	 */
	void setPNamingContext(String fn, PNamingContext pnc, boolean codingonly)
		throws PExceptionTyping, UnsupportedOperationException;
}
