/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.api;

import org.objectweb.jorm.type.api.PExceptionTyping;

import java.io.Serializable;
import java.util.Date;

/**
 * The <b>PIndexedElem</b> interface defines the representation of an
 * element of a generic class. It provides a way to access the element value,
 * as well as its index. It also allows the association of a status for such
 * elements when the generic class accessor implemented by the MI supports
 * delta modifications (@see PAccessorGenClass#deltaSupported).
 * @author	R. Basset, P. Dchamboux
 */
public interface PIndexedElem {
	/**
	 * It specifies that this element is a new one within this generic class
	 * object. This means that is new index has been created.
	 */
	final static byte ELEM_CREATED = 1;

	/**
	 * It specifies that this element has been deleted from this generic class
	 * object. This means that no more element with this index exists anymore
	 * wiothin this generic class object.
	 */
	final static byte ELEM_DELETED = 2;

	/**
	 * It specifies that this element as been modified within this generic class
	 * object. This means that only the value of the indexed element has been
	 * modified, and especially not the index.
	 */
	final static byte ELEM_MODIFIED = 3;

	/**
	 * It specifies that this element existed at load time and that it has not
	 * been modified until then. This means that neither the value of the
	 * indexed element nor its index has been modified.
	 */
	final static byte ELEM_UNMODIFIED = 4;

	/**
	 * It yields the status of this indexed element. The status value belongs to
	 * the following ones:
	 * <ul>
	 * <li>ELEM_CREATED</li>
	 * <li>ELEM_DELETED</li>
	 * <li>ELEM_MODIFIED</li>
	 * <li>ELEM_UNMODIFIED</li>
	 * </ul>
	 */
	byte getElemStatus();

	/**
	 * It retrieves the value of a boolean element.
	 * @return	The boolean value of the field as stored within the memory
	 * instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	boolean pieGetBooleanElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a byte element.
	 * @return	The byte value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	byte pieGetByteElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a byte index field.
	 * @param fn	A String containing the name of the index field.
	 * @return	The byte value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	byte pieGetByteIndexField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a char element.
	 * @return	The char value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	char pieGetCharElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a char index field.
	 * @param fn	A String containing the name of the index field.
	 * @return	The char value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	char pieGetCharIndexField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a short element.
	 * @return	The short value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	short pieGetShortElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a short index field.
	 * @param fn	A String containing the name of the index field.
	 * @return	The short value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	short pieGetShortIndexField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of an int element.
	 * @return	The int value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	int pieGetIntElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of an int index field.
	 * @param fn	A String containing the name of the index field.
	 * @return	The int value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	int pieGetIntIndexField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a long element.
	 * @return	The long value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	long pieGetLongElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a long index field.
	 * @param fn	A String containing the name of the index field.
	 * @return	The long value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	long pieGetLongIndexField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a float element.
	 * @return	The float value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	float pieGetFloatElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a double element.
	 * @return	The double value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	double pieGetDoubleElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a string element.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			memory instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	String pieGetStringElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a string index field.
	 * @param fn	A String containing the name of the index field.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	String pieGetStringIndexField(String fn) throws PExceptionTyping;

	/**
	 * It retrieves the value of a date element.
	 * @return	The java.sql.Date value of the field as stored within the
	 * 			memory instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	Date pieGetDateElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a serialized element.
	 * @return	The java.io.Serializable value of the field as stored within
	 * 			the memory instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	Serializable pieGetSerializedElem() throws PExceptionTyping;

	/**
	 * It retrieves the value of a reference element.
	 * @return	The reference value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	Object pieGetRefElem() throws PExceptionTyping;

	/**
	 * It assigns the value to a boolean element.
	 * @param value	The boolean value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetBooleanElem(boolean value) throws PExceptionTyping;

	/**
	 * It assigns the value to a byte element.
	 * @param value	The byte value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetByteElem(byte value) throws PExceptionTyping;

	/**
	 * It assigns the value to a byte index field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The byte value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	void pieSetByteIndexField(String fn, byte value) throws PExceptionTyping;

	/**
	 * It assigns the value to a char element.
	 * @param value	The char value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetCharElem(char value) throws PExceptionTyping;

	/**
	 * It assigns the value to a char index field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The char value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	void pieSetCharIndexField(String fn, char value) throws PExceptionTyping;

	/**
	 * It assigns the value to a short element.
	 * @param value	The short value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetShortElem(short value) throws PExceptionTyping;

	/**
	 * It assigns the value to a short index field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The short value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	void pieSetShortIndexField(String fn, short value) throws PExceptionTyping;

	/**
	 * It assigns the value to an int element.
	 * @param value	The int value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetIntElem(int value) throws PExceptionTyping;

	/**
	 * It assigns the value to an int index field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The int value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	void pieSetIntIndexField(String fn, int value) throws PExceptionTyping;

	/**
	 * It assigns the value to a long element.
	 * @param value	The long value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetLongElem(long value) throws PExceptionTyping;

	/**
	 * It assigns the value to a long index field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The long value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	void pieSetLongIndexField(String fn, long value) throws PExceptionTyping;

	/**
	 * It assigns the value to a float element.
	 * @param value	The float value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetFloatElem(float value) throws PExceptionTyping;

	/**
	 * It assigns the value to a double element.
	 * @param value	The double value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetDoubleElem(double value) throws PExceptionTyping;

	/**
	 * It assigns the value to a string element.
	 * @param value	The java.lang.String value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetStringElem(String value) throws PExceptionTyping;

	/**
	 * It assigns the value to a string index field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.lang.String value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this index.
	 */
	void pieSetStringIndexField(String fn, String value) throws PExceptionTyping;

	/**
	 * It assigns the value to a date element.
	 * @param value	The java.sql.Date value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetDateElem(Date value) throws PExceptionTyping;

	/**
	 * It assigns the value to a serialized element.
	 * @param value	The java.io.Serializable value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetSerializedElem(Serializable value) throws PExceptionTyping;

	/**
	 * It assigns the value to a reference element.
	 * @param value	The reference value to assign.
	 * @exception	PExceptionTyping	Elements are not of this type.
	 */
	void pieSetRefElem(Object value) throws PExceptionTyping;
}
