/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.api;

import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.parser.api.Parser;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PTypeSpace;

/**
 * The <b>PMapper</b> interface defines a JORM mapper as an object that mainly
 * maintains the set of JORM classes that it maps on a particular DS. It relies
 * on JCA as a means to access a DS.
 * It enables to access to the meta-information associated to JORM classes at
 * runtime.
 * Before being able to request the load of meta-information at mapping time,
 * the mapper should have been assigned a mapping name, a meta-information
 * manager, as well as a parser.
 * @author	P. Dchamboux
 */
public interface PMapper {
	/**
	 * The name yielded by this method defines what kind of PMapper is used
	 * on a particular DS.
	 * @return	The name defining the PMapper kind.
	 */
	String getMapperName();

	/**
	 * It retrieves the name of the mapping that is instanciated through this
	 * mapper.
	 * @return	The name of the mapping.
	 */
	String getMappingName();

	/**
	 * It looks for the PClassMapping object associated to a JORM class within
	 * this PMapper.
	 * @return	The PClassMapping that maps the corresponding class, or null if
	 * 			none exists
	 */
	PClassMapping lookup(String classname);

	/**
	 * It registers a PClassMapping associated to a particular JORM class within
	 * this PMapper. It initializes the storage structures if necessary. It
	 * previously removes them if the "dsrem" flag is set. If the mapper is
	 * ready to manage the meta-information, it is loaded for this JORM class.
	 * @param conn	The connection to access to the underlying DS.
	 * @param pcm	The class mapping to add to this mapper.
	 * @param cleanup	Tells if structures and/or data should be removed at
	 * 					initialization time.
	 * @exception PExceptionIO	It is raised when an I/O problem occurs within
	 * 							the underlying DS.
	 * @exception PExceptionProtocol	It is raised when a problem occurs
	 * 									during initialisation of the class
	 * 									mapping.
	 * @exception PExceptionTyping	It is raised when a typing problem occurs
	 * 								during initialisation of the class mapping.
	 */
	void map(Object conn, PClassMapping pcm, byte cleanup)
		throws PExceptionIO, PExceptionProtocol, PExceptionTyping;

	/**
	 * It registers a PClassMapping associated to a particular JORM class within
	 * this PMapper. It initializes the storage structures if necessary. It
	 * previously removes them if the "dsrem" flag is set. It also loads the
	 * meta-information related to this class if the "loadmeta" flag is set.
	 * @param conn	The connection to access to the underlying DS.
	 * @param pcm		The class mapping to add to this mapper.
	 * @param cleanup	Tells if structures and/or data should be removed at
	 * 					initialization time.
	 * @param loadmeta	It specifies that meta-information should be loaded for
	 * 					this class to map.
	 * @exception PExceptionIO	It is raised when an I/O problem occurs within
	 * 							the underlying DS.
	 * @exception PExceptionProtocol	It is raised when a problem occurs
	 * 									during initialisation of the class
	 * 									mapping.
	 * @exception PExceptionTyping	It is raised when a typing problem occurs
	 * 								during initialisation of the class mapping.
	 */
	void map(Object conn, PClassMapping pcm, byte cleanup, boolean loadmeta)
		throws PExceptionIO, PExceptionProtocol, PExceptionTyping;

	/**
	 * It assigns the name of the mapper that is instanciated. It is mandatory
	 * to assign it at mapper initialization as the load of meta-information
	 * depends on it.
	 * @param mappingname	The name of the maping as defined within the XML
	 * 						JORM definition files.
	 */
	void setMapperName(String mappername);

	/**
	 * It assigns the name of the mapping that is instanciated. It is mandatory
	 * to assign it at mapper initialization as the load of meta-information
	 * depends on it.
	 * @param mappingname	The name of the maping as defined within the XML
	 * 						JORM definition files.
	 */
	void setMappingName(String mappingname);

	/**
	 * It assigns a meta-information Manager that may used by a parser in order
	 * to load the meta-information associated to JORM classes mapped within
	 * this PMapper.
	 * @param m	The meta-information Manager to be assigned.
	 */
	void setMetaInfoManager(Manager m);

	/**
	 * It assigns a Parser that may be used to create the meta-information
	 * associated to JORM classes mapped within this PMapper.
	 * @param p	The Parser to be assigned.
	 */
	void setParser(Parser p);

	/**
	 * It assigns a PTypeSpace that may be used to create PType
	 * associated to JORM classes mapped within this PMapper.
	 * @param pts	The PTypeSpace to be assigned.
	 */
	void setPTypeSpace(PTypeSpace pts);

	/**
	 * It returns a PTypeSpace that may be used to create PType
	 * associated to JORM classes mapped within this PMapper.
	 */
	PTypeSpace getPTypeSpace();

	/**
	 * It unmaps a particular JORM class within this PMapper.
	 * @param classname	The name of the JORM class to be unmapped.
	 */
	void unmap(String classname) throws PException;
}
