/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.api;

/**
 * This class, which is entirely static, defines the state graphs for various
 * JORM objects.
 * @author	P. Dchamboux
 */
public class PStateGraph {
	/**
	 * It defines the state graph that represents the lifecycle of PBinding
	 * objects. This means that, knowing a particular state and the action to
	 * be performed, it computes the transition to the next state that should be
	 * taken by this PBinding after the execution of this action.
	 * @param currentstate	The state before the execution of the action.
	 * @param action		The action to be performed.
	 * @return	The state after the execution of the action.
	 */
	public static byte nextStatePBinding(byte currentstate, byte action) {
		switch (currentstate) {
		case PBinding.LIFECYCLE_ACTIVEFORIO:
			switch (action) {
			case PBinding.ACTION_WRITE:
			case PBinding.ACTION_READ:
			case PBinding.ACTION_EXIST:
			case PBinding.ACTION_BIND:
			case PBinding.ACTION_SETPA:
				return currentstate;
			case PBinding.ACTION_EXPORT:
				return PBinding.LIFECYCLE_NEWTOWRITE;
			case PBinding.ACTION_UNSETPA:
				return PBinding.LIFECYCLE_NOACC;
			case PBinding.ACTION_UNBIND:
				return PBinding.LIFECYCLE_NOTBOUND;
			case PBinding.ACTION_UNEXPORT:
				return PBinding.LIFECYCLE_DELTOWRITE;
			default :
				return PBinding.LIFECYCLE_ERROR;
			}
		case PBinding.LIFECYCLE_NEWTOWRITE:
			switch (action) {
			case PBinding.ACTION_WRITE:
			case PBinding.ACTION_BIND:
				return PBinding.LIFECYCLE_ACTIVEFORIO;
			case PBinding.ACTION_EXPORT:
			case PBinding.ACTION_SETPA:
				return currentstate;
			case PBinding.ACTION_UNBIND:
			case PBinding.ACTION_UNEXPORT:
				return PBinding.LIFECYCLE_NOTBOUND;
			case PBinding.ACTION_READ:
			case PBinding.ACTION_EXIST:
			case PBinding.ACTION_UNSETPA:
			default :
				return PBinding.LIFECYCLE_ERROR;
			}
		case PBinding.LIFECYCLE_DELTOWRITE:
			switch (action) {
			case PBinding.ACTION_WRITE:
			case PBinding.ACTION_UNBIND:
				return PBinding.LIFECYCLE_NOTBOUND;
			case PBinding.ACTION_BIND:
				return PBinding.LIFECYCLE_ACTIVEFORIO;
			case PBinding.ACTION_EXPORT:
				return PBinding.LIFECYCLE_NEWTOWRITE;
			case PBinding.ACTION_SETPA:
			case PBinding.ACTION_UNEXPORT:
				return currentstate;
			case PBinding.ACTION_UNSETPA:
				return PBinding.LIFECYCLE_DELTOWRITENOACC;
			case PBinding.ACTION_READ:
			case PBinding.ACTION_EXIST:
			default :
				return PBinding.LIFECYCLE_ERROR;
			}
		case PBinding.LIFECYCLE_DELTOWRITENOACC:
			switch (action) {
			case PBinding.ACTION_WRITE:
			case PBinding.ACTION_UNBIND:
				return PBinding.LIFECYCLE_NOTBOUNDNOACC;
			case PBinding.ACTION_BIND:
				return PBinding.LIFECYCLE_NOACC;
			case PBinding.ACTION_SETPA:
				return PBinding.LIFECYCLE_DELTOWRITE;
			case PBinding.ACTION_UNSETPA:
			case PBinding.ACTION_UNEXPORT:
				return currentstate;
			case PBinding.ACTION_READ:
			case PBinding.ACTION_EXIST:
			case PBinding.ACTION_EXPORT:
			default :
				return PBinding.LIFECYCLE_ERROR;
			}
		case PBinding.LIFECYCLE_NOACC:
			switch (action) {
			case PBinding.ACTION_EXIST:
			case PBinding.ACTION_BIND:
			case PBinding.ACTION_UNSETPA:
				return currentstate;
			case PBinding.ACTION_SETPA:
				return PBinding.LIFECYCLE_ACTIVEFORIO;
			case PBinding.ACTION_UNBIND:
				return PBinding.LIFECYCLE_NOTBOUNDNOACC;
			case PBinding.ACTION_UNEXPORT:
				return PBinding.LIFECYCLE_DELTOWRITENOACC;
			case PBinding.ACTION_WRITE:
			case PBinding.ACTION_READ:
			case PBinding.ACTION_EXPORT:
			default :
				return PBinding.LIFECYCLE_ERROR;
			}
		case PBinding.LIFECYCLE_NOTBOUND:
			switch (action) {
			case PBinding.ACTION_BIND:
				return PBinding.LIFECYCLE_ACTIVEFORIO;
			case PBinding.ACTION_EXPORT:
				return PBinding.LIFECYCLE_NEWTOWRITE;
			case PBinding.ACTION_SETPA:
			case PBinding.ACTION_UNBIND:
				return currentstate;
			case PBinding.ACTION_UNSETPA:
				return PBinding.LIFECYCLE_NOTBOUNDNOACC;
			case PBinding.ACTION_WRITE:
			case PBinding.ACTION_READ:
			case PBinding.ACTION_EXIST:
			case PBinding.ACTION_UNEXPORT:
			default :
				return PBinding.LIFECYCLE_ERROR;
			}
		case PBinding.LIFECYCLE_NOTBOUNDNOACC:
			switch (action) {
			case PBinding.ACTION_BIND:
				return PBinding.LIFECYCLE_NOACC;
			case PBinding.ACTION_SETPA:
				return PBinding.LIFECYCLE_NOTBOUND;
			case PBinding.ACTION_UNSETPA:
			case PBinding.ACTION_UNBIND:
				return currentstate;
			case PBinding.ACTION_WRITE:
			case PBinding.ACTION_READ:
			case PBinding.ACTION_EXIST:
			case PBinding.ACTION_EXPORT:
			case PBinding.ACTION_UNEXPORT:
			default :
				return PBinding.LIFECYCLE_ERROR;
			}
		default :
			return PBinding.LIFECYCLE_ERROR;
		}
	}
}
