/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.cmdline;

import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.compiler.lib.JormCompiler;
import org.objectweb.jorm.util.lib.BasicClassLoader;
import org.objectweb.jorm.util.lib.CompilerLogger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.FileInputStream;

/**

 * Compiler of the persistence tool. This object manages the command line

 * parsing, the creation of an object schema, the generation of java files,

 * and the compilation to obtain class files.
 * @author	X. Spengler

 */
public class CommandLine {
	/**

	 * compilationContext object contains informations read from the

	 * command line of the pmc compiler.

	 */
	private CompilerParameter compilerParameter;

	/**

	 * a simple logger to log

	 */
	private Logger logger;

	/**
	 * the classloader which will be used in the jorm process
	 */
	private BasicClassLoader classLoader;

	/**
	 * Builds a new CommandLine object.
	 */
	public CommandLine() {
		classLoader = new BasicClassLoader();
	}

	/**

	 * Builds a new CommandLine object.
	 *

	 * @param args	the arguments of the pmc command line

	 */
	public static void main(String[] args) {
		CommandLine commandLine = new CommandLine();
		//commandLine.defineLogger();
		commandLine.parseCommandLine(args);
		commandLine.process();
	}

	/**
	 * Defines a logger for the comp.compiler package.
	 */
	private void defineLogger() {
		logger = CompilerLogger.getLoggerFactory().getLogger(
			"org.objectweb.jorm.compiler");
		logger.setIntLevel(BasicLevel.INFO);
	}

	/**
	 * Creates a new JormCompiler object and launches the process to generate
	 * files and compile them.
	 */
	private void process() {
		JormCompiler jormCompiler =
			new JormCompiler(compilerParameter, classLoader);
		jormCompiler.setLogger(logger);
		jormCompiler.setLoggerFactory(CompilerLogger.getLoggerFactory());

		if (compilerParameter.getInputFileNames() == null) {
            if (logger != null)
                logger.log(BasicLevel.WARN,
                        "no description files given, can not continue");
            CommandLineInfo.usage();
			System.exit(-9);
		}
		jormCompiler.process();
	}

	/**

	 * Parses the command line and initializes the compilation context object.
	 * This method verifies if no argument is missing to continue the process.
	 *

	 * @param args	the arguments of the command line

	 */
	private void parseCommandLine(String[] args) {
		CommandLineParser clp = new CommandLineParser(classLoader);

		try {
			new FileInputStream("jorm.properties");
			clp.getCompilerParameter().loadConfFile("jorm.properties");
		}
		catch (Exception e) {
			// nothing for the moment
		}
		if (CompilerLogger.getLoggerFactory()!=null) {
			clp.setLogger(CompilerLogger.getLoggerFactory()
				.getLogger("org.objectweb.jorm.cmdline"));
		}

		clp.parse(args, clp.getCompilerParameter().getCPExtension());
		compilerParameter = clp.getCompilerParameter();
	}
}

