/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.cmdline;

import org.objectweb.jorm.comp.cmdline.api.CmdLineParser;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.compiler.lib.BasicCompilerParameter;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.util.lib.BasicClassLoader;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.io.api.PathExplorer;
import org.objectweb.util.io.lib.DirJavaExplorer;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Hashtable;
import java.util.Vector;

/**
 * The CommandLineParser object parses the command line of the compiler
 * to build a compilation context object. This object will be used to parse
 * xml files, to generate java source files and to compile java files into
 * class files.
 * @author	X. Spengler

 */
public class CommandLineParser implements Loggable {
	/**
	 * a CompilerParameter to fill in when the command line
	 * is parsed. Informations are results of parsed options.
	 */
	private BasicCompilerParameter compilerParameter;

	/**
	 * the path separator used on the current system
	 */
	static String pathSeparator = System.getProperty("path.separator");

	/**
	 * a simple logger to log
	 */
	private Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	private LoggerFactory loggerFactory;

	/**
	 * exts hashtable contains commandLineParser extension objects.
	 * key: the name of the class (without the packages)
	 * value: the CmdLineParser object (an implementation)
	 */
	private Hashtable exts;

	/**
	 * a basic classloader to load classes
	 */
	private BasicClassLoader classLoader;

	/**
	 * Builds a new CommandLineParser object.
	 * This constructor initializes the compilationcontext to fill in.
	 *
	 * @param classLoader   the classloader to use to load classes
     * @deprecated BasicClassloader is not a class loader! Use default constructor (Christophe Ney)
	 */
    public CommandLineParser(BasicClassLoader classLoader) {
		this.classLoader = classLoader;
		compilerParameter = new BasicCompilerParameter(classLoader);
	}

    /**
     * default constructor
     */
    public CommandLineParser() {
        this(new BasicClassLoader());
	}

	/**
	 * Returns the current compilation context object.
	 *
	 * @return	the current CompilerParameter
	 */
	public CompilerParameter getCompilerParameter() {
		return compilerParameter;
	}

	/**

	 * Sets the java compiler to the compilation context.
	 *

	 * @param jc	the java compiler passed to the command line

	 */
	private void treatJavaCompiler(String jc) {
		compilerParameter.setJavaCompiler(jc);
	}

	/**

	 * Adds a filename to the compilation context file list.

	 * In the case the file is already present in this list, the file is
	 * not added.
	 *

	 * @param inputFileName	the name of a file passed to the command line

	 */
	private void treatInputFileName(String inputFileName) {
		if (!compilerParameter.addInputFileName(inputFileName)) {
			logger.log(BasicLevel.WARN, "Warning: The file " + inputFileName +
				" is given more than one time to process.");
        }
	}

	/**

	 * Prints the version number of the current release of the persistence
	 * compiler and returns.

	 */
	private void treatPrintVersion() {
		CommandLineInfo.printVersion();
		System.exit(0);
	}

	/**

	 * Displays to the programmer the usage of the persistence compiler
	 * and returns.

	 */
	private void treatHelp() {
		CommandLineInfo.usage();
		System.exit(0);
	}

	/**

	 * Sets to the compilation context the path used to process the
	 * global system.
	 *

	 * @param jormPath	the path used to search or retrieve files

	 */
	private void treatJormPaths(String jormPath) {
		if (System.getProperty("os.name").indexOf("Windows") >= 0)
			compilerParameter.setClasspath(parsePath(jormPath + ";."));
		else
			compilerParameter.setClasspath(parsePath(jormPath + ":."));

	}

	/**

	 * Parses a path and seperates the different elements to return a Vector
	 * representation of the path string.
	 *

	 * @param path	the path string to parse

	 */
	private PathExplorer parsePath(String path) {
		Vector result = new Vector();
		int previous, current;
		PathExplorer pathExplorer = new DirJavaExplorer();

		previous = 0;
		current = path.indexOf(pathSeparator, previous);
		while (current != -1) {
			result.add(path.substring(previous, current));
			previous = current + 1;
			current = path.indexOf(pathSeparator, previous);
		}
		result.add(path.substring(previous));
		pathExplorer.addPath(result);
		return pathExplorer;
	}

	/**

	 * Parses the command line arguments and builds the compilation context.
	 *

	 * @param args	the arguments of the command line (options)

	 * @param i		the current argument

	 * @return	    the number of the next arguments to proceed

	 */
	private int parseArgument(String args[], int i) {
		try {
			if (args[i].equals("-classpath"))
				treatJormPaths(args[++i]);
			else if (args[i].equals("-javac"))
				treatJavaCompiler(args[++i]);
			else if (args[i].equals("-d"))
				compilerParameter.setOutputDirectory(args[++i]);
			else if (args[i].equals("-v"))
				compilerParameter.setVerbose(true);
			else if (args[i].equals("-V"))
				treatPrintVersion();
			else if (args[i].equals("-h"))
				treatHelp();
			else if (args[i].equals("--help"))
				treatHelp();
			else if (args[i].equals("-help"))
				treatHelp();
			else if (args[i].equals("-keepsrc"))
				compilerParameter.setKeepSrc(true);
			else if (args[i].endsWith(".pd"))
				treatInputFileName(args[i]);
			else if (args[i].equals("-nobuild"))
				compilerParameter.setNoBuild(true);
			else if (args[i].equals("-mappername"))
				compilerParameter.setMapperName(args[++i]);
			else if (args[i].equals("-mappingname"))
				compilerParameter.setMappingName(args[++i]);
			else if (args[i].equals("-nocompile"))
				compilerParameter.setNoCompile(true);
			else if (args[i].equals("-parser"))
				compilerParameter.setParserName(args[++i]);
			else if (args[i].equals("-loglevel"))
				compilerParameter.setLogLevel(Integer.parseInt(args[++i]));
			else if (args[i].equals("-mappingfactory"))
				compilerParameter.setMappingFactory(args[++i]);
			else if (args[i].equals("-conf"))
				if (! treatConfFile(args[++i]))
                    throw new Exception();
			else if (args[i].equals("-noverify"))
				compilerParameter.setVerifyMetaInfo(false);
			else if (args[i].equals("-verifier"))
				compilerParameter.setVerifierName(args[++i]);
			else if (args[i].equals("-binderinheritance"))
				compilerParameter.setPClassMappingInheritance(args[++i]);
			else if (args[i].equals("-bindinginheritance"))
				compilerParameter.setPBindingInheritance(args[++i]);
            else if (args[i].equals("-metainfo"))
                compilerParameter.setMetaInfoName(args[++i]);
            else if (args[i].equals("-nodtdverify"))
                compilerParameter.setDtdVerify(false);
            else if (args[i].equals("-bindingabstract"))
                compilerParameter.setBindingAbstract(true);
			else if (args[i].equals("-ext"))
				extProcess(args[++i]);
			else {
				String option = args[i];
				String newOption = null;

				/**
				 we verify that's an option with a '-'
				 but verify if the option contains a '.'
				 and verify that the corresponding ext object exists in the
				 system
				 */
                logger.log(BasicLevel.ERROR, option.substring(1, option.lastIndexOf('.')));

				if ((option.startsWith("-")) &&
					(option.indexOf('.') != -1) &&
                        (exts.get(newOption = option.substring(1, option.lastIndexOf('.'))) != null)) {
					//(exts.get(newOption = option.substring(1, option.indexOf('.'))) != null)) {

					CmdLineParser obj = (CmdLineParser) (exts.get(newOption));
					i = obj.parseArgument(args, i);
					i--;

				}
				else {
					// it's not an option, a warning is sent to the user
                    if (logger != null)
                        logger.log(BasicLevel.WARN,
                                "the following option is not allowed :" + args[i]);
					CommandLineInfo.usage();
					System.exit(2);
				}
			}
		}
		catch (java.lang.ArrayIndexOutOfBoundsException e) {
			if (logger != null)
                logger.log(BasicLevel.ERROR,
                        "wrong argument");
			CommandLineInfo.usage();
			System.exit(-4);
		}
        catch (Exception e) {
            CommandLineInfo.usage();
			System.exit(-4);
        }
		return ++i;
	}

	private void extProcess(String ext) {
		//String key = ext.substring(ext.lastIndexOf('.') + 1, ext.length());
        String key = ext;

		// verify that a similar object (with similar name does not exist in the
		// system.
		if (exts.get(key) != null) {
            if (logger != null)
                logger.log(BasicLevel.ERROR, "a similar object is already existing" +
                        " in the system (" + key + ")");
			System.exit(-12);
		}
		else {
			CmdLineParser obj = null;
			try {
				obj = (CmdLineParser) (classLoader.load(ext));
				((Loggable)obj).setLogger(logger);
			}
			catch (Exception e) {
				e.printStackTrace();
                if (logger != null)
                    logger.log(BasicLevel.ERROR, "problem to load class");
				System.exit(-11);
			}
			exts.put(key, obj);
		}
	}

	public boolean treatConfFile(String file) {
		return compilerParameter.loadConfFile(file);
	}

	/**

	 * Parses all the arguments of the command line.
	 *

	 * @param args	the arguments of the command line

	 */
	public void parse(String args[], Hashtable exts) {
		this.exts = exts;
		int i = 0;

		while (i < args.length)
			i = parseArgument(args, i);
	}


	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////


	/**

	 * Defines a logger object.
	 *

	 * @param logger	the logger object

	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
		compilerParameter.setLogger(logger);
	}

	/**

	 * Defines the logger factory to obtain new logger.
	 *

	 * @param loggerFactory	the LoggerFactory object to obtain a logger object

	 */
	public void setLoggerFactory(LoggerFactory loggerFactory) {
		this.loggerFactory = loggerFactory;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}
}