/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.compiler.api;

import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.io.api.PathExplorer;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * The CompilerParameter interface defines methods to set an get informations
 * read by the command line parser. These informations are needed to
 * the compiler system to parse, generate, and compile.
 * @author	X. Spengler
 */
public interface CompilerParameter {
	/**
	 * The pattern that represents the JORM class name to be replaced into
	 * the string defining the inherited class of a either the PBinding or the
	 * PClassMapping that is generated.
	 */
	public final String INHERITANCECNPATTERN = "%c";

	/**
	 * Returns the directory where files must be generated.
	 *
	 * @return  null if the directory is not specified or a path with system
	 *          separator
	 */
	String getOutputDirectory();

	/**
	 * Defines the output directory for the generation of the java files.
	 *
	 * @param directory the name of the output directory
	 */
	void setOutputDirectory(String directory);

	/**
	 * Returns the class name which the generated PClassMapping classes must be
	 * extended.
	 * @return  null if the generated PClassMapping classes do not inherit of any
	 *          class or the full class name (with package name) using dot as
	 *          package separator.
	 */
	String getPClassMappingInheritance();

	/**
	 * Computes the actual class name which the generated PClassMapping class
	 * must extend.
	 * @param cn	The class name of the persistent class for which the
	 * 				code is generated.
	 * @return	The computed class name.
	 */
	String computePClassMappingInheritance(String cn);

	/**
	 * Returns the class name which the generated PBinding class must
	 * extend.
	 * @return  null if the generated PBinding classes do not inherit of any
	 *          class or the full class name (with package name) using dot as
	 *          package separator.
	 */
	String getPBindingInheritance();

	/**
	 * Computes the actual class name which the generated PBinding class
	 * must extend.
	 * @param cn	The class name of the persistent class for which the
	 * 				code is generated.
	 * @return	The computed class name.
	 */
	String computePBindingInheritance(String cn);

	/**
	 * Sets the name of the generated PBinder class name.
	 *
	 * @param v the class name which must be extended by the generated PBinder.
	 *          If the parameter is null, then the generated PBinder does not
	 *          inherit any class.
	 */
	void setPClassMappingInheritance(String v);

	/**
	 * Sets the generated PBinding class name.
	 *
	 * @param classname the class name which must be extended by the generated
	 *                  PBinding. If the parameter is null, then the generated
	 *                  PBinding does not inherit any class.
	 */
	void setPBindingInheritance(String classname);

	/**
	 * Returns the log level specified by the user.
	 *
	 * @return  the integer value of the log level
	 */
	int getLogLevel();

	/**
	 * Defines the level of the logger.
	 *
	 * @param level	the integer representation of the log level
	 */
	void setLogLevel(int level);

	/**
	 * Returns true if the user doesn't want that the generated files must be
	 * compiled.
	 *
	 * @return  true, if the user does not want that the generated files must
	 *          be compiled, else false
	 */
	boolean isNoCompile();

	/**
	 * Returns true if the user doesn't want to generate files.
	 *
	 * @return  true, if the user does not want to generate files
	 */
	boolean isNoBuild();

	/**
	 * Returns the class of the xml parser specified by the user.
	 *
	 * @return  the string representation of the parser class name
	 */
	String getParserName();

	/**
	 * Returns the mapper name specified by the user.
	 *
	 * @return  the name of the mapper
	 */
	String getMapperName();

	/**
	 * Returns the name of the mapping specified by the user.
	 *
	 * @return  the name of the mapping
	 */
	String getMappingName();

	/**
	 * Sets the keep source flag.
	 *
	 * @param keepSrc	true, the sources will not be destroyed, else false,
	 *                  the sources will be removed
	 */
	void setKeepSrc(boolean keepSrc);

	/**
	 * Allows to know if the programmer wants to keep the sources files or not.
	 *
	 * @return	true, the source files are not removed, else false, the source
	 *          files are removed
	 */
	boolean isKeepSrc();

	/**
	 * Sets the java compiler used to compile files when the generation process
	 * will be finished.
	 *
	 * @param jc	the Java compiler
	 */
	void setJavaCompiler(String jc);

	/**
	 * Returns the java compiler sets by the programmer on the command line.
	 *
	 * @return	the string representation of the java compiler name
	 */
	String getJavaCompiler();

	/**
	 * Sets the verbose mode.
	 *
	 * @param mode	true, verbose informations will be furnished, else false,
	 *              no informations are sent to the programmer
	 */
	void setVerbose(boolean mode);

	/**
	 * Allows to know if the user want to know verbose informations.
	 *
	 * @return	true, the verbose informations will be furnished, else false,
	 *          no informations are sent to the programmer
	 */
	boolean isVerbose();

	/**
	 * Adds a file name to the list of the input file names.
	 *
	 * @param fileName	the name of the file to add to the list
	 * @return          true, if the file is added to the list (was not yet
	 *                  present), else false, if the file is already in the list
	 *                  (and has not been added)
	 */
	boolean addInputFileName(String fileName);

	/**
	 * Allows to get a list of input file names (.pd files).
	 *
	 * @return	an iterator over the input file names (iterator of String)
	 */
	Iterator getInputFileNames();

	/**
	 * Prints informations managed by the current object.
	 */
	void print();

	/**
	 * Defines the classpath for the current jorm compilation.
	 *
	 * @param classpath	the string representation of the classpath
	 */
	void setClasspath(PathExplorer path);

	/**
	 * Returns the classpath used to search files or to compile.
	 *
	 * @return	the string representation of the classpath
	 */
	PathExplorer getClasspath();

	/**
	 * Defines the name of the mapper.
	 *
	 * @param mapperName	the name of the mapper
	 */
	void setMapperName(String mapperName);

	/**
	 * Defines the name of the mapping
	 *
	 * @param mappingName	the name of the mapping
	 */
	void setMappingName(String mappingName);

	/**
	 * Defines the name of the mapping factory used by the meta information
	 * to build mapping information.
	 *
	 * @param mappingFactory    the name of the mapping factory class
	 */
	void setMappingFactory(String mappingFactory);

	/**
	 * Returns the name of the mapping factory used by the meta information
	 * to build mapping information.
	 *
	 * @return  the name of the mapping factory class
	 */
	String getMappingFactory();

	/**
	 * Defines the name of the parser class which is able to parse the meta
	 * information related to a given mapping.
	 *
	 * @param parser_   the name of the parser
	 */
	void setParserName(String parser_);

	/**
	 * Sets if the user does not want to compile generated files.
	 *
	 * @param noCompile true, no files are compiled, else false
	 */
	void setNoCompile(boolean noCompile);

	/**
	 * Sets if the user does not want to generate files.
	 *
	 * @param noBuild   true, no files are generated, else false
	 */
	void setNoBuild(boolean nobuild);

	/**
	 * Loads a jorm configuration file.
	 *
	 * @param file  the name of the jorm configuration file
	 * @return      true, if everything is ok, else false
	 */
	boolean loadConfFile(String file);

	/**
	 * Sets if the user wants or not the verification of the meta information.
	 *
	 * @param verify    if true, the global check is made, else false
	 */
	void setVerifyMetaInfo(boolean verify);

	/**
	 * Returns if the user wants or not the verification of the meta
	 * information.
	 *
	 * @return  true, the user want the verification of the meta information,
	 *          else false
	 */
	boolean isVerifyMetaInfo();

	/**
	 * Sets the name of the class which makes the verification of the specific
	 * part of the meta-information.
	 *
	 * @param verifier  the name of the class verifier
	 */
	void setVerifierName(String verifier);

	/**
	 * Returns the name of the class which makes the verification of the
	 * specific part of the meta-information.
	 *
	 * @return  the name of the class verifier
	 */
	String getVerifierName();

	/**
	 * Returns an object, in most cases, it should be a
	 * com.mypackage.myobjects.CmdLineParser object.
	 * This object is created by the CommandLineParser to parse specific
	 * option on the command line. The returned object is composed by a list
	 * of getter, to obtain object values.<p>
	 * example: if a -myoption <value> is parsed by the a MyCmdLineParser object
	 * wich is an implementation of CmdLineParser. This object is returned by
	 * the getCPExtension method when it calls with "MyCmdLineParser" as
	 * argument.
	 *
	 * @param name  the name of the class with makes the extension of the
	 *              command line
	 */
	Object getCPExtension(String name);

	/**
	 * Allows to add into the hashtable object, the CommandLineParser
	 * extension objects which is identified by the @parameter name.
	 */
	void addCPExtension(String name, Object ext);

	/**
	 * Allows to get the hashtable object which contains the CommandLineParser
	 * extension objects.
	 * The key of the hashtable is the name of the extension class, and the
	 * value, the Class itself.
	 *
	 * @param exts_ the Hashtable object which contains the CommandLineParser
	 *              extension
	 */
	Hashtable getCPExtension();

    /**
     * Sets the class name of the meta information system.
     * The Class must be an implementation of the Manager api.
     *
     * @param metainfo  the class name of the meta information manager
     *                  implementation
     */
    void setMetaInfoName(String metainfo);

    /**
     * Returns the class name of the meta information system.
     * The Class is an implementation of the Manager api.
     *
     * @return  the String representation of the meta information manager
     *          implementation
     */
    String getMetaInfoName();

    /**
     * Allows to set if the dtd must be used or not to validate the xml files.
     *
     * @param verif if true, the dtd is used to verify the xml files, else false
     */
    void setDtdVerify(boolean verif);

    /**
     * Allows to know if the dtd is used or not to validate the xml files.
     *
     * @return  true, if the dtd must be used to validate xml files,<p>
     *          false, if the dtd is not used.
     */
    boolean isDtdVerify();

    /**
     * Enables or disables the generation of the dependances.
     *
     * @param gendep    true, the dependances are generated, else false
     */
    void setGenDep(boolean gendep);

    /**
     * Allows to know if the dependances must be generated or not.
     *
     * @return  true, if the dependances are generated, else false
     */
    boolean isGenDep();

    /**
     * Defines if the Binding object is defined abstract or not.
     *
     * @param bindingAbstract   true, the Binding object is abstract, else false
     */
    void setBindingAbstract(boolean bindingAbstract);

    /**
     * Allows to know if the Binding object is abstract or not.
     *
     * @return  true, the Binding object is abstract, else false
     */
    boolean isBindingAbstract();

    /**
     * Verifies environment context before the parsing must be launched.
     *
     * @return  true, if the context must be checked before using the parser,
     *          else false
     */
    boolean isVerifyContext();

    /**
     * Defines if the environment context must be verified before the parsing
     * must be launched.
     *
     * @param verifyContext if true, the context will be checked before to use
     *                      the parser, else false
     */
    void setVerifyContext(boolean verifyContext);
}
