/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.compiler.api;

import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.io.api.PathExplorer;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

/**
 * Compiler is an object used to process a Jorm compilation without using
 * the command line. This object is built by the user inside its own object
 * and the process() method is called to launch the computation.
 * @author	X. Spengler
 */
public interface JCompiler {
	/**
	 * Executes the jorm process.
	 * It creates a new JormCompiler object, and launches the generation of the
	 * java files and compiles them.
	 */
	boolean execute();

	/**
	 * Sets the keep source flag.
	 *
	 * @param keepSrc	true, the sources will not be destroyed, else false,
	 *                  the sources will be removed
	 */
	void setKeepSrc(boolean keepSrc);

	/**
	 * Sets the java compiler used to compile files when the generation
	 * process will be finished.
	 *
	 * @param jc	the Java compiler
	 */
	void setJavaCompiler(String jc);

	/**
	 * Sets the verbose mode.
	 *
	 * @param mode	true, verbose informations will be furnished, else false,
	 *              no informations are sent to the programmer
	 */
	void setVerbose(boolean mode);

	/**
	 * Adds a file name to the list of the input file names.
	 *
	 * @param fileName	the name of the file to add to the list
	 * @return          true, if the file is added to the list (was not yet
	 *                  present), else false, if the file is already in the
	 *                  list (and has not been added)
	 */
	boolean addInputFileName(String fileName);

	/**
	 * Defines the classpath for the current jorm compilation.
	 *
	 * @param classpath	the string representation of the classpath
	 */
	void setClasspath(PathExplorer path);

	/**
	 * Defines the name of the mapper.
	 *
	 * @param mapperName	the name of the mapper
	 */
	void setMapperName(String mapperName);

	/**
	 * Defines the name of the parser class which is able to parse the meta
	 * information related to a given mapping.
	 *
	 * @param parser_   the name of the parser
	 */
	void setParserName(String parser_);

	/**
	 * Returns the class of the xml parser specified by the user.
	 *
	 * @return  the name of the parser class
	 */
	String getParserName();

	/**
	 * Defines the name of the mapping.
	 *
	 * @param mappingName	the name of the mapping
	 */
	void setMappingName(String mappingName);

	/**
	 * Defines the output directory for the generation of the java files.
	 *
	 * @param directory the name of the output directory
	 */
	void setOutputDirectory(String directory);

	/**
	 * Defines the name of the mapping factory used by the meta information
	 * to build mapping information.
	 *
	 * @param mappingFactory    the name of the mapping factory
	 */
	void setMappingFactory(String mappingFactory);

	/**
	 * Sets if the user does not want to compile generated files.
	 *
	 * @param noCompile true, no files are compiled, else false
	 */
	void setNoCompile(boolean noCompile);

	/**
	 * Sets if the user does not want to generate files.
	 *
	 * @param noBuild   true, no files are generated, else false
	 */
	void setNoBuild(boolean noBuild);

	/**
	 * Defines a logger name and a list of aliases for a specific object.
	 *
	 * @param Logger	the logger object
	 */
	void setLogger(Logger logger);

	/**
	 * Defines the logger factory to obtain new logger.
	 *
	 * @param loggerFactory	the LoggerFactory object to obtain a logger object
	 */
	void setLoggerFactory(LoggerFactory loggerFactory);

	/**
	 * Returns the CompilerParameter object of the current compiler.
	 *
	 * @return  the CompilerParameter object or null if no exist
	 */
	CompilerParameter getCompilerParameter();

	/**
	 *
	 */
	void setPClassMappingInheritance(String classmapping);

	/**
	 * Defines the class name which the generated PBinding classes must be
	 * extended.
	 *
	 * @param binding	the string representation of the super class
	 */
	void setPBindingInheritance(String binding);

        /**
     * Sets the class name of the meta information system.
     * The Class must be an implementation of the Manager api.
     *
     * @param metainfo  the class name of the meta information manager
     *                  implementation
     */
    void setMetaInfoName(String metainfo);

    /**
     * Allows to set if the dtd must be used or not to validate the xml files.
     *
     * @param verif if true, the dtd is used to verify the xml files, else false
     */
    void setDtdVerify(boolean verif);

    /**
     * Defines if the Binding object is defined abstract or not
     * @param bindingAbstract   true, the Binding object is abstract, else false
     */
    void setBindingAbstract(boolean bindingAbstract);
}
