/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.compiler.lib;

import org.objectweb.jorm.comp.cmdline.api.CmdLineParser;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.util.lib.BasicClassLoader;
import org.objectweb.jorm.util.lib.CompilerLogger;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.io.api.PathExplorer;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.jorm.util.api.Loggable;
import java.io.FileInputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

/**

 * A BasicCompilerParameter object is an object which manages informations

 * read by the command line parser. These informations are needed to

 * the compiler system to parse, generate, and compile.
 * @author	X. Spengler

 */
public class BasicCompilerParameter implements CompilerParameter, Loggable {
	/**

	 * classpath is an object which stores a set of paths (or jar files).

	 */
	private PathExplorer classpath;

	/**

	 * path where files will be generated

	 */
	private String targetPath;

	/**

	 * verbose flag, set by the user

	 */
	private boolean verbose = false;

	/**

	 * a vector of input file names (.pd files)

	 */
	private Vector inputFileNames;

	/**

	 * keep the source files or not.

	 */
	private boolean keepSrc = true;

	/**

	 * javaCompiler is the name of the javac compiler used to compile

	 * the java sources.

	 */
	private String javaCompiler;

	/**

	 * mapperName is the name mapper (i.e: OR)

	 */
	private String mapperName;

	/**

	 * mappingName is the name mapping (i.e: org.objectweb. ...)

	 */
	private String mappingName;

	/**

	 * parser is the name of the class which manages

	 * the parser system.

	 */
	private String parser;

	/**

	 * if the compile flag is set to true, the java sources will be compiled,

	 * if false, there will be no compilation.

	 */
	private boolean noCompile;

	/**
	 * if the build flag is set to true, files are generated, else false.
	 */
	private boolean noBuild;

	/**

	 * the string separator for the current system.

	 */
	static String pathSeparator = System.getProperty("path.separator");

	/**
	 * the initial log level to log information
	 */
	private int logLevel;

	/**
	 * the name of the mapping factory, able to build mapping object.
	 */
	private String mappingFactory;

	/**
	 * if the verifyMetaInfo flag is set to true, the meta information is
	 * checked.
	 */
	private boolean verifyMetaInfo;

	/**
	 * the name of the class which makes the verification of the meta
	 * information.
	 */
	private String verifier;

	/**
	 *
	 */
	private String bindingInheritance = null;

	/**
	 *
	 */
	private String classMappingInheritance = null;

	/**
	 * exts hashtable contains commandLineParser extension objects.
	 * key: the name of the class (without the packages)
	 * value: the CmdLineParser object (an implementation)
	 */
	private Hashtable exts = null;

	/**
	 * a basic classloader to load classes
	 */
	private BasicClassLoader classLoader = null;

	/**
	 * a simple logger to log
	 */
	Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	LoggerFactory loggerFactory;

    /**
     * the class name of the meta information manager implementation
     */
    private String metaInfoName;

    /**
     * if the dtdVerify flag is set to true, the dtd is used to verify the xml
     * files, else false
     */
    private boolean dtdVerify;

    /**
     * if the genDep flag is set to true, the dependance will be generated
     */
    private boolean genDep;

    /**
     * if the bindingAbstract is set to true, the binding object will be
     * abstract, else false
     */
    private boolean bindingAbstract;

    /**
     * if the verifyContext flag is set to true, the environment context will
     * be checked before to use the parser
     */
    private boolean verifyContext;

	/**

	 * builds a new instance of a BasicCompilerParameter object. All attributes
	 * are initialized here.
	 *
	 * @param classLader    the classloader to use to load dynamicaly classes
     * @deprecated BasicClassloader is not a class loader! Use default constructor (Christophe Ney)
	 */
	public BasicCompilerParameter(BasicClassLoader classLoader) {
		inputFileNames = new Vector();
		classpath = null;
		targetPath = null;
		verbose = false;
		keepSrc = false;
		javaCompiler = "javac";
		parser = null;
		mapperName = null;
		mappingName = null;
		noCompile = false;
		noBuild = false;
		logLevel = 0;
		mappingFactory = null;
		verifyMetaInfo = true;
		verifier = null;
		this.classLoader = classLoader;
		exts = new Hashtable();
        metaInfoName = "org.objectweb.jorm.metainfo.lib.BasicManager";
        dtdVerify = true;
        genDep = false;
        bindingAbstract = false;
        verifyContext = true;
	}

    /**
     * Default constructor
     * @author Christophe Ney cney@batisseurs.com
     */
    public BasicCompilerParameter() {
        // don't be mistaken, basic classloader is not a classloader!
        // but this comes to using JORM code classloader to load extensions
        this(new BasicClassLoader());
    }


	///////////////////////////////////////////////////////////////////
	//                      from CompilerParameter interface
	///////////////////////////////////////////////////////////////////

	/**

	 * Sets the keep source flag.
	 *

	 * @param keepSrc	true, the sources will not be destroyed, else false,
	 *                  the sources will be removed

	 */
	public void setKeepSrc(boolean keepSrc) {
		this.keepSrc = keepSrc;
	}

	/**

	 * Allows to know if the programmer wants to keep the sources files or not.
	 *

	 * @return	true, the source files are not removed, else false, the source
	 *          files are removed

	 */
	public boolean isKeepSrc() {
		return keepSrc;
	}

	/**

	 * Sets the java compiler used to compile files when the generation

	 * process will be finished.
	 *

	 * @param jc	the Java compiler

	 */
	public void setJavaCompiler(String jc) {
		this.javaCompiler = jc;
	}

	/**

	 * Returns the java compiler sets by the programmer on the command line.
	 *

	 * @return	the string representation of the java compiler name

	 */
	public String getJavaCompiler() {
		return javaCompiler;
	}

	/**

	 * Sets the verbose mode.
	 *

	 * @param mode	true, verbose informations will be furnished, else false,
	 *              no informations are sent to the programmer

	 */
	public void setVerbose(boolean mode) {
		verbose = mode;
	}

	/**

	 * Allows to know if the user want to know verbose informations.
	 *

	 * @return	true, the verbose informations will be furnished, else false,
	 *          no informations are sent to the programmer

	 */
	public boolean isVerbose() {
		return verbose;
	}

	/**

	 * Adds a file name to the list of the input file names.
	 *

	 * @param fileName	the name of the file to add to the list

	 * @return	        true, if the file is added to the list (was not yet
	 *                  present), else false, if the file is already in the
	 *                  list (and has not been added)

	 */
	public boolean addInputFileName(String fileName) {
		if (!inputFileNames.contains(fileName)) {
			inputFileNames.add(fileName);
			return true;
		}
		else
			return false;
	}

	/**

	 * Allows to get a list of input file names (.pd files)
	 *

	 * @return	a vector representation of the input file names

	 */
	public Iterator getInputFileNames() {
        return inputFileNames.iterator();
	}

	/**

	 * Allows to know if the input file names list is empty or not.
	 *

	 * @return	true, if the input file names list is not empty, else false,
	 *          if the input file names list is empty

	 */
	public boolean existInputFileNames() {
		return !inputFileNames.isEmpty();
	}

	/**

	 * Allows to know if a target path name is defined to generate source files.
	 *

	 * @return	true, if a target path name is defined, else false, if a target
	 *          path name is yet defined

	 */
	public boolean existTargetPath() {
		return targetPath != null;
	}

	/**

	 * Parses a path, and returns a vector of entries.
	 * Entries are files, directories, etc...
	 *

	 * @param path	the path to parse

	 * @return	    a vector of entry

	 */
	public Vector parsePath(String path) {
		Vector result = new Vector();
		int previous, current;

		previous = 0;
		current = path.indexOf(pathSeparator, previous);
		while (current != -1) {
			result.add(path.substring(previous, current));

			previous = current + 1;
			current = path.indexOf(pathSeparator, previous);
		}
		result.add(path.substring(previous));
		return result;
	}

	/**

	 * Prints informations managed by the current object.

	 */
	public void print() {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "targetPath: " + targetPath);
            logger.log(BasicLevel.DEBUG, "verbose : " + verbose);
            logger.log(BasicLevel.DEBUG, "inputFileNames : ");
            for (int i = 0; i < inputFileNames.size(); i++)
                logger.log(BasicLevel.DEBUG, ((String) inputFileNames.elementAt(i)));
        }
	}

	/**

	 * Returns the directory where files must be generated.
	 *

	 * @return  null if the directory is not specified or a path with system

	 *          separator

	 */
	public String getOutputDirectory() {
		return targetPath;
	}

	/**

	 * Defines the directory where files are generated.
	 *

	 * @param outputdir	the string representation of the directory

	 */
	public void setOutputDirectory(String outputdir) {
		targetPath = outputdir;
	}

	/**

	 * Returns the class name which the generated PClassMapping classes must be
	 * extended.
	 *

	 * @return  null if the generated PClassMapping classes do not inherit of any
	 *          class or the full class name (with package name) using dot as
	 *          package separator

	 */
	public String getPClassMappingInheritance() {
		return classMappingInheritance;
	}

	/**
	 * Computes the actual class name which the generated PClassMapping class
	 * must extend.
	 * @param cn	The class name of the persistent class for which the
	 * 				code is generated.
	 * @return	The computed class name.
	 */
	public String computePClassMappingInheritance(String cn) {
		int ind = classMappingInheritance.indexOf(INHERITANCECNPATTERN);
		if (ind == -1)
			return classMappingInheritance;
		return classMappingInheritance.substring(0, ind) + cn
			+ classMappingInheritance.substring(ind + INHERITANCECNPATTERN.length());
	}

	public void setPClassMappingInheritance(String classmapping) {
		classMappingInheritance = classmapping;
	}

	/**

	 * Returns the class name which the generated PBinding classes must be
	 * extended.
	 *

	 * @return  null if the generated PBinding classes do not inherit of any
	 *          class or the full class name (with package name) using dot as
	 *          package separator

	 */
	public String getPBindingInheritance() {
		return bindingInheritance;
	}

	/**
	 * Computes the actual class name which the generated PBinding class
	 * must extend.
	 * @param cn	The class name of the persistent class for which the
	 * 				code is generated.
	 * @return	The computed class name.
	 */
	public String computePBindingInheritance(String cn) {
		int ind = bindingInheritance.indexOf(INHERITANCECNPATTERN);
		if (ind == -1)
			return bindingInheritance;
		return bindingInheritance.substring(0, ind) + cn
			+ bindingInheritance.substring(ind + INHERITANCECNPATTERN.length());
	}

	/**

	 * Defines the class name which the generated PBinding classes must be
	 * extended.
	 *
	 * @param binding	the string representation of the super class

	 */
	public void setPBindingInheritance(String binding) {
		bindingInheritance = binding;
	}

	/**

	 * Returns the log level specified by the user.
	 *
	 * @return  the integer representation of the log level

	 */
	public int getLogLevel() {
		return logLevel;
	}

	/**

	 * Defines the level of the logger.
	 *

	 * @param level	the integer representation of the log level

	 */
	public void setLogLevel(int level) {
		logLevel = level;
	}

	/**

	 * Returns true if the user doesn't want that the genrated files must be
	 * compiled.
	 */
	public boolean isNoCompile() {
		return noCompile;
	}

	/**

	 * Returns the class of the xml parser specified by the user.
	 *
	 * @return  the string representation of the parser class name

	 */
	public String getParserName() {
		return parser;
	}

	/**

	 * Defines the name of the parser class which is able to parse the meta
	 * information related to a given mapping.
	 *
	 * @param parser   the name of the parser class name

	 */
	public void setParserName(String parser) {
		this.parser = parser;
	}

	/**

	 * Returns the class of the mapper specified by the user.
	 *
	 * @return  the name of the mapper

	 */
	public String getMapperName() {
		return mapperName;
	}

	/**

	 * Defines the name of the mapper used for the meta information.
	 *

	 * @param map	the name of the mapper

	 */
	public void setMapperName(String mapperName) {
		this.mapperName = mapperName;
	}

	/**

	 * Defines the name of the mapping.
	 *

	 * @param mappingName	the name of the mapping

	 */
	public void setMappingName(String mappingName) {
		this.mappingName = mappingName;
	}

	/**

	 * Returns the class of the mapping specified by the user.
	 *
	 * @return  the name of the mapping

	 */
	public String getMappingName() {
		return mappingName;
	}

	/**

	 * Defines the classpath for the current jorm compilation.
	 *

	 * @param classpath	the string representation of the classpath

	 */
	public void setClasspath(PathExplorer path) {
		classpath = path;
	}

	/**
	 * Defines the name of the mapping factory used by the meta information
	 * to build mapping information.
	 *
	 * @param mappingFactory    the class name of the mapping factory
	 */
	public void setMappingFactory(String mappingFactory) {
		this.mappingFactory = mappingFactory;
	}

	/**
	 * Returns the name of the mapping factory used by the meta information
	 * to build mapping information.
	 *
	 * @return  the name of the mapping factory class
	 */
	public String getMappingFactory() {
		return mappingFactory;
	}

	/**

	 * Returns the classpath used to search files or to compile.
	 *

	 * @return	the string representation of the classpath

	 */
	public PathExplorer getClasspath() {
		return classpath;
	}

	/**
	 * Dets if the user does not want to compile generated files.
	 *
	 * @param noCompile true, no files are compiled, else false
	 */
	public void setNoCompile(boolean noCompile) {
		this.noCompile = noCompile;
	}

	/**
	 * Sets if the user does not want to generate files.
	 *
	 * @param noBuild   true, no files are generated, else false
	 */
	public void setNoBuild(boolean noBuild) {
		this.noBuild = noBuild;
	}

	/**
	 * Returns true if the user doesn't want to generate files.
	 *
	 * @return  true, if the user does not want to generate files
	 */
	public boolean isNoBuild() {
		return noBuild;
	}

	/**
	 * Sets if the user wants or not the verification of the meta information.
	 *
	 * @param verify    if true, the global check is made, else false
	 */
	public void setVerifyMetaInfo(boolean verify) {
		verifyMetaInfo = verify;
	}

	/**
	 * Returns if the user wants or not the verification of the meta
	 * information.
	 *
	 * @return  true, the user want the verification of the meta information,
	 *          else false
	 */
	public boolean isVerifyMetaInfo() {
		return verifyMetaInfo;
	}

	/**
	 * Sets the name of the class which makes the verification of the specific
	 * part of the meta-information.
	 *
	 * @param verifier  the name of the class verifier
	 */
	public void setVerifierName(String verifier) {
		this.verifier = verifier;
	}

	/**
	 * Returns the name of the class which makes the verification of the
	 * specific part of the meta-information.
	 *
	 * @return  the name of the class verifier
	 */
	public String getVerifierName() {
		return verifier;
	}

	/**
	 * Load a jorm configuration file.
	 *
	 * @param file  the name of the jorm configuration file
	 * @return      true, if everything is ok, else false
	 */
	public boolean loadConfFile(String file) {
		Properties p = new Properties();
		try {
			p.load(new FileInputStream(file));
			// passes to the CompilerLogger new information about the
			// log configuration
            String s = null;
            if ((s = p.getProperty("log.config.file")) != null) {
                Properties plog = new Properties();
                plog.load(new FileInputStream(s));
                CompilerLogger.configure(plog);
            }
			if (CompilerLogger.getLoggerFactory()!=null && logger==null) {
				logger = CompilerLogger.getLoggerFactory().getLogger(
					"org.objectweb.jorm.compiler");
			}


			if ((s = p.getProperty("keepsrc")) != null)
				setKeepSrc(
					(new Boolean(s)).booleanValue());

			if ((s = p.getProperty("verbose")) != null)
				setVerbose(
					(new Boolean(s)).booleanValue());

			if ((s = p.getProperty("javacompiler")) != null)
				setJavaCompiler(s);

			if ((s = p.getProperty("mappername")) != null)
				setMapperName(s);

			if ((s = p.getProperty("mappingname")) != null)
				setMappingName(s);

			if ((s = p.getProperty("mappingfactory")) != null)
				setMappingFactory(s);

			if ((s = p.getProperty("parser")) != null)
				setParserName(s);

			if ((s = p.getProperty("bindinginheritance")) != null)
				setPBindingInheritance(s);

			if ((s = p.getProperty("classmappinginheritance")) != null)
				setPClassMappingInheritance(s);

            if ((s = p.getProperty("metainfo")) != null)
				setMetaInfoName(s);

            if ((s = p.getProperty("dtdverify")) != null)
                setDtdVerify(
                        (new Boolean(s)).booleanValue());

			if ((s = p.getProperty("inputfile")) != null) {
				StringTokenizer tok = new StringTokenizer(s, ",");
				while (tok.hasMoreTokens()) {
					String inputFileName = tok.nextToken().trim();
					addInputFileName(inputFileName);
				}
			}

			// this case concerns the extension of the commandlineparser object
			if ((s = p.getProperty("cpextension")) != null) {
				StringTokenizer tok = new StringTokenizer(s, ",");
				while (tok.hasMoreTokens()) {
					String ext = tok.nextToken().trim();
					String key = ext.substring(ext.lastIndexOf('.') + 1, ext.length());
					// verify that a similar object (with similar name does not exist in the
					// system.
					if ((exts.size() == 0) || (exts.get(key) == null)) {
						CmdLineParser obj = null;
						try {
							obj = (CmdLineParser) (classLoader.load(ext));
							((Loggable)obj).setLogger(logger);
						}
						catch (Exception e) {
							e.printStackTrace();
							if (logger!=null)
								logger.log(BasicLevel.ERROR,
									"problem to load class " + ext);
							System.exit(-11);
						}
						exts.put(key, obj);
					}
				}
			}

			// now, try to load other properties from other commandlineparser
			for (Enumeration e = exts.keys(); e.hasMoreElements();) {
				String key = (String) (e.nextElement());
				for (Enumeration en = p.propertyNames(); en.hasMoreElements();) {
					String elem = (String) (en.nextElement());
					if (elem.startsWith(key + ".")) {
						CmdLineParser obj = (CmdLineParser) (exts.get(key));
						obj.parseProperty(elem, p.getProperty(elem));
					}

				}
			}

		}
		catch (Exception e) {
			return false;
		}
		return true;
	}

	/**
	 * Returns an object, in most cases, it should be a
	 * com.mypackage.myobjects.CmdLineParser object.
	 * This object is created by the CommandLineParser to parse specific
	 * option on the command line. The returned object is composed by a list
	 * of getter, to obtain object values.<p>
	 * example: if a -myoption <value> is parsed by the a MyCmdLineParser object
	 * wich is an implementation of CmdLineParser. This object is returned by
	 * the getCPExtension method when it calls with "MyCmdLineParser" as
	 * argument.
	 *
	 * @param name  the name of the class with makes the extension of the
	 *              command line
	 */
	public Object getCPExtension(String name) {
		return exts.get(name);
	}

	/**
	 * Allows to add into the hashtable object, the CommandLineParser
	 * extension objects which is identified by the @parameter name.
	 */
	public void addCPExtension(String name, Object ext) {
		exts.put(name, ext);
	}

	/**
	 * Allows to get the hashtable object which contains the CommandLineParser
	 * extension objects.
	 * The key of the hashtable is the name of the extension class, and the
	 * value, the Class itself.
	 *
	 * @param exts_ the Hashtable object which contains the CommandLineParser
	 *              extension.
	 */
	public Hashtable getCPExtension() {
		return exts;
	}

    /**
     * Sets the class name of the meta information system.
     * The Class must be an implementation of the Manager api.
     *
     * @param metainfo  the class name of the meta information manager
     *                  implementation
     */
    public void setMetaInfoName(String metainfo) {
        metaInfoName = metainfo;
    }

    /**
     * Returns the class name of the meta information system.
     * The Class is an implementation of the Manager api.
     *
     * @return  the String representation of the meta information manager
     *          implementation
     */
    public String getMetaInfoName() {
        return metaInfoName;
    }

    /**
     * Allows to set if the dtd must be used or not to validate the xml files.
     *
     * @param verif if true, the dtd is used to verify the xml files, else false
     */
    public void setDtdVerify(boolean verif) {
        dtdVerify = verif;
    }

    /**
     * Allows to know if the dtd is used or not to validate the xml files.
     *
     * @return  true, if the dtd must be used to validate xml files,<p>
     *          false, if the dtd is not used.
     */
    public boolean isDtdVerify() {
        return dtdVerify;
    }

    /**
     * Enables or disables the generation of the dependances.
     *
     * @param gendep    true, the dependances are generated, else false
     */
    public void setGenDep(boolean gendep) {
        genDep = gendep;
    }

    /**
     * Allows to know if the dependances must be generated or not.
     *
     * @return  true, if the dependances are generated, else false
     */
    public boolean isGenDep() {
        return genDep;
    }

    /**
     * Defines if the Binding object is defined abstract or not.
     *
     * @param bindingAbstract   true, the Binding object is abstract, else false
     */
    public void setBindingAbstract(boolean bindingAbstract) {
        this.bindingAbstract = bindingAbstract;
    }

    /**
     * Allows to know if the Binding object is abstract or not.
     *
     * @return  true, the Binding object is abstract, else false
     */
    public boolean isBindingAbstract() {
        return bindingAbstract;
    }

    /**
     * Verifies environment context before the parsing must be launched.
     *
     * @return  true, if the context must be checked before using the parser,
     *          else false
     */
    public boolean isVerifyContext() {
        return verifyContext;
    }

    /**
     * Defines if the environment context must be verified before the parsing
     * must be launched.
     *
     * @param verifyContext if true, the context will be checked before to use
     *                      the parser, else false
     */
    public void setVerifyContext(boolean verifyContext) {
        this.verifyContext = verifyContext;
    }

	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	/**
	 * Defines the logger factory to obtain new logger.
	 *
	 * @param loggerFactory	the LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerFactory) {
		this.loggerFactory = loggerFactory;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}
}