/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.compiler.lib;

import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.compiler.api.JCompiler;
import org.objectweb.jorm.util.lib.BasicClassLoader;
import org.objectweb.jorm.util.lib.CompilerLogger;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.io.api.PathExplorer;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Properties;
import java.io.FileInputStream;
import java.io.IOException;

/**
 * Compiler is an object used to process a Jorm compilation without using
 * the command line. This object is built by the user inside its own object
 * and the process() method is called to launch the computation.
 * @author	X. Spengler
 */
public class
        Compiler implements JCompiler, Loggable {
	/**
	 * the jorm compiler object is the main object of the jorm compilation.
	 * its launches all the different processes.
	 */
	private JormCompiler jormCompiler = null;

	/**
	 * the compiler parameter object is used to store options and parameters
	 * used by the jorm compiler to make the process.
	 */
	private CompilerParameter compilerParameter;

	/**
	 * a simple logger to log
	 */
	private Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	private LoggerFactory loggerFactory;

	private BasicClassLoader classLoader;

	/**
	 * Builds a new compiler.
	 *
	 * @param confFile  the string representation of a properties file to load
	 */
	public Compiler(String confFile) {
		classLoader = new BasicClassLoader();
		compilerParameter = new BasicCompilerParameter(classLoader);
		compilerParameter.loadConfFile(confFile);
        if (CompilerLogger.getLoggerFactory() != null)
            logger = CompilerLogger.getLoggerFactory().getLogger(
                    "org.objectweb.jorm.compiler");
	}

	/**
	 * Builds a new compiler.
	 *
	 * @param jormConfFile  the string representation of a properties file to
	 * load
	 * @param logConfFile  the string representation of a monolog properties
	 * file to load
	 */
	public Compiler(String jormConfFile, String logConfFile)
		throws IOException {
		Properties p = new Properties();
		p.load(new FileInputStream(logConfFile));
		CompilerLogger.configure(p);

		classLoader = new BasicClassLoader();
		compilerParameter = new BasicCompilerParameter(classLoader);
		compilerParameter.loadConfFile(jormConfFile);
        if (CompilerLogger.getLoggerFactory() != null)
            logger = CompilerLogger.getLoggerFactory().getLogger(
                    "org.objectweb.jorm.compiler");
	}

    /**
     * @deprecated BasicClassloader is not a class loader! need refactoring (Christophe Ney)
	 */
	public BasicClassLoader getClassLoader() {
		return classLoader;
	}

	///////////////////////////////////////////////////////////////////
	//                      from JCompiler interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Sets the keep source flag.
	 *
	 * @param keepSrc	true, the sources will not be destroyed, else false,
	 *                  the sources will be removed
	 */
	public void setKeepSrc(boolean keepSrc) {
		compilerParameter.setKeepSrc(keepSrc);
	}

	/**
	 * Sets the java compiler used to compile files when the generation
	 * process will be finished.
	 *
	 * @param jc	the Java compiler
	 */
	public void setJavaCompiler(String jc) {
		compilerParameter.setJavaCompiler(jc);
	}

	/**
	 * Sets the verbose mode.
	 *
	 * @param mode	true, verbose informations will be furnished, else false,
	 *              no informations are sent to the programmer
	 */
	public void setVerbose(boolean mode) {
		compilerParameter.setVerbose(mode);
	}

	/**
	 * Adds a file name to the list of the input file names.
	 *
	 * @param fileName	the name of the file to add to the list
	 * @return	        true, if the file is added to the list (was not yet
	 *                  present), else false, if the file is already in the
	 *                  list (and has not been added)
	 */
	public boolean addInputFileName(String fileName) {
		return compilerParameter.addInputFileName(fileName);
	}

	/**
	 * Defines the classpath for the current jorm compilation.
	 *
	 * @param classpath	the string representation of the classpath
	 */
	public void setClasspath(PathExplorer path) {
		compilerParameter.setClasspath(path);
	}

	/**
	 * Defines the name of the mapper.
	 *
	 * @param mapperName	the name of the mapper
	 */
	public void setMapperName(String mapperName) {
		compilerParameter.setMapperName(mapperName);
	}

	/**
	 * Defines the name of the mapping.
	 *
	 * @param mappingName	the name of the mapping
	 */
	public void setMappingName(String mappingName) {
		compilerParameter.setMappingName(mappingName);
	}

	/**
	 * Defines the output directory for the generation of the java files.
	 *
	 * @param directory the name of the output directory
	 */
	public void setOutputDirectory(String directory) {
		compilerParameter.setOutputDirectory(directory);
	}

	/**
	 * Defines the name of the mapping factory used by the meta information
	 * to build mapping information.
	 *
	 * @param mappingFactory    the class name of the mapping factory
	 */
	public void setMappingFactory(String mappingFactory) {
		compilerParameter.setMappingFactory(mappingFactory);
	}

	/**
	 * Sets if the user does not want to compile generated files.
	 *
	 * @param noCompile true, no files are compiled, else false
	 */

	public void setNoCompile(boolean noCompile) {
		compilerParameter.setNoCompile(noCompile);
	}

	/**
	 * Sets if the user does not want to generate files.
	 *
	 * @param noBuild   true, no files are generated, else false
	 */
	public void setNoBuild(boolean noBuild) {
		compilerParameter.setNoBuild(noBuild);
	}

	/**
	 * Executes the jorm process.
	 * It creates a new JormCompiler object, and launches the generation of the
	 * java files and compiles them.
	 */
	public boolean execute() {
        if (CompilerLogger.getLoggerFactory() != null) {
            logger = CompilerLogger.getLoggerFactory().getLogger(
                    "org.objectweb.jorm.comp.compiler");
            logger.setIntLevel(BasicLevel.WARN);
        }

        JormCompiler jormCompiler =
			new JormCompiler(compilerParameter, classLoader);
		jormCompiler.setLogger(logger);
		jormCompiler.setLoggerFactory(CompilerLogger.getLoggerFactory());

		if (compilerParameter.getInputFileNames() == null) {
			//logger.log(BasicLevel.WARN,
			//	"no description files given, can not continue");
			//System.exit(0);
            return true;
		}
		return jormCompiler.process();
	}

	/**
	 * Returns the CompilerParameter object of the current compiler.
	 *
	 * @return  the CompilerParameter object or null if no exist
	 */
	public CompilerParameter getCompilerParameter() {
		return compilerParameter;
	}

	/**
	 * Returns the class of the xml parser specified by the user.
	 *
	 * @return  the class name of the parser
	 */
	public String getParserName() {
		return compilerParameter.getParserName();
	}

	/**
	 * Defines the name of the parser class which is able to parse the meta
	 * information related to a given mapping.
	 *
	 * @param parser   the name of the parser class
	 */
	public void setParserName(String parser) {
		compilerParameter.setParserName(parser);
	}

	public void setPClassMappingInheritance(String classmapping) {
		compilerParameter.setPClassMappingInheritance(classmapping);
	}

	/**
	 * Defines the class name which the generated PBinding classes must
	 * be extended.
	 *
	 * @param binding	the string representation of the super class
	 */
	public void setPBindingInheritance(String binding) {
		compilerParameter.setPBindingInheritance(binding);
    }

    /**
     * Sets the class name of the meta information system.
     * The Class must be an implementation of the Manager api.
     *
     * @param metainfo  the class name of the meta information manager
     *                  implementation
     */
    public void setMetaInfoName(String metainfo) {
        compilerParameter.setMetaInfoName(metainfo);
    }

    /**
     * Allows to set if the dtd must be used or not to validate the xml files.
     *
     * @param verif if true, the dtd is used to verify the xml files, else false
     */
    public void setDtdVerify(boolean verif) {
        compilerParameter.setDtdVerify(verif);
    }

    /**
     * Defines if the Binding object is defined abstract or not
     * @param bindingAbstract   true, the Binding object is abstract, else false
     */
    public void setBindingAbstract(boolean bindingAbstract) {
        compilerParameter.setBindingAbstract(bindingAbstract);
    }

    ///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
    public void setLogger(Logger logger) {
        this.logger = logger;
        if (compilerParameter.isVerbose())
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.setIntLevel(BasicLevel.DEBUG);
    }

    /**
     * Defines the logger factory to obtain new logger.
     *
	 * @param loggerFactory	the LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerFactory) {
		this.loggerFactory = loggerFactory;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}

}
