/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.compiler.lib;

import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.generator.api.JormGenerator;
import org.objectweb.jorm.comp.generator.api.MOPFactory;
import org.objectweb.jorm.comp.generator.lib.JormGeneratorImpl;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.parser.api.MappingParser;
import org.objectweb.jorm.parser.api.Parser;
import org.objectweb.jorm.parser.lib.BasicDomParser;
import org.objectweb.jorm.parser.lib.ParserErrorHandler;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.util.lib.BasicClassLoader;
import org.objectweb.jorm.verifier.api.MappingVerifier;
import org.objectweb.jorm.verifier.api.Verifier;
import org.objectweb.jorm.verifier.lib.BasicVerifier;
import org.objectweb.util.io.api.TargetHolder;
import org.objectweb.util.io.lib.JavaFileHolder;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;


/**
 * Compiler of the persistence tool. This object manages the command line
 * parsing, the creation of an object schema, the generation of java files,
 * and the compilation to obtain class files.
 * @author	X. Spengler
 */
public class JormCompiler implements Loggable {

    public final static String RDB_MAPPER_NAME = "rdb";
    public final static String FILE_MAPPER_NAME = "file";
    public final static String MEMORY_MAPPER_NAME = "memory";

    /**
     * parser object is able to parse XML files and build an internal
     * representation of its content.
     */
    Parser parser;

    /**
     * manager object is an objet charged to manage meta objects.
     */
    Manager manager;

    /**
     * generator is a simple JormGenerator object, which is able to generate
     * java file and to compile them.
     */
    JormGenerator generator;

    /**
     * compilerParameter object contains informations read from the
     * command line of the pmc compiler.
     */
    CompilerParameter compilerParameter;

    /**
     * a basic class loader to load class passed by parameters to the pmc
     * command line. This class loader is able to pick a class name and
     * to return an object representation of the class.
     */
    BasicClassLoader classLoader;

    /**
     * a simple logger to log
     */
    Logger logger;

    /**
     * a logger factory to create other loggers if needed
     */
    LoggerFactory loggerFactory;

    /**
     * the logger for the meta information
     */
    Logger metaLogger;

    /**
     * the logger for the parser system
     */
    Logger parserLogger;

    /**
     * This hashmap describes the associations between a mapper name and a
     * MOPFactory.
     */
    HashMap mapper2mop = null;

    /**
     * basicVerifier is the verifier read on the command line or the
     * jorm.properties file.
     */
    Verifier basicVerifier;

    Collection genDepCollection = null;

    /**
     * Builds a new JormCompiler object.
     * @param compileParameter	the list of arguments read on the command line
     * @deprecated BasicClassloader is not a classloader! Use another constructor (Christophe Ney)
	 */
    public JormCompiler(CompilerParameter compilerParameter,
                        BasicClassLoader classLoader) {
        this.compilerParameter = compilerParameter;
        this.classLoader = classLoader;
        mapper2mop = new HashMap(2);
        mapper2mop.put(RDB_MAPPER_NAME,
                "org.objectweb.jorm.comp.generator.rdb.RdbMOPFactory");
        mapper2mop.put(FILE_MAPPER_NAME,
                "org.objectweb.jorm.comp.generator.file.FileMOPFactory");
        mapper2mop.put(MEMORY_MAPPER_NAME,
                "org.objectweb.jorm.comp.generator.memory.MemoryMOPFactory");
    }

    /**
     * Builds a new JormCompiler object.
     * @param compileParameter	the list of arguments read on the command line
     */
    public JormCompiler(CompilerParameter compilerParameter) {
        this(compilerParameter,new BasicClassLoader());
    }

    /**
     * Builds a new JormCompiler object.
     * @param compileParameter	the list of arguments read on the command line
     * @param manager a meta info manager instance
     */
    public JormCompiler(CompilerParameter compilerParameter,Manager manager) {
        this(compilerParameter,new BasicClassLoader());
        this.manager = manager;

    }

    /**
     * Creates class instances for the XML parser, the Meta Information Manager
     * and manages the JormGenerator system.
     *
     * @return  true, if everything works properly, else false
     */
    public boolean process() {
        // verifies that all the needed parameters are available for a
        // correct processus
        if (compilerParameter.isVerifyContext())
            verifyContext();

        // parser part of the Jorm Compiler System
        parseFiles();
        verifyMetaInfo();

        if (!compilerParameter.isNoBuild()) {
            try {
                if (!generateFiles(null))
                    return false;
            }
            catch (Exception e) {
                if (logger != null)
                    logger.log(BasicLevel.ERROR, "Problem during generation:", e);
                return false;
            }
        }
        if (!compilerParameter.isNoCompile()) {
            compileFiles();
        }
        return true;
    } // method process

    /**
     * Parses the input files and build the meta information.
     */
    public void parseFiles() {
        parser = new BasicDomParser(compilerParameter.isDtdVerify());
        try {
            manager = (Manager) classLoader.load(compilerParameter.getMetaInfoName());
            manager.init(classLoader);
        } catch (Exception e) {
            e.printStackTrace();
            if (logger != null)
                logger.log(BasicLevel.ERROR, "problem to load class");
            System.exit(-11);
        }
        if (loggerFactory != null)
            metaLogger = loggerFactory.getLogger("org.objectweb.jorm.metainfo");

        ((Loggable) manager).setLogger(metaLogger);

        if (compilerParameter.getMappingFactory() != null)
            manager.addMappingFactory(compilerParameter.getMapperName(),
                    compilerParameter.getMappingFactory());

        // parser part of the Jorm Compiler System
        if (loggerFactory != null)
            parserLogger = loggerFactory.getLogger("org.objectweb.jorm.parser");

        ((Loggable) parser).setLogger(parserLogger);
        parser.setMetaInfoManager(manager);
        parser.setPathExplorer(compilerParameter.getClasspath());
        parser.setMapperName(compilerParameter.getMapperName());
        parser.setMappingName(compilerParameter.getMappingName());
        if (compilerParameter.getParserName() != null)
            try {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "try to load Parser object (" +
                            compilerParameter.getParserName() + ")");
                parser.setMappingParser((MappingParser) (classLoader.load
                        (compilerParameter.getParserName())));
            } catch (Exception e) {
                e.printStackTrace();
                if (logger != null)
                    logger.log(BasicLevel.ERROR, "problem to load class");
                System.exit(-11);
            }
        parser.setGenDep(compilerParameter.isGenDep());
        genDepCollection = parser.parse(compilerParameter.getInputFileNames());
    }

    /**
     * Verifies if the needed parameters are available for a normal processing.
     */
    private void verifyContext() {
        if (compilerParameter.getParserName() == null) {
            if (logger != null)
                logger.log(BasicLevel.ERROR,
                        "the parser name can not be null");
            System.exit(-31);
        }
        if (compilerParameter.getMappingFactory() == null) {
            if (logger != null)
                logger.log(BasicLevel.ERROR,
                        "the mapping factory name can not be null");
            System.exit(-32);
        }
        if (compilerParameter.getMapperName() == null) {
            if (logger != null)
                logger.log(BasicLevel.ERROR,
                        "the mapper name can not be null");
            System.exit(-33);
        }
        if (compilerParameter.getMappingName() == null) {
            if (logger != null)
                logger.log(BasicLevel.ERROR,
                        "the mapping name can not be null");
            System.exit(-34);
        }
    }

    /**
     * Verifies if the metainformation is in a properly state.
     */
    private void verifyMetaInfo() {
        if (compilerParameter.isVerifyMetaInfo()) {
            basicVerifier = new BasicVerifier();
            Logger verifierLogger = null;
            if (loggerFactory != null)
                verifierLogger = loggerFactory.getLogger(
                        "org.objectweb.jorm.verifier");

            //verifierLogger.setIntLevel(logger.getCurrentIntLevel());
            ((Loggable) basicVerifier).setLogger(verifierLogger);

            basicVerifier.setMetaInfoManager(manager);

            if (compilerParameter.getVerifierName() != null)
                try {
                    if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                        logger.log(BasicLevel.DEBUG, "try to load Verfier object (" +
                                compilerParameter.getVerifierName() + ")");
                    basicVerifier.setMappingVerifier(
                            (MappingVerifier) (classLoader.load
                            (compilerParameter.getVerifierName())));
                } catch (Exception e) {
                    e.printStackTrace();
                    if (logger != null)
                        logger.log(BasicLevel.ERROR, "problem to load class");
                    System.exit(-41);
                }
            basicVerifier.verify();

        }
    }

    /**

     * Sets the generator that the current compiler has to be used.
     * This method has to be used with care because a CodeGenerator object add
     * factories to the Generator.
     *

     * @param gen	a simple JormGenerator object

     */
    public void setJormGenerator(JormGenerator gen) {
        generator = gen;
    }

    /**

     * Returns the generator used by the current compiler.
     *

     * @return		a simple generator

     */
    public JormGenerator getJormGenerator() {
        return generator;
    }

    /**

     * Generates java files for all selected sources.
     * Sources are determined by the input files (.pd files) and the Object
     * Generation Factories added to the JormGenerator object.

     */
    public boolean generateFiles(Collection mo) throws Exception {

        // Find the right MOPFactory class associated to the mapper name

        java.lang.Class mopFactClass = null;
        String mopFcn = (String) mapper2mop.get(
                compilerParameter.getMapperName());
        if (mopFcn == null) {
            if (logger != null)
                logger.log(BasicLevel.ERROR, "No MOPFactory associated to the mapper"
                        + compilerParameter.getMapperName());
            System.exit(-1);
        }
        try {
            mopFactClass = java.lang.Class.forName(mopFcn);
        } catch (Exception e) {
            if (logger != null)
                logger.log(BasicLevel.ERROR, "problem to load class " + mopFcn);
            throw new PException(e, "problem to load class " + mopFcn);
        }

        // The target holder is shared by all classes

        TargetHolder targetHolder = null;
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "Output directory="
                    + compilerParameter.getOutputDirectory());
        try {
            targetHolder =
                    new JavaFileHolder(compilerParameter.getOutputDirectory());
        } catch (Exception e) {
            if (logger != null)
                logger.log(BasicLevel.ERROR, "Invalid output directory"
                        + compilerParameter.getOutputDirectory());
            throw new PException(e, "Invalid output directory"
                        + compilerParameter.getOutputDirectory());
        }
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                    "The mapper name is " + compilerParameter.getMapperName());
            logger.log(BasicLevel.DEBUG,
                    "The mapping name is " + compilerParameter.getMappingName());
        }

        // cney : support for invocation with meta-info instead of parser
        Iterator classes = null;
        if (mo != null) {
            classes = mo.iterator();
        }
        else if (genDepCollection!=null) {
            classes = genDepCollection.iterator();
        }
        else {
            classes = manager.getAllClassIterator();
        }
        while (classes.hasNext()) {
            //Class c = (Class) classes.next();
            Object o = classes.next();
            if (logger != null)
                if (o instanceof CompositeName)
                    logger.log(BasicLevel.INFO, "Generation for the " + ((CompositeName) o).getName() + " composite name.");
                else if (o instanceof Class)
                    logger.log(BasicLevel.INFO, "Generation for the " + ((Class) o).getName() + " class.");
                else {
                    logger.log(BasicLevel.ERROR, "Meta Object is not a class or a composite name : "+o);
                    throw new PException(
                        "Meta Object is not a class or a composite name : "+o);
                }

            // Instanciate a new Generator and a new MOPFactory in order to

            // have a better isolation between generation steps

            MOPFactory mopF = null;
            try {
                mopF = (MOPFactory) mopFactClass.newInstance();
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, e);
                System.exit(-1);
            }
            JormGeneratorImpl jormGen = new JormGeneratorImpl(mopF);
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG,
                        "JormGenerator and MOPFactory instancied");

            if (loggerFactory != null) {
                jormGen.setLogger(loggerFactory.getLogger("org.objectweb.jorm.generator"));
            }
            else {
                jormGen.setLogger(logger);
            }
            jormGen.generate(((MetaObject) o), targetHolder, compilerParameter);
        }
        return true;
    }

    /**
     * Compiles java files generated by the generator system.
     */
    public void compileFiles() {
        // TODO: need the list of classes to compile...
        throw new Error("NOT IMPLEMENTED!"); // added by Christophe Ney for safety (hopefully, did I miss something?)
    }

    /**
     * Returns the Manager which is filled in by the parser.
     *
     * @return  an object which contains all the meta information
     */
    public Manager getMetaInformation() {
        return manager;
    }

    /**
     * Returns the current class loader used to load class in the current
     * jorm process.
     *
     * @return  an existing classloader
     */
    public BasicClassLoader getClassLoader() {
        return classLoader;
    }

    /**
     * returns the exception which is reached during the parsing of a file
     *
     * @return  an Exception object
     */
    public Exception getParserException() {
        return parser.getParserException();
    }

    /**
     * allows to know the list of error the parser encounters.
     *
     * @return  a ParserErrorHandler object which countains all errors from
     *          the xml parser
     */
    public ParserErrorHandler getParserErrorHandler() {
        return parser.getParserErrorHandler();
    }


    ///////////////////////////////////////////////////////////////////
    //                      from Loggable interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Defines a logger object.
     *
     * @param logger	the logger object
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
        if ((compilerParameter.isVerbose()) && (logger != null)) {
            logger.setIntLevel(BasicLevel.DEBUG);
        }
    }

    /**

     * Defines the logger factory to obtain new logger.
     *

     * @param loggerFactory	the LoggerFactory object to obtain a logger object

     */
    public void setLoggerFactory(LoggerFactory loggerFactory) {
        this.loggerFactory = loggerFactory;
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }
}

