/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.generator.lib;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.context.Context;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.generator.api.MOP;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Schema;
import org.objectweb.util.io.api.TargetHolder;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.File;
import java.io.FileWriter;

/**
 * This class is a generator of XXXBinding objects. It is generic either the
 * mapper type. This generator use the velocity tools. The used template is
 * Binding.vm . Associated to this template this
 * generator builds a velocity context which contains the following
 * information:<br/>
 * <table border=1>
 * 	<tr><td>Key</td><td>Value</td></tr>
 * 	<tr><td>"class"</td><td>The reference to the Class meta object</td></tr>
 * 	<tr>
 * 		<td>"tools"</td>
 * 		<td>The reference to the <a href="CommonHelper.html">CommonHelper<a/></td>
 * </tr>
 * 	<tr>
 * 		<td>"header"</td>
 * 		<td>The file name of the common template which contains the header of
 * 			the generate files.</td>
 * </tr>
 * 	<tr>
 * 		<td>"bindingTools"</td>
 * 		<td>The reference to current instance of BindingGenerator.</td>
 * </tr>
 * 	<tr>
 * 		<td>"cparam"</td>
 * 		<td>The reference to a CompilerParameter instance. This instance
 * 		describes the user compilation parameter.</td>
 * </tr>
 * </table>
 *
 * As a XXXBinding class is specific to a mapper, the Binding.vm template uses
 * externals macros which must be defined as an external template library in
 * the velocity tools.
 */
public class BindingGenerator extends ObjectGenerator {

	/**
	 * The mapper specialization
	 */
	protected MOP bmop;

	/**
	 * This constrcutor permits to specify which BindingMOP must be used
	 */
	public BindingGenerator(MOP _bmop) throws Exception {
		super();
		bmop = _bmop;
	}

	/**
	 * This sette method permits to specify which BindingMOP must be used
	 */
	public void setBindingMOP(MOP _bmop) {
		bmop = _bmop;
	}

	// IMPLEMENTATION OF THE Generator INTERFACE //

	/**
	 * This method generates a XPBinding file corresponding to the pod
	 * parameter in the directory parameter
	 * @param pod	the meta object interface
	 * @param holder	The target holder which allows to create files
	 * @param cp	This parameter permits to reach the compilation parameters
	 */
	public void generate(MetaObject pod, TargetHolder holder,
						 CompilerParameter cp) throws Exception {
		if (bmop == null)
			throw new Exception("A BindingMOP is necessary to generate a PBinding");

		// The MetaObject parameter must be a Class.
		Class co = (Class) pod;

		// File name
		String fileName = cp.getMapperName() + File.separatorChar
			+ co.getName() + "Binding";
		String packName = ((Schema) co.getParent()).getName();
		if (packName != null && packName.length() > 0) {
			fileName = packName + File.separatorChar + fileName;
			fileName = fileName.replace('.', File.separatorChar);
		}
		logger.log(BasicLevel.DEBUG, "Generate the " + fileName + " class");

		// Fetch a block with the fileName
		FileWriter fw = holder.getFileWriter(fileName + ".java");
		VelocityEngine ve = initVelocity();

		// Signal to Velocity to use a list template which contain macros
		String[] libs = bmop.getTemplateLibraries();
		String strlibs = "";
		for (int i = 0; i < libs.length; i++)
			strlibs += (i == 0 ? "" : ",") + libs[i];
		if (strlibs.length() > 0)
			ve.setProperty(Velocity.VM_LIBRARY, strlibs);
		ve.init();

		Context ctx = new VelocityContext();
		ctx.put("class", co);
		ctx.put("cparam", cp);
		CommonHelper ch = new CommonHelper();
		ch.setLogger(logger);
		ctx.put("tools", ch);
		ctx.put("header", TEMPLATE_DIR + "/lib/Header.vm");
		ctx.put("bindingTools", this);
		bmop.initContext(ctx);
		ve.getTemplate(TEMPLATE_DIR + "/lib/Binding.vm").merge(ctx, fw);
		fw.flush();
		fw.close();
	}
}
