/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.generator.lib;

import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.generator.api.GeneratorFactory;
import org.objectweb.jorm.comp.generator.api.JormGenerator;
import org.objectweb.jorm.comp.generator.api.MOP;
import org.objectweb.jorm.comp.generator.api.MOPFactory;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.io.api.TargetHolder;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.jorm.metainfo.api.Class;
import java.util.Vector;

/**
 * This class start the generations of the jorm objects and user object. The
 * jorm objects are the mapping, the binding and the accessor. It is needed to
 * specify a MOPFactory before start the generation.The generate method
 * starts the generation. To generate user objects it is necessary to specify
 * their GeneratorFactory by the addGeneratorFactory method.
 */
public class JormGeneratorImpl implements JormGenerator, Loggable {

    /**
     * the list of optional GeneratorFactory
     */
    private Vector gfs = null;

    private Logger logger = null;

    private LoggerFactory loggerFactory = null;

    /**
     * The mopFactory which specialize jorm for a given support
     */
    private MOPFactory mopFactory = null;

    /**
     * This constructor allows to specify which MOPFactory must be used in this
     * JormGenerator instance.
     */
    public JormGeneratorImpl(MOPFactory mopf) {
        mopFactory = mopf;
    }

    /**
     * defines the logger factory to obtain new logger.
     * @param loggerFactory	the LoggerFactory object to obtain a logger object
     */
    public void setLogger(Logger _logger) {
        logger = _logger;
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }

    public void setLoggerFactory(LoggerFactory lf) {
        loggerFactory = lf;
    }
    // IMPLEMENTATION OF THE JormGenerator INTERFACE //
    //-----------------------------------------------//

    /**
     * This method permits to add GeneratorFactory instance. This is the way to
     * extend the JormGenerator to generate supplementary files.
     */
    public void addGeneratorFactory(GeneratorFactory gf) {
        if (gfs == null)
            gfs = new Vector();
        gfs.add(gf);
    }

    /**
     * This setter method permit to specify which MOPFactory must be used in
     * this JormGenerator instance.
     */
    public void setMOPFactory(MOPFactory mopf) {
        mopFactory = mopf;
    }


    // IMPLEMENTATION OF THE Generator INTERFACE //
    //-------------------------------------------//

    /**
     * This method generates the jorm objects by delegation to the
     * AccessorGenerator, the BindingGenerator and the MappingGenerator. Its
     * second task is to launch the user objects generator.
     *
     * @param pod	This meta object must be a Class
     * @param th	The target holder which allows to create files
     * @param cp	The compiler parameter
     * @throw Exception if an error occurs during the generation.
     */
    public void generate(MetaObject mo,
                         TargetHolder holder,
                         CompilerParameter cp) throws Exception {
        if (mopFactory == null) {
            throw new Exception("It's necessary to specify a MOPFactory");
        }

        if (mo instanceof Class) {
            // Generate jorm objects (binder, binding, accessor)
            ObjectGenerator gen = null;

            if (logger!=null) {
                logger.log(BasicLevel.DEBUG, "BinderInheritance = " + cp.getPClassMappingInheritance());
                logger.log(BasicLevel.DEBUG, "BindingInheritance = " + cp.getPBindingInheritance());
                logger.log(BasicLevel.DEBUG, "Mapper name = " + cp.getMapperName());
                logger.log(BasicLevel.DEBUG, "Mapping name = " + cp.getMappingName());
            }
            gen = new AccessorGenerator();
            if (loggerFactory!=null) {
                gen.setLogger(
                    loggerFactory.getLogger(logger.getName() + ".accessor"));
                gen.setLoggerFactory(loggerFactory);
            }
            else {
                gen.setLogger(logger);
            }
            gen.generate(mo, holder, cp);
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG, "XXXAccessor generated");


            MOP mMop = mopFactory.createMappingMOP(cp);
            ((Loggable) mMop).setLogger(logger);
            gen = new MappingGenerator(mMop);
            if (loggerFactory!=null) {
                gen.setLogger(
                    loggerFactory.getLogger(logger.getName() + ".mapping"));
                gen.setLoggerFactory(loggerFactory);
            }
            else {
                gen.setLogger(logger);
            }
            gen.generate(mo, holder, cp);
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG, "XXXMapping generated");

            MOP bMop = mopFactory.createBindingMOP(cp);
            ((Loggable) bMop).setLogger(logger);
            gen = new BindingGenerator(bMop);
            if (loggerFactory!=null) {
                gen.setLogger(
                    loggerFactory.getLogger(logger.getName() + ".binding"));
                gen.setLoggerFactory(loggerFactory);
            }
            else {
                gen.setLogger(logger);
            }
            gen.generate(mo, holder, cp);
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG, "XXXBinding generated");


            // generate optional or user objects
            if (gfs != null) {
                for (int i = 0; i < gfs.size(); i++) {
                    GeneratorFactory gf = (GeneratorFactory) gfs.elementAt(i);
                    gf.createGenerator().generate(mo, holder, cp);
                }
            }
        } else { //This is a composite name

            // generate the specific png interface
            ObjectGenerator gen = null;
            gen = new CompositeNameGenerator(); // generate the specific png interface
            if (loggerFactory!=null) {
                gen.setLogger(
                    loggerFactory.getLogger(logger.getName() + ".compositename"));
                gen.setLoggerFactory(loggerFactory);
            }
            else {
                gen.setLogger(logger);
            }
            gen.generate(mo, holder, cp);
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG, "XXXPNG generated");

            // generate the pname

            gen = new PNameGenerator(); // generate the specific png interface
            if (loggerFactory!=null) {
                gen.setLogger(
                    loggerFactory.getLogger(logger.getName() + ".pname"));
                gen.setLoggerFactory(loggerFactory);
            }
            else {
                gen.setLogger(logger);
            }
            gen.generate(mo, holder, cp);
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG, "XXXPName generated");

            // generate the binder

            gen = new BinderGenerator(); // generate the specific png interface
            if (loggerFactory!=null) {
                gen.setLogger(
                    loggerFactory.getLogger(logger.getName() + ".binder"));
                gen.setLoggerFactory(loggerFactory);
            }
            else {
                gen.setLogger(logger);
            }
            gen.generate(mo, holder, cp);
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG, "XXXbinder generated");
        }
    }
}
