/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.generator.lib;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.context.Context;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.generator.api.MOP;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.GenClass;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Schema;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.io.api.Block;
import org.objectweb.util.io.api.TargetHolder;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.File;
import java.io.FileWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Iterator;

/**
 * This class is a generator of XPMapping objects. It is generic either the
 * mapper type. This generator use the velocity tools. The used template is
 * Mapping.vm . Associated to this template, this
 * generator builds a velocity context which contains the following
 * information:<br/>
 * <table border=1>
 * 	<tr><td>Key</td><td>Value</td></tr>
 * 	<tr><td>"class"</td><td>The reference to the Class meta object</td></tr>
 * 	<tr>
 * 		<td>"tools"</td>
 * 		<td>The reference to the <a href="CommonHelper.html">CommonHelper<a/></td>
 * </tr>
 * 	<tr>
 * 		<td>"header"</td>
 * 		<td>The file name of the common template which contains the header of
 * 			the generate files.</td>
 * </tr>
 * 	<tr>
 * 		<td>"mappingTools"</td>
 * 		<td>The reference to current instance of MappingGenerator.</td>
 * </tr>
 * 	<tr>
 * 		<td>"cparam"</td>
 * 		<td>The reference to a CompilerParameter instance. This instance
 * 		describes the user compilation parameter.</td>
 * </tr>
 * </table>
 *
 * As a XXXMapping class is specific to a mapper, the Mapping.vm template uses
 * externals macros which must be defined as an external template library in
 * the velocity tools.
 */
public class MappingGenerator extends ObjectGenerator {

	/**
	 * The mapper specialization
	 */
	private MOP bmop;

	/**
	 * This constrcutor permits to specify which MappingMOP must be used
	 */
	public MappingGenerator(MOP _bmop) throws Exception {
		super();
		bmop = _bmop;
	}

	/**
	 * This sette method permits to specify which MappingMOP must be used
	 */
	public void setMappingMOP(MOP _bmop) {
		bmop = _bmop;
	}

	// IMPLEMENTATION OF THE Generator INTERFACE //

	/**
	 * This method generates a XXXPMapping file corresponding to the pod
	 * parameter in the directory parameter
	 * @param pod	the meta object interface
	 * @param holder	The target holder which allows to create files
	 * @param cp	This parameter permits to reach the compilation parameters
	 */
	public void generate(MetaObject pod, TargetHolder holder,
						 CompilerParameter cp) throws Exception {

		if (bmop == null)
			throw new Exception("A MappingMOP is necessary to generate a PBinder");

		// The MetaObject parameter must be a Class.
		Class co = (Class) pod;

		// File name
		String fileName = cp.getMapperName() + File.separatorChar
			+ co.getName() + "Mapping";
		String packName = ((Schema) co.getParent()).getName();
		if (packName != null && packName.length() > 0) {
			fileName = packName + File.separatorChar + fileName;
			fileName = fileName.replace('.', File.separatorChar);
		}
		logger.log(BasicLevel.DEBUG, "Generate the " + fileName + " class");

		// Fetch a block with the fileName
		FileWriter fw = holder.getFileWriter(fileName + ".java");
		VelocityEngine ve = initVelocity();

		// Signal to Velocity to use a list template which contain macros
		String[] libs = bmop.getTemplateLibraries();
		String strlibs = "";
		for (int i = 0; i < libs.length; i++)
			strlibs += (i == 0 ? "" : ",") + libs[i];
		if (strlibs.length() > 0)
			ve.setProperty(Velocity.VM_LIBRARY, strlibs);
		ve.init();

		StringWriter sw = new StringWriter();
		Context ctx = new VelocityContext();
		ctx.put("class", co);
		ctx.put("cparam", cp);
		CommonHelper ch = new CommonHelper();
		ch.setLogger(logger);
		ctx.put("tools", ch);
		ctx.put("header", TEMPLATE_DIR + "/lib/Header.vm");
		ctx.put("mappingTools", this);
		bmop.initContext(ctx);
		ve.getTemplate(TEMPLATE_DIR + "/lib/Mapping.vm").merge(ctx, fw);
		fw.flush();
		fw.close();
	}

	// METHOD USED BY THE TEMPLATE //
	//-----------------------------//

	/**
	 * This method is a helper to reach the getInheritedClasses(Class, HashMap)
	 * method. Indeed in the velocity template is not possible to instanciate
	 * an object.
	 */
	public HashMap getInheritedClasses(Class c) {
		return getInheritedClasses(c, new HashMap());
	}

	/**
	 * This recursive method add in the HashMap @calculated all inherited class
	 * of the @current class parameter.
	 * @return an hashmap where keys are class names and their values are the
	 * direct inhertied class.
	 */
	private HashMap getInheritedClasses(Class current, HashMap calculated) {
		int parentNumber = current.getInheritedClassNumber();
		// Base case : No inherited classes
		if (parentNumber == 0)
			return calculated;

		String[] list = new String[parentNumber + 1];
		// Put the current class in first position into the list
		list[0] = current.getName();
		// Put the list into the structure
		calculated.put(current.getName(), list);
		//Full the list of inherited classes for the current class
		Iterator it = current.iterateInheritedClass();
		for (int i = 1; it.hasNext() && i < (parentNumber + 1); i++) {
			Class parent = (Class) it.next();
			list[i] = parent.getName();
			if (!calculated.containsKey(parent.getName()))
				getInheritedClasses(parent, calculated);
		}
		return calculated;
	}

	/**
	 * This generation helper method retrieves the string corresponding
	 * to the codding type (short value) associated to a PType.
	 */
	public String ptype2CTString(PType pt) throws PExceptionTyping {
		if (pt==null) {
			throw new PExceptionTyping("null parameter forbidden");
		}
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_CHAR:
			return "PNamingContext.CTCHAR";
		case PType.TYPECODE_BYTE:
			return "PNamingContext.CTBYTE";
		case PType.TYPECODE_SHORT:
			return "PNamingContext.CTSHORT";
		case PType.TYPECODE_INT:
			return "PNamingContext.CTINT";
		case PType.TYPECODE_LONG:
			return "PNamingContext.CTLONG";
		case PType.TYPECODE_STRING:
			return "PNamingContext.CTSTRING";
		default:
			throw new PExceptionTyping("The following " + pt.getJormName()
				+ " type has'nt been allowed to be used as a PName field");
		}

	}

	/**
	 * This generation helper method retrieves the string corresponding
	 * to the instance of the PType.
	 */
	public String ptype2String(PType pt) throws PExceptionTyping {
		if (pt==null) {
			throw new PExceptionTyping("null parameter forbidden");
		}
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_CHAR:
			return "PTypeSpace.CHAR";
		case PType.TYPECODE_BYTE:
			return "PTypeSpace.BYTE";
		case PType.TYPECODE_SHORT:
			return "PTypeSpace.SHORT";
		case PType.TYPECODE_INT:
			return "PTypeSpace.INT";
		case PType.TYPECODE_LONG:
			return "PTypeSpace.LONG";
		case PType.TYPECODE_STRING:
			return "PTypeSpace.STRING";
		default:
			throw new PExceptionTyping("The following " + pt.getJormName()
				+ " type has'nt been allowed to be used as a PName field");
		}

	}

	/**
	 * This method is a generation helper. It returns a string like the
	 * following regular expression: (".getGenClassMapping()")*
	 * where the number of repetition of the string is equals to the level
	 * parameter.
	 */
	public String getGenclass(int level) {
		String res = "";
		for (int i = 0; i < level; i++) {
			res += ".getGenClassMapping()";
		}
		return res;
	}

	/**
	 * This method calculates the number of indexes in a GenClass. This method
	 * should be moved into the meta object GenClass.
	 */
	public int getIndexNumber(GenClass gc) {
		Iterator it = gc.iterateIndexField();
		int res = 0;
		while (it.hasNext()) {
			res++;
			it.next();
		}
		return res;
	}

	/**
	 * This method writes the declaraction of an array. More exatcly the list
	 * of values is written. The values stored in the iterator are the elements
	 * of the array. Each iterator elements are also an array of String
	 */
	private void writeArrayDef(Block block, Iterator lists) throws Exception {
		while (lists.hasNext()) {
			String[] list = (String[]) lists.next();
			String stringList = "{";
			for (int i = 0; i < list.length; i++)
				stringList += (i != 0 ? ", " : "") + "\"" + list[i] + "\"";
			stringList += "}" + (lists.hasNext() ? "," : "");
			block.writeLine(stringList);
		}
	}

	/**
	 * It retrieves the fully qualified class name of a Class meta object.
	 */
	public String getFQName(Class c) {
		return ((Schema) c.getParent()).getName() + "." + c.getName();
	}
}
