/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.generator.rdb;

import org.apache.velocity.context.Context;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.generator.api.MOP;
import org.objectweb.jorm.comp.generator.lib.CommonHelper;
import org.objectweb.jorm.comp.generator.lib.ObjectGenerator;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.rdb.RDBClassMapping;
import org.objectweb.jorm.metainfo.api.rdb.RDBColumnSpec;
import org.objectweb.jorm.metainfo.api.rdb.RDBValueMapping;
import org.objectweb.jorm.runtime.mapper.rdb.adapter.api.RdbAdpater;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

/**
 * This class implements the BindingMOP interface in order to specialize the
 * generation of the PBinding for the relationnal database spport.
 *
 * @author	<A HREF="mailto:sebastien.chassande@inrialpes.fr><b>
 *			Sebastien Chassande-Barrioz
 *			</b></A>
 * TODO :
 * - le code gnr doit utiliser des constantes pour les noms des champs et
 * le nom des colonnes:
 * 		=> Maintenir dans le generateur une hashMap (colName, constanteName)
 * - Commenter
 */
public class RdbBindingMOP extends CommonHelper implements MOP {

	/**
	 * This field permits to access to the converstion type.
	 */
	protected RdbAdpater adapter = null;

	/**
	 * @_adapter is the RdbAdpater instance to specialize the type conversion
	 * either a data base
	 */
	public RdbBindingMOP(RdbAdpater _adapter) {
		super();
		adapter = _adapter;
	}

	// IMPLEMENTATION OF THE MOP INTERFACE //
	//-------------------------------------//

	public void initContext(Context ctx) {
        try {
            Class cl = (Class) ctx.get("class");
            CompilerParameter cp = (CompilerParameter) ctx.get("cparam");

            ctx.put("bindingMOPTools", this);
            ctx.put("rdbAdater", adapter);
            ArrayList colinfo = getColInfo(cl, cp);
            ctx.put("colinfo",colinfo);
            ctx.put("writeCreateRequest",getWriteCreateRequest(cl, colinfo, cp));
            ctx.put("writeRemoveRequest",getWriteRemoveRequest(cl, colinfo, cp));
            ctx.put("writeUpdateRequest",getWriteUpdateRequest(cl, colinfo, cp));
        } catch (Exception e) {
            System.out.println(e);
        }
    }

	public String[] getTemplateLibraries() {
		String[] libs = { ObjectGenerator.TEMPLATE_DIR + "/rdb/RdbBinding.vm"};
		return libs;
	}

	// ------ HELPER METHODS -------//
	//------------------------------//

	public String getSqlGet(String resultSet, PType type, int pos)
		throws Exception {
		switch (type.getTypeCode()) {
		case PType.TYPECODE_CHAR:
			return resultSet + "." + adapter.getSqlGetterName(type)
				+ "(" + pos + ").charAt(0)";
		case PType.TYPECODE_INT_OBJ:
			return "((Integer)" + resultSet + "." +
                adapter.getSqlGetterName(type) + "(" + pos + "))";
		case PType.TYPECODE_SERIALIZED:
			return "(Serializable) new ObjectInputStream(" + resultSet
			+ ".getBlob(" + pos + ").getBinaryStream()).readObject()";
		default:
			return resultSet + "." + adapter.getSqlGetterName(type)
				+ "(" + pos + ")";
		}
	}

	/**
	 * This method provides the name of the column where the fiels of the class passed as parameter are mapped.
     * @moc is the meta object representing the jorm class.
     * @cp  is the compoiler parameter
     * @return an array list that provides the columns name as string.
	 */
	public ArrayList getColumnListOfRead(Class moc, CompilerParameter cp)
		throws Exception {

		ArrayList cols = new ArrayList();
		Iterator it = moc.iterateAllField();
        PrimitiveElement pe;
		while (it.hasNext()) {
			Object o = (TypedElement) it.next();
			if (o instanceof Reference) {
				NameDef ndf = ((Reference) o).getNameDef(cp.getMappingName());
                if (ndf.isFieldName()) {
                    // this is a single name, this name is defined from a primitive field or an hidden field.
                    pe = (PrimitiveElement) moc.getTypedElement(ndf.getFieldName());
                    RDBColumnSpec col = getColumnName(pe, cp);
                    if (!(cols.contains(col.getName())))
                        cols.add(col.getName());
                } else {
                    // this is a composite name
                    NameRef nr = ndf.getNameRef();
                    // get the projection of this composite name onto the field of the class
                    Map proj = nr.getProjection();
                    // classfields represents the name of the field related to the name structure.
                    // classfields are either primitive field or hiddenfield of the class.
                    Iterator classfields = proj.values().iterator();
                    while (classfields.hasNext()) {
                        pe = (PrimitiveElement) moc.getTypedElement((String)(classfields.next()));
                        RDBColumnSpec col = getColumnName(pe, cp);
                        if (!(cols.contains(col.getName())))
                            cols.add(col.getName());
                    }
                }
			}
			else if (o instanceof PrimitiveElement) {
				RDBColumnSpec col = getColumnName((PrimitiveElement) o, cp);
                if (!(cols.contains(col.getName())))
                    cols.add(col.getName());
			}
		}
		return cols;
	}

	/**
	 * This method calculates the number of serialized fields in the class. Be
	 * carefull the pnames cannot be composed of serialized fields.
     * @c the meta object representing the jorm class.
	 */
	private int numberOfSerializedField(Class c)
		throws Exception {
		int res = 0;
		Iterator it = c.iterateAllField();
		while (it.hasNext()) {
			PType pt = ((TypedElement) it.next()).getType();
			if (pt.getTypeCode() == PType.TYPECODE_SERIALIZED)
				res++;
		}
		return res;
	}

    /**
     * It build an sql request used to read instance of the class co.
     * @co the meta object representing the class
     * @cols the list of columns name where the fiels of the class are mapped
     * @cp the compiler parameters
     * @return an iterator over all the possible request (generic png, specific png)
     */
	public String getReadRequest(Class co, ArrayList cols, CompilerParameter cp, int pngspe)
	//public String getReadRequest(Class co, ArrayList cols, CompilerParameter cp)
        throws Exception {

		String request = "SELECT ";
		// list of the column names
		for (int i = 0; i < (cols.size() - 1); i++) {
			request += (String) cols.get(i) + ", ";
		}
		// Do not add the comma at the end of the list !
		request += (String) cols.get(cols.size() - 1);
		// Table name
		RDBClassMapping cm = (RDBClassMapping)
			co.getClassMapping(cp.getMapperName(), cp.getMappingName());
		request += " FROM " + cm.getTableName() + " WHERE ";

		NameDef nd = co.getNameDef(cp.getMappingName());

		return getNameDefRequest(co, nd, "pname", cp, "AND", request, true, pngspe);
	}

	public String getExistRequest(Class moc, CompilerParameter cp, int pngspe)
		throws Exception {
		RDBClassMapping cm = (RDBClassMapping)
			moc.getClassMapping(cp.getMapperName(), cp.getMappingName());

		String request = "SELECT * FROM " + cm.getTableName() + " WHERE ";
		//Write the NameDef of the current class
		NameDef nd = moc.getNameDef(cp.getMappingName());
		return getNameDefRequest(moc, nd, "pname", cp, "AND", request, true, pngspe);
	}

	/**
	 * This method does two things:
	 * - First it writes an sql request by the encoding a PName which the
	 * variable name is specified by the parameter @pNameVarName. The result is
	 * returned by this method.
	 * - Second it writes inside the block specified in parameter the code
	 * needed by the request (variable definitions, ...)
	 *
	 * @moc is the current Class. This parameter is used in the case where the
	 * PName contains a reference to a field of the current class.
	 * @nd is the meta object which represents the PName that must be written
	 * @pNameVarName is the string which represent the name of the PName in the
	 * generated code.
	 * @cp is the compiler parameter of the current generation
	 * @block is the block where the code needed to execute the sql request
	 * is written
	 * @request is the sql request before the call
     * @pngspe  the request must be generated for specific pnamegetter.
	 * @return the sql resquest returned
	 *
	 * This implementation takes in account the difference between a PName
	 * composed by a single field or a PName composed by multiple fields.
	 *
	 * The result seems to the following examples:
	 *
	 * Example 1: PName coding in long
	 * "YYY = '"+xxx.encodeLong()+"'"
	 * where YYYY is the col name of the field of the PName
	 * and xxx the PName variable name
	 *
	 * Example 2: PName is last name and first name
	 * "LAST_NAME_COL = '"+pnameGetter.pngetStringField("lastNameField")+"'"
	 */
	private String getNameDefRequest(Class moc,
									 NameDef nd,
									 String pNameVarName,
									 CompilerParameter cp,
									 String separator,
									 String request,
									 boolean writeFieldName,
                                     int pngspe) throws Exception {

		// A PName can be encoded by several fields. This information is
		// contained in the NameDef


        boolean isMultipleField = nd.isNameRef();
        PrimitiveElement pe;

		if (!isMultipleField) {
			logger.log(BasicLevel.DEBUG,
				"The PName of this Object is composed of an unique field");
            pe = (PrimitiveElement) moc.getTypedElement(nd.getFieldName());
            RDBColumnSpec col = getColumnName(pe, cp);
            logger.log(BasicLevel.DEBUG, "The PName col = " + col.getName());
            request += col.getName() + " = '\" + ";
            request += pNameVarName + "."
					+ getPNameEncodeFunction(pe.getType()) + "()";
            request += " + \"'";
		} else {
            logger.log(BasicLevel.DEBUG,
				"The PName of this Object is composed of multiple field");
            NameRef nr = nd.getNameRef();
            // get the projection of this composite name onto the field of the class
            Map proj = nr.getProjection();
            // classfields represents the name of the field related to the name structure.
            // classfields are either primitive field or hiddenfield of the class.
            Iterator cnfields = proj.keySet().iterator();
            String cnfieldname ;
            String classfieldsname ;
            while (cnfields.hasNext()) {
                cnfieldname = (String)cnfields.next();
                classfieldsname = (String) proj.get(cnfieldname);
                pe = (PrimitiveElement) moc.getTypedElement(classfieldsname);
                RDBColumnSpec col = getColumnName(pe, cp);
                logger.log(BasicLevel.DEBUG, "The PName col = " + col.getName());
                if (pngspe==0) {
                    // request for generic png
                    request += col.getName() + " = '\" + ";
                    request += "((PNameGetter)"+pNameVarName + "Getter)."
                        + getPNameGetterGetFunction(pe.getType())
                        + "(pClassMapping." + pe.getName() + "_NAME, null)";
                    request += " + \"'";
                    if (cnfields.hasNext()){
                        request += separator + " ";
                    }
                }
                else {
                    //request for specific png
                    request += col.getName() + " = '\" + ";
                    request += "(("+nr.getName()+"PNG)"+pNameVarName + "Getter).pnGet"
                        + upperFL(cnfieldname)
                        + "(null)";
                    request += " + \"'";

                    if (cnfields.hasNext()){
                        request += separator + " ";
                    }
                }
            }
        }
		return request;
	}

    /**
     * This method allow to retrieve a primitive element from th meta object mo.
     * if this metaobject is a primitive element, it is returned itself.
     * if this meta object is a single namedef the primitive element composing the namedef is returned.
     * This method does not work with composite pname.
     * @mo the meta object embeding the primitive element
     * @c the class embeding the meta object mo
     * @return the primitive element
     */

	public PrimitiveElement getPrimitiveElement(Object mo,
												Class c,
												boolean canBeFieldName)
		throws PException {
		if (mo instanceof PrimitiveElement)
			return (PrimitiveElement) mo;
        else if (mo instanceof NameDef) {
			NameDef nd = ((NameDef) mo);
			if (nd.isFieldName())
                return (PrimitiveElement) (c.getTypedElement(nd.getFieldName()));
			else
				throw new PException("This is composite name !" + mo);
		} else
			throw new PException(
				"impossible to obtain a PrimitiveElement with " + mo);
	}

	public class ColInfo {
        public final static byte PRIM_FIELD = 1;
        public final static byte REF_SINGLE = 2;
        public final static byte REF_COMPOSITE = 3;
        public final static byte PNAME_SINGLE = 4;
        public final static byte PNAME_COMPOSITE = 5;


        public int kase = 0;
        public String[] accessor;
        public String fieldName;
        public String value;


        /**
         * case 1 (primitive field) : Coldesc contains a ColDesc object.
         * case 2 (ref single) : Coldesc contains a ColDesc object.
         * case 3 (ref composite) : Coldesc contains an arraylist of ColDesc object.
         * case 4 (pname single) : Coldesc contains a ColDesc object.
         * case 5 (pname composite) : Coldesc contains an arraylist of ColDesc object.
         */
        public Object coldesc;

        public ColInfo() {
            kase = 0;
            accessor = new String[2];
            accessor[0] = null;
            accessor[1] = null;
            fieldName = null;
            value = null;
            coldesc = null;

        }
/*
        public boolean equals(Object o) {

        }


        public int hashCode() {
            return (fieldName == null ? 0 : fieldName.length());
        }*/

        public int getKase() {
            return kase;
        }

        public String[] getAccessor() {
            return accessor;
        }

        public String getAccessor(int i) {
            return accessor[i];
        }

        public String getFieldName() {
            return fieldName;
        }

        public String getValue() {
            return value;
        }

        public Object getColDesc() {
            return coldesc;
        }

        public void toDebug(Logger l) {
            l.log(BasicLevel.DEBUG, "");
            l.log(BasicLevel.DEBUG, "kase=" + kase);
            l.log(BasicLevel.DEBUG, "fieldName=" + fieldName);
            l.log(BasicLevel.DEBUG, "value=" + value);
            l.log(BasicLevel.DEBUG, "accessor[0]=" + accessor[0]);
            l.log(BasicLevel.DEBUG, "accessor[1]=" + accessor[1]);
        }
    }

    public class ColDesc {

        public String colName;
        public PType type;
        public String[] PnameGetters;
        public boolean mustBePrepared = false;

        public ColDesc() {

            colName = null;
            type = null;
            PnameGetters = new String[2];
            PnameGetters[0] = null; // specific PNG
            PnameGetters[1] = null; // generic PNG
            mustBePrepared = false;

        }
/*
        public boolean equals(Object o) {
            return o != null
                    && (o instanceof ColDesc)
                    && (colName != null && colName.equals(((ColDesc) o).colName));
        }


        public int hashCode() {
            return (colName.length());
        }
*/


        public String getColName() {
            return colName;
        }

        public String[] getPnameGetter() {
            return PnameGetters;
        }

        public String getPnameGetter(int i) {
            return PnameGetters[i];
        }

        public PType getType() {
            return type;
        }

        public boolean getMustBePrepared() {
            return mustBePrepared;
        }
    }

    /**
     * It build an arraylist providing information over the column where the class is mapped.
     * These informations are customized for the write operation.
     * The array list contains a set of colinfo.
     * The structure of a colinfo instance is the following :
     *  <colname     case    specifique_accessor    generic_accessor value fieldname>
     *
     * case 1 : a primitive field
     *          colname     case    specific_accessor       generic_accessor    nil             fieldname
     *
     * case 2 : a ref single
     *          colname     case    specific_accessor       generic_accessor    encode_function fieldname
     *
     * case 3 : a ref composite
     *          (the field is a ref toward a jorm object identified with a composite pname. This pname
     *          is mapped over several columns)
     *
     *          nil         case    specific_accessor       generic_accessor    nil             fieldname
     *          colname1    case    pnamegetterspecific1    pnamegettergeneric1 nil             fieldname
     *          colname2    case    pnamegetterspecific2    pnamegettergeneric2 nil             fieldname
     *          ...
     *
     * case 4 : pname of the class (this is a single pname)
     *          colname     case    nill                    nil                 encode_function nil
     *
     * case 5 : pname of the class (this is a composite pname)
     *          colname1    case    pnamegetterspecific1    pnamegettergeneric1 nil             nil
     *          colname2    case    pnamegetterspecific2    pnamegettergeneric2 nil             nil
     *          ...
     */
    public ArrayList getColInfo(Class moc, CompilerParameter cp)
            throws Exception {
        logger.log(BasicLevel.DEBUG, "");
        ArrayList al = new ArrayList();
        ArrayList cilist = new ArrayList();


        // Add the pname class columns
        // get the col info of the namedef of the class
        NameDef nd = moc.getNameDef(cp.getMappingName());
        al = getNameDefColInfo(moc, nd, null, "pname", cp, true, al, cilist);

        // Add the columns of the reference field
        Iterator it = moc.iterateAllField();
        while (it.hasNext()) {
            TypedElement te = (TypedElement) it.next();
            if (te instanceof Reference) {
                logger.log(BasicLevel.DEBUG, "Field reference: " + te.getName());
                al = getNameDefColInfo(
                        moc,
                        ((Reference) te).getNameDef(cp.getMappingName()),
                        te.getName(),
                        te.getName() + "PN",
                        cp,
                        true,
                        al,
                        cilist
                );
            }
        }

        // Add the columns of the PrimitiveElement
        it = moc.iterateAllField();
        while (it.hasNext()) {
            TypedElement te = (TypedElement) it.next();
            if (te instanceof PrimitiveElement) {
                logger.log(BasicLevel.DEBUG, "Primitive field: " + te.getName());
                ColInfo ci = new ColInfo();
                ColDesc cd = new ColDesc();

                cd.colName = getColumnName((PrimitiveElement) te, cp).getName();
                cd.type = te.getType();
                cd.mustBePrepared =
                        (te.getType().getTypeCode() == PType.TYPECODE_SERIALIZED);


                ci.fieldName = te.getName();
                ci.kase = ColInfo.PRIM_FIELD;
                ci.accessor[0] = "paccessor.paGet" + upperFL(te.getName()) + "()";
                ci.accessor[1] = "paccessorGen."
                        + this.getPaGetMethod(te.getType())
                        + "(pClassMapping." + te.getName() + "_NAME)";

                ci.coldesc = cd;
                if (!cilist.contains(cd.colName)) {
                    logger.log(BasicLevel.DEBUG, "add ColInfo into the list");
                    al.add(ci);
                    cilist.add(cd.colName);
                } else {
                    logger.log(BasicLevel.DEBUG, "Do not add ColInfo into the list");
                }
            }
        }
        return al;
    }


    /**
     * Build a prepared request for a write create action. This method is used by the pbinding
     * generator to build the write request.
     * @param moc the meta class object
     * @param colslist an arraylist of structured colinfo.
     * @cp the compiler parameter
     */
    public String getWriteCreateRequest(Class moc, ArrayList colsinfolist, CompilerParameter cp)
            throws Exception {
        // INSERT INTO TClient (Id, GOI, GBI, LN) VALUES ("+ values +")

        RDBClassMapping cm = (RDBClassMapping)
                moc.getClassMapping(cp.getMapperName(), cp.getMappingName());
        String request = "INSERT INTO " + cm.getTableName() + " (";
        ArrayList colsinfo;
        String cols = "";
        String vals = "";
        boolean first = true;
        ColInfo ci;
        ColDesc cd;
        Iterator it = colsinfolist.iterator();
        Iterator cdit;
        // it contains colinfo objetc
        while (it.hasNext()) {
            ci = (ColInfo) (it.next());
            if (first) {
                if (ci.kase == ColInfo.PRIM_FIELD ||
                    ci.kase == ColInfo.PNAME_SINGLE ||
                    ci.kase == ColInfo.REF_SINGLE) {

                    cd = (ColDesc) ci.coldesc;
                    cols += cd.getColName();
                    vals += "?";
                    first = false;
                } else {
                    cdit = ((ArrayList)ci.coldesc).iterator();
                    while (cdit.hasNext()) {

                        cd = (ColDesc) (cdit.next());
                        if (first) {
                            cols += cd.getColName();
                            vals += "?";
                            first = false;
                        } else {
                            cols += ", " + cd.getColName();
                            vals += ", ?";
                        }
                    }
                }
            } else if (ci.kase == ColInfo.PRIM_FIELD ||
                    ci.kase == ColInfo.PNAME_SINGLE ||
                    ci.kase == ColInfo.REF_SINGLE) {

                    cd = (ColDesc) ci.coldesc;
                    cols += ", "+cd.getColName();
                    vals += ", ?";

                } else {
                    cdit = ((ArrayList)ci.coldesc).iterator();
                    while (cdit.hasNext()) {
                        cd = (ColDesc) (cdit.next());
                        cols += ", " + cd.getColName();
                        vals += ", ?";
                    }
                }
        }
        return request + cols + ") VALUES (" + vals + ")";
    }


    /**
     * Build a prepared request for a write remove action. This method is used by the pbinding
     * generator to build the write request.
     * @param moc the meta class object
     * @param colslist an arraylist of structured colinfo.
     * @cp the compiler parameter
     */
    public String getWriteRemoveRequest(Class moc, ArrayList colinfolist, CompilerParameter cp)
            throws Exception {
        // DELETE FROM ${tableName} WHERE ${pnameRequest0}

        String pnrq = "";
        boolean first = true;
        ColInfo ci;
        ColDesc cd;
        Iterator itcd;
        Iterator it = colinfolist.iterator();

        RDBClassMapping cm = (RDBClassMapping)
                moc.getClassMapping(cp.getMapperName(), cp.getMappingName());
        String request = "DELETE FROM " + cm.getTableName() + " WHERE ";

        while (it.hasNext()) {
            ci = (ColInfo) it.next();
            if (ci.kase == ColInfo.PNAME_SINGLE) {

                cd = (ColDesc) ci.coldesc;
                if (first) {
                    pnrq += cd.getColName()+" = ? ";
                    first = false;
                } else {
                    pnrq += "AND "+cd.getColName()+" = ? ";
                }
            } else if (ci.kase == ColInfo.PNAME_COMPOSITE) {
                itcd = ((ArrayList)  ci.coldesc).iterator();

                while (itcd.hasNext()) {
                   cd = (ColDesc) itcd.next();
                   if (first) {
                      pnrq += cd.getColName()+" = ? ";
                      first = false;
                   } else {
                      pnrq += "AND "+cd.getColName()+" = ? ";
                   }
                }
            }
        }
        return request + pnrq;
    }

    /**
     * Build a prepared request for a write update action. This method is used by the pbinding
     * generator to build the write request.
     * @param moc the meta class object
     * @param colslist an arraylist of structured colinfo.
     * @cp the compiler parameter
     */
    public String getWriteUpdateRequest(Class moc, ArrayList colinfolist, CompilerParameter cp)
           throws Exception {
        // "UPDATE EMPLOYEES SET SALARY = ? WHERE ID = ?");

        RDBClassMapping cm = (RDBClassMapping)
                moc.getClassMapping(cp.getMapperName(), cp.getMappingName());
        String request = "UPDATE " + cm.getTableName() + " SET ";

        String colsrq = "";
        String pnrq = "";
        boolean firstCol = true;
        boolean firstPn = true;
        ColDesc cd;
        ColInfo ci;
        Iterator itcd;
        Iterator it = colinfolist.iterator();

        while (it.hasNext()) {
            ci = (ColInfo) it.next();
            switch (ci.kase) {
                case ColInfo.PRIM_FIELD: {
                    cd = (ColDesc) ci.coldesc;
                    if (firstCol) {
                        colsrq += cd.getColName()+" = ? ";
                        firstCol = false;
                    } else {
                        colsrq += ", "+cd.getColName()+" = ? ";
                    }
                    break;
                } case ColInfo.PNAME_SINGLE :{
                    cd = (ColDesc) ci.coldesc;
                    if (firstPn) {
                        pnrq += cd.getColName()+" = ? ";
                        firstPn = false;
                    } else {
                        pnrq += "AND "+cd.getColName()+" = ? ";
                    }
                    break;
                } case ColInfo.REF_COMPOSITE : {
                    // in the case of a ref composite, we have a list of coldesc.
                    itcd = ((ArrayList)  ci.coldesc).iterator();
                    while (itcd.hasNext()) {
                        cd = (ColDesc) itcd.next();
                        if (firstCol) {
                            firstPn = false;
                            colsrq += cd.getColName()+" = ? ";
                        } else {
                            colsrq += ", "+cd.getColName()+" = ? ";
                        }
                    }
                    break;

                } case ColInfo.REF_SINGLE: {
                    cd = (ColDesc) ci.coldesc;
                    if (firstCol) {
                        colsrq += cd.getColName()+" = ? ";
                        firstCol = false;
                    } else {
                        colsrq += ", "+cd.getColName()+" = ? ";
                    }
                    break;
                } case ColInfo.PNAME_COMPOSITE : {
                    // in the case of a pn composite, we have a list of coldesc.
                    itcd = ((ArrayList)  ci.coldesc).iterator();
                    while (itcd.hasNext()) {
                        cd = (ColDesc) itcd.next();
                        if (firstPn) {
                            pnrq += cd.getColName()+" = ? ";
                            firstPn = false;
                        } else {
                            pnrq += "AND "+cd.getColName()+" = ? ";
                        }
                    }
                    break;
                }
            }
        }
        return request + colsrq + " WHERE " + pnrq;
    }

    /**
     * Compute the start index for the pname setter in the case of an update request.
     * This index is used to generate a setter for the prepared statement of the update request.
     * @param moc the meta class object
     * @param colinfolist an arraylist of structured colinfo.
     * @cp the compiler parameter
     */
    public int getPnameIndexForUpdateRequest(Class moc, ArrayList colinfolist, CompilerParameter cp){
        int colsidx = 0;
        ColInfo ci;
        Iterator itcd;

        Iterator it = colinfolist.iterator();
        while (it.hasNext()) {
            ci = (ColInfo) it.next();
            switch (ci.kase) {
                case ColInfo.PRIM_FIELD: {
                    colsidx++;
                    break;
                } case ColInfo.REF_COMPOSITE : {
                    // in the case of a ref composite, we have a list of coldesc.
                    itcd = ((ArrayList)  ci.coldesc).iterator();
                    while (itcd.hasNext()) {
                        itcd.next();
                        colsidx++;
                    }
                    break;
                } case ColInfo.REF_SINGLE: {
                    colsidx++;
                    break;
                }
                default : break;
            }
        }
        return ++colsidx;
    }

    /**
     * it build an array list of colinfo from the namedef passed as parameter.
     * This pname can be a single or a composite one.
     * @moc the meta object representing the jorm class
     * @nd the namedef to be analysed
     * @pnamevar the name of the pname variable declared in the binding.
     * @cp the compiler parameter
     * @colinfo the arraylist to fill
     */
    public ArrayList getNameDefColInfo(Class moc,
                                       NameDef nd,
                                       String fieldname,
                                       String pnamevar,
                                       CompilerParameter cp,
                                       boolean canBeFieldName,
                                       ArrayList res,
                                       ArrayList cilist) throws Exception {

        logger.log(BasicLevel.DEBUG, "");
        PrimitiveElement pe = null;
        boolean first = true;
        ColInfo ci;
        ColDesc cd;

        // for each field of the PName
        if (!nd.isNameRef()) {
            // the pname is not a composite
            // getTypedElement return the element, it can be a primitve elem, an hiddenfield
            ci = new ColInfo();
            cd = new ColDesc();
            pe = (PrimitiveElement) moc.getTypedElement(nd.getFieldName());
            // pe is a field or an hiddenfield.
            logger.log(BasicLevel.DEBUG, "Field reference to: " + pe.getName());
            if (fieldname != null) {
                // it is not the pn of the class, it is a pn related to a field ref
                // it is a single pn, a ref single
                logger.log(BasicLevel.DEBUG, "Single reference field, the field : "
                        + fieldname);

                ci.kase = ColInfo.REF_SINGLE;
                ci.fieldName = fieldname;
                ci.accessor[0] = "paccessor.paGet" + upperFL(fieldname) + "(conn)";
                ci.accessor[1] = "paccessorGen.paGetRefField(pClassMapping." + fieldname + "_NAME, conn)";
                ci.value = pnamevar + "." + getPNameEncodeFunction(pe.getType()) + "()";

                cd.colName = getColumnName((PrimitiveElement) pe, cp).getName();
                cd.type = pe.getType();
                cd.mustBePrepared = (pe.getType().getTypeCode() == PType.TYPECODE_SERIALIZED);
            } else {
                // it is the pn of the class, but it is a single pn.
                // we do not need the accessor but the encode function
                logger.log(BasicLevel.DEBUG, "Single PName.");

                ci.kase = ColInfo.PNAME_SINGLE;
                ci.value = pnamevar + "." + getPNameEncodeFunction(pe.getType()) + "()";

                cd.colName = getColumnName((PrimitiveElement) pe, cp).getName();
                cd.type = pe.getType();
                cd.mustBePrepared = (pe.getType().getTypeCode() == PType.TYPECODE_SERIALIZED);
            }
            ci.coldesc = cd;
            if (!(cilist.contains(cd.colName))) {
                logger.log(BasicLevel.DEBUG, "add ColInfo into the list");
                res.add(ci);
                cilist.add(cd.colName);
            }
        } else { //it is a name ref
            // the pname is a composite
            // the name ref contain the projection of the name over the field of the class
            // the pname can be the pname of the class or it can be a pname related to a field.

            ArrayList cdlisttmp;
            NameRef nr = nd.getNameRef();
            Map proj = nr.getProjection();
            Iterator namefields = proj.keySet().iterator();
            // namefields is the list of the field belonging to the name structure in the
            // composite name definition.
            ci = new ColInfo();
            cdlisttmp = new ArrayList(proj.size());
            while (namefields.hasNext()) {
                String currentkey = (String) namefields.next();
                pe = (PrimitiveElement) moc.getTypedElement((String) proj.get(currentkey));
                cd = new ColDesc();

                if (fieldname != null) { // ref compo
                    // create the fisrt col for in the case of a ref composite (a composite field ref)
                    if (first) {
                        ci.kase = ColInfo.REF_COMPOSITE;
                        ci.fieldName = fieldname;
                        ci.accessor[0] = "paccessor.paGet" + upperFL(fieldname) + "(conn)";
                        ci.accessor[1] = "paccessorGen.paGetRefField(pClassMapping." + fieldname + "_NAME, conn)";
                        ci.value = nr.getName();
                        first = false;
                    }

                    cd.colName = getColumnName((PrimitiveElement) pe, cp).getName();
                    cd.type = pe.getType();
                    cd.mustBePrepared = (pe.getType().getTypeCode() == PType.TYPECODE_SERIALIZED);
                    cd.PnameGetters[0] = "((" + ci.value + "PNG)" + fieldname + "PNG).pnGet" +
                            upperFL(getProjectionField(nd, pe.getName())) + "(null)";
                    cd.PnameGetters[1] = "((PNameGetter)" + fieldname + "PNG)." + getPNameGetterGetFunction(pe.getType())
                            + "(\"" + getProjectionField(nd, pe.getName()) + "\",null)";

                    if (!cilist.contains(cd.colName)) {
                        cdlisttmp.add(cd);
                        cilist.add(cd.colName);
                    }
                } else { // pname compo
                    if (first) {
                        ci.kase = ColInfo.PNAME_COMPOSITE;
                        ci.value = nr.getName();
                        first = false;
                    }
                    cd.colName = getColumnName((PrimitiveElement) pe, cp).getName();
                    cd.type = pe.getType();
                    cd.mustBePrepared = (pe.getType().getTypeCode() == PType.TYPECODE_SERIALIZED);

                    cd.PnameGetters[0] = "((" + ci.value + "PNG)pnamegetter).pnGet" +
                            upperFL(getProjectionField(nd, pe.getName())) + "(null)";
                    cd.PnameGetters[1] = "((PNameGetter)pnamegetter)." + getPNameGetterGetFunction(pe.getType())
                            + "(\"" + getProjectionField(nd, pe.getName()) + "\", null)";

                    if (!cilist.contains(cd.colName)) {
                        cdlisttmp.add(cd);
                        cilist.add(cd.colName);
                    }
                }
            }

            if (!cdlisttmp.isEmpty()) {
                ci.coldesc = cdlisttmp;
                logger.log(BasicLevel.DEBUG, "add ColInfo into the list");
                res.add(ci);
            }
        }
        return res;
    }

    /**
     * build a setter used to initialize a prepared statement.
     * @param ci the colinfo describing a list of col desc
     * @param cl a collumn description
     * @param index the index of the setter
     * @param specific
     *          - in the case of a primitive field, a single ref or a pn single :
     *              specific says if the setter function is build from a specific accessor or a generic one.
     *              (0 : specific acc / 1 : generic accessor)
     *          - in the case of a composite pname of a composite reference
     *              specific says if the setter function is build from a specific pnamegetter or a generic one.
     *        (0 : specific png / 1 : generic png)
     * @return a string such as setInt(2, fn) where 2 is the index and fn the function
     *         to retreive the value.
     */

    public String getPreparedSetter(ColInfo ci, ColDesc o, int index,
                                    int specific) throws PExceptionProtocol {
        String value = "";
        if (ci.kase == ColInfo.PRIM_FIELD) {
            value = ci.getAccessor(specific);
        }
        else if ((ci.kase == ColInfo.REF_SINGLE)
                || (ci.kase == ColInfo.PNAME_SINGLE)) {
            value = ci.getValue();
        }
        else if ((ci.kase == ColInfo.REF_COMPOSITE)
                || (ci.kase == ColInfo.PNAME_COMPOSITE)) {
            value = o.getPnameGetter(specific);
        }
        else {
            throw new PExceptionProtocol("Inner jorm error: Bad ColInfo case");
        }
        switch (o.type.getTypeCode()) {
        case PType.TYPECODE_CHAR:
            value = "String.valueOf(" + value + ")";
        case PType.TYPECODE_SERIALIZED:
            value = "new RdbBlob(" + value + ")";
        }
        return adapter.getSqlSetterName(o.type)
            + "(" + index + ", " + value + ")";
    }

	/**
	 * This method permits to fetch the column name of the PrimitiveElement
	 * specified as parameter. This method checks that there is only one column
	 * because this generator version does not support the mapping of a field
	 * on several column.
	 */
	public RDBColumnSpec getColumnName(PrimitiveElement pe, CompilerParameter cp)
		throws Exception {
		RDBValueMapping vm = (RDBValueMapping)
			pe.getValueMapping(cp.getMapperName(), cp.getMappingName());
		Iterator cols = vm.iterateRDBColumnSpec();
		if (!cols.hasNext()) {
			throw new PException("A columnSpec must be specified");
		}
		RDBColumnSpec col = (RDBColumnSpec) cols.next();
		if (cols.hasNext()) {
			throw new PException("This generator version is not able to manage the mapping of a PName field onto multiple column");
		}
		return col;
	}

}
