/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.comp.generator.rdb;

import org.apache.velocity.context.Context;
import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.generator.api.MOP;
import org.objectweb.jorm.comp.generator.lib.CommonHelper;
import org.objectweb.jorm.comp.generator.lib.ObjectGenerator;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.metainfo.api.rdb.RDBClassMapping;
import org.objectweb.jorm.metainfo.api.rdb.RDBColumnSpec;
import org.objectweb.jorm.metainfo.api.rdb.RDBValueMapping;
import org.objectweb.jorm.runtime.mapper.rdb.adapter.api.RdbAdpater;
import org.objectweb.util.monolog.api.BasicLevel;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;


public class RdbMappingMOP extends CommonHelper implements MOP {

    /**
     * This field permits to access to the converstion type.
     */
    protected RdbAdpater adapter = null;

    protected final static String[] IMPLEMENTS = {};

    public RdbMappingMOP(RdbAdpater _tc) {
        super();
        adapter = _tc;
    }

    // IMPLEMENTATION OF THE MOP INTERFACE //
    //-------------------------------------//

    public void initContext(Context ctx) {
        ctx.put("mappingMOPTools", this);
        ctx.put("rdbAdater", adapter);
    }

    public String[] getTemplateLibraries() {
        String[] libs = {ObjectGenerator.TEMPLATE_DIR + "/rdb/RdbMapping.vm"};
        return libs;
    }

    // ------ HELPER METHODS -------//
    //------------------------------//

    /**
     * This method permits to fetch the column name of the PrimitiveElement
     * specified as parameter. This method checks that there is only one column
     * because this generator version does not support the mapping of a field
     * on several column.
     */
    public RDBColumnSpec getColumnName(PrimitiveElement pe, CompilerParameter cp)
            throws Exception {
        RDBValueMapping vm = (RDBValueMapping)
                pe.getValueMapping(cp.getMapperName(), cp.getMappingName());
        if (vm == null)
            logger.log(BasicLevel.ERROR, "The primitive element "
                    + pe.getName() + "/" + pe.getType().getJormName()
                    + " has no ValueMapping ("
                    + cp.getMapperName() + "," + cp.getMappingName() + ")");
        Iterator cols = vm.iterateRDBColumnSpec();
        if (!cols.hasNext()) {
            throw new PException("A columnSpec must be specified");
        }
        RDBColumnSpec col = (RDBColumnSpec) cols.next();
        if (cols.hasNext()) {
            throw new PException("This generator version is not able to manage the mapping of a PName field onto multiple column");
        }
        return col;
    }

    public String getCreateRequest(Class moc, CompilerParameter cp)
            throws Exception {
        logger.log(BasicLevel.DEBUG, "begin");

        RDBClassMapping cm = (RDBClassMapping)
                moc.getClassMapping(cp.getMapperName(), cp.getMappingName());
        if (cm == null)
            logger.log(BasicLevel.ERROR, "The class "
                    + moc.getName() + " has no ClassMapping ("
                    + cp.getMapperName() + "," + cp.getMappingName() + ")");

        String tableName = cm.getTableName();
        logger.log(BasicLevel.DEBUG, "table Name = " + tableName);
        String request = "CREATE TABLE " + tableName + " (";

        ArrayList colNames = new ArrayList();
        PrimitiveElement sf;
        //---- for each scalar field of the PName ----//
        NameDef classnd = moc.getNameDef(cp.getMappingName());
        if (Debug.ON)
            logger.log(BasicLevel.DEBUG, "The scalar fields of the pname: ");

        if (classnd.isFieldName()) {
            // this is a single name, this name is defined from a primitive field or an hidden field.
            sf = (PrimitiveElement) moc.getTypedElement(classnd.getFieldName());

            if (Debug.ON)
                logger.log(BasicLevel.DEBUG, "scalar field(" + sf.getName()
                        + ", " + sf.getType().getJormName() + ")");
            RDBColumnSpec col = getColumnName(sf, cp);
            if (Debug.ON)
                logger.log(BasicLevel.DEBUG, "col Name: "
                        + (col == null?"col is null!":col.getName()));
            if (colNames.contains(col.getName())) {
                logger.log(BasicLevel.WARN,
                        "Column " + col.getName() + " already defined");
                //continue; ???
            } else {
                colNames.add(col.getName());
                String sqlType = col.getType();
                if (sqlType == null || sqlType.length() == 0) {
                    sqlType = adapter.getSqlType(sf.getType());
                }
                String colName = col.getName();
                if (col.isNotNull()) {
                    sqlType += " NOT NULL";
                }
                request += colName + " " + sqlType;
                if (Debug.ON)
                    logger.log(BasicLevel.DEBUG, colName + " " + sqlType);
            }
        } else {
            boolean first = true;
            // this is a composite name
            NameRef nr = classnd.getNameRef();
            // get the projection of this composite name onto the field of the class
            Map proj = nr.getProjection();
            // classfields represents the name of the field related to the name structure.
            // classfields are either primitive field or hiddenfield of the class.
            Iterator classfields = proj.values().iterator();
            while (classfields.hasNext()) {
                sf = (PrimitiveElement) moc.getTypedElement((String) (classfields.next()));
                if (Debug.ON)
                    logger.log(BasicLevel.DEBUG, "scalar field(" + sf.getName()
                            + ", " + sf.getType().getJormName() + ")");
                RDBColumnSpec col = getColumnName(sf, cp);
                if (Debug.ON)
                    logger.log(BasicLevel.DEBUG, "col Name: "
                            + (col == null?"col is null!":col.getName()));
                if (colNames.contains(col.getName())) {
                    logger.log(BasicLevel.WARN,
                            "Column " + col.getName() + " already defined");
                    continue;
                }
                colNames.add(col.getName());
                String sqlType = col.getType();
                if (sqlType == null || sqlType.length() == 0) {
                    sqlType = adapter.getSqlType(sf.getType());
                }
                String colName = col.getName();
                if (col.isNotNull()) {
                    sqlType += " NOT NULL";
                }
                if (first) {
                    first = false;

                } else {
                    request += ", ";
                }
                request += colName + " " + sqlType;
                if (Debug.ON)
                    logger.log(BasicLevel.DEBUG, colName + " " + sqlType);

            }
        }

        if (Debug.ON)
            logger.log(BasicLevel.DEBUG, "Each field of the class:");
        //---- for each field ----//
        Iterator it = moc.iterateAllField();

        boolean first = true;
        while (it.hasNext()) {
            TypedElement te = (TypedElement) it.next();
            logger.log(BasicLevel.DEBUG, "Treat the field "
                    + te.getType().getJormName() + " " + te.getName());
            if (te instanceof Reference) {
                // The field is a reference => use a NameDef
                NameDef fieldnd = ((Reference) te).getNameDef(cp.getMappingName());


                if (fieldnd.isFieldName()) {
                    // this is a single name, this name is defined from a primitive field or an hidden field.
                    sf = (PrimitiveElement) moc.getTypedElement(fieldnd.getFieldName());
                    RDBColumnSpec col = getColumnName(sf, cp);
                    if (colNames.contains(col.getName())) {
                        logger.log(BasicLevel.WARN,
                                "Column " + col.getName() + " already defined");
                        //continue;
                    } else {
                        colNames.add(col.getName());
                        String sqlType = col.getType();
                        if (sqlType == null || sqlType.length() == 0) {
                            sqlType = adapter.getSqlType(sf.getType());
                        }
                        String colName = col.getName();


                        request += ", " + colName + " " + sqlType;
                        if (Debug.ON)
                            logger.log(BasicLevel.DEBUG, colName + " " + sqlType);
                    }
                } else { // this is a composite name
                    // this is a composite name
                    NameRef nr = fieldnd.getNameRef();
                    // get the projection of this composite name onto the field of the class
                    Map proj = nr.getProjection();
                    // classfields represents the name of the field related to the name structure.
                    // classfields are either primitive field or hiddenfield of the class.
                    Iterator classfields = proj.values().iterator();
                    while (classfields.hasNext()) {
                        sf = (PrimitiveElement) moc.getTypedElement((String) (classfields.next()));
                        RDBColumnSpec col = getColumnName(sf, cp);
                        if (colNames.contains(col.getName())) {
                            logger.log(BasicLevel.WARN,
                                    "Column " + col.getName() + " already defined");
                            continue;
                        }
                        colNames.add(col.getName());
                        String sqlType = col.getType();
                        if (sqlType == null || sqlType.length() == 0) {
                            sqlType = adapter.getSqlType(sf.getType());
                        }
                        String colName = col.getName();

                        request += ", " + colName + " " + sqlType;
                        if (Debug.ON)
                            logger.log(BasicLevel.DEBUG, colName + " " + sqlType);
                    }
                }
            } else {
                RDBColumnSpec col = getColumnName((PrimitiveElement) te, cp);
                if (colNames.contains(col.getName())) {
                    logger.log(BasicLevel.WARN,
                            "Column " + col.getName() + " already defined");
                    continue;
                }
                colNames.add(col.getName());
                String sqlType = col.getType();
                if (sqlType == null || sqlType.length() == 0) {
                    sqlType = adapter.getSqlType(te.getType());
                }
                String colName = col.getName();
                request += ", " + colName + " " + sqlType;
                if (Debug.ON)
                    logger.log(BasicLevel.DEBUG, colName + " " + sqlType);
            }
        }

        //---- Primary key ----//
        it = ((RDBClassMapping) moc.getClassMapping(
                cp.getMapperName(), cp.getMappingName())
                ).iteratePrimaryKey();
        if (it.hasNext()) {
            request += ", PRIMARY KEY (";
            while (it.hasNext()) {
                RDBColumnSpec col = (RDBColumnSpec) it.next();
                request += ", " + col.getName();
            }
            request += ")";
        }

        logger.log(BasicLevel.DEBUG, "end");
        return request + ")";

    }


}
