/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */


package org.objectweb.jorm.facility.naming.gobid.binder;


import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.naming.api.PBinderDelegate;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.naming.cache.api.Cache;
import org.objectweb.jorm.naming.cache.lib.SimpleHashedCache;
import org.objectweb.jorm.facility.naming.gobid.generator.LGenerator;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import java.util.Iterator;

/**
 * The class <b>BinderLong</b> defines a basic binder used by JORM to manage persistent long name.
 * A binderlong hold a persistent long id. This id is used to identify the binder in a given namingcontext.
 * Furthermore it embed a reference to a LGenerator used to produce persitant long id when creating new Pname.
 * This binder works only with the BinderPNLong class.
 * @author	<A HREF="mailto:noel.depalma@rd.francetelecom.fr><b>
 *			Noel De Palma
 *			</b></A>
 */

public class LBinder implements PBinderDelegate, Loggable {

	/**
	 *   id of this binder. This is a persistent id. It identify the binder.
	 */
	long id;

	/**
	 * represent the null value for the name managed by this binder.
	 */
	private static long NULLID = -1;
	/**
	 * generator used by this binder to produce persistent long id.
	 * This field contain the mi of a longIdGenerator (which is also a pbinding).
	 * This reference is managed specifiquely by binderlong accessor so it is not visible in the BinderLong.pd file.
	 */
	protected LGenerator generator = null; // will be set by the binderlongfactrory through the PALongGen

	/**
	 * inner cache to store the association beetween paname and the related infoitem.
	 */
	Cache bindingcache = null;

	/**
	 * Pclassmapping which used this binderlong
	 */
	PClassMapping pcmbinderlong = null;

	/**
	 * This is the PType associated to the JORM class which is the type of
	 * the reference managed by this naming context.
	 */
	PType pt;

	/**
	 * The logger which collects traces of this class
	 */
	protected Logger logger = null;

	/**
	 * a constructeur that need to be initialized.
	 * a binderlong has to be initialized with the setBinderClassMapping method.
	 */
	public LBinder() {
		this.bindingcache = new SimpleHashedCache();
	}

	/**
	 * It initialize this binder with the pclassmapping which use this binder and it initialize internals structures.
	 * @param pcm the pclassmapping which use this binder
	 */
	public void setPClassMapping(PClassMapping pcm) {
		pcmbinderlong = pcm;
	}

	/**
	 * Retrieve the id of this binder
	 * @return a long representing the id of this binder.
	 */
	public long getId() {
		return this.id;
	}

	public void setId(long i) {
		id = i;
	}

	/**
	 * It retrieve the generator related to this LBinder.
	 * @return the generator.
	public Object getIdGen() {
		return this.generator;
	}

	/**
	 * It set the generator related to this LBinder.
	 * @param gen the generator.
	public void setIdGen(Object gen) {
		generator = (LGenerator) gen;
	}
	 */

	/**
	 * It test the equality of this binder with another passed as parameter.
	 * The binder are equal if they have the same id.
	 * @param obj the object to compare
	 */
	public boolean equals(Object obj) {
		return obj instanceof LBinder
			&& this.id == ((LBinder) obj).getId();
	}


	// IMPLEMENTATION OF THE PBinder INTERFACE //
	//-----------------------------------------//

	/**
	 * Implementation of the PBinder interface.
	 * @see org.objectweb.jorm.naming.api.PBinder
	 */
	public void bind(PName pn, PBinding pb) throws PExceptionNaming {
		if (pn == null)
			throw new PExceptionNaming("This pname is null");
		if (pb == null)
			throw new PExceptionNaming("This pbinding is null");
		if (!(pn instanceof LName))
			throw new PExceptionNaming("Can only bind PNString");
		if (!((pn.getPNamingContext()).equals(pcmbinderlong)))
			throw new PExceptionNaming("This pname is not valid in this binder");
		if (!pb.getPClassMapping().equals(pcmbinderlong))
			throw new PExceptionNaming("This pbinding is not valid in this binder");
		if (pn.isNull())
			throw new PExceptionNaming("BinderLong.bind : This PName represent the null value");
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "");

		bindingcache.register(pn, pb);
	}

	/**
	 * Implementation of the PBinder interface.
	 * @see org.objectweb.jorm.naming.api.PBinder
	 */
	public Iterator iteratePBinding() throws UnsupportedOperationException {
		return bindingcache.getIterator();
	}

	/**
	 * Implementation of the PBinder interface.
	 * @see org.objectweb.jorm.naming.api.PBinder
	 */
	public void unbind(PBinding pb) throws PExceptionNaming {
		if (pb != null) {
			if (!pb.getPClassMapping().equals(pcmbinderlong))
				throw new PExceptionNaming("This pbinding is not valid in this binder (wrong PClassMapping/pb)");
			PName pn = pb.getPName();
			if (pn != null) {
				if (!(pn instanceof LName))
					throw new PExceptionNaming("This PBinding cannot be managed by this PBinder (wrong PName)");
				if (!((pn.getPNamingContext()).equals(pcmbinderlong)))
					throw new PExceptionNaming("This pname is not valid in this binder (wrong PClassMapping/pn)");

				bindingcache.unregister(pn);
			}
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "");
		}

	}

	/**
	 * Implementation of the PBinder interface.
	 * @see org.objectweb.jorm.naming.api.PBinder
	 */
	public PClassMapping getBinderClassMapping() {
		return this.pcmbinderlong;
	}


	// IMPLEMENTATION OF THE PNamingContext INTERFACE //
	//------------------------------------------------//

	/**
	 * Implementation of the PNamingContext interface.
	 * The coding supported by this binder is CTLONG and CTSTRING.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public boolean codingSupported(short codingtype) {
		return (codingtype == PNamingContext.CTLONG
			|| codingtype == PNamingContext.CTSTRING);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * This operation is Unsupported.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeByte(byte en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Byte identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * This operation is Unsupported.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeChar(char en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Char identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * This operation is Unsupported.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeInt(int en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An Int identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * This operation is Unsupported.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeIntObj(Integer en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An Integer identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException {
		if (en == NULLID) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "PName Null");
			return getNull();
		}
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "PName = " + en);
		return new LName(en, pcmbinderlong);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeShort(short en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Short identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeString(String en) throws PExceptionNaming {
		if (Long.toString(NULLID).equals(en)) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "PName Null");
			return getNull();
		}
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "PName = " + en);
		return new LName(Long.parseLong(en), pcmbinderlong);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Object encodeAbstract(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An Abstract identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public byte encodeByte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Byte identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public char encodeChar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Char identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public int encodeInt(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An Int identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Integer encodeIntObj(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An Int identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (!((pn.getPNamingContext()).equals(pcmbinderlong)))
			throw new PExceptionNaming("This pname is not valid in this binder");
		if (pn.isNull()) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "PName = null");
			return NULLID;
		}
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "PName = " + pn);
		return ((LName) pn).getId();

	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public short encodeShort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Short identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public String encodeString(PName pn) throws PExceptionNaming {
		if (!((pn.getPNamingContext()).equals(pcmbinderlong)))
			throw new PExceptionNaming("This pname is not valid in this binder");
		if (pn.isNull()) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "PName = null");
			return Long.toString(NULLID);
		}
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "PName = " + pn);
		return String.valueOf(((LName) pn).getId());

	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeAbstract(Object en, Object ctx) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An abstract identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public void setPType(PType pt) {
		this.pt = pt;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public boolean supportDynamicComposite() {
		return false;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public boolean supportCompositeField(String fn, PType ft) {
		return false;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public boolean supportStaticComposite() {
		return false;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName getNull() {
		return new LName(NULLID, pcmbinderlong);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName export(Object conn, Object infoitem)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		if (infoitem == null)
			return getNull();
		if (!(infoitem instanceof PBinding))
			throw new PExceptionNaming(
				"The item cannot be exported by this binder");
		if (!((PBinding) infoitem).getPClassMapping().equals(pcmbinderlong))
			throw new PExceptionNaming("This pbinding is not valid in this binder");
		// check if it is already exported
		byte status = ((PBinding) (infoitem)).getStatus();
		if ((status == PBinding.LIFECYCLE_NOACC) ||
			(status == PBinding.LIFECYCLE_NEWTOWRITE) ||
			(status == PBinding.LIFECYCLE_ACTIVEFORIO)) {

			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "Already exported");

			// This pname is valid in this binder.
			PName pn = ((PBinding) (infoitem)).getPName();
			if (!((pn.getPNamingContext()).equals(pcmbinderlong)))
				throw new PExceptionNaming("This pname is not valid in this binder");
			return pn;
		}
		else {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "Get  new Id and create LName");

			// generate a new persistant id for the pname
			long newid;

			try {
				newid = generator.genId(conn); // sync here
			}
			catch (PExceptionNaming pen) {
				throw pen;
			}
			catch (PExceptionIO peio) {
				throw peio;
			}
			catch (PExceptionProtocol pep) {
				throw pep;
			}
			catch (PException pe) {
				throw new PExceptionIO(pe);
			}

			// build the new pname
			LName pnres = new LName(newid, pcmbinderlong);

			// register the pname in the cache with the related pbinding
			bindingcache.register(pnres, infoitem);
			return pnres;
		}
	}

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context.
	 * @param hints		Any information relevant for name creation.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 *							the data store.
	 */
	public PName export(Object conn, Object infoitem, Object hints)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		throw new PExceptionNaming("Not supported yet.");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Object resolve(Object conn, PName pn) throws PExceptionNaming, PExceptionIO {
		if (pn == null)
			throw new PExceptionNaming("BinderLong.Resolve : This pname is null");
		if (!((pn.getPNamingContext()).equals(pcmbinderlong)))
			throw new PExceptionNaming("This pname is not valid in this binder");
		if (pn.isNull()) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "PName is null then return null");
			return null;
		}
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "lookup in the cache");
		return (PBinding) bindingcache.lookup(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public void unexport(Object conn, Object obj)
		throws PExceptionNaming, PExceptionIO {
		if (obj == null)
			throw new PExceptionNaming("A null object cannot be unexported");
		PName pn = null;
		if (obj instanceof LName)
			pn = (PName) obj;
		else if (obj instanceof PBinding)
			pn = ((PBinding) obj).getPName();
		else
			throw new PExceptionNaming(
				"This object type cannot be managed by this binder:" + obj);
		if (!pn.getPNamingContext().equals(pcmbinderlong))
			throw new PExceptionNaming(
				"This pname is not valid in this binder");
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "unregister in the cache");
		bindingcache.unregister(pn);
	}


	// IMPLEMENTATION OF THE Loggable INTERFACE //
	//------------------------------------------//

	/**
	 * It assigns a logger to an component that wants to log things.
	 * @param Logger	The logger object.
	 */
	public void setLogger(Logger _logger) {
		logger = _logger;
	}

	/**
	 * It assigns a logger factory that allows the creation of new loggers.
	 * @param loggerfactory	The LoggerFactory object to obtain new loggers.
	 */
	public void setLoggerFactory(LoggerFactory loggerfactory) {
		logger = loggerfactory.getLogger(this.getClass().getName());
	}
	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return null;
	}
}

