/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.facility.naming.gobid.binder;

import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.facility.meminst.lifecycle.lazyload.LazyLoad;
import org.objectweb.jorm.facility.meminst.manager.basic.api.Manager;
import org.objectweb.jorm.facility.naming.gobid.generator.LGenerator;
import org.objectweb.jorm.facility.naming.gobid.generator.LGeneratorManager;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.util.monolog.api.BasicLevel;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionFactory;

/**
 * The class <b>PALBinder</b> defines an accessor to access the mi of a LBinder.
 * This accessor represent also the pbinding of the LBinder.
 * The pname of this object is a wellknown name encoded as a string. This mean that the
 * LBindermapping which manage the persistence of a LBinder uses a binderstring.
 * The generated LBinderPBinding must inherit from this class.
 * This class is abstract because it must be extended by its generated PBinding.
 *
 * @author nono de palma.
 */
public abstract class PALBinder
	extends LBinder
	implements PBinding, PBinder, LBinderAccessor, LazyLoad {

	private static long NULLID = -1;

	private PName generatorName = null;

	/**
	 *  The loaded flag is used to manage the loading of this object from the DS if necessary
	 */
	private boolean loaded = false;

	/**
	 * This connection factory is used to fetch connection on the data store.
	 * The connections are used to store or to load the data of this object from
	 * the support.
	 */
	protected ConnectionFactory connectionFactory = null;

	public void setConnectionFactory (ConnectionFactory cf) {
		connectionFactory = cf;
	}

	/**
	 * Retrieve the id of this binder
	 * @return a long representing the id of this binder.
	 */
	public long getId(Object conn) throws PException{
		load(conn);
		return super.getId();
	}

	/**
	 * Retrieve the id of this binder
	 * @return a long representing the id of this binder.
	 */
	public long getId() {
		try {
			load(null);
			return super.getId();
		}
		catch (PException pe) {
			if (logger != null)
				logger.log(BasicLevel.ERROR,"Impossible to load the binder",pe);
			return -1;
		}
	}

	/**
	 * Write the mi through this pbinding to the DS.
	 */
	public void store(Object conn) throws PException {
        Object c = conn;
		loaded = true;
		try {
			if (conn == null)
					c = connectionFactory.getConnection();
			write((conn==null?c:conn));
		}
		catch (ResourceException e) {
			throw new PExceptionIO(e,"Cannot reload this mi");
		}

		finally {
			if (c != null && conn == null)
				try {
					((Connection)c).close();
				}
				catch (Exception e) {
					logger.log(BasicLevel.FATAL, "Cannot close connection!!!");
				}
		}
	}

	public boolean isLoaded() {
		return loaded;
	}

	/**
	 * Load the mi from the DS if necessary.
	 */
	public void load(Object conn) throws PException {
        Object c = null;

		if (!loaded) {
			if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "Loading..." + this);

			try {
				if (conn == null) {
						c = connectionFactory.getConnection();
				}
				PMapper mapper = this.getPClassMapping().getPMapper();
				read((conn==null?c:conn));
				if (generatorName==null)
					throw new PExceptionProtocol("The LBinder does not have generator");
				// generator_jormname = name of the jorm class managing all
				// the binder
				PClassMapping pcm = (PClassMapping)
					mapper.lookup(LGeneratorManager.GENERATOR_JORM_NAME);
				Manager manager = (Manager) pcm.getMIManager();
				generator = (LGenerator)
					manager.getPBinding(c, generatorName).getPAccessor();
				loaded = true;
			}
			catch (ResourceException e) {
				throw new PExceptionIO(e,"Cannot load the persistent state of a naming context.");
			}
			finally {
				if (c!=null && conn == null)
					try {
						((Connection)c).close();
					}
					catch (Exception e) {
						logger.log(BasicLevel.FATAL, "Cannot close connection!!!");
					}
			}
		}
		else if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "Already Load");
	}



	// IMPLEMENTATION OF THE PNamingContext INTERFACE //
	//------------------------------------------------//

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeByte(byte en) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeByte(en);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeLong(en);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeChar(char en) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeChar(en);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeInt(int en) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeInt(en);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeIntObj(Integer en) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeIntObj(en);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeShort(short en) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeShort(en);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeString(String en) throws PExceptionNaming {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeString(en);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Object encodeAbstract(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeAbstract(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public byte encodeByte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeByte(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public char encodeChar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeChar(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public int encodeInt(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeInt(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Integer encodeIntObj(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeIntObj(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeLong(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public short encodeShort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeShort(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public String encodeString(PName pn) throws PExceptionNaming {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeString(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeAbstract(Object en, Object ctx) throws PExceptionNaming, UnsupportedOperationException {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeAbstract(en, ctx);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName export(Object conn, Object infoitem)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.export(conn, infoitem);
	}

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context.
	 * @param hints		Any information relevant for name creation.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 *							the data store.
	 */
	public PName export(Object conn, Object infoitem, Object hints)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		throw new PExceptionNaming("Not supported yet.");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Object resolve(Object conn, PName pn) throws PExceptionNaming, PExceptionIO {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.resolve(conn, pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public void unexport(Object conn, PName pn) throws PExceptionNaming, PExceptionIO {
		try {
			load(null);
		}
		catch(PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		super.unexport(conn, pn);
	}


	// IMPLEMENTATION OF THE PAccessor INTERFACE //
	//-------------------------------------------//

	/**
	 * Implementation of the paccessor interface.
	 * @see org.objectweb.jorm.api.PAccessor
	 */
	public Object getMemoryInstance() {
		return this;
	}


	// IMPLEMENTATION OF THE GENERATED LBinderAccessor INTERFACE //
	//--------------------------------------------------------------//

	/**
	 * It retrieve the id of this binder.
	 * @return the long value identifying this binder.
	 */
	public long paGetId() {
		return id;
	}

	/**
	 * It set the id of this binder.
	 * @param id the long value identifying this binder.
	 */
	public void paSetId(long id) {
		this.id = id;
	}

	public Object paGetGenerator(Object connection) throws PException {
		return ((PBinding) generator).getPName();
	}

	public void paSetGenerator(Object val, Object connection)
		throws PException {
		generatorName = (PName) val;
		generator = (LGenerator) ((PName) generatorName).resolve(connection);
		if (Debug.ON && logger!=null)
			logger.log(BasicLevel.DEBUG, "generator=" + generator);
	}
}