/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.facility.naming.gobid.generator;


import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.facility.naming.gobid.generator.LGenerator;
import org.objectweb.jorm.facility.naming.gobid.generator.LGeneratorImpl;
import org.objectweb.jorm.facility.meminst.lifecycle.lazyload.LazyLoad;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.monolog.api.BasicLevel;
import javax.resource.cci.ConnectionFactory;
import javax.resource.cci.Connection;
import javax.resource.ResourceException;

/**
 * Define an accessor to access to a long id generator.This class act as an interceptor to handle LongGen reloading.
 * PName of LongGen objects are composed by the pname of the binderlong that use the longGen instance + ".LGenerator"
 * This PName is managed by a BinderString. The PNStringSupported is implemented to allow the binderstring to export
 * the wellknown name of the long generator. This class is abstract seens it must be inherited by the pbinding
 * generated by jorm to handle long generator.
 * @author	<A HREF="mailto:noel.depalma@rd.francetelecom.fr><b>
 *			Noel De Palma
 *			</b></A>
 */
public abstract class PALGenerator
	extends LGeneratorImpl
	implements LazyLoad, PBinding, LGenerator, LGeneratorAccessor {

	/**
	 * Define the range increment. This is the number of long id generation
	 * before the generator writing.
	 */
	private static final long range = 10;

	private long maxRange = range;

	/**
	 *  The loaded flag is used to manage the loading of this object from the DS if necessary.
	 *  If loaded is true then a call to the load method will reload the object from the DS.
	 */
	private boolean loaded = false;

	/**
	 * This connection factory is used if the connection given as parameter
	 * of methods is null.
	 */
	ConnectionFactory connectionFactory = null;

	public void setConnectionFactory (ConnectionFactory cf) {
		connectionFactory = cf;
	}

	public boolean isLoaded() {
		return loaded;
	}


	/**
	 * read the mi through its pbinding from the DS if necessary.
	 * @param conn The connexion to the DS.
	 */

	public void load(Object conn) throws PException {
        Object c = conn;
		try {
			if (!loaded) {
				if (conn == null)
					c = connectionFactory.getConnection();
				read(c);
				super.setNextId(maxRange);
				maxRange += range;
				store(c);
				// loaded = true; it is done in the store method
			}
		}
		catch (ResourceException e) {
			throw new PExceptionIO(e,"Cannot reload this mi");
		}

		finally {
			if (c != null && conn == null)
				try {
					((Connection)c).close();
				}
				catch (Exception e) {
					logger.log(BasicLevel.FATAL, "Cannot close connection!!!");
				}
		}
	}

	/**
	 * Write the object through this its pbinding from the DS.
	 * @param conn The connexion to the DS.
	 */
	public void store(Object conn) throws PException {
		if (Debug.ON && logger!=null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "");
		loaded = true;
        Object c = conn;
		try {
			if (conn == null)
				c = connectionFactory.getConnection();
			write(c);
		}
		catch (ResourceException e) {
			throw new PExceptionIO(e,"Cannot reload this mi");
		}
		finally {
			if (c != null && conn == null)
				try {
					((Connection)c).close();
				}
				catch (Exception e) {
					logger.log(BasicLevel.FATAL, "Cannot close connection!!!");
				}
		}
	}



	/*****************************************************/
	/*     LGenerator Methods overidding              	 */
	/*****************************************************/

	/**
	 * LGenerator Methods overidding. It takes into account the reloading of the generator if necessary.
	 * @param conn the connection to the DS.
	 */
	public long genId(Object conn) throws PException {
			load(conn);
			long nextid = super.genId(conn);
			if (nextid == maxRange) {
				maxRange += range;
				store(conn);
			}
			return nextid;
	}

	/************************************************/
	/****************** PAccessor itf ****************/
	/************************************************/

	/**
	 * Implementation of the paccessor interface.
	 * @see org.objectweb.jorm.api.PAccessor
	 */
	public Object getMemoryInstance() {
		return this;
	}


	/**************************************************/
	/******** LongGenAccessor implem ************/
	/**************************************************/

	/**
	 * Implementation of the LongGenAccessor interface.
	 * It returns the value of the nextid field.
	 * @return	The actual value of the field.
	 */

	public long paGetNextId() throws PException {

		return maxRange;
	}

	/**
	 * Implementation of the LongGenAccessor interface.
	 * It set the value of the nextid field.
	 * @param id The new value of the field.
	 */

	public void paSetNextId(long id) throws PException {
		maxRange = id;
	}

}
