/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */
/**
 * Copyright (C) 2000
 */

package org.objectweb.jorm.facility.naming.gobid.pnc;

import org.objectweb.jorm.api.*;
import org.objectweb.jorm.facility.naming.gobid.binder.LBinder;
import org.objectweb.jorm.facility.naming.gobid.binder.LName;
import org.objectweb.jorm.facility.naming.gobid.binder.PALBinder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Iterator;
import java.util.Set;

/**
 * The class <b>LLPnc</b> defines a basic pnc used by JORM to manage persistent long long name.
 * This pnc works only with the LLName class.
 * This pnc is also a persistent class. the pname of a LLPnc is a wellknown name.
 * @author nono depalma
 */

public class LLPnc implements PNamingContext, Loggable {

    static long NULLID = -1;

    // persistent data
    // pa + mi
    Set classBinders = null;  // gen class used to store binder reference
    PType pt;

    protected Logger logger = null;


    public void setClassBinders(Set s) {
        classBinders = s;
    }

    public Set getClassBinders() {
        return classBinders;
    }

    // IMPLEMENTATION OF THE PNamingContext INTERFACE //
    //------------------------------------------------//

    public boolean codingSupported(short codingtype) {
        return (codingtype == PNamingContext.CTCOMPOSITE);
    }

    public PName decodeAbstract(Object en, Object ctx)
            throws PExceptionNaming, UnsupportedOperationException {
        // The PNCLong manages only PNPNCLong objects
        LLName pn = new LLName();
        long idobj;
        long idbinder;
        try {
            idobj = ((PNameGetter) en).pngetLongField("objId", ctx);
            idbinder = ((PNameGetter) en).pngetLongField("binderId", ctx);
        } catch (PExceptionTyping e) {
            throw new PExceptionNaming(e, "Wrong name structure!!");
        } catch (PExceptionIO e) {
            throw new PExceptionNaming(e, "Wrong name structure!!");
        }
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "objId=" + idobj + "  / binderId=" + idbinder);
        if (idobj == NULLID && idbinder == NULLID)
            return getNull();

        //Lookup the binder in the set of binder
        PNamingContext prevpnc;
        Iterator it = classBinders.iterator();
        while (it.hasNext()) {
            prevpnc = (PNamingContext) (it.next());
            //if (prevpnc != null) {
                if (Debug.ON && logger != null)
                    logger.log(BasicLevel.DEBUG,
                            "Binder Id: " + ((LBinder) prevpnc).getId());
                if (((LBinder) prevpnc).getId() == idbinder) {
                    pn.init(idobj, this, prevpnc);
                    return pn;
                }
            //}
        }
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "The binder was not found: return null");
        return null;
    }

    public PName decodeByte(byte en) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be decoded with this method");
    }

    public PName decodeChar(char en) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be decoded with this method");
    }

    public PName decodeInt(int en) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be decoded with this method");
    }

    public PName decodeIntObj(Integer en) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be decoded with this method");
    }

    public PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be decoded with this method");
    }

    public PName decodeShort(short en) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be decoded with this method");
    }

    public PName decodeString(String en) throws PExceptionNaming {
        LLName pn = new LLName();
        long idobj;
        long idbinder;
        idobj = Long.parseLong(en.substring(0, en.indexOf(":") - 1));
        idbinder = Long.parseLong(en.substring(en.indexOf(":")));
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "objId=" + idobj + "  / binderId=" + idbinder);
        if (idobj == NULLID && idbinder == NULLID)
            return getNull();

        //Lookup the binder in the set of binder
        PNamingContext prevpnc;
        Iterator it = classBinders.iterator();
        while (it.hasNext()) {
            prevpnc = (PNamingContext) (it.next());
            if (((LBinder) prevpnc).getId() == idbinder) {
                pn.init(idobj, this, prevpnc);
                return pn;
            }
        }
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "The binder was not found: return null");
        return null;
    }

    public Object encodeAbstract(PName pn) throws PExceptionNaming, UnsupportedOperationException {
        // pn must be a LLName
        // encode binder to id and return a getter
        // nothing to do : the pname is a getter, and he knows the binder to encode.
        if (!((pn.getPNamingContext()).equals(this)))
            throw new PExceptionNaming("This pname is not valid in this pnc");
        return pn;
    }

    public byte encodeByte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be encode with this method");
    }

    public char encodeChar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be encode with this method");
    }

    public int encodeInt(PName pn) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be encode with this method");
    }

    public Integer encodeIntObj(PName pn) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be encode with this method");
    }

    public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be encode with this method");
    }

    public short encodeShort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "A composite identifier cannot be encode with this method");
    }

    public String encodeString(PName pn) throws PExceptionNaming {
        return Long.toString(((LLName) pn).getIdObj()) + ":"
                + Long.toString(((LBinder) ((LLName) pn).getPrevPnc()).getId());
    }

    public PName getNull() {
        return new LLName(NULLID, this, null);
    }

    public void setPType(PType pt) {
        this.pt = pt;
    }

    public boolean supportDynamicComposite() {
        return true;
    }

    public boolean supportCompositeField(String fn, PType pt) {
        return (
                ("objId".equals(fn) || "binderId".equals(fn))
                && pt.getTypeCode() == PType.TYPECODE_LONG
                );
    }

    public boolean supportStaticComposite() {
        return false;
    }

    /**
     * It associates a PName to an information item within a naming context.
     * @param conn		The connection that can be used to access the data store.
     * @param infoitem	The information item that must be accessible through a
     * 					PName within a naming context.
     * @return	The <code>PName</code> associated to the given information
     * 			item.
     * @exception PExceptionNaming	Raised in case of problem within the
     * 								naming context.
     */
    public PName export(Object conn, Object infoitem)
            throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
        if (infoitem == null)
            return getNull();
        if (!(infoitem instanceof LName))
            throw new PExceptionNaming("Only LName can be exported in this PNC:" + infoitem);

        // if the binder is not registred yet in the classbinder set, we must
        //register it.
        long idbinder = -1;
        PNamingContext pnc = ((PName) infoitem).getPNamingContext();
        if (pnc instanceof PClassMapping)
        // terminal pname
            try {
                idbinder = ((PALBinder)
                        ((PClassMapping) pnc).getPBinder()).getId(conn);
            } catch (PExceptionNaming pn) {
                throw pn;
            } catch (PExceptionIO pio) {
                throw pio;
            } catch (PExceptionProtocol pp) {
                throw pp;
            } catch (PException pe) {
                throw new PExceptionIO(pe);
            }
        else
            throw new PExceptionProtocol(
                    "The naming does not manage intermediate PNamingContext in this version");

        PNamingContext prevpnc = null;
        // refresh sur genclass
        Iterator it = classBinders.iterator();
        boolean found = false;
        while (it.hasNext() && !found) {
            prevpnc = (PNamingContext) (it.next());
            try {
                if (((PALBinder) prevpnc).getId(conn) == idbinder)
                    found = true;
            } catch (PExceptionNaming pn) {
                throw pn;
            } catch (PExceptionIO pio) {
                throw pio;
            } catch (PExceptionProtocol pp) {
                throw pp;
            } catch (PException pe) {
                throw new PExceptionIO(pe);
            }
        }
        if (!found) {
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG, "add a binder: id=" + idbinder);
            classBinders.add(
                    (PALBinder) ((PClassMapping) pnc).getPBinder());
        }

        LLName pnres = new LLName();
        pnres.init(((LName) infoitem).getId(),
                this,
                (PALBinder) ((PClassMapping) ((PName) infoitem).getPNamingContext()).getPBinder());
        return pnres;
    }

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context.
	 * @param hints		Any information relevant for name creation.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 *							the data store.
	 */
	public PName export(Object conn, Object infoitem, Object hints)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		throw new PExceptionNaming("Not supported yet.");
	}

	/**
     * It retrieves the information item associated to the given PName (if any
     * exists).
     * @param conn	The connection that can be used to access the data store.
     * @param pn	The PName to be looked up.
     * @return	The information item associated with the given PName within
     * 			a naming context.
     * @exception PExceptionNaming	Raised in case of problem within the
     * 								naming context.
     */

    public Object resolve(Object conn, PName infoitem) throws PExceptionNaming, PExceptionIO {
        if (infoitem == null)
            throw new PExceptionNaming("This pname is null");
        if (infoitem.isNull())
            return null;
        if (!(infoitem instanceof LLName))
            throw new PExceptionNaming("Only PNlong can be resolved in this PNC");
        if (!((infoitem.getPNamingContext()).equals(this)))
            throw new PExceptionNaming("This pname is not valid in this binder");
        // rebuild the pname
        return ((LLName) infoitem).getPrevPnc().decodeLong(((LLName) infoitem).getIdObj());
    }

    /**
     * It removes the PName and all the associated information within a naming
     * context (i.e., the association between the PName and its information
     * item).
     * @param conn	The connection that can be used to access the data store.
     * @param pn	A PName valid in this naming context.
     * @exception PExceptionNaming	Raised in case of problem within the
     * 								naming context.
     */
    public void unexport(Object conn, Object pn)
            throws PExceptionNaming, PExceptionIO {

        //TODO
    }

    // IMPLEMENTATION OF THE Loggable INTERFACE //
    //------------------------------------------//

    /**
     * It assigns a logger to an component that wants to log things.
     * @param Logger	The logger object.
     */
    public void setLogger(Logger _logger) {
        logger = _logger;
    }

    /**
     * It assigns a logger factory that allows the creation of new loggers.
     * @param loggerfactory	The LoggerFactory object to obtain new loggers.
     */
    public void setLoggerFactory(LoggerFactory loggerfactory) {
        logger = loggerfactory.getLogger(this.getClass().getName());
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return null;
    }
}

