/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */
/**
 * Copyright (C) 2000 France Tlcom R&D
 * DTL/ASR
 */

package org.objectweb.jorm.facility.naming.gobid.pnc;

import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.facility.meminst.lifecycle.lazyload.LazyLoad;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Set;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionFactory;

//import org.objectweb.jorm.genclassmi.set.genset.PAGenSet;
//import org.objectweb.jorm.genclassmi.set.reference.PARefSet;

/**
 * The class <b>PALLPnc</b> defines a pnc used by JORM to reference persistent object identified by a long name.
 * This pnc can only export Basicpnclong name and resolve pnlong name.
 * This pnc is also a persistent class. the pname this pnc is a wellknown name. This mean that the
 * pnclong_mapping uses a binderstring to manage this pnc.
 * This pnc hold a generic class embeding persistent reference to binderlong object.
 * The MI, the accessor and the PBinding is merged in this object.
 * The generated PNCLongPBinding must inherit from this class.
 * @author nono de palma.
 */

public abstract class PALLPnc extends LLPnc
	implements LazyLoad, PBinding, PNamingContext, LLPncAccessor {

	private static long NULLID = -1;

	/**
	 *  The loaded flag is used to manage the loading of this object from the DS if necessary
	 */
	boolean loaded = false;

	/**
	 * This connection factory is used if the connection given as parameter
	 * of methods is null.
	 */
	protected ConnectionFactory connectionFactory = null;

	public Set getClassBinders() {
		try {
			load(null);
			((LazyLoad) classBinders).load(null);
		}
		catch (PException pen) {
		}
		return super.getClassBinders();
	}

	public void setConnectionFactory(ConnectionFactory cf) {
		connectionFactory = cf;
	}

	public boolean isLoaded() {
		return loaded;
	}

	/**
	 * read the mi through its pbinding from the DS if necessary.
	 * @param conn The connexion to the DS.
	 */
	public void load(Object conn) throws PException {
		if (Debug.ON && logger != null)
			logger.log(BasicLevel.DEBUG, "begin");
		Connection c = null;
		try {
			if (!loaded) {
				if (conn == null) {
					c = connectionFactory.getConnection();
				}
				read((conn == null?c:conn));
				loaded = true;
			}
		}
		catch (ResourceException e) {
			throw new PExceptionIO(e, "Cannot reload this mi");
		}
		finally {
			if (c != null && conn == null)
				try {
					((Connection) c).close();
				}
				catch (Exception e) {
					logger.log(BasicLevel.FATAL, "Cannot close connection!!!");
				}
		}
		if (Debug.ON && logger != null)
			logger.log(BasicLevel.DEBUG, "end");
	}

	/**
	 * Write the object through this its pbinding from the DS.
	 * @param conn The connexion to the DS.
	 */
	public void store(Object conn) throws PException {
		if (Debug.ON && logger != null)
			logger.log(BasicLevel.DEBUG, "begin");
		loaded = true;
		Connection c = null;
		try {
			if (conn == null)
				c = connectionFactory.getConnection();
			write((conn == null?c:conn));
		}
		catch (ResourceException e) {
			throw new PExceptionIO(e, "Cannot reload this mi");
		}
		finally {
			if (c != null && conn == null)
				try {
					((Connection) c).close();
				}
				catch (Exception e) {
					logger.log(BasicLevel.FATAL, "Cannot close connection!!!");
				}
		}
		if (Debug.ON && logger != null)
			logger.log(BasicLevel.DEBUG, "end");
	}

	// OVERRIDING THE PNCLong CLASS //
	//------------------------------//

	/**
	 * intercept the decodeabstract method. Does not forward on the inner pnclong.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */

	public PName decodeAbstract(Object en, Object ctx)
		throws UnsupportedOperationException, PExceptionNaming {
		try {
			load(null);
			((LazyLoad) classBinders).load(null);
		}
		catch (PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeAbstract(en, ctx);

	}

	public PName decodeString(String en)
		throws UnsupportedOperationException, PExceptionNaming {
		try {
			load(null);
			((LazyLoad) classBinders).load(null);
		}
		catch (PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.decodeString(en);

	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Object encodeAbstract(PName pn)
		throws UnsupportedOperationException, PExceptionNaming {
		try {
			load(null);
		}
		catch (PExceptionNaming pen) {
			throw pen;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the loading of the object");
		}
		return super.encodeAbstract(pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */

	public PName export(Object conn, Object infoitem)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		try {
			load(conn);
			((LazyLoad) classBinders).load(conn);
			PName pnres = super.export(conn, infoitem);
			((LazyLoad) classBinders).store(conn);
			return pnres;
		}
		catch (PExceptionNaming pen) {
			throw pen;
		}
		catch (PExceptionIO peio) {
			throw peio;
		}
		catch (PExceptionProtocol pep) {
			throw pep;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the life cycle of the object");
		}
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Object resolve(Object conn, PName pn) throws PExceptionNaming, PExceptionIO {
		try {
			load(conn);
		}
		catch (PExceptionNaming pen) {
			throw pen;
		}
		catch (PExceptionIO peio) {
			throw peio;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the life cycle of the object");
		}
		return super.resolve(conn, pn);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public void unexport(Object conn, Object pn) throws PExceptionNaming, PExceptionIO {
		try {
			load(conn);
		}
		catch (PExceptionNaming pen) {
			throw pen;
		}
		catch (PExceptionIO peio) {
			throw peio;
		}
		catch (PException pe) {
			throw new PExceptionNaming(pe,
				"An error occured during the life cycle of the object");
		}
		super.unexport(conn, pn);
	}

	// IMPLEMENTATION OF THE PAccessor INTERFACE //
	//-------------------------------------------//

	/**
	 * Implementation of the paccessor interface.
	 * @see org.objectweb.jorm.api.PAccessor
	 */
	public Object getMemoryInstance() {
		return this;
	}


	// IMPLEMENTATION OF THE PNCLongAccessor INTERFACE //
	//-------------------------------------------------//

	/**
	 * It retrieve the value of the classBinders field.
	 * @return the refset representing the set of binder used by this pnclong.
	 */

	public Object paGetClassBinders(Object connection) throws PException {
		return ((org.objectweb.jorm.facility.meminst.genclass.java.util.Set)
			classBinders).getPBinding().getPName();
	}

	/**
	 * It set the value of the classBinders field.
	 * @param classBinders the refset representing the set of binder used by this pnclong.
	 */

	public void paSetClassBinders(Object cb, Object connection) throws PException {
		PBinding pb = (PBinding) ((PName) cb).resolve(connection);
		if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "cb=" + cb + "\n cb.getpa="
				+ pb.getPAccessor());
		classBinders = (Set) pb.getPAccessor();
	}
}
