/**
 * Copyright (C) 2000
 */

package org.objectweb.jorm.facility.naming.usid;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.naming.api.PBinderDelegate;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.naming.cache.api.Cache;
import org.objectweb.jorm.naming.cache.lib.SimpleHashedCache;
import org.objectweb.jorm.type.api.PType;
import java.util.Iterator;

/**
 * The class <b>BinderString</b> defines the basic string binder used by JORM.
 * It maintain the association between a string and a pbinding. This association is
 * not persistent.
 * @author nono de palma.
 */
public class StrBinder implements PBinderDelegate {
	private static String NULLNAME = "BinderString.NULLPNAME#";
	private Cache bindingcache = null;
	private PClassMapping pcmbinderstring = null; //pclassmapping using this pbinder
	PType pt;

	/**
	 * A constructor that must be initialized through the init method.
	 */
	public StrBinder() {
		this.bindingcache = new SimpleHashedCache();
	}

	/**
	 * A constructor that need not to be initialized through the init method.
	 */
	public StrBinder(PClassMapping pcm) {
		this.bindingcache = new SimpleHashedCache();
		this.pcmbinderstring = pcm;
	}

	/**
	 * It initialize this pbinder by  setting the pclassmapping for which it is the binder.
	 * @param pcm    The pclassmapping related to the pbinder.
	 */
	public void setPClassMapping(PClassMapping pcm) {
		this.pcmbinderstring = pcm;
	}



	// IMPLEMENTATION OF THE PBinder INTERFACE //
	//-----------------------------------------//

	/**
	 * Implementation of the PBinder interface.
	 * @see org.objectweb.jorm.naming.api.PBinder
	 */

	public void bind(PName pn, PBinding pb) throws PExceptionNaming {
		if (pn == null)
			throw new PExceptionNaming("This pname is null");
		if (pb == null)
			throw new PExceptionNaming("This pbinding is null");
		if (!(pn instanceof StrName))
			throw new PExceptionNaming("Can only bind StrName");
		if (!pn.getPNamingContext().equals(pcmbinderstring))
			throw new PExceptionNaming("This pname is not valid in this binder");
		if (!pb.getPClassMapping().equals(pcmbinderstring))
			throw new PExceptionNaming("This pbinding is not valid in this binder");
		if (pn.isNull())
			throw new PExceptionNaming("This PName represent the null value");

		// it is a rebind operation
		// we overwrite the previous binding
		bindingcache.register(pn, pb);
	}

	/**
	 * Implementation of the PBinder interface.
	 * @see org.objectweb.jorm.naming.api.PBinder
	 */
	public PClassMapping getBinderClassMapping() {
		return this.pcmbinderstring;
	}

	/**
	 * Implementation of the PBinder interface.
	 * @see org.objectweb.jorm.naming.api.PBinder
	 */
	public Iterator iteratePBinding() throws UnsupportedOperationException {
		return bindingcache.getIterator();

	}

	/**
	 * Implementation of the PBinder interface.
	 * @see org.objectweb.jorm.naming.api.PBinder
	 */
	public void unbind(PBinding pb) throws PExceptionNaming {
		if (pb == null)
			throw new PExceptionNaming("This pbinding is null");
		PName pn = pb.getPName();
		if (pn == null)
			throw new PExceptionNaming("This PBinding is not bounded with a pname");
		if (!pb.getPClassMapping().equals(pcmbinderstring))
			throw new PExceptionNaming("This pbinding is not valid in this binder");
		bindingcache.unregister(pn);
	}


	// IMPLEMENTATION OF THE PNamingContext INTERFACE //
	//------------------------------------------------//

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public boolean codingSupported(short codingtype) {
		return codingtype == PNamingContext.CTSTRING;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeByte(byte en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Byte identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeChar(char en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Char identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeInt(int en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An int identifier cannot be decode with this method");
	}
	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeIntObj(Integer en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An Integer identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Long identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeShort(short en) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Short identifier cannot be decode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeString(String en) throws PExceptionNaming {
		if (NULLNAME.equals(en))
			return getNull();
		else
			return new StrName(en, (PNamingContext) pcmbinderstring);
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Object encodeAbstract(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An abstract identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public byte encodeByte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Byte identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public char encodeChar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Char identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public int encodeInt(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An Int identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public Integer encodeIntObj(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An Int identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Long identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public short encodeShort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"A Short identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface. The pname must be a StrName.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public String encodeString(PName pn) throws PExceptionNaming {
		// the PNAME must be a StrName
		if (!(pn instanceof StrName))
			throw new PExceptionNaming("Can only bind StrName");

		if (!pn.getPNamingContext().equals(pcmbinderstring))
			throw new PExceptionNaming("This pname is not valid in this binder");
		if (pn.isNull())
			return NULLNAME;
		else
			return ((StrName) pn).getId();

	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName decodeAbstract(Object en, Object ctx) throws PExceptionNaming, UnsupportedOperationException {
		throw new UnsupportedOperationException(
			"An abstract identifier cannot be encode with this method");
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public boolean supportDynamicComposite() {
		return false;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public boolean supportCompositeField(String fn, PType ft) {
		return false;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public boolean supportStaticComposite() {
		return false;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public void setPType(PType pt) {
		this.pt = pt;
	}

	/**
	 * Implementation of the PNamingContext interface.
	 * It associates a PName to an information item within a naming context.
	 * the infoitem must support the StrNameSupported interface. This interface is used by
	 * the binder to retrieve the wellknown name of the object.
	 *
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context. In the case of a the binder string,
	 *                  the infoitem should be a pbinding.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item. With this naming context hte pname is a StrName.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName export(Object conn, Object infoitem)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		// in the case of a binderstring,
		// we need not to check if the infoitem is already exported.
		// we build always the same pname.
		if (infoitem == null) {
			return getNull();
		}
		if (infoitem instanceof StrName
			&& ((PName) infoitem).getPNamingContext().equals(pcmbinderstring))
			return (PName) infoitem;
		if (!(infoitem instanceof PBinding))
			throw new PExceptionNaming("The item cannot be exported by this binder");
		// This pname is valid in this binder ?
		PName pn = ((PBinding) (infoitem)).getPName();

		if (pn instanceof StrName
			&& pn != null
			&& pn.getPNamingContext().equals(pcmbinderstring)) {

			bindingcache.register(pn, infoitem);
			return pn;
		}
		else
			throw new PExceptionNaming("This item cannot be exported by this binder");
	}

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context.
	 * @param hints		Any information relevant for name creation.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 *							the data store.
	 */
	public PName export(Object conn, Object infoitem, Object hints)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		throw new PExceptionNaming("Not supported yet.");
	}

	/**
	 * It yields a "null" representation of a name within this naming context.
	 * This is a null StrName
	 * @return	The "null" PName associated with this naming context.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public PName getNull() {
		return new StrName(NULLNAME, pcmbinderstring);
	}

	/**
	 * It retrieves the information item associated to the given PName (if any
	 * exists). The pname must be a StrName and must support the hashcode method.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be looked up.
	 * @return	The PBinding associated with the given PName within
	 * 		this binder
	 * @exception PExceptionNaming	Raised if the Pname is not registered into this binder.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */

	public Object resolve(Object conn, PName pn) throws PExceptionNaming, PExceptionIO {
		if (pn == null)
			throw new PExceptionNaming("This pname is null");
		if (pn.isNull())
			return null;
		if (!pn.getPNamingContext().equals(pcmbinderstring))
			throw new PExceptionNaming("This pname is not valid in this binder");
		return bindingcache.lookup(pn);
	}

	/**
	 * It removes the PName and all the associated information within a naming
	 * context (i.e., the association between the PName and its information
	 * item). in the case of the StrBinder, the unexport operation do nothing more
	 * than the unbind operation.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	A PName valid in this naming context.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
	public void unexport(Object conn, Object pn) throws PExceptionNaming, PExceptionIO {
		// in the case of the StrBinder, the unexport operation do nothing more than the unbind operation.
		if (!(pn instanceof StrName) || pn == null)
			throw new PExceptionNaming("This pname cannot be managed by this PBinder");
		if (!((PName) pn).getPNamingContext().equals(pcmbinderstring))
			throw new PExceptionNaming("This pname is not valid in this binder");
		bindingcache.unregister(pn);
	}
}

