/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.interp.genclass.lib;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.GenClass;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.interp.genclass.api.FieldDesc;
import org.objectweb.jorm.interp.genclass.api.GenClassProp;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Iterator;

/**
 * This abstract class provides all the generic code for the PClassMapping
 * associated with JORM generic classes. It implements all methods except the
 * following ones: createPBinding, initMappingStructures, isConform, and
 * removeMappingStructures.
 * @author	P. Dchamboux
 */
public abstract class GenClassMapping
	implements PClassMapping, GenClassProp, Loggable {
	/**
	 * This field references the mapper. The mapper is used to know the mapper
	 * name and the mapping name.
	 */
	protected PMapper mapper = null;

	/**
	 * It references the meta-information associated with this generic class.
	 */
	protected GenClass genClassMetaInfo = null;

	/**
	 * This field is the PType of the involved generic class.
	 */
	protected PType genClassPType = null;

	/**
	 * This array contains the ScalarField meta object elements which
	 * represent the index fields of the generic class instance.
	 */
	protected FieldDesc[] indexFields = null;

	/**
	 * This array contains the ScalarField meta object elements which
	 * represent the identifier fields of the generic class instance.
	 */
	protected FieldDesc[] identifierFields = null;

	/**
	 * This array contains the ScalarField meta object elements which
	 * represent the element fields of the generic class instance. If it is
	 * a reference, there must be at least two fields: one describing the
	 * reference field, and one describing its structural projection. This
	 * structural projection may be a composite name in which case it should
	 * describes all the fields of this composite name.
	 */
	protected FieldDesc[] elemFields = null;

	/**
	 * The actual PBinder.
	 */
	protected PBinder binder = null;

	/**
	 * The PClassMapping of the generic class instance representing the element
	 * stored within this generic class object. This is valid only if the
	 * elements of this generic class object are also generic class objects.
	 */
	protected PClassMapping elemClassMapping = null;

	/**
	 * The naming context associated to the reference elements of this generic
	 * class instance.
	 */
	protected PNamingContext elemNamingContext = null;
	protected boolean elemCodingOnly = true;

	/**
	 * The memory instance manager
	 */
	protected Object manager = null;

	/**
	 * The logger to log all information related to the mapping of this class.
	 */
	protected Logger logger = null;

	/**
	 * The loggerFactory.
	 */
	protected LoggerFactory loggerFactory = null;


	public FieldDesc[] getElemFields() {
		return elemFields;
	}

	public FieldDesc[] getIdentifierFields() {
		return identifierFields;
	}

	public FieldDesc[] getIndexFields() {
		return indexFields;
	}

	/**
	 * It associates a PBinding object with a PName within a PBinder. This
	 * sets/resets the link between a PBinding and a DSI.
	 * @param pn    The PName to associate to the PBinding. This PName must be
	 * 				valid within this binder just like with a naming context.
	 * @param pb	The PBinding to bind with the given PName; this PBinding
	 * 				must have been created or initialised by this PBinder.
	 * @exception PExceptionNaming	The given PName is not valid in this
	 * 								PBinder.
	 * @exception PExceptionProto	This PBinding cannot be managed by this
	 * 								PBinder.
	 */
	public void bind(PName pn, PBinding pb) throws PExceptionNaming {
		GenClassBinding gcb = (GenClassBinding) pb;
		byte nextstate = PStateGraph.nextStatePBinding(gcb.status,
			PBinding.ACTION_BIND);
		// Perform the transition
		binder.bind(pn, pb);
		gcb.pName = pn;
		gcb.status = nextstate;
	}

	/**
	 * It tests if a given coding type is supported by this naming context.
	 * @param codingtype	One of the coding type defined by CTxxx constants
	 * 						defined within this interface.
	 * @return	It returns true if the coding type is supported by this naming
	 * 			context.
	 */
	public boolean codingSupported(short codingtype) {
		return binder.codingSupported(codingtype);
	}

	/**
	 * The <b>decodeAbstract</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is abstract, which means that this is a Java
	 * object of any complexity. One example of such object is a getter
	 * interface representing a relational primary key (also called composite
	 * names).
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The Object to be decoded.
	 * @return	The resulting PName.
	 */
	public PName decodeAbstract(Object en, Object ctx) throws PExceptionNaming, UnsupportedOperationException {
		return binder.decodeAbstract(en, ctx);
	}

	/**
	 * The <b>decodeByte</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a byte.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The byte to be decoded.
	 * @return	The resulting PName.
	 */
	public PName decodeByte(byte en) throws PExceptionNaming, UnsupportedOperationException {
		return binder.decodeByte(en);
	}

	/**
	 * The <b>decodeChar</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a char.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The char to be decoded.
	 * @return	The resulting PName.
	 */
	public PName decodeChar(char en) throws PExceptionNaming, UnsupportedOperationException {
		return binder.decodeChar(en);
	}

	/**
	 * The <b>decodeInt</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is an int.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The int to be decoded.
	 * @return	The resulting PName.
	 */
	public PName decodeInt(int en) throws PExceptionNaming, UnsupportedOperationException {
		return binder.decodeInt(en);
	}

	/**
	 * The <b>decodeIntObj</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is an int.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The int to be decoded.
	 * @return	The resulting PName.
	 */
	public PName decodeIntObj(Integer en) throws PExceptionNaming, UnsupportedOperationException {
		return binder.decodeIntObj(en);
	}

	/**
	 * The <b>decodeLong</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a long.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The long to be decoded.
	 * @return	The resulting PName.
	 */
	public PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException {
		return binder.decodeLong(en);
	}

	/**
	 * The <b>decodeShort</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a short.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The short to be decoded.
	 * @return	The resulting PName.
	 */
	public PName decodeShort(short en) throws PExceptionNaming, UnsupportedOperationException {
		return binder.decodeShort(en);
	}

	/**
	 * The <b>decodeString</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a String.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The String to be decoded.
	 * @return	The resulting PName.
	 */
	public PName decodeString(String en) throws PExceptionNaming {
		return binder.decodeString(en);
	}

	/**
	 * The <b>encodeAbstract</b> method produces an encoded representation of a PName
	 * within a naming context. The encoding format is the case of this method
	 * is a Java object of any complexity. Example of such object is a getter
	 * interface giving access of the fields of a relational primary key (also
	 * called composite names).
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded.
	 * @return	The Object corresponding to the encoded representation.
	 */
	public Object encodeAbstract(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return binder.encodeAbstract(pn);
	}

	/**
	 * The <b>encodeByte</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a byte.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The byte corresponding to the encoded representation.
	 */
	public byte encodeByte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return binder.encodeByte(pn);
	}

	/**
	 * The <b>encodeChar</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a char.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The char corresponding to the encoded representation.
	 */
	public char encodeChar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return binder.encodeChar(pn);
	}

	/**
	 * The <b>encodeInt</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a int.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The int corresponding to the encoded representation.
	 */
	public int encodeInt(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return binder.encodeInt(pn);
	}

	/**
	 * The <b>encodeInt</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a int.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The int corresponding to the encoded representation.
	 */
	public Integer encodeIntObj(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return binder.encodeIntObj(pn);
	}

	/**
	 * The <b>encodeLong</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a long.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The long corresponding to the encoded representation.
	 */
	public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return binder.encodeLong(pn);
	}

	/**
	 * The <b>encodeShort</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a short.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The short corresponding to the encoded representation.
	 */
	public short encodeShort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return binder.encodeShort(pn);
	}

	/**
	 * The <b>encodeString</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a String.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The String corresponding to the encoded representation.
	 */
	public String encodeString(PName pn) throws PExceptionNaming {
		return binder.encodeString(pn);
	}

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 */
	public PName export(Object conn, Object infoitem)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		GenClassBinding gcb = (GenClassBinding) infoitem;
		byte nextstate = PStateGraph.nextStatePBinding(gcb.status,
			PBinding.ACTION_EXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized operation");
		// Perform the transition
		PName pn = binder.export(conn, infoitem);
		gcb.pName = pn;
		gcb.status = nextstate;
		return pn;
	}

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context.
	 * @param hints		Any information relevant for name creation.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 *							the data store.
	 */
	public PName export(Object conn, Object infoitem, Object hints)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		GenClassBinding gcb = (GenClassBinding) infoitem;
		byte nextstate = PStateGraph.nextStatePBinding(gcb.status,
			PBinding.ACTION_EXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized operation");
		// Perform the transition
		PName pn = binder.export(conn, infoitem, hints);
		gcb.pName = pn;
		gcb.status = nextstate;
		return pn;
	}

	/**
	 * It retrieves the name of the class mapped by this PClassMapping. It
	 * should be the same as the one get when calling
	 * "getPType().getJormName()" if the PType exists.
	 */
	public String getClassName() {
		return genClassPType.getJavaName();
	}

	/**
	 * It gives access to the PClassMapping associated to the element of a
	 * generic class whose type is a generic class.
	 * @return	The PClassMapping of the generic class instance.
	 */
	public PClassMapping getGenClassMapping() {
		return elemClassMapping;
	}

	/**
	 * It gives access to the PClassMapping associated to a field whose type is
	 * a generic class. It has no meaning within this context.
	 * @param fn	The name of the generic class field.
	 * @return	The PClassMapping of the generic class instance: always null
	 * 			here.
	 */
	public PClassMapping getGenClassMapping(String fn) {
		return null;
	}

	/**
	 * It returns the memory instance manager. This manager could manage the
	 * life cycle of memorty instances. A manager is not always required,
	 * therefore this method can return null.
	 */
	public Object getMIManager() {
		return manager;
	}

	/**
	 * It returns the Class/GenClass object representing the meta-information
	 * associated to the JORM class/generic class whose mapping is defined by
	 * this PClassMapping. It may be null if not requested at mapping time.
	 * @return	The Class/GenClass meta-object or null.
	 * @see	org.objectweb.jorm.api.PMapper#map
	 */
	public MetaObject getMetaInfo() {
		return genClassMetaInfo;
	}

	/**
	 * It yields a "null" representation of a name within this naming context.
	 * @return	The "null" PName associated with this naming context.
	 */
	public PName getNull() {
		return binder.getNull();
	}

	/**
	 * It yields the PBinder associated to this PClassMapping. As a PBinder
	 * itself, this PClassMapping delegates this behaviour to this PBinder.
	 * @return	The PBinder associated to this PClassMapping.
	 */
	public PBinder getPBinder() {
		return binder;
	}

	/**
	 * It returns the PClassMapping for which it is the binder.
	 * @return	The PClassMapping to which it is associated.
	 */
	public PClassMapping getBinderClassMapping() {
		return this;
	}

	/**
	 * It gives access to the mapper that has mapped the associated JORM class
	 * as defined by this PClassMapping.
	 * @return	The associated PMapper.
	 */
	public PMapper getPMapper() {
		return mapper;
	}

	/**
	 * It returns the naming context that manages the references contained by
	 * the generic class instance whose mapping is defined by this
	 * PClassMapping.
	 * @return	The naming context that manages references contained by a
	 * 			generic class instance.
	 * @exception	UnsupportedOperationException	It is raised when this
	 * 												PClassMapping does not map
	 * 												a generic class instance, or
	 * 												when the elements of this
	 * 												generic class instance are
	 * 												not references.
	 */
	public PNamingContext getPNamingContext()
		throws UnsupportedOperationException {
		return elemNamingContext;
	}

	/**
	 * It returns the naming context that manages the reference associated to
	 * a field whose type is a JORM class.
	 * @return	The naming context that manages the reference defined by a
	 * 			field.
	 * @exception	UnsupportedOperationException	It is raised when no naming
	 * 												context is associated to
	 * 												the corresponding field.
	 */
	public PNamingContext getPNamingContext(String fn)
		throws UnsupportedOperationException {
		throw new UnsupportedOperationException();
	}

	/**
	 * It intialises the class mapping. It especially assigns a mapper as well
	 * as a meta-object describing either a class or a generic class.
	 * @param mapper	The mapper to be assigned to this class mapping.
	 * @param metaclass	The meta-object describing the associated class. It may
	 * 					be null.
	 */
	public void init(PMapper mapper, MetaObject metaclass) {
		this.mapper = mapper;
		genClassMetaInfo = (GenClass) metaclass;
	}

	/**
	 * It yields an Iterator that gives access to all PBinding objects that
	 * are managed by this PBinder. It may not be supported by all PBinder.
	 * @return	The relevant Iterator.
	 * @exception UnsupportedOperationException	It is raised if this method is
	 * 											not supported by this PBinder.
	 */
	public Iterator iteratePBinding() throws UnsupportedOperationException {
		return binder.iteratePBinding();
	}

	/**
	 * It retrieves the information item associated to the given PName (if any
	 * exists).
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be looked up.
	 * @return	The information item associated with the given PName within
	 * 			a naming context.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 */
	public Object resolve(Object conn, PName pn)
		throws PExceptionNaming, PExceptionIO {
		return binder.resolve(conn, pn);
	}

	/**
	 * It assigns the memory instance manager. This manager could manage the
	 * life cycle of memorty instances. A manager is not always required.
	 */
	public void setMIManager(Object m) {
		manager = m;
	}

	/**
	 * It assigns the PBinder associated to this PClassMapping. As a PBinder
	 * itself, this PClassMapping delegates this behaviour to this PBinder.
	 * @param pb	The PBinder associated to this PClassMapping.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the name that identifies the DSI for
	 * 									the PClassMapping is not supported by
	 * 									this PBinder that should be connected
	 * 									to this PClassMapping.
	 */
	public void setPBinder(PBinder pb) throws PExceptionTyping {
		binder = pb;
	}

	/**
	 * It assigns the PNamingContext used to manage the references hold by a
	 * generic class instance whose mapping is described by this PClassMapping.
	 * It should verify that the name structure for this reference is supported
	 * by the corresponding naming context.
	 * @param pnc	The PNamingContext to be associated to this generic class
	 * 				PClassMapping for managing its reference elements.
	 * @param codingonly	It specifies that when using this naming context for
	 * 						managing a reference, it only executes coding or
	 * 						both coding and export/resolve phase.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the names that reference the DSI stored
	 * 									within the generic class instance is not
	 * 									supported by this naming context.
	 * @exception	UnsupportedOperationException	It is raised when this
	 * 												PClassMapping does not map
	 * 												a generic class instance, or
	 * 												when the elements of this
	 * 												generic class instance are
	 * 												not references.
	 */
	public void setPNamingContext(PNamingContext pnc, boolean codingonly)
		throws PExceptionTyping, UnsupportedOperationException {
		elemNamingContext = pnc;
		elemCodingOnly = codingonly;
	}

	/**
	 * It assigns the PNamingContext used to manage the reference hold by a
	 * field whose type is a JORM class.
	 * It should verify that the name structure for this reference is supported
	 * by the corresponding naming context.
	 * @param fn	The name of the involved field.
	 * @param pnc	The PNamingContext to be associated to this field for
	 * 				managing the reference it may hold.
	 * @param codingonly	It specifies that when using this naming context for
	 * 						managing a reference, it only executes coding or
	 * 						both coding and export/resolve phase.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the names that reference the DSI stored
	 * 									within the generic class instance is not
	 * 									supported by this naming context.
	 * @exception	UnsupportedOperationException	It is raised when no naming
	 * 												context is associated to
	 * 												the corresponding field.
	 */
	public void setPNamingContext(String fn, PNamingContext pnc,
								  boolean codingonly)
		throws PExceptionTyping, UnsupportedOperationException {
		throw new UnsupportedOperationException();
	}

	/**
	 * It assigns the GenClassMapping used to manage the reference hold by a
	 * field whose type is a JORM class.
	 */
	public void setGenClassMapping(PClassMapping pcm)
		throws UnsupportedOperationException {
		// TODO: verifies with the PType that the elements are gen class
		// themselves
		elemClassMapping = pcm;
	}

	/**
	 * It assigns the GenClassMapping used to manage the reference hold by a
	 * field whose type is a JORM class.
	 */
	public void setGenClassMapping(String fn, PClassMapping pcm)
		throws UnsupportedOperationException {
		throw new UnsupportedOperationException("A generic class never contain several fields");
	}

	/**
	 * It assigns the PType associated to the JORM class which is the type of
	 * the reference managed by this naming context.
	 * @param pt	The PType of the JORM class reference managed by this naming
	 * 				context.
	 */
	public void setPType(PType pt) {
		binder.setPType(pt);
	}

	/**
	 * Tests if this naming context supports comosite name through a dynamic
	 * approach, which means that encodingAbstract/decodingAbstract use getter
	 * objects that implement the PNameGetter interface.
	 * @return	true if dynamic getter is supported by composite name coding.
	 */
	public boolean supportDynamicComposite() {
		return binder.supportDynamicComposite();
	}

	/**
	 * In case of a naming context that supports composite names (i.e., names
	 * composed of several typed fields, whose type is limited to scalar ones),
	 * it tests if a particular field is defined by this composite name.
	 * It should be used by PClassMapping objects in order to verify, at
	 * deployment time, that naming contexts that are defined to manage the
	 * various references supports the same encoding name structure.
	 * @return	true if the proposed field is supported by this composite name.
	 */
	public boolean supportCompositeField(String fn, PType ft) {
		return binder.supportCompositeField(fn, ft);
	}

	/**
	 * Tests if this naming context supports comosite name through a static
	 * approach, which means that encodingAbstract/decodingAbstract use getter
	 * objects that implement specific NsFieldGetter interface generated for
	 * a particular field "Field" within a particular NameDef space "Ns".
	 * @return	true if static getter is supported by composite name coding.
	 */
	public boolean supportStaticComposite() {
		return binder.supportStaticComposite();
	}

	/**
	 * It releases the link between the PBinding passed as parameter and its
	 * associated DSI represented by the PName referenced by this PBinding.
	 * @param pb	The PBinding to unbind.
	 * @exception PExceptionProto	This PBinding cannot be managed by this
	 * 								PBinder.
	 */
	public void unbind(PBinding pb) throws PExceptionNaming {
		GenClassBinding gcb = (GenClassBinding) pb;
		byte nextstate = PStateGraph.nextStatePBinding(gcb.status,
			PBinding.ACTION_UNBIND);
		// If the state does not change, do nothing
		if (nextstate == gcb.status)
			return;
		// Perform the transition
		binder.unbind(pb);
		gcb.status = nextstate;
	}

	/**
	 * It removes the PName and all the associated information within a naming
	 * context (i.e., the association between the PName and its information
	 * item).
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	In case of a binder, the unexport is done on a PBinding.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 */
	public void unexport(Object conn, Object pn)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol {
		GenClassBinding gcb = (GenClassBinding) pn;
		byte nextstate = PStateGraph.nextStatePBinding(gcb.status,
			PBinding.ACTION_UNEXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized operation");
		// Perform the transition
		binder.unexport(conn, pn);
		gcb.status = nextstate;
	}

	// IMPLMENTATION OF THE GenClassProp INTERFACE

	/**
	 * Defines the type property of a generic class instance.
	 * @param pt	The associated type.
	 */
	public void setGcType(PType pt) {
		genClassPType = pt;
	}

	/**
	 * Defines the fields property specifying the fields composing the
	 * identifier structure of a generic class instance.
	 * @param fds	The array of descriptions of the associated fields.
	 */
	public void setIdentifierDesc(FieldDesc[] fds) {
		identifierFields = fds;
	}

	/**
	 * Defines the fields property specifying the fields composing the
	 * index structure of a generic class instance elements.
	 * @param fds	The array of descriptions of the associated fields.
	 */
	public void setIndexDesc(FieldDesc[] fds) {
		indexFields = fds;
	}

	/**
	 * Defines the fields property specifying the fields composing the
	 * element value structure of a generic class instance elements.
	 * @param fds	The array of descriptions of the associated fields.
	 */
	public void setElemDesc(FieldDesc[] fds) {
		elemFields = fds;
	}

	/**
	 * It assigns a logger factory that allows the creation of new loggers.
	 * @param loggerfactory	The LoggerFactory object to obtain new loggers.
	 */
	public void setLoggerFactory(LoggerFactory lf) {
		loggerFactory = lf;
	}

	/**
	 * It assigns a logger to an component that wants to log things.
	 * @param Logger	The logger object.
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}
}
