/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.interp.genclass.rdb.lib;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.interp.genclass.lib.GenClassBinding;
import org.objectweb.jorm.interp.genclass.lib.GenClassMapping;
import org.objectweb.jorm.type.api.PExceptionTyping;

import java.sql.Blob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;

/**
 * This class provides the implementation of PBinding for RDBMS.
 * @author	P. Dchamboux
 */

public class RdbGenClassBinding extends GenClassBinding {

	/**
	 * It constructs a RdbGenClassBinding.
	 */
	public RdbGenClassBinding() {
		super();
	}

	/**
	 * It constructs a RdbGenClassBinding.
	 */
	public RdbGenClassBinding(GenClassMapping mapping)
		throws PExceptionProtocol {
		super(mapping);
	}

	/**
	 * It verifies if the DSI attached to this binding exists within the DS.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @return	It returns true if the DSI exists.
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_NOACC
	 */
	public boolean exist(Object conn) throws PExceptionIO, PExceptionProtocol {
		byte nextstate = PStateGraph.nextStatePBinding(status, ACTION_EXIST);
		if (nextstate == LIFECYCLE_ERROR)
			throw new PExceptionProtocol(
				"Cannot perform exist on this GenClass PBinding");
		boolean exist = true; // TO BE REWORKED IN FUTURE !!
//		try {
//			RdbGenClassMapping gcmapping = (RdbGenClassMapping) genClassMapping;
//			Connection sqlconn = gcmapping.getConnection(conn);
//			StringBuffer sqlstring = new StringBuffer("SELECT ");
//			gcmapping.defineAllNames(sqlstring, gcmapping.ID);
//			sqlstring.append(" FROM " + gcmapping.tableName + " WHERE ");
//			gcmapping.defineIdentifierExpr(sqlstring, pName, gcmapping.SEPAND, sqlconn);
//			sqlstring.append(";");
//			Statement stmt = sqlconn.createStatement();
//			exist = stmt.executeQuery(sqlstring.toString()).next();
//		}
//		catch (SQLException e) {
//			throw new PExceptionIO(e);
//		}
//		catch (PExceptionTyping e) {
//			throw new PExceptionProtocol("Typing exception at runtime!!");
//		}
//		status = nextstate;
		return exist;
	}

	/**
	 * It reads the data store instance designated by the PName of this
	 * binding and transfers each attribute value to the accessor object.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionNaming	It is raised when there is a problem for the
	 * 								associated binder to manage this PName.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 */
	public void read(Object conn)
		throws PExceptionIO, PExceptionNaming, PExceptionProtocol {
		byte nextstate = PStateGraph.nextStatePBinding(status, ACTION_READ);
		if (nextstate == LIFECYCLE_ERROR)
			throw new PExceptionProtocol(
				"Cannot perform read on this GenClass PBinding");
		try {
			RdbGenClassMapping gcmapping = (RdbGenClassMapping) genClassMapping;
			Connection sqlconn = gcmapping.getConnection(conn);
			StringBuffer sqlstring = new StringBuffer("SELECT ");
			gcmapping.defineAllNames(sqlstring,
				(byte) (gcmapping.INDEX | gcmapping.ELEM));
			sqlstring.append(" FROM " + gcmapping.tableName + " WHERE ");
			gcmapping.defineIdentifierExpr(sqlstring, pName, gcmapping.SEPAND, sqlconn);
			sqlstring.append(";");
			boolean nextvalid;
			ResultSet rs = sqlconn.createStatement()
				.executeQuery(sqlstring.toString());
			int size = ((RdbGenClassMapping) genClassMapping)
				.typeConverter.fetchResultSetSize(rs);
			paccessorGenClass.paSetNbElem(size);
			PNameGetter png = gcmapping.getPNameGetter(rs);
			while (rs.next()) {
				gcmapping.passValuesToAccessor(conn, rs, paccessorGenClass,
					png);
			}
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		catch (PExceptionTyping e) {
			throw new PExceptionProtocol("Typing exception at runtime!!");
		}
		status = nextstate;
	}

	/**
	 * It transfers each attribute value taken from the PAccessor
	 * object to the data store instance designated by the PName
	 * associated with this binding.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @exception   e1 if the state of the binding does not allow writing
	 *                 to be completed (no valid name for example)
	 *              e2 if the pc parameter does not allow acces to the
	 *                 data store
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_NEWTOWRITE
	 * 									LIFECYCLE_DELTOWRITE
	 * 									LIFECYCLE_DELTOWRITENOACC
	 */
	public void write(Object conn)
		throws PExceptionIO, PExceptionNaming, PExceptionProtocol {
		byte nextstate = PStateGraph.nextStatePBinding(status, ACTION_WRITE);
		if (nextstate == LIFECYCLE_ERROR)
			throw new PExceptionProtocol(
				"Cannot perform write on this GenClass PBinding");
		try {
			RdbGenClassMapping gcmapping = (RdbGenClassMapping) genClassMapping;
			Connection sqlconn = gcmapping.getConnection(conn);
			switch (status) {
			case LIFECYCLE_NEWTOWRITE:
				createGcObject(gcmapping, sqlconn);
				break;
			case LIFECYCLE_DELTOWRITE:
			case LIFECYCLE_DELTOWRITENOACC:
				deleteGcObject(gcmapping, sqlconn);
				break;
			default :
				if (paccessorGenClass.paDeltaSupported())
					updateGcObject(gcmapping, sqlconn);
				else {
					deleteGcObject(gcmapping, sqlconn);
					createGcObject(gcmapping, sqlconn);
				}
				break;
			}
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		catch (PExceptionTyping e) {
			throw new PExceptionProtocol(e, "Typing exception at runtime!!");
		}
		status = nextstate;
	}

	/**
	 *
	 */
	private void createGcObject(RdbGenClassMapping gcmapping,
								Connection sqlconn) throws SQLException,
		PExceptionProtocol, PExceptionTyping, PExceptionNaming, PExceptionIO {
		int size = paccessorGenClass.paGetNbElem();
		if (size == 0) {
			// Inserts the empty generic class instance marker
			return;
		}
		// Inserts all elements of the MI of this generic class instance
		StringBuffer sqlstring = new StringBuffer("INSERT INTO "
			+ gcmapping.tableName + " VALUES (");
		gcmapping.defineIdentifierExpr(sqlstring, pName, gcmapping.SEPSCOL,
			sqlconn);
		String idprefix = sqlstring.toString();
		Iterator iterelem = paccessorGenClass.paIterator();
		while (iterelem.hasNext()) {
			PIndexedElem pie = (PIndexedElem) iterelem.next();
			sqlstring = new StringBuffer(idprefix);
			Blob elemblob
				= gcmapping.defineElemExpr(sqlconn, sqlstring, pie, false, RdbGenClassMapping.SEPSCOL);
			gcmapping.defineIndexExpr(sqlstring, pie, gcmapping.SEPSCOL);
			sqlstring.append(");");
			if (elemblob != null) {
				PreparedStatement stmt
					= sqlconn.prepareStatement(sqlstring.toString());
				stmt.setBlob(1, elemblob);
				stmt.execute();
			}
			else {
				sqlconn.createStatement().executeUpdate(sqlstring.toString());
			}
		}
	}

	/**
	 *
	 */
	private void deleteGcObject(RdbGenClassMapping gcmapping,
								Connection sqlconn) throws SQLException,
		PExceptionIO, PExceptionProtocol, PExceptionTyping, PExceptionNaming {
		StringBuffer sqlstring = new StringBuffer(
			"DELETE FROM " + gcmapping.tableName + " WHERE ");
		gcmapping.defineIdentifierExpr(sqlstring, pName, gcmapping.SEPAND,
			sqlconn);
		sqlstring.append(";");
		sqlconn.createStatement().executeUpdate(sqlstring.toString());
	}

	/**
	 *
	 */
	private void updateGcObject(RdbGenClassMapping gcmapping,
								Connection sqlconn) throws SQLException,
		PExceptionTyping, PExceptionProtocol, PExceptionNaming, PExceptionIO {
		StringBuffer sqlstring = new StringBuffer("INSERT INTO "
			+ gcmapping.tableName + " VALUES (");
		gcmapping.defineIdentifierExpr(sqlstring, pName, gcmapping.SEPSCOL,
			sqlconn);
		String insprefix = sqlstring.toString();
		sqlstring = new StringBuffer("DELETE FROM "
			+ gcmapping.tableName + " WHERE ");
		gcmapping.defineIdentifierExpr(sqlstring, pName, gcmapping.SEPAND,
			sqlconn);
		String delprefix = sqlstring.toString();
		Iterator iterelem = paccessorGenClass.paIterator();
		while (iterelem.hasNext()) {
			PIndexedElem pie = (PIndexedElem) iterelem.next();
			Blob elemblob = null;
			switch (pie.getElemStatus()) {
			case PIndexedElem.ELEM_CREATED:
				sqlstring = new StringBuffer(insprefix);
				elemblob
					= gcmapping.defineElemExpr(sqlconn, sqlstring, pie, false, RdbGenClassMapping.SEPSCOL);
				gcmapping.defineIndexExpr(sqlstring, pie, gcmapping.SEPSCOL);
				sqlstring.append(");");
				break;
			case PIndexedElem.ELEM_DELETED:
				sqlstring = new StringBuffer(delprefix);
				if (gcmapping.defineIndexExpr(sqlstring, pie, gcmapping.SEPAND)) {
					elemblob = gcmapping.defineElemExpr(sqlconn, sqlstring, pie, true, RdbGenClassMapping.SEPAND);
				}
				sqlstring.append(";");
				break;
			case PIndexedElem.ELEM_MODIFIED:
				sqlstring = new StringBuffer("UPDATE "
					+ gcmapping.tableName + " SET ");
				gcmapping.defineElemExpr(sqlconn, sqlstring, pie, true, RdbGenClassMapping.SEPSCOL);
				sqlstring.append(" WHERE ");
				gcmapping
					.defineIdentifierExpr(sqlstring, pName, gcmapping.SEPAND,
						sqlconn);
				gcmapping.defineIndexExpr(sqlstring, pie, gcmapping.SEPAND);
				sqlstring.append(";");
				break;
			case PIndexedElem.ELEM_UNMODIFIED:
				continue;
			default :
				throw new PExceptionProtocol(
					"Unknown modification status for PIndexedElem.");
			}
			if (elemblob != null) {
				PreparedStatement stmt
					= sqlconn.prepareStatement(sqlstring.toString());
				stmt.setBlob(1, elemblob);
				stmt.execute();
			}
			else {
				sqlconn.createStatement().executeUpdate(sqlstring.toString());
			}
		}
	}
}
