/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.interp.genclass.rdb.lib;

import org.objectweb.jorm.api.PAccessorGenClass;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.interp.genclass.api.FieldDesc;
import org.objectweb.jorm.interp.genclass.lib.GenClassMapping;
import org.objectweb.jorm.runtime.mapper.rdb.RdbBlob;
import org.objectweb.jorm.runtime.mapper.rdb.adapter.api.RdbAdpater;
import org.objectweb.jorm.interp.genclass.rdb.api.RdbFieldDesc;
import org.objectweb.jorm.interp.genclass.rdb.api.RdbGenClassProp;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.transaction.api.ConnectionWrapper;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

/**
 * This class provides the implementation of PClassMapping for RDBMS.
 * @author	P. Dchamboux
 */

public class RdbGenClassMapping extends GenClassMapping
	implements RdbGenClassProp {
	static final byte ID = 0x01;
	static final byte INDEX = 0x02;
	static final byte ELEM = 0x04;
	static final String SEPAND = " AND ";
	static final String SEPSCOL = ", ";
	static final String QUOTE = "'";
	/**
	 * The name of the table into which to store generic class object elements.
	 * It should be null in case of "legacy" mapping.
	 */
	protected String tableName = null;
	/**
	 * This is the type converter defined for a particular kind of RDB.
	 */
	protected RdbAdpater typeConverter = null;

	/**
	 * It creates a new PBinding. Not all PClassMapping need to support this
	 * method.
	 * @return	The new PBinding with lifecycle state set to
	 * 			LIFECYCLE_NOTBOUNDNOACC.
	 * @exception UnsupportedOperationException	Raised when the method is not
	 * 											supported by the PClassMapping.
	 */
	public PBinding createPBinding() throws UnsupportedOperationException {
		try {
			return new RdbGenClassBinding(this);
		}
		catch (PExceptionProtocol e) {
			return null;
		}
	}

	/**
	 * It returns an SQL connection from the connection received from the
	 * application, which should be either an actual SQL one or a
	 * ConnectionWrapper one.
	 * @param conn	The connection received from the application.
	 * @exception PExceptionIO	It is raised if no valid SQL connection can
	 * 							be returned to access to the RDB.
	 */
	Connection getConnection(Object conn) throws PExceptionIO {
		if (conn == null)
			throw new PExceptionIO("Needs a connection to access to the RDB");
		else if (conn instanceof Connection)
			return (Connection) conn;
		else if (conn instanceof ConnectionWrapper)
			return (Connection) ((ConnectionWrapper) conn).getNonCCIConnection();
		else
			throw new PExceptionIO("Wrong connection type to access to the RDB.");
	}

	/**
	 * It initializes the persistent structures required to map a JORM class
	 * within a particular Data Store.
	 * This method has no effect when the structures already exist, or when the
	 * structure must not be manipulated via JORM (e.g., case of a legacy DS).
	 * The deletion of these structures may be requested before initializing
	 * them.
	 * @param conn		The connection that can be used to access the DS.
	 * @param cleanup	Tells if structures and/or data should be removed at
	 * 					initialization time.
	 * @exception PExceptionIO	A problem occured during the initialization of
	 * 							the DS.
	 */
	public void initMappingStructures(Object conn, byte cleanup)
		throws PExceptionProtocol, PExceptionIO {
		Connection sqlconn = getConnection(conn);
		if (tableName == null)
			return;
		try {
			if (cleanup != CLEANUP_DONOTHING)
				removeMappingStructures(conn, cleanup == CLEANUP_REMOVEDATA);
			else if (sqlconn.getMetaData()
				.getTables(null, null, tableName, null).next())
				return;
			StringBuffer sqlstring = new StringBuffer("CREATE TABLE "
				+ tableName + " (");
			defineColumns(identifierFields, sqlstring, (short) 0);
			defineColumns(elemFields, sqlstring,
				(short) (elemFields[0].type == PTypeSpace.REFTOP ? 1 : 0));
			defineColumns(indexFields, sqlstring, (short) 0);
			sqlstring.append("PRIMARY KEY(");
			if (indexFields.length > 0) {
				//There is index fields, therefore the primary key is composed by
				// the genclass identifier and the index fields
				defineNames(indexFields, sqlstring,
					defineNames(identifierFields, sqlstring, true, (short) 0),
						(short) 0);
			} else {
				//There is no index, therefore the primary key is composed by
				// the genclass identifier and the element
				//The start of the elemFields depends on the element type.
				defineNames(elemFields, sqlstring,
					defineNames(identifierFields, sqlstring, true, (short) 0),
					(short) (elemFields[0].type.getTypeCode()==PType.TYPECODE_REFERENCE ? 1 : 0));
			}
			sqlstring.append("));");
			Statement stmt = sqlconn.createStatement();
			stmt.execute(sqlstring.toString());
			stmt.close();
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
	}

	/**
	 * It removes the persistent structures required to map a JORM class
	 * within a particular Data Store.
	 * This method has no effect when the structures do not exist, or when the
	 * structure must not be manipulated via JORM (e.g., case of a legacy DS).
	 * @param conn			The connection that can be used to access the DS.
	 * @param contentonly	It specifies if only the content of the persistent
	 * 						structures should be removed (it only removes data).
	 */
	public void removeMappingStructures(Object conn, boolean contentonly)
		throws PExceptionIO {
		Connection sqlconn = getConnection(conn);
		if (tableName == null)
			return;
		try {
			if (!sqlconn.getMetaData().getTables(null, null, tableName, null)
				.next())
				return;
			Statement stmt = sqlconn.createStatement();
			if (contentonly)
			// Remove all data only
				stmt.execute("DELETE FROM " + tableName + ";");
			else
			// Remove all data and the structures"
				stmt.execute("DROP TABLE " + tableName + ";");
			stmt.close();
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
	}

	/**
	 * It specifies if the given kind of PMapper is supported by this
	 * PClassMapping. In this case, all kinds of RDB mappers are supported.
	 * @param mappername	The name defining the PMapper kind.
	 * @return	true if this kind of mapper is supported.
	 */
	public boolean isConform(String mappername) {
		if (mappername.equals("Rdb"))
			return true;
		else
			return false;
	}

	/**
	 * It appends to the given sqlstring the definition of columns that
	 * corresponds to the given fields.
	 * @param fields	The fields for which the columns must be defined.
	 * @param sqlstring	The SQL string to which to append these column
	 * 					definitions.
	 * @param start		The first field from which to start for defining the
	 * 					columns.
	 */
	void defineColumns(FieldDesc[] fields, StringBuffer sqlstring, short start)
		throws PExceptionProtocol {
		RdbFieldDesc field;
		for (short i = start; i < fields.length; i++) {
			field = (RdbFieldDesc) fields[i];
			sqlstring.append(field.columnName);
			sqlstring.append(" ");
			sqlstring.append(typeConverter.getSqlType(field.type));
			sqlstring.append(", ");
		}
	}

	/**
	 * It appends to the given sqlstring the names of the columns that
	 * composes the primary key.
	 * @param fields	The fields for which the columns must belong to the
	 * 					primary key.
	 * @param sqlstring	The SQL string to which to append these column names.
	 * @param firstname	It specifies that the first name has not already been
	 * 					defined for the primary key.
	 * @return 	The returned boolean value specifies if a column name was
	 * 			already added into the sqlstring. The value depends on the
	 * 			firstname parameter and if column has been added during the
	 * 			method execution.
	 */
	boolean defineNames(FieldDesc[] fields, StringBuffer sqlstring,
						boolean firstname, short start) {
		RdbFieldDesc field;
		for (short i = start; i < fields.length; i++) {
			field = (RdbFieldDesc) fields[i];
			if (firstname)
				firstname = false;
			else
				sqlstring.append(", ");
			sqlstring.append(field.columnName);
		}
		return firstname;
	}

	/**
	 * It appends to the given sqlstring the names of all columns that
	 * composes the indexed element (some or all those that compose an element).
	 * @param sqlstring		The SQL string to which to append all column names.
	 * @param whichnames	It specifies the names to appends. This is an OR
	 * 						operation between the ID, ELEM, and INDEX flags.
	 */
	void defineAllNames(StringBuffer sqlstring, byte whichnames) {
		boolean firstname = true;
		if ((whichnames & ID) == ID) {
			firstname = defineNames(identifierFields, sqlstring, firstname,
				(short) 0);
		}
		if ((whichnames & ELEM) == ELEM) {
			//In case where the element is a reference to a class the elemFields
			// array contains at least two element which the first is the real
			// type of the element. Then the column description starts to the
			// index 1.
			firstname = defineNames(elemFields, sqlstring, firstname,
				(short) (elemFields.length==1 ? 0 : 1) );
		}
		if ((whichnames & INDEX) == INDEX) {
			firstname = defineNames(indexFields, sqlstring, firstname,
				(short) 0);
		}
	}

	/**
	 * It appends to the given sqlstring the expression for matching a GC
	 * identifier represented by the given PName.
	 * @param sqlstring	The involved SQL string.
	 * @param pname		The PName representing the identifier of the GC to
	 * 					access.
	 * @execption PExceptionProtocol	It is raised when an encoding type for
	 * 									one of the identifier fields is wrong.
	 * @execption PExceptionTyping		It is raised when an encoding type for
	 * 									one of the fields is wrong.
	 */
	void defineIdentifierExpr(StringBuffer sqlstring, PName pname,
							  String sep, Connection sqlconn)
		throws PExceptionProtocol, PExceptionTyping, PExceptionIO, PExceptionNaming {
		if (identifierFields.length == 1) {
			if (sep == SEPAND) {
				sqlstring.append(
					((RdbFieldDesc) identifierFields[0]).columnName);
				sqlstring.append(" = ");
			}
			defineNameFieldValue(sqlstring, (RdbFieldDesc) identifierFields[0],
				pname, sqlconn);
			return;
		}
		// The identifier is a composite name
		PNameGetter png = (PNameGetter) pname.encodeAbstract();
		for (short i = 0; i < identifierFields.length; i++) {
			if (i != 0)
				sqlstring.append(sep);
			if (sep == SEPAND) {
				sqlstring.append(
					((RdbFieldDesc) identifierFields[i]).columnName);
				sqlstring.append(" = ");
			}
			defineGetterFieldValue(sqlstring,
				(RdbFieldDesc) identifierFields[i], png, null);
		}
	}

	/**
	 * It appends to the given sqlstring the expression for defining the list
	 * of valu.
	 * @param conn		The connection that can be used to access the DS.
	 * @param sqlstring	The involved SQL string.
	 * @param pie		The PIndexedElem giving access to the index of this
	 * 					element.
	 * @param withfieldnameassign	If true, adds the column name to be assigned
	 * 								before each element value.
	 * @execption PExceptionProtocol	It is raised when an encoding type for
	 * 									one of the identifier fields is wrong.
	 * @execption PExceptionTyping		It is raised when an encoding type for
	 * 									one of the fields is wrong.
	 */
	Blob defineElemExpr(Object conn, StringBuffer sqlstring, PIndexedElem pie,
						boolean withfieldnameassign, String sep)
		throws PExceptionProtocol, PExceptionTyping, PExceptionNaming,
		PExceptionIO {
		Blob res = null;
		if (elemFields.length <= 2) {
			sqlstring.append(sep);
			if (withfieldnameassign)
				sqlstring.append(
					((RdbFieldDesc) elemFields[elemFields.length - 1]).columnName
					+ " = ");
		}
		switch (elemFields[0].type.getTypeCode()) {
		case PType.TYPECODE_BOOLEAN:
			sqlstring.append(pie.pieGetBooleanElem());
			break;
		case PType.TYPECODE_CHAR:
			sqlstring.append(pie.pieGetCharElem());
			break;
		case PType.TYPECODE_BYTE:
			sqlstring.append(pie.pieGetByteElem());
			break;
		case PType.TYPECODE_SHORT:
			sqlstring.append(pie.pieGetShortElem());
			break;
		case PType.TYPECODE_INT:
			sqlstring.append(pie.pieGetIntElem());
			break;
		case PType.TYPECODE_LONG:
			sqlstring.append(pie.pieGetLongElem());
			break;
		case PType.TYPECODE_FLOAT:
			sqlstring.append(pie.pieGetFloatElem());
			break;
		case PType.TYPECODE_DOUBLE:
			sqlstring.append(pie.pieGetDoubleElem());
			break;
		case PType.TYPECODE_STRING:
			sqlstring.append(QUOTE);
			sqlstring.append(pie.pieGetStringElem());
			sqlstring.append(QUOTE);
			break;
		case PType.TYPECODE_DATE:
			sqlstring.append(pie.pieGetDateElem());
			break;
		case PType.TYPECODE_SERIALIZED:
			sqlstring.append("?");
			try {
				res = new RdbBlob(pie.pieGetSerializedElem());
			}
			catch (IOException e) {
				throw new PExceptionIO(e);
			}
			break;
		case PType.TYPECODE_REFERENCE:
			if (elemFields.length == 2) {
				switch (elemFields[1].type.getTypeCode()) {
				case PType.TYPECODE_CHAR:
					if (elemCodingOnly)
						sqlstring.append(
							((PName)pie.pieGetRefElem()).encodeChar());
					else
						sqlstring.append(elemNamingContext.export(
							conn, pie.pieGetRefElem()).encodeChar());
					break;
				case PType.TYPECODE_BYTE:
					if (elemCodingOnly)
						sqlstring.append(
							((PName)pie.pieGetRefElem()).encodeByte());
					else
						sqlstring.append(elemNamingContext.export(
							conn, pie.pieGetRefElem()).encodeByte());
					break;
				case PType.TYPECODE_SHORT:
					if (elemCodingOnly)
						sqlstring.append(
							((PName)pie.pieGetRefElem()).encodeShort());
					else
						sqlstring.append(elemNamingContext.export(
							conn, pie.pieGetRefElem()).encodeShort());
					break;
				case PType.TYPECODE_INT:
					if (elemCodingOnly)
						sqlstring.append(
							((PName)pie.pieGetRefElem()).encodeInt());
					else
						sqlstring.append(elemNamingContext.export(
							conn, pie.pieGetRefElem()).encodeInt());
					break;
				case PType.TYPECODE_LONG:
					if (elemCodingOnly)
						sqlstring.append(
							((PName)pie.pieGetRefElem()).encodeLong());
					else
						sqlstring.append(elemNamingContext.export(
							conn, pie.pieGetRefElem()).encodeLong());
					break;
				case PType.TYPECODE_STRING:
					sqlstring.append(QUOTE);
					if (elemCodingOnly)
						sqlstring.append(
							((PName)pie.pieGetRefElem()).encodeString());
					else
						sqlstring.append(elemNamingContext.export(
							conn, pie.pieGetRefElem()).encodeString());
					sqlstring.append(QUOTE);
					break;
				default :
					throw new PExceptionProtocol(
						"Wrong type for an reference element field of a GenClass");
				}
			}
			else {
				// This is a composite reference
				PNameGetter png;
				if (elemCodingOnly)
					png = (PNameGetter)
						((PName)pie.pieGetRefElem()).encodeAbstract();
				else
					png = (PNameGetter)
						elemNamingContext.export(conn, pie.pieGetRefElem())
						.encodeAbstract();
				for (int i = 1; i < elemFields.length; i++) {
					sqlstring.append(sep);
					if (withfieldnameassign) {
						sqlstring.append(
							((RdbFieldDesc) elemFields[i]).columnName);
						sqlstring.append(" = ");
					}
					defineGetterFieldValue(sqlstring,
						(RdbFieldDesc) elemFields[i], png, null);
				}
			}
			break;
		default :
			throw new PExceptionProtocol(
				"Wrong type for an element field of a GenClass");
		}
		return res;
	}

	/**
	 * It appends to the given sqlstring the expression for matching a GC
	 * identifier represented by the given PName.
	 * @param sqlstring	The involved SQL string.
	 * @param pie		The PIndexedElem giving access to the index of this
	 * 					element.
	 * @return			True if no index has been added.
	 * @execption PExceptionProtocol	It is raised when an encoding type for
	 * 									one of the identifier fields is wrong.
	 * @execption PExceptionTyping		It is raised when an encoding type for
	 * 									one of the fields is wrong.
	 */
	boolean defineIndexExpr(StringBuffer sqlstring, PIndexedElem pie,
						 String sep)
		throws PExceptionProtocol, PExceptionTyping {
		for (short i = 0; i < indexFields.length; i++) {
			sqlstring.append(sep);
			if (sep == SEPAND) {
				sqlstring.append(((RdbFieldDesc) indexFields[i]).columnName);
				sqlstring.append(" = ");
			}
			defineGetterFieldValue(sqlstring, (RdbFieldDesc) indexFields[i],
				null, pie);
		}
		return indexFields.length == 0;
	}

	/**
	 * It appends to the given sqlstring the value of a field accessing through
	 * a getter. This can be either fields of an index of a PIndexedElem, or
	 * fields of a PNameGetter representing a composite name.
	 * @param sqlstring	The involved SQL string.
	 * @param rdf		The descriptor of the field.
	 * @param png		The PNameGetter for accessing the values of the fields
	 * 					of a composite name. It is null if the involved getter
	 * 					is the one of an PIndexedElem.
	 * @param pie		The PIndexedElem being the getter for the index fields
	 * 					of a PIndexedElem. It is null if the involved getter
	 * 					is PName one.
	 * @execption PExceptionProtocol	It is raised when a problem occurs
	 * 									while accessing the relevant field
	 * 									value.
	 * @execption PExceptionTyping		It is raised when an encoding type for
	 * 									one of the fields is wrong.
	 */
	void defineGetterFieldValue(StringBuffer sqlstring, RdbFieldDesc rfd,
								PNameGetter png, PIndexedElem pie)
		throws PExceptionProtocol, PExceptionTyping {
		if (((png == null) && (pie == null))
			|| ((png != null) && (pie != null)))
			throw new PExceptionProtocol(
				"One of the PNG or PIE paramter should be defined exclusively from each other");
		try {
			switch (rfd.type.getTypeCode()) {
			case PType.TYPECODE_BYTE:
				sqlstring.append((int) ((png != null)
					? png.pngetByteField(rfd.name, null)
					: pie.pieGetByteIndexField(rfd.name)));
				break;
			case PType.TYPECODE_CHAR:
				sqlstring.append((png != null)
					? png.pngetCharField(rfd.name, null)
					: pie.pieGetCharIndexField(rfd.name));
				break;
			case PType.TYPECODE_SHORT:
				sqlstring.append((int) ((png != null)
					? png.pngetShortField(rfd.name, null)
					: pie.pieGetShortIndexField(rfd.name)));
				break;
			case PType.TYPECODE_INT:
				sqlstring.append((png != null)
					? png.pngetIntField(rfd.name, null)
					: pie.pieGetIntIndexField(rfd.name));
				break;
			case PType.TYPECODE_LONG:
				sqlstring.append((png != null)
					? png.pngetLongField(rfd.name, null)
					: pie.pieGetLongIndexField(rfd.name));
				break;
			case PType.TYPECODE_STRING:
				sqlstring.append(QUOTE);
				sqlstring.append((png != null)
					? png.pngetStringField(rfd.name, null)
					: pie.pieGetStringIndexField(rfd.name));
				sqlstring.append(QUOTE);
				break;
			default :
				throw new PExceptionTyping(
					"Wrong field type for an identifier PName");
			}
		}
		catch (PExceptionIO e) {
			throw new PExceptionProtocol(
				"Should never raise an IO exception here!!");
		}
	}

	/**
	 * It appends to the given sqlstring the value of a field defined by a
	 * PName.
	 * @param sqlstring	The involved SQL string.
	 * @param rdf		The descriptor of the field.
	 * @param pname		The PName representing the identifier of the GC to
	 * 					access.
	 * @execption PExceptionTyping		It is raised when an encoding type for
	 * 									one of the identifier fields is wrong.
	 */
	void defineNameFieldValue(StringBuffer sqlstring, RdbFieldDesc rfd,
							  PName pname, Connection sqlconn)
		throws PExceptionTyping, PExceptionIO, PExceptionNaming {
		switch (rfd.type.getTypeCode()) {
		case PType.TYPECODE_BYTE:
			sqlstring.append((int) pname.encodeByte());
			break;
		case PType.TYPECODE_CHAR:
			sqlstring.append(pname.encodeChar());
			break;
		case PType.TYPECODE_SHORT:
			sqlstring.append((int) pname.encodeShort());
			break;
		case PType.TYPECODE_INT:
			sqlstring.append(pname.encodeInt());
			break;
		case PType.TYPECODE_LONG:
			sqlstring.append(pname.encodeLong());
			break;
		case PType.TYPECODE_STRING:
			sqlstring.append(QUOTE);
			sqlstring.append(pname.encodeString());
			sqlstring.append(QUOTE);
			break;
		default :
			throw new PExceptionTyping(
				"Wrong field type for an identifier PName");
		}
	}

	/**
	 * It tests if the current rwo of this result corresponds to the marker
	 * of empty generic class instance. This marker is defined by the fact that
	 * values of all fields of the GC element are null.
	 * @param rs	The ResultSet that points to the first row of those that
	 * compose the value of a generic class instance.
	 * @return	true if this is the marker of an empty generic class instance.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							accessing the SQL ResultSet.
	 */
	boolean testEmptyGcMarker(ResultSet rs) throws PExceptionIO {
		boolean notmarker = false;
		try {
			for (short j = 0; j < elemFields.length; j++) {
				switch (elemFields[j].type.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
					rs.getBoolean(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_CHAR:
					rs.getInt(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_BYTE:
					rs.getByte(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_SHORT:
					rs.getShort(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_INT:
					rs.getInt(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_LONG:
					rs.getLong(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_FLOAT:
					rs.getFloat(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_DOUBLE:
					rs.getDouble(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_STRING:
					rs.getString(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_DATE:
					rs.getDate(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				case PType.TYPECODE_SERIALIZED:
					rs.getClob(j + 1);
					if (!rs.wasNull())
						return false;
					break;
				default :
				}
			}
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		return true;
	}

	/**
	 * It passes the values of the index and the generic class element through
	 * a PIndexedElem to the accessor associated to this generic class instance.
	 * @param rs		The ResultSet that points to the current indexed element
	 * 					to add to the MI through the given PAccessorGenClass
	 * 					accessor.
	 * @param accessor	The relevant generic class accessor associated with a
	 * 					valid MI.
	 * @param png		The PNameGetter that gives access to the values of the
	 * 					fields of a composite name, when the elements of this
	 * 					generic class instance are composite references. It is
	 * 					null if it is not a composite name.
	 * @exception PExceptionProtocol	It is raised if there is a problem with
	 * 									the DSI structure with respect to the MI
	 * 									one.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							accessing the SQL ResultSet.
	 */
	void passValuesToAccessor(Object conn, ResultSet rs,
							  PAccessorGenClass accessor, PNameGetter png)
		throws PExceptionProtocol, PExceptionIO, PExceptionTyping,
		PExceptionNaming {
		PIndexedElem pie = accessor.createPIndexedElem();
		if (Debug.ON && logger!=null) logger.log(BasicLevel.DEBUG,"");
		try {
			// The begining of the index depends of the size of the element:
			// If the element is the primitive element then the size is 1
			// If the element is a reference field then the first element of the
			// elemFields array is just a tag.
			short indexStart = (short)
				(elemFields.length==1 ? 1 : elemFields.length-1);

			for (short i = 0; i < indexFields.length; i++) {
				if (Debug.ON && logger!=null)
					logger.log(BasicLevel.DEBUG,
						"index["+i+"]:" + indexFields[i].name);
				switch (indexFields[i].type.getTypeCode()) {
				case PType.TYPECODE_CHAR:
					pie.pieSetCharIndexField(indexFields[i].name,
						(char) rs.getInt(i + indexStart));
					break;
				case PType.TYPECODE_BYTE:
					pie.pieSetByteIndexField(indexFields[i].name,
						rs.getByte(i + indexStart));
					break;
				case PType.TYPECODE_SHORT:
					pie.pieSetShortIndexField(indexFields[i].name,
						rs.getShort(i + indexStart));
					break;
				case PType.TYPECODE_INT:
					pie.pieSetIntIndexField(indexFields[i].name,
						rs.getInt(i + indexStart));
					break;
				case PType.TYPECODE_LONG:
					pie.pieSetLongIndexField(indexFields[i].name,
						rs.getLong(i + indexStart));
					break;
				case PType.TYPECODE_STRING:
					pie.pieSetStringIndexField(indexFields[i].name,
						rs.getString(i + indexStart));
					break;
				default :
					throw new PExceptionProtocol(
						"Wrong type for an index field of a GenClass");
				}
			}
			if (Debug.ON && logger!=null)
				logger.log(BasicLevel.DEBUG,
					"element type:" + elemFields[0].type.getJormName());
			switch (elemFields[0].type.getTypeCode()) {
			case PType.TYPECODE_BOOLEAN:
				pie.pieSetBooleanElem(rs.getBoolean(1));
				break;
			case PType.TYPECODE_CHAR:
				pie.pieSetCharElem((char) rs.getInt(1));
				break;
			case PType.TYPECODE_BYTE:
				pie.pieSetByteElem(rs.getByte(1));
				break;
			case PType.TYPECODE_SHORT:
				pie.pieSetShortElem(rs.getShort(1));
				break;
			case PType.TYPECODE_INT:
				pie.pieSetIntElem(rs.getInt(1));
				break;
			case PType.TYPECODE_LONG:
				pie.pieSetLongElem(rs.getLong(1));
				break;
			case PType.TYPECODE_FLOAT:
				pie.pieSetFloatElem(rs.getFloat(1));
				break;
			case PType.TYPECODE_DOUBLE:
				pie.pieSetDoubleElem(rs.getDouble(1));
				break;
			case PType.TYPECODE_STRING:
				pie.pieSetStringElem(rs.getString(1));
				break;
			case PType.TYPECODE_DATE:
				pie.pieSetDateElem((java.util.Date) rs.getDate(1));
				break;
			case PType.TYPECODE_SERIALIZED:
				pie.pieSetSerializedElem((Serializable)
					(new ObjectInputStream(rs.getBlob(1).getBinaryStream()))
					.readObject());
				break;
			case PType.TYPECODE_REFERENCE:
				if (Debug.ON && logger!=null)
					logger.log(BasicLevel.DEBUG,
						"reference size:" + (elemFields.length - 1));
				if (elemFields.length == 2)
					switch (elemFields[1].type.getTypeCode()) {
					case PType.TYPECODE_CHAR:
						if (elemCodingOnly)
							pie.pieSetRefElem(
								elemNamingContext
								.decodeChar((char) rs.getInt(1)));
						else
							pie.pieSetRefElem(
								elemNamingContext
								.decodeChar((char) rs.getInt(1)).resolve(conn));
						break;
					case PType.TYPECODE_BYTE:
						if (elemCodingOnly)
							pie.pieSetRefElem(
								elemNamingContext.decodeByte(rs.getByte(1)));
						else
							pie.pieSetRefElem(
								elemNamingContext.decodeByte(rs.getByte(1))
								.resolve(conn));
						break;
					case PType.TYPECODE_SHORT:
						if (elemCodingOnly)
							pie.pieSetRefElem(
								elemNamingContext.decodeShort(rs.getShort(1)));
						else
							pie.pieSetRefElem(
								elemNamingContext.decodeShort(rs.getShort(1))
								.resolve(conn));
						break;
					case PType.TYPECODE_INT:
						if (elemCodingOnly)
							pie.pieSetRefElem(
								elemNamingContext.decodeInt(rs.getInt(1)));
						else
							pie.pieSetRefElem(
								elemNamingContext.decodeInt(rs.getInt(1))
								.resolve(conn));
						break;
					case PType.TYPECODE_LONG:
						if (elemCodingOnly)
							pie.pieSetRefElem(
								elemNamingContext.decodeLong(rs.getLong(1)));
						else
							pie.pieSetRefElem(
								elemNamingContext.decodeLong(rs.getLong(1))
								.resolve(conn));
						break;
					case PType.TYPECODE_STRING:
						if (elemCodingOnly)
							pie.pieSetRefElem(
								elemNamingContext.decodeString(rs.getString(1)));
						else
							pie.pieSetRefElem(
								elemNamingContext.decodeString(rs.getString(1))
								.resolve(conn));
							break;
					default :
						throw new PExceptionProtocol(
							"Wrong type for an reference element field of a GenClass");
					}
				else if (elemCodingOnly)
					pie.pieSetRefElem(elemNamingContext.decodeAbstract(png, null));
				else
					pie.pieSetRefElem(elemNamingContext.decodeAbstract(png,null).resolve(conn));
				break;
			default :
				throw new PExceptionProtocol(
					"Wrong type for an element field of a GenClass");
			}
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		catch (IOException e) {
			throw new PExceptionIO(e);
		}
		catch (ClassNotFoundException e) {
			throw new PExceptionIO(e);
		}
		accessor.paAdd(pie, conn);
	}

	/**
	 * It retrieves a PNameGetter that gives access to the values of the
	 * fields of a composite name, when the elements of this generic class
	 * instance are composite references. It yields null if it is not a
	 * composite name.
	 * @return	The allocated PNameGetter.
	 */
	PNameGetter getPNameGetter(ResultSet rs) {
		if (elemFields.length > 2)
			return new RdbGenClassNameGetter(rs, elemFields);
		return null;
	}

	/**
	 * Defines the table name property of a generic class instance mapped to
	 * a RDB.
	 * @param pt	The associated type.
	 */
	public void setTableName(String tn) {
		tableName = tn;
	}

	/**
	 * Defines the type converter property of a generic class instance mapped to
	 * a RDB.
	 * @param pt	The associated type.
	 */
	public void setTypeConverter(RdbAdpater tc) {
		typeConverter = tc;
	}
}
