/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.interp.genclass.rdb.lib;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.interp.genclass.api.FieldDesc;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;

import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * This is a class that implements the PNameGetter interface and that provides
 * a generic access based on the description of fields of the element of the
 * relevant generic class instance.
 * @author	P. Dchamboux
 */

class RdbGenClassNameGetter implements PNameGetter {
	/**
	 * The ResultSet pointing to the row from which to fetch the value of the
	 * fields of a composite name.
	 */
	private ResultSet resultSet;
	/**
	 * The description of the fields of a composite name.
	 */
	private FieldDesc[] elemFields;

	/**
	 * Constructs a getter that enables to retrieve the values of the fields
	 * of a composite name.
	 */
	RdbGenClassNameGetter(ResultSet rs, FieldDesc[] rfds) {
		resultSet = rs;
		elemFields = rfds;
	}

	/**
	 * It retrieves the value of a byte field.
	 * @param fn	A String containing the name of the field.
     * @param context A context used by PNameGetter to extract the related data.
	 * @return	The byte value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public byte pngetByteField(String fn, Object ctx)
		throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].name))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_BYTE)
						return resultSet.getByte(i);
					else
						break;
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		throw new PExceptionTyping("Wrong composite name field in a GenClass.");
	}

	/**
	 * It retrieves the value of a char field.
	 * @param fn	A String containing the name of the field.
     * @param context A context used by PNameGetter to extract the related data.
	 * @return	The char value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public char pngetCharField(String fn, Object ctx)
		throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].name))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_CHAR)
						return (char) resultSet.getInt(i);
					else
						break;
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		throw new PExceptionTyping("Wrong composite name field in a GenClass.");
	}

	/**
	 * It retrieves the value of a short field.
	 * @param fn	A String containing the name of the field.
     * @param context A context used by PNameGetter to extract the related data.
	 * @return	The short value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public short pngetShortField(String fn, Object ctx)
		throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].name))
					if (elemFields[i].type.getTypeCode()
						== PType.TYPECODE_SHORT)
						return resultSet.getShort(i);
					else
						break;
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		throw new PExceptionTyping("Wrong composite name field in a GenClass.");
	}

	/**
	 * It retrieves the value of an int field.
	 * @param fn	A String containing the name of the field.
     * @param context A context used by PNameGetter to extract the related data.
	 * @return	The int value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public int pngetIntField(String fn, Object ctx)
		throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].name))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_INT)
						return resultSet.getInt(i);
					else
						break;
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		throw new PExceptionTyping("Wrong composite name field in a GenClass.");
	}

    /**
	 * It retrieves the value of an int field.
	 * @param fn	A String containing the name of the field.
     * @param context A context used by PNameGetter to extract the related data.
	 * @return	The int value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public Integer pngetIntObjField(String fn, Object ctx)
		throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].name))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_INT_OBJ)
                        return (Integer) resultSet.getObject(i);
					else
						break;
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		throw new PExceptionTyping("Wrong composite name field in a GenClass.");
	}

	/**
	 * It retrieves the value of a long field.
	 * @param fn	A String containing the name of the field.
     * @param context A context used by PNameGetter to extract the related data.
	 * @return	The long value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public long pngetLongField(String fn, Object ctx)
		throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].name))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_LONG)
						return resultSet.getLong(i);
					else
						break;
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		throw new PExceptionTyping("Wrong composite name field in a GenClass.");
	}

	/**
	 * It retrieves the value of a string field.
	 * @param fn	A String containing the name of the field.
     * @param context A context used by PNameGetter to extract the related data.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public String pngetStringField(String fn, Object ctx)
		throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].name))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_STRING)
						return resultSet.getString(i);
					else
						break;
		}
		catch (SQLException e) {
			throw new PExceptionIO(e);
		}
		throw new PExceptionTyping("Wrong composite name field in a GenClass.");
	}
}
