/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.api;

import org.objectweb.jorm.type.api.PType;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * A Class interface defines methods to describe a Class meta object.
 * A Class is composed by a set of fields (primitive, reference class or
 * reference generic class), a set of inherited classes, a mapping class,
 * can be abstract or not and has a name.
 * @author	X. Spengler
 */
public interface Class extends MetaObject {

    /**
     * Allows to know if the current class is an abstract class or not.
     *
     * @return	true, if the current class is an abstract class, and false,
     *          if the current class is not abstract
     */
    boolean isAbstract();

    /**
     * Allows to know the name of the current class object.
     * @return	the string representation of the class name
     */
    String getName();

    /**
     * Allows to obtain an existing Field.
     * A TypedElement is a class field, and this method finds a field with its
     * name (fieldName). This method is a generic method to return a
     * TypedElement object.
     *
     * @param fieldName	the name of the class field
     * @return	a field object. If the field does not exist, null is returned.
     */
    TypedElement getTypedElement(String fieldName);

    /**
     * Creates a new NameDef object for the current Class object.
     * Here we can define if the namedef is defined outside jorm or not.
     * In the case where the namedef is already defined for the current class,
     * the old one is used and returned.
     *
     * @param nameDefName   the name of the nameDef read in the persistence
     *                      description file
     * @param isSystem	true, the namedef is defined outside jorm, and false,
     *                  the namedef is define inside jorm
     * @return	a new object used to describe the PName projection for
     *          the current Class
     */
    NameDef createNameDef();

    /**
     * Creates an hiddenfield for the current Class object.
     * An hidden field is used to map the pname of the class or the pname of a ref field.
     * An hidden field is a scalarfield.
     *
     * @param fieldname     the name of the hiddenfield
     * @param type	        the ptype of the hiddenfield
     * @return	a new scalarfield object used to describe the hiddenfield
     */

    ScalarField createHiddenField(String fieldName, PType type);

    /**
     * Provides an iterator over the all the hiddenfield of the class including inherited fields.
     * @return the iterator embeding scalarfield object describing the hiddenfields.
     */
    Iterator iterateAllHiddenField();

    /**
     * Provides an iterator over the all the hiddenfield of the class excluding inherited fields.
     * @return the iterator embeding scalarfield object describing the hiddenfields.
     */
    Iterator iterateHiddenField();

    /**
     * Retrieve the hiddenfield identified by its name in the current class.
     * @param fieldname the name of the hiddenfield.
     * @return the scalarfield object describing the hiddenfield or null if the field does not exist in the class.
     */
    ScalarField getHiddenField(String fieldName);


    /**
     * Returns an iterator on existing NameDef for the current Class.
     * This method returns an empty iterator if no namedef exists in this Class.
     *
     * @return	an iterator on NameDef, or an empty iterator if no namedef is
     *          defined
     */
    Iterator iterateNameDef();

    /**
     * Returns a NameDef object from a given mapperName and mappingName.
     * If the namedef does not exist, null is returned.
     *
     * @param mappingName   the name of the mapping
     * @return  an existing namedef object if exists, either null
     */
    NameDef getNameDef(String mappingName);

    /**
     * Allows to obtain a new PrimitiveElement from a class field name an its
     * type.
     * A PrimitiveElement is a Class field. If the class field already exists,
     * it is returned.
     *
     * @param fieldName	the name of the primitive field
     * @param type      the PType of the primitive field
     * @return	a new PrimitiveElement object, or the existing one if already
     *          defined
     */
    PrimitiveElement createPrimitiveElement(String fieldName, PType type);

    /**
     * Allows to obtain a new ClassRef from a field name an its class.
     * A ClassRef object is a field which is represented by a reference to a
     * Class. If the field already exists, it is returned.
     *
     * @param fieldName	the name of the field
     * @param className	the string representation of the class which represents
     *                  the reference of the field
     * @return	a new ClassRef object, or the existing one if already defined
     */
    ClassRef createClassRef(String fieldName, Class class_);

    /**
     * Allows to obtain a new GenClassRef from a field name an its generic
     * class.
     * A GenClassRef object is a field which is represented by a
     * reference to a generic class. If the field already exists,
     * it is returned.
     *
     * @param fieldName	the name of the field
     * @param genName	the GenClass object which represents the reference of
     *                  the field
     * @return	a new GenClassRef object, or the existing one if already
     *          defined
     */
    GenClassRef createGenClassRef(String fieldName, String genName);

    /**
     * Returns an iterator on existing class fields.
     * If no field is defined, an empty iterator is returned.
     * This iterator contains PrimitiveElements, ClassRef and GenClassRef
     * objects. The top level object is TypedElement.
     *
     * @return	an iterator on fields, or null if there is no defined field
     */
    Iterator iterateField();

    /**
     * Returns an iterator on existing class fields + fields from super classes.
     * This iterator contains PrimitiveElement, ClassRef and GenClassRef
     * objects.
     *
     * @return	an iterator on all defined fields, plus fields from super
     */
    Iterator iterateAllField();

    /**
     * Allows to obtain all the inherited classes from the current class object.
     * This iterator contains Class objects. In the case where there is no
     * inherited class, an empty iterator is returned.
     *
     * @return	an iterator on inherited classes (Class objets). If no object
     *          exists, an empty iterator is returned.
     */
    Iterator iterateInheritedClass();

    /**
     * Adds a new inherited class to the list of the inherited classes for the
     * current class object.
     * If the class already exists, nothing is done.
     *
     * @param class_	the inherited class to add to inherited classes list
     */
    void addInheritedClass(Class class_);

    /**
     * Returns the number of inherited classes for the current Class object.
     *
     * @return  the number of inherited classes
     */
    int getInheritedClassNumber();

    /**
     * Returns an existing ClassMapping object for the current class object.
     * If the ClassMapping object does not exist, null is returned.
     *
     * @param mapperName	the name of the mapper class name
     * @return	a ClassMapping object. If the ClassMapping does not exist for the
     *          mapper, null is returned.
     */
    ClassMapping getClassMapping(String mapperName, String mappingName);

    /**
     * Creates a new ClassMapping object for the current class.
     * This mapping is defined by the name of the mapper, and the name of
     * the mapping. If the ClassMapping object already exists, the existing
     * one is returned.
     *
     * @param mapperName	the name of the class mapper
     * @param mappingName	the name of the class mapping
     * @return	a new ClassMapping object. This object is dependent from
     *	    	the mapper. In the case the ClassMapping exists, it is returned.
     */
    ClassMapping createClassMapping(String mapperName, String mappingName);

    /**
     * Allows to know all the ClassMapping object used for the current class.
     * If no ClassMapping object are defined, an empty iterator is returned.
     *
     * @return	an Iterator object on ClassMapping object. If there is no
     *          object, an empty iterator is returned.
     */
    Iterator iterateClassMapping();

    /**
     * Constructs an Hashtable with inherited class
     * @param table the hashtable to build
     */
    void addAllInheritedClass(Hashtable table);

    /**
     * Returns the number of fields which are references to classes.
     *
     * @return  the number of fieldRef
     */
    int getFieldRefNumber();

    /**
     * Specifies that this class is inherited by another one within the
     * set of compiled ones.
     */
    void setInherited();

    /**
     * Tests if this class is inherited by another one.
     */
    boolean isInherited();

}
