/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.api;

import org.objectweb.jorm.type.api.PType;

import java.util.Iterator;

/**
 * A GenClassRef is an interface which describes method to define
 * a class field which can be a reference to a generic class.
 * A GenClassRef object is composed by a PrimitiveElement (a primitive type),
 * or a ClassRef (a reference to a Class), or an another GenClassRef (a
 * reference to a generic class.
 * A recursion is defined when a GenClassRef object is composed by an other
 * GenClassRef object. A final object of the recursion is a PrimitiveElement
 * or a ClassRef objects.
 * @author	X. Spengler
 */
public interface GenClassRef extends Reference {
	/**
	 * Sets the type of the GenClassRef with a PrimitiveElement object.
	 *
	 * @param primitiveElement	the primitive element which defines the current
	 *                          GenClassRef object. This object is the final
	 *                          object of the recursion.
	 */
	void setPrimitiveElement(PrimitiveElement primitiveElement);

	/**
	 * Sets the type of the GenClassRef with a ClassRef object.
	 *
	 * @param class_	the reference to the Class object which defines the
	 *                  current GenClassRef object. This object is the final
	 *                  object of the recursion.
	 */
	void setClassRef(ClassRef class_);

	/**
	 * Creates a new GenClassRef object.
	 * This method must be used when the recursion is not finished (we are on
	 * not final object).
	 *
	 * @param genClassRefName	the name of the generic class used to create a
	 *                          new GenClassRef object
	 * @return	a new GenClassRef object
	 */
	GenClassRef createGenClassRef(String genClassRefName);

	/**
	 * Returns the PrimitiveElement object which defines the current GenClassRef
	 * object.
	 * This method is used after a call to isPrimitive() method, to determine
	 * if the current object is a "basic" type.
	 *
	 * @return	the current type of the field. This method always returns a
	 *          PType object, and can not return null value because of the
	 *          previous test.
	 */
	PrimitiveElement getPrimitiveElement();

	/**
	 * Returns the ClassRef object which defines the current GenClassRef object.
	 * This method is used after a call to isClassRef() method, to determine if
	 * the current object is defined by a final class or not.
	 *
	 * @return	the final Class which describes the type of the object which
	 *          defines the GenClassRef object. This method always returns a
	 *          Class, and can not return null value because of the previous
	 *          test.
	 */
	ClassRef getClassRef();

	/**
	 * Returns the GenClassRef object which defines the current GenClassRef
	 * object.
	 * This object exists when it is in a recursion. This method is used after a
	 * call to isGenClassRef() method, to determine if the current object is a
	 * reference to a generic class or not.
	 *
	 * @return	the underlying reference to the generic class. This method
	 *          always returns a GenClassRef, and can not return null value
	 *          because of the previous test.
	 */
	GenClassRef getGenClassRef();

	/**
	 * Allows to know if the current object is primitive or not (final and
	 * "basic").
	 *
	 * @return	true, if the current object is primitive, else false, if the
	 *          object is not primitive (either a reference to a Class, or
	 *		    a reference to a generic class).
	 */
	boolean isPrimitive();

	/**
	 * Allows to know if the current object is a reference to a Class or not.
	 *
	 * @return	true, if the object is a reference to a Class, else false, if
	 *          the object is not a reference to a Class (either a primitive
	 *          object, or a reference to a generic class).
	 */
	boolean isClassRef();

	/**
	 * Allows to know if the current object is a reference to a generic class
	 * or not.
	 *
	 * @return	true, if the object is a reference to a generic class, else
	 *	      	false, if the object is not a reference to a generic class
	 *		    (either a primitive object, or a reference to a class).
	 */
	boolean isGenClassRef();

	/**
	 * Creates a new NameDef object for the new GenClassRef object.
	 * If the specified namedef object already exists, it is returned, else
	 * a new NameDef object is created.
	 *
	 * @param nameDefName   the name of the nameDef read in the persistence
	 *                      description file
	 * @param isSystem	    true, the namedef is defined outside jorm, else
	 *          			false, the namedef is define inside jorm
	 * @return	a new object used to describe the PName projection for
	 *	    	the current Class, or an existing object if already defined
	 */
	NameDef createRefNameDef();

	/**
	 * Returns an iterator on existing NameDef of GenClassRef object.
	 * If no NameDef exists, an empty iterator is returned.
	 *
	 * @return	an iterator on NameDef object
	 */
	Iterator iterateRefNameDef();

	/**
	 * Returns a NameDef object from a given mapperName and mappingName.
	 * If the corresponding NameDef does not exist, null is returned.
	 *
	 * @param mapperName    the name of the mapper
	 * @param mappingName   the name of the mapping
	 * @return  an existing namedef object if exists, either null
	 */
	NameDef getRefNameDef(String mappingName);

    /**
	 * Returns the underlying GenClass object which describes the generic class
	 * object used.
	 *
	 * @return	the underlying generic class object
	 */
	GenClass getGenClass();

	/**
	 * Returns an existing ClassMapping object for the current reference object
	 * from a given mapper name and mapping name.
	 *
	 * @param mapperName	the name of the mapper reference name
	 * @param mappingName   the name of the mapping reference name
	 * @return	a ClassMapping object. If the ClassMapping does not exist for
	 *          the	mapper, null is returned.
	 */
	ClassMapping getClassMapping(String mapperName, String mappingName);

	/**
	 * Creates a new ClassMapping object for the current reference.
	 * This mapping is defined by the name of the mapper, and the name of the
	 * mapping.
	 *
	 * @param mapperName	the name of the reference mapper
	 * @param mappingName	the name of the reference mapping
	 * @return	a new ClassMapping object. This object is dependent from
	 *		the mapper.
	 */
	ClassMapping createClassMapping(String mapperName, String mappingName);

	/**
	 * Allows to know all the ClassMapping objects used for the current
	 * reference.
	 * If no ClassMapping is defined, an empty iterator is returned.
	 *
	 * @return	an iterator object on ClassMapping object. If there is no
	 *          object,	null is returned.
	 */
	Iterator iterateClassMapping();

	/**
	 * Returns the name of the GenClass object.
	 *
	 * @returns the name of the GenClassRef object
	 */
	String getGenClassName();

     /**
	 * Create a new hidden field for the refgenclass.
	 *
	 * @returns the ScalarField
	 */
    ScalarField createHiddenField(String fieldName, PType type);

    /**
	 * return an iterator over the hiddenfield of the refgenclass definition.
	 *
	 * @returns the iterator containing the hiddenfield.
	 */
    Iterator iterateHiddenField();

    /**
	 * retrieve an hiddenfield from its name.
	 *
	 * @returns the hiddenfield.
	 */
    ScalarField getHiddenField(String fieldName);
}
