/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.api;

import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.util.lib.BasicClassLoader;

import java.util.Iterator;
import java.util.Enumeration;

/**
 * The Manager object is the main interface to create the meta information.
 * This object is able to register mapping factories, and to create schema
 * objects.
 * @author	X. Spengler
 */
public interface Manager extends MetaObject {
    /**
     * Initializes the current Manager.
     * This method replaces the constructor method, with a single parameter,
     * which is a class loader.
     *
     * @param classloader   a single class loader to load class into the
     *                      meta information
     * @deprecated Basic ClassLoader is not a classloader. Need refactoring (Christophe Ney)
     */
    void init(BasicClassLoader classloader);

    void init();

	/**
	 * Returns an existing schema.
	 * If no Schema corresponds to the given name, null is returned.
	 *
	 * @param schemaName	the name of the schema
	 * @return	a Schema object. If the schema does not exist, null is returned.
	 */
	Schema getSchema(String schemaName);

	/**
	 * Returns a new schema.
	 * This method is the factory method to create schema object. If the
	 * schemaName schema already exists, it is returned.
	 *
	 * @param schemaName	the name of the schema
	 * @return	a new schema object, or the existing one if already defined
	 */
	Schema createSchema(String schemaName);

	/**
	 * Allows to know all the declared schemas into the current metainformation
	 * manager.
	 * This iterator contains Schema object.
	 *
	 * @return	an iterator on registered Schema object. If there is no schema,
	 *		    an empty iterator is returned.
	 */
	Iterator iterateSchema();

    Enumeration enumSchema();

	/**
	 * Adds a mapping factory to the current metainformation manager.
	 * A mapping factory is a factory to map object (class, field, ...).
	 *
	 * @param mapperName	the name of the mapper (i.e: OR for rdb)
	 * @param mappingFactory	the mapping factory to add to the list of
	 *                          existing mapping factories
	 */
	void addMappingFactory(String mapperName, String mappingFactory);

	/**
	 * Returns a mapping factory corresponding to a mapper name.
	 * If no MappingFactory corresponds to the mapper name, null is returned.
	 *
	 * @param mapperName	the name of the mapper to obtain
	 * @return	a mapping factory. If the mapping factory does not exist for
	 *          the mapper, null is returned.
	 */
	MappingFactory getMappingFactory(String mapperName);

	/**
	 * Searches a class into all the declared schemas and returns it.
	 * If no class corresponds to this name, null is returned.
	 *
	 * @param className	the name of the class to search
	 * @return	the Class object found, null if not found
	 */
	Class getClass(String className);


    /**
     *  Build an iterator to iterate all classes in the meta information
     * (composite names are excluded)
     */
    Iterator getAllClassIterator();

	/**
	 * Returns the PTypeSpace structure.
	 */
	PTypeSpace getPTypeSpace();
}
