/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.api;

import java.util.Iterator;

/**
 * Schema object defines an object able to gather a set of classes.
 * It is a similar approach to the package keyword from java.
 * @author	X. Spengler
 */
public interface Schema extends MetaObject {
    /**
     * Returns the name of the schema.
     *
     * @return	the string representation of the name of the schema.
     *		    Null is returned if there is no schema name else a not empty
     *          dotted string.
     */
    String getName();

    /**
     * Returns a Class created with a class name.
     *
     * @param className	the string representation of the class name
     * @return	        a Class corresponding to its name. If the Class does
     *                  not exist, null is returned.
     */
    Class getClass(String className);

    /**
     * Returns an hidden class.
     * An Hidden class is a class which is not yet totally defined. If the class
     * is not found into the current schema, null is returned.
     *
     * @param className the name of the class to search
     * @return          the Class object corresponding to its name or null if
     *                  not present in the current structure
     */
    Class getHiddenClass(String className);

    /**
     * Moves an hidden class to a known class.
     * A known class is known when it is defined in an XML description file.
     *
     * @param className the name of the class to move
     */
    void moveHiddenToClass(String className);

    /**
     * Returns a new Class created with a class name.
     * The MetaObject is added to the list of classes managed by the current
     * schema. An isAbstract boolean is defined to know if the current class
     * is an abstract class or not.
     *
     * @param className     the string representation of the class name
     * @param isAbstract	true, if the current class is an abstract class,
     *                      else false, if the current class is not an abstract
     *                      class
     * @return	            a Class corresponding to the created class and
     *                      registered, or an existing one if already defined
     */
    Class createClass(String className, boolean isAbstract);

    /**
     * Returns a new Class created with a class name.
     * The MetaObject is added to the list of unknown classes managed by the
     * current schema. An isAbstract boolean is defined to know if the current
     * class is an abstract class or not.
     *
     * @param className	    the string representation of the class name
     * @param isAbstract	true, if the current class is an abstract class,
     *                      else false, if the current class is not an abstract
     *                      class
     * @return          	a Class corresponding to the created class and
     *                      registered, or an existing one if already defined
     */
    Class createHiddenClass(String className, boolean isAbstract);

    /**
     * Returns a GenClass created with a generic class name.
     *
     * @param genClassName	the string representation of the generic class name
     * @return	            a GenClass corresponding to its name. If the
     *                      GenClass does not exist, null is returned.
     */
    GenClass getGenClass(String genClassName);

    /**
     * Returns a new GenClass created with a generic class name.
     * The MetaObject is added to the list of generic classes managed by the
     * current schema. If the generic class already exists, it is returned.
     *
     * @param genClassName	the string representation of the generic class name
     * @param isAbstract    true, if the generic class is abstract, else false,
     *                      if the generic class is not abstract
     * @return	            a GenClass corresponding to the created generic
     *                      class and registered, or an existing one if already
     *                      defined
     */
    GenClass createGenClass(String genClassName, boolean isAbstract);

    /**
     * Allows to know all the registered classes into the current schema.
     * This method returns an iterator on Class object.
     *
     * @return	an iterator for Class object. If there is no class, an empty
     *          iterator is returned.
     */
    Iterator iterateClass();

    /**
     * Allows to know all the registered generic classes into the current
     * schema.
     * This method returns an iterator on GenClass object.
     *
     * @return	an iterator for GenClass object. If there is no generic class,
     *          an empty iterator is returned.
     */
    Iterator iterateGenClass();

    /**
     * Allows to know all the not yet completly registered classes into the
     * current schema.
     * This method returns an iterator on Class object.
     *
     * @return	an iterator for Class object. If there is no class, an empty
     *          iterator is returned.
     */
    Iterator iterateHiddenClass();


    /**
     * Moves an hidden compositename to a known compositename.
     * A known compositename is known when it is defined in an XML description file.
     *
     * @param cn_Name the name of the compositename to move
     */
    void moveHiddenToCompositeName(String cn_Name);

    /**
     * Returns a compositename identified with its name.
     *
     * @param cn_Name	    the string representation of the name of the composite name.
     * @return	            a compositename corresponding to its name. If the
     *                      compositename does not exist, null is returned.
     */
    CompositeName getCompositeName(String cn_Name);

    /**
     * Returns an hidden compositename.
     * An Hidden compositename is a compositename which is not yet totally defined.
     * If the compositename is not found into the current schema, null is returned.
     *
     * @param cn_Name   the name of the compositename to search
     * @return          the compositename object corresponding to its name or null if
     *                  not present in the current structure
     */
    CompositeName getHiddenCompositeName(String cn_Name);

    /**
     * Returns a new compositename identified by its name.
     * The MetaObject is added to the list of compositename classes managed by the
     * current schema. If the compositename already exists, it is returned.
     *
     * @param cn_Name	    the string representation of the name of the compositename
     * @return	            a compositename corresponding to the created compositename
     *                      and registered, or an existing one if already
     *                      defined
     */
    CompositeName createCompositeName(String CN_Name);

    /**
     * Returns a new compositename created with its name.
     * The MetaObject is added to the list of unknown compositename managed by the
     * current schema.
     *
     * @param cn_Name	    the string representation of the name of the compositename
     * @return          	a Class corresponding to the created class and
     *                      registered, or an existing one if already defined
     */
    CompositeName createHiddenCompositeName(String cn_Name);

    /**
     * Allows to know all the registered compositename into the current schema.
     * This method returns an iterator on compositename object.
     *
     * @return	an iterator for compositename object. If there is no class, an empty
     *          iterator is returned.
     */
    Iterator iterateCompositeName();

    /**
     * Allows to know all the not yet completly registered compositename into the
     * current schema.
     * This method returns an iterator on compositename object.
     *
     * @return	an iterator for compositename object. If there is no compositename, an empty
     *          iterator is returned.
     */
    Iterator iterateHiddenCompositeName();
}
