/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib;

/*
import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
*/

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MappingFactory;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.jorm.util.api.Loggable;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

/**
 * BasicClass is an implementation of the Class interface defined
 * in the meta information system. This object is defined by a list
 * of fields, a name, a list of mapping.
 * @author	X. Spengler
 */
public class BasicClass implements Class, Loggable {
    /**
     * abstractClass boolean defines if the current class is an abstract
     * class or not.
     */
    private boolean abstractClass;

    /**
	 * specifies that this class is inherited by one of those that are defined
	 * within this MetaInfo Manager.
	 */
	private boolean isInherited = false;

    /**
     * name is the name of the current class object.
     */
    private String name;

    /**
     * nameDefs defines a list of NameDef that the user can potentially
     * used for the mapping.
     * key: the name of the NameDef object
     * value: a NameDef object
     */

    private ArrayList nameDefs;

    /**
     * classRefs defines the list of the fields reprensented by
     * a reference to a Class.
     * key: the name of the ClassRef object
     * value: the ClassRef object
     */
    private Hashtable classRefs;

    /**
     * genClassRefs defines the list of the fields which are defined
     * by a Generic Class.
     * key: the name of the GenClassRef object
     * value: the GenClassRef object
     */
    private Hashtable genClassRefs;

    /**
     * primitiveElements defines the list of the fields with a primitive type.
     * key: the name of the PrimitiveElement object
     * value: the PrimitiveElement object
     */
    private Hashtable primitiveElements;

    /**
     * hiddenFields defines the list of the hiddenfields.
     * These fields are scalar fields.
     * key: the name of the hiddenField object
     * value: the scalarfield object
     */
    private Hashtable hiddenFields;

    /**
     * the parent of the current Class.
     */
    private MetaObject parent;

    /**
     * inheritedClasses represents the list of the inherited classes by
     * the current class object.
     * element: a Class MetaObject object
     */
    private ArrayList inheritedClasses;

    /**
     * classMappings defines the list of the class mapping defined by
     * the user. A ClassMapping is declared by a mapper name and its
     * mapping name.
     * key: the name of the mapping
     * value: a ClassMapping object
     */
    private Hashtable classMappings;

    /**
     * a simple logger to log
     */
    private Logger logger;

    /**
     * a logger factory to create other loggers if needed
     */
    private LoggerFactory loggerFactory;

    /**
     * Builds a new Class object.
     * This object is mainly defined by its name, if the class is abstract or
     * not, and its parent (a schema).
     *
     * @param className     the name of the current class
     * @param abstractClass true if the current class is abstract, else false
     *                      if the current class is not abstract
     * @param parent    the parent of the current object
     */
    public BasicClass(String className,
                      boolean abstractClass,
                      MetaObject parent) {
        this.name = className;
        this.abstractClass = abstractClass;
        this.parent = parent;

        classRefs = new Hashtable();
        genClassRefs = new Hashtable();
        classMappings = new Hashtable();
        nameDefs = new ArrayList();
        primitiveElements = new Hashtable();
        hiddenFields = new Hashtable();
        inheritedClasses = new ArrayList();
    }


    ///////////////////////////////////////////////////////////////////
    //                      from Class interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Allows to know if the current class is an abstract class or not.
     *
     * @return	true, if the current class is an abstract class, and false,
     *          if the current class is not abstract
     */
    public boolean isAbstract() {
        return abstractClass;
    }

    /**
     * Allows to know the name of the current class object.
     * @return	the string representation of the class name
     */
    public String getName() {
        return name;
    }

    /**
     * Allows to obtain an existing Field.
     * A TypedElement is a class field, and this method finds a field with its
     * name (fieldName). This method is a generic method to return a
     * TypedElement object. The field can be hidden or not.
     *
     * @param fieldName	the name of the class field
     * @return	a field object. If the field does not exist, null is returned.
     */
    public TypedElement getTypedElement(String fieldName) {
        if (logger != null)
            logger.log(BasicLevel.DEBUG,"try to return field (" + fieldName + ")");

        // this method finds in the 3 structures where is the field
        // fieldName could be a PrimitiveElement
        if (primitiveElements.get(fieldName) != null) {
            if (logger != null)
                logger.log(BasicLevel.DEBUG, "the field is a PrimitiveElement object");
            return (PrimitiveElement) primitiveElements.get(fieldName);
        } else if (hiddenFields.get(fieldName) != null) {
            if (logger != null)
                logger.log(BasicLevel.DEBUG, "the field is an hiddenfield object");
            return (PrimitiveElement) hiddenFields.get(fieldName);
        }
        // fieldName could be a reference to a class
        else if (classRefs.get(fieldName) != null) {
            if (logger != null)
                logger.log(BasicLevel.DEBUG, "the field is a ClassRef object");
            return (ClassRef) classRefs.get(fieldName);
        }
        // fieldName could be a reference to a generic class
        else if (genClassRefs.get(fieldName) != null) {
            if (logger != null)
                logger.log(BasicLevel.DEBUG, "the field is a GenClassRef object");
            return (GenClassRef) genClassRefs.get(fieldName);
        }
        // attention, the fieldname does not exist. where is it ?
        if (logger != null)
            logger.log(BasicLevel.WARN,
                        "the field (" + fieldName + ") is not defined in the current class");
        return null;
    }

    /**
     * Creates a new NameDef object for the current Class object.
     * Here we can define if the namedef is defined outside jorm or not.
     * In the case where the namedef is already defined for the current class,
     * the old one is used and returned.
     *
     * @param nameDefName   the name of the nameDef read in the persistence
     *                      description file
     * @param isSystem	true, the namedef is defined outside jorm, and false,
     *                  the namedef is define inside jorm
     * @return	a new object used to describe the PName projection for
     *          the current Class
     */
    public NameDef createNameDef() {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new NameDef () for the current Class (" + name + ")");

        NameDef newNameDef = new BasicNameDef(this);
        ((Loggable) newNameDef).setLogger(logger);
        nameDefs.add(newNameDef);
        return newNameDef;
    }

    /**
     * register the namedef in the currentclass with the mapping name passed as parameter.
     * @param mappingname the name of the mapping
     * @param nd the namedef to register.
     */
    /*public void registerNameDef(String mappingName, NameDef nd) {
        if (nameDefs.get(mappingName) != null) {

            logger.log(BasicLevel.ERROR,
                    "attention, the mappingName (" + mappingName +
                    ") is already binded to an existing NameDef in the class (" + name + ")");
        } else {
            nameDefs.put(mappingName, nd);
        }
    }*/

    /**
     * Returns a NameDef object from a given mapperName and mappingName.
     * If the namedef does not exist, null is returned.
     *
     * @param mappingName   the name of the mapping
     * @return  an existing namedef object if exists, either null
     */
    public NameDef getNameDef(String mappingName) {

		Iterator iter = nameDefs.iterator();
		while (iter.hasNext()) {
			NameDef nameDef = (NameDef) iter.next();
			// verify if the nameDef object contains the mapperName and
			// mappingName, if it is the case, we return it.
			if (nameDef.arePresents(mappingName)) {
				return nameDef;
			}
		}
		return null;
	}

    /**
     * Returns an iterator on existing NameDef for the current Class.
     * This method returns an empty iterator if no namedef exists in this Class.
     *
     * @return	an iterator on NameDef, or an empty iterator if no namedef is
     *          defined
     */

    public Iterator iterateNameDef() {
        return nameDefs.iterator();
    }

    /**
     * Allows to obtain a new PrimitiveElement from a class field name an its
     * type.
     * A PrimitiveElement is a Class field. If the class field already exists,
     * it is returned.
     *
     * @param fieldName	the name of the primitive field
     * @param type      the PType of the primitive field
     * @return	a new PrimitiveElement object, or the existing one if already
     *          defined
     */
    public PrimitiveElement createPrimitiveElement(String fieldName,
                                                   PType type) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "create a new PrimitiveElement (" +
                    fieldName + ") for the current Class (" + name + ")");

        // try to get the fieldname from the current structure
        if (primitiveElements.get(fieldName) != null) {
            // be careful, the primitive element already exists, we return it.
            if (logger != null) logger.log(BasicLevel.WARN,
                    "attention, try to create an existing PrimitiveElement (" +
                    fieldName + "), return existing one");
            return (PrimitiveElement) primitiveElements.get(fieldName);
        }

        // oh, the primitive element does not exist, we must create it.
        PrimitiveElement newPrimitiveElement =
                new BasicPrimitiveElement(fieldName, type, this);
        ((Loggable) newPrimitiveElement).setLogger(logger);
        primitiveElements.put(fieldName, newPrimitiveElement);
        return newPrimitiveElement;
    }

    /**
     * Creates a new scalar field and adds it to the list of hidden fields.
     * If the field already exists, it is returned, else a new one is built.
     *
     * @param fieldName	the name of the field to create
     * @param type	    the type of the field to create
     * @return	a new ScalarField object if this object does not already exist
     */
    public ScalarField createHiddenField(String fieldName, PType type) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new HiddenField (" + fieldName + ") for the current class (" + name + ")");

        // verify if the field is not yet in the database
        if (hiddenFields.get(fieldName) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing HiddenField (" +
                    fieldName + "), return existing one");
            return (ScalarField) hiddenFields.get(fieldName);
        }

        // else create a new one
        ScalarField newHiddenField =
                new BasicScalarField(fieldName, type, this);
        ((Loggable) newHiddenField).setLogger(logger);
        hiddenFields.put(fieldName, newHiddenField);
        return newHiddenField;
    }

    /**
     * Allows to obtain a new ClassRef from a field name an its class.
     * A ClassRef object is a field which is represented by a reference to a
     * Class. If the field already exists, it is returned.
     *
     * @param fieldName	the name of the field
     * @param className	the string representation of the class which represents
     *                  the reference of the field
     * @return	a new ClassRef object, or the existing one if already defined
     */
    public ClassRef createClassRef(String fieldName, Class class_) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "create a new ClassRef (" + fieldName +
                    ") for the current Class (" + name + ")");

        // just verify if the fielname does not exist in the current structure.
        if (classRefs.get(fieldName) != null) {
            // be careful, the object exists, we return it
            if (logger != null) logger.log(BasicLevel.WARN,
                    "attention, try to create an existing ClassRef (" +
                    fieldName + "), return existing one");
            return (ClassRef) classRefs.get(fieldName);
        }

        // we have to create a new object
        ClassRef newClassRef = new BasicClassRef(fieldName, class_, this);
        ((Loggable) newClassRef).setLogger(logger);
        classRefs.put(fieldName, newClassRef);
        return newClassRef;
    }

    /**
     * Allows to obtain a new GenClassRef from a field name an its generic
     * class.
     * A GenClassRef object is a field which is represented by a
     * reference to a generic class. If the field already exists,
     * it is returned.
     *
     * @param fieldName	the name of the field
     * @param genName	the GenClass object which represents the reference of
     *                  the field
     * @return	a new GenClassRef object, or the existing one if already
     *          defined
     */
    public GenClassRef createGenClassRef(String fieldName, String genName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "create a new GenClassRef (" +
                    genName + ") from field (" + fieldName + ") from Class (" +
                    name + ")");

        // verify if the fieldname already exists
        if (genClassRefs.get(fieldName) != null) {
            // be careful, the object already exists, we return it
            if (logger != null) logger.log(BasicLevel.WARN,
                    "attention, try to create an existing GenClassRef (" + fieldName +
                    "), return existing one");
            return (GenClassRef) genClassRefs.get(fieldName);
        }

        // we have to create a new object and return it
        GenClassRef newGenClassRef =
                new BasicGenClassRef(genName, fieldName, this);
        ((Loggable) newGenClassRef).setLogger(logger);
        genClassRefs.put(fieldName, newGenClassRef);
        return newGenClassRef;
    }

    /**
     * Returns an iterator on existing class fields.
     * If no field is defined, an empty iterator is returned.
     * This iterator contains PrimitiveElements, ClassRef and GenClassRef
     * objects. The top level object is TypedElement.
     *
     * @return	an iterator on fields, or null if there is no defined field
     */
    public Iterator iterateField() {
        return listField().iterator();
    }

    /**
     * returns an ArrayList which contains the current fields.
     * @return  a list of current fields
     */
    private ArrayList listField() {
        ArrayList allFields = new ArrayList();
        Iterator iter = primitiveElements.values().iterator();
        while (iter.hasNext()) {
            allFields.add(iter.next());
        }
        iter = null;
        iter = classRefs.values().iterator();
        while (iter.hasNext()) {
            allFields.add(iter.next());
        }
        iter = null;
        iter = genClassRefs.values().iterator();
        while (iter.hasNext()) {
            allFields.add(iter.next());
        }
        iter = null;

        return allFields;
    }

    /**
     * Returns an iterator on existing class fields + fields from super classes.
     * This iterator contains PrimitiveElement, ClassRef and GenClassRef
     * objects. This iterator does not contains the hiddenfield of the class.
     *
     * @return	an iterator on all defined fields, plus fields from super
     */
    public Iterator iterateAllField() {

        // In the other case, we start to calculate the current fields
        ArrayList allFields = listField();

        // and add the fields from the inherited classes
        Iterator iter = iterateInheritedClass();
        while (iter.hasNext()) {
            Class aClass = (Class) iter.next();
            Iterator itera = aClass.iterateAllField();
            while (itera.hasNext()) {
                allFields.add(itera.next());
            }
        }

        return allFields.iterator();
    }

    /**
     * Provides an iterator over the all the hiddenfield of the class including inherited fields.
     * @return the iterator embeding scalarfield object describing the hiddenfields.
     */
    public Iterator iterateAllHiddenField() {

        ArrayList allFields = new ArrayList();
        Iterator iter = hiddenFields.values().iterator();

        while (iter.hasNext()) {
            allFields.add(iter.next());
        }
        // and add the fields from the inherited classes
        iter = iterateInheritedClass();
        while (iter.hasNext()) {
            Class aClass = (Class) iter.next();
            Iterator itera = aClass.iterateAllField();
            while (itera.hasNext()) {
                allFields.add(itera.next());
            }
        }
        return allFields.iterator();
    }

    /**
     * return an iterator over the hiddenfield of the class definition.
     *
     * @returns the iterator containing the hiddenfield.
     */
    public Iterator iterateHiddenField() {
        return hiddenFields.values().iterator();
    }

    /**
     * retrieve an hiddenfield from its name.
     *
     * @returns the hiddenfield or null if the field does not exist.
     */
    public ScalarField getHiddenField(String fieldName) {
        logger.log(BasicLevel.DEBUG,
                "try to return an hidden field (" + fieldName + ")");
        // this method finds in the 3 structures where is the field
        return (ScalarField) hiddenFields.get(fieldName);

    }

    /**
     * Allows to obtain all the inherited classes from the current class object.
     * This iterator contains Class objects. In the case where there is no
     * inherited class, an empty iterator is returned.
     *
     * @return	an iterator on inherited classes (Class objets). If no object
     *          exists, an empty iterator is returned.
     */
    public Iterator iterateInheritedClass() {
        return inheritedClasses.iterator();
    }

    /**
     * Adds a new inherited class to the list of the inherited classes for the
     * current class object.
     * If the class already exists, nothing is done.
     *
     * @param class_	the inherited class to add to inherited classes list
     */
    public void addInheritedClass(Class class_) {
        if (class_ != null) {
			inheritedClasses.add(class_);
			class_.setInherited();
		}
    }

    /**
     * Returns the number of inherited classes for the current Class object.
     *
     * @return  the number of inherited classes
     */
    public int getInheritedClassNumber() {
        return inheritedClasses.size();
    }

    /**
     * Returns an existing ClassMapping object for the current class object.
     * If the ClassMapping object does not exist, null is returned.
     *
     * @param mapperName	the name of the mapper class name
     * @return	a ClassMapping object. If the ClassMapping does not exist for the
     *          mapper, null is returned.
     */
    public ClassMapping getClassMapping(String mapperName, String mappingName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "try to get ClassMapping from mapperName=<" + mapperName + "> " +
                    "mappingName=<" + mappingName + ">");

        // verify that the ClassMapping object exists
        if (classMappings.get(mappingName) == null) {
             if (logger != null) logger.log(BasicLevel.WARN,
                    "The ClassMapping (" + mapperName + " - " + mappingName + ") does not " +
                    "exists.");
            return null;
        }

        return (ClassMapping) classMappings.get(mappingName);
    }

    /**
     * Creates a new ClassMapping object for the current class.
     * This mapping is defined by the name of the mapper, and the name of
     * the mapping. If the ClassMapping object already exists, the existing
     * one is returned.
     *
     * @param mapperName	the name of the class mapper
     * @param mappingName	the name of the class mapping
     * @return	a new ClassMapping object. This object is dependent from
     *	    	the mapper. In the case the ClassMapping exists, it is returned.
     */
    public ClassMapping createClassMapping(String mapperName,
                                           String mappingName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new ClassMapping for the current Class (" + name +
                    ") mapperName=<" + mapperName + "> " +
                    "mappingName=<" + mappingName + ">");

        // just verify that the classmapping does not exists in the structure
        if (classMappings.get(mappingName) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing ClassMapping (" +
                    mappingName + "), return existing one");
            return (ClassMapping) classMappings.get(mappingName);
        }

        // in the other case, we have to create a new from the MappingFactory
        MetaObject tmp = parent;
        MetaObject par = null;
        while (tmp != null) {
            par = tmp;
            tmp = tmp.getParent();
        }

        ClassMapping classMapping =
                ((MappingFactory) ((Manager) par).getMappingFactory(mapperName))
                .createClassMapping(mapperName, mappingName, this);

        ((Loggable) classMapping).setLogger(logger);
        classMappings.put(mappingName, classMapping);
        return classMapping;
    }

    /**
     * Allows to know all the ClassMapping object used for the current class.
     * If no ClassMapping object are defined, an empty iterator is returned.
     *
     * @return	an Iterator object on ClassMapping object. If there is no
     *          object, an empty iterator is returned.
     */
    public Iterator iterateClassMapping() {
        return classMappings.values().iterator();
    }

    /**
     * Constructs an Hashtable with inherited class
     *
     * @param table the hashtable to build
     */
    public void addAllInheritedClass(Hashtable table) {
        Iterator iter = iterateInheritedClass();
        String all[] = new String[inheritedClasses.size() + 1];
        int count = 0;
        all[count] = name;
        count++;

        while (iter.hasNext()) {
            Class theClass = ((Class) iter.next());
            theClass.addAllInheritedClass(table);

            all[count] = theClass.getName();
            count++;

        }
        table.put(name, all);
    }

    /**
     * Returns the number of fields which are references to classes.
     *
     * @return  the number of fieldRef
     */
    public int getFieldRefNumber() {
        return classRefs.size();
    }

    /**
	 * Specifies that this class is inherited by another one within the
	 * set of compiled ones.
	 */
	public void setInherited() {
		isInherited = true;
	}

	/**
	 * Tests if this class is inherited by another one.
	 */
	public boolean isInherited() {
		return isInherited;
	}

    ///////////////////////////////////////////////////////////////////
    //                      from MetaObject interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Allows to know the parent MetaObject of the current MetaObject.
     *
     * @return	the MetaObject corresponding to the parent of the
     *  		current object. If there is no parent, null is returned.
     */
    public MetaObject getParent() {
        return parent;
    }

    ///////////////////////////////////////////////////////////////////
    //                      from Loggable interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Defines a logger object.
     *
     * @param logger	the logger object
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * Defines the logger factory to obtain new logger.
     *
     * @param loggerFactory	the LoggerFactory object to obtain a logger object
     */
    public void setLoggerFactory(LoggerFactory loggerFactory) {
        this.loggerFactory = loggerFactory;
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }
}

