/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

/**
 * BasicClassRef defines a reference to a class which is class field.
 * This object is created by the Class object to define its fields.
 * @author	X. Spengler
 */
public class BasicClassRef implements ClassRef, Loggable {
    /**
     *  the name of the reference for the class field name
     */
    private String fieldName;

    /**
     * the name of the class reference
     */
    private Class currentClass;

    /**
     * parent is the parent of the current object
     */
    private MetaObject parent;

    /**
     * a simple logger to log
     */
    private Logger logger;

    /**
     * a logger factory to create other loggers if needed
     */
    private LoggerFactory loggerFactory;

    /**
     * the list of the name definition declared by the user in the
     * current reference to the class.
     * value: a NameDef object
     */
    private ArrayList nameDefs;


    /**
     * Builds a new BasicClassRef object.
     * This object is defined by its name, the reference to the class and
     * its parent.
     *
     * @param fieldName	the string representation of the field name
     * @param class_	the reference to the class
     * @param parent	the parent of the current object
     */
    public BasicClassRef(String fieldName, Class currentClass, MetaObject parent) {
        this.fieldName = fieldName;
        this.currentClass = currentClass;
        this.parent = parent;
        nameDefs = new ArrayList();
    }

    ///////////////////////////////////////////////////////////////////
    //                      from ClassRef interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Returns the Class object which describes the ClassRef entity.
     * This object is used as the type of a field.
     *
     * @return  the Class Object contained by the current class ref object
     */
    public Class getMOClass() {
        return currentClass;
    }

    /**
     * Returns the name of the class ref.
     *
     * @return the name of the class reference meta object
     */
    public String getClassName() {
        return currentClass.getName();
    }

    ///////////////////////////////////////////////////////////////////
    //                      from MetaObject interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Allows to know the parent MetaObject of the current MetaObject.
     *
     * @return	the MetaObject corresponding to the parent of the
     *	    	current object. If there is no parent, null is returned
     */
    public MetaObject getParent() {
        return parent;
    }

    ///////////////////////////////////////////////////////////////////
    //                      from TypedElement interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Returns the name of the current typed element.
     *
     * @return	the string representation of the typed element
     */
    public String getName() {
        return fieldName;
    }

    /**
     * Allows to know the type of the field.
     *
     * @return	a PType object which defines the type of the current field
     */
    public PType getType() {
        MetaObject tmp = parent;
        MetaObject par = null;
        while (tmp != null) {
            par = tmp;
            tmp = tmp.getParent();
        }

        try {
            if (currentClass.getInheritedClassNumber() == 0) {
                // no super class, just use the name of the class.
                return ((Manager) par).getPTypeSpace().createPType(currentClass.getName());
            } else { // minimum 1 super class
                Hashtable table = new Hashtable();
                currentClass.addAllInheritedClass(table);
                String allNames [][] = new String[table.size()][];
                int count = 0;

                Iterator iter = table.values().iterator();
                while (iter.hasNext()) {
                    String[] names = (String[]) iter.next();
                    allNames[count][0] = names[0];
                    allNames[count] = names;
                }
                return ((Manager) par).getPTypeSpace().createPType(
                        currentClass.getName(),
                        allNames);
            }
        } catch (Exception e) {

        }
        return null;
    }

    ///////////////////////////////////////////////////////////////////
    //                      from Reference interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Creates a new NameDef object for the current reference class object.
     * If this NameDef already exists for the current object, it is returned,
     * in other case, a new NameDef object is created and returned.
     *
     * @param nameDefName   the name of the nameDef read in the persistence
     *                      description file
     * @param isSystem	    true, the namedef is defined outside jorm, and
     *                      false, the namedef is define inside jorm
     * @return	a new object used to describe the name projection for
     *	      	the current reference class, or an existing one if it
     *          already exists
     */

    public NameDef createNameDef() {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new NameDef () for the current ClassRef (" + fieldName + ")");
        // in the case it does not exist, we create it.
        NameDef newNameDef = new BasicNameDef(this);
        ((Loggable) newNameDef).setLogger(logger);
        nameDefs.add(newNameDef);
        return newNameDef;
    }

    /**
     * register the namedef in the reference with the mapping name passed as parameter.
     * the reference can be a ref to a class or a genclass.
     * @param mappingname the name of the mapping
     * @param nd the namedef to register.
     */
    /*public void registerNameDef(String mappingName, NameDef nd) {
        if (nameDefs.get(mappingName) != null) {

            logger.log(BasicLevel.ERROR,
                    "attention, try to create an existing NameDef (" +
                    mappingName + ")");
        } else {
            nameDefs.put(mappingName, nd);
        }
    }*/

    /**
     * Returns a NameDef object from a given mapperName and mappingName.
     * If the namedef does not exist, null is returned.
     *
     * @param mappingName   the name of the mapping
     * @return  an existing namedef object if exists, either null
     */
    public NameDef getNameDef(String mappingName) {

        Iterator iter = nameDefs.iterator();
        while (iter.hasNext()) {
            NameDef nameDef = (NameDef) iter.next();
            // verify if the nameDef object contains the
            // mappingName, if it is the case, we return it.
            if (nameDef.arePresents(mappingName)) {
                return nameDef;
            }
        }
        return null;
    }

    /**
     * Returns an iterator on existing NameDef for the current Class.
     * This method returns an empty iterator if no namedef exists in this Class.
     *
     * @return	an iterator on NameDef, or an empty iterator if no namedef is
     *          defined
     */

    public Iterator iterateNameDef() {
        return nameDefs.iterator();
    }


    ///////////////////////////////////////////////////////////////////
    //                      from Loggable interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Defines a logger object.
     *
     * @param logger	the logger object
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * Defines the logger factory to obtain new logger.
     *
     * @param loggerFactory	the LoggerFactory object to obtain a logger object
     */
    public void setLoggerFactory(LoggerFactory loggerFactory) {
        this.loggerFactory = loggerFactory;
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }
}
