/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

/**
 * BasicGenClassRef defines a reference to a generic class which is class field.
 * This object is created by the GenClass object to define its fields.
 * @author	X. Spengler
 */
public class BasicGenClassRef implements GenClassRef, Loggable {
    /**
     *  the name of the reference for the class field
     */
    private String name;

    /**
     * the name of the generic class which composed the GenClassRef object
     */
    private String genClassName;

    /**
     * parent is the parent of the current object
     */
    private MetaObject parent;

    /**
     * the list of the name definition declared by the user in the
     * current reference to the generic class.
     * key: the name of the NameDef
     * value: a NameDef object
     */
    private ArrayList nameDefs;

    /**
     * the list of the name definition declared by the user in the
     * current reference to the generic class.
     * key: the name of the NameDef
     * value: a NameDef object
     */
    private ArrayList refNameDefs;

    /**
     * hiddenFields defines the list of the hiddenfields.
     * These fields are scalar fields.
     * key: the name of the hiddenField object
     * value: the scalarfield object
     */
    private Hashtable hiddenFields;


    /**
     * a simple logger to log
     */
    private Logger logger;

    /**
     * a logger factory to create other loggers if needed
     */
    private LoggerFactory loggerFactory;

    /**
     * if true, the current GenClassRef is a primitive field
     */
    private boolean primitive;

    /**
     * if true, the current GenClassRef is a reference to a class
     */
    private boolean classRef;

    /**
     * if true, the current GenClassRef is a reference to a generic class
     */
    private boolean genClassRef;

    /**
     * currentGenClassRef is the reference to the generic class which defines
     * the type of the current field
     */
    private GenClassRef currentGenClassRef;

    /**
     * currentGenClass is the reference to the generic class which defines
     * the type of the current field
     */
    private GenClass currentGenClass;

    /**
     * currentClassRef is the reference to the class which defines the type
     * of the current field
     */
    private ClassRef currentClassRef;

    /**
     * classMappings defines all the mapping of classes
     * key: the name of the mapping
     * value: a ClassMapping object
     */
    private Hashtable classMappings;

    /**
     * primitiveType defines the type of the current field
     */
    private PrimitiveElement primitiveElement;

    /**
     * Builds a new BasicGenClassRef object.
     * This object is defined by its name, the reference to the generic class
     * and its parent.
     *
     * @param genClassName  the name of the generic class reference
     * @param name	        the string representation of the generic class
     *                      field name
     * @param parent	the parent of the current object
     */
    public BasicGenClassRef(String genClassName,
                            String name,
                            MetaObject parent) {
        this.name = name;
        this.genClassName = genClassName;
        this.parent = parent;
        hiddenFields = new Hashtable();
        nameDefs = new ArrayList();
        refNameDefs = new ArrayList();
        classMappings = new Hashtable();
    }

    ///////////////////////////////////////////////////////////////////
    //                      from GenClassRef interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Sets the type of the GenClassRef with a PrimitiveElement object.
     *
     * @param primitiveElement	the primitive element which defines the current
     *                          GenClassRef object. This object is the final
     *                          object of the recursion.
     */
    public void setPrimitiveElement(PrimitiveElement primitiveElement) {
        ((Loggable) primitiveElement).setLogger(logger);

        // set the flags, primitive to true, the others to false
        primitive = true;
        classRef = false;
        genClassRef = false;

        this.primitiveElement = primitiveElement;
        createGenClass();
    }

    /**
     * Sets the type of the GenClassRef with a ClassRef object.
     *
     * @param class_	the reference to the Class object which defines the
     *                  current GenClassRef object. This object is the final
     *                  object of the recursion.
     */
    public void setClassRef(ClassRef class_) {
        // set the flags, classRef to true, the others to false
        classRef = true;
        genClassRef = false;
        primitive = false;

        currentClassRef = class_;
        createGenClass();
    }

    /**
     * Creates a new GenClassRef object.
     * This method must be used when the recursion is not finished (we are on
     * not final object).
     *
     * @param genClassRefName	the name of the generic class used to create a
     *                          new GenClassRef object
     * @return	a new GenClassRef object
     */
    public GenClassRef createGenClassRef(String genClassRefName) {
        // set the flags, genClassRef to true, the others to false
        classRef = false;
        genClassRef = true;
        primitive = false;

        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "create a new GenClassRef (" +
                    genClassRefName + ") for the current GenClassRef (" +
                    name + ")");

        if (currentGenClassRef != null)
            return currentGenClassRef;
        else {
            currentGenClassRef = new BasicGenClassRef(genClassRefName, null, this);
            ((Loggable) currentGenClassRef).setLogger(logger);

            createGenClass();

            return currentGenClassRef;
        }
    }

    /**
     * Returns the PrimitiveElement object which defines the current GenClassRef
     * object.
     * This method is used after a call to isPrimitive() method, to determine
     * if the current object is a "basic" type.
     *
     * @return	the current type of the field. This method always returns a
     *          PType object, and can not return null value because of the
     *          previous test.
     */
    public PrimitiveElement getPrimitiveElement() {
        return primitiveElement;
    }

    /**
     * Returns the ClassRef object which defines the current GenClassRef object.
     * This method is used after a call to isClassRef() method, to determine if
     * the current object is defined by a final class or not.
     *
     * @return	the final Class which describes the type of the object which
     *          defines the GenClassRef object. This method always returns a
     *          Class, and can not return null value because of the previous
     *          test.
     */
    public ClassRef getClassRef() {
        return currentClassRef;
    }

    /**
     * Returns the GenClassRef object which defines the current GenClassRef
     * object.
     * This object exists when it is in a recursion. This method is used after a
     * call to isGenClassRef() method, to determine if the current object is a
     * reference to a generic class or not.
     *
     * @return	the underlying reference to the generic class. This method
     *          always returns a GenClassRef, and can not return null value
     *          because of the previous test.
     */
    public GenClassRef getGenClassRef() {
        return currentGenClassRef;
    }

    /**
     * Allows to know if the current object is primitive or not (final and
     * "basic").
     *
     * @return	true, if the current object is primitive, else false, if the
     *          object is not primitive (either a reference to a Class, or
     *		    a reference to a generic class).
     */
    public boolean isPrimitive() {
        return primitive;
    }

    /**
     * Allows to know if the current object is a reference to a Class or not.
     *
     * @return	true, if the object is a reference to a Class, else false, if
     *          the object is not a reference to a Class (either a primitive
     *          object, or a reference to a generic class).
     */
    public boolean isClassRef() {
        return classRef;
    }

    /**
     * Allows to know if the current object is a reference to a generic class
     * or not.
     *
     * @return	true, if the object is a reference to a generic class, else
     *	      	false, if the object is not a reference to a generic class
     *		    (either a primitive object, or a reference to a class).
     */
    public boolean isGenClassRef() {
        return genClassRef;
    }

    /**
     * Creates a new NameDef object for the new GenClassRef object.
     * If the specified namedef object already exists, it is returned, else
     * a new NameDef object is created.
     *
     * @param nameDefName   the name of the nameDef read in the persistence
     *                      description file
     * @param isSystem	    true, the namedef is defined outside jorm, else
     *          			false, the namedef is define inside jorm
     * @return	a new object used to describe the PName projection for
     *	    	the current Class, or an existing object if already defined
     */

    public NameDef createRefNameDef() {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new RefNameDef () for the current GenClassRef (" + name + ")");
        // in the case it does not exist, we create it.
        NameDef newNameDef = new BasicNameDef(this);
        ((Loggable) newNameDef).setLogger(logger);
        refNameDefs.add(newNameDef);
        return newNameDef;
    }

    /**
     * register the namedef in the currentclass with the mapping name passed as parameter.
     * @param mappingname the name of the mapping
     * @param nd the namedef to register.
     */
    /* public void registerRefNameDef(String mappingName, NameDef nd) {
        if (refNameDefs.get(mappingName) != null) {

            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing RefNameDef (" +
                    mappingName + ")");
        } else {
            refNameDefs.put(mappingName, nd);
        }
    }*/



    /**
     * Returns a NameDef object from a given mapperName and mappingName.
     * If the namedef does not exist, null is returned.
     *
     * @param mappingName   the name of the mapping
     * @return  an existing namedef object if exists, either null
     */
    public NameDef getRefNameDef(String mappingName) {

        Iterator iter = refNameDefs.iterator();
        while (iter.hasNext()) {
            NameDef nameDef = (NameDef) iter.next();
            // verify if the nameDef object contains the mapperName and
            // mappingName, if it is the case, we return it.
            if (nameDef.arePresents(mappingName)) {
                return nameDef;
            }
        }
        return null;
    }

    /**
     * Returns an iterator on existing NameDef of GenClassRef object.
     * If no NameDef exists, an empty iterator is returned.
     *
     * @return	an iterator on NameDef object
     */
    public Iterator iterateRefNameDef() {
        return refNameDefs.iterator();
    }


    /**
     * Create a new hidden field for the refgenclass.
     *
     * @returns the ScalarField
     */
    public ScalarField createHiddenField(String fieldName, PType type) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new HiddenField (" + fieldName + ") for the current refgenclass (" + name + ")");

        // verify if the field is not yet in the database
        if (hiddenFields.get(fieldName) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing HiddenField (" +
                    fieldName + "), return existing one");
            return (ScalarField) hiddenFields.get(fieldName);
        }

        // else create a new one
        ScalarField newHiddenField =
                new BasicScalarField(fieldName, type, this);
        ((Loggable) newHiddenField).setLogger(logger);
        hiddenFields.put(fieldName, newHiddenField);
        return newHiddenField;
    }

    /**
     * return an iterator over the hiddenfield of the refgenclass definition.
     *
     * @returns the iterator containing the hiddenfield.
     */
    public Iterator iterateHiddenField() {
        return hiddenFields.values().iterator();
    }


    /**
     * retrieve an hiddenfield from its name.
     *
     * @returns the hiddenfield.
     */
    public ScalarField getHiddenField(String fieldName) {
        logger.log(BasicLevel.DEBUG,
                "try to return an hidden field (" + fieldName + ")");
        return (ScalarField) hiddenFields.get(fieldName);
    }


    /**
     * Returns the underlying GenClass object which describes the generic class
     * object used.
     *
     * @return	the underlying generic class object
     */
    public GenClass getGenClass() {
        return currentGenClass;
    }

    /**
     * Returns an existing ClassMapping object for the current reference object
     * from a given mapper name and mapping name.
     *
     * @param mapperName	the name of the mapper reference name
     * @param mappingName   the name of the mapping reference name
     * @return	a ClassMapping object. If the ClassMapping does not exist for
     *          the	mapper, null is returned.
     */
    public ClassMapping getClassMapping(String mapperName,
                                        String mappingName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "try to get ClassMapping from mapperName=<" + mapperName +
                    "> mappingName=<" + mappingName + ">");

        return (ClassMapping) classMappings.get(mappingName);
    }

    /**
     * Creates a new ClassMapping object for the current reference.
     * This mapping is defined by the name of the mapper, and the name of the
     * mapping.
     *
     * @param mapperName	the name of the reference mapper
     * @param mappingName	the name of the reference mapping
     * @return	a new ClassMapping object. This object is dependent from
     *		the mapper.
     */
    public ClassMapping createClassMapping(String mapperName,
                                           String mappingName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new ClassMapping for the current GenClassRef (" +
                    name + ") mapperName=<" + mapperName + "> " +
                    "mappingName=<" + mappingName + ">");

        // verify if there is already a classmapping with the good mappingName
        if (classMappings.get(mappingName) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing ClassMapping (" +
                    mappingName + "), return existing one");
            return (ClassMapping) classMappings.get(mappingName);
        }

        // else create a new one
        MetaObject tmp = parent;
        MetaObject par = null;
        while (tmp != null) {
            par = tmp;
            tmp = tmp.getParent();
        }

        ClassMapping ClassMapping =
                ((MappingFactory) ((Manager) par).getMappingFactory(mapperName))
                .createClassMapping(mapperName, mappingName, this);

        ((Loggable) ClassMapping).setLogger(logger);
        classMappings.put(mappingName, ClassMapping);
        return ClassMapping;
    }

    /**
     * Allows to know all the ClassMapping objects used for the current
     * reference.
     * If no ClassMapping is defined, an empty iterator is returned.
     *
     * @return	an iterator object on ClassMapping object. If there is no
     *          object,	null is returned.
     */
    public Iterator iterateClassMapping() {
        return classMappings.values().iterator();
    }

    /**
     * Returns the name of the GenClass object.
     *
     * @returns the name of the GenClassRef object
     */
    public String getGenClassName() {
        return genClassName;
    }


    /**
     * Allows to know the parent MetaObject of the current MetaObject.
     *
     * @return	the MetaObject corresponding to the parent of the
     *  		current object. If there is no parent, null is returned.
     */
    public MetaObject getParent() {
        return parent;
    }

    /**
     * Creates a new NameDef object for the current reference class object.
     * If this NameDef already exists for the current object, it is returned,
     * in other case, a new NameDef object is created and returned.
     *
     * @param nameDefName   the name of the nameDef read in the persistence
     *                      description file
     * @param isSystem	    true, the namedef is defined outside jorm, and
     *                      false, the namedef is define inside jorm
     * @return	a new object used to describe the name projection for
     *	      	the current reference class, or an existing one if it
     *          already exists
     */

    public NameDef createNameDef() {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new NameDef () for the current Class (" + name + ")");



        // in the case it does not exist, we create it.
        NameDef newNameDef = new BasicNameDef(this);
        ((Loggable) newNameDef).setLogger(logger);
        nameDefs.add(newNameDef);
        return newNameDef;
    }

    /**
     * register the namedef in the reference with the mapping name passed as parameter.
     * the reference can be a ref to a class or a genclass.
     * @param mappingname the name of the mapping
     * @param nd the namedef to register.
     */
    /*public void registerNameDef(String mappingName, NameDef nd) {
        if (nameDefs.get(mappingName) != null) {

            logger.log(BasicLevel.ERROR,
                    "attention, try to create an existing NameDef (" +
                    mappingName + ")");
        } else {
            nameDefs.put(mappingName, nd);

        }
    }*/

    /**
     * Returns a NameDef object from a given mapperName and mappingName.
     * If the namedef does not exist, null is returned.
     *
     * @param mappingName   the name of the mapping
     * @return  an existing namedef object if exists, either null
     */
    public NameDef getNameDef(String mappingName) {

        Iterator iter = nameDefs.iterator();
        while (iter.hasNext()) {
            NameDef nameDef = (NameDef) iter.next();
            // verify if the nameDef object contains the mapperName and
            // mappingName, if it is the case, we return it.
            if (nameDef.arePresents(mappingName)) {
                return nameDef;
            }
        }
        return null;
    }

    /**
     * Returns an iterator on existing NameDef for the current Class.
     * This method returns an empty iterator if no namedef exists in this Class.
     *
     * @return	an iterator on NameDef, or an empty iterator if no namedef is
     *          defined
     */

    public Iterator iterateNameDef() {
        return nameDefs.iterator();
    }



    ///////////////////////////////////////////////////////////////////
    //                      from Loggable interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Defines a logger object.
     *
     * @param logger	the logger object
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * Defines the logger factory to obtain new logger.
     *
     * @param loggerFactory	the LoggerFactory object to obtain a logger object
     */
    public void setLoggerFactory(LoggerFactory loggerFactory) {
        this.loggerFactory = loggerFactory;
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }

    ///////////////////////////////////////////////////////////////////
    //                      from TypedElement interface
    ///////////////////////////////////////////////////////////////////


    /**
     * Returns the name of the current typed element.
     *
     * @return	the string representation of the typed element
     */
    public String getName() {
        return name;
    }

    /**
     * Allows to know the type of the field.
     *
     * @return	a PType object which defines the type of the current field
     */
    public PType getType() {
        MetaObject tmp = parent;
        MetaObject par = null;
        while (tmp != null) {
            par = tmp;
            tmp = tmp.getParent();
        }

        try {
            if (primitive) {
                return primitiveElement.getType();
            } else if (classRef) {
                return currentClassRef.getType();
            } else { //genClassRef
                return null;
                // return ((Manager)par).getPTypeSpace().createPType(currentClass.getName(), allNames);
            }

        } catch (Exception e) {

        }
        return null;
    }

    ///////////////////////////////////////////////////////////////////
    //                      local methods
    ///////////////////////////////////////////////////////////////////

    /**
     * Creates a new GenClass object to describe the generic object used.
     *
     * @param name	the name of the generic class
     * @return	a new GenClass object
     */
    private void createGenClass() {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new GenClass (" + name +
                    ") for the current GenClassRef");
        if (currentGenClass == null) {
            currentGenClass = new BasicGenClass(name, false, this);
            ((Loggable) currentGenClass).setLogger(logger);
        }
    }
}
