/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MappingFactory;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Schema;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.type.lib.PTypeSpacePAAH;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.util.lib.BasicClassLoader;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Implementation object of the Manager interface.
 * This interface defines methods to create or find objects which construct
 * the meta information from the description files.
 * This object is managed by the parser system.
 * @author	X. Spengler
 */
public class BasicManager implements Manager, Loggable {
	/**
	 * this structure defines all declared schema in the current manager.
	 * key: the name of the schema
	 * value: a Schema object
	 */
	protected Hashtable schemas;

	/**
	 * this structure defines all the mapping factories declared in the meta
	 * information manager.
	 * key: the name of the mapper
	 * value: a MappingFactory object
	 */
	protected Hashtable mappingFactories;

	/**
	 * the unique classloader in jorm compilation part to dynamicaly
	 * load classes (for mapping factories).
	 */
	protected BasicClassLoader classLoader;

	/**
	 * a simple logger to log
	 */
	protected Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	protected LoggerFactory loggerFactory;

	/**
	 * TBD
	 */
	protected PTypeSpace ptypeSpace;


	///////////////////////////////////////////////////////////////////
	//                      from Manager interface
	///////////////////////////////////////////////////////////////////

    /**
     * Initializes the current Manager.
     * This method replaces the constructor method, with a single parameter,
     * which is a class loader.
     *
     * @param classloader   a single class loader to load class into the
     *                      meta information
     * @deprecated Basic ClassLoader is not a classloader. Need refactoring (Christophe Ney)
     */
    public void init(BasicClassLoader classloader) {
        this.classLoader = classloader;
		schemas = new Hashtable();
		mappingFactories = new Hashtable();
		ptypeSpace = new PTypeSpacePAAH();
    }

    /**
     * Initializes the current Manager with no specific classloader
     **/
    public void init() {
        init(new BasicClassLoader());
    }

	/**
	 * Returns an existing schema.
	 * If no Schema corresponds to the given name, null is returned.
	 *
	 * @param schemaName	the name of the schema
	 * @return	a Schema object. If the schema does not exist, null is returned.
	 */
	public Schema getSchema(String schemaName) {
		return (Schema) schemas.get(schemaName);
	}

	/**
	 * Returns a new schema.
	 * This method is the factory method to create schema object. If the
	 * schemaName schema already exists, it is returned.
	 *
	 * @param schemaName	the name of the schema
	 * @return	a new schema object, or the existing one if already defined
	 */
	public Schema createSchema(String schemaName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new Schema (" + schemaName + ") for the current Manager");

		// verify if the schema is not yet in the database
        if (schemas.get(schemaName) != null) {
           if (loggerFactory!=null)
            logger.log(BasicLevel.DEBUG,
            "The Schema already exist (" + schemaName +
            "), return existing one");

            return (Schema) schemas.get(schemaName);
        }

		// else create a new one
		Schema newSchema = new BasicSchema(schemaName, this);
		((Loggable)newSchema).setLogger(logger);
		schemas.put(schemaName, newSchema);
		return newSchema;
	}

	/**
	 * Allows to know all the declared schemas into the current metainformation
	 * manager.
	 * This iterator contains Schema object.
	 *
	 * @return	an iterator on registered Schema object. If there is no schema,
	 *		    an empty iterator is returned.
	 */
	public Iterator iterateSchema() {
		return schemas.values().iterator();
	}

    public Enumeration enumSchema() {
        return schemas.elements();
    }

	/**
	 * Adds a mapping factory to the current metainformation manager.
	 * A mapping factory is a factory to map object (class, field, ...).
	 *
	 * @param mapperName	the name of the mapper (i.e: OR for rdb)
	 * @param mappingFactory	the mapping factory to add to the list of
	 *                          existing mapping factories
	 */
	public void addMappingFactory(String mapperName, String mappingFactory) {
		try {
			MappingFactory factory =
				(MappingFactory) classLoader.load(mappingFactory);
			factory.init();
			if (factory instanceof Loggable) {
				if (loggerFactory!=null) {
					((Loggable) factory).setLoggerFactory(loggerFactory);
				}
				else if (logger!=null) {
					((Loggable) factory).setLogger(logger);
				}
			}
			mappingFactories.put(mapperName, factory);
		}
		catch (Exception e) {
			logger.log(BasicLevel.ERROR,
				"problem during loading of " + mappingFactory + " class");
			e.printStackTrace();
		}
	}

	/**
	 * Returns a mapping factory corresponding to a mapper name.
	 * If no MappingFactory corresponds to the mapper name, null is returned.
	 *
	 * @param mapperName	the name of the mapper to obtain
	 * @return	a mapping factory. If the mapping factory does not exist for
	 *          the mapper, null is returned.
	 */
	public MappingFactory getMappingFactory(String mapperName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "get the mappingFactory from " +
                    mapperName);
		return (MappingFactory) mappingFactories.get(mapperName);
	}

	/**
	 * searches a class into all the declared schemas and returns it.
	 * If no class corresponds to this name, null is returned.
	 * @param className	the name of the class to search
	 * @return	the Class object found, null if not found.
	 */
	public Class getClass(String className) {
		if (iterateSchema() == null)
			return null;

		Iterator allSchemas = iterateSchema();
		Class myClass = null;
		Schema aSchema = null;

		while (allSchemas.hasNext()) {
			aSchema = (Schema) allSchemas.next();
			myClass = aSchema.getClass(className);
			if (myClass != null) {
				return myClass;
			}
			myClass = aSchema.getHiddenClass(className);
			if (myClass != null) {
				logger.log(BasicLevel.WARN,
					"Be careful, the returned class is an hidden class, which " +
					"has not been yet defined.");
				return myClass;
			}
		}
		return null;
	}

	/**
	 * TBD
	 */
	public PTypeSpace getPTypeSpace() {
		return ptypeSpace;
	}

	///////////////////////////////////////////////////////////////////
	//                      from MetaObject interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Allows to know the parent MetaObject of the current MetaObject.
	 *
	 * @return	the MetaObject corresponding to the parent of the
	 *		    current object. If there is no parent, null is returned.
	 */
	public MetaObject getParent() {
		// the manager has not parent, orphan, God ?
		return null;
	}

	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	/**
	 * Defines the logger factory to obtain new logger.
	 *
	 * @param loggerFactory	the LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerFactory) {
		this.loggerFactory = loggerFactory;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}

    /**
     * build an iterator to iterates all classes of the meta-info
     * @return a new iterator
     */
    public Iterator getAllClassIterator() {
        Vector v = new Vector();
        if (iterateSchema() != null) {
            Iterator allSchemas = iterateSchema();
            while (allSchemas.hasNext()) {
                Schema aSchema = (Schema) allSchemas.next();
                Iterator it = aSchema.iterateClass();
                while (it.hasNext()) {
                    v.add(it.next());
                }
                it = aSchema.iterateHiddenClass();
                while (it.hasNext()) {
                    v.add(it.next());
                }
                it = aSchema.iterateGenClass();
                while (it.hasNext()) {
                    v.add(it.next());
                }
            }
        }
        return v.iterator();
    }
}