/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * The BasicNameDef object describes the mapping of a reference field,
 * It can be the PName of a Class. It can be a Field which is an attribute
 * of the class and is a reference; in this case, the reference can be
 * to another Class (ClassObject) or to a GenClass.
 * @author	X. Spengler
 */
public class BasicNameDef implements NameDef, Loggable {
    /**
     * the parent of the current object
     */
    private MetaObject parent;

    /**
     * isSystem defines is the current namedef is defined outside jorm or not.
     * if true, the name is defined outside jorm,
     * if false, the name is defined by jorm.
     */
    private boolean isSystem = false;

    /**
     * isNameRef defines if the namedef is describe by a composite name.
     * @return true the namedef is a composite one.
     */
    private boolean isNameRef = false;

    /**
     * isFieldName defines the namedef is describe by a single field.
     * @return true the namedef is a single one.
     */
    private boolean isFieldName = false;

    /**
     * nameRef contains a reference to a composite name.
     */
    private NameRef nameRef = null;

    /**
     * the fieldRefs contains FieldName  which is iether an hiddenfield or a field
     * added to the current list.
     * element: a FieldName object
     */
    private String fieldName = null;

    /**
     * a simple logger to log
     */
    private Logger logger;

    /**
     * a logger factory to create other loggers if needed
     */
    private LoggerFactory loggerFactory;

    /**
     * the String representation of the namedef name
     */
    private String name;

    /**
	 * this arrays contains differents set of mappingName/mapperName tuple
	 * for a given namedef object.
	 */
	private ArrayList mapps;

    /**
     * Builds a new BasicNameDef objects.
     * This object is defined by its name, if it is defined outside jorm or not,
     * and its parent.
     *
     * @param name	the name of the name definition object
     * @param isSystem	if true, the name is defined outside jorm, else false,
     *                  the name is defined by jorm
     * @param parent	the parent object of the current NameDef object
     */
    public BasicNameDef(MetaObject parent) {
        this.parent = parent;
        mapps = new ArrayList();
    }


    ///////////////////////////////////////////////////////////////////
    //                      from NameDef interface
    ///////////////////////////////////////////////////////////////////


    /**
     * Returns the number of hidden fields which composed the current object.
     *
     * @return	the number of hidden fields
     */
    public int getFieldNumber() {
        if (!(isNameRef()))
            return 1;
        else
            return nameRef.getProjection().size();
    }

    /**
     * Allows to know if the mapping of the referenced field is defined by
     * another system (could be defined by Versant system for instance).
     *
     * @return	true, if the mapping is defined by another system, else false,
     *          if the mapping is defined inside JORM (by ScalarField(s) and
     *          FieldName(s)).
     */
    public boolean isSystem() {
        return isSystem;
    }

    /**
     * Allows to set if the current NameDef is system defined (defined outside
     * jorm) or not.
     *
     * @param system	true, if the mapping is defined by another system, else
     *	        		false, if the mapping is defined inside JORM (by
     *                  ScalarField(s) and FieldName(s)).
     */
    public void setSystem(boolean system) {
        isSystem = system;
    }

    /**
     * allow to know if the namedef is describe by a composite name.
     * @return true the namedef is a composite one.
     */
    public boolean isNameRef() {
        return isNameRef;
    }

    /**
     * allow to know if the namedef is describe by a single field.
     * @return true the namedef is a single one.
     */
    public boolean isFieldName() {
        return isFieldName;
    }

    /**
     * Build a reference toward a composite name describing the namedef.
     * @param name the name of the compsoite name
     * @param cn_Class the meta object representing the composite name.
     * @return the nameref metaobject describing the reference.
     */
    public NameRef createNameRef(String name, CompositeName cn_Class) {
        isNameRef = true;
        BasicNameRef bnr = new BasicNameRef(name, cn_Class, this);
        this.nameRef = bnr;
        return bnr;
    }

    /**
     * Retrieves the nameref if the namedef is a composite one.
     * @return the nameref or null if the namedef is a single one or system.
     */
    public NameRef getNameRef() {
        return this.nameRef;
    }


     /**
     * Retrieve an iterator over the mapping name defined for this NameDef.
     * @return an iterator containing the mapping name as string.
     */
    public Iterator getMappNames(){
        return mapps.iterator();
    }

    /**
     * Retrieve the name of the field describing the namedef.
     * The fieldname can be the name of a field or an hiddenfield.
     * @return the fieldname or null is the namedef is a system or a composite one.
     */
    public String getFieldName() {
        return this.fieldName;
    }

    /**
     * Allow to set the fieldname describing the namedef single field.
     * @param fieldname the fieldname
     */
    public void setFieldName(String fieldname) {
        this.fieldName = fieldname;
        this.isFieldName = true;
    }

    /**
     * Returns an iterator on all fields (HiddenFields and FieldRef).
     * This iterator contains FieldName and ScalarField objects. If no object
     * fields exists, an empty iterator is returned.
     *
     * @return  an iterator on field
     */

    public Iterator iterateField() {
        if (!(isNameRef())) {
            ArrayList allInOne = new ArrayList();
            allInOne.add(fieldName);
            return allInOne.iterator();
        } else
            return nameRef.getProjection().values().iterator();
    }

    /**
         * Sets the names of the mapper and the mapping defined for this NameDef.
         * This method replaces two setters, which are not required because the two
         * String are set in same time.
         *
         * @param mapperName    the name of the used mapper
         * @param mappingName   the name of the used mapping
         */
        public void setMappNames(String mappingName) {
            mapps.add(mappingName);
        }

    /**
	 * Verifies if the tuple mappername/mappingname is present in the
	 * current namedef object.
	 * If the current NameDef supports the tuple, true is returned, else false.
	 *
	 * @param mappingName   the name of the mapping
	 * @return  true, if the tuple is present in the current NameDef, else
	 *          false, if the tuple does not exist in the current NameDef
	 */
	public boolean arePresents(String mappingName) {
		Iterator iter = mapps.iterator();
		while (iter.hasNext()) {
			if (((String) iter.next()).equals(mappingName)) {
				return true;
			}
		}
		return false;
	}



    ///////////////////////////////////////////////////////////////////
    //                      from MetaObject interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Allows to know the parent MetaObject of the current MetaObject.
     *
     * @return	the MetaObject corresponding to the parent of the
     *		    current object. If there is no parent, null is returned.
     */
    public MetaObject getParent() {
        return parent;
    }

    ///////////////////////////////////////////////////////////////////
    //                      from Loggable interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Defines a logger object.
     *
     * @param logger	the logger object
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * Defines the logger factory to obtain new logger.
     *
     * @param loggerFactory	the LoggerFactory object to obtain a logger object
     */
    public void setLoggerFactory(LoggerFactory loggerFactory) {
        this.loggerFactory = loggerFactory;
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }


}
