/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MappingFactory;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.ValueMapping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.jorm.util.api.Loggable;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * The BasicPrimitiveElement object is used to define fields which
 * are primitive types. This object is defined by its name, its type,
 * its parent, its size (if used) and a list of mappings for the field.
 * @author	X. Spengler
 */
public class BasicPrimitiveElement implements PrimitiveElement, Loggable {
	/**
	 * the name of the primitive field
	 */
	private String name;

	/**
	 * the type of the primitive field
	 */
	private PType type;

	/**
	 * the parent of current object
	 */
	private MetaObject parent;

	/**
	 * the list of mappings for the current field.
	 * key: the name of the mapping
	 * value: a ValueMapping object
	 */
	private Hashtable valueMappings;

	/**
	 * a simple logger to log
	 */
	private Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	private LoggerFactory loggerFactory;

	/**
	 * Builds a new PrimitiveElement object.
	 * This object is mainly defined by its name, its type and its parent.
	 *
	 * @param name	the name of the primitive field
	 * @param type	the type of the field (PType object)
	 * @param parent	the parent of the current object
	 */
	public BasicPrimitiveElement(String name, PType type, MetaObject parent) {
		this.name = name;
		this.type = type;
		this.parent = parent;
		valueMappings = new Hashtable();
	}

	///////////////////////////////////////////////////////////////////
	//                      from PrimitiveElement interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Allows to know if the current Field is a scalar field or not.
	 *
	 * @return	true, if the Field is scalar, else false, if the Field is not
	 *          scalar
	 */
	public boolean isScalar() {
		// TBD return false for the moment
		return false;
	}

	/**
	 * Returns a ValueMapping object for the current element object.
	 * If the corresponding object does not exist, null is returned.
	 *
	 * @param mapperName	the name of the mapping element name
	 * @param mappingName	the name of the  mapping name
	 * @return	a ValueMapping object. If there is no value mapping object,
	 *  		null is returned.
	 */
	public ValueMapping getValueMapping(String mapperName,
										String mappingName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "try to get ValueMapping from mapperName=<" + mapperName +
                    "> mappingName=<" + mappingName + "> for " + name +
                    " element.");

		if (valueMappings.get(mappingName) == null) {
            if (logger != null)
                logger.log(BasicLevel.DEBUG,
                        "no ValueMapping to return, try again ... !");
		}
		return (ValueMapping) valueMappings.get(mappingName);
	}

	/**
	 * Creates a new ValueMapping object for the current element.
	 * This mapping is defined by the name of the mapper, and the name of the
	 * mapping. If the object already exists, it is returned.
	 *
	 * @param mapperName	the name of the mapper name
	 * @param mappingName	the name of the  mapping name
	 * @return	a new ValueMapping object. This object is dependent from
	 *		    the mapper.
	 */
	public ValueMapping createValueMapping(String mapperName,
										   String mappingName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new ValueMapping for the current "+
                    "PrimitiveElement ("+this.name + ") mapperName=<" +
                    mapperName + "> mappingName=<" + mappingName + ">");

		if (valueMappings.get(mappingName) != null) {
            if (logger != null)
                logger.log(BasicLevel.WARN,
                        "attention, try to create an existing ValueMapping (" +
                        mapperName + "), return existing one");
			return (ValueMapping) valueMappings.get(mappingName);
		}

		MetaObject tmp = parent;
		MetaObject par = null;
		while (tmp != null) {
			par = tmp;
			tmp = tmp.getParent();
		}

		ValueMapping valueMapping =
			((MappingFactory) (((Manager) par).getMappingFactory(mapperName)))
			.createValueMapping(mapperName, mappingName, this);

		((Loggable)valueMapping).setLogger(logger);
		valueMappings.put(mappingName, valueMapping);
		return valueMapping;
	}

	/**
	 * Allows to know all the ValueMapping object used for the current element.
	 *
	 * @return	an iterator object on ValueMapping object. If no ValueMapping
	 *	    	is declared, an empty iterator is returned.
	 */
	public Iterator iterateValueMapping() {
		return valueMappings.values().iterator();
	}

	///////////////////////////////////////////////////////////////////
	//                      from MetaObject interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Allows to know the parent MetaObject of the current MetaObject.
	 *
	 * @return	the MetaObject corresponding to the parent of the
	 *	    	current object. If there is no parent, null is returned.
	 */
	public MetaObject getParent() {
		return parent;
	}

	///////////////////////////////////////////////////////////////////
	//                      from TypedElement interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Returns the name of the current typed element.
	 *
	 * @return	the string representation of the typed element
	 */
	public String getName() {
		return name;
	}

	/**
	 * Allows to know the type of the field.
	 *
	 * @return	a PType object which defines the type of the current field
	 */
	public PType getType() {
		return type;
	}

	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	/**
	 * Defines the logger factory to obtain new logger.
	 *
	 * @param loggerFactory	the LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerFactory) {
		this.loggerFactory = loggerFactory;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}
}
