/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * BasicSchema is an implementation of the Schema interface.
 * This object defines a set of classes and generic classes declared in a Schema
 * tag. Its parent is the Manager.
 * @author	X. Spengler
 */
public class BasicSchema implements Schema, Loggable {
    /**
     * This structure contains all declared classes for the
     * current schema object.
     * key: the name of the class
     * value: the Class MetaObject object
     */
    protected Hashtable classes;

    /**
     * This structure contains all undeclared classes for the current schema
     * object. An Hidden class is not generated, and must be defined in a
     * separated XML description file.
     * key: the name of the class
     * value: the Class MetaObject object
     */
    protected Hashtable hiddenClasses;

    /**
     * This structure contains all declared compositename for the
     * current schema object.
     * key: the name of the class
     * value: the Class MetaObject object
     */
    protected Hashtable compositeNames;

    /**
     * This structure contains all undeclared compositename for the current schema
     * object. An Hidden compositename is not generated, and must be defined in a
     * separated XML description file.
     * key: the name of the compositename
     * value: the compositename MetaObject object
     */
    protected Hashtable hiddenCompositeNames;

    /**
     * This structure contains all declared generic classes
     * for the current schema object.
     * key: the name of the generic class
     * value: a GenClass object
     */
    protected Hashtable genClasses;

    /**
     * The name of the schema, this name could be null if
     * the name is not defined by the user.
     */
    protected String name;

    /**
     * the parent object for the schema (Manager)
     * which is a MetaObject.
     */
    protected MetaObject parent;

    /**
     * a simple logger to log
     */
    public Logger logger;

    /**
     * a logger factory to create other loggers if needed
     */
    protected LoggerFactory loggerFactory;

    /**
     * Builds a new Schema object.
     * This object is defined by its name and its parent.
     *
     * @param name      the name of the schema to create
     * @param parent    the parent meta-object
     */
    public BasicSchema(String name, Manager parent) {
        this.name = name;
        this.parent = parent;
        classes = new Hashtable();
        hiddenClasses = new Hashtable();
        genClasses = new Hashtable();
        compositeNames = new Hashtable();
        hiddenCompositeNames = new Hashtable();
    }

    ///////////////////////////////////////////////////////////////////
    //                      from Schema interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Returns the name of the schema.
     *
     * @return	the string representation of the name of the schema.
     *		    Null is returned if there is no schema name else a not empty
     *          dotted string.
     */
    public String getName() {
        return name;
    }

    /**
     * Returns a Class created with a class name.
     *
     * @param className	the string representation of the class name
     * @return	        a Class corresponding to its name. If the Class does
     *                  not exist, null is returned.
     */
    public Class getClass(String className) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            if (classes.get(className) == null)
                logger.log(BasicLevel.DEBUG,
                        "There is no class with this name (" + className +
                        ") for the current Schema (" + name + ")");

        return (Class) classes.get(className);
    }

    /**
     * Returns an hidden class.
     * An Hidden class is a class which is not yet totally defined. If the class
     * is not found into the current schema, null is returned.
     *
     * @param className the name of the class to search
     * @return          the Class object corresponding to its name or null if
     *                  not present in the current structure
     */
    public Class getHiddenClass(String className) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
            if (hiddenClasses.get(className) != null)
                logger.log(BasicLevel.DEBUG,
                        "An hiddenclass exist for the class (" + className +
                        ") for the current Schema (" + name + ")");
            else
                logger.log(BasicLevel.DEBUG,
                        "There is no hiddenclass for the class(" + className +
                        ") for the current Schema (" + name + ")");
        }
        return (Class) hiddenClasses.get(className);
    }

    /**
     * Moves an hidden class to a known class.
     * A known class is known when it is defined in an XML description file.
     *
     * @param className the name of the class to move
     */
    public void moveHiddenToClass(String className) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "Moving the hiddenclass (" + className +
                    ") to Class for the current Schema (" + name + ")");

        Class hidden = (Class) hiddenClasses.remove(className);
        classes.put(className, hidden);
    }

    /**
     * Moves an hidden compositename to a known compositename.
     * A known compositename is known when it is defined in an XML description file.
     *
     * @param cn_Name the name of the compositename to move
     */
    public void moveHiddenToCompositeName(String cn_Name) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "Moving the hidden composite name (" + cn_Name +
                    ") to compositename for the current Schema (" + name + ")");

        CompositeName hidden = (CompositeName) hiddenCompositeNames.remove(cn_Name);
        compositeNames.put(cn_Name, hidden);
    }

    /**
     * Returns a compositename identified with its name.
     *
     * @param cn_Name	    the string representation of the name of the composite name.
     * @return	            a compositename corresponding to its name. If the
     *                      compositename does not exist, null is returned.
     */
    public CompositeName getCompositeName(String cn_Name) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            if (compositeNames.get(cn_Name) == null)
                logger.log(BasicLevel.DEBUG,
                        "There is no composite name with this name (" + cn_Name +
                        ") for the current Schema (" + name + ")");

        return (CompositeName) compositeNames.get(cn_Name);
    }

    /**
     * Returns an hidden compositename.
     * An Hidden compositename is a compositename which is not yet totally defined.
     * If the compositename is not found into the current schema, null is returned.
     *
     * @param cn_Name   the name of the compositename to search
     * @return          the compositename object corresponding to its name or null if
     *                  not present in the current structure
     */
    public CompositeName getHiddenCompositeName(String cn_Name) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
            if (hiddenCompositeNames.get(cn_Name) != null)
                logger.log(BasicLevel.DEBUG,
                        "An hidden composite name exist for the composite name (" + cn_Name +
                        ") for the current Schema (" + name + ")");
            else
                logger.log(BasicLevel.DEBUG,
                        "There is no hidden composite name exist for the composite name (" + cn_Name +
                        ") for the current Schema (" + name + ")");
        }
        return (CompositeName) hiddenCompositeNames.get(cn_Name);
    }

    /**
     * Returns a new compositename identified by its name.
     * The MetaObject is added to the list of compositename classes managed by the
     * current schema. If the compositename already exists, it is returned.
     *
     * @param cn_Name	    the string representation of the name of the compositename
     * @return	            a compositename corresponding to the created compositename
     *                      and registered, or an existing one if already
     *                      defined
     */
    public CompositeName createCompositeName(String CN_Name) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new CompositeName (" + CN_Name +
                    ") for the current Schema (" + name + ")");

        // verify the class is not yet in the current schema
        if (compositeNames.get(CN_Name) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing CompositeName (" + CN_Name +
                    "), return existing one");

            return (CompositeName) compositeNames.get(CN_Name);
        }

        // else create a new one
        CompositeName newCName = new BasicCompositeName(CN_Name, this);
        ((Loggable) newCName).setLogger(logger);
        compositeNames.put(CN_Name, newCName);
        return newCName;
    }

    /**
     * Returns a new compositename created with its name.
     * The MetaObject is added to the list of unknown compositename managed by the
     * current schema.
     *
     * @param cn_Name	    the string representation of the name of the compositename
     * @return          	a Class corresponding to the created class and
     *                      registered, or an existing one if already defined
     */
    public CompositeName createHiddenCompositeName(String cn_Name) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create an hidden Composite name (" + cn_Name +
                    ") for the current Schema (" + name + ")");

        // verify the class is not yet in the current schema
        if (hiddenCompositeNames.get(cn_Name) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing hidden compositename (" +
                    cn_Name + ")," + " return existing one");
            return (CompositeName) hiddenCompositeNames.get(cn_Name);
        }

        // else create a new one
        CompositeName newCN = new BasicCompositeName(cn_Name, this);
        ((Loggable) newCN).setLogger(logger);
        hiddenCompositeNames.put(cn_Name, newCN);
        return newCN;
    }

    /**
     * Allows to know all the registered compositename into the current schema.
     * This method returns an iterator on compositename object.
     *
     * @return	an iterator for compositename object. If there is no class, an empty
     *          iterator is returned.
     */
    public Iterator iterateCompositeName() {
        return compositeNames.values().iterator();
    }

    /**
     * Allows to know all the not yet completly registered compositename into the
     * current schema.
     * This method returns an iterator on compositename object.
     *
     * @return	an iterator for compositename object. If there is no compositename, an empty
     *          iterator is returned.
     */
    public Iterator iterateHiddenCompositeName() {
        return hiddenCompositeNames.values().iterator();
    }


    /**
     * Returns a new Class created with a class name.
     * The MetaObject is added to the list of classes managed by the current
     * schema. An isAbstract boolean is defined to know if the current class
     * is an abstract class or not.
     *
     * @param className     the string representation of the class name
     * @param isAbstract	true, if the current class is an abstract class,
     *                      else false, if the current class is not an abstract
     *                      class
     * @return	            a Class corresponding to the created class and
     *                      registered, or an existing one if already defined
     */
    public Class createClass(String className, boolean isAbstract) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new Class (" + className +
                    ") for the current Schema (" + name + ")");

        // verify the class is not yet in the current schema
        if (classes.get(className) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing Class (" + className +
                    "), return existing one");
            return (Class) classes.get(className);
        }

        // else create a new one
        Class newClass = new BasicClass(className, isAbstract, this);
        ((Loggable) newClass).setLogger(logger);
        classes.put(className, newClass);
        return newClass;
    }


    /**
     * Returns a new Class created with a class name.
     * The MetaObject is added to the list of unknown classes managed by the
     * current schema. An isAbstract boolean is defined to know if the current
     * class is an abstract class or not.
     *
     * @param className	    the string representation of the class name
     * @param isAbstract	true, if the current class is an abstract class,
     *                      else false, if the current class is not an abstract
     *                      class
     * @return          	a Class corresponding to the created class and
     *                      registered, or an existing one if already defined
     */
    public Class createHiddenClass(String className, boolean isAbstract) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create an hidden Class (" + className +
                    ") for the current Schema (" + name + ")");

        // verify the class is not yet in the current schema
        if (hiddenClasses.get(className) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing hidden Class (" +
                    className + ")," + " return existing one");
            return (Class) hiddenClasses.get(className);
        }

        // else create a new one
        Class newClass = new BasicClass(className, isAbstract, this);
        ((Loggable) newClass).setLogger(logger);
        hiddenClasses.put(className, newClass);
        return newClass;
    }

    /**
     * Returns a GenClass created with a generic class name.
     *
     * @param genClassName	the string representation of the generic class name
     * @return	            a GenClass corresponding to its name. If the
     *                      GenClass does not exist, null is returned.
     */
    public GenClass getGenClass(String genClassName) {
        return (GenClass) genClasses.get(genClassName);
    }

    /**
     * Returns a new GenClass created with a generic class name.
     * The MetaObject is added to the list of generic classes managed by the
     * current schema. If the generic class already exists, it is returned.
     *
     * @param genClassName	the string representation of the generic class name
     * @param isAbstract    true, if the generic class is abstract, else false,
     *                      if the generic class is not abstract
     * @return	            a GenClass corresponding to the created generic
     *                      class and registered, or an existing one if already
     *                      defined
     */
    public GenClass createGenClass(String genClassName, boolean isAbstract) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "create a new GenClass (" +
                    genClassName + ") for the current Schema (" + name + ")");

        // verify the genclass is not yet in the current schema
        if (genClasses.get(genClassName) != null) {
            logger.log(BasicLevel.WARN,
                    "attention, try to create an existing GenClass (" +
                    genClassName + "), return existing one");
            return (GenClass) genClasses.get(genClassName);
        }

        // else create a new one
        GenClass newGenClass =
                new BasicGenClass(genClassName, isAbstract, this);
        ((Loggable) newGenClass).setLogger(logger);
        genClasses.put(genClassName, newGenClass);
        return newGenClass;
    }

    /**
     * Allows to know all the registered classes into the current schema.
     * This method returns an iterator on Class object.
     *
     * @return	an iterator for Class object. If there is no class, an empty
     *          iterator is returned.
     */
    public Iterator iterateClass() {
        return classes.values().iterator();
    }

    /**
     * Allows to know all the registered generic classes into the current
     * schema.
     * This method returns an iterator on GenClass object.
     *
     * @return	an iterator for GenClass object. If there is no generic class,
     *          an empty iterator is returned.
     */
    public Iterator iterateGenClass() {
        return genClasses.values().iterator();
    }

    /**
     * Allows to know all the not yet completly registered classes into the
     * current schema.
     * This method returns an iterator on Class object.
     *
     * @return	an iterator for Class object. If there is no class, an empty
     *          iterator is returned.
     */
    public Iterator iterateHiddenClass() {
        return hiddenClasses.values().iterator();
    }

    ///////////////////////////////////////////////////////////////////
    //                      from MetaObject interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Allows to know the parent MetaObject of the current MetaObject.
     *
     * @return	the MetaObject corresponding to the parent of the
     *	       	current object. If there is no parent, null is returned.
     */
    public MetaObject getParent() {
        return parent;
    }

    ///////////////////////////////////////////////////////////////////
    //                      from Loggable interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Defines a logger object.
     *
     * @param logger	the logger object
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * Defines the logger factory to obtain new logger.
     *
     * @param loggerFactory	the LoggerFactory object to obtain a logger object
     */
    public void setLoggerFactory(LoggerFactory loggerFactory) {
        this.loggerFactory = loggerFactory;
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }
}
