/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib.rdb;

import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.rdb.RDBClassMapping;
import org.objectweb.jorm.metainfo.lib.BasicClassMapping;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * BasicRDBClassMapping defines methods to describe how the a class is mapped
 * into a relational database system.
 * @author	X. Spengler
 */
public class BasicRDBClassMapping extends BasicClassMapping
	implements RDBClassMapping {
	/**
	 * the name of the table
	 */
	private String tableName;

	/**
	 * a list of primary keys
	 * element: a String object
	 */
	private ArrayList primaryKeys;

	/**
	 * defines if the table can be null or not
	 */
	private boolean notNull;

	/**
	 * Builds a new BasicRDBClassMapping.
	 * This object is defined by the name of the mapper, the name of the mapping
	 * class and its parent.
	 *
	 * @param mapperName	the name of the mapper.
	 * @param mappingName	the name of the class mapping class.
	 * @param parent	the parent of the current object.
	 */
	public BasicRDBClassMapping(String mapperName,
								String mappingName,
								MetaObject parent) {
		super(mapperName, mappingName, parent);
		this.parent = parent;
		tableName = null;
		primaryKeys = new ArrayList();
		notNull = false;
	}

	///////////////////////////////////////////////////////////////////
	//                      from RDBClassMapping interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Allows to know the name of the table.
	 *
	 * @return	the name of the table for the current relational class mapping
	 */
	public String getTableName() {
		return tableName;
	}

	/**
	 * Sets the name of the relational table
	 *
	 * @param tableName the name of the table
	 */
	public void setTableName(String tableName) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "Set the name of the table (" + tableName + ")");
		this.tableName = tableName;
	}

	/**
	 * Defines if the current table can be null or not.
	 *
	 * @param notnull   true, the table can not be null, else false, the table
	 *                  can be null.
	 */
	public void setNotNull(boolean notnull) {
		this.notNull = notnull;
	}

	/**
	 * Allows to know if the table can be null or not.
	 *
	 * @return  true, the table can not be null, else false, the table can
	 *          be null.
	 */
	public boolean isNotNull() {
		return notNull;
	}

	/**
	 * Adds a primary key to the list of existing primary key for the current
	 * table.
	 *
	 * @param pk    the name of the new primary key
	 */
	public void addPrimaryKey(String pk) {
		if (!primaryKeys.contains(pk))
			primaryKeys.add(pk);
	}

	/**
	 * Returns an iterator on all the primary keys from the current table.
	 * This iterator contains String objects.
	 *
	 * @return  an iterator on primary keys, or an empty iterator if no key
	 *          exists
	 */
	public Iterator iteratePrimaryKey() {
		return primaryKeys.iterator();
	}
}
