/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.metainfo.lib.rdb;

import org.objectweb.jorm.metainfo.lib.BasicValueMapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.rdb.RDBColumnSpec;
import org.objectweb.jorm.metainfo.api.rdb.RDBValueMapping;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * BasicRDBValueMapping defines all the columns which compose
 * a value mapping.
 * @author	X. Spengler
 */
public class BasicRDBValueMapping extends BasicValueMapping
	implements RDBValueMapping {
	/**
	 * a table of columns
	 * key: the name of the column
	 * value: the RDBColumnSpec which corresponds to its name
	 */
	private Hashtable rdbColumnSpecs;

	/**
	 * Build a new BasicRDBValueMapping object.
	 *
	 * @param mapperName    the name of the mapper
	 * @param mappingName   the name of the mapping
	 * @param parent    the parent of the current object
	 */
	public BasicRDBValueMapping(String mapperName,
								String mappingName,
								MetaObject parent) {
        super(mapperName, mappingName, parent);
		rdbColumnSpecs = new Hashtable();
	}

	///////////////////////////////////////////////////////////////////
	//                      from RDBValueMapping interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Creates a new RDBColumnSpec for the current RDBValueMapping object.
	 * If the column already exists, it is returned.
	 *
	 * @param columnName	the name of the column
	 * @param type	the type of the column
	 * @param notNull	true, if "notnull" is defined, else false
	 * @return	a new RDBColumnSpec object or if the column already exists,
	 *          it is returned
	 */
	public RDBColumnSpec createRDBColumnSpec(String name,
											 String type,
											 boolean notNull) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "create a new RDBColumnSpec for the current RDBValueMapping");

		// verify if the object is known or not
		if (rdbColumnSpecs.get(name) != null) {
            if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG,
                        "attention, try to create an existing RDBColumnSpec (" +
                        name + "), return existing one");
			return (RDBColumnSpec) rdbColumnSpecs.get(name);
		}

		// else creates a new one
		RDBColumnSpec rdbColumnSpec =
			new BasicRDBColumnSpec(name, type, notNull, this);
		rdbColumnSpecs.put(name, rdbColumnSpec);
		return rdbColumnSpec;
	}

	/**
	 * Returns the RDBColumnSpec object corresponding to its name.
	 * If the column object does not exist, null is returned.
	 *
	 * @param columnName	the name of the RDBColumnSpec object (column name)
	 * @return	the corresponding RDBColumnSpec object or null if it does
	 *	    	not exist
	 */
	public RDBColumnSpec getRDBColumnSpec(String name) {
		return (RDBColumnSpec) rdbColumnSpecs.get(name);
	}

	/**
	 * Allows to know all the RDBColumnSpec defined in the current
	 * RDBValueMapping object.
	 * If no object is defined, an empty iterator is returned. For the moment,
	 * only one element is returned. This iterator contains RDBColumnSpec
	 * objects.
	 *
	 * @return	an iterator on RDBColumnSpec object
	 */
	public Iterator iterateRDBColumnSpec() {
		return rdbColumnSpecs.values().iterator();
	}
}
