/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.naming.api;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;

import java.util.Iterator;

/**
 * This interface defines a binder as an extension of naming context. This
 * extension allows the association of bindings to PName. Each JORM class
 * managed by a mapper have an associated PBinder.
 * Binders provide binding objects specific to the persistent class.
 * Binders are naming context keeping the association between the persistent
 * name and the binding object that gives access to an associated DSI.
 * @author	R. Basset, P. Dchamboux
 */
public interface PBinder extends PNamingContext {
	/**
	 * It associates a PBinding object with a PName within a PBinder. This
	 * sets/resets the link between a PBinding and a DSI.
	 * @param pn    The PName to associate to the PBinding. This PName must be
	 * 				valid within this binder just like with a naming context.
	 * @param pb	The PBinding to bind with the given PName; this PBinding
	 * 				must have been created or initialised by this PBinder.
	 * @exception PExceptionNaming	The given PName is not valid in this
	 * 								PBinder or this PBinding cannot be managed
	 * 								by this PBinder.
	 */
	void bind(PName pn, PBinding pb)
		throws PExceptionNaming;

	/**
	 * It returns the PClassMapping for which it is the binder.
	 * @return	The PClassMapping to which it is associated.
	 */
	PClassMapping getBinderClassMapping();

	/**
	 * It yields an Iterator that gives access to all PBinding objects that
	 * are managed by this PBinder. It may not be supported by all PBinder.
	 * @return	The relevant Iterator.
	 * @exception UnsupportedOperationException	It is raised if this method is
	 * 											not supported by this PBinder.
	 */
	Iterator iteratePBinding() throws UnsupportedOperationException;

	/**
	 * It releases the link between the PBinding passed as parameter and its
	 * associated DSI represented by the PName referenced by this PBinding.
	 * @param pb	The PBinding to unbind.
	 * @exception PExceptionNaming	This PBinding cannot be managed by this
	 * 								PBinder.
	 */
	void unbind(PBinding pb) throws PExceptionNaming;
}
