/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.naming.api;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.type.api.PType;

/**
 * The <b>PName</b> interface defines the behaviour of a persistent name as
 * it is managed by its associated naming context. Indeed, such a behaviour
 * is supported partly within the name implementation, and partly within its
 * naming context.
 * @author	R. Basset, P. Dchamboux
 */
public interface PName extends Cloneable {
	/**
	 * It tests if a given coding type is supported for this PName.
	 * @param codingtype	One of the coding type defined by
	 * 						PNamingContext.CTxxx constants defined within the
	 * 						PNamingContext interface.
	 * @return	It returns true if the coding type is supported for this PName.
	 */
	boolean codingSupported(short codingtype);

	/**
	 * The <b>encodeAbstract</b> method produces an encoded representation of
	 * this PName. The encoding format is the case of this method is a Java
	 * object of any complexity. Example of such object is a getter interface
	 * giving access of the fields of a composite relational primary key.
	 *
	 * @param conn	The connection which can be used to access the data store.
	 * @return	The Object corresponding to the encoded representation.
	 */
	Object encodeAbstract() throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeByte</b> method produces an encoded representation of this
	 * PName within its naming context. The encoding format in the case of this
	 * method is a byte.
	 *
	 * @param conn	The connection which can be used to access the data store.
	 * @return	The byte corresponding to the encoded representation.
	 */
	byte encodeByte() throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeChar</b> method produces an encoded representation of this
	 * PName within its naming context. The encoding format in the case of this
	 * method is a char.
	 *
	 * @param conn	The connection which can be used to access the data store.
	 * @return	The char corresponding to the encoded representation.
	 */
	char encodeChar() throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeChar</b> method produces an encoded representation of this
	 * PName within its naming context. The encoding format in the case of this
	 * method is a int.
	 *
	 * @param conn	The connection which can be used to access the data store.
	 * @return	The int corresponding to the encoded representation.
	 */
	int encodeInt() throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeIntObj</b> method produces an encoded representation of this
	 * PName within its naming context. The encoding format in the case of this
	 * method is a Integer.
	 *
	 * @param conn	The connection which can be used to access the data store.
	 * @return	The Integer corresponding to the encoded representation.
	 */
	Integer encodeIntObj() throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeChar</b> method produces an encoded representation of this
	 * PName within its naming context. The encoding format in the case of this
	 * method is a long.
	 *
	 * @param conn	The connection which can be used to access the data store.
	 * @return	The long corresponding to the encoded representation.
	 */
	long encodeLong() throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeChar</b> method produces an encoded representation of this
	 * PName within its naming context. The encoding format in the case of this
	 * method is a short.
	 *
	 * @param conn	The connection which can be used to access the data store.
	 * @return	The short corresponding to the encoded representation.
	 */
	short encodeShort() throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeChar</b> method produces an encoded representation of this
	 * PName within its naming context. The encoding format in the case of this
	 * method is a String.
	 * @param conn	The connection which can be used to access the data store.
	 * @return	The String corresponding to the encoded representation.
	 */
	String encodeString() throws PExceptionNaming;

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pnc	Thhe naming context which must provide a PName for
	 * 				this PName.
	 * @return	The new PName associated to this PName.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 * the data store
	 */
	PName export(Object conn, PNamingContext pnc)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol;

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pnc	Thhe naming context which must provide a PName for
	 * 				this PName.
	 * @param hints	Any information relevant for name creation.
	 * @return	The new PName associated to this PName.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 * the data store
	 */
	PName export(Object conn, PNamingContext pnc, Object hints)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol;

	/**
	 * It retrieves the naming context associated to this PName.
	 * @return	The naming context which provide this pname
	 */
	PNamingContext getPNamingContext();

	/**
	 * It retrieves the PType associated to this PName if any exists. If none
	 * is associated, it returns null.
	 * @return	The PType associated to this PName.
	 */
	PType getPType();

	/**
	 * It tells if this name corresponds to "null" representation within a
	 * its naming context.
	 * @return	true if it represents a "null" name.
	 */
	boolean isNull();

	/**
	 * It retrieves the information item associated to the given PName (if any
	 * exists).
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be looked up.
	 * @return	The information item associated with the given PName within
	 * 			a naming context.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 * the data store
	 */
	Object resolve(Object conn) throws PExceptionNaming, PExceptionIO;

	/**
	 * It removes this PName and all the associated information within a naming
	 * context (i.e., the association between the PName and its information
	 * item).
	 * @param conn	The connection that can be used to access the data store.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 * the data store
	 */
	void unexport(Object conn)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol;
}
