/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.naming.api;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.type.api.PExceptionTyping;

/**
 * The <b>PNameGetter</b> interface defines a generic accessor to be used by
 * persistent bindings to access fields of composite names it manages.
 * @author	P. Dchamboux
 */
public interface PNameGetter {
	/**
	 * It retrieves the value of a byte field.
	 * @param fn	A String containing the name of the field.
     * @param context A context used by PNameGetter to extract the related data.
     *                In in the case of RDB, it can be the result set.
	 * @return	The byte value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	byte pngetByteField(String fn, Object context) throws PExceptionTyping, PExceptionIO;

	/**
	 * It retrieves the value of a char field.
	 * @param fn	A String containing the name of the field.
	 * @param context A context used by PNameGetter to extract the related data.
     *                In in the case of RDB, it can be the result set.
     * @return	The char value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	char pngetCharField(String fn, Object context) throws PExceptionTyping, PExceptionIO;

	/**
	 * It retrieves the value of a short field.
	 * @param fn	A String containing the name of the field.
	 * @param context A context used by PNameGetter to extract the related data.
     *                In in the case of RDB, it can be the result set.
     * @return	The short value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	short pngetShortField(String fn, Object context) throws PExceptionTyping, PExceptionIO;

    /**
	 * It retrieves the value of an int field.
	 * @param context A context used by PNameGetter to extract the related data.
     *                In in the case of RDB, it can be the result set.
     * @param fn	A String containing the name of the field.
	 * @return	The int value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	int pngetIntField(String fn, Object context) throws PExceptionTyping, PExceptionIO;

	/**
	 * It retrieves the value of an Integer field.
	 * @param context A context used by PNameGetter to extract the related data.
     *                In in the case of RDB, it can be the result set.
     * @param fn	A String containing the name of the field.
	 * @return	The Integer value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	Integer pngetIntObjField(String fn, Object context) throws PExceptionTyping, PExceptionIO;

	/**
	 * It retrieves the value of a long field.
	 * @param context A context used by PNameGetter to extract the related data.
     *                In in the case of RDB, it can be the result set.
     * @param fn	A String containing the name of the field.
	 * @return	The long value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	long pngetLongField(String fn, Object context) throws PExceptionTyping, PExceptionIO;

	/**
	 * It retrieves the value of a string field.
	 * @param context A context used by PNameGetter to extract the related data.
     *                In in the case of RDB, it can be the result set.
     * @param fn	A String containing the name of the field.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	String pngetStringField(String fn, Object context) throws PExceptionTyping, PExceptionIO;
}
