/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.naming.api;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.type.api.PType;

/**
 * The <b>PNamingContext</b> interface provides a means to create and manage
 * PName. It ensures that each of the names it manages unambiguously
 * designates another persistent name or any other persistent information item.
 * A naming context defines the external structure of the name it manages,
 * which corresponds to the type involved in the encoding/decoding process.
 * Furthermore, several name format may be supported by a particular naming
 * context.
 * @author	R. Basset, P. Dchamboux
 */
public interface PNamingContext {
	/**
	 * The coding type of the naming context is a <code>byte</code>.
	 */
	static final short CTBYTE = 0x0001;
	/**
	 * The coding type of the naming context is an <code>char</code>.
	 */
	static final short CTCHAR = 0x0002;
	/**
	 * The coding type of the naming context is a <code>short</code>.
	 */
	static final short CTSHORT = 0x0004;
	/**
	 * The coding type of the naming context is an <code>int</code>.
	 */
	static final short CTINT = 0x0008;
	/**
	 * The coding type of the naming context is a <code>long</code>.
	 */
	static final short CTLONG = 0x0010;
	/**
	 * The coding type of the naming context is a <code>string</code>.
	 */
	static final short CTSTRING = 0x0020;
	/**
	 * The coding type of the naming context is an <code>Object</code> that
	 * is DS-specific.
	 */
	static final short CTSYSTEM = 0x0040;

	/**
	 * The coding type of the naming context is an <code>Object</code> that
	 * is composite name (e.g., a multi-fields relational primary key).
	 */
	static final short CTCOMPOSITE = 0x0080;

    /**
	 * The coding type of the naming context is an <code>int</code>.
	 */
	static final short CTINT_OBJ = 0x0100;

	/**
	 * It tests if a given coding type is supported by this naming context.
	 * @param codingtype	One of the coding type defined by CTxxx constants
	 * 						defined within this interface.
	 * @return	It returns true if the coding type is supported by this naming
	 * 			context.
	 */
	boolean codingSupported(short codingtype);

	/**
	 * The <b>decodeAbstract</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is abstract, which means that this is a Java
	 * object of any complexity. One example of such object is a getter
	 * interface representing a relational primary key (also called composite
	 * names).
     * @param context A context used to decode the related data.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The Object to be decoded.
	 * @return	The resulting PName.
	 */
	PName decodeAbstract(Object en, Object context) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>decodeByte</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a byte.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The byte to be decoded.
	 * @return	The resulting PName.
	 */
	PName decodeByte(byte en) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>decodeChar</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a char.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The char to be decoded.
	 * @return	The resulting PName.
	 */
	PName decodeChar(char en) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>decodeInt</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is an int.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The int to be decoded.
	 * @return	The resulting PName.
	 */
	PName decodeInt(int en) throws PExceptionNaming, UnsupportedOperationException;

    /**
	 * The <b>decodeIntObj</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is an Integer.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The Integer to be decoded.
	 * @return	The resulting PName.
	 */
	PName decodeIntObj(Integer en) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>decodeLong</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a long.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The long to be decoded.
	 * @return	The resulting PName.
	 */
	PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>decodeShort</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a short.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The short to be decoded.
	 * @return	The resulting PName.
	 */
	PName decodeShort(short en) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>decodeString</b> methods recreates PName within a naming
	 * context from an encoded representation. In the case of this method, the
	 * encoded representation is a String.
	 * @param conn	The connection that can be used to access the data store.
	 * @param en	The String to be decoded.
	 * @return	The resulting PName.
	 */
	PName decodeString(String en) throws PExceptionNaming;

	/**
	 * The <b>encodeAbstract</b> method produces an encoded representation of a PName
	 * within a naming context. The encoding format is the case of this method
	 * is a Java object of any complexity. Example of such object is a getter
	 * interface giving access of the fields of a relational primary key (also
	 * called composite names).
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded.
	 * @return	The Object corresponding to the encoded representation.
	 */
	Object encodeAbstract(PName pn) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeByte</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a byte.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The byte corresponding to the encoded representation.
	 */
	byte encodeByte(PName pn) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeChar</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a char.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The char corresponding to the encoded representation.
	 */
	char encodeChar(PName pn) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeInt</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a int.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The int corresponding to the encoded representation.
	 */
	int encodeInt(PName pn) throws PExceptionNaming, UnsupportedOperationException;

    /**
	 * The <b>encodeInt</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a int.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The int corresponding to the encoded representation.
	 */
	Integer encodeIntObj(PName pn) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeLong</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a long.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The long corresponding to the encoded representation.
	 */
	long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeShort</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a short.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The short corresponding to the encoded representation.
	 */
	short encodeShort(PName pn) throws PExceptionNaming, UnsupportedOperationException;

	/**
	 * The <b>encodeString</b> method produces an encoded representation of a
	 * PName within a naming context. The encoding format in the case of this
	 * method is a String.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be encoded
	 * @return	The String corresponding to the encoded representation.
	 */
	String encodeString(PName pn) throws PExceptionNaming;

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 *							the data store.
	 */
	PName export(Object conn, Object infoitem)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol;

	/**
	 * It associates a PName to an information item within a naming context.
	 * @param conn		The connection that can be used to access the data store.
	 * @param infoitem	The information item that must be accessible through a
	 * 					PName within a naming context.
	 * @param hints		Any information relevant for name creation.
	 * @return	The <code>PName</code> associated to the given information
	 * 			item.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 *							the data store.
	 */
	PName export(Object conn, Object infoitem, Object hints)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol;

	/**
	 * It yields a "null" representation of a name within this naming context.
	 * @return	The "null" PName associated with this naming context.
	 */
	PName getNull();

	/**
	 * It retrieves the information item associated to the given PName (if any
	 * exists).
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be looked up.
	 * @return	The information item associated with the given PName within
	 * 			a naming context.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 * the data store
	 */
	Object resolve(Object conn, PName pn) throws PExceptionNaming, PExceptionIO;

	/**
	 * It assigns the PType associated to the JORM class which is the type of
	 * the reference managed by this naming context.
	 * @param pt	The PType of the JORM class reference managed by this naming
	 * 				context.
	 */
	void setPType(PType pt);

	/**
	 * Tests if this naming context supports comosite name through a dynamic
	 * approach, which means that encodingAbstract/decodingAbstract use getter
	 * objects that implement the PNameGetter interface.
	 * @return	true if dynamic getter is supported by composite name coding.
	 */
	boolean supportDynamicComposite();

	/**
	 * In case of a naming context that supports composite names (i.e., names
	 * composed of several typed fields, whose type is limited to scalar ones),
	 * it tests if a particular field is defined by this composite name.
	 * It should be used by PClassMapping objects in order to verify, at
	 * deployment time, that naming contexts that are defined to manage the
	 * various references supports the same encoding name structure.
	 * @return	true if the proposed field is supported by this composite name.
	 */
	boolean supportCompositeField(String fn, PType ft);

	/**
	 * Tests if this naming context supports comosite name through a static
	 * approach, which means that encodingAbstract/decodingAbstract use getter
	 * objects that implement specific NsFieldGetter interface generated for
	 * a particular field "Field" within a particular NameDef space "Ns".
	 * @return	true if static getter is supported by composite name coding.
	 */
	boolean supportStaticComposite();

	/**
	 * It removes the PName and all the associated information within a naming
	 * context (i.e., the association between the PName and its information
	 * item).
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	A PName valid in this naming context. In case of a PBinder,
	 * 				export is done on a PBinding.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @exception PExceptionIO	Raised in case of problem during the access of
	 * the data store
	 */
	void unexport(Object conn, Object pn)
		throws PExceptionNaming, PExceptionIO, PExceptionProtocol;
}
