/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.parser.api;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.util.io.api.PathExplorer;
import org.w3c.dom.Element;

/**
 * The MappingParser interface has methods to parse specific mapping information
 * from an Element object. The methods use a basic MetaObject, and fill-in it
 * with corresponding mapping information.
 * @author	X. Spengler
 */
public interface MappingParser {
	/**
	 * Parses the mapping part of the XML file and builds the corresponding
	 * meta-information for a given Class object.
	 * Manager, PathExplorer, MapperName and MappingName must be set before the
	 * call of this method.
	 *
	 * @param node	    the node to treat
	 * @param class_    the class to complete
	 */
	void parseMapping(Element node, Class class_);

	/**
	 * Parses the mapping part of the XML file and builds the corresponding
	 * meta-information for a given PrimitiveElement object.
	 * Manager, PathExplorer, MapperName and MappingName must be set before the
	 * call of this method.
	 *
	 * @param node              the node to treat
	 * @param primitiveElement  the primitive element to complete
	 */
	void parseMapping(Element node, PrimitiveElement primitiveElement);

	/**
	 * Parses the mapping part of the XML file and builds the corresponding
	 * meta-information for a given GenClassRef object.
	 * Manager, PathExplorer, MapperName and MappingName must be set before the
	 * call of this method.
	 *
	 * @param node          the node to treat
	 * @param genClassRef   the generic class reference to complete
	 */
	void parseMapping(Element node, GenClassRef genClassRef);

	/**
	 * Assigns a meta-information manager to a parser in order from it to
	 * create the meta-information related to the XML files it parses.
	 *
	 * @param mim	the meta-information Manager to be associated to this
	 * 				parser
	 */
	void setMetaInfoManager(Manager mim);

	/**
	 * Assigns a PathExplorer object for locating files that have to be parsed.
	 *
	 * @param pathexpl	The PathExplorer to be used for file location
	 */
	void setPathExplorer(PathExplorer pathexpl);

	/**
	 * Sets the name of the mapper to use.
	 * This name is passed on the command line or defined in a properties file.
	 *
	 * @param mapperName	the name of the mapper
	 */
	void setMapperName(String mapperName);

	/**
	 * Sets the name of the mapping to use.
	 * This name is passed on the command line or defined in a properties file.
	 *
	 * @param mappingName	the name of the mapping
	 */
	void setMappingName(String mappingName);
}
