/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.parser.lib;

import org.apache.xerces.parsers.DOMParser;
import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.metainfo.lib.BasicClass;
import org.objectweb.jorm.metainfo.lib.BasicClassRef;
import org.objectweb.jorm.metainfo.lib.BasicPrimitiveElement;
import org.objectweb.jorm.metainfo.lib.BasicSchema;
import org.objectweb.jorm.parser.api.MappingParser;
import org.objectweb.jorm.parser.api.Parser;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.io.api.PathExplorer;
import org.objectweb.util.io.lib.DirJavaExplorer;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;

import java.util.*;

/**
 * The BasicDomParser class implements the Parser interface.
 * It uses the DOM trees associated to the XML files that are parsed in order
 * to build the meta-objects.
 * @author	P. Dechamboux, X. Spengler
 */
public class BasicDomParser implements Parser, Loggable {
    /**
     * The meta information manager for managing meta-objects within the
     * context of a mapper.
     */
    private Manager metaInfoManager;

    /**
     * A dom parser (here from xerses)
     */
    private DOMParser parser;

    /**
     * a simple logger to log
     */
    private Logger logger;

    /**
     * a logger factory to create other loggers if needed
     */
    private LoggerFactory loggerFactory;

    /**
     * pathExplorer is able to return input or output stream from a given file.
     */
    private PathExplorer pathExplorer;

    /**
     * the name of the mapping structure
     */
    private String mappingName;

    /**
     * the name of the mapper.
     */
    private String mapperName;

    /**
     * the parser which is able to parse the mapping part of the XML file.
     */
    private MappingParser mappingParser;

    static String fileSeparator = System.getProperty("file.separator");
    static String pathSeparator = System.getProperty("path.separator");

    /**
     * the parser exception reached during the parsing of a file
     */
    private Exception parserException = null;


    /**
     * This structure contains MetaObject which will be generated in the future
     */
    private Hashtable motable = null;

    /**
     * if the genDep flag is set to true, the dependance will be generated
     */
    private boolean genDep = false;

    /**
     * Main constructor for BasicDomParser. In order to compile XML files, it
     * creates a DOMParser and initialzes it.
     *
     * @param dtdVerify true, the dtd must be used to validate xml files, else
     *                  false
     */
    public BasicDomParser(boolean dtdVerify) {
        parser = new DOMParser();
        mapperName = null;
        mappingName = null;

        try {
            parser.setFeature(
                    "http://xml.org/sax/features/validation",
                    dtdVerify);

        } catch (SAXNotRecognizedException e) {
            logger.log(BasicLevel.ERROR,
                    "there is a problem (exception) from the SAX module " +
                    "(SAXNotRecognized)");
        } catch (SAXNotSupportedException e) {
            logger.log(BasicLevel.ERROR,
                    "there is a problem (exception) from the SAX module " +
                    "(SAXNotSupported)");
        }
        ParserErrorHandler parserErrorHandler = new ParserErrorHandler();
        parser.setErrorHandler(parserErrorHandler);
        motable = new Hashtable();
    }

    ///////////////////////////////////////////////////////////////////
    //                      from Parser interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Launches the parsing process. It reads and parses all the XML files
     * and builds a schema of meta-objects for all of them.
     * Before to use this method, the following object need to be set : Manager
     * (with setMetaInfoManager), the PathExplorer (with setPathExplorer), the
     * mapper name (with setMapperName), the mapping name (with setMappingName),
     * the mapping parser object (with setMappingParser).
     *
     * @param files	an iterator over the name the XML files to be parsed
     */
    public Collection parse(Iterator files) {
        while (files.hasNext()) {
            String filename = (String) files.next();
            MetaObject mo = parse(filename);
            if (mo != null)
                motable.put(filename, mo);
        }
        try {
            resolveHiddenClass();
            resolveHiddenCompositeName();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return motable.values();
    }

    /**
     * Launches the parsing process.
     * It reads and parse an XML file and builds a schema of meta-objects
     * for it.
     * Before to use this method, the following object need to be set : Manager
     * (with setMetaInfoManager), the PathExplorer (with setPathExplorer), the
     * mapper name (with setMapperName), the mapping name (with setMappingName),
     * the mapping parser object (with setMappingParser).
     *
     * @param file	the name of the XML file to be parsed
     */
    public MetaObject parse(String file) {
        Document doc;
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.LEVEL_DEBUG,
                    "try to get input stream from <" + file + ">");
        InputSource inputsource = new InputSource(pathExplorer.getInputStream(file));

        if (inputsource.getByteStream() == null) {
            logger.log(BasicLevel.WARN, "Attention, the file \"" + file +
                    "\" has not been found in the given CLASSPATH, " +
                    "the file is missing !");
            return null;
        }

        ((ParserErrorHandler) parser.getErrorHandler()).initialize();
        ((Loggable) mappingParser).setLogger(logger);
        mappingParser.setPathExplorer(pathExplorer);
        mappingParser.setMapperName(mapperName);
        mappingParser.setMappingName(mappingName);

        // parse the file inputsource
        try {
            parser.parse(inputsource);
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR,
                    "there is a problem (exception) during the parsing of <"
                    + file + "> from the SAX module (SAXNotRecognized)");
            e.printStackTrace();
            parserException = e;
        }
        if (((ParserErrorHandler) parser.getErrorHandler()).numErrors() != 0)
            logger.log(BasicLevel.ERROR,
                    "the parser handler has encountered "
                    + ((ParserErrorHandler) parser.getErrorHandler()).numErrors()
                    + " errors, please fix it(them)");

        // display errors
        for (int i = 0;
             i < ((ParserErrorHandler) parser.getErrorHandler()).numErrors();
             i++) {
            logger.log(BasicLevel.ERROR,
                    (((ParserErrorHandler) parser.getErrorHandler()).
                    getErrors()).get(i));
        }

        if (((ParserErrorHandler) parser.getErrorHandler()).numErrors() == 0)
            if (Debug.ON && logger != null &&
                    logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG,
                        "no error found by the DOM parser, continuing...");
        // Build the meta-information
        if (parser.getDocument().getDocumentElement() != null)
            return build(parser.getDocument(), file);

        return null;
    }

    /**
     * Assigns a meta-information manager to a parser in order from it to
     * create the meta-information related to the XML files it parses.
     *
     * @param mim	the meta-information Manager to be associated to this parser
     */
    public void setMetaInfoManager(Manager mim) {
        this.metaInfoManager = mim;
    }

    /**
     * Assigns a PathExplorer object for locating files that have to be parsed.
     *
     * @param pathexpl	the PathExplorer to be used for file location
     */
    public void setPathExplorer(PathExplorer pathExpl) {
        pathExplorer = pathExpl;
        if (pathExplorer == null)
            pathExplorer = new DirJavaExplorer();
    }

    /**
     * Sets the name of the mapper to use.
     * This name is passed on the command line or defined in a properties file.
     *
     * @param mapperName	the name of the mapper
     */
    public void setMapperName(String mapperName) {
        this.mapperName = mapperName;
    }

    /**
     * Sets the name of the mapping to use.
     * This name is passed on the command line or defined in a properties file.
     *
     * @param mappingName	the name of the mapping
     */
    public void setMappingName(String mappingName) {
        this.mappingName = mappingName;
    }

    /**
     * Sets the parser which is able to parse the mapping part of the XML file.
     *
     * @param mappingParser	the parser which parses the mapping object
     */
    public void setMappingParser(MappingParser mappingParser) {
        this.mappingParser = mappingParser;
    }

    /**
     * Enables or disables the generation of the dependances.
     *
     * @param gendep    true, the dependances are generated, else false
     */
    public void setGenDep(boolean gendep) {
        genDep = gendep;
    }

    /**
     * Allows to know if the dependances must be generated or not.
     *
     * @return  true, if the dependances are generated, else false
     */
    public boolean isGenDep() {
        return genDep;
    }

    /**
     * returns the exception which is reached during the parsing of a file
     *
     * @return  an Exception object
     */
    public Exception getParserException() {
        return parserException;
    }

    /**
     * allows to know the list of error the parser encounters.
     *
     * @return  a ParserErrorHandler object which countains all errors from
     *          the xml parser
     */
    public ParserErrorHandler getParserErrorHandler() {
        return (ParserErrorHandler) parser.getErrorHandler();
    }

    ///////////////////////////////////////////////////////////////////
    //                      local methods
    ///////////////////////////////////////////////////////////////////

    /**
     * builds the meta-information into the meta information manager.
     * @param doc	the document to read and move into meta information
     * @param name	the name of the document (file name).
     */
    private MetaObject build(Document doc, String name) {
        if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "build the meta information for <" + name +
                    "> description file");


        Schema schema = null;
        Class clazz = null;
        CompositeName aname = null;
        GenClass genclazz = null;

        NodeList nodes = doc.getDocumentElement().getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            String nodeName = node.getNodeName();

            if (nodeName.equals("Schema")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                schema = metaInfoManager.createSchema(((Element) node).getAttribute("Name"));
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");

            } else if (nodeName.equals("Class")) {
                String className = ((Element) node).getAttribute("Name");

                // compare the name of the file with the name of the class
                int index = name.lastIndexOf(fileSeparator);
                String sub = name.substring(index + 1, name.length() - 3);

                if (sub.compareTo(className) != 0)
                    logger.log(BasicLevel.WARN, "Be careful, the name of the class" +
                            " is different from the name of the file : " +
                            name + " and " + className +
                            "(" + className + " is used)");

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                String isAbstractString = ((Element) node).getAttribute("Abstract");
                if (schema == null) {
                    schema = metaInfoManager.createSchema("NoName");
                }

                // try to find the class in the schema
                clazz = schema.getClass(className);
                if (clazz == null) {
                    clazz = schema.getHiddenClass(className);
                    if (clazz == null) {
                        // if does not exist, create the class
                        if (isAbstractString.equals(""))
                            clazz = schema.createClass(className, false);
                        else
                            clazz = schema.createClass(className, Boolean.getBoolean(isAbstractString));
                    } else {
                        // the class is in the list of hidden classes, move it
                        // to the normal list
                        schema.moveHiddenToClass(className);
                    }
                }
                processClass(clazz, (Element) node);
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
                return clazz;

            } else if (nodeName.equals("CompositeName")) {
                String CN_name = ((Element) node).getAttribute("Name");
                // compare the name of the file with the name of the class
                int index = name.lastIndexOf(fileSeparator);
                String sub = name.substring(index + 1, name.length() - 3);

                if (sub.compareTo(CN_name) != 0)
                    logger.log(BasicLevel.WARN, "Be careful, the name of the composite name" +
                            " is different from the name of the file : " +
                            name + " and " + CN_name +
                            "(" + CN_name + " is used)");

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                if (schema == null) {
                    schema = metaInfoManager.createSchema("NoName");
                }
                // try to find the name in the schema
                aname = schema.getCompositeName(CN_name);

                if (aname == null) {
                    // if does not exist, try to find the name in the list of
                    // hidden name
                    aname = schema.getHiddenCompositeName(CN_name);

                    if (aname == null) {
                        // if does not exist, create the name
                        aname = schema.createCompositeName(CN_name);
                    } else {
                        // the class is in the list of hidden classes, move it
                        // to the normal list

                        schema.moveHiddenToCompositeName(CN_name);
                    }
                }
                // process the name tag and sub-tag
                processCompositeName(aname, (Element) node);
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
                return aname;
            } else if (nodeName.equals("GenClass")) {
                String genClassName = ((Element) node).getAttribute("Name");

                // compare the name of the file and the name of the GenClass
                int index = name.lastIndexOf(fileSeparator);
                String sub = name.substring(index + 1, name.length() - 3);

                if (sub.compareTo(genClassName) != 0)
                    logger.log(BasicLevel.WARN, "Be careful, the name of the genclass" +
                            " is different from the name of the file : " +
                            name + " and " + genClassName +
                            "(" + genClassName + " is used)");

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                genclazz = schema.createGenClass(genClassName, false);


                NodeList genNodes = ((Element) node).getChildNodes();
                for (int j = 0; j < genNodes.getLength(); j++) {
                    Node genNode = genNodes.item(j);
                    String genNodeName = genNode.getNodeName();
                    if (genNodeName.equals("ScalarField"))
                        processHiddenField(genclazz, genNode);
                }
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
                return genclazz;
            }

        }
        return null;
    }


    private void processCompositeName(CompositeName aname, Element element) {

        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            String nodeName = node.getNodeName();

            if (nodeName.equals("ScalarField")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                processScalarField(aname, node);

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("Inherit")) {
                // now, try to get the compositename extension and add it to the list
                // of inherited compositename
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                String CN_name = ((Element) node).getAttribute("Name");
                int lastindexofpoint = CN_name.lastIndexOf('.');
                String singleName = null;
                Schema newSchema = null;
                // search the schema of the inherited compositename
                if (lastindexofpoint != -1) {
                    // if the name of the schema is specified
                    String schemaName = CN_name.substring(0, CN_name.lastIndexOf('.'));
                    newSchema = metaInfoManager.createSchema(schemaName);
                    singleName = CN_name.substring(CN_name.lastIndexOf('.') + 1, CN_name.length());
                } else {
                    // if the name of the schema is not specified
                    newSchema = (Schema) aname.getParent();
                    logger.log(BasicLevel.WARN, "Be careful, " +
                            "the name of package (Schema name) is not specified, " +
                            "continue with " + newSchema.getName() + " schema.");
                    singleName = CN_name;
                }
                // try to get the class from the schema
                CompositeName singleCN = newSchema.getCompositeName(singleName);
                if (singleCN == null) {
                    // the class does not exist in the list of classes of the schema
                    // try to get the class in the list of hidden class of the schema
                    singleCN = newSchema.getHiddenCompositeName(singleName);
                    if (singleCN == null) {
                        // the class does not exist, create an hidden class
                        singleCN = newSchema.createHiddenCompositeName(singleName);
                    }
                }
                aname.addInheritedCompositeName(singleCN);
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }
        }
    }


    /**
     * parses a given element to fill in the Class structure. The Class is
     * created by the build method.
     * @param clazz	the Class object to fill in with the element object.
     * @param element	the DOM element to process (detail of a Class)
     */
    private void processClass(Class clazz, Element element) {

        NodeList nodes = element.getChildNodes();

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            String nodeName = node.getNodeName();
            if (nodeName.equals("Field")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                processField(clazz, (Element) node);
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }

            if (nodeName.equals("HiddenField")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                processHiddenField(clazz, (Element) node);
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("Extension")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                    logger.log(BasicLevel.DEBUG, "name of the extension=<" +
                            node.getFirstChild().getNodeValue() + ">");
                }
                clazz.addInheritedClass(
                        new BasicClass(node.getFirstChild().getNodeValue(), false, null));
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");

            } else if (nodeName.equals("NameDef")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                NameDef nameDef = clazz.createNameDef();
                processNameDef(nameDef, ((Element) node));
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");

            } else if (nodeName.equals("ClassMapping")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                mappingParser.setMetaInfoManager(metaInfoManager);
                mappingParser.parseMapping((Element) node, clazz);

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }
        }
    }


    private void processHiddenField(MetaObject clazz, Node node) {
        String nodeName = node.getNodeName();
        String HiddenFieldName = ((Element) node).getAttribute("Name");
        NodeList hiddenNodes = ((Element) node).getChildNodes();
        ScalarField hiddenField = null;

        for (int j = 0; j < hiddenNodes.getLength(); j++) {
            Node hiddenNode = hiddenNodes.item(j);
            String hiddenNodeName = hiddenNode.getNodeName();
            if (hiddenNodeName.equals("ScalarType")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + hiddenNodeName + ">");

                String scalarType = ((Element) hiddenNode).getAttribute("Type");
                if (clazz instanceof Class)
                    hiddenField = ((Class)clazz).createHiddenField(HiddenFieldName, getPType(scalarType));
                else if (clazz instanceof GenClass)
                    hiddenField = (ScalarField)((GenClass)clazz).createIndexField(HiddenFieldName, getPType(scalarType));
                else  if (logger != null)
                    logger.log(BasicLevel.ERROR, "The hiddenfield " +
                               HiddenFieldName + " should be declared in a class or a genclass");

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + hiddenNodeName + ">");
            } else if (hiddenNodeName.equals("ValueMapping")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + hiddenNodeName + ">");
                mappingParser.setMetaInfoManager(metaInfoManager);
                mappingParser.parseMapping((Element) hiddenNode, hiddenField);
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + hiddenNodeName + ">");
            }
        }
    }

    private void processHiddenField(GenClassRef gclr, Node node) {
        String nodeName = node.getNodeName();


        String HiddenFieldName = ((Element) node).getAttribute("Name");
        NodeList hiddenNodes = ((Element) node).getChildNodes();
        ScalarField hiddenField = null;

        for (int j = 0; j < hiddenNodes.getLength(); j++) {
            Node hiddenNode = hiddenNodes.item(j);
            String hiddenNodeName = hiddenNode.getNodeName();

            if (hiddenNodeName.equals("ScalarType")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + hiddenNodeName + ">");

                String scalarType = ((Element) hiddenNode).getAttribute("Type");
                hiddenField = gclr.createHiddenField(HiddenFieldName, getPType(scalarType));

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + hiddenNodeName + ">");
            } else if (hiddenNodeName.equals("ValueMapping")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + hiddenNodeName + ">");

                mappingParser.setMetaInfoManager(metaInfoManager);
                mappingParser.parseMapping((Element) hiddenNode, hiddenField);

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + hiddenNodeName + ">");
            }
        }
    }

    private void processField(Class clazz, Element node) {
        /**
         <!ELEMENT Field (PrimitiveType | (RefClass, NameDef+) | (RefGenClass, NameDef+) ) >
         <!ATTLIST Field
         Name CDATA #REQUIRED
         >
         */

        String itsName = ((Element) node).getAttribute("Name");
        NodeList nodesField = ((Element) node).getChildNodes();
        boolean refclassfield = true;

        ClassRef classRef = null;
        GenClassRef genClassRef = null;

        for (int i = 0; i < nodesField.getLength(); i++) {
            Node nodeField = nodesField.item(i);
            String nodeName = nodeField.getNodeName();

            if (nodeName.equals("PrimitiveType")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                PrimitiveElement primitiveElement = clazz.createPrimitiveElement(itsName,
                        getPType(((Element) nodeField).getAttribute("TypeCat")));

                processPrimitiveElement(primitiveElement, (Element) nodeField);

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");

            } else if (nodeName.equals("RefClass")) {
                /**
                 (RefClass, NameDef+)
                 */
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                String className = ((Element) nodeField).getAttribute("ClassName");
                int lastindexofpoint = className.lastIndexOf('.');
                String singleName = null;
                Schema newSchema = null;
                // search the schema of the refclass
                if (lastindexofpoint != -1) {
                    // if the name of the schema is specified
                    String schemaName = className.substring(0, className.lastIndexOf('.'));
                    newSchema = metaInfoManager.createSchema(schemaName);
                    singleName = className.substring(className.lastIndexOf('.') + 1, className.length());
                } else {
                    // if the name of the schema is not specified
                    newSchema = (Schema) clazz.getParent();
                    logger.log(BasicLevel.WARN, "Be careful, " +
                            "the name of package (Schema name) is not specified, " +
                            "continue with " + newSchema.getName() + " schema.");
                    singleName = className;
                }
                // try to get the class from the schema
                Class singleClass = newSchema.getClass(singleName);
                if (singleClass == null) {
                    // the class does not exist in the list of classes of the schema
                    // try to get the class in the list of hidden class of the schema
                    singleClass = newSchema.getHiddenClass(singleName);
                    if (singleClass == null) {
                        // the class does not exist, create an hidden class
                        singleClass = newSchema.createHiddenClass(singleName, false);
                    }
                }

                classRef = clazz.createClassRef(itsName, singleClass);

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("RefGenClass")) {
                /**
                 (RefGenClass, NameDef+)
                 */
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                refclassfield = false;
                String genName = ((Element) nodeField).getAttribute("Name");
                genClassRef = clazz.createGenClassRef(itsName, genName);
                processGenClassRef(genClassRef, (Element) nodeField);

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("NameDef")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                // either from RefClass or RefGenClass
                if (refclassfield) {
                    NameDef nameDef = classRef.createNameDef();

                    // classRef.createNameDef(((Element) nodeField).getAttribute("Name"));
                    processNameDef(nameDef, ((Element) nodeField));

                } else {
                    NameDef nameDef = genClassRef.createNameDef();

                    // genClassRef.createNameDef(((Element) nodeField).getAttribute("Name"));
                    processNameDef(nameDef, ((Element) nodeField));

                }
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }
        }
    }

    /**
     * returns the PType of a given name. If the type does not exist,
     * null is returned.
     * @param name	the string representation of the name.
     * @return	the corresponding PType of its name.
     */
    private PType getPType(String name) {
        PType type;
        if (name.equals(PTypeSpace.BOOLEAN.getJormName()))
            type = PTypeSpace.BOOLEAN;
        else if (name.equals(PTypeSpace.CHAR.getJormName()))
            type = PTypeSpace.CHAR;
        else if (name.equals(PTypeSpace.BYTE.getJormName()))
            type = PTypeSpace.BYTE;
        else if (name.equals(PTypeSpace.SHORT.getJormName()))
            type = PTypeSpace.SHORT;
        else if (name.equals(PTypeSpace.INT.getJormName()))
            type = PTypeSpace.INT;
        else if (name.equals(PTypeSpace.LONG.getJormName()))
            type = PTypeSpace.LONG;
        else if (name.equals(PTypeSpace.FLOAT.getJormName()))
            type = PTypeSpace.FLOAT;
        else if (name.equals(PTypeSpace.DOUBLE.getJormName()))
            type = PTypeSpace.DOUBLE;
        else if (name.equals(PTypeSpace.STRING.getJormName()))
            type = PTypeSpace.STRING;
        else if (name.equals(PTypeSpace.DATE.getJormName()))
            type = PTypeSpace.DATE;
        else if (name.equals(PTypeSpace.SERIALIZED.getJormName()))
            type = PTypeSpace.SERIALIZED;
        else
            type = null;
        return type;
    }

    private void processPrimitiveElement(PrimitiveElement primitiveElement, Element node) {
        NodeList nodesField = ((Element) node).getChildNodes();
        for (int i = 0; i < nodesField.getLength(); i++) {
            Node nodeField = nodesField.item(i);
            String nodeName = nodeField.getNodeName();
            if (nodeName.equals("ValueMapping")) {

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                mappingParser.setMetaInfoManager(metaInfoManager);
                mappingParser.parseMapping((Element) nodeField, primitiveElement);
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");

            }
        }
    }

    private void processNameDef(NameDef nameDef, Element element) {

        boolean defaultmappingname = true;
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            String nodeName = node.getNodeName();

            if (nodeName.equals("UsedByMapping")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                    logger.log(BasicLevel.DEBUG,
                            "adding mapping name =<" +
                            node.getFirstChild().getNodeValue() + ">");
                }

                defaultmappingname = false;
                nameDef.setMappNames(node.getFirstChild().getNodeValue());
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }
            else if (nodeName.equals("NameRef")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                String cn_Name = ((Element) node).getAttribute("Name");
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "nameref name =<" + cn_Name + ">");

                int lastindexofpoint = cn_Name.lastIndexOf('.');
                String singleName = null;
                Schema newSchema = null;
                // search the schema of the refclass
                if (lastindexofpoint != -1) {
                    // if the name of the schema is specified
                    String schemaName = cn_Name.substring(0, cn_Name.lastIndexOf('.'));
                    newSchema = metaInfoManager.createSchema(schemaName);
                    singleName = cn_Name.substring(cn_Name.lastIndexOf('.') + 1, cn_Name.length());
                } else {
                    // if the name of the schema is not specified TJRS = 2 ?

                    Object tmp = nameDef.getParent();

                    // get the schema
                    while (!(tmp instanceof Schema)) {

                        tmp = ((MetaObject) tmp).getParent();
                    }
                    newSchema = (Schema) tmp;

                    logger.log(BasicLevel.WARN, "Be careful, " +
                            "the name of package (Schema name) is not specified, " +
                            "continue with " + newSchema.getName() + " schema.");
                    singleName = cn_Name;
                }
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "schema name =<" + newSchema.getName() + ">");

                // try to get the class from the schema
                CompositeName singleCName = newSchema.getCompositeName(singleName);
                if (singleCName == null) {
                    // the class does not exist in the list of classes of the schema
                    // try to get the class in the list of hidden class of the schema
                    singleCName = newSchema.getHiddenCompositeName(singleName);
                    if (singleCName == null) {
                        // the class does not exist, create an hidden class
                        singleCName = newSchema.createHiddenCompositeName(singleName);
                    }
                }
                NameRef nameRef = nameDef.createNameRef(singleName, singleCName);
                processNameRef(nameRef, ((Element) node));

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("FieldName")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                String fieldName = ((Element) node).getAttribute("Name");
                nameDef.setFieldName(fieldName);

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("System")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                nameDef.setSystem(true);
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }
        }
        if (defaultmappingname) nameDef.setMappNames(mappingName);
    }

    private void processNameRef(NameRef nameRef, Element element) {

        NodeList nodes = element.getChildNodes();

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            String nodeName = node.getNodeName();

            if (nodeName.equals("NameFieldProjection")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                String namefield = ((Element) node).getAttribute("NameField");
                String classfield = ((Element) node).getAttribute("ClassField");
                nameRef.addProjection(namefield, classfield);


                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");

            }
        }
    }


    private void processScalarField(MetaObject mo, Node node) {
        String nodeName = node.getNodeName();
        String scalarFieldName = ((Element) node).getAttribute("Name");
        NodeList scalarNodes = ((Element) node).getChildNodes();
        ScalarField primitiveElement = null;

        for (int j = 0; j < scalarNodes.getLength(); j++) {
            Node scalarNode = scalarNodes.item(j);
            String scalarNodeName = scalarNode.getNodeName();

            if (scalarNodeName.equals("ScalarType")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + scalarNodeName + ">");

                String scalarType = ((Element) scalarNode).getAttribute("Type");

                if (mo instanceof CompositeName)
                    primitiveElement = ((CompositeName) mo).createCompositeNameField(scalarFieldName,
                            getPType(scalarType));

                /*else
                    primitiveElement = ((GenClass) mo).createIndexField(scalarFieldName,
                            getPType(scalarType));*/

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + scalarNodeName + ">");

            } else if (scalarNodeName.equals("ValueMapping")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + scalarNodeName + ">");

                mappingParser.setMetaInfoManager(metaInfoManager);
                mappingParser.parseMapping((Element) scalarNode, primitiveElement);

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + scalarNodeName + ">");
            } else if (scalarNodeName.equals("NullValue")) {
                if (mo instanceof CompositeName){
                    primitiveElement.setNullValue(((Element) scalarNode).getAttribute("Value"));

                }

            }
        }

    }

    /**
     * builds a RegGenClassField object with a DOM element.
     * @param genClassRef	the GenClassRef object to fill in with the DOM element
     * @param element	the DOM element to process to fill in the GenClassRef object.
     */
    private void processGenClassRef(GenClassRef genClassRef, Element element) {
        /**
         <!ELEMENT RefGenClass ( NameDef+, (PrimitiveType | (RefClass, NameDef+) | (RefGenClass, NameDef+)) ) >
         <!ATTLIST RefGenClass
         Name CDATA #REQUIRED
         Shared (TRUE | FALSE) "TRUE"
         >
         */

        NodeList nodes = element.getChildNodes();
        int where = 0;
        ClassRef classRef = null;

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            String nodeName = node.getNodeName();

            // only the first namedef (if other NameDef found,
            // it depends from RefClass or RefGenClass
            if (nodeName.equals("NameDef")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                if (where == 0) {
                    NameDef nameDef =
                            genClassRef.createRefNameDef();
                    processNameDef(nameDef, ((Element) node));

                } else if (where == 1) {
                    NameDef nameDef = classRef.createNameDef();
                    //classRef.createNameDef(((Element) node).getAttribute("Name"));
                    processNameDef(nameDef, (Element) node);

                } else if (where == 2) {
                    NameDef nameDef = genClassRef.createNameDef();
                    //genClassRef.createNameDef(((Element) node).getAttribute("Name"));
                    processNameDef(nameDef, (Element) node);

                }

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("HiddenField")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                processHiddenField(genClassRef, (Element) node);
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("PrimitiveType")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                PrimitiveElement primitiveElement =
                        new BasicPrimitiveElement(
                                genClassRef.getName(),
                                getPType(((Element) node).getAttribute("TypeCat")),
                                genClassRef);

                genClassRef.setPrimitiveElement(primitiveElement);
                processPrimitiveElement(primitiveElement, (Element) node);

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("RefClass")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG,
                            "begin =<" + nodeName + ">" + ((Element) node).getAttribute("ClassName"));

                String className = ((Element) node).getAttribute("ClassName");


                int lastindexofpoint = className.lastIndexOf('.');
                String singleName = null;
                Schema newSchema = null;
                if (lastindexofpoint != -1) {
                    String schemaName = className.substring(0, className.lastIndexOf('.'));
                    newSchema = metaInfoManager.createSchema(schemaName);
                    singleName = className.substring(className.lastIndexOf('.') + 1, className.length());
                } else {

                    MetaObject superSchema = null;
                    MetaObject obj = genClassRef;
                    while (obj.getParent() != null) {
                        superSchema = obj;
                        obj = obj.getParent();
                    }
                    if (superSchema instanceof Schema) {
                        newSchema = (Schema) superSchema;

                        logger.log(BasicLevel.WARN, "Be careful, " +
                                "the name of package (Schema name) is not specified, " +
                                "continue with " + newSchema.getName() + " schema.");
                    }
                    singleName = className;
                }

                Class singleClass = newSchema.getClass(singleName);
                if (singleClass == null) {
                    singleClass = newSchema.getHiddenClass(singleName);
                    if (singleClass == null) {
                        singleClass = newSchema.createHiddenClass(singleName, false);
                    }
                }

                classRef = new BasicClassRef(singleName, singleClass, newSchema);
                ((Loggable) classRef).setLogger(((BasicSchema) newSchema).logger);
                genClassRef.setClassRef(classRef);
                where = 1;

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("RefGenClass")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                String genName = ((Element) node).getAttribute("Name");
                GenClassRef newGenClassRef = genClassRef.createGenClassRef(genName);
                processGenClassRef(newGenClassRef, (Element) node);
                where = 2;

                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            } else if (nodeName.equals("ClassMapping")) {
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
                mappingParser.setMetaInfoManager(metaInfoManager);
                mappingParser.parseMapping((Element) node, genClassRef);
                if (Debug.ON && logger != null && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }
        }
    }

    /**
     * Tries to resolve the hidden classes from schema objects.
     * All schemas are parsed, and all hidden classes from them are resolved.
     */
    private void resolveHiddenClass() {
        boolean onClass = false;

        Enumeration enum_schema = metaInfoManager.enumSchema();

        // iterate while all hidden classes from all schemas exist
        while ((!onClass) && (enum_schema.hasMoreElements())) {
            boolean onSchema = true;

            //iterate on schemas
            for (Enumeration e = enum_schema; e.hasMoreElements();) {
                //while (iter_schema.hasNext()) {
                Schema aSchema = (Schema) enum_schema.nextElement();

                // iterate on hidden classes
                Iterator iter_hiddenClasses = aSchema.iterateHiddenClass();
                while (iter_hiddenClasses.hasNext()) {
                    Class aClass = (Class) iter_hiddenClasses.next();
                    String itsName = aClass.getName();
                    // parse the persistence description file with the .pd
                    // extension
                    MetaObject mo =
                            parse(aSchema.getName().replace('.', fileSeparator.charAt(0)) +
                            fileSeparator.charAt(0) + itsName + ".pd");
                    if (genDep)
                        motable.put(itsName, mo);
                }

                // if there is no hidden class into the current schema
                // the class validation is true only if the class validation
                // on previous schema is true too.
                if (!aSchema.iterateHiddenClass().hasNext()) {
                    onClass = true & onSchema;
                }
                // else, the class validation and the schema validation
                // are both false
                else {
                    onClass = false;
                    onSchema = false;
                }

            }
        }
    }

    private void resolveHiddenCompositeName() {
        boolean onCN = false;

        Enumeration enum_schema = metaInfoManager.enumSchema();
        // iterate while all hidden classes from all schemas exist
        while ((!onCN) && (enum_schema.hasMoreElements())) {
            boolean onSchema = true;

            //iterate on schemas
            for (Enumeration e = enum_schema; e.hasMoreElements();) {
                //while (iter_schema.hasNext()) {
                Schema aSchema = (Schema) enum_schema.nextElement();

                // iterate on hidden classes
                Iterator iter_hiddenCN = aSchema.iterateHiddenCompositeName();
                while (iter_hiddenCN.hasNext()) {
                    CompositeName aCN = (CompositeName) iter_hiddenCN.next();
                    String itsName = aCN.getName();
                    // parse the persistence description file with the .pd
                    // extension
                    MetaObject mo =
                            parse(aSchema.getName().replace('.', fileSeparator.charAt(0)) +
                            fileSeparator.charAt(0) + itsName + ".pd");
                    if (genDep)
                        motable.put(itsName, mo);
                }

                // if there is no hidden class into the current schema
                // the class validation is true only if the class validation
                // on previous schema is true too.
                if (!aSchema.iterateHiddenCompositeName().hasNext()) {
                    onCN = true & onSchema;
                }
                // else, the class validation and the schema validation
                // are both false
                else {
                    onCN = false;
                    onSchema = false;
                }

            }
        }
    }

    ///////////////////////////////////////////////////////////////////
    //                      from Loggable interface
    ///////////////////////////////////////////////////////////////////

    /**
     * Defines a logger object.
     *
     * @param logger	the logger object
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * Defines the logger factory to obtain new loggers.
     *
     * @param loggerfactory	The LoggerFactory object to obtain a logger object
     */
    public void setLoggerFactory(LoggerFactory loggerfactory) {
        this.loggerFactory = loggerfactory;
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }
}