/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.parser.lib;

import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.parser.api.MappingParser;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.io.api.PathExplorer;
import org.objectweb.util.io.lib.DirJavaExplorer;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.util.Hashtable;

/**
 * ProxyParser is an implementation of the Parser interface. This object defines
 * a parser which is able to dispatch operation to specific mapping parser,
 * and able to build the corresponding meta information.
 * @author	X. Spengler
 */
public class ProxyParser implements MappingParser, Loggable {
    /**
     * a proxyParser table is used to store specific parser to parse mapping
     * information
     * key: the name of the mapper (i.e. rdb)
     * value: the class of the mapping parser (i.e. RDBParser)
     */
    private static Hashtable proxyParser;

	/**
	 * The meta information manager for managing meta-objects within the
	 * context of a mapper.
	 */
	private Manager metaInfoManager;

	/**
	 * a simple logger to log
	 */
	private Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	private LoggerFactory loggerFactory;

	/**
	 * pathExplorer is able to return input or output stream from a given file.
	 */
	private PathExplorer pathExplorer;

	/**
	 * the name of the mapping structure
	 */
	private String mappingName;

	/**
	 * the name of the mapper.
	 */
	private String mapperName;

	/**
	 * Builds a new proxy parser.
	 * this class is loaded by a classloader.
	 */
	public ProxyParser() {
		// keep empty
	}

    static {
        try {
            proxyParser = new Hashtable();

            proxyParser.put("rdb",
                    (java.lang.Class.forName("org.objectweb.jorm.parser.lib.rdb.RDBParser")).newInstance());
            proxyParser.put("file",
                    (java.lang.Class.forName("org.objectweb.jorm.parser.lib.file.FileParser")).newInstance());
            proxyParser.put("memory",
                    (java.lang.Class.forName("org.objectweb.jorm.parser.lib.memory.MemoryParser")).newInstance());

        } catch (Exception e) {
            System.err.println("Problem during loading of classes");
        }
    }

    ///////////////////////////////////////////////////////////////////
    //                      from MappingParser interface
    ///////////////////////////////////////////////////////////////////

    /**
	 * Parses the mapping part of the XML file and builds the corresponding
	 * meta-information for a given Class object.
	 * Manager, PathExplorer, MapperName and MappingName must be set before the
	 * call of this method.
	 *
	 * @param node	    the node to treat
	 * @param class_    the class to complete
	 */
	public void parseMapping(Element element, org.objectweb.jorm.metainfo.api.Class class_) {
		NodeList nodes = element.getChildNodes();
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			String nodeName = node.getNodeName();

            int pos = nodeName.indexOf("ClassMapping");
            if (pos > 0) {
                String spos = nodeName.substring(0, pos );
                String emapping = ((Element) node).getAttribute("MappingName");
                MappingParser mp = (MappingParser) proxyParser.get(spos);
                setterMethod(mp, spos, emapping);
                mp.parseMapping(element,class_);
            }
        }
	}

    public void setterMethod(MappingParser mp, String spos, String emapping) {
        mp.setMetaInfoManager(metaInfoManager);

        if (mapperName == null)
            mp.setMapperName(spos);
        else
            mp.setMapperName(mapperName);

        if (mappingName == null)
            mp.setMappingName("null");
        else
            mp.setMappingName(mappingName);

        mp.setPathExplorer(pathExplorer);
        ((Loggable)mp).setLogger(logger);
    }

    /**
	 * Parses the mapping part of the XML file and builds the corresponding
	 * meta-information for a given PrimitiveElement object.
	 * Manager, PathExplorer, MapperName and MappingName must be set before the
	 * call of this method.
	 *
	 * @param node              the node to treat
	 * @param primitiveElement  the primitive element to complete
	 */
	public void parseMapping(Element element,
							 PrimitiveElement primitiveElement) {
		NodeList nodes = element.getChildNodes();
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			String nodeName = node.getNodeName();
            int pos = nodeName.indexOf("ValueMapping");
            if (pos > 0) {
                String spos = nodeName.substring(0, pos);
                String emapping = ((Element) node).getAttribute("MappingName");
                MappingParser mp = (MappingParser) proxyParser.get(spos);
                setterMethod(mp, spos, emapping);
                mp.parseMapping(element,primitiveElement);
            }

        }
    }

    /**
     * Parses the mapping part of the XML file and builds the corresponding
     * meta-information for a given GenClassRef object.
     * Manager, PathExplorer, MapperName and MappingName must be set before the
     * call of this method.
     *
     * @param node          the node to treat
     * @param genClassRef   the generic class reference to complete
     */
    public void parseMapping(Element element, GenClassRef genClassRef) {
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
			String nodeName = node.getNodeName();
            int pos = nodeName.indexOf("ClassMapping");
            if (pos > 0) {
                String spos = nodeName.substring(0, pos );
                String emapping = ((Element) node).getAttribute("MappingName");
                MappingParser mp = (MappingParser) proxyParser.get(spos);
                setterMethod(mp, spos, emapping);
                mp.parseMapping(element,genClassRef);
            }
		}
	}

	/**
	 * Assigns a meta-information manager to a parser in order from it to
	 * create the meta-information related to the XML files it parses.
	 *
	 * @param mim	the meta-information Manager to be associated to this
	 * 				parser
	 */
	public void setMetaInfoManager(Manager mim) {
		metaInfoManager = mim;
        metaInfoManager.addMappingFactory("rdb",
                "org.objectweb.jorm.metainfo.lib.rdb.BasicRDBMappingFactory");
        metaInfoManager.addMappingFactory("file",
                "org.objectweb.jorm.metainfo.lib.file.BasicFileMappingFactory");
        metaInfoManager.addMappingFactory("memory",
                "org.objectweb.jorm.metainfo.lib.memory.BasicMemoryMappingFactory");
    }

	/**
	 * Assigns a PathExplorer object for locating files that have to be parsed.
	 *
	 * @param pathexpl	The PathExplorer to be used for file location
	 */
	public void setPathExplorer(PathExplorer pathExpl) {
		pathExplorer = pathExpl;
		if (pathExplorer == null)
			pathExplorer = new DirJavaExplorer();
	}

	/**
	 * Sets the name of the mapper to use.
	 * This name is passed on the command line or defined in a properties file.
	 *
	 * @param mapperName	the name of the mapper
	 */
	public void setMapperName(String mapperName) {
		this.mapperName = mapperName;
	}

	/**
	 * Sets the name of the mapping to use.
	 * This name is passed on the command line or defined in a properties file.
	 *
	 * @param mappingName	the name of the mapping
	 */
	public void setMappingName(String mappingName) {
		this.mappingName = mappingName;
	}


	/**
	 * returns the PType of a given name. If the type does not exist,
	 * null is returned.
	 * @param name	the string representation of the name.
	 * @return	the corresponding PType of its name.
	 */
	private PType getPType(String name) {
		PType type;
		if (name.equals(PTypeSpace.BOOLEAN.getJormName()))
			type = PTypeSpace.BOOLEAN;
		else if (name.equals(PTypeSpace.CHAR.getJormName()))
			type = PTypeSpace.CHAR;
		else if (name.equals(PTypeSpace.BYTE.getJormName()))
			type = PTypeSpace.BYTE;
		else if (name.equals(PTypeSpace.SHORT.getJormName()))
			type = PTypeSpace.SHORT;
		else if (name.equals(PTypeSpace.INT.getJormName()))
			type = PTypeSpace.INT;
		else if (name.equals(PTypeSpace.LONG.getJormName()))
			type = PTypeSpace.LONG;
		else if (name.equals(PTypeSpace.FLOAT.getJormName()))
			type = PTypeSpace.FLOAT;
		else if (name.equals(PTypeSpace.DOUBLE.getJormName()))
			type = PTypeSpace.DOUBLE;
		else if (name.equals(PTypeSpace.STRING.getJormName()))
			type = PTypeSpace.STRING;
		else if (name.equals(PTypeSpace.DATE.getJormName()))
			type = PTypeSpace.DATE;
		else if (name.equals(PTypeSpace.SERIALIZED.getJormName()))
			type = PTypeSpace.SERIALIZED;
		else
			type = null;
		return type;
	}

	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	/**
	 * Defines the logger factory to obtain new loggers.
	 *
	 * @param loggerfactory	The LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerfactory) {
		this.loggerFactory = loggerfactory;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}
}
