/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.parser.lib.rdb;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.ValueMapping;
import org.objectweb.jorm.metainfo.api.rdb.RDBClassMapping;
import org.objectweb.jorm.metainfo.api.rdb.RDBValueMapping;
import org.objectweb.jorm.parser.api.MappingParser;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.api.Debug;
import org.objectweb.util.io.api.PathExplorer;
import org.objectweb.util.io.lib.DirJavaExplorer;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.jorm.util.api.Loggable;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * RDBParser is an implementation of the Parser interface. This object defines
 * a parser which is able to parse meta information related to an RDB system,
 * and able to build the corresponding meta information.
 * @author	X. Spengler
 */
public class RDBParser implements MappingParser, Loggable {
	/**
	 * The meta information manager for managing meta-objects within the
	 * context of a mapper.
	 */
	private Manager metaInfoManager;

	/**
	 * a simple logger to log
	 */
	private Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	private LoggerFactory loggerFactory;

	/**
	 * pathExplorer is able to return input or output stream from a given file.
	 */
	private PathExplorer pathExplorer;

	/**
	 * the name of the mapping structure
	 */
	private String mappingName;

	/**
	 * the name of the mapper.
	 */
	private String mapperName;

	/**
	 * Builds a new rdb parser.
	 * this class is loaded by a classloader.
	 */
	public RDBParser() {
		// keep empty
	}

	///////////////////////////////////////////////////////////////////
	//                      from MappingParser interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Parses the mapping part of the XML file and builds the corresponding
	 * meta-information for a given Class object.
	 * Manager, PathExplorer, MapperName and MappingName must be set before the
	 * call of this method.
	 *
	 * @param node	    the node to treat
	 * @param class_    the class to complete
	 */
	public void parseMapping(Element element, Class class_) {
		NodeList nodes = element.getChildNodes();
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			String nodeName = node.getNodeName();
			if (nodeName.equals(mapperName + "ClassMapping")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                RDBClassMapping rdbClassMapping = (RDBClassMapping)
                        class_.createClassMapping(mapperName, ((Element) node).getAttribute("MappingName"));

                rdbClassMapping.setNotNull((new Boolean(
                        ((Element) node).getAttribute(
                                "NotNull"))).booleanValue());
                NodeList enodes = ((Element) node).getChildNodes();
                for (int j = 0; j < enodes.getLength(); j++) {
                    Node enode = enodes.item(j);
                    String enodeName = enode.getNodeName();
                    /**
                     * <!ELEMENT OR:ClassMapping (OR:TableName, OR:ColumnName*) >
                     */

                    if (enodeName.equals(mapperName + "TableName")) {
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "begin =<" + enodeName + ">");
                        rdbClassMapping.setTableName(
                                ((Element) enode).getAttribute("Name"));
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "end   =<" + enodeName + ">");
                    }
                    else if (enodeName.equals(mapperName + "ColumnName")) {
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "begin =<" + enodeName + ">");
                        rdbClassMapping.addPrimaryKey(
                                ((Element) enode).getAttribute("Name"));
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "end   =<" + enodeName + ">");
                    }
                }

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }
        }
	}

	/**
	 * Parses the mapping part of the XML file and builds the corresponding
	 * meta-information for a given PrimitiveElement object.
	 * Manager, PathExplorer, MapperName and MappingName must be set before the
	 * call of this method.
	 *
	 * @param node              the node to treat
	 * @param primitiveElement  the primitive element to complete
	 */
	public void parseMapping(Element element,
							 PrimitiveElement primitiveElement) {
		NodeList nodes = element.getChildNodes();
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			String nodeName = node.getNodeName();
			if (nodeName.equals(mapperName + "ValueMapping")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                String emapping = ((Element) node).getAttribute("MappingName");

                NodeList enodes = ((Element) node).getChildNodes();
                for (int j = 0; j < enodes.getLength(); j++) {
                    Node enode = enodes.item(j);
                    String enodeName = enode.getNodeName();
                    if (enodeName.equals(mapperName + "ColumnSpec")) {
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "begin =<" + enodeName + ">");
                        ValueMapping valueMapping = null;
                        if ((valueMapping =
                                primitiveElement.getValueMapping(mapperName, emapping)) == null)
                            valueMapping =
                                    primitiveElement.createValueMapping(mapperName, emapping);

                        String theName = ((Element) enode).getAttribute("Name");
                        String theType = ((Element) enode).getAttribute("Type");
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "ColumnSpec name=<" + theName +
                                    "> type=<" + theType + ">");

                        ((RDBValueMapping) valueMapping).createRDBColumnSpec(
                                theName,
                                theType,
                                (new Boolean(
                                        ((Element) enode).getAttribute(
                                                "NotNull"))).booleanValue());
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "end   =<" + enodeName + ">");
                    }
                }

                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
            }
        }
    }

    /**
     * Parses the mapping part of the XML file and builds the corresponding
     * meta-information for a given GenClassRef object.
     * Manager, PathExplorer, MapperName and MappingName must be set before the
     * call of this method.
     *
     * @param node          the node to treat
     * @param genClassRef   the generic class reference to complete
     */
    public void parseMapping(Element element, GenClassRef genClassRef) {
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
			String nodeName = node.getNodeName();
			if (nodeName.equals(mapperName + "ClassMapping")) {
                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

                RDBClassMapping rdbClassMapping = (RDBClassMapping)
                        genClassRef.createClassMapping(mapperName, ((Element) node).getAttribute("MappingName"));

                rdbClassMapping.setNotNull((new Boolean(
                        ((Element) node).getAttribute(
                                "NotNull"))).booleanValue());
                NodeList enodes = ((Element) node).getChildNodes();
                for (int j = 0; j < enodes.getLength(); j++) {
                    Node enode = enodes.item(j);
                    String enodeName = enode.getNodeName();
                    /**
                     * <!ELEMENT OR:ClassMapping (OR:TableName, OR:ColumnName*) >
                     */

                    if (enodeName.equals(mapperName + "TableName")) {
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "begin =<" + enodeName + ">");
                        rdbClassMapping.setTableName(
                                ((Element) enode).getAttribute("Name"));
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "end   =<" + enodeName + ">");
                    }
                    else if (enodeName.equals(mapperName + "ColumnName")) {
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "begin =<" + enodeName + ">");
                        rdbClassMapping.addPrimaryKey(
                                ((Element) enode).getAttribute("Name"));
                        if (Debug.ON && logger != null &&
                                logger.isLoggable(BasicLevel.DEBUG))
                            logger.log(BasicLevel.DEBUG, "end   =<" + enodeName + ">");
                    }
                }


                if (Debug.ON && logger != null &&
                        logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			}
		}
	}

	/**
	 * Assigns a meta-information manager to a parser in order from it to
	 * create the meta-information related to the XML files it parses.
	 *
	 * @param mim	the meta-information Manager to be associated to this
	 * 				parser
	 */
	public void setMetaInfoManager(Manager mim) {
		this.metaInfoManager = mim;
	}

	/**
	 * Assigns a PathExplorer object for locating files that have to be parsed.
	 *
	 * @param pathexpl	The PathExplorer to be used for file location
	 */
	public void setPathExplorer(PathExplorer pathExpl) {
		pathExplorer = pathExpl;
		if (pathExplorer == null)
			pathExplorer = new DirJavaExplorer();
	}

	/**
	 * Sets the name of the mapper to use.
	 * This name is passed on the command line or defined in a properties file.
	 *
	 * @param mapperName	the name of the mapper
	 */
	public void setMapperName(String mapperName) {
		this.mapperName = mapperName;
	}

	/**
	 * Sets the name of the mapping to use.
	 * This name is passed on the command line or defined in a properties file.
	 *
	 * @param mappingName	the name of the mapping
	 */
	public void setMappingName(String mappingName) {
		this.mappingName = mappingName;
	}

	///////////////////////////////////////////////////////////////////
	//                      local methods
	///////////////////////////////////////////////////////////////////

	/**
	 * returns the PType of a given name. If the type does not exist,
	 * null is returned.
	 * @param name	the string representation of the name.
	 * @return	the corresponding PType of its name.
	 */
	private PType getPType(String name) {
		PType type;
		if (name.equals(PTypeSpace.BOOLEAN.getJormName()))
			type = PTypeSpace.BOOLEAN;
		else if (name.equals(PTypeSpace.CHAR.getJormName()))
			type = PTypeSpace.CHAR;
		else if (name.equals(PTypeSpace.BYTE.getJormName()))
			type = PTypeSpace.BYTE;
		else if (name.equals(PTypeSpace.SHORT.getJormName()))
			type = PTypeSpace.SHORT;
		else if (name.equals(PTypeSpace.INT.getJormName()))
			type = PTypeSpace.INT;
		else if (name.equals(PTypeSpace.LONG.getJormName()))
			type = PTypeSpace.LONG;
		else if (name.equals(PTypeSpace.FLOAT.getJormName()))
			type = PTypeSpace.FLOAT;
		else if (name.equals(PTypeSpace.DOUBLE.getJormName()))
			type = PTypeSpace.DOUBLE;
		else if (name.equals(PTypeSpace.STRING.getJormName()))
			type = PTypeSpace.STRING;
		else if (name.equals(PTypeSpace.DATE.getJormName()))
			type = PTypeSpace.DATE;
		else if (name.equals(PTypeSpace.SERIALIZED.getJormName()))
			type = PTypeSpace.SERIALIZED;
		else
			type = null;
		return type;
	}

	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	/**
	 * Defines the logger factory to obtain new loggers.
	 *
	 * @param loggerfactory	The LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerfactory) {
		this.loggerFactory = loggerfactory;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}
}
