/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.runtime;

import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.Debug;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.parser.api.Parser;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.type.lib.PTypeSpacePAAH;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.File;
import java.util.HashMap;
import javax.resource.cci.ConnectionFactory;

/**
 * @author	P. Dchamboux
 */
public class Mapper implements PMapper, Loggable {
	ConnectionFactory connectionFactory;
	/**
	 * This is the initial size of the map used to store the mapped classes.
	 */
	static final int INITIALMAPSIZE = 100;
	/**
	 * This is the name of the kind of mapper used (e.g., file, rdb, etc).
	 */
	String mapperName;
	/**
	 * This is the name of the mapping that this mapper executes within.
	 */
	String mappingName = null;
	/**
	 * This is the Manager that manages the meta-information associated with
	 * classes thnat are mapped within this mapper. It may stay null.
	 */
	Manager metainfoManager = null;
	/**
	 * This is the Parser that is used to load the mate-information associated
	 * with classes thnat are mapped within this mapper. It may stay null.
	 */
	Parser parser = null;
	/**
	 * This is the TypeSpace associated with this mapper. It is mandatory for
	 * mapping classes.
	 */
	PTypeSpace typeSpace = null;
	/**
	 * This is the structure that registers the classes that have been mapped
	 * within this mapper.
	 */
	HashMap mappedClasses = new HashMap(INITIALMAPSIZE);

	/**
	 * The logger associated with this mapper.
	 */
	Logger logger = null;
	/**
	 * The logger factory associated with this mapper.
	 */
	LoggerFactory loggerFactory = null;

	/**
	 * The name yielded by this method defines what kind of PMapper is used
	 * on a particular DS.
	 * @return	The name defining the PMapper kind.
	 */
	public String getMapperName() {
		return mapperName;
	}

	/**
	 * It retrieves the name of the mapping that is instanciated through this
	 * mapper.
	 * @return	The name of the mapping.
	 */
	public String getMappingName() {
		return mappingName;
	}

	/**
	 * It looks for the PClassMapping object associated to a JORM class within
	 * this PMapper.
	 * @return	The PClassMapping that maps the corresponding class, or null if
	 * 			none exists
	 */
	public PClassMapping lookup(String classname) {
		return (PClassMapping) mappedClasses.get(classname);
	}

	/**
	 * It registers a PClassMapping associated to a particular JORM class within
	 * this PMapper. It initializes the storage structures if necessary. It
	 * previously removes them if the "dsrem" flag is set. If the mapper is
	 * ready to manage the meta-information, it is loaded for this JORM class.
	 * @param conn	The connection to access to the underlying DS.
	 * @param pcm	The class mapping to add to this mapper.
	 * @param cleanup	Tells if structures and/or data should be removed at
	 * 					initialization time.
	 * @exception PExceptionIO	It is raised when an I/O problem occurs within
	 * 							the underlying DS.
	 * @exception PExceptionProtocol	It is raised when a problem occurs
	 * 									during initialisation of the class
	 * 									mapping.
	 * @exception PExceptionTyping	It is raised when a typing problem occurs
	 * 								during initialisation of the class mapping.
	 */
	public void map(Object conn, PClassMapping pcm, byte cleanup)
		throws PExceptionIO, PExceptionProtocol, PExceptionTyping {
		if (pcm == null)
			throw new PExceptionProtocol(
				"Cannot map null class mapping.");
		PClassMapping ecm
			= (PClassMapping) mappedClasses.get(pcm.getClassName());
		if (ecm == pcm)
			return;
		if (ecm != null)
			throw new PExceptionProtocol(
				"Class already mapped within this mapper.");
		if (!pcm.isConform(mapperName))
			throw new PExceptionProtocol(
				"Class mapping cannot be managed by this mapper.");
		if (typeSpace == null)
			typeSpace = new PTypeSpacePAAH();
		((Loggable) pcm).setLogger(loggerFactory.getLogger("org.objectweb.jorm."
			+ mapperName + (mappingName != null?"." + mappingName:"")
			+ "." + pcm.getClassName()));
		pcm.init(this, null);
		pcm.initMappingStructures(conn, cleanup);
		mappedClasses.put(pcm.getClassName(), pcm);
		if (logger!=null)
			logger.log(BasicLevel.INFO,
				"New mapped class: " + pcm.getClassName());
	}

	/**
	 * It registers a PClassMapping associated to a particular JORM class within
	 * this PMapper. It initializes the storage structures if necessary. It
	 * previously removes them if the "dsrem" flag is set. It also loads the
	 * meta-information related to this class if the "loadmeta" flag is set.
	 * @param conn	The connection to access to the underlying DS.
	 * @param pcm		The class mapping to add to this mapper.
	 * @param cleanup	Tells if structures and/or data should be removed at
	 * 					initialization time.
	 * @param loadmeta	It specifies that meta-information should be loaded for
	 * 					this class to map.
	 * @exception PExceptionIO	It is raised when an I/O problem occurs within
	 * 							the underlying DS.
	 * @exception PExceptionProtocol	It is raised when a problem occurs
	 * 									during initialisation of the class
	 * 									mapping.
	 * @exception PExceptionTyping	It is raised when a typing problem occurs
	 * 								during initialisation of the class mapping.
	 */
	public void map(Object conn, PClassMapping pcm, byte cleanup,
					boolean loadmeta)
		throws PExceptionIO, PExceptionProtocol, PExceptionTyping {
		if (loadmeta && ((metainfoManager == null) || (parser == null))) {
			throw new PExceptionProtocol(
				"Not enough means to load meta-information.");
		}
		map(conn, pcm, cleanup);
		if (!loadmeta)
			return;
		parser.parse(
			pcm.getClassName().replace('.', File.pathSeparatorChar) + ".pd");
	}

	/**
	 * It assigns the name of the mapper that is instanciated. It is mandatory
	 * to assign it at mapper initialization as the load of meta-information
	 * depends on it.
	 * @param mappingname	The name of the maping as defined within the XML
	 * 						JORM definition files.
	 */
	public void setMapperName(String mappername) {
		mapperName = mappername;
	}

	/**
	 * It assigns the name of the mapping that is instanciated through this
	 * mapper. It is mandatory to assign it at mapper initialization as the
	 * load of meta-information depends on it.
	 * @param mappingname	The name of the maping as defined within the XML
	 * 						JORM definition files.
	 */
	public void setMappingName(String mappingname) {
		mappingName = mappingname;
	}

	/**
	 * It assigns a meta-information Manager that may used by a parser in order
	 * to load the meta-information associated to JORM classes mapped within
	 * this PMapper.
	 * @param m	The meta-information Manager to be assigned.
	 */
	public void setMetaInfoManager(Manager m) {
		metainfoManager = m;
	}

	/**
	 * It assigns a Parser that may be used to create the meta-information
	 * associated to JORM classes mapped within this PMapper.
	 * @param p	The Parser to be assigned.
	 */
	public void setParser(Parser p) {
		parser = p;
	}

	/**
	 * It assigns a PTypeSpace that may be used to create PType
	 * associated to JORM classes mapped within this PMapper.
	 * @param pts	The PTypeSpace to be assigned.
	 */
	public void setPTypeSpace(PTypeSpace pts) {
		typeSpace = pts;
	}

	/**
	 * It returns a PTypeSpace that may be used to create PType
	 * associated to JORM classes mapped within this PMapper.
	 */
	public PTypeSpace getPTypeSpace() {
		return typeSpace;
	}

	/**
	 * It unmaps a particular JORM class within this PMapper.
	 * @param classname	The name of the JORM class to be unmapped.
	 */
	public void unmap(String classname) throws PException {
		mappedClasses.remove(classname);
	}

	/**
	 * It assigns a logger to an component that wants to log things.
	 * @param Logger	The logger object.
	 */
	public void setLogger(Logger logger) {
	}

	/**
	 * It assigns a logger factory that allows the creation of new loggers.
	 * @param loggerfactory	The LoggerFactory object to obtain new loggers.
	 */
	public void setLoggerFactory(LoggerFactory loggerfactory) {
		loggerFactory = loggerfactory;
		logger = loggerFactory.getLogger("org.objectweb.jorm." + mapperName
			+ "." + mappingName);
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		 return loggerFactory;
	}
}
