/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.runtime.mapper.rdb;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.sql.Blob;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 * @author	P. Dchamboux
 */
public class RdbBlob extends OutputStream implements Blob {

	public final static int DEFAULT_BUFFER_SIZE = 1024;

	/**
	 * The size of the buffer stored in the buffer list
	 */
	protected int bufferSize = DEFAULT_BUFFER_SIZE;

	/**
	 * This arrayList contains buffers. A buffer is a byte[] which the size
	 * is bufferSize.
	 */
	protected ArrayList bufferList = new ArrayList();

	/**
	 * The current buffer.
	 */
	private byte[] currentBuffer;

	/**
	 * The current position in the last buffer
	 */
	private int currentPos;

	/**
	 * The total length
	 */
	protected long size;

	/**
	 * This constructor calculates the size of the serialized object.
	 */
	public RdbBlob(Serializable serobj) throws IOException {
		this(serobj, DEFAULT_BUFFER_SIZE);
	}

	/**
	 * This constructor calculates the size of the serialized object.
	 */
	public RdbBlob(Serializable serobj, int buffersize) throws IOException {
		bufferSize = buffersize;
		currentBuffer = new byte[bufferSize];
		bufferList.add(currentBuffer);
		currentPos = 0;
		size = 0;
		(new ObjectOutputStream(this)).writeObject(serobj);
	}

	// IMPLEMETATION OF THE Blob INTERFACE //

	public long length() throws SQLException {
		return size;
	}

	public byte[] getBytes(long pos, int length) throws SQLException {
		println("RdbBlob.getBytes(pos=" + pos + ", length=" + length + ")");

		if (pos > size)
			return new byte[0];
		long len = ((pos + length) > size ? size - pos : length);
		println("RdbBlob.getBytes:len=" + len);

		byte[] res = new byte[(int) len];
		//The first buffer index
		long fb = pos / bufferSize;
		println("RdbBlob.getBytes:fb=" + fb);

		//The last buffer index
		long lb = len / bufferSize;
		println("RdbBlob.getBytes:lb=" + lb);

		int current = (int) (bufferSize - pos * bufferSize + fb);
		current = (int) (current > len ? len : current);
		println("RdbBlob.getBytes:current=" + current);

		// Copy the first buffer
		System.arraycopy(bufferList.get((int) fb), (int) (pos - fb * bufferSize),
			res, 0, current);

		// Copy middle buffer
		for (long i = (fb + 1); i < lb; i++) {
			System.arraycopy(bufferList.get((int) i), 0, res, current,
				bufferSize);
			current += bufferSize;
			println("RdbBlob.getBytes:current=" + current);
		}

		// Copy the last buffer
		if (fb != lb)
			System.arraycopy(bufferList.get((int) lb), 0, res, current,
				res.length - current);

		return res;
	}

	public InputStream getBinaryStream() throws SQLException {
		return new RdbInputStream(this);
	}

	public long position(byte[] bytes, long l) throws SQLException {
		if (l > size)
			return -1;
		if (bytes.length > (size - l))
			return -1;

		int curBufId = (int) l / bufferSize;
		byte[] curBuf = (byte[]) bufferList.get(curBufId);
		int curPos = (int) l % bufferSize;
		boolean found = false;
		while (!found
			|| (bytes.length - size + curPos + curBufId * bufferSize) <= 0) {

			int cp = curPos;
			int cb = curBufId;
			int i = 0;
			while (i < bytes.length
				&& (bytes[i] == ((byte[]) bufferList.get(cb))[cp])) {
				i++;
				cp++;
				if (cp == bufferSize) {
					cp = 0;
					cb++;
				}
			}
			if (i == bytes.length)
				return curPos + curBufId * bufferSize;

			// Increment
			if (!found) {
				curPos++;
				if (curPos == bufferSize) {
					curPos = 0;
					curBufId++;
				}
			}
		}
		return -1;
	}

	public long position(Blob blob, long l) throws SQLException {
		throw new SQLException("Unsupported operation");
	}

	// Blob extensions for JDBC3
	// Still needs to be implemented !!!

	public OutputStream setBinaryStream(long pos) throws SQLException {
		throw new SQLException("Unsupported operation");
	}

	public int setBytes(long pos, byte[] bytes) throws SQLException {
		throw new SQLException("Unsupported operation");
	}

	public int setBytes(long pos, byte[] bytes, int offset, int len)
		throws SQLException{
		throw new SQLException("Unsupported operation");
	}

	public void truncate(long len) throws SQLException {
		throw new SQLException("Unsupported operation");
	}

	// OVERRIDING OF THE OutputStream CLASS //

	/**
	 * do nothing
	 */
	public void close() {
		println("RdbBlob.close()");
	}

	/**
	 * do nothing
	 */
	public void flush() {
		println("RdbBlob.flush()");
	}

//TODO: optimize the following method
	/**
	 *
	 */
	public void write(byte[] b) {
		println("RdbBlob.write([])");
		for (int i = 0; i < b.length; i++)
			write(b[i]);
	}

//TODO: optimize the following method
	/**
	 *
	 */
	public void write(byte[] b, int off, int len) {
		println("RdbBlob.write([], " + off + ", " + len + ")");
		for (int i = off; i < len; i++)
			write(b[i]);
	}

	/**
	 *
	 */
	public void write(int b) {
		println("RdbBlob.write(" + b + ")");
		currentBuffer[currentPos] = (byte) b;
		size++;
		currentPos++;
		if (currentPos == bufferSize) {
			println("New buffer");
			currentBuffer = new byte[bufferSize];
			bufferList.add(currentBuffer);
			currentPos = 0;
		}
	}

	private void println(String msg) {
		//System.out.println(msg);
	}
}

class RdbInputStream extends InputStream {

	private RdbBlob blob = null;

	private int bufferId = 0;
	private int cursor = 0;

	RdbInputStream(RdbBlob _blob) {
		blob = _blob;
	}

	public int getSize() {
		return (int) blob.size;
	}

	// OVERRIDING OF THE Inputstream CLASS //

	public int available() throws IOException {
		return (int) blob.size - (bufferId * blob.bufferSize + cursor);
	}

	/**
	 * Does not support by this implementation
	 */
	public synchronized void mark(int readlimit) {
	}

	/**
	 * Does not support by this implementation
	 */
	public synchronized void reset() throws IOException {
		throw new IOException("mark/reset not supported");
	}

	/**
	 * This implementation does not support the mark/reset operations
	 */
	public boolean markSupported() {
		return false;
	}

	public long skip(long n) throws IOException {
		if (n <= 0)
			return 0;
		long oldPos = bufferId * blob.bufferSize + cursor;
		long newPos = oldPos + n;
		newPos = ((newPos >= blob.size)? blob.size : newPos);
		cursor = (int) newPos % blob.bufferSize;
		bufferId = (int) newPos / blob.bufferSize;
		return newPos - oldPos;
	}

	public int read() throws IOException {
		//check if there is a non read byte at least
		if ((bufferId * blob.bufferSize + cursor) == (blob.size))
			return -1;
		byte res = ((byte[]) blob.bufferList.get(bufferId))[cursor];

		// increment the cursor
		if (++cursor == blob.bufferSize) {
			cursor = 0;
			bufferId++;
		}
		return res;
	}

//TODO: implement the following method to optimize the reading
/*	public int read(byte b[], int off, int len) throws IOException {

		if (b == null) {
			throw new NullPointerException();
		}
		else if ((off < 0) || (off > b.length) || (len < 0) ||
			((off + len) > b.length) || ((off + len) < 0)) {
			throw new IndexOutOfBoundsException();
		}
		else if (len == 0) {
			return 0;
		}
	}
*/
}
