/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.type.api;

/**
 * The <b>PType</b> class defines the minimal information required to define
 * a JORM type that belongs to a PTypeSpace. It should be derived for PType
 * representing JORM classes. This derivation should be different between
 * classes and generic classes.
 * @author	P. Dchamboux
 */
public class PType {
	public final static short TYPECODE_BOOLEAN = 0;
	public final static short TYPECODE_CHAR = 1;
	public final static short TYPECODE_BYTE = 2;
	public final static short TYPECODE_SHORT = 3;
	public final static short TYPECODE_INT = 4;
	public final static short TYPECODE_LONG = 5;
	public final static short TYPECODE_FLOAT = 6;
	public final static short TYPECODE_DOUBLE = 7;
	public final static short TYPECODE_STRING = 8;
	public final static short TYPECODE_DATE = 9;
/*
	public final static short TYPECODE_TIME = 10;
	public final static short TYPECODE_TIMESTAMP = 11;
*/
	public final static short TYPECODE_SERIALIZED = 12;
	public final static short TYPECODE_REFERENCE = 13;

    public final static short TYPECODE_INT_OBJ = 24;

	// Each PType has a type code
	private short typeCode;
	// Each PType has a JORM name
	private String jormName;
	// Each PType has an associated java type
	private String javaName;
	// Each PType representing a JORM class belongs to a PTypeSpace
	private PTypeSpace typeSpace;

	/**
	 * It constructs the types that represent basic JORM types defined as
	 * constants within the PTypeSpace interface.
	 * @param typecode	The type code as defined within PType.
	 * @param jormname	The JORM name of this basic type.
	 * @param javaname	The Java associated with this basic type.
	 */
	protected PType(short typecode, String jormname, String javaname) {
		typeCode = typecode;
		jormName = jormname;
		javaName = javaname;
		typeSpace = null;
	}

	/**
	 * It constructs the types that represent JORM classes, be it generic or
	 * not.
	 * @param classname	The name of the class for which a PType must be
	 * 					constructed.
	 * @param typespace	The type space to which the constructed PType belongs.
	 */
	protected PType(String classname, PTypeSpace typespace) {
		typeCode = TYPECODE_REFERENCE;
		jormName = classname;
		javaName = "java.lang.Object";
		typeSpace = typespace;
	}

	/**
	 * It compares if the current PType conforms to the given PType. In case
	 * of PType representing JORM classes, the comparison is delegated to the
	 * associated PTypeSpace.
	 * @param pt	The PType to be compared with.
	 * @return	It returns true if this PType conforms to the one passed as
	 * parameter.
	 */
	public boolean isa(PType pt) {
		switch (this.typeCode) {
		case TYPECODE_BOOLEAN:
			return pt.typeCode == TYPECODE_BOOLEAN;
		case TYPECODE_CHAR:
			return pt.typeCode == TYPECODE_CHAR;
		case TYPECODE_BYTE:
			switch (pt.typeCode) {
			case TYPECODE_BYTE:
			case TYPECODE_SHORT:
			case TYPECODE_INT:
			case TYPECODE_LONG:
			case TYPECODE_FLOAT:
			case TYPECODE_DOUBLE:
				return true;
			default :
				return false;
			}
		case TYPECODE_SHORT:
			switch (pt.typeCode) {
			case TYPECODE_SHORT:
			case TYPECODE_INT:
			case TYPECODE_LONG:
			case TYPECODE_FLOAT:
			case TYPECODE_DOUBLE:
				return true;
			default :
				return false;
			}
		case TYPECODE_INT:
			switch (pt.typeCode) {
			case TYPECODE_INT:
			case TYPECODE_LONG:
			case TYPECODE_DOUBLE:
				return true;
			default :
				return false;
			}
		case TYPECODE_LONG:
			switch (pt.typeCode) {
			case TYPECODE_LONG:
				return true;
			default :
				return false;
			}
		case TYPECODE_FLOAT:
			switch (pt.typeCode) {
			case TYPECODE_FLOAT:
			case TYPECODE_DOUBLE:
				return true;
			default :
				return false;
			}
		case TYPECODE_DOUBLE:
			return pt.typeCode == TYPECODE_DOUBLE;
		case TYPECODE_STRING:
			return (pt.typeCode == TYPECODE_STRING)
				|| (pt.typeCode == TYPECODE_SERIALIZED);
		case TYPECODE_DATE:
			return (pt.typeCode == TYPECODE_DATE)
				|| (pt.typeCode == TYPECODE_SERIALIZED);
        case TYPECODE_INT_OBJ:
			return (pt.typeCode == TYPECODE_INT_OBJ)
				|| (pt.typeCode == TYPECODE_SERIALIZED);
/*		case TYPECODE_TIME:
			return (pt.typeCode == TYPECODE_TIME)
					|| (pt.typeCode == TYPECODE_SERIALIZED);
		case TYPECODE_TIMESTAMP:
			return (pt.typeCode == TYPECODE_TIMESTAMP)
					|| (pt.typeCode == TYPECODE_SERIALIZED);
*/
		case TYPECODE_SERIALIZED:
			return pt.typeCode == TYPECODE_SERIALIZED;
		case TYPECODE_REFERENCE:
			return (pt.typeCode == TYPECODE_REFERENCE)
				&& typeSpace.isa(this, pt);
		}
		return false;
	}

	public boolean equals(Object o) {
		if (!(o instanceof PType)) {
			return false;
		}
		PType p = (PType) o;
		return (p.getTypeCode() == typeCode)
			&& (typeCode == TYPECODE_REFERENCE
			? jormName.equals(p.getJormName())
			: true);
	}

	/**
	 * It retrieves the PType directly inherited by a PType representing a
	 * JORM class.
	 * @return	The array of directly inherited PType.
	 */
	public PType[] getInheritedPType() {
		return null;
	}

	/**
	 * It retrieves the JORM name associated to this PType.
	 * @return	It returns the JORM name.
	 */
	public String getJormName() {
		return jormName;
	}

	/**
	 * It retrieves the Java name associated to this PType.
	 * @return	It returns the Java name.
	 */
	public String getJavaName() {
		return javaName;
	}

	/**
	 * It retrieves the nested PType in case of a generic class PType.
	 * @return	The nested PType.
	 */
	public PType getNestedPType() {
		return null;
	}

	/**
	 * It retrieves the type space associated to this PType.
	 * @return	The associated type space.
	 */
	public PTypeSpace getPTypeSpace() {
		return typeSpace;
	}

	/**
	 * It retrieves the code associated to this PType.
	 * @return	It returns the Java name.
	 */
	public int getTypeCode() {
		return typeCode;
	}
}
