/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.type.api;

/**
 * The <b>PTypeSpace</b> interface defines a space of JORM type along with
 * the inheritance relationships that may exist between them.
 * @author	P. Dchamboux
 */
public interface PTypeSpace {
	final static PType BOOLEAN = new PType(PType.TYPECODE_BOOLEAN,
		"boolean",
		"boolean");
	final static PType CHAR = new PType(PType.TYPECODE_CHAR,
		"char",
		"char");
	final static PType BYTE = new PType(PType.TYPECODE_BYTE,
		"byte",
		"byte");
	final static PType SHORT = new PType(PType.TYPECODE_SHORT,
		"short",
		"short");
	final static PType INT = new PType(PType.TYPECODE_INT,
		"int",
		"int");
	final static PType LONG = new PType(PType.TYPECODE_LONG,
		"long",
		"long");
	final static PType FLOAT = new PType(PType.TYPECODE_FLOAT,
		"float",
		"float");
	final static PType DOUBLE = new PType(PType.TYPECODE_DOUBLE,
		"double",
		"double");
	final static PType STRING = new PType(PType.TYPECODE_STRING,
		"string",
		"java.lang.String");
	final static PType DATE = new PType(PType.TYPECODE_DATE,
		"date",
		"java.util.Date");
	final static PType SERIALIZED = new PType(PType.TYPECODE_SERIALIZED,
		"serialized",
		"java.io.Serializable");
	final static PType REFTOP = new PType(PType.TYPECODE_REFERENCE,
		"object",
		"java.lang.Object");
    final static PType INT_OBJ = new PType(PType.TYPECODE_INT_OBJ,
		"Integer",
		"java.lang.Integer");

	/**
	 *
	 */
	final static PType[] NULLSUPERTYPES = new PType[0];

	/**
	 *
	 */
	final static String[][] NULLSUPERCLASSNAMES = new String[0][];

	/**
	 * It creates a new PType for a JORM class that has no inherited class.
	 * @param classname		The name of the JORM class.
	 * @return	The created PType or the existing one if it has the same
	 * 			definition (same name and same super types).
	 * @exception PExceptionTyping	Raised when a PType with this class name
	 * 								exists with a different array of super
	 * 								types.
	 */
	PType createPType(String classname) throws PExceptionTyping;

	/**
	 * It creates a new PType for a JORM class. It is defined by its name and
	 * its inheritance hierarchy. For example, if "Engineer" inherits from
	 * "Employee" that inherits from "Person", then the following stands:
	 * <bl>
	 * <li>classname = "schm1.Engineer"</li>
	 * <li>superclassnames = {{"schm1.Engineer", "schm2.Employee"},
	 * {"schm2.Employee", "schm3.Person"}}</li>
	 * </bl>
	 * @param classname			The name of the JORM class to be defined.
	 * @param superclassnames	The hierarchy of of classes inherited by the
	 * 							previous class (defined by classname) if any.
	 * @return	The created PType or the existing one if it has the same
	 * 			definition (same name and same super types).
	 * @exception PExceptionTyping	Raised when a PType with this class name
	 * 								exists with a different array of super
	 * 								types.
	 */
	PType createPType(String classname, String[][] superclassnames)
		throws PExceptionTyping;

	/**
	 * It creates a new PType for a JORM generic class instance. Such an
	 * instance may have a nested structure like
	 * "list<array<set<int>>>", which means that:
	 * <bl>
	 * <li>gcnames = {"list", "array", "set"}</li>
	 * <li>ptypename = INT</li>
	 * </bl>
	 * @param gcnames	The array of names of the nested generic classes.
	 * @param ptypename	The PType of the type of the elements contained
	 * 					within this generic class, which is one of the
	 * 					following: BOOLEAN, CHAR, BYTE, SHORT, INT, LONG,
	 * 					FLOAT, DOUBLE, STRING, DATE, or SERIALIZED (see these
	 * 					constants in this interface definition).
	 * @return	The created PType or the existing one if it has the same
	 * 			definition (same name and same element type).
	 * @exception PExceptionTyping	Raised when a PType with this class name
	 * 								exists with a different inner PType
	 */
	PType createPType(String[] gcnames, PType ptype) throws PExceptionTyping;

	/**
	 * It creates a new PType for a JORM generic class instance. Such an
	 * instance may have a nested structure like
	 * "list<array<set<schm1.Engineer>>>", which means that:
	 * <bl>
	 * <li>gcnames = {"list", "array", "set"}</li>
	 * <li>typename = "schm1.Engineer"</li>
	 * <li>superclassnames = {{"schm1.Engineer", "schm2.Employee"},
	 * {"schm2.Employee", "schm3.Person"}}</li>
	 * </bl>
	 * @param gcnames			The array of names of the nested generic
	 * 							classes.
	 * @param typename			The name of the type of the elements contained
	 * 							within this generic class.
	 * @param superclassnames	The hierarchy of of classes inherited by the
	 * 							previous class (defined by typename) if any.
	 * @return	The created PType or the existing one if it has the same
	 * 			definition (same name and same element type).
	 * @exception PExceptionTyping	Raised when a PType with this class name
	 * 								exists with a different inner PType
	 */
	PType createPType(String[] gcnames, String typename,
					  String[][] superclassnames) throws PExceptionTyping;

	/**
	 * It tries to retrieve a PType representing a JORM class with the given
	 * class name.
	 * @param classname		The name of a JORM class the PType is looking for.
	 * @return	The relevant PType if one has been found or null if none found.
	 */
	PType getPType(String classname);

	/**
	 * It compares to PType representing JORM classes. It verifies if the first
	 * one conforms to the second one.
	 * @param pt1	The first JORM class PType.
	 * @param pt2	The second JORM class PType.
	 * @return	It returns true if pt1 conforms to pt2.
	 */
	boolean isa(PType pt1, PType pt2);
}
