/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.type.lib;

import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;

import java.util.HashMap;

/*
 * This implementation of the PTypeSpace considers that this is the PType which
 * contains its ancestor. Then class contains only a simple list of PType
 * defined in this PType space. The PAAH letters mean PType As Ancestor Holder.
 * This implementation uses the PTypePAAH class to create PType of jorm class.
 *
 * @author	Pascal Dchamboux, Sebastien Chassande-Barrioz
 */

public class PTypeSpacePAAH implements PTypeSpace {

	private static final PType[] SUPERNULL = new PType[0];

	/**
	 * This hash map contains the list of PType defined in this PType space.
	 * keys = the jorm names of PType
	 * values = The PType objects
	 */
	private HashMap types = null;

	public PTypeSpacePAAH() {
		types = new HashMap();
		// Puts basic JORM types into the types map (for GC support purpose)
		types.put(PTypeSpace.BOOLEAN.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.BOOLEAN));
		types.put(PTypeSpace.BYTE.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.BYTE));
		types.put(PTypeSpace.CHAR.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.CHAR));
		types.put(PTypeSpace.SHORT.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.SHORT));
		types.put(PTypeSpace.INT.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.INT));
		types.put(PTypeSpace.LONG.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.LONG));
		types.put(PTypeSpace.FLOAT.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.FLOAT));
		types.put(PTypeSpace.DOUBLE.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.DOUBLE));
		types.put(PTypeSpace.STRING.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.STRING));
		types.put(PTypeSpace.DATE.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.DATE));
		types.put(PTypeSpace.SERIALIZED.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.SERIALIZED));
		types.put(PTypeSpace.REFTOP.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.REFTOP));
        types.put(PTypeSpace.INT_OBJ.getJormName(),
			new PTypeHashMapEntry(PTypeSpace.INT_OBJ));
	}

	/**
	 * It creates a new PType for a JORM class that has no inherited class.
	 * @param classname		The name of the JORM class.
	 * @return	The created PType or the existing one if it has the same
	 * 			definition (same name and same super types).
	 * @exception PExceptionTyping	Raised when a PType with this class name
	 * 								exists with a different array of super
	 * 								types.
	 */
	public synchronized PType createPType(String classname)
		throws PExceptionTyping {
		if (isConflict(classname, NULLSUPERCLASSNAMES))
			throw new PExceptionTyping("Cannot create type <" + classname
				+ ">: it conflicts with existing types.");
		return definePType(classname, NULLSUPERCLASSNAMES).entryPType;
	}

	/**
	 * It creates a new PType for a JORM class. It is defined by its name and
	 * its inheritance hierarchy. For example, if "Engineer" inherits from
	 * "Employee" that inherits from "Person", then the following stands:
	 * <bl>
	 * <li>classname = "schm1.Engineer"</li>
	 * <li>superclassnames = {{"schm1.Engineer", "schm2.Employee"},
	 * {"schm2.Employee", "schm3.Person"}}</li>
	 * </bl>
	 * @param classname			The name of the JORM class to be defined.
	 * @param superclassnames	The hierarchy of of classes inherited by the
	 * 							previous class (defined by classname) if any.
	 * @return	The created PType or the existing one if it has the same
	 * 			definition (same name and same super types).
	 * @exception PExceptionTyping	Raised when a PType with this class name
	 * 								exists with a different array of super
	 * 								types.
	 */
	public synchronized PType createPType(String classname,
										  String[][] superclassnames)
		throws PExceptionTyping {
		if (isConflict(classname, superclassnames))
			throw new PExceptionTyping("Cannot create type <" + classname
				+ ">: it conflicts with existing types.");
		return definePType(classname, superclassnames).entryPType;
	}

	/**
	 * It creates a new PType for a JORM generic class instance. Such an
	 * instance may have a nested structure like
	 * "list<array<set<int>>>", which means that:
	 * <bl>
	 * <li>gcnames = {"list", "array", "set"}</li>
	 * <li>ptypename = INT</li>
	 * </bl>
	 * @param gcnames	The array of names of the nested generic classes.
	 * @param ptypename	The PType of the type of the elements contained
	 * 					within this generic class, which is one of the
	 * 					following: BOOLEAN, CHAR, BYTE, SHORT, INT, LONG,
	 * 					FLOAT, DOUBLE, STRING, DATE, or SERIALIZED (see these
	 * 					constants in this interface definition).
	 * @return	The created PType or the existing one if it has the same
	 * 			definition (same name and same element type).
	 * @exception PExceptionTyping	Raised when a PType with this class name
	 * 								exists with a different inner PType
	 */
	public PType createPType(String[] gcnames, PType ptype)
		throws PExceptionTyping {
		return ((PTypeHashMapEntry) types.get(ptype.getJormName()))
			.addGcPType(gcnames, (short) (gcnames.length - 1));
	}

	/**
	 * It creates a new PType for a JORM generic class instance. Such an
	 * instance may have a nested structure like
	 * "list<array<set<schm1.Engineer>>>", which means that:
	 * <bl>
	 * <li>gcnames = {"list", "array", "set"}</li>
	 * <li>typename = "schm1.Engineer"</li>
	 * <li>superclassnames = {{"schm1.Engineer", "schm2.Employee"},
	 * {"schm2.Employee", "schm3.Person"}}</li>
	 * </bl>
	 * @param gcnames			The array of names of the nested generic
	 * 							classes.
	 * @param typename			The name of the type of the elements contained
	 * 							within this generic class.
	 * @param superclassnames	The hierarchy of of classes inherited by the
	 * 							previous class (defined by typename) if any.
	 * @return	The created PType or the existing one if it has the same
	 * 			definition (same name and same element type).
	 * @exception PExceptionTyping	Raised when a PType with this class name
	 * 								exists with a different inner PType
	 */
	public synchronized PType createPType(String[] gcnames, String typename,
										  String[][] superclassnames)
		throws PExceptionTyping {
		if (isConflict(typename, superclassnames))
			throw new PExceptionTyping("Cannot create type <" + typename
				+ ">: it conflicts with existing types.");
		return defineGcPType(gcnames, typename, superclassnames);
	}

	/**
	 * It tries to retrieve a PType representing a JORM class with the given
	 * class name.
	 * @param classname		The name of a JORM class the PType is looking for.
	 * @return	The relevant PType if one has been found or null if none found.
	 */
	public PType getPType(String classname) {
		synchronized (types) {
			return (PType) types.get(classname);
		}
	}

	/**
	 * It compares to PType representing JORM classes. It verifies if the first
	 * one conforms to the second one.
	 * @param pt1	The first JORM class PType.
	 * @param pt2	The second JORM class PType.
	 * @return	It returns true if pt1 conforms to pt2.
	 */
	public boolean isa(PType pt1, PType pt2) {
		return pt1 != null && pt2 == null && pt1.isa(pt2);
	}

	/**
	 *
	 */
	private boolean isConflict(String classname, String[][] superclasses) {
		PTypeHashMapEntry entry = (PTypeHashMapEntry) types.get(classname);
		if (entry == null)
			return false;
		short i = 0;
		while ((i < superclasses.length) &&
			!(superclasses[i][0].equals(classname)))
			i++;
		// This a ROOT PType (no inheritance)
		if (i == superclasses.length) {
			if (((PTypePAAH) entry.entryPType).superTypes.length == 0)
				return false;
			return true;
		}
		if (((PTypePAAH) entry.entryPType).superTypes.length
			!= superclasses[i].length)
			return true;
		PType subpt;
		short j;
		for (short k = 0; k < ((PTypePAAH) entry.entryPType).superTypes.length;
			 k++) {
			j = 1;
			while (j < superclasses[i].length) {
				if (((PTypePAAH) entry.entryPType).superTypes[k].getJormName()
					.equals(superclasses[i][j])) {
					if (isConflict(superclasses[i][j], superclasses))
						return true;
					break;
				}
				j++;
			}
			if (j == superclasses[i].length)
				return true;
		}
		return false;
	}

	/**
	 *
	 */
	private PTypeHashMapEntry definePType(String classname,
										  String[][] superclasses) {
		PTypeHashMapEntry entry = (PTypeHashMapEntry) types.get(classname);
		if (entry != null)
			return entry;
		short i = 0;
		while ((i < superclasses.length) &&
			!(superclasses[i++][0].equals(classname)))
			;
		// This a ROOT PType (no inheritance)
		if (i == superclasses.length) {
			return new PTypeHashMapEntry(new PTypePAAH(classname, this,
				NULLSUPERTYPES));
		}
		PType[] supertypes = new PType[superclasses[i].length - 1];
		for (short j = 1; j < superclasses[i].length; j++)
			supertypes[j - 1]
				= definePType(superclasses[i][j], superclasses).entryPType;
		return
			new PTypeHashMapEntry(new PTypePAAH(classname, this, supertypes));
	}

	/**
	 *
	 */
	private PType defineGcPType(String[] gcnames, String typename,
								String[][] superclasses) {
		PTypeHashMapEntry typeentry = (PTypeHashMapEntry) types.get(typename);
		if (typeentry == null) {
			typeentry = definePType(typename, superclasses);
		}
		return typeentry.addGcPType(gcnames, (short) (gcnames.length - 1));
	}
}

/**
 * This class represents a PType and defines a tree of PType. A instance is
 * node of PTypeHashMapEntry tree. A Path in this tree defines a PType. Be
 * carefull the path is reversed: for example the PType array(set(String)) will
 * be stored in the following tree:<br/>
 * String                     <br/>
 *   \--set                   <br/>
 *       \--array             <br/>
 *
 * If after you add the PType array(String) the tree will be completed as here:
 * String                     <br/>
 *   |--set                   <br/>
 *   |   \--array             <br/>
 *   \--array                 <br/>
 *
 * To optimize the adding operation, a PTypeHashMapEntry manages an array of
 * several childrens. This array is increased by increment.
 *
 * @author	Pascal Dchamboux, Sebastien Chassande-Barrioz
 */
class PTypeHashMapEntry {
	/**
	 * The size of the increment.
	 */
	public final static int INCREMENT = 5;

	/**
	 * The current PType defined by the reverse path from the root
	 */
	PType entryPType;

	/**
	 * The children of the current PTypeHashMapEntry
	 */
	private PTypeHashMapEntry[] gcInstances = null;

	public PTypeHashMapEntry(PType ptype) {
		entryPType = ptype;
	}

	/**
	 * It Adds a PType in the tree.
	 * @param gcnames is the list of the gen class names which must be inserted
	 * @param curgc is an index into the gcnames array. It is used to go all
	 * over elements of the array. The gcnames placed before the index, have
	 * been already inserted in the tree.
	 * @return The PType inserted or found in the tree.
	 */
	public PType addGcPType(String[] gcnames, short curgc) {
		if (curgc < 0) {
			// The searched element is the current.
			return entryPType;
		}

		// No GC already associated to this type
		if (gcInstances == null) {
			gcInstances = new PTypeHashMapEntry[INCREMENT];
			for (short i = 1; i < INCREMENT; i++)
				gcInstances[i] = null;
			gcInstances[0]
				= new PTypeHashMapEntry(new PTypeGenClass(gcnames[curgc],
					entryPType, entryPType.getPTypeSpace()));
			return gcInstances[0].addGcPType(gcnames, (short) (curgc - 1));
		}

		// Looks if the currrent GC already exists
		int i = 0;
		while ((i < gcInstances.length)
			&& gcInstances[i] != null
			&& !gcInstances[i++].entryPType.getJormName().equals(gcnames[curgc]))
			;

		if (i == gcInstances.length) {
			// It does not exist and no free slot exists: increases size
			if (i == gcInstances.length) {
				PTypeHashMapEntry[] newgcinstances
					= new PTypeHashMapEntry[gcInstances.length + INCREMENT];
				short j = 0;
				while (j < gcInstances.length)
					newgcinstances[j++] = gcInstances[j];
				j++;
				while (j < newgcinstances.length)
					newgcinstances[j++] = null;
			}
			gcInstances[i] = new PTypeHashMapEntry(
				new PTypeGenClass(gcnames[curgc],entryPType,
					entryPType.getPTypeSpace()));
		}
		else if (gcInstances[i]==null) {
			// It does not exsit but a slot is available
			gcInstances[i] = new PTypeHashMapEntry(new PTypeGenClass(gcnames[curgc],
				entryPType, entryPType.getPTypeSpace()));
		}

		// In all cases recall with the next gcnames
		return gcInstances[i].addGcPType(gcnames, (short) (curgc - 1));
	}
}
