/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.util.lib;

import org.apache.tools.ant.AntClassLoader;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.taskdefs.MatchingTask;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.Reference;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;

/**
 *
 */
public class AntJorm extends MatchingTask {

	/**
	 * the jorm configuration file
	 */
	private String confFile;

	/**
	 * the output directory for the generation of the java files and class files
	 */
	private String output;

	private String mappingName;

	private String mapperName;

	private String bindingInheritance;

	private String binderInheritance;

	private String mappingfactory = null;

	private String parser = null;

	private String verifier = null;

	private boolean verbose = false;

	private boolean keepsrc = true;

	private String logPropFile = null;

	private File pdDir = null;

	private Path classpath = null;

    private String jormclasspath = null;

	private AntClassLoader loader = null;

	private boolean localpropertiesfile = false;

	private boolean nojavac = false;

    private boolean bindingAbstract = false;

	private ArrayList extensions = new ArrayList();

	/**
	 * execute method of the Ant task
	 */
	public void execute() throws BuildException {
		loader = new AntClassLoader(AntJorm.class.getClassLoader(), false);

		// Full the class loader
		String[] pathelems = classpath.list();
		int j = 0;
		for (; j < pathelems.length; j++)
			loader.addPathElement(pathelems[j]);

		Object exec = null;
		try {
			Class comp =
				loader.loadClass("org.objectweb.jorm.util.lib.AntJormExec");
			Constructor[] cons = comp.getDeclaredConstructors();
			Constructor co = cons[0];
			Class[] c = co.getParameterTypes();
			exec = co.newInstance(new Object[]{
				getProject(),
				mappingName,
				mapperName,
				mappingfactory,
				parser,
				verifier,
				new Boolean(verbose),
				new Boolean(keepsrc),
                new Boolean(bindingAbstract),
				confFile,
				logPropFile,
				output,
				bindingInheritance,
				binderInheritance,
				pdDir,
				super.getDirectoryScanner(pdDir),
				classpath,
				(localpropertiesfile ? Boolean.TRUE : Boolean.FALSE),
				this,
				(nojavac ? Boolean.TRUE : Boolean.FALSE),
				extensionsParams(extensions),
				extensionNames(extensions),
                jormclasspath
			});
		}
		catch (InvocationTargetException e) {
			if (e.getTargetException() instanceof BuildException)
				throw (BuildException) e.getTargetException();
		}
		catch (Exception e) {
			throw new BuildException(e);
		}
	}

	public void setPdDir(File d) {
		pdDir = d;
	}

	public void setLocalPropertiesFile(boolean l) {
		localpropertiesfile = l;
	}

	public void setNoJavac(boolean l) {
		nojavac = l;
	}

	/**
	 * Set the classpath to be used for this compilation.
	 */
	public void setClasspath(Path cp) {
		if (classpath == null)
			classpath = cp;
		else
			classpath.append(cp);
	}

	public void setJormclasspath(String cp) {
			jormclasspath = cp;
	}

	/**
	 * Gets the classpath to be used for this compilation.
	 */
	public Path getClasspath() {
		return classpath;
	}

	/**
	 * Maybe creates a nested classpath element.
	 */
	public Path createClasspath() {
		if (classpath == null)
			classpath = new Path(project);
		return classpath.createPath();
	}

	/**
	 * Adds a reference to a CLASSPATH defined elsewhere.
	 */
	public void setClasspathRef(Reference r) {
		createClasspath().setRefid(r);
	}

	/**
	 * allows to set the output directory for the current task
	 * @param s the string representation of the output directory
	 */
	public void setOutput(String s) {
		output = s;
	}

	/**
	 * allows to set the configuration file for Jorm.
	 * @param s the string representation of the configuration file.
	 */
	public void setConf(String s) {
		confFile = s;
	}

	public void setMappingname(String s) {
		mappingName = s;
	}

	public void setMappername(String s) {
		mapperName = s;
	}

	public void setBindinginheritance(String s) {
		bindingInheritance = s;
	}

	public void setBinderinheritance(String s) {
		binderInheritance = s;
	}

	public void setMappingfactory(String mappingfactory) {
		this.mappingfactory = mappingfactory;
	}

	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}

	public void setKeepsrc(boolean keepsrc) {
		this.keepsrc = keepsrc;
	}

	public void setParser(String parser) {
		this.parser = parser;
	}

	public void setVerifier(String verifier) {
		this.verifier = verifier;
	}

    public void setBindingAbstract(boolean bindingAbstract) {
        this.bindingAbstract = bindingAbstract;
    }

	public void setLogPropFile(String logPropFile) {
		this.logPropFile = logPropFile;
	}

	public DirectoryScanner getDirectoryScanner(File p) {
		return super.getDirectoryScanner(p);
	}

	public Extension createExtension() {
		Extension e = new Extension();
		extensions.add(e);
		return e;
	}

	public class Extension extends Task {
		public String extensionName = null;
		public ArrayList params = null;
		public void setName(String n) {
			extensionName = n;
		}
		public Param createParam() {
			Param res = new Param();
			if (params==null)
				params = new ArrayList();
			params.add(res);
			return res;
		}
		public String[][] params2StringArray() {
			String[][] res = new String[params.size()][];
			Iterator it = params.iterator();
			int pos = 0;
			while(it.hasNext()) {
				Param p = (Param) it.next();
				res[pos] =
					p.toStringArray(extensionName + '.' + p.paramName);
				pos ++;
			}
			return res;
		}
		public String getExtensionName() {
			return extensionName;
		}

	}

	public class Param extends Task {
		public String paramName = null;
		public ArrayList values = null;
		public void setName(String n) {
			paramName = n;
		}
		public Value createValue() {
			if (values==null)
				values = new ArrayList();
			Value v = new Value();
			values.add(v);
			return v;
		}

		public String[] toStringArray(String firstelem) {
			String[] res = new String[values.size() + 1];
			Iterator it = values.iterator();
			int pos = 1;
			res[0] = firstelem;
			while(it.hasNext()) {
				res[pos] = ((Value) it.next()).value;
			}
			return res;
		}
	}

	public class Value extends Task {
		public String value = null;
		public void setValue(String v) {
			value = v;
		}
	}

	private String[][][] extensionsParams(ArrayList al) {
		String[][][] res = new String[al.size()][][];
		for (int i=0; i<al.size(); i++) {
			res[i] = ((Extension) al.get(i)).params2StringArray();
		}
		return res;
	}

	private String[] extensionNames(ArrayList al) {
		String[] res = new String[al.size()];
		for (int i=0; i<al.size(); i++) {
			res[i] = ((Extension) al.get(i)).getExtensionName();
		}
		return res;
	}
}


