/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: R. Basset, S. Chassande-Barrioz, P. Dechamboux, N. De Palma,
 *          A. Lefebvre, X. Spengler.
 *
 */

package org.objectweb.jorm.util.lib;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.taskdefs.Javac;
import org.apache.tools.ant.taskdefs.UpToDate;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.types.Path;
import org.objectweb.jorm.comp.cmdline.api.CmdLineParser;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.compiler.lib.Compiler;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.io.api.PathExplorer;
import org.objectweb.util.io.lib.DirJavaExplorer;

import java.io.File;
import java.io.IOException;

/**
 * This class executes the generation and the compilation of a set of .pd
 * All parameter must be specified in the constructor because the actions are
 * done by the constructor.
 *
 * Sebastien Chassande-Barrioz
 */
public class AntJormExec {

	/**
	 * @param project is the ant project
	 * @param mappingName is the
	 */
	public AntJormExec(
		Project project,
		String mappingName,
		String mapperName,
		String mappingfactory,
		String parser,
		String verifier,
		Boolean verbose,
		Boolean keepsrc,
        Boolean bindingAbstract,
		String confFile,
		String logPropFile,
		String output,
		String bindingInheritance,
		String binderInheritance,
		File pdDir,
		DirectoryScanner ds,
		Path classpath,
		Boolean localpropertiesfile,
		Task task,
		Boolean nojavac,
		String[][][] extensionsParams,
		String[] extensionNames,
		String jormclasspath
		) throws BuildException {

		task.log("output=" + output, project.MSG_DEBUG);
		task.log("pdDir=" + pdDir, project.MSG_DEBUG);

		String[] pdFiles = ds.getIncludedFiles();
		boolean isUpToDate = false;
		Compiler compiler = null;
		for (int i = 0; i < pdFiles.length; i++) {
			File pdFile = new File(pdDir, pdFiles[i]);
			task.log("pdfile=" + pdFile, project.MSG_DEBUG);

			String packageName = getPackageName(pdFiles[i]);
			task.log("\tpackageName=" + packageName, project.MSG_DEBUG);

			String className = getClassName(pdFiles[i]);
			task.log("\tclassName=" + className, project.MSG_DEBUG);

			task.log("\tlog Properties file=" + logPropFile, project.MSG_DEBUG);
			String f = null;
			if (localpropertiesfile.booleanValue())
				f = pdFile.getParent() + "/" + confFile;
			else
				f = confFile;
			task.log("\tlocal properties file=" + f, project.MSG_DEBUG);
			try {
				if (logPropFile != null)
					compiler = new Compiler(f, logPropFile);
				else
					compiler = new Compiler(f);
			}
			catch (IOException e) {
				throw new BuildException(
					"Impossible to create a new Compiler: ", e);
			}

			CompilerParameter cp = compiler.getCompilerParameter();
			task.log("nb of extensions: " + extensionsParams.length,
				project.MSG_DEBUG);
			for (int ext = 0; ext < extensionsParams.length; ext++) {
				//Instanciate the extension module
				CmdLineParser clp = null;
				try {
					task.log("extension[" + ext + "] name: " + extensionNames[ext],
						project.MSG_DEBUG);
					clp = (CmdLineParser)
						compiler.getClassLoader().load(extensionNames[ext]);
				}
				catch (Exception e) {
					throw new BuildException("Impossible to get an instance of "
						+ extensionNames[ext], e);
				}
				// Add the extension module
				cp.addCPExtension(extensionNames[ext], clp);
				task.log("extension[" + ext + "]: nb of parameter: "
					+ extensionsParams[ext].length, project.MSG_DEBUG);
				// Parse the parameters with the extension module
				for (int param = 0; param < extensionsParams[ext].length; param++) {
					task.log("extension[" + ext + "]: param[" + param + "]: "
						+ extensionsParams[ext][param], project.MSG_DEBUG);
					((Loggable) clp).setLogger(CompilerLogger.getLoggerFactory().getLogger(
						"org.objectweb.jorm.compiler"));
					clp.parseArgument(extensionsParams[ext][param], 0);
				}
			}
			PathExplorer pathE = new DirJavaExplorer();
			if (jormclasspath != null) {
				pathE.addPath(jormclasspath);
			}
			cp.setClasspath(pathE);
			compiler.setOutputDirectory(output);
			if (pdDir == null)
				throw new BuildException("The pdDir is needed");
			if (mappingName != null)
				cp.setMappingName(mappingName);
			if (mapperName != null)
				cp.setMapperName(mapperName);
			if (mappingfactory != null)
				cp.setMappingFactory(mappingfactory);
			if (parser != null)
				cp.setParserName(parser);
			if (verifier != null)
				cp.setVerifierName(verifier);
			if (parser != null)
				cp.setParserName(parser);
			if (keepsrc != null)
				cp.setKeepSrc(keepsrc.booleanValue());
            if (bindingAbstract != null)
                cp.setBindingAbstract(bindingAbstract.booleanValue());
			if (verbose != null)
				cp.setVerbose(verbose.booleanValue());
			if (bindingInheritance != null)
				cp.setPBindingInheritance(bindingInheritance);
			if (binderInheritance != null)
				cp.setPClassMappingInheritance(binderInheritance);
			if (nojavac != null)
				cp.setNoCompile(nojavac.booleanValue());
			UpToDate u = new UpToDate();
			u.setProject(project);

			//Add the pd file
			FileSet fs = new FileSet();
			fs.setProject(project);
			fs.setDir(pdDir);
			fs.setIncludes(pdFiles[i]);
			u.addSrcfiles(fs);
			// The target is the generated accessor file
			String bindingFileName = output + "/" + packageName + "/"
				+ cp.getMapperName() + "/" + className + "Binding.java";
			task.log("\tbindingFileName=" + bindingFileName, project.MSG_DEBUG);
			u.setTargetFile(new File(bindingFileName));
			if (!u.eval()) {
				task.log("upToDate", project.MSG_DEBUG);
				compiler.addInputFileName(pdFile.getAbsolutePath());
				isUpToDate = true;
			}
			else
				task.log("no upToDate", project.MSG_DEBUG);

			if (isUpToDate) {
				compiler.execute();
			}
		}

		// Compil all new files at the same time.
		if (isUpToDate && !compiler.getCompilerParameter().isNoCompile()) {
			task.log("compil", project.MSG_DEBUG);
			// compiles the generated java files
			Javac javac = new Javac();
			Path path = new Path(project, output);
			javac.setProject(project);
			javac.setSrcdir(path);
			javac.setDestdir(new File(output));
			javac.setDebug(true);
			javac.setIncludes("**/*.java");
			javac.setClasspath(classpath);
			javac.execute();
		}
	}

	private String getPackageName(String f) {
		return f.substring(0, f.lastIndexOf(File.separatorChar));
	}

	private String getClassName(String f) {
		return f.substring(f.lastIndexOf(File.separatorChar) + 1, f.lastIndexOf('.'));
	}

}
